# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import oser
from htf.communication.hart.protocol import HartFrame
import logging
logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)

class HartApplicationLayer(object):
    """
    HART Application-Layer.
    """

    def __init__(self, interface, decoder_generator=None):
        """
        Args:
            interface (HartInterface or compatible): the HART interface to
                be used.
            decoder_generator=None (callable or None): a callable that returns
                a HartFrame instance with or without device specific commands.
        """
        self._interface = interface
        self.set_decoder_generator(decoder_generator)
        self.set_device_address(0)
        self.set_master_address("primary_master")

    def write(self, frame, number_of_preambles=5, preamble=b"\xff"):
        """
        Write a HART frame. The HART frame is encoded automatically.

        Except for command 0 the HART device address is set automatically
        before sending the frame content.

        Args:
            frame (HartFrame): the hart frame to be sent.
            number_of_preambles=5 (int): the number of preambles.
            preamble='\xff' (bytes): ``number_of_preambles`` times
                ``preamble`` is sent before the frame content is sent.
        """
        binary = frame
        if isinstance(frame, oser.ByteStruct):
            # use last device address if needed
            if frame.command != 0:
                logger.info("Using device address 0x%0x", self._device_address)
                frame.address.address.set(self._device_address)
            frame.address.master_address.set(self._master_address)
            binary = frame.encode()
        logger.debug("request:\n%s", frame)
        self._interface.write(binary, number_of_preambles, preamble)

    def read(self, timeout=1.0):
        """
        Read a message using the interface and decode it using
        the decoder.

        Args:
            timeout=1.0 (float or int): the timeout in seconds.

        Returns:
            HartFrame or None: if no frame was recieved ``None`` is returned.
        """
        response = self._interface.read(timeout=timeout)
        logger.debug("response:\n%s", response)
        if response is None:
            return response

        decoder = self._decoder_generator()
        decoder.decode(response)

        return decoder

    def query(self, request, number_of_preambles=5, preamble=b"\xff",
              timeout=1.0):
        """
        Query a HART command. The HART frame is encoded automatically.

        Args:
            request (HartFrame): the hart frame to be sent.
            number_of_preambles=5 (int): the number of preambles.
            preamble='\xff' (bytes): ``number_of_preambles`` times
                ``preamble`` is sent before the frame content is sent.
            timeout=1.0 (float or int): the timeout in seconds.

        Returns:
            HartFrame or None: if no frame was recieved ``None`` is returned.
        """
        self.write(request, number_of_preambles, preamble)
        response = self.read(timeout=timeout)
        return response

    def close(self):
        """
        Close the HART application layer and close the HART interface
        supplied.
        Is called in __del__, too.
        """
        self._interface.close()

    def __del__(self):
        try:
            self.close()
        except:
            pass

    def set_decoder_generator(self, decoder_generator=None):
        """
        Set the frame generator.

        Args:
            decoder_generator=HartFrame (callable): when called returns
        """
        if decoder_generator is None:
            decoder_generator = HartFrame
        self._decoder_generator = decoder_generator

    def set_device_address(self, device_address):
        """
        Set the current HART device address.

        Args:
            device_address (int): the HART device address to be used
                (<= 38 bits)
        """
        self._device_address = device_address

    def set_master_address(self, master_address):
        """
        Set the current simulated master address.

        Args:
            master_address (str): "primary_master" or "secondary_master"
        """
        assert master_address in ["primary_master", "secondary_master"]
        self._master_address = master_address

    def poll_devices(self, polling_addresses=None):
        """
        Poll ``polling_addresses`` to find HART devices.

        Args:
            polling_addresses=[0] (list of int): list of HART polling addresses

        Yields:
            int: HART device addresses
        """
        if polling_addresses is None:
            polling_addresses = [0, ]

        if not hasattr(polling_addresses, "__iter__"):
            polling_addresses = [polling_addresses]

        for polling_address in polling_addresses:
            frame = self._decoder_generator()
            frame.delimiter.address_type.set('polling')
            frame.delimiter.frame_type.set('stx')

            logger.info("polling address %i" % polling_address)
            frame.address.address.set(polling_address)
            try:
                response = self.query(frame)
                logger.debug("polling response: %s", response)
                if response is not None:
                    device_type = response.payload.device_type.get()
                    device_id = response.payload.device_id.value.get()
                    address = (device_type << 24) + device_id
                    logger.debug("yield address: 0x%0x", address)
                    yield address
            except Exception as e:
                print(e)
                pass

    def find_device(self, polling_addresses=None):
        """
        Find the first HART device by polling all addresses stated
        in ``polling_addresses`` and use the first device found.

        If no device is found an ``AssertionError`` is raised.

        Args:
            polling_addresses=[0] (list of int or int): the polling
                addresses used to find a HART device.

        Returns:
            int: Unique HART address of the first device found in polling.
        """

        devices = list(self.poll_devices(polling_addresses=polling_addresses))
        logger.info("Found %i HART devices: %s", len(devices), devices)
        if len(devices) == 0:
            logger.info("No device was found.")
            raise Exception("No device was found.")
        address = devices[0]
        self.set_device_address(address)
        logger.info("Using first device with address 0x%08x",
                    self._device_address)

        return address
