# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from oser import (
    ByteStruct,
    BitStruct,
    BitField,
    Nibble,
    Flag,
    UBInt8,
    UBInt16,
    UBInt32,
    IfElse,
    If,
    Array,
    String,
    BFloat,
    Lazy
)
from .data_types import PackedASCII, Date, Time, ResponseCode, FieldDeviceStatus, Delimiter, \
    _EmptyRequest, UnitAndVariable, DeviceVariableCodeUnitAndValue, \
    AnalogChannelNumberUnitAndLevel, TrendValueAndStatus, CommandByteCountAndData, HartResponse
import datetime


class Command33_ReadDeviceVariablesRequest(ByteStruct):
    """
    HART command 33 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command33_ReadDeviceVariablesRequest, self).__init__()
        self.device_variable_codes = Array(length=lambda self: self.up().byte_count.get(),
                                           prototype=UBInt8)


class Command33_ReadDeviceVariablesResponse(HartResponse):
    """
    HART command 33 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command33_ReadDeviceVariablesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_codes_units_and_values = Array(
            length=lambda self: self.up().byte_count.get() // 6,
            prototype=DeviceVariableCodeUnitAndValue)


class Command34_WritePrimaryVariableDampingValueRequest(ByteStruct):
    """
    HART command 34 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command34_WritePrimaryVariableDampingValueRequest, self).__init__()
        self.primary_variable_damping_value = BFloat()


class Command34_WritePrimaryVariableDampingValueResponse(HartResponse):
    """
    HART command 34 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command34_WritePrimaryVariableDampingValueResponse, self).__init__()
        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_access_restricted=16,
            error_busy=32)

        self.field_device_status = FieldDeviceStatus()
        self.actual_primary_variable_damping_value = BFloat()


class Command35_WritePrimaryVariableRangeValuesRequest(ByteStruct):
    """
    HART command 35 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command35_WritePrimaryVariableRangeValuesRequest, self).__init__()
        self.primary_variable_upper_and_lower_range_value_units_code = UBInt8()
        self.primary_variable_upper_range_value = BFloat()
        self.primary_variable_lower_range_value = BFloat()


class Command35_WritePrimaryVariableRangeValuesResponse(HartResponse):
    """
    HART command 35 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command35_WritePrimaryVariableRangeValuesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_lower_range_value_too_high=9,
            error_lower_range_value_too_low=10,
            error_upper_range_value_too_high=11,
            error_upper_range_value_too_low=12,
            error_upper_and_lower_range_values_out_of_limits=13,
            warning_span_too_small=14,
            error_access_restricted=16,
            error_invalid_units_code=18,
            error_invalid_span=29,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_upper_and_lower_range_value_units_code = UBInt8()
        self.primary_variable_upper_range_value = BFloat()
        self.primary_variable_lower_range_value = BFloat()


class Command36_SetPrimaryVariableUpperRangeValueRequest(_EmptyRequest):
    """
    HART command 36 (Common practice) request abstraction
    """


class Command36_SetPrimaryVariableUpperRangeValueResponse(HartResponse):
    """
    HART command 36 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command36_SetPrimaryVariableUpperRangeValueResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_applied_process_too_high=9,
            error_applied_process_too_low=10,
            warning_span_too_small=14,
            error_access_restricted=16,
            error_invalid_span=29,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command37_SetPrimaryVariableLowerRangeValueRequest(_EmptyRequest):
    """
    HART command 37 (Common practice) request abstractio
    """


class Command37_SetPrimaryVariableLowerRangeValueResponse(HartResponse):
    """
    HART command 37 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command37_SetPrimaryVariableLowerRangeValueResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_applied_process_too_high=9,
            error_applied_process_too_low=10,
            warning_new_lower_range_value_pushed=14,
            error_access_restricted=16,
            error_invalid_span=29,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command39_EEPROMControlRequest(ByteStruct):
    """
    HART command 39 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command39_EEPROMControlRequest, self).__init__()
        self.eeprom_control_code = UBInt8()


class Command39_EEPROMControlResponse(HartResponse):
    """
    HART command 39 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command39_EEPROMControlResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.eeprom_control_code = UBInt8()


class Command40_EnterExitFixedCurrentModeRequest(ByteStruct):
    """
    HART command 40 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command40_EnterExitFixedCurrentModeRequest, self).__init__()
        self.primary_variable_fixed_current_level = BFloat()


class Command40_EnterExitFixedCurrentModeResponse(HartResponse):
    """
    HART command 40 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command40_EnterExitFixedCurrentModeResponse, self).__init__()

        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_loop_current_not_active=11,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.actual_primary_variable_current_level = BFloat()


class Command41_PerformSelfTestRequest(_EmptyRequest):
    """
    HART command 41 (Common practice) request abstraction
    """


class Command41_PerformSelfTestResponse(HartResponse):
    """
    HART command 41 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command41_PerformSelfTestResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command42_PerformDeviceResetRequest(_EmptyRequest):
    """
    HART command 42 (Common practice) request abstraction
    """


class Command42_PerformDeviceResetResponse(HartResponse):
    """
    HART command 42 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command42_PerformDeviceResetResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command43_SetPrimaryVariableZeroRequest(_EmptyRequest):
    """
    HART command 43 (Common practice) request abstraction
    """


class Command43_SetPrimaryVariableZeroResponse(HartResponse):
    """
    HART command 43 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command43_SetPrimaryVariableZeroResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_applied_process_too_high=9,
            error_applied_process_too_low=10,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command44_WritePrimaryVariableUnitsRequest(ByteStruct):
    """
    HART command 44 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command44_WritePrimaryVariableUnitsRequest, self).__init__()
        self.primary_variable_units_code = UBInt8()


class Command44_WritePrimaryVariableUnitsResponse(HartResponse):
    """
    HART command 44 (Common practice) response abstractiosn.
    """

    def __init__(self):
        super(Command44_WritePrimaryVariableUnitsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_units_code = UBInt8()


class Command45_TrimLoopCurrentZeroRequest(ByteStruct):
    """
    HART command 45 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command45_TrimLoopCurrentZeroRequest, self).__init__()
        self.externally_measured_primary_variable_loop_current_level = BFloat()


class Command45_TrimLoopCurrentZeroResponse(HartResponse):
    """
    HART command 45 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command45_TrimLoopCurrentZeroResponse, self).__init__()
        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_incorrect_loop_current_mode_or_value=9,
            error_loop_current_not_active=11,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.actual_measured_primary_variable_loop_current_level = BFloat()


class Command46_TrimLoopCurrentGainRequest(ByteStruct):
    """
    HART command 46 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command46_TrimLoopCurrentGainRequest, self).__init__()
        self.externally_measured_primary_variable_loop_current_level = BFloat()


class Command46_TrimLoopCurrentGainResponse(HartResponse):
    """
    HART command 46 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command46_TrimLoopCurrentGainResponse, self).__init__()
        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_incorrect_loop_current_mode_or_value=9,
            error_loop_current_not_active=11,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.actual_measured_primary_variable_loop_current_level = BFloat()


class Command47_WritePrimaryVariableTransferFunctionRequest(ByteStruct):
    """
    HART command 47 (Common practice) request abstraction
    """

    def __init__(self):
        super(Command47_WritePrimaryVariableTransferFunctionRequest, self).__init__()
        self.primary_variable_transfer_function_code = UBInt8()


class Command47_WritePrimaryVariableTransferFunctionResponse(HartResponse):
    """
    HART command 47 (Common practice) response abstraction.
    """

    def __init__(self):
        super(Command47_WritePrimaryVariableTransferFunctionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_transfer_function_code = UBInt8()


class Command48_ReadAdditionalDeviceStatusRequest(ByteStruct):
    """
    HART command 48 request abstraction.
    """

    def __init__(self):
        super(Command48_ReadAdditionalDeviceStatusRequest, self).__init__()

        class DeviceSpecificStatus(BitStruct):
            def __init__(self):
                super(DeviceSpecificStatus, self).__init__()
                self.value = BitField(length=48)

        self.device_specific_status = If(lambda self: self.up().byte_count.get() > 0,
                                         if_true=DeviceSpecificStatus())
        self.extended_device_status = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.device_operating_mode = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.status0 = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.status1 = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.analog_channel_saturated = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.status2 = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.status3 = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())
        self.analog_channel_fixed = If(lambda self: self.up().byte_count.get() > 0, if_true=UBInt8())

        class AdditionalDeviceSpecificStatus(BitStruct):
            def __init__(self):
                super(AdditionalDeviceSpecificStatus, self).__init__()
                self.value = BitField(length=88)

        self.additional_device_specific_status = If(lambda self: self.up().byte_count.get() > 0,
                                                    if_true=AdditionalDeviceSpecificStatus())


class Command48_ReadAdditionalDeviceStatusResponse(HartResponse):
    """
    HART command 48 response abstraction.
    """

    def __init__(self):
        super(Command48_ReadAdditionalDeviceStatusResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_in_progress=8,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_specific_status = BitStruct()
        self.device_specific_status.value = BitField(length=48)
        self.extended_device_status = UBInt8()
        self.device_operating_mode = UBInt8()
        self.status0 = UBInt8()
        self.status1 = UBInt8()
        self.analog_channel_saturated = UBInt8()
        self.status2 = UBInt8()
        self.status3 = UBInt8()
        self.analog_channel_fixed = UBInt8()
        self.additional_device_specific_status = BitStruct()
        self.additional_device_specific_status.value = BitField(length=88)


class Command49_WritePrimaryVariableTransducerSerialNumberRequest(ByteStruct):
    """
    HART command 49 request abstraction.
    """

    def __init__(self):
        super(Command49_WritePrimaryVariableTransducerSerialNumberRequest, self).__init__()
        self.primary_variable_transducer_serial_number = BitStruct()
        self.primary_variable_transducer_serial_number.value = BitField(length=24)


class Command49_WritePrimaryVariableTransducerSerialNumberResponse(HartResponse):
    """
    HART command 49 response abstraction.
    """

    def __init__(self):
        super(Command49_WritePrimaryVariableTransducerSerialNumberResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_transducer_serial_number = BitStruct()
        self.primary_variable_transducer_serial_number.value = BitField(length=24)


class Command50_ReadDynamicVariableAssignmentsRequest(_EmptyRequest):
    """
    HART command 50 request abstraction.
    """


class Command50_ReadDynamicVariableAssignmentsResponse(HartResponse):
    """
    HART command 50 response abstraction.
    """

    def __init__(self):
        super(Command50_ReadDynamicVariableAssignmentsResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_device_variable = UBInt8()
        self.secondary_variable_device_variable = UBInt8()
        self.tertiary_variable_device_variable = UBInt8()
        self.quaternary_variable_device_variable = UBInt8()


class Command51_WriteDynamicVariableAssignmentsRequest(ByteStruct):
    """
    HART command 51 request abstraction.
    """

    def __init__(self):
        super(Command51_WriteDynamicVariableAssignmentsRequest, self).__init__()

        self.device_variables = Array(length=lambda self: self.up().byte_count.get(),
                                      prototype=UBInt8)


class Command51_WriteDynamicVariableAssignmentsResponse(HartResponse):
    """
    HART command 51 response abstraction.
    """

    def __init__(self):
        super(Command51_WriteDynamicVariableAssignmentsResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variables = Array(length=lambda self: self.up().byte_count.get() - 2,
                                      prototype=UBInt8)


class Command52_SetDeviceVariableZeroRequest(ByteStruct):
    """
    HART command 52 request abstraction.
    """

    def __init__(self):
        super(Command52_SetDeviceVariableZeroRequest, self).__init__()

        self.device_variable_to_zero = UBInt8()


class Command52_SetDeviceVariableZeroResponse(HartResponse):
    """
    HART command 52 response abstraction.
    """

    def __init__(self):
        super(Command52_SetDeviceVariableZeroResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_applied_process_too_high=8,
            error_applied_process_too_low=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_to_zero = UBInt8()


class Command53_WriteDeviceVariableUnitsRequest(ByteStruct):
    """
    HART command 53 request abstraction.
    """

    def __init__(self):
        super(Command53_WriteDeviceVariableUnitsRequest, self).__init__()

        self.device_variable_code = UBInt8()
        self.device_variable_units_code = UBInt8()


class Command53_WriteDeviceVariableUnitsResponse(HartResponse):
    """
    HART command 53 response abstraction.
    """

    def __init__(self):
        super(Command53_WriteDeviceVariableUnitsResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_device_variable_code=11,
            error_invalid_units_code=12,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.device_variable_units_code = UBInt8()


class Command54_ReadDeviceVariableInformationRequest(ByteStruct):
    """
    HART command 54 request abstraction.
    """

    def __init__(self):
        super(Command54_ReadDeviceVariableInformationRequest, self).__init__()

        self.device_variable_code = UBInt8()


class Command54_ReadDeviceVariableInformationResponse(HartResponse):
    """
    HART command 54 response abstraction.
    """

    def __init__(self):
        super(Command54_ReadDeviceVariableInformationResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.device_variable_transducer_serial_number = BitStruct()
        self.device_variable_transducer_serial_number.value = BitField(length=24)
        self.device_variable_limits = UBInt8()
        self.device_variable_upper_transducer_limit = BFloat()
        self.device_variable_lower_transducer_limit = BFloat()
        self.device_variable_damping_value = BFloat()
        self.device_variable_minimum_span = BFloat()
        self.device_variable_classification = UBInt8()
        self.device_variable_family = UBInt8()
        self.acquisition_period = Time(datetime.time())
        self.device_variable_properties = UBInt8()


class Command55_WriteDeviceVariableDampingValueRequest(ByteStruct):
    """
    HART command 55 request abstraction.
    """

    def __init__(self):
        super(Command55_WriteDeviceVariableDampingValueRequest, self).__init__()

        self.device_variable_code = UBInt8()
        self.device_variable_damping_value = BFloat()


class Command55_WriteDeviceVariableDampingValueResponse(HartResponse):
    """
    HART command 55 response abstraction.
    """

    def __init__(self):
        super(Command55_WriteDeviceVariableDampingValueResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.device_variable_damping_value = BFloat()


class Command56_WriteDeviceVariableTransducerSerialNumberRequest(ByteStruct):
    """
    HART command 56 request abstraction.
    """

    def __init__(self):
        super(Command56_WriteDeviceVariableTransducerSerialNumberRequest, self).__init__()

        self.device_variable_code = UBInt8()
        self.device_variable_transducer_serial_number = BitStruct()
        self.device_variable_transducer_serial_number.value = BitField(length=24)


class Command56_WriteDeviceVariableTransducerSerialNumberResponse(HartResponse):
    """
    HART command 56 response abstraction.
    """

    def __init__(self):
        super(Command56_WriteDeviceVariableTransducerSerialNumberResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.device_variable_transducer_serial_number = BitStruct()
        self.device_variable_transducer_serial_number.value = BitField(length=24)


class Command57_ReadUnitTagDescriptorDateRequest(_EmptyRequest):
    """
    HART command 57 request abstraction.
    """


class Command57_ReadUnitTagDescriptorDateResponse(HartResponse):
    """
    HART command 57 response abstraction.
    """

    def __init__(self):
        super(Command57_ReadUnitTagDescriptorDateResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.unit_tag = PackedASCII(length=6, value=b"?" * 8)
        self.unit_descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.unit_date = BitStruct()
        self.unit_date.value = BitField(length=24)


class Command58_WriteUnitTagDescriptorDateRequest(ByteStruct):
    """
    HART command 58 request abstraction.
    """

    def __init__(self):
        super(Command58_WriteUnitTagDescriptorDateRequest, self).__init__()

        self.unit_tag = PackedASCII(length=6, value=b"?" * 8)
        self.unit_descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.unit_date = BitStruct()
        self.unit_date.value = BitField(length=24)


class Command58_WriteUnitTagDescriptorDateResponse(HartResponse):
    """
    HART command 58 response abstraction.
    """

    def __init__(self):
        super(Command58_WriteUnitTagDescriptorDateResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.unit_tag = PackedASCII(length=6, value=b"?" * 8)
        self.unit_descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.unit_date = BitStruct()
        self.unit_date.value = BitField(length=24)


class Command59_WriteNumberOfResponsePreamblesRequest(ByteStruct):
    """
    HART command 59 request abstraction.
    """

    def __init__(self):
        super(Command59_WriteNumberOfResponsePreamblesRequest, self).__init__()

        self.number_of_preambles = UBInt8()


class Command59_WriteNumberOfResponsePreamblesResponse(HartResponse):
    """
    HART command 59 response abstraction.
    """

    def __init__(self):
        super(Command59_WriteNumberOfResponsePreamblesResponse, self).__init__()

        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.number_of_preambles = UBInt8()


class Command60_ReadAnalogChannelAndPercentOfRangeRequest(ByteStruct):
    """
    HART command 60 request abstraction.
    """

    def __init__(self):
        super(Command60_ReadAnalogChannelAndPercentOfRangeRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()


class Command60_ReadAnalogChannelAndPercentOfRangeResponse(HartResponse):
    """
    HART command 60 response abstraction.
    """

    def __init__(self):
        super(Command60_ReadAnalogChannelAndPercentOfRangeResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.analog_channel_level = BFloat()
        self.analog_channel_percent_of_range = BFloat()


class Command61_ReadDynamicVariablesAndPrimaryVariableAnalogChannelRequest(_EmptyRequest):
    """
    HART command 61 request abstraction.
    """


class Command61_ReadDynamicVariablesAndPrimaryVariableAnalogChannelResponse(HartResponse):
    """
    HART command 61 response abstraction.
    """

    def __init__(self):
        super(Command61_ReadDynamicVariablesAndPrimaryVariableAnalogChannelResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_analog_channel_units_code = UBInt8()
        self.primary_variable_analog_level = BFloat()
        self.dynamic_variables = Array(
            length=lambda self: (self.up().byte_count.get() - 7) // 5,
            prototype=UnitAndVariable)


class Command62_ReadAnalogChannelsRequest(ByteStruct):
    """
    HART command 62 request abstraction.
    """

    def __init__(self):
        super(Command62_ReadAnalogChannelsRequest, self).__init__()

        self.analog_channel_number_codes = Array(
            length=lambda self: self.up().byte_count.get(),
            prototype=UBInt8)


class Command62_ReadAnalogChannelsResponse(HartResponse):
    """
    HART command 62 response abstraction.
    """

    def __init__(self):
        super(Command62_ReadAnalogChannelsResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_update_failure=8)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_levels = Array(
            length=lambda self: self.up().byte_count.get() // 6,
            prototype=AnalogChannelNumberUnitAndLevel)


class Command63_ReadAnalogChannelInformationRequest(ByteStruct):
    """
    HART command 63 request abstraction.
    """

    def __init__(self):
        super(Command63_ReadAnalogChannelInformationRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()


class Command63_ReadAnalogChannelInformationResponse(HartResponse):
    """
    HART command 63 response abstraction.
    """

    def __init__(self):
        super(Command63_ReadAnalogChannelInformationResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_alarm_selection_code = UBInt8()
        self.analog_channel_transfer_function_code = UBInt8()
        self.analog_channel_upper_and_lower_range_value_units_code = UBInt8()
        self.analog_channel_upper_range_value = BFloat()
        self.analog_channel_lower_range_value = BFloat()
        self.analog_channel_damping_value = BFloat()
        self.analog_channel_flags = UBInt8()


class Command64_WriteAnalogChannelAdditionalDampingValueRequest(ByteStruct):
    """
    HART command 64 request abstraction.
    """

    def __init__(self):
        super(Command64_WriteAnalogChannelAdditionalDampingValueRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_additional_damping_value = BFloat()


class Command64_WriteAnalogChannelAdditionalDampingValueResponse(HartResponse):
    """
    HART command 64 response abstraction.
    """

    def __init__(self):
        super(Command64_WriteAnalogChannelAdditionalDampingValueResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_additional_damping_value = BFloat()


class Command65_WriteAnalogChannelRangeValuesRequest(ByteStruct):
    """
    HART command 65 request abstraction.
    """

    def __init__(self):
        super(Command65_WriteAnalogChannelRangeValuesRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_upper_and_lower_range_value_units_code = UBInt8()
        self.analog_channel_upper_range_value = BFloat()
        self.analog_channel_lower_range_value = BFloat()


class Command65_WriteAnalogChannelRangeValuesResponse(HartResponse):
    """
    HART command 65 response abstraction.
    """

    def __init__(self):
        super(Command65_WriteAnalogChannelRangeValuesResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_lower_range_value_too_high=9,
            error_lower_range_value_too_low=10,
            error_upper_range_value_too_high=11,
            error_upper_range_value_too_low=12,
            error_upper_and_lower_range_values_out_of_limits=13,
            warning_span_too_small=14,
            error_invalid_analog_channel_code_number=15,
            error_access_restricted=16,
            error_invalid_range_units_code=28,
            error_invalid_span=29,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_upper_and_lower_range_value_units_code = UBInt8()
        self.analog_channel_upper_range_value = BFloat()
        self.analog_channel_lower_range_value = BFloat()


class Command66_EnterExitFixedAnalogChannelModeRequest(ByteStruct):
    """
    HART command 66 request abstraction.
    """

    def __init__(self):
        super(Command66_EnterExitFixedAnalogChannelModeRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.fixed_analog_channel_level = BFloat()


class Command66_EnterExitFixedAnalogChannelModeResponse(HartResponse):
    """
    HART command 66 response abstraction.
    """

    def __init__(self):
        super(Command66_EnterExitFixedAnalogChannelModeResponse, self).__init__()

        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_in_multidrop_mode=11,
            error_invalid_units_code=12,
            error_invalid_analog_channel_code_number=15,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.fixed_analog_channel_level = BFloat()


class Command67_TrimAnalogChannelZeroRequest(ByteStruct):
    """
    HART command 67 request abstraction.
    """

    def __init__(self):
        super(Command67_TrimAnalogChannelZeroRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.externally_measured_analog_channel_level = BFloat()


class Command67_TrimAnalogChannelZeroResponse(HartResponse):
    """
    HART command 67 response abstraction.
    """

    def __init__(self):
        super(Command67_TrimAnalogChannelZeroResponse, self).__init__()

        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_not_in_proper_analog_channel_mode=9,
            error_in_multidrop_mode=11,
            error_invalid_units_code=12,
            error_invalid_analog_channel_code_number=15,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.actual_channel_level = BFloat()


class Command68_TrimAnalogChannelGainRequest(ByteStruct):
    """
    HART command 68 request abstraction.
    """

    def __init__(self):
        super(Command68_TrimAnalogChannelGainRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.externally_measured_analog_channel_level = BFloat()


class Command68_TrimAnalogChannelGainResponse(HartResponse):
    """
    HART command 68 response abstraction.
    """

    def __init__(self):
        super(Command68_TrimAnalogChannelGainResponse, self).__init__()

        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_not_in_proper_analog_channel_mode=9,
            error_in_multidrop_mode=11,
            error_invalid_units_code=12,
            error_invalid_analog_channel_code_number=15,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_units_code = UBInt8()
        self.externally_measured_analog_channel_level = BFloat()


class Command69_WriteAnalogChannelTransferFunctionRequest(ByteStruct):
    """
    HART command 69 request abstraction.
    """

    def __init__(self):
        super(Command69_WriteAnalogChannelTransferFunctionRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_transfer_function_code = UBInt8()


class Command69_WriteAnalogChannelTransferFunctionResponse(HartResponse):
    """
    HART command 69 response abstraction.
    """

    def __init__(self):
        super(Command69_WriteAnalogChannelTransferFunctionResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_transfer_function_code=13,
            error_invalid_analog_channel_code_number=15,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_transfer_function_code = UBInt8()


class Command70_ReadAnalogChannelEndpointValuesRequest(ByteStruct):
    """
    HART command 70 request abstraction.
    """

    def __init__(self):
        super(Command70_ReadAnalogChannelEndpointValuesRequest, self).__init__()

        self.analog_channel_number_code = UBInt8()


class Command70_ReadAnalogChannelEndpointValuesResponse(HartResponse):
    """
    HART command 70 response abstraction.
    """

    def __init__(self):
        super(Command70_ReadAnalogChannelEndpointValuesResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.analog_channel_number_code = UBInt8()
        self.analog_channel_upper_and_lower_range_value_units_code = UBInt8()
        self.analog_channel_upper_endpoint_value = BFloat()
        self.analog_channel_lower_endpoint_value = BFloat()
        self.analog_channel_upper_limit_value = BFloat()
        self.analog_channel_lower_limit_value = BFloat()


class Command71_LockDeviceRequest(ByteStruct):
    """
    HART command 71 request abstraction.
    """

    def __init__(self):
        super(Command71_LockDeviceRequest, self).__init__()

        self.lock_code = UBInt8()


class Command71_LockDeviceResponse(HartResponse):
    """
    HART command 71 response abstraction.
    """

    def __init__(self):
        super(Command71_LockDeviceResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_lock_code=10,
            error_cannot_lock_device=11,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.lock_code = UBInt8()


class Command72_SquawkRequest(ByteStruct):
    """
    HART command 72 request abstraction.
    """

    def __init__(self):
        super(Command72_SquawkRequest, self).__init__()

        self.squawk_control = UBInt8()


class Command72_SquawkResponse(HartResponse):
    """
    HART command 72 response abstraction.
    """

    def __init__(self):
        super(Command72_SquawkResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_device_specific_command_error=6,
            error_unable_to_squawk=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.squawk_control = UBInt8()


class Command73_FindDeviceRequest(_EmptyRequest):
    """
    HART command 73 request abstraction.
    """


class Command73_FindDeviceResponse(HartResponse):
    """
    HART command 73 response abstraction.
    """

    def __init__(self):
        super(Command73_FindDeviceResponse, self).__init__()
        self.response_code = ResponseCode()
        self.field_device_status = FieldDeviceStatus()

        self.data_object_id = UBInt8()
        self.device_type = UBInt16()
        self.minimum_preambles_request = UBInt8()
        self.protocol_revision = UBInt8()

        self.device_revision_level = UBInt8()
        self.software_revision_level = UBInt8()

        self.hardware_revision_level_and_physical_signaling_code = UBInt8()
        self.flags = UBInt8()

        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)

        self.minimum_preambles_response = If(lambda self: self.protocol_revision > 5,
                                             if_true=UBInt8())
        self.maximum_number_of_device_variables = If(lambda self: self.protocol_revision > 5,
                                                     if_true=UBInt8())
        self.configuration_change_counter = If(lambda self: self.protocol_revision > 5,
                                               if_true=UBInt16())
        self.extended_flags = If(lambda self: self.protocol_revision > 5,
                                 if_true=UBInt8())

        self.manufacturer_id = If(lambda self: self.protocol_revision >= 7,
                                  if_true=UBInt16())
        self.private_label_distributor_code = If(lambda self: self.protocol_revision >= 7,
                                                 if_true=UBInt16())
        self.device_profile = If(lambda self: self.protocol_revision >= 7,
                                 if_true=UBInt8())


class Command74_ReadIOSystemCapabilitiesRequest(_EmptyRequest):
    """
    HART command 74 request abstraction.
    """


class Command74_ReadIOSystemCapabilitiesResponse(HartResponse):
    """
    HART command 74 response abstraction.
    """

    def __init__(self):
        super(Command74_ReadIOSystemCapabilitiesResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.maximum_number_of_io_cards = UBInt8()
        self.maximum_number_of_channels_per_card = UBInt8()
        self.maximum_number_of_subdevices_per_channel = UBInt8()
        self.number_of_devices = UBInt16()
        self.maximum_number_of_delayed_responses = UBInt8()
        self.master_mode = UBInt8()
        self.retry_count = UBInt8()


class Command75_PollSubDeviceRequest(ByteStruct):
    """
    HART command 75 request abstraction.
    """

    def __init__(self):
        super(Command75_PollSubDeviceRequest, self).__init__()

        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.subdevice_polling_address = UBInt8()


class Command75_PollSubDeviceResponse(HartResponse):
    """
    HART command 75 response abstraction.
    """

    def __init__(self):
        super(Command75_PollSubDeviceResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_9_no_subdevice_found=9,
            error_access_restricted=16,
            error_invalid_io_card_number=21,
            error_invalid_channel_number=22,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.data_object_id = UBInt8()
        self.device_type = UBInt16()
        self.minimum_preambles_request = UBInt8()
        self.protocol_revision = UBInt8()

        self.device_revision_level = UBInt8()
        self.software_revision_level = UBInt8()

        self.hardware_revision_level_and_physical_signaling_code = UBInt8()
        self.flags = UBInt8()

        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)

        self.minimum_preambles_response = If(lambda self: self.protocol_revision > 5,
                                             if_true=UBInt8())
        self.maximum_number_of_device_variables = If(lambda self: self.protocol_revision > 5,
                                                     if_true=UBInt8())
        self.configuration_change_counter = If(lambda self: self.protocol_revision > 5,
                                               if_true=UBInt16())
        self.extended_flags = If(lambda self: self.protocol_revision > 5,
                                 if_true=UBInt8())

        self.manufacturer_id = If(lambda self: self.protocol_revision >= 7,
                                  if_true=UBInt16())
        self.private_label_distributor_code = If(lambda self: self.protocol_revision >= 7,
                                                 if_true=UBInt16())
        self.device_profile = If(lambda self: self.protocol_revision >= 7,
                                 if_true=UBInt8())


class Command76_ReadLockDeviceStateRequest(_EmptyRequest):
    """
    HART command 76 request abstraction.
    """


class Command76_ReadLockDeviceStateResponse(HartResponse):
    """

    HART command 76 response abstraction.
    """

    def __init__(self):
        super(Command76_ReadLockDeviceStateResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.lock_status = UBInt8()


class Command77_SendCommandToSubDeviceRequest(ByteStruct):
    """
    HART command 77 request abstraction.
    """

    def __init__(self):
        super(Command77_SendCommandToSubDeviceRequest, self).__init__()

        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.transmit_preamble_count = UBInt8()
        self.delimiter = Delimiter()
        self.address = IfElse(condition=lambda self: self.delimiter.address_type == "polling",
                              if_true=UBInt8(),
                              if_false=Array(length=5, prototype=UBInt8))
        self.command = UBInt8()
        self.byte_count = UBInt8()
        self.data = IfElse(condition=lambda self: self.delimiter.address_type == "polling",
                           if_true=Array(
                               length=lambda self: self.up().byte_count.get() - 7,
                               prototype=UBInt8),
                           if_false=Array(
                               length=lambda self: self.up().byte_count.get() - 11,
                               prototype=UBInt8))


class Command77_SendCommandToSubDeviceResponse(HartResponse):
    """
    HART command 77 response abstraction.
    """

    def __init__(self):
        super(Command77_SendCommandToSubDeviceResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_access_restricted=16,
            error_invalid_io_card_number=21,
            error_invalid_channel_number=22,
            error_subdevice_response_too_long=23,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.delimiter = Delimiter()
        self.address = IfElse(condition=lambda self: self.delimiter.address_type == "polling",
                              if_true=UBInt8(),
                              if_false=Array(length=5, prototype=UBInt8))
        self.command = UBInt8()
        self.byte_count = UBInt8()
        self.data = IfElse(condition=lambda self: self.delimiter.address_type == "polling",
                           if_true=Array(length=lambda self: self.up().byte_count.get() - 8,
                                         prototype=UBInt8),
                           if_false=Array(length=lambda self: self.up().byte_count.get() - 12,
                                          prototype=UBInt8))


class Command78_ReadAggregatedCommandsRequest(ByteStruct):
    """
    HART command 78 request abstraction.
    """

    def __init__(self):
        super(Command78_ReadAggregatedCommandsRequest, self).__init__()

        self.requested_commands = UBInt8()
        self.commands = Array(length=lambda self: self.requested_commands.get(),
                              prototype=CommandByteCountAndData)


class Command78_ReadAggregatedCommandsResponse(HartResponse):
    """
    HART command 78 response abstraction.
    """

    def __init__(self):
        super(Command78_ReadAggregatedCommandsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_invalid_command_requested=9,
            warning_command_response_truncated=30,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.extended_field_device_status = UBInt8()
        self.requested_commands = UBInt8()
        self.commands = Array(length=lambda self: self.requested_commands.get(),
                              prototype=CommandByteCountAndData)


class Command79_WriteDeviceVariableRequest(ByteStruct):
    """
    HART command 79 request abstraction.
    """

    def __init__(self):
        super(Command79_WriteDeviceVariableRequest, self).__init__()

        self.device_variable_code = UBInt8()
        self.write_device_variable_command_code = UBInt8()
        self.units_code = UBInt8()
        self.device_variable_value = BFloat()
        self.device_variable_status = UBInt8()


class Command79_WriteDeviceVariableResponse(HartResponse):
    """

    HART command 79 response abstraction.
    """

    def __init__(self):
        super(Command79_WriteDeviceVariableResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_device_family_status_bits_not_set=8,
            error_invalid_write_device_variable_command_code=10,
            warning_rate_of_change_limit_exceeded=14,
            error_access_restricted=16,
            error_invalid_device_variable_index=17,
            error_invalid_units_code=18,
            error_device_variable_index_not_allowed=19,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.write_device_variable_command_code = UBInt8()
        self.units_code = UBInt8()
        self.device_variable_value = BFloat()
        self.device_variable_status = UBInt8()


class Command80_ReadDeviceVariableTrimPointsRequest(ByteStruct):
    """
    HART command 80 request abstraction.
    """

    def __init__(self):
        super(Command80_ReadDeviceVariableTrimPointsRequest, self).__init__()

        self.device_variable_code = UBInt8()


class Command80_ReadDeviceVariableTrimPointsResponse(HartResponse):
    """

    HART command 80 response abstraction.
    """

    def __init__(self):
        super(Command80_ReadDeviceVariableTrimPointsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_device_variable_index=17,
            error_device_variable_index_not_allowed=19,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.trim_points_units_code = UBInt8()
        self.lower_trim_point = BFloat()
        self.upper_trim_point = BFloat()


class Command81_ReadDeviceVariableTrimGuidelinesRequest(ByteStruct):
    """
    HART command 81 request abstraction.
    """

    def __init__(self):
        super(Command81_ReadDeviceVariableTrimGuidelinesRequest, self).__init__()

        self.device_variable_guidelines = UBInt8()


class Command81_ReadDeviceVariableTrimGuidelinesResponse(HartResponse):
    """
    HART command 81 response abstraction.
    """

    def __init__(self):
        super(Command81_ReadDeviceVariableTrimGuidelinesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_device_variable_index=17,
            error_device_variable_index_not_allowed=19,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_guidelines = UBInt8()
        self.trim_points_supported = UBInt8()
        self.trim_points_units_code = UBInt8()
        self.minimum_lower_trim_point_value = BFloat()
        self.maximum_lower_trim_point_value = BFloat()
        self.minimum_upper_trim_point_value = BFloat()
        self.maximum_upper_trim_point_value = BFloat()
        self.minimum_differential = BFloat()


class Command82_WriteDeviceVariableTrimPointRequest(ByteStruct):
    """
    HART command 82 request abstraction.
    """

    def __init__(self):
        super(Command82_WriteDeviceVariableTrimPointRequest, self).__init__()

        self.device_variable_code = UBInt8()
        self.trim_point = UBInt8()
        self.trim_point_units_code = UBInt8()
        self.trim_point_value = BFloat()


class Command82_WriteDeviceVariableTrimPointResponse(HartResponse):
    """
    HART command 82 response abstraction.
    """

    def __init__(self):
        super(Command82_WriteDeviceVariableTrimPointResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_applied_process_too_high=9,
            error_applied_process_too_low=10,
            error_trim_error_excess_correction_attempted=11,
            error_computation_error_trim_values_were_not_changed=13,
            warning_span_too_small=14,
            error_access_restricted=16,
            error_invalid_device_variable_index=17,
            error_invalid_units_code=18,
            error_device_variable_index_not_allowed=19,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()
        self.trim_point = UBInt8()
        self.trim_point_units_code = UBInt8()
        self.trim_point_value = BFloat()


class Command83_ResetDeviceVariableTrimRequest(ByteStruct):
    """
    HART command 83 request abstraction.
    """

    def __init__(self):
        super(Command83_ResetDeviceVariableTrimRequest, self).__init__()

        self.device_variable_code = UBInt8()


class Command83_ResetDeviceVariableTrimResponse(HartResponse):
    """
    HART command 83 response abstraction.
    """

    def __init__(self):
        super(Command83_ResetDeviceVariableTrimResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_invalid_device_variable_index=17,
            error_device_variable_index_not_allowed=19,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code = UBInt8()


class Command84_ReadSubDeviceIdentitySummaryRequest(ByteStruct):
    """
    HART command 84 request abstraction.
    """

    def __init__(self):
        super(Command84_ReadSubDeviceIdentitySummaryRequest, self).__init__()

        self.subdevice_index = UBInt16()


class Command84_ReadSubDeviceIdentitySummaryResponse(HartResponse):
    """
    HART command 84 response abstraction.
    """

    def __init__(self):
        super(Command84_ReadSubDeviceIdentitySummaryResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_duplicate_id=8)
        self.field_device_status = FieldDeviceStatus()

        self.subdevice_index = UBInt16()
        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.manufacturer_id = UBInt16()
        self.expanded_device_type_code = UBInt16()
        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)
        self.universal_command_revision_level = UBInt8()
        self.long_tag = String(length=32)
        self.device_revision = UBInt8()
        self.device_profile = UBInt8()
        self.private_label_distributor_code = UBInt16()


class Command85_ReadIOChannelStatisticsRequest(ByteStruct):
    """
    HART command 85 request abstraction.
    """

    def __init__(self):
        super(Command85_ReadIOChannelStatisticsRequest, self).__init__()

        self.io_card = UBInt8()
        self.channel = UBInt8()


class Command85_ReadIOChannelStatisticsResponse(HartResponse):
    """
    HART command 85 response abstraction.
    """

    def __init__(self):
        super(Command85_ReadIOChannelStatisticsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_io_card_number=21,
            error_invalid_channel_number=22,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.stx_count = UBInt16()
        self.ack_count = UBInt16()
        self.ostx_count = UBInt16()
        self.oack_count = UBInt16()
        self.back_count = UBInt16()


class Command86_ReadSubDeviceStatisticsRequest(ByteStruct):
    """
    HART command 86 request abstraction.
    """

    def __init__(self):
        super(Command86_ReadSubDeviceStatisticsRequest, self).__init__()

        self.subdevice_index = UBInt16()


class Command86_ReadSubDeviceStatisticsResponse(HartResponse):
    """
    HART command 86 response abstraction.
    """

    def __init__(self):
        super(Command86_ReadSubDeviceStatisticsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.subdevice_index = UBInt16()
        self.stx_count = UBInt16()
        self.ack_count = UBInt16()
        self.back_count = UBInt16()


class Command87_WriteIOSystemMasterModeRequest(ByteStruct):
    """
    HART command 87 request abstraction.
    """

    def __init__(self):
        super(Command87_WriteIOSystemMasterModeRequest, self).__init__()

        self.master_mode = UBInt8()


class Command87_WriteIOSystemMasterModeResponse(HartResponse):
    """
    HART command 87 response abstraction.
    """

    def __init__(self):
        super(Command87_WriteIOSystemMasterModeResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.master_mode = UBInt8()


class Command88_WriteIOSystemRetryCountRequest(ByteStruct):
    """
    HART command 88 request abstraction.
    """

    def __init__(self):
        super(Command88_WriteIOSystemRetryCountRequest, self).__init__()

        self.retry_count = UBInt8()


class Command88_WriteIOSystemRetryCountResponse(HartResponse):
    """
    HART command 88 response abstraction.
    """

    def __init__(self):
        super(Command88_WriteIOSystemRetryCountResponse, self).__init__()
        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.retry_count = UBInt8()


class Command89_SetRealTimeClockRequest(ByteStruct):
    """
    HART command 89 request abstraction.
    """

    def __init__(self):
        super(Command89_SetRealTimeClockRequest, self).__init__()

        self.time_set_code = UBInt8()
        self.date = Date(datetime.date(1900, 1, 1))
        self.time = Time(datetime.time())
        self.padding = UBInt16()


class Command89_SetRealTimeClockResponse(HartResponse):
    """
    HART command 89 response abstraction.
    """

    def __init__(self):
        super(Command89_SetRealTimeClockResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_data_code_detected=9,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.time_set_code = UBInt8()
        self.date = Date(datetime.date(1900, 1, 1))
        self.time = Time(datetime.time())


class Command90_ReadRealTimeClockRequest(_EmptyRequest):
    """
    HART command 90 request abstraction.
    """


class Command90_ReadRealTimeClockResponse(HartResponse):
    """
    HART command 90 response abstraction.
    """

    def __init__(self):
        super(Command90_ReadRealTimeClockResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.date = Date(datetime.date(1900, 1, 1))
        self.time = Time(datetime.time())
        self.date_clock_last_set = Date(datetime.date(1900, 1, 1))
        self.time_clock_last_set = Time(datetime.time())
        self.rtc_flags = UBInt8()


class Command91_ReadTrendConfigurationRequest(ByteStruct):
    """
    HART command 91 request abstraction.
    """

    def __init__(self):
        super(Command91_ReadTrendConfigurationRequest, self).__init__()

        self.trend_number = UBInt8()


class Command91_ReadTrendConfigurationResponse(HartResponse):
    """
    HART command 91 response abstraction.
    """

    def __init__(self):
        super(Command91_ReadTrendConfigurationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_trend_number=11)
        self.field_device_status = FieldDeviceStatus()

        self.trend_number = UBInt8()
        self.total_trends_supported = UBInt8()
        self.trend_control_code = UBInt8()
        self.device_variable_code = UBInt8()
        self.trend_sample_interval = Time(datetime.time())


class Command92_WriteTrendConfigurationRequest(ByteStruct):
    """
    HART command 92 request abstraction.
    """

    def __init__(self):
        super(Command92_WriteTrendConfigurationRequest, self).__init__()

        self.trend_number = UBInt8()
        self.trend_control_code = UBInt8()
        self.device_variable_code = UBInt8()
        self.trend_sample_interval = Time(datetime.time())


class Command92_WriteTrendConfigurationResponse(HartResponse):
    """
    HART command 92 response abstraction.
    """

    def __init__(self):
        super(Command92_WriteTrendConfigurationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_invalid_trend_number=11,
            error_access_restricted=16,
            error_invalid_device_variable_index=17,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.trend_number = UBInt8()
        self.trend_control_code = UBInt8()
        self.device_variable_code = UBInt8()
        self.trend_sample_interval = Time(datetime.time())


class Command93_ReadTrendRequest(ByteStruct):
    """
    HART command 93 request abstraction.
    """

    def __init__(self):
        super(Command93_ReadTrendRequest, self).__init__()

        self.trend_number = UBInt8()


class Command93_ReadTrendResponse(HartResponse):
    """
    HART command 93 response abstraction.
    """

    def __init__(self):
        super(Command93_ReadTrendResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_trend_not_active=8,
            error_invalid_trend_number=11)
        self.field_device_status = FieldDeviceStatus()

        self.trend_number = UBInt8()
        self.device_variable_code = UBInt8()
        self.device_variable_classification = UBInt8()
        self.device_variable_units_code = UBInt8()
        self.date_trend_value0 = Date(datetime.date(1900, 1, 1))
        self.time_trend_value0 = Time(datetime.time())
        self.time_sample_interval = Time(datetime.time())
        self.trend_values = Array(length=12, prototype=TrendValueAndStatus)


class Command94_ReadIOSystemClientSideCommunicationStatisticsRequest(_EmptyRequest):
    """
    HART command 94 request abstraction.
    """


class Command94_ReadIOSystemClientSideCommunicationStatisticsResponse(HartResponse):
    """
    HART command 94 response abstraction.
    """

    def __init__(self):
        super(Command94_ReadIOSystemClientSideCommunicationStatisticsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.messages_received = UBInt32()
        self.messages_returned = UBInt32()
        self.requests_forwared = UBInt32()
        self.responses_returned = UBInt32()


class Command95_ReadDeviceCommunicationStatisticsRequest(_EmptyRequest):
    """
    HART command 95 request abstraction.
    """


class Command95_ReadDeviceCommunicationStatisticsResponse(HartResponse):
    """
    HART command 95 response abstraction.
    """

    def __init__(self):
        super(Command95_ReadDeviceCommunicationStatisticsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.stx_count = UBInt16()
        self.ack_count = UBInt16()
        self.back_count = UBInt16()


class Command96_ReadSynchronousActionRequest(ByteStruct):
    """
    HART command 96 request abstraction.
    """

    def __init__(self):
        super(Command96_ReadSynchronousActionRequest, self).__init__()

        self.action_number = UBInt8()


class Command96_ReadSynchronousActionResponse(HartResponse):
    """
    HART command 96 response abstraction.
    """

    def __init__(self):
        super(Command96_ReadSynchronousActionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.action_number = UBInt8()
        self.total_action = UBInt8()
        self.action_control = UBInt8()
        self.device_variable_code = UBInt8()
        self.command = UBInt16()
        self.trigger_date = Date(datetime.date(1900, 1, 1))
        self.trigger_time = Time(datetime.time())


class Command97_ConfigureSynchronousActionRequest(ByteStruct):
    """
    HART command 97 request abstraction.
    """

    def __init__(self):
        super(Command97_ConfigureSynchronousActionRequest, self).__init__()

        self.action_number = UBInt8()
        self.action_control = UBInt8()
        self.device_variable_code = UBInt8()
        self.command = UBInt16()
        self.trigger_date = Date(datetime.date(1900, 1, 1))
        self.trigger_time = Time(datetime.time())


class Command97_ConfigureSynchronousActionResponse(HartResponse):
    """
    HART command 97 response abstraction.
    """

    def __init__(self):
        super(Command97_ConfigureSynchronousActionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_sampling_time_adjusted=8,
            error_bad_trigger_action=9,
            error_invalid_date=10,
            error_invalid_time=11,
            error_invalid_device_variable=12,
            error_command_number_not_supported=13,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.action_number = UBInt8()
        self.action_control = UBInt8()
        self.device_variable_code = UBInt8()
        self.command = UBInt16()
        self.trigger_date = Date(datetime.date(1900, 1, 1))
        self.trigger_time = Time(datetime.time())


class Command98_ReadCommandActionRequest(ByteStruct):
    """
    HART command 98 request abstraction.
    """

    def __init__(self):
        super(Command98_ReadCommandActionRequest, self).__init__()

        self.action_number = UBInt8()


class Command98_ReadCommandActionResponse(HartResponse):
    """
    HART command 98 response abstraction.
    """

    def __init__(self):
        super(Command98_ReadCommandActionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.action_number = UBInt8()
        self.command = UBInt16()
        self.byte_count = UBInt8()
        self.data = Array(length=lambda self: self.up().byte_count.get() - 6,
                          prototype=UBInt8)


class Command99_ConfigureCommandActionRequest(ByteStruct):
    """
    HART command 99 request abstraction.
    """

    def __init__(self):
        super(Command99_ConfigureCommandActionRequest, self).__init__()

        self.action_number = UBInt8()
        self.command = UBInt16()
        self.byte_count = UBInt8()
        self.data = Array(length=lambda self: self.up().byte_count.get() - 4,
                          prototype=UBInt8)


class Command99_ConfigureCommandActionResponse(HartResponse):
    """
    HART command 99 response abstraction.
    """

    def __init__(self):
        super(Command99_ConfigureCommandActionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_bad_trigger_action=9,
            error_bad_command_data=10,
            error_command_number_not_supported=13,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.action_number = UBInt8()
        self.command = UBInt16()
        self.byte_count = UBInt8()
        self.data = Array(length=lambda self: self.up().byte_count.get() - 6,
                          prototype=UBInt8)


class Command100_WritePrimaryVariableAlarmCodeRequest(ByteStruct):
    """
    HART command 100 request abstraction.
    """

    def __init__(self):
        super(Command100_WritePrimaryVariableAlarmCodeRequest, self).__init__()

        self.primary_variable_alarm_selection_code = UBInt8()


class Command100_WritePrimaryVariableAlarmCodeResponse(HartResponse):
    """
    HART command 100 response abstraction.
    """

    def __init__(self):
        super(Command100_WritePrimaryVariableAlarmCodeResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_alarm_selection_code = UBInt8()


class Command101_ReadSubDeviceToBurstMessageMapRequest(ByteStruct):
    """
    HART command 101 request abstraction.
    """

    def __init__(self):
        super(Command101_ReadSubDeviceToBurstMessageMapRequest, self).__init__()
        self.burst_message = UBInt8()


class Command101_ReadSubDeviceToBurstMessageMapResponse(HartResponse):
    """
    HART command 101 response abstraction.
    """

    def __init__(self):
        super(Command101_ReadSubDeviceToBurstMessageMapResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.burst_message = UBInt8()
        self.subdevice_index = UBInt16()


class Command102_MapSubdeviceToBurstMessageRequest(ByteStruct):
    """
    HART command 102 request abstraction.
    """

    def __init__(self):
        super(Command102_MapSubdeviceToBurstMessageRequest, self).__init__()

        self.burst_message = UBInt8()
        self.subdevice_index = UBInt16()


class Command102_MapSubdeviceToBurstMessageResponse(HartResponse):
    """
    HART command 102 response abstraction.
    """

    def __init__(self):
        super(Command102_MapSubdeviceToBurstMessageResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_subdevice_index=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.burst_message = UBInt8()
        self.subdevice_index = UBInt16()


class Command103_WriteBurstPeriodRequest(ByteStruct):
    """
    HART command 103 request abstraction.
    """

    def __init__(self):
        super(Command103_WriteBurstPeriodRequest, self).__init__()

        self.burst_message = UBInt8()
        self.update_period = Time(datetime.time())
        self.maximum_update_period = Time(datetime.time())


class Command103_WriteBurstPeriodResponse(HartResponse):
    """
    HART command 103 response abstraction.
    """

    def __init__(self):
        super(Command103_WriteBurstPeriodResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_update_time_adjusted=8,
            error_invalid_burst_message=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.burst_message = UBInt8()
        self.update_period = Time(datetime.time())
        self.maximum_update_period = Time(datetime.time())


class Command104_WriteBurstTriggerRequest(ByteStruct):
    """
    HART command 104 request abstraction.
    """

    def __init__(self):
        super(Command104_WriteBurstTriggerRequest, self).__init__()

        self.burst_message = UBInt8()
        self.burst_trigger_mode_selection_code = UBInt8()
        self.device_variable_classification = UBInt8()
        self.units_code = UBInt8()
        self.trigger_level = BFloat()


class Command104_WriteBurstTriggerResponse(HartResponse):
    """
    HART command 104 response abstraction.
    """

    def __init__(self):
        super(Command104_WriteBurstTriggerResponse, self).__init__()
        self.response_code = ResponseCode(
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_burst_condition_conflict=8,
            error_invalid_burst_message=9,
            error_invalid_device_variable_classification=11,
            error_invalid_units_code=12,
            error_invalid_burst_trigger_mode_selection=13,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.burst_message = UBInt8()
        self.burst_trigger_mode_selection_code = UBInt8()
        self.device_variable_classification = UBInt8()
        self.units_code = UBInt8()
        self.trigger_level = BFloat()


class Command105_ReadBurstModeConfigurationRequest(ByteStruct):
    """
    HART command 105 request abstraction.
    """

    def __init__(self):
        super(Command105_ReadBurstModeConfigurationRequest, self).__init__()

        self.burst_message = UBInt8()


class Command105_ReadBurstModeConfigurationResponse(HartResponse):
    """
    HART command 105 response abstraction.
    """

    def __init__(self):
        super(Command105_ReadBurstModeConfigurationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_invalid_burst_message=9,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.burst_mode_control_code = UBInt8()
        self.command_number_expansion_flag = UBInt8()
        self.device_variable_code_slot0 = UBInt8()
        self.device_variable_code_slot1 = UBInt8()
        self.device_variable_code_slot2 = UBInt8()
        self.device_variable_code_slot3 = UBInt8()
        self.device_variable_code_slot4 = UBInt8()
        self.device_variable_code_slot5 = UBInt8()
        self.device_variable_code_slot6 = UBInt8()
        self.device_variable_code_slot7 = UBInt8()
        self.burst_message = UBInt8()
        self.maximum_burst_messages = UBInt8()
        self.extended_command_number = UBInt16()
        self.update_time = Time(datetime.time())
        self.maximum_update_time = Time(datetime.time())
        self.burst_trigger_mode_code = UBInt8()
        self.device_variable_classification = UBInt8()
        self.units_code = UBInt8()
        self.trigger_value = BFloat()


class Command106_FlushDelayedResponsesRequest(_EmptyRequest):
    """
    HART command 106 request abstraction.
    """


class Command106_FlushDelayedResponsesResponse(HartResponse):
    """
    HART command 106 response abstraction.
    """

    def __init__(self):
        super(Command106_FlushDelayedResponsesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_running_responses_not_flushed=8,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()


class Command107_WriteBurstDeviceVariablesRequest(ByteStruct):
    """
    HART command 107 request abstraction.
    """

    def __init__(self):
        super(Command107_WriteBurstDeviceVariablesRequest, self).__init__()

        self.device_variable_code_slot0 = UBInt8()
        self.device_variable_code_slot1 = UBInt8()
        self.device_variable_code_slot2 = UBInt8()
        self.device_variable_code_slot3 = UBInt8()
        self.device_variable_code_slot4 = UBInt8()
        self.device_variable_code_slot5 = UBInt8()
        self.device_variable_code_slot6 = UBInt8()
        self.device_variable_code_slot7 = UBInt8()
        self.burst_message = UBInt8()


class Command107_WriteBurstDeviceVariablesResponse(HartResponse):
    """
    HART command 107 response abstraction.
    """

    def __init__(self):
        super(Command107_WriteBurstDeviceVariablesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_burst_condition_conflict=8,
            error_invalid_burst_message=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.device_variable_code_slot0 = UBInt8()
        self.device_variable_code_slot1 = UBInt8()
        self.device_variable_code_slot2 = UBInt8()
        self.device_variable_code_slot3 = UBInt8()
        self.device_variable_code_slot4 = UBInt8()
        self.device_variable_code_slot5 = UBInt8()
        self.device_variable_code_slot6 = UBInt8()
        self.device_variable_code_slot7 = UBInt8()
        self.burst_message = UBInt8()


class Command108_WriteBurstModeCommandNumberRequest(ByteStruct):
    """
    HART command 108 request abstraction.
    """

    def __init__(self):
        super(Command108_WriteBurstModeCommandNumberRequest, self).__init__()

        self.command = UBInt16()
        self.burst_message = UBInt8()


class Command108_WriteBurstModeCommandNumberResponse(HartResponse):
    """
    HART command 108 response abstraction.
    """

    def __init__(self):
        super(Command108_WriteBurstModeCommandNumberResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_burst_condition_conflict=8,
            error_invalid_burst_message=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.burst_message = UBInt8()


class Command109_BurstModeControlRequest(ByteStruct):
    """
    HART command 109 request abstraction.
    """

    def __init__(self):
        super(Command109_BurstModeControlRequest, self).__init__()

        self.burst_mode_control_code = UBInt8()
        self.burst_message = UBInt8()


class Command109_BurstModeControlResponse(HartResponse):
    """
    HART command 109 response abstraction.
    """

    def __init__(self):
        super(Command109_BurstModeControlResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_update_period_increased=8,
            error_no_bandwidth_available=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.burst_mode_control_code = UBInt8()
        self.burst_message = UBInt8()


class Command110_ReadAllDynamicVariablesRequest(_EmptyRequest):
    """
    HART command 110 request abstraction.
    """


class Command110_ReadAllDynamicVariablesResponse(HartResponse):
    """
    HART command 110 response abstraction.
    """

    def __init__(self):
        super(Command110_ReadAllDynamicVariablesResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.dynamic_variables = Array(
            length=lambda self: self.up().byte_count.get() // 5,
            prototype=UnitAndVariable)


class Command111_TransferServiceControlRequest(ByteStruct):
    """
    HART command 111 request abstraction.
    """

    def __init__(self):
        super(Command111_TransferServiceControlRequest, self).__init__()

        self.port = UBInt8()
        self.function_code = UBInt8()
        self.maximum_segment_length = UBInt8()
        self.starting_master_byte_counter = UBInt16()


class Command111_TransferServiceControlResponse(HartResponse):
    """
    HART command 111 response abstraction.
    """

    def __init__(self):
        super(Command111_TransferServiceControlResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_segment_length_changed=8,
            error_port_not_found=9,
            error_port_in_use=10,
            error_maximum_ports_in_use=11,
            error_segment_length_too_small=12,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.port = UBInt8()
        self.function_code = UBInt8()
        self.maximum_segment_length = UBInt8()
        self.starting_master_byte_counter = UBInt16()
        self.starting_field_device_byte_counter = UBInt16()


class Command112_BlockTransferRequest(ByteStruct):
    """
    HART command 112 request abstraction.
    """

    def __init__(self):
        super(Command112_BlockTransferRequest, self).__init__()

        self.port = UBInt8()
        self.function_code = UBInt8()
        self.number_of_bytes_transferred = UBInt8()
        self.master_byte_count = UBInt16()
        self.slave_byte_count = UBInt16()
        self.data = Array(
            length=lambda self: self.up().byte_count.get() - 7,
            prototype=UBInt8)


class Command112_BlockTransferResponse(HartResponse):
    """
    HART command 112 response abstraction.
    """

    def __init__(self):
        super(Command112_BlockTransferResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_port_not_found=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.port = UBInt8()
        self.function_code = UBInt8()
        self.number_of_bytes_transferred = UBInt8()
        self.master_byte_count = UBInt16()
        self.slave_byte_count = UBInt16()
        self.data = Array(
            length=lambda self: self.up().byte_count.get() - 9,
            prototype=UBInt8)


class Command113_CatchDeviceVariableRequest(ByteStruct):
    """
    HART command 113 request abstraction.
    """

    def __init__(self):
        super(Command113_CatchDeviceVariableRequest, self).__init__()

        self.destination_device_variable = UBInt8()
        self.capture_mode_code = UBInt8()
        self.source_slave_expanded_device_type_code = UBInt16()
        self.source_slave_device_id = BitStruct()
        self.source_slave_device_id.value = BitField(length=24)
        self.command_number_expansion_flag = UBInt8()
        self.source_slot_number = UBInt8()
        self.shed_time = BFloat()
        self.source_command_number = UBInt16()


class Command113_CatchDeviceVariableResponse(HartResponse):
    """
    HART command 113 response abstraction.
    """

    def __init__(self):
        super(Command113_CatchDeviceVariableResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_catch_data_from_back_message=8,
            error_invalid_capture_mode_code=10,
            error_invalid_slot_number=11,
            error_command_number_not_supported=12,
            error_access_restricted=16,
            error_invalid_device_variable_index=17,
            error_device_variable_index_not_allowed=19,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.destination_device_variable = UBInt8()
        self.capture_mode_code = UBInt8()
        self.source_slave_address = BitStruct()
        self.source_slave_address.value = BitField(length=40)
        self.command_number_expansion_flag = UBInt8()
        self.source_slot_number = UBInt8()
        self.shed_time = BFloat()
        self.source_command_number = UBInt16()


class Command114_ReadCaughtDeviceVariableRequest(ByteStruct):
    """
    HART command 114 request abstraction.
    """

    def __init__(self):
        super(Command114_ReadCaughtDeviceVariableRequest, self).__init__()

        self.destination_device_variable = UBInt8()


class Command114_ReadCaughtDeviceVariableResponse(HartResponse):
    """
    HART command 114 response abstraction.
    """

    def __init__(self):
        super(Command114_ReadCaughtDeviceVariableResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_invalid_device_variable_index=17,
            error_device_variable_index_not_allowed=19,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.destination_device_variable = UBInt8()
        self.capture_mode_code = UBInt8()
        self.source_slave_address = BitStruct()
        self.source_slave_address.value = BitField(length=40)
        self.LSB_of_source_command_number = UBInt8()
        self.source_slot_number = UBInt8()
        self.shed_time = BFloat()
        self.source_command_number = UBInt16()


class Command115_ReadEventNotificationSummaryRequest(ByteStruct):
    """
    HART command 115 request abstraction.
    """

    def __init__(self):
        super(Command115_ReadEventNotificationSummaryRequest, self).__init__()

        self.event_specification_number = UBInt8()


class Command115_ReadEventNotificationSummaryResponse(HartResponse):
    """
    HART command 115 response abstraction.
    """

    def __init__(self):
        super(Command115_ReadEventNotificationSummaryResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.event_specification_number = UBInt8()
        self.number_of_events_supported = UBInt8()
        self.event_status_notification_control = BitStruct()
        self.event_status_notification_control.event_status = Nibble()
        self.event_status_notification_control.notification_control_code = Nibble()
        self.first_event_trigger_time = Time(datetime.time())
        self.event_notification_retry_time = Time(datetime.time())
        self.maximum_update_time = Time(datetime.time())
        self.event_debounce_interval = Time(datetime.time())
        self.device_status_mask = BitStruct()
        self.device_status_mask.value = BitField(length=8)
        self.event_mask = Array(
            length=lambda self: self.up().byte_count.get() - 22,
            prototype=UBInt8)


class Command116_WriteEventNotificationBitMaskRequest(ByteStruct):
    """
    HART command 116 request abstraction.
    """

    def __init__(self):
        super(Command116_WriteEventNotificationBitMaskRequest, self).__init__()

        self.event_specification_number = UBInt8()
        self.device_status_mask = BitStruct()
        self.device_status_mask.value = BitField(length=8)
        self.event_mask = Array(
            length=lambda self: self.up().byte_count.get() - 2,
            prototype=UBInt8)


class Command116_WriteEventNotificationBitMaskResponse(HartResponse):
    """
    HART command 116 response abstraction.
    """

    def __init__(self):
        super(Command116_WriteEventNotificationBitMaskResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.event_specification_number = UBInt8()
        self.device_status_mask = BitStruct()
        self.device_status_mask.value = BitField(length=8)
        self.event_mask = Array(
            length=lambda self: self.up().byte_count.get() - 4,
            prototype=UBInt8)


class Command117_WriteEventNotificationTimingRequest(ByteStruct):
    """
    HART command 117 request abstraction.
    """

    def __init__(self):
        super(Command117_WriteEventNotificationTimingRequest, self).__init__()

        self.event_specification_number = UBInt8()
        self.event_notification_retry_time = Time(datetime.time())
        self.maximum_update_time = Time(datetime.time())
        self.event_debounce_interval = Time(datetime.time())


class Command117_WriteEventNotificationTimingResponse(HartResponse):
    """
    HART command 117 response abstraction.
    """

    def __init__(self):
        super(Command117_WriteEventNotificationTimingResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_update_period_or_debounce_interval_adjusted=8,
            error_invalid_event_specification_number=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.event_specification_number = UBInt8()
        self.event_notification_retry_time = Time(datetime.time())
        self.maximum_update_time = Time(datetime.time())
        self.event_debounce_interval = Time(datetime.time())


class Command118_EventNotificationControlRequest(ByteStruct):
    """
    HART command 118 request abstraction.
    """

    def __init__(self):
        super(Command118_EventNotificationControlRequest, self).__init__()

        self.event_specification_number = UBInt8()
        self.event_notification_control_code = UBInt8()


class Command118_EventNotificationControlResponse(HartResponse):
    """
    HART command 118 response abstraction.
    """

    def __init__(self):
        super(Command118_EventNotificationControlResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_update_time_adjusted=8,
            error_no_bandwidth_available=9,
            warning_update_rate_uncertain=14,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.event_specification_number = UBInt8()
        self.event_notification_control_code = UBInt8()


class Command119_AcknowledgeEventNotificationRequest(ByteStruct):
    """
    HART command 119 request abstraction.
    """

    def __init__(self):
        super(Command119_AcknowledgeEventNotificationRequest, self).__init__()

        self.event_specification_number = UBInt8()
        self.first_event_trigger_time = Time(datetime.time())
        self.configuration_change_counter = UBInt16()
        self.device_status = BitStruct()
        self.device_status.value = BitField(length=8)
        self.command48_data = Array(
            length=lambda self: self.up().byte_count.get() - 8,
            prototype=UBInt8)


class Command119_AcknowledgeEventNotificationResponse(HartResponse):
    """
    HART command 119 response abstraction.
    """

    def __init__(self):
        super(Command119_AcknowledgeEventNotificationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_not_all_events_cleared=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.event_specification_number = UBInt8()
        self.first_event_trigger_time = Time(datetime.time())
        self.configuration_change_counter = UBInt16()
        self.device_status = BitStruct()
        self.device_status.value = BitField(length=8)
        self.command48_data = Array(
            length=lambda self: self.up().byte_count.get() - 10,
            prototype=UBInt8)


class _EmptyExtendedRequest(ByteStruct):
    """
    HART emptry extended request abstraction.
    """

    def __init__(self):
        super(_EmptyExtendedRequest, self).__init__()
        self.command = UBInt16()


class Command512_ReadCountryCodeRequest(_EmptyExtendedRequest):
    """
    HART command 512 request abstraction.
    """


class Command512_ReadCountryCodeResponse(HartResponse):
    """
    HART command 512 response abstraction.
    """

    def __init__(self):
        super(Command512_ReadCountryCodeResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.country_code = String(length=2)
        self.si_units_only = UBInt8()


class Command513_WriteCountryCodeRequest(ByteStruct):
    """
    HART command 513 request abstraction.
    """

    def __init__(self):
        super(Command513_WriteCountryCodeRequest, self).__init__()

        self.command = UBInt16()
        self.country_code = String(length=2)
        self.si_units_only = UBInt8()


class Command513_WriteCountryCodeResponse(HartResponse):
    """
    HART command 513 response abstraction.
    """

    def __init__(self):
        super(Command513_WriteCountryCodeResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_si_units_restriction_failed=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.country_code = String(length=2)
        self.si_units_only = UBInt8()


class Command514_RegisterEventManagerRequest(ByteStruct):
    """
    HART command 514 request abstraction.
    """

    def __init__(self):
        super(Command514_RegisterEventManagerRequest, self).__init__()

        self.command = UBInt16()
        self.event_manager_registration_control_code = UBInt8()


class Command514_RegisterEventManagerResponse(HartResponse):
    """
    HART command 514 response abstraction.
    """

    def __init__(self):
        super(Command514_RegisterEventManagerResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.event_manager_registration_control_code = UBInt8()


class Command515_ReadEventManagerRegistrationStatusRequest(_EmptyExtendedRequest):
    """
    HART command 515 request abstraction.
    """


class Command515_ReadEventManagerRegistrationStatusResponse(HartResponse):
    """
    HART command 515 response abstraction.
    """

    def __init__(self):
        super(Command515_ReadEventManagerRegistrationStatusResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.event_manager_registration_status = UBInt8()


class Command516_ReadDeviceLocationRequest(_EmptyExtendedRequest):
    """
    HART command 516 request abstraction.
    """


class Command516_ReadDeviceLocationResponse(HartResponse):
    """
    HART command 516 response abstraction.
    """

    def __init__(self):
        super(Command516_ReadDeviceLocationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.latitude = BFloat()
        self.longitude = BFloat()
        self.location_method = UBInt8()
        self.altitude = BFloat()


class Command517_WriteDeviceLocationRequest(ByteStruct):
    """
    HART command 517 request abstraction.
    """

    def __init__(self):
        super(Command517_WriteDeviceLocationRequest, self).__init__()

        self.command = UBInt16()
        self.latitude = BFloat()
        self.longitude = BFloat()
        self.location_method = UBInt8()
        self.altitude = BFloat()


class Command517_WriteDeviceLocationResponse(HartResponse):
    """
    HART command 517 response abstraction.
    """

    def __init__(self):
        super(Command517_WriteDeviceLocationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_passed_parameter_too_large=3,
            error_passed_parameter_too_small=4,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.latitude = BFloat()
        self.longitude = BFloat()
        self.location_method = UBInt8()
        self.altitude = BFloat()


class Command518_ReadLocationDescriptionRequest(_EmptyExtendedRequest):
    """
    HART command 518 request abstraction.
    """


class Command518_ReadLocationDescriptionResponse(HartResponse):
    """
    HART command 518 response abstraction.
    """

    def __init__(self):
        super(Command518_ReadLocationDescriptionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.location_description = String(length=32)


class Command519_WriteLocationDescriptionRequest(ByteStruct):
    """
    HART command 519 request abstraction.
    """

    def __init__(self):
        super(Command519_WriteLocationDescriptionRequest, self).__init__()

        self.command = UBInt16()
        self.location_description = String(length=32)


class Command519_WriteLocationDescriptionResponse(HartResponse):
    """
    HART command 519 response abstraction.
    """

    def __init__(self):
        super(Command519_WriteLocationDescriptionResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.location_description = String(length=32)


class Command520_ReadProcessUnitTagRequest(_EmptyExtendedRequest):
    """
    HART command 520 request abstraction.
    """


class Command520_ReadProcessUnitTagResponse(HartResponse):
    """
    HART command 520 response abstraction.
    """

    def __init__(self):
        super(Command520_ReadProcessUnitTagResponse, self).__init__()
        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.process_unit_tag = String(length=32)


class Command521_WriteProcessUnitTagRequest(ByteStruct):
    """
    HART command 521 request abstraction.
    """

    def __init__(self):
        super(Command521_WriteProcessUnitTagRequest, self).__init__()

        self.command = UBInt16()
        self.process_unit_tag = String(length=32)


class Command521_WriteProcessUnitTagResponse(HartResponse):
    """
    HART command 521 response abstraction.
    """

    def __init__(self):
        super(Command521_WriteProcessUnitTagResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.process_unit_tag = String(length=32)


class Command522_WriteVolumetricFlowClassificationRequest(ByteStruct):
    """
    HART command 522 request abstraction.
    """

    def __init__(self):
        super(Command522_WriteVolumetricFlowClassificationRequest, self).__init__()

        self.command = UBInt16()
        self.device_variable_code = UBInt8()
        self.classification_code = UBInt8()


class Command522_WriteVolumetricFlowClassificationResponse(HartResponse):
    """
    HART command 522 response abstraction.
    """

    def __init__(self):
        super(Command522_WriteVolumetricFlowClassificationResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.device_variable_code = UBInt8()
        self.classification_code = UBInt8()


class Command523_ReadCondensedStatusMappingArrayRequest(ByteStruct):
    """
    HART command 523 request abstraction.
    """

    def __init__(self):
        super(Command523_ReadCondensedStatusMappingArrayRequest, self).__init__()

        self.command = UBInt16()
        self.starting_status_map_index = UBInt8()
        self.number_of_entries_to_read = UBInt8()


class Command523_ReadCondensedStatusMappingArrayResponse(HartResponse):
    """
    HART command 523 response abstraction.
    """

    def __init__(self):
        super(Command523_ReadCondensedStatusMappingArrayResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            warning_set_to_nearest_possible_value=8)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.starting_status_map_index = UBInt8()
        self.number_of_entries_returned = UBInt8()
        self.status_map_codes = BitStruct()
        self.status_map_codes.value = Array(
            length=lambda self: self.up().number_of_entries_returned.get(),
            prototype=Nibble)


class Command524_WriteCondensedStatusMappingRequest(ByteStruct):
    """
    HART command 524 request abstraction.
    """

    def __init__(self):
        super(Command524_WriteCondensedStatusMappingRequest, self).__init__()

        self.command = UBInt16()
        self.starting_status_map_index = UBInt8()
        self.number_of_entries_to_write = UBInt8()
        self.status_map_codes = BitStruct()
        self.status_map_codes.value = Array(
            length=lambda self: self.up().number_of_entries_to_write.get(),
            prototype=Nibble)


class Command524_WriteCondensedStatusMappingResponse(HartResponse):
    """
    HART command 524 response abstraction.
    """

    def __init__(self):
        super(Command524_WriteCondensedStatusMappingResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            warning_set_to_nearest_possible_value=8,
            error_starting_status_map_index_must_be_even_value=9,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.starting_status_map_index = UBInt8()
        self.number_of_entries_returned = UBInt8()
        self.status_map_codes = BitStruct()
        self.status_map_codes.value = Array(
            length=lambda self: self.up().number_of_entries_returned.get(),
            prototype=Nibble)


class Command525_ResetCondensedStatusMapRequest(_EmptyExtendedRequest):
    """
    HART command 525 request abstraction.
    """


class Command525_ResetCondensedStatusMapResponse(HartResponse):
    """
    HART command 525 response abstraction.
    """

    def __init__(self):
        super(Command525_ResetCondensedStatusMapResponse, self).__init__()
        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()


class Command526_WriteStatusSimulationModeRequest(ByteStruct):
    """
    HART command 526 request abstraction.
    """

    def __init__(self):
        super(Command526_WriteStatusSimulationModeRequest, self).__init__()

        self.command = UBInt16()
        self.status_simulation_mode_code = UBInt8()


class Command526_WriteStatusSimulationModeResponse(HartResponse):
    """
    HART command 526 response abstraction.
    """

    def __init__(self):
        super(Command526_WriteStatusSimulationModeResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.status_simulation_mode_code = UBInt8()


class Command527_SimulateStatusBitRequest(ByteStruct):
    """
    HART command 527 request abstraction.
    """

    def __init__(self):
        super(Command527_SimulateStatusBitRequest, self).__init__()

        self.command = UBInt16()
        self.simulated_bit_index = UBInt8()
        self.simulated_value_code = UBInt8()


class Command527_SimulateStatusBitResponse(HartResponse):
    """
    HART command 527 response abstraction.
    """

    def __init__(self):
        super(Command527_SimulateStatusBitResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.simulated_bit_index = UBInt8()
        self.simulated_value_code = UBInt8()


class Command528_ReadSubdeviceAssignmentListInformationRequest(_EmptyExtendedRequest):
    """
    HART command 528 request abstraction.
    """


class Command528_ReadSubdeviceAssignmentListInformationResponse(HartResponse):
    """
    HART command 528 response abstraction.
    """

    def __init__(self):
        super(Command528_ReadSubdeviceAssignmentListInformationResponse, self).__init__()
        self.response_code = ResponseCode()
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.number_of_subdevices = UBInt16()
        self.maximum_number_of_subdevices = UBInt16()
        self.subdevice_assignment_status = UBInt8()


class Command529_ReadSubdeviceAssignmentRequest(ByteStruct):
    """
    HART command 529 request abstraction.
    """

    def __init__(self):
        super(Command529_ReadSubdeviceAssignmentRequest, self).__init__()

        self.command = UBInt16()
        self.subdevice_index = UBInt16()


class Command529_ReadSubdeviceAssignmentResponse(HartResponse):
    """
    HART command 529 response abstraction.
    """

    def __init__(self):
        super(Command529_ReadSubdeviceAssignmentResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.subdevice_index = UBInt16()
        self.subdevice_assignment_status = UBInt8()
        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.manufacturer_id = UBInt16()
        self.expanded_device_type_code = UBInt16()
        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)
        self.long_tag = String(length=32)
        self.device_revision = UBInt8()


class Command530_WriteSubdeviceAssingmentRequest(ByteStruct):
    """
    HART command 530 request abstraction.
    """

    def __init__(self):
        super(Command530_WriteSubdeviceAssingmentRequest, self).__init__()

        self.command = UBInt16()
        self.subdevice_index = UBInt16()
        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.manufacturer_id = UBInt16()
        self.expanded_device_type_code = UBInt16()
        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)
        self.long_tag = String(length=32)
        self.device_revision = UBInt8()


class Command530_WriteSubdeviceAssingmentResponse(HartResponse):
    """
    HART command 530 response abstraction.
    """

    def __init__(self):
        super(Command530_WriteSubdeviceAssingmentResponse, self).__init__()
        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_subdevice_already_exists=9,
            error_bad_device_identity=10,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.subdevice_index = UBInt16()
        self.io_card = UBInt8()
        self.channel = UBInt8()
        self.manufacturer_id = UBInt16()
        self.expanded_device_type_code = UBInt16()
        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)
        self.long_tag = String(length=32)
        self.device_revision = UBInt8()


class Command531_TransferLiveListToAssignmentListRequest(ByteStruct):
    """
    HART command 531 request abstraction.
    """

    def __init__(self):
        super(Command531_TransferLiveListToAssignmentListRequest, self).__init__()

        self.command = UBInt16()
        self.subdevice_assignment_transfer_code = UBInt8()


class Command531_TransferLiveListToAssignmentListResponse(HartResponse):
    """
    HART command 531 response abstraction.
    """

    def __init__(self):
        super(Command531_TransferLiveListToAssignmentListResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            warning_duplicate_id=8,
            error_access_restricted=16,
            error_busy=32,
            error_delayed_response_initiated=33,
            error_delayed_response_running=34,
            error_delayed_response_dead=35,
            error_delayed_response_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.subdevice_assignment_transfer_code = UBInt8()
        self.number_of_subdevices = UBInt16()


class Command532_ReadClientSubscriptionSummaryRequest(ByteStruct):
    """
    HART command 532 request abstraction.
    """

    def __init__(self):
        super(Command532_ReadClientSubscriptionSummaryRequest, self).__init__()

        self.command = UBInt16()
        self.target_field_device_unique_id = BitStruct()
        self.target_field_device_unique_id.value = BitField(length=40)


class Command532_ReadClientSubscriptionSummaryResponse(HartResponse):
    """
    HART command 532 response abstraction.
    """

    def __init__(self):
        super(Command532_ReadClientSubscriptionSummaryResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_target_unique_id_must_be_broadcast_address=9,
            error_unknown_unique_id=65)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.target_field_device_unique_id = BitStruct()
        self.target_field_device_unique_id.value = BitField(length=40)
        self.client_subscription_flags = UBInt16()


class Command533_WriteClientSubscriptionFlagsRequest(ByteStruct):
    """
    HART command 533 request abstraction.
    """

    def __init__(self):
        super(Command533_WriteClientSubscriptionFlagsRequest, self).__init__()

        self.command = UBInt16()
        self.target_field_device_unique_id = BitStruct()
        self.target_field_device_unique_id.value = BitField(length=40)
        self.client_subscription_flags = UBInt16()


class Command533_WriteClientSubscriptionFlagsResponse(HartResponse):
    """
    HART command 533 response abstraction.
    """

    def __init__(self):
        super(Command533_WriteClientSubscriptionFlagsResponse, self).__init__()
        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_target_unique_id_must_be_broadcast_address=9,
            error_unknown_unique_id=65)
        self.field_device_status = FieldDeviceStatus()

        self.command = UBInt16()
        self.target_field_device_unique_id = BitStruct()
        self.target_field_device_unique_id.value = BitField(length=40)
        self.client_subscription_flags = UBInt16()


def requests_generator():
    yield 33, Lazy(Command33_ReadDeviceVariablesRequest)
    yield 34, Lazy(Command34_WritePrimaryVariableDampingValueRequest)
    yield 35, Lazy(Command35_WritePrimaryVariableRangeValuesRequest)
    yield 36, Lazy(Command36_SetPrimaryVariableUpperRangeValueRequest)
    yield 37, Lazy(Command37_SetPrimaryVariableLowerRangeValueRequest)
    yield 39, Lazy(Command39_EEPROMControlRequest)
    yield 40, Lazy(Command40_EnterExitFixedCurrentModeRequest)
    yield 41, Lazy(Command41_PerformSelfTestRequest)
    yield 42, Lazy(Command42_PerformDeviceResetRequest)
    yield 43, Lazy(Command43_SetPrimaryVariableZeroRequest)
    yield 44, Lazy(Command44_WritePrimaryVariableUnitsRequest)
    yield 45, Lazy(Command45_TrimLoopCurrentZeroRequest)
    yield 46, Lazy(Command46_TrimLoopCurrentGainRequest)
    yield 47, Lazy(Command47_WritePrimaryVariableTransferFunctionRequest)
    yield 48, Lazy(Command48_ReadAdditionalDeviceStatusRequest)
    yield 49, Lazy(Command49_WritePrimaryVariableTransducerSerialNumberRequest)
    yield 50, Lazy(Command50_ReadDynamicVariableAssignmentsRequest)
    yield 51, Lazy(Command51_WriteDynamicVariableAssignmentsRequest)
    yield 52, Lazy(Command52_SetDeviceVariableZeroRequest)
    yield 53, Lazy(Command53_WriteDeviceVariableUnitsRequest)
    yield 54, Lazy(Command54_ReadDeviceVariableInformationRequest)
    yield 55, Lazy(Command55_WriteDeviceVariableDampingValueRequest)
    yield 56, Lazy(Command56_WriteDeviceVariableTransducerSerialNumberRequest)
    yield 57, Lazy(Command57_ReadUnitTagDescriptorDateRequest)
    yield 58, Lazy(Command58_WriteUnitTagDescriptorDateRequest)
    yield 59, Lazy(Command59_WriteNumberOfResponsePreamblesRequest)
    yield 60, Lazy(Command60_ReadAnalogChannelAndPercentOfRangeRequest)
    yield 61, Lazy(Command61_ReadDynamicVariablesAndPrimaryVariableAnalogChannelRequest)
    yield 62, Lazy(Command62_ReadAnalogChannelsRequest)
    yield 63, Lazy(Command63_ReadAnalogChannelInformationRequest)
    yield 64, Lazy(Command64_WriteAnalogChannelAdditionalDampingValueRequest)
    yield 65, Lazy(Command65_WriteAnalogChannelRangeValuesRequest)
    yield 66, Lazy(Command66_EnterExitFixedAnalogChannelModeRequest)
    yield 67, Lazy(Command67_TrimAnalogChannelZeroRequest)
    yield 68, Lazy(Command68_TrimAnalogChannelGainRequest)
    yield 69, Lazy(Command69_WriteAnalogChannelTransferFunctionRequest)
    yield 70, Lazy(Command70_ReadAnalogChannelEndpointValuesRequest)
    yield 71, Lazy(Command71_LockDeviceRequest)
    yield 72, Lazy(Command72_SquawkRequest)
    yield 73, Lazy(Command73_FindDeviceRequest)
    yield 74, Lazy(Command74_ReadIOSystemCapabilitiesRequest)
    yield 75, Lazy(Command75_PollSubDeviceRequest)
    yield 76, Lazy(Command76_ReadLockDeviceStateRequest)
    yield 77, Lazy(Command77_SendCommandToSubDeviceRequest)
    yield 78, Lazy(Command78_ReadAggregatedCommandsRequest)
    yield 79, Lazy(Command79_WriteDeviceVariableRequest)
    yield 80, Lazy(Command80_ReadDeviceVariableTrimPointsRequest)
    yield 81, Lazy(Command81_ReadDeviceVariableTrimGuidelinesRequest)
    yield 82, Lazy(Command82_WriteDeviceVariableTrimPointRequest)
    yield 83, Lazy(Command83_ResetDeviceVariableTrimRequest)
    yield 84, Lazy(Command84_ReadSubDeviceIdentitySummaryRequest)
    yield 85, Lazy(Command85_ReadIOChannelStatisticsRequest)
    yield 86, Lazy(Command86_ReadSubDeviceStatisticsRequest)
    yield 87, Lazy(Command87_WriteIOSystemMasterModeRequest)
    yield 88, Lazy(Command88_WriteIOSystemRetryCountRequest)
    yield 89, Lazy(Command89_SetRealTimeClockRequest)
    yield 90, Lazy(Command90_ReadRealTimeClockRequest)
    yield 91, Lazy(Command91_ReadTrendConfigurationRequest)
    yield 92, Lazy(Command92_WriteTrendConfigurationRequest)
    yield 93, Lazy(Command93_ReadTrendRequest)
    yield 94, Lazy(Command94_ReadIOSystemClientSideCommunicationStatisticsRequest)
    yield 95, Lazy(Command95_ReadDeviceCommunicationStatisticsRequest)
    yield 96, Lazy(Command96_ReadSynchronousActionRequest)
    yield 97, Lazy(Command97_ConfigureSynchronousActionRequest)
    yield 98, Lazy(Command98_ReadCommandActionRequest)
    yield 99, Lazy(Command99_ConfigureCommandActionRequest)
    yield 100, Lazy(Command100_WritePrimaryVariableAlarmCodeRequest)
    yield 101, Lazy(Command101_ReadSubDeviceToBurstMessageMapRequest)
    yield 102, Lazy(Command102_MapSubdeviceToBurstMessageRequest)
    yield 103, Lazy(Command103_WriteBurstPeriodRequest)
    yield 104, Lazy(Command104_WriteBurstTriggerRequest)
    yield 105, Lazy(Command105_ReadBurstModeConfigurationRequest)
    yield 106, Lazy(Command106_FlushDelayedResponsesRequest)
    yield 107, Lazy(Command107_WriteBurstDeviceVariablesRequest)
    yield 108, Lazy(Command108_WriteBurstModeCommandNumberRequest)
    yield 109, Lazy(Command109_BurstModeControlRequest)
    yield 110, Lazy(Command110_ReadAllDynamicVariablesRequest)
    yield 111, Lazy(Command111_TransferServiceControlRequest)
    yield 112, Lazy(Command112_BlockTransferRequest)
    yield 113, Lazy(Command113_CatchDeviceVariableRequest)
    yield 114, Lazy(Command114_ReadCaughtDeviceVariableRequest)
    yield 115, Lazy(Command115_ReadEventNotificationSummaryRequest)
    yield 116, Lazy(Command116_WriteEventNotificationBitMaskRequest)
    yield 117, Lazy(Command117_WriteEventNotificationTimingRequest)
    yield 118, Lazy(Command118_EventNotificationControlRequest)
    yield 119, Lazy(Command119_AcknowledgeEventNotificationRequest)
    yield 512, Lazy(Command512_ReadCountryCodeRequest)
    yield 513, Lazy(Command513_WriteCountryCodeRequest)
    yield 514, Lazy(Command514_RegisterEventManagerRequest)
    yield 515, Lazy(Command515_ReadEventManagerRegistrationStatusRequest)
    yield 516, Lazy(Command516_ReadDeviceLocationRequest)
    yield 517, Lazy(Command517_WriteDeviceLocationRequest)
    yield 518, Lazy(Command518_ReadLocationDescriptionRequest)
    yield 519, Lazy(Command519_WriteLocationDescriptionRequest)
    yield 520, Lazy(Command520_ReadProcessUnitTagRequest)
    yield 521, Lazy(Command521_WriteProcessUnitTagRequest)
    yield 522, Lazy(Command522_WriteVolumetricFlowClassificationRequest)
    yield 523, Lazy(Command523_ReadCondensedStatusMappingArrayRequest)
    yield 524, Lazy(Command524_WriteCondensedStatusMappingRequest)
    yield 525, Lazy(Command525_ResetCondensedStatusMapRequest)
    yield 526, Lazy(Command526_WriteStatusSimulationModeRequest)
    yield 527, Lazy(Command527_SimulateStatusBitRequest)
    yield 528, Lazy(Command528_ReadSubdeviceAssignmentListInformationRequest)
    yield 529, Lazy(Command529_ReadSubdeviceAssignmentRequest)
    yield 530, Lazy(Command530_WriteSubdeviceAssingmentRequest)
    yield 531, Lazy(Command531_TransferLiveListToAssignmentListRequest)
    yield 532, Lazy(Command532_ReadClientSubscriptionSummaryRequest)
    yield 533, Lazy(Command533_WriteClientSubscriptionFlagsRequest)


def responses_generator():
    yield 33, Lazy(Command33_ReadDeviceVariablesResponse)
    yield 34, Lazy(Command34_WritePrimaryVariableDampingValueResponse)
    yield 35, Lazy(Command35_WritePrimaryVariableRangeValuesResponse)
    yield 36, Lazy(Command36_SetPrimaryVariableUpperRangeValueResponse)
    yield 37, Lazy(Command37_SetPrimaryVariableLowerRangeValueResponse)
    yield 39, Lazy(Command39_EEPROMControlResponse)
    yield 40, Lazy(Command40_EnterExitFixedCurrentModeResponse)
    yield 41, Lazy(Command41_PerformSelfTestResponse)
    yield 42, Lazy(Command42_PerformDeviceResetResponse)
    yield 43, Lazy(Command43_SetPrimaryVariableZeroResponse)
    yield 44, Lazy(Command44_WritePrimaryVariableUnitsResponse)
    yield 45, Lazy(Command45_TrimLoopCurrentZeroResponse)
    yield 46, Lazy(Command46_TrimLoopCurrentGainResponse)
    yield 47, Lazy(Command47_WritePrimaryVariableTransferFunctionResponse)
    yield 48, Lazy(Command48_ReadAdditionalDeviceStatusResponse)
    yield 49, Lazy(Command49_WritePrimaryVariableTransducerSerialNumberResponse)
    yield 50, Lazy(Command50_ReadDynamicVariableAssignmentsResponse)
    yield 51, Lazy(Command51_WriteDynamicVariableAssignmentsResponse)
    yield 52, Lazy(Command52_SetDeviceVariableZeroResponse)
    yield 53, Lazy(Command53_WriteDeviceVariableUnitsResponse)
    yield 54, Lazy(Command54_ReadDeviceVariableInformationResponse)
    yield 55, Lazy(Command55_WriteDeviceVariableDampingValueResponse)
    yield 56, Lazy(Command56_WriteDeviceVariableTransducerSerialNumberResponse)
    yield 57, Lazy(Command57_ReadUnitTagDescriptorDateResponse)
    yield 58, Lazy(Command58_WriteUnitTagDescriptorDateResponse)
    yield 59, Lazy(Command59_WriteNumberOfResponsePreamblesResponse)
    yield 60, Lazy(Command60_ReadAnalogChannelAndPercentOfRangeResponse)
    yield 61, Lazy(Command61_ReadDynamicVariablesAndPrimaryVariableAnalogChannelResponse)
    yield 62, Lazy(Command62_ReadAnalogChannelsResponse)
    yield 63, Lazy(Command63_ReadAnalogChannelInformationResponse)
    yield 64, Lazy(Command64_WriteAnalogChannelAdditionalDampingValueResponse)
    yield 65, Lazy(Command65_WriteAnalogChannelRangeValuesResponse)
    yield 66, Lazy(Command66_EnterExitFixedAnalogChannelModeResponse)
    yield 67, Lazy(Command67_TrimAnalogChannelZeroResponse)
    yield 68, Lazy(Command68_TrimAnalogChannelGainResponse)
    yield 69, Lazy(Command69_WriteAnalogChannelTransferFunctionResponse)
    yield 70, Lazy(Command70_ReadAnalogChannelEndpointValuesResponse)
    yield 71, Lazy(Command71_LockDeviceResponse)
    yield 72, Lazy(Command72_SquawkResponse)
    yield 73, Lazy(Command73_FindDeviceResponse)
    yield 74, Lazy(Command74_ReadIOSystemCapabilitiesResponse)
    yield 75, Lazy(Command75_PollSubDeviceResponse)
    yield 76, Lazy(Command76_ReadLockDeviceStateResponse)
    yield 77, Lazy(Command77_SendCommandToSubDeviceResponse)
    yield 78, Lazy(Command78_ReadAggregatedCommandsResponse)
    yield 79, Lazy(Command79_WriteDeviceVariableResponse)
    yield 80, Lazy(Command80_ReadDeviceVariableTrimPointsResponse)
    yield 81, Lazy(Command81_ReadDeviceVariableTrimGuidelinesResponse)
    yield 82, Lazy(Command82_WriteDeviceVariableTrimPointResponse)
    yield 83, Lazy(Command83_ResetDeviceVariableTrimResponse)
    yield 84, Lazy(Command84_ReadSubDeviceIdentitySummaryResponse)
    yield 85, Lazy(Command85_ReadIOChannelStatisticsResponse)
    yield 86, Lazy(Command86_ReadSubDeviceStatisticsResponse)
    yield 87, Lazy(Command87_WriteIOSystemMasterModeResponse)
    yield 88, Lazy(Command88_WriteIOSystemRetryCountResponse)
    yield 89, Lazy(Command89_SetRealTimeClockResponse)
    yield 90, Lazy(Command90_ReadRealTimeClockResponse)
    yield 91, Lazy(Command91_ReadTrendConfigurationResponse)
    yield 92, Lazy(Command92_WriteTrendConfigurationResponse)
    yield 93, Lazy(Command93_ReadTrendResponse)
    yield 94, Lazy(Command94_ReadIOSystemClientSideCommunicationStatisticsResponse)
    yield 95, Lazy(Command95_ReadDeviceCommunicationStatisticsResponse)
    yield 96, Lazy(Command96_ReadSynchronousActionResponse)
    yield 97, Lazy(Command97_ConfigureSynchronousActionResponse)
    yield 98, Lazy(Command98_ReadCommandActionResponse)
    yield 99, Lazy(Command99_ConfigureCommandActionResponse)
    yield 100, Lazy(Command100_WritePrimaryVariableAlarmCodeResponse)
    yield 101, Lazy(Command101_ReadSubDeviceToBurstMessageMapResponse)
    yield 102, Lazy(Command102_MapSubdeviceToBurstMessageResponse)
    yield 103, Lazy(Command103_WriteBurstPeriodResponse)
    yield 104, Lazy(Command104_WriteBurstTriggerResponse)
    yield 105, Lazy(Command105_ReadBurstModeConfigurationResponse)
    yield 106, Lazy(Command106_FlushDelayedResponsesResponse)
    yield 107, Lazy(Command107_WriteBurstDeviceVariablesResponse)
    yield 108, Lazy(Command108_WriteBurstModeCommandNumberResponse)
    yield 109, Lazy(Command109_BurstModeControlResponse)
    yield 110, Lazy(Command110_ReadAllDynamicVariablesResponse)
    yield 111, Lazy(Command111_TransferServiceControlResponse)
    yield 112, Lazy(Command112_BlockTransferResponse)
    yield 113, Lazy(Command113_CatchDeviceVariableResponse)
    yield 114, Lazy(Command114_ReadCaughtDeviceVariableResponse)
    yield 115, Lazy(Command115_ReadEventNotificationSummaryResponse)
    yield 116, Lazy(Command116_WriteEventNotificationBitMaskResponse)
    yield 117, Lazy(Command117_WriteEventNotificationTimingResponse)
    yield 118, Lazy(Command118_EventNotificationControlResponse)
    yield 119, Lazy(Command119_AcknowledgeEventNotificationResponse)
    yield 512, Lazy(Command512_ReadCountryCodeResponse)
    yield 513, Lazy(Command513_WriteCountryCodeResponse)
    yield 514, Lazy(Command514_RegisterEventManagerResponse)
    yield 515, Lazy(Command515_ReadEventManagerRegistrationStatusResponse)
    yield 516, Lazy(Command516_ReadDeviceLocationResponse)
    yield 517, Lazy(Command517_WriteDeviceLocationResponse)
    yield 518, Lazy(Command518_ReadLocationDescriptionResponse)
    yield 519, Lazy(Command519_WriteLocationDescriptionResponse)
    yield 520, Lazy(Command520_ReadProcessUnitTagResponse)
    yield 521, Lazy(Command521_WriteProcessUnitTagResponse)
    yield 522, Lazy(Command522_WriteVolumetricFlowClassificationResponse)
    yield 523, Lazy(Command523_ReadCondensedStatusMappingArrayResponse)
    yield 524, Lazy(Command524_WriteCondensedStatusMappingResponse)
    yield 525, Lazy(Command525_ResetCondensedStatusMapResponse)
    yield 526, Lazy(Command526_WriteStatusSimulationModeResponse)
    yield 527, Lazy(Command527_SimulateStatusBitResponse)
    yield 528, Lazy(Command528_ReadSubdeviceAssignmentListInformationResponse)
    yield 529, Lazy(Command529_ReadSubdeviceAssignmentResponse)
    yield 530, Lazy(Command530_WriteSubdeviceAssingmentResponse)
    yield 531, Lazy(Command531_TransferLiveListToAssignmentListResponse)
    yield 532, Lazy(Command532_ReadClientSubscriptionSummaryResponse)
    yield 533, Lazy(Command533_WriteClientSubscriptionFlagsResponse)
