# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from oser import (
    ByteStruct,
    Enum,
    IfElse,
    UBInt16,
    UBInt32,
    Array,
    BFloat,
    String,
    UBInt8,
    BitStruct,
    Flag,
    BitField
)
from oser.core import _ord, _chr, DecodeException
import datetime
import struct


class PackedASCII(String):
    """
    The HART Packed ASCII data type
    """

    def __init__(self, length=None, value=b'', padding=b'\x00'):
        """
        Args:
            length=None: states the string length. Can be a callable (e.g. lambda) or a scalar.
                If set to an Integer the result is a fixed length string padded with ``padding``.
                If set to a ``callable`` (lambda or function) the result is a fixed or variable
                length string padded with ``padding``.
                The length must be a multiple of three.
            value="": the initial string value.
            padding="\x00": that padding pattern used for filling the encoded data if `value` is shorter than `length`.
        """
        ret = String.__init__(self,
                              length=length,
                              value=value,
                              padding=padding)
        if self._evaluate_length() % 3:
            raise AssertionError("length must be a multiple of 3")
        return ret

    def encode(self, full_data=b'', context_data=b''):
        length = self._evaluate_length()
        encoded = (self._value + self._padding * length)[:int(length * 4 / 3)]

        byte_list = list(bytearray(encoded))
        ret = []

        # take 4 bytes and make three of it
        for i in range(len(byte_list) // 4):
            b0, b1, b2, b3 = byte_list[i * 4:(i + 1) * 4]
            r0 = ((b0 & 0x3f) << 2) | ((b1 & 0x30) >> 4)
            r1 = ((b1 & 0x0f) << 4) | ((b2 & 0x3c) >> 2)
            r2 = ((b2 & 0x03) << 6) | (b3 & 0x3f)
            ret += [r0, r1, r2]

        ret = bytes(bytearray(ret))
        return ret

    def decode(self, data, full_data=b'', context_data=b''):
        length = self._evaluate_length(data)
        encoded = data[:length]

        byte_list = list(map(_ord, encoded))
        ret = ""

        # take 4 bytes and make three of it
        for i in range(len(byte_list) // 3):
            b0, b1, b2 = byte_list[i * 3:(i + 1) * 3]

            r0 = (b0 & 0xfc) >> 2
            r1 = ((b0 & 0x03) << 4) | ((b1 & 0xf0) >> 4)
            r2 = ((b1 & 0x0f) << 2) | ((b2 & 0xc0) >> 6)
            r3 = (b2 & 0x3f)

            if not (r0 & (1 << 5)):
                r0 |= (1 << 6)

            if not (r1 & (1 << 5)):
                r1 |= (1 << 6)

            if not (r2 & (1 << 5)):
                r2 |= (1 << 6)

            if not (r3 & (1 << 5)):
                r3 |= (1 << 6)

            ret += _chr(r0) + _chr(r1) + _chr(r2) + _chr(r3)

        if not isinstance(ret, bytes):  # pragma: no cover
            ret = ret.encode()
        self.set(ret)

        return length


class Date(String):

    def __init__(self, date):
        """
        The HART Date data type

        Args:
            date (datetime.date): an instance of :func:`datetime.date` or :func:`datetime.datetime`
                between ``1900-01-01`` and ``2155-12-31``.
        """
        ret = String.__init__(self, length=3, value=date, padding=b"\x00")
        return ret

    def set(self, value):
        assert isinstance(value, (datetime.date, datetime.datetime)), "value must be of type datetime.date"
        self._value = value

    def encode(self, full_data=b'', context_data=b''):
        date = self.get()
        tmp = bytes(bytearray([date.day, date.month, date.year - 1900]))
        return tmp

    def decode(self, data, full_data=b'', context_data=b''):
        length = self._evaluate_length(data)
        encoded = data[:length]

        date = datetime.date(_ord(encoded[2]) + 1900,
                             _ord(encoded[1]),
                             _ord(encoded[0]))
        self.set(date)

        return length


class Time(UBInt32):

    def __init__(self, time):
        """
        The HART Time data type.

        Args:
            time (datetime.time): an instance of :func:`datetime.time` or :func:`datetime.datetime`
                between ``00:00:00`` and ``23:59:59``.
        """
        assert isinstance(time, (datetime.time, datetime.datetime)),\
            "value must be of type datetime.time or datetime.datetime"
        UBInt32.__init__(self, value=time)

    def set(self, value):
        assert isinstance(value, (datetime.time, datetime.datetime)),\
            "value must be of type datetime.time or datetime.datetime"
        self._value = value

    def encode(self, full_data=b'', context_data=b''):
        time = self.get()
        tmp = time.hour
        tmp = tmp * 60 + time.minute
        tmp = tmp * 60 + time.second
        tmp = tmp * 32000 + int(round((float(time.microsecond) / 1000.0) * 32.0))

        return struct.pack(self._struct_representation, tmp)

    def decode(self, data, full_data=b'', context_data=b''):
        try:
            value = struct.unpack(self._struct_representation,
                                  data[:self._byte_count])[0]
        except struct.error as e:
            raise DecodeException(value=str(e), object=self)

        value = int(round((float(value) / 32.0) * 1000.0))
        microsecond = int(value % 1000000)
        value = int(value / 1000000)
        second = value % 60
        value = int(value / 60)
        minute = value % 60
        hour = int(value / 60)

        time = datetime.time(hour, minute, second, microsecond)
        self._value = time
        return self._byte_count


class ResponseCode(Enum):
    def __init__(self, **kwargs):
        """
        HART device specific response code generator.

        ``0: "success"`` is added automatically.
        """
        values = {"success": 0,
                  "error_command_not_implemented": 64}
        values.update(kwargs)
        super(ResponseCode, self).__init__(prototype=UBInt8(),
                                           values=values,
                                           value="success",
                                           strict=False)

    def get(self):
        """
        Return the value.

        Returns:
            the value
        """
        value = self._value
        if value & 0x80:
            # indication of communication error
            active_communication_errors = []
            if value & 0x40:
                active_communication_errors.append("vertical_perity_error")
            if value & 0x20:
                active_communication_errors.append("overrun_error")
            if value & 0x10:
                active_communication_errors.append("framing_error")
            if value & 0x08:
                active_communication_errors.append("longitudinal_error")
            # 0x04 is reserved
            if value & 0x02:
                active_communication_errors.append("buffer_overflow")
            # 0x01 is reserved

            ret = "; ".join(active_communication_errors)
            return ret
        # default behaviour in the other case
        for k, v in self._values.items():
            if v == self._value:
                return k

        return value

    def is_error(self):
        """
        Checks if the response code is of class 'error' or not.
        Warning response codes must have the 'warning\_'-prefix.
        If there is no prefix the response is considered to be an error.
        """
        try:
            key = self.get().lower()
            if key == 'success' or key.startswith('warning_'):
                return False
        except AttributeError:
            # Fallback for an unknown reason
            pass
        return True


class FieldDeviceStatus(UBInt8):
    """
    HART Field Device Status data type.
    """

    def __str__(self, indent=0, name=None, stop_at=None):
        ret = " " * indent * self._indentation_level

        if name:
            ret += name + ": "

        ret += str(self._get_string_representation()) + " (" + self._type_as_string() + ")\n"

        return ret

    def _get_string_representation(self):
        """
        Return the value as string.

        Returns:
            the value as string
        """
        value = self._value
        # indication field device status
        field_device_status = []
        if value & 0x80:
            field_device_status.append("device_malfunction")
        if value & 0x40:
            field_device_status.append("configuration_changed")
        if value & 0x20:
            field_device_status.append("cold_start")
        if value & 0x10:
            field_device_status.append("more_status_available")
        if value & 0x08:
            field_device_status.append("loop_current_fixed")
        if value & 0x04:
            field_device_status.append("loop_current_saturated")
        if value & 0x02:
            field_device_status.append("non-primary_variable_out_of_limits")
        if value & 0x01:
            field_device_status.append("primary_variable_out_of_limits")

        ret = "; ".join(field_device_status)
        return ret


class FrameType(Enum):
    def set(self, value):
        """
        Set the value.

        Args:
            value: the new value
        """
        root = self.root()
        command = 0

        if root is not self:
            # save old extended command
            try:
                command = root.command.get()
            except AttributeError:
                pass
        ret = super(FrameType, self).set(value)
        # payload might change due to change of request/response

        if root is not self:
            # set old extended command
            try:
                root.command.set(command)
            except AttributeError:
                pass

        return ret


class Delimiter(BitStruct):
    """
    HART PDU Delimiter abstraction.
    """

    def __init__(self):
        super(Delimiter, self).__init__()

        self.address_type = Enum(prototype=Flag,
                                 values={"polling": 0, "unique": 1},
                                 value="unique",
                                 strict=True)
        self.number_of_expansion_bytes = BitField(length=2)

        self.physical_layer_type = Enum(prototype=BitField(length=2),
                                        values={"asynchronous": 0,
                                                "synchronous": 1},
                                        value="asynchronous",
                                        strict=True)
        self.frame_type = FrameType(prototype=BitField(length=3),
                                    values={"back": 1,
                                            "stx": 2,
                                            "ack": 6},
                                    value="stx",
                                    strict=True)


class PollingAddress(BitField):
    def __init__(self, length=6):
        super(PollingAddress, self).__init__(length=length)

    def __str__(self, indent=0, name=None, stop_at=None):
        ret = " " * (indent)*self._indentation_level

        if name:  # pragma: no cover
            ret += name + ": "

        ret += hex(self.get()) + " (" + self._type_as_string() + ")\n"

        return ret


class SlaveAddress(PollingAddress):
    def __init__(self, length=38):
        super(SlaveAddress, self).__init__(length=length)


class Address(BitStruct):
    """
    HART PDU address abstraction.
    """

    def __init__(self):
        super(Address, self).__init__()

        self.master_address = Enum(prototype=Flag,
                                   values={"secondary_master": 0, "primary_master": 1},
                                   value="primary_master",
                                   strict=True)
        self.burst_mode = Flag(0)
        self.address = IfElse(condition=lambda self: self.up().delimiter.address_type == "polling",
                              if_true=PollingAddress(length=6),
                              if_false=SlaveAddress(length=38))


class UBInt8Enum(Enum):
    def __init__(self, values={}, value=None, strict=False):
        """
        HART 1-Byte enumeration as an unsigned 8-Bit integer.
        Contains all mandatory default values.
        """
        default_values = {'Not Used': 250,
                          'None': 251,
                          'Unknown': 252,
                          'Special': 253,
                          'Expansion': 254,
                          'Reserved': 255}
        all_values = {}
        all_values.update(default_values)
        all_values.update(values)
        super(UBInt8Enum, self).__init__(prototype=UBInt8,
                                         values=all_values,
                                         value=value,
                                         strict=strict)

    def get(self):
        value = super(UBInt8Enum, self).get()
        if type(value) is int:
            return 'Reserved'
        return value

class _EmptyRequest(ByteStruct):
    pass


class HartResponse(ByteStruct):
    """
    A ByteStruct for HART responses which checks for error response codes in _items() and stops iteration
    after field_device_status.
    All HART responses should be derived from this class.
    """
    def _items(self):
        for name, value in super(HartResponse, self)._items():
            yield name, value
            is_error = self.response_code.is_error()
            if is_error is True and isinstance(value, FieldDeviceStatus):
                return


class UnitAndVariable(ByteStruct):
    def __init__(self):
        super(UnitAndVariable, self).__init__()

        self.units_code = UBInt8()
        self.variable = BFloat()


class DeviceVariableCodeUnitAndValue(ByteStruct):
    def __init__(self):
        super(DeviceVariableCodeUnitAndValue, self).__init__()

        self.device_variable_code = UBInt8()
        self.units_code = UBInt8()
        self.device_variable_value = BFloat()


class AnalogChannelNumberUnitAndLevel(ByteStruct):
    def __init__(self):
        super(AnalogChannelNumberUnitAndLevel, self).__init__()

        self.analog_channel_number_code = UBInt8()
        self.units_code = UBInt8()
        self.level = BFloat()


class TrendValueAndStatus(ByteStruct):
    def __init__(self):
        super(TrendValueAndStatus, self).__init__()

        self.trend_value = BFloat()
        self.trend_status = UBInt8()


class CommandByteCountAndData(ByteStruct):
    def __init__(self):
        super(CommandByteCountAndData, self).__init__()

        self.command = UBInt16()
        self.byte_count = UBInt8()
        self.data = Array(length=lambda self: self.byte_count.get(),
                          prototype=UBInt8)
