# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.communication.hart.interface import HartInterface
from htf.communication.hart.application_layer import HartApplicationLayer
from htf.communication.hart.protocol import HartFrame
import logging
logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)


class HartDeviceCommunication(object):
    """
    A communication to a HART device usable as a context block.
    """

    def __init__(self, comport, polling_addresses=[0], device_address=None, decoder_generator=None):
        """
        Args:
            comport (str): the comport to be used where the HART modem is
                connected to.
            polling_addresses=[0] (list of int or int): the polling
                addresses used to find a HART device.
            device_address=None (int or None): the initial HART device address.
                If set to an interger no device must be looked up.
            decoder_generator=None (callable or None): a callable that returns
                a HartFrame instance with or without device specific commands.
        """
        self._polling_addresses = polling_addresses
        self._device_address = device_address

        if decoder_generator is None:
            decoder_generator = self._get_default_decoder_generator()
        self._decoder_generator = decoder_generator

        self._interface = HartInterface(comport)
        self._application_layer = HartApplicationLayer(
            interface=self._interface,
            decoder_generator=self._decoder_generator)

    @staticmethod
    def _get_default_decoder_generator():
        return HartFrame

    def __enter__(self):
        if self._device_address is None:
            self.find_device(polling_addresses=self._polling_addresses)
        # find first device
        return self

    def __exit__(self, exceptionType, exceptionValue, tb):
        self.close()
        if exceptionType is not None:
            return False

        return True

    def close(self):
        """
        Close the HART application layer and close the HART interface
        supplied.
        Is called in __del__, too.
        """
        self._application_layer.close()

    def __del__(self):
        self.close()

    def set_device_address(self, device_address):
        """
        Set the current HART device address.

        Args:
            device_address (int): the HART device address to be used (<= 38 bits)
        """
        self._device_address = device_address
        self._application_layer.set_device_address(device_address)

    def query(self, request, number_of_preambles=5, preamble=b"\xff", timeout=1.0):
        """
        Query a HART command. The HART frame is encoded automatically.

        Args:
            request (HartFrame): the hart frame to be sent.
            number_of_preambles=5 (int): the number of preambles.
            preamble='\xff' (bytes): ``number_of_preambles`` times
                ``preamble`` is sent before the frame content is sent.
            timeout=1.0 (float or int): the timeout in seconds.

        Returns:
            HartFrame or None: if no frame was recieved ``None`` is returned.
        """
        if request.command != 0:
            request.address.address.set(self._device_address)
        logger.debug("query() request:\n%s", request)
        response = self._application_layer.query(request,
                                                 number_of_preambles,
                                                 preamble,
                                                 timeout)
        logger.debug("query() response:\n%s", response)
        return response

    def find_device(self, polling_addresses=None):
        """
        Find the first HART device by polling all addresses stated
        in ``polling_addresses`` and use the first device found.

        If no device is found an ``AssertionError`` is raised.

        Args:
            polling_addresses=[0] (list of int or int): the polling
                addresses used to find a HART device.

        Returns:
            int: Unique HART address of the first device found in polling.
        """
        device_address = self._application_layer.find_device(polling_addresses)
        self.set_device_address(device_address)
        logging.info("Using first device with address 0x%08x", self._device_address)
        return device_address
