# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import time
import oser
import serial
import serial.rs485
from htf.communication.hart.protocol import Delimiter
import logging
logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)


class HartInterface(object):
    """
    Data-Link-Layer and Physical-Layer for HART.
    """

    def __init__(self, comport, timeout=.3):
        """
        Args:
            comport (str): the comport to be used where the HART modem is
                connected to. Must support RS485 mode.
            timeout=.3 (float): the time out for a single byte. Timeouts in
                ``read()`` are multiple of this timeout.
        """
        self._comport = comport

        self._serial = None
        self._timeout = timeout
        self._rs485_support = True
        s = serial.rs485.RS485(
                port=None,  # do not open automatically,
                baudrate=1200,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_ODD,
                rtscts=False,
                stopbits=serial.STOPBITS_ONE,
                timeout=self._timeout
        )
        s.setRTS(False)

        s.port = comport

        try:
            s.rs485_mode = serial.rs485.RS485Settings()
            s._rs485_mode = s.rs485_mode  # workaround for a bug _rs485_mode is read but never set (issue #1024)

            # open the port
            s.open()
        except (ValueError, OSError):  # Workaround when no driver support for RS485 Mode is available (issue #4115)
            logger.debug("RS485 Mode is not supported using RTS set in software.")
            s = serial.Serial(
                port=None,  # do not open automatically,
                baudrate=1200,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_ODD,
                rtscts=False,
                stopbits=serial.STOPBITS_ONE,
                timeout=self._timeout
            )
            s.setRTS(False)

            s.port = comport
            self._rs485_support = False
            s.open()

            time.sleep(0.1)

        self._serial = s

        self._delimiter = Delimiter()
        self._byte_count = oser.UBInt8()
        self._last_preamble_count = 0

    def write(self, data, number_of_preambles=5, preamble=b"\xff"):
        r"""
        Write a HART frame.

        Args:
            data (bytes): the bytes to be sent.
            number_of_preambles=5 (int): the number of preambles.
            preamble='\\xff' (bytes): ``number_of_preambles`` times
                ``preamble`` is sent before the frame content is sent.
        """
        assert isinstance(data, type(b"")), "data must be bytes"
        the_preamble = preamble * number_of_preambles
        message = the_preamble + data
        logger.debug("write bytes:\n%s", oser.to_hex(message))
        self._serial.flush()
        self._serial.reset_input_buffer()

        if not self._rs485_support:
            self._serial.setRTS(True)
            self._serial.write(message)
            self._serial.flush()
            self._serial.setRTS(False)
        else:
            self._serial.write(message)

    def _next_bytes(self, size=1):
        """
        Because HART communication is really slow a timeout may occur within the
        transmission of a long HART frame.

        Thus a timeout is the case that no byte is received or less bytes are received
        than requested.

        If less bytes than requested are received in a first try and after that the rest is
        transmitted in one or more chunks that's ok.
        """
        ret = b""
        bytes_left = size
        while bytes_left > 0:
            bytes_ = self._serial.read(bytes_left)
            logger.debug("read bytes (bytes_left: %i):\n%s", bytes_left, oser.to_hex(bytes_))
            if bytes_ in (b"", None):
                # timeout at the begging or within a transmission
                return b""
            ret += bytes_
            bytes_left -= len(bytes_)
        return ret

    def get_last_preamble_count(self):
        """
        Gets the count of preambles that was most recently received.
        This does **NOT** require a full HART frame to be read,
        but does require the preambles to be followed by a valid delimiter.

        Returns:
            int: the most recently received number of consecutive preambles
        """
        return self._last_preamble_count

    def read(self, timeout=1.0):
        """
        Read a frame.

        Args:
            timeout=1.0 (float or int): the timeout in seconds.

        Returns:
            bytes or None: if no frame was recieved ``None`` is returned else
                a HART frame is returned.
        """

        # clear input buffer first
        self._serial.reset_input_buffer()

        # read until first preamble was read
        number_of_consecutive_preambles = 0
        delimiter = None
        logger.debug("wait for first preamble (\xff)")
        n = int(float(timeout) / float(self._timeout))
        while n > 0:
            logger.debug("tries left: %i", n)
            # print("wait for preamble %i" % i)
            byte = self._next_bytes(1)
            logger.debug("received byte is %s", repr(byte))
            if len(byte) > 1:
                logger.debug("wrong number of bytes was received")
                return None
            if byte in (b"", None):
                n -= 1
            if byte == b"\xff":
                number_of_consecutive_preambles += 1
                # first preamble was found
                # print("first preamble was found", repr(preamble))
                # print("%i byte before the preamble were received" % i)
                # if number_of_consecutive_preambles >= 2:
                #    break
            else:
                if number_of_consecutive_preambles >= 2:
                    # try to decode the delimiter
                    delimiter = byte
                    try:
                        logging.debug("read delimiter")
                        self._delimiter.decode(delimiter)
                        n = 0
                        break
                    except Exception as e:
                        # print(e)
                        pass
                number_of_consecutive_preambles = 0

        self._last_preamble_count = number_of_consecutive_preambles
        if number_of_consecutive_preambles < 2:
            logger.debug("not enough consecutive preambles were received (%i)", number_of_consecutive_preambles)
            logger.debug("no frame was received")
            return None

        # check the delimiter
        self._delimiter.decode(delimiter)
        logger.debug("delimiter:\n%s", self._delimiter)

        address_length = None

        logging.debug("read address")
        if self._delimiter.address_type == 'polling':
            address_length = 1
            logger.debug("reading a polling address -> %i bytes", address_length)
        else:  # 'unique':
            address_length = 5
            logger.debug("reading a unique address -> %i bytes", address_length)

        address = self._next_bytes(address_length)
        logging.debug("address: %s", repr(address))
        if address in (b"", None) or len(address) != address_length:
            logger.debug("no address was received")
            return None

        logging.debug("read expansion bytes")
        expansion_bytes_length = self._delimiter.number_of_expansion_bytes.get()
        expansion_bytes = self._next_bytes(expansion_bytes_length)
        if len(expansion_bytes) != expansion_bytes_length:
            logger.debug("no expansion bytes were received")
            return None
        logging.debug("expansion bytes: %s", repr(expansion_bytes))

        logging.debug("read command")
        command = self._next_bytes(1)
        logging.debug("command: %s", repr(command))
        if command in (b"", None) or len(command) != 1:
            logger.debug("no command was received")
            return None

        logging.debug("read byte count")
        byte_count = self._next_bytes(1)
        logging.debug("byte count: %s", repr(byte_count))
        if byte_count in (b"", None) or len(byte_count) != 1:
            logger.debug("no byte count was received")
            return None

        # decode byte count
        self._byte_count.decode(byte_count)
        logging.debug("byte count: %i", self._byte_count.get())

        logging.debug("read payload")
        payload_length = self._byte_count.get()

        logger.debug("reading %i more bytes until the end", payload_length)
        payload = self._next_bytes(payload_length)
        logging.debug("payload: %s", repr(payload))
        if len(payload) != payload_length:
            logger.debug("no payload was received")
            return None

        logging.debug("read checksum")
        checksum = self._next_bytes(1)
        logging.debug("checksum: %s", repr(checksum))
        if checksum in (b"", None) or len(checksum) != 1:
            logger.debug("no checksum was received")
            return None

        response = delimiter + address + expansion_bytes + command + byte_count + payload + checksum

        logger.debug("read message:\n%s", oser.to_hex(response))
        return response

    def query(self, data, number_of_preambles=5, preamble=b"\xff", timeout=1.0):
        r"""
        Query a HART command.

        Args:
            request (bytes): the hart frame to be sent.
            number_of_preambles=5 (int): the number of preambles.
            preamble='\\xff' (bytes): ``number_of_preambles`` times
                ``preamble`` is sent before the frame content is sent.
            timeout=1.0 (float or int): the timeout in seconds.

        Returns:
            bytes or None: if no frame was recieved ``None`` is returned else
                a HART frame is returned.
        """
        self.write(data, number_of_preambles, preamble)
        response = self.read(timeout=timeout)
        return response

    def close(self):
        """
        Close the HART interface.
        Is called in __del__, too.
        """
        logger.debug("close HART interface")

        if self._serial is not None:
            self._serial.close()

    def __del__(self):
        try:
            self.close()
        except:
            pass
