# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from .interface import HartInterface
from .application_layer import HartApplicationLayer
import datetime
import logging
from htf.communication.hart.protocol import HartFrame
import argparse
import htf
import sys

logger = logging.getLogger(__name__)


class HartMonitor(object):
    """
    A HART Monitor that can be used interactively.
    """

    def __init__(self,
                 comport=None,
                 decoder_generator=None,
                 protocol_factory=None,
                 interactive=False,
                 argv=None):
        """
        Args:
            comport (str): the comport to be used where the HART modem is
                connected to.
            decoder_generator=None (callable or None): a callable that returns
                a HartFrame instance with or without device specific commands that
                is used to decode all frames (if ``protocol_factory`` is ``None``).
            protocol_factory=None (callable): a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device address. For every received frame the frame is
                decoded using the ``decoder_generator`` that may be ``None``
                and after that a device specific decoder is created respecting
                the device address.
            interactive=True (bool): if set to ``True`` the HART Monitor becomes
                interactive
            argv=None (list of str or None): used for testing, etc.
        """
        if argv is None:
            argv = sys.argv

        self._interface = None
        self._application_layer = None
        self._comport = comport
        self._show_version = False
        self._database = None
        self._protocol_factory = None
        self._interactive = interactive

        self.set_protocol_factory(protocol_factory)

        if self._interactive is True:
            self._parse_arguments(argv)

        self._interface = HartInterface(self._comport)
        self._application_layer = HartApplicationLayer(
            interface=self._interface,
            decoder_generator=decoder_generator
        )

        self.set_decoder_generator(decoder_generator)

        if self._interactive is True:
            self._run_interactively()

    def close(self):
        if self._application_layer is not None:
            self._application_layer.close()

    def __del__(self):
        self.close()

    def _parse_arguments(self, argv):
        self._main_parser = argparse.ArgumentParser(
            description='hart-monitor - HILSTER Testing Framework (HILSTER GmbH, http://hilster.de)')

        self._main_parser.add_argument('-c', '--comport',
                                       default=self._comport,
                                       dest="_comport",
                                       metavar="COMPORT",
                                       required=True,
                                       help="comport for HART communication")

        self._main_parser.add_argument('-d', '--database',
                                       default=None,
                                       dest="_database",
                                       metavar="Protocol database",
                                       help="An import string to the protocol database where device specific "
                                            "HART frames can be found. See docs for details.")

        self._main_parser.add_argument('-V', '--version',
                                       default=self._comport,
                                       dest="_show_version",
                                       action="store_true",
                                       help="Show version and exit")

        self._main_parser.parse_args(argv[1:], self)

        if self._show_version:
            print("hart-monitor is part of htf-" + htf.__version__)
            exit(0)

        if self._database is not None:
            print("Loading protocol database from", self._database)
            try:
                parts = self._database.split(".")
                import_string = ".".join(parts[:-1])
                logger.info("Importing database %s" % self._database)
                module = __import__(import_string)
                for part in parts[1:-1]:
                    module = getattr(module, part)
                protocol_factory = getattr(module, parts[-1])
                self.set_protocol_factory(protocol_factory)
            except ImportError:
                logger.info("Could not load database from '%s'" % self._database)
        # exit(0)

    def _run_interactively(self):
        self.run(0)
        self.close()
        exit(0)

    def set_protocol_factory(self, protocol_factory):
        """
        Set the protocol factory.

        Args:
            protocol_factory=None (callable): a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device address.
        """
        self._protocol_factory = protocol_factory

    def set_decoder_generator(self, decoder_generator=None):
        """
        Set the decoder generator. If ``decoder_generator`` is ``None``
        ``HartFrame`` is used.

        Args:
            decoder_generator=HartFrame (callable): a parser for HART frames.
        """
        if decoder_generator is None:
            decoder_generator = HartFrame
        self._decoder_generator = decoder_generator
        self._application_layer.set_decoder_generator(decoder_generator)

    def _parse_frame(self, frame):
        if frame.command.get() == 0:
            return frame
        if self._protocol_factory is None:
            return frame
        # a protocol factory was supplied
        decoder = self._protocol_factory(frame.address.address.get())()
        decoder.decode(frame.encode())
        return decoder

    def handle_frame(self, frame):  # pragma: no cover
        """
        The frame handler that prints the formatted output.
        This method can be overwritten.

        Args:
            frame (HartFrame): the HART frame to be monitored.
        """
        print("-" * 80)
        print(datetime.datetime.utcnow())
        # print(frame.introspect())
        print(frame)
        logger.info(str(frame))

    def handle_exception(self, frame, exc):
        """
        The exception handler that prints the formatted output.
        This method can be overwritten.

        This overwritten method may or may not call ``raise`` at
        the end to raise the exception again.

        Args:
            frame (HartFrame): the HART frame to be monitored.
            exc (BaseException): the exception that can be formatted.
        """
        print("An exception was raised:", exc)
        # raise

    def run(self, frame_count=0):
        """
        Run the monitor and display received HART frames.

        Args:
            frame_count (int): The number of frames to be received. ``0`` means forever.
        """
        if frame_count == 0:
            def condition(x): return True
        else:
            def condition(x): return x <= frame_count
        i = 0
        while condition(i):
            frame = None
            try:
                frame = self._application_layer.read()
                if frame is not None:
                    frame = self._parse_frame(frame)
                    self.handle_frame(frame)
            except KeyboardInterrupt:
                print("Keyboard interrupt was received. Stopping.")
                return
            except BaseException as e:
                self.handle_exception(frame, e)
            i += 1
