# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from oser import (
    ByteStruct,
    UBInt8,
    UBInt16,
    IfElse,
    Array,
    String,
    Switch,
    Lazy,
)

from oser.checksum import ChecksumMixin
from oser.core import _ord
from .data_types import Delimiter, Address
import htf.communication.hart.universal_commands
import htf.communication.hart.common_practice_commands
from .core import _HartFrame


class CheckByte(UBInt8, ChecksumMixin):
    """
    The HART PDU checkbyte abstraction.
    """
    _original_type = UBInt8

    def __init__(self, strict=True, automatic_calculation=True):
        self._strict = strict
        self.set_automatic_calculation(automatic_calculation)
        super(CheckByte, self).__init__()

    def _calculate_checksum(self, data):
        checksum = 0
        for byte in data:
            checksum ^= _ord(byte)

        return checksum

    def encode(self, full_data=b"", context_data=b""):
        """
        Return the encoded binary string.

        Args:
            full_data (bytes): the binary data string until the part to be
                encoded. The user normally does not need to supply this.
            context_data (bytes): the binary data of the current context.
                The user normally does not need to supply this.
        """
        return ChecksumMixin.encode(self, full_data, context_data)

    def decode(self, data, full_data=b"", context_data=b""):
        """
        Decode a binary string and return the number
        of bytes that were decoded.

        Args:
            data (bytes): the data buffer that is decoded.
            full_data (bytes): the binary data string until the part to be
                decoded. The user normally does not need to supply this.
            context_data (bytes): the binary data of the current context.
                The user normally does not need to supply this.

        Returns:
            int: the number of bytes that were decoded.
        """
        return ChecksumMixin.decode(self, data, full_data, context_data)

    def _fuzzing_value_iterator(self):
        """
        A generator for all fuzzing values set.
        This sets strict temporarly to False
        if fuzzing values are set and restores the original values after the
        fuzzing iterator ends.

        Returns:
            self: if no fuzzing values are set

        Yields:
            fuzzing values
        """
        # store old values
        self._tmp_strict = self._strict

        if self._fuzzing_values is not None:
            # disable strict and automatic_calculation
            self._strict = False

        for value in super(UBInt8, self)._fuzzing_value_iterator():
            yield value

        # restore values
        self._strict = self._tmp_strict


class Command(UBInt8):
    """
    Command parser that supports simple and extended commands.
    """

    def __init__(self, value=0):
        self._command = None
        ret = super(Command, self).__init__(value=value)
        self.set(value)
        self._extended_command_decoder = UBInt16()
        return ret

    def set(self, value):
        """
        Set the value.

        Args:
            value: the new value

        """
        self._command = value
        if value > 255:
            self._value = 0x1F
            self.up().payload.command.set(value)
        else:
            self._value = value

    def get(self):
        """
        Return the value.

        Returns:
            the value.
        """
        if self._value == 0x1F:
            # self._command = self.up().payload.command.get()
            # this leads to an infinite loop.
            # value is set in decode() which calls set()
            pass
        else:
            self._command = self._value
        return self._command

    def decode(self, data, full_data=b'', context_data=b''):
        ret = UBInt8.decode(self,
                            data,
                            full_data=full_data,
                            context_data=context_data)
        # the self._value is set now
        if self._value == 0x1F:
            # read beyond the scope to pre-read the extended command
            if self.up().delimiter.frame_type.get() in ["ack", "back"]:
                # response
                self._extended_command_decoder.decode(data[4:4 + 2])
                pass
            else:
                # request
                self._extended_command_decoder.decode(data[2:2 + 2])
            self.set(self._extended_command_decoder.get())
        return ret


class HartFrame(_HartFrame):
    """
    A HART frame abstraction.
    """

    def __init__(self,
                 command=None,
                 auto_length=True,
                 auto_checkbyte=True):
        """
        Args:
            command=None (int or None): the HART command number.
            auto_length=True (bool): if set to ``True`` the byte count of the
                HART frame is calculated automatically when calling
                ``encode``.
            auto_checkbyte=True (bool): if set to ``True`` the checkbyte of
                the HART frame is calculated automatically when calling
                ``encode``.
        """
        super(HartFrame, self).__init__()

        self._auto_length = None
        self.set_auto_length_enable(auto_length)

        self.delimiter = Delimiter()
        self.address = Address()
        self.expansion_bytes = Array(length=lambda self: self.delimiter.number_of_expansion_bytes.get(),
                                     prototype=UBInt8)
        self.command = Command()
        self.byte_count = UBInt8()

        self._setup_parsers()

        self.payload = IfElse(
            condition=lambda self: self.delimiter.frame_type.get() in ["ack", "back"],
            if_true=self._response_parser,
            if_false=self._request_parser)

        self.check_byte = CheckByte()

        self._update_universal_requests()
        self._update_universal_responses()

        self._update_common_practice_universal_requests()
        self._update_common_practice_universal_responses()

        self._update_requests()
        self._update_responses()

        self.set_auto_check_byte_enable(auto_checkbyte)

        if command is not None:
            self.command.set(command)
            if command == 0:
                self.delimiter.address_type.set("polling")

    def set_command_request_parser(self, command, parser):
        """
        Set a command request parser. This overwrites exising command parsers.

        Args:
            command (int): the command number to be set.
            parser (ByteStruct): the parser for the command.
        """
        self._request_parser.set_value(command, parser)

    def set_command_response_parser(self, command, parser):
        """
        Set a command response parser. This overwrites exising command parsers.

        Args:
            command (int): the command number to be set.
            parser (ByteStruct): the parser for the command.
        """
        self._response_parser.set_value(command, parser)

    def get_command_request_parser(self, command):
        """
        Get a command request parser.

        Args:
            command (int): the command number to be returned.

        Returns:
            ByteStruct: the command parser.

        Raises:
            KeyError: Raises a ``KeyError`` if the command if the command
                is not set.
        """
        return self._request_parser.get_value(command)

    def get_command_response_parser(self, command):
        """
        Get a command response parser.

        Args:
            command (int): the command number to be returned.

        Returns:
            ByteStruct: the command parser.

        Raises:
            KeyError: Raises a ``KeyError`` if the command if the command
                is not set.
        """
        return self._response_parser.get_value(command)

    def _universal_requests_generator(self):
        """
        A generator that yields command, parser tuples for requests.

        Yields:
            tuples of command and parser
        """
        for command, parser in htf.communication.hart.universal_commands.requests_generator():
            yield command, parser

    def _universal_responses_generator(self):
        """
        A generator that yields command, parser tuples for responses.

        Yields:
            tuples of command and parser
        """
        for command, parser in htf.communication.hart.universal_commands.responses_generator():
            yield command, parser

    def _common_practice_requests_generator(self):
        """
        A generator that yields command, parser tuples for requests.

        Yields:
            tuples of command and parser
        """
        for command, parser in htf.communication.hart.common_practice_commands.requests_generator():
            yield command, parser

    def _common_practice_responses_generator(self):
        """
        A generator that yields command, parser tuples for responses.

        Yields:
            tuples of command and parser
        """
        for command, parser in htf.communication.hart.common_practice_commands.responses_generator():
            yield command, parser

    def requests_generator(self):
        """
        A generator that yields command, parser tuples for requests.

        This method shall be overwritten to realize device specific
        commands.

        Yields:
            tuples of command and parser
        """
        return []

    def responses_generator(self):
        """
        A generator that yields command, parser tuples for responses.

        This method shall be overwritten to realize device specific
        commands.

        Yields:
            tuples of command and parser
        """
        return []

    def _update_universal_requests(self):
        for command, parser in self._universal_requests_generator():
            self.set_command_request_parser(command, parser)

    def _update_universal_responses(self):
        for command, parser in self._universal_responses_generator():
            self.set_command_response_parser(command, parser)

    def _update_common_practice_universal_requests(self):
        for command, parser in self._common_practice_requests_generator():
            self.set_command_request_parser(command, parser)

    def _update_common_practice_universal_responses(self):
        for command, parser in self._common_practice_responses_generator():
            self.set_command_response_parser(command, parser)

    def _update_requests(self):
        """
        Update the request parsers. This method can also be overwritten.
        """
        for command, parser in self.requests_generator():
            self.set_command_request_parser(command, parser)

    def _update_responses(self):
        """
        Update the responses parsers. This method can also be overwritten.
        """
        for command, parser in self.responses_generator():
            self.set_command_response_parser(command, parser)

    def set_auto_length_enable(self, enable_auto_length=True):
        """
        Enable or disable the auto length feature.

        Args:
            enable_auto_length=True (bool): the new enable state.
        """
        self._auto_length = enable_auto_length

    def set_auto_check_byte_enable(self, enable_auto_check_byte=True):
        """
        Enable or disable the automatic checkbyte calculation on ``encode``.

        Args:
            enable_auto_check_byte=True (bool): the new enable state.
        """
        self.check_byte.set_automatic_calculation(enable_auto_check_byte)

    def encode(self, full_data=b"", context_data=b""):
        """
        Return the encoded binary string.
        ``byte_count`` is set automatically if ``auto_length`` is supplied.

        Args:
            full_data (bytes): the binary data string until the part to be
                encoded. The user normally does not need to supply this.
            context_data (bytes): the binary data of the current context.
                The user normally does not need to supply this.
        """
        self._set_byte_count()
        return ByteStruct.encode(self, full_data=full_data, context_data=context_data)
