# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import oser
import htf
from htf.communication.hart.protocol import HartFrame
from htf.communication.hart.interface import HartInterface
from htf.communication.hart.application_layer import HartApplicationLayer
from IPython import embed
import logging
from htf.util.oser_prompt import fill_oser_instance
import inspect
import argparse
import sys

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class HartShell(object):
    """
    A HART Shell that can be used interactively.
    """

    def __init__(self,
                 comport=None,
                 decoder_generator=None,
                 protocol_factory=None,
                 timeout=None,
                 interactive=True,
                 loglevel=2,
                 argv=None):
        """
        Args:
            comport (str): the comport to be used where the HART modem is
                connected to.
            decoder_generator=None (callable or None): a callable that returns
                a HartFrame instance with or without device specific commands.
            protocol_factory=None (callable): a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device type.
            timeout=None (float or int): the default timeout in seconds.
            interactive=True (bool): if set to ``True`` the HART Shell becomes
                interactive.
            loglevel=2 (int): set loglevel for loggnig to ``loglevel``.
            argv=None (list of str or None): used for testing, etc.
        """
        if argv is None:
            argv = sys.argv

        self._interface = None
        self._application_layer = None
        self._comport = comport
        self._show_version = False
        self._database = None
        self._decoder_generator = None
        self._protocol_factory = None
        self._interactive = interactive
        self._loglevel = loglevel
        self._timeout = None
        self._globals = None
        self._number_of_preambles = 5
        self.response = None

        self.set_loglevel(self._loglevel)

        self.set_protocol_factory(protocol_factory)

        self.set_timeout(timeout)

        if self._interactive is True:
            self._parse_arguments(argv)

        self._interface = HartInterface(self._comport)
        self._application_layer = HartApplicationLayer(
            interface=self._interface,
            decoder_generator=decoder_generator
        )

        self.set_decoder_generator(decoder_generator)

        self._setup_commands()

        if self._interactive is True:
            self._run_interactively()

    def close(self):
        """
        Close the HART application layer and close the HART interface
        supplied.
        Is called in __del__, too.
        """
        if self._application_layer is not None:
            self._application_layer.close()

    def __del__(self):
        self.close()

    def _parse_arguments(self, argv):
        self._main_parser = argparse.ArgumentParser(
            description='hart-shell - HILSTER Testing Framework (HILSTER GmbH, http://hilster.de)')

        self._main_parser.add_argument('-c', '--comport',
                                       default=self._comport,
                                       dest="_comport",
                                       metavar="COMPORT",
                                       required=True,
                                       help="comport for HART communication")

        self._main_parser.add_argument('-d', '--database',
                                       default=None,
                                       dest="_database",
                                       metavar="Protocol database",
                                       help="An import string to the protocol database where device specific "
                                            "HART frames can be found. See docs for details.")

        self._main_parser.add_argument('-t', '--timeout',
                                       default=1.0,
                                       dest="_timeout",
                                       metavar="timeout",
                                       type=float,
                                       help="The default timeout for commands.")

        self._main_parser.add_argument('-v', '--loglevel',
                                       default=0,
                                       dest="_loglevel",
                                       metavar="loglevel",
                                       type=int,
                                       help="The loglevel: 0: quiet, 1: binary, 2: binary+payload of interest, 3: binary+debug")

        self._main_parser.add_argument('-V', '--version',
                                       default=self._comport,
                                       dest="_show_version",
                                       action="store_true",
                                       help="Show version and exit")

        self._main_parser.parse_args(argv[1:], self)

        if self._show_version:
            print("hart-shell is part of htf-" + htf.__version__)
            exit(0)

        if self._database is not None:
            try:
                parts = self._database.split(".")
                import_string = ".".join(parts[:-1])
                logger.info("Importing database %s" % self._database)
                module = __import__(import_string)
                for part in parts[1:-1]:
                    module = getattr(module, part)
                protocol_factory = getattr(module, parts[-1])
                self.set_protocol_factory(protocol_factory)
            except ImportError as e:
                logger.info("Could not load database from '%s'" % self._database)
                # raise
        # exit(0)

    def _setup_commands(self):
        self._globals = globals()
        for command in range(256):
            def _create_hart_command(c):
                def _hart_command(inputs=None, edit_all=False, timeout=None):
                    return self.cmd(command=c, inputs=inputs, edit_all=edit_all, timeout=timeout)
                return _hart_command
            cmd_callable = _create_hart_command(c=command)
            # return lambda inputs=None: hs.cmd(command=c, inputs=inputs)
            # cmd_callable = (lambda c=command: lambda inputs=None: hs.cmd(command=c, inputs=inputs))()
            cmd_callable.__doc__ = """Query HART command %i interactively

    Args:
        inputs=None (list or ``None``): the inputs to fill the payload
        edit_all=False (bool): if set to ``True`` the whole request is edited
            interactively
        timeout=1.0 (float): the timeout to wait for the answer

    Returns:
        HartFrame: the response
""" % command
            if self._interactive is True:
                self._globals["cmd%i" % command] = cmd_callable
                # print("Created globals method", "cmd%i" % command)
            setattr(self, "cmd%i" % command, cmd_callable)

        # iterate all methods:
        if self._interactive:
            for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
                if not name.startswith("_") and name != "close":
                    # print("name:", name)
                    self._globals[name] = method

    def _run_interactively(self):
        self._run_embedded_ipython_shell()
        self.close()
        exit(0)

    def _run_embedded_ipython_shell(self):  # pragma: no cover
        embed(header="""HART-Shell %s""" %
              htf.__version__)  # this call anywhere in your program will start IPython

    def set_loglevel(self, loglevel):
        """
        Set the loglevel.

        Loglevel ``0`` means ``quiet``.

        Loglevel ``1`` means ``binary`` output. All packets are printed in
        a compact binary format.

        Loglevel ``2`` means ``binary+payload of interest`` output. All packets are printed in
        a compact binary format.
        Additionally the interesting payload is printed. ``response_code`` is left out if it
        is ``success`` and ``field_device_status`` is left out in case of ``0``.
        Other payload is printed line by line.

        Loglevel ``3`` means ``binary+whole packet`` output. All packets are printed in
        a compact binary format.
        Additionally the packets are fully printed.

        Args:
            loglevel (int): Set loglevel to ``loglevel``.
        """
        self._loglevel = loglevel

    def set_protocol_factory(self, protocol_factory=None):
        """
        Set the protocol factory.

        Args:
            protocol_factory=None (callable): a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device type.
        """
        self._protocol_factory = protocol_factory

    def set_decoder_generator(self, decoder_generator=None):
        """
        Set the frame generator.

        Args:
            decoder_generator=HartFrame (callable): a parser for HART frames.
        """
        if decoder_generator is None:
            decoder_generator = HartFrame
        self._decoder_generator = decoder_generator
        self._application_layer.set_decoder_generator(decoder_generator)

    def set_decoder_by_device_address(self, device_address):
        """
        Select the frame decoder and encoder by the device address

        Args:
            device_address (int): the device address
        """
        decoder_generator = self._decoder_generator
        if self._protocol_factory is not None:
            device_type = (device_address >> 24) & 0xFFFF
            decoder_generator = self._protocol_factory(device_type)
        self.set_decoder_generator(decoder_generator=decoder_generator)

    def set_timeout(self, timeout):
        """
        Set the default timeout.

        Args:
            timeout (float or int): the default timeout in seconds
        """
        if timeout is None:
            timeout = 1.0
        self._timeout = timeout

    def set_device_address(self, device_address):
        """
        Set the current HART device address.

        Args:
            device_address (int): the HART device address to be used (<= 38 bits)
        """
        self._application_layer.set_device_address(device_address)
        self.set_decoder_by_device_address(self._application_layer._device_address)

    def set_master_address(self, master_address):
        """
        Set the current simulated master address.

        Args:
            master_address (str): "primary_master" or "secondary_master"
        """
        self._application_layer.set_master_address(master_address)

    def set_number_of_preambles(self, number_of_preambles):
        """
        Set the number of preambles for requests.

        Args:
            number_of_preambles (int): the number of preambles for requests.
        """
        self._number_of_preambles = number_of_preambles

    def find_device(self, polling_addresses=None):
        """
        Find the first HART device by polling all addresses stated
        in ``polling_addresses`` and use the first device found.

        If no device is found an ``AssertionError`` is raised.

        Args:
            polling_addresses=[0] (list of int or int): the polling
                addresses used to find a HART device.

        Returns:
            int: Unique HART address of the first device found in polling.
        """
        try:
            ret = self._application_layer.find_device(polling_addresses)
            self.set_decoder_by_device_address(self._application_layer._device_address)
            print("Using first device with address 0x%08x" % self._application_layer._device_address)
            return ret
        except Exception as e:
            "No HART device was found: %s", str(e)

        print("No HART device was found.")

    def poll_devices(self, polling_addresses=None):
        """
        Poll ``polling_addresses`` to find HART devices.

        Args:
            polling_addresses=[0] (list of int): list of HART polling addresses

        Yields:
            int: HART device addresses
        """
        addresses = list(self._application_layer.poll_devices(polling_addresses))
        if len(addresses) == 1:
            print("Found %i HART device." % len(addresses))
        else:
            print("Found %i HART devices." % len(addresses))

        for address in sorted(addresses):
            print("0x%010x" % address)

    def _print_hex_line(self, packet, number_of_preambles):
        line = ""
        line += packet.delimiter.frame_type.get().upper() + ": "
        line += "%iFF " % number_of_preambles
        line += "%02X " % bytearray(packet.delimiter.encode())[0]
        for x in bytearray(packet.address.encode()):
            line += "%02X" % x
        line += " "
        if len(packet.expansion_bytes):
            for x in bytearray(packet.expansion_bytes.encode()):
                line += "%02X" % x
            line += " "

        if packet.payload.size():
            for x in bytearray(packet.payload.encode()):
                line += "%02X" % x
            line += " "

        line += "%02X " % bytearray(packet.check_byte.encode())[0]
        print(line)

    def _print_packet(self, packet, number_of_preambles):
        if self._loglevel is None:
            return

        if self._loglevel >= 1:
            self._print_hex_line(packet, number_of_preambles)

        if self._loglevel == 2:
            for name, value in packet.payload._items():
                if name == "response_code" and value.get() == "success":
                    continue
                if name == "field_device_status" and value.get() in ["", 0]:
                    continue
                print(name + ":", value.get())

        if self._loglevel == 3:
            print(packet)

    def _query_command(self, command, inputs=None, edit_all=False, interactive=True, timeout=1.0):
        request = self._decoder_generator(command=command)

        if self._interactive is True and interactive is True:
            request.encode()  # build checkbyte for oser-prompt
            
            if edit_all is True:
                fill_oser_instance(instance=request,
                                   inputs=inputs)
            else:
                fill_oser_instance(instance=request.payload,
                                   prefix=request.__class__.__name__ + ".payload.",
                                   inputs=inputs)

        # use last device address if needed
        if command != 0:
            request.address.address.set(self._application_layer._device_address)

        request.address.master_address.set(self._application_layer._master_address)
        request.encode()  # build checkbyte

        self._print_packet(request, self._number_of_preambles)

        response = self._application_layer.query(
                                request,
                                timeout=timeout,
                                number_of_preambles=self._number_of_preambles)

        number_of_recieved_preambles = self.last_preamble_count()

        if response is not None:
            self.response = response
            self._print_packet(response, number_of_recieved_preambles)

        return response

    def cmd(self, command, inputs=None, edit_all=False, interactive=True, timeout=None):
        """
        Query a HART command.

        Args:
            command (int): the HART command number.
            inputs=None (None or list of object): the inputs to be set in order
                to be put into the HART request.
            edit_all=False (bool): if set to ``True`` and the HART shell runs
                interactively all request fields are edited interactively.
            interactive=True (bool): if set to ``True`` the request is edited interactively.
            timeout=None (float or int): the timeout in seconds. If ``None`` is
                supplied the default timeout is used. If the default timeout
                is not set 1.0 is used.
        """
        #print("Running Command %i" % command)

        if timeout is None:
            timeout = self._timeout
        if timeout is None:
            timeout = 1.0
        return self._query_command(command, inputs, edit_all, interactive, timeout)

    def last_preamble_count(self):
        """
        Returns the number of preambles most recently received by the
        HartInterface.

        Returns:
            int: number of preambles
        """
        return self._interface.get_last_preamble_count()

    def last_response(self):
        """
        Get the last response.

        The last received packet can also be access via
        ``self.response``.

        Returns:
            object: the last received packet.
        """
        return self.response

    def list_device_commands(self, commands=None):
        """
        List the implemented commands of a HART device.
        Helps to find undocumented commands.

        Args:
            commands (list of int): the list of commands to be probed. If set to ``None`` ``0 .. 255`` is used.

        Returns:
            list of int: the implemented commands.
        """
        if commands is None:
            commands = range(256)

        supported_commands = []

        try:
            loglevel = self._loglevel
            self.set_loglevel(0)  # be quiet

            for command in commands:
                print("command", command, ".. ", end='')
                response = None
                try:
                    response = self.cmd(command, interactive=False)
                except Exception:
                    pass
                if response is not None:
                    supported_commands.append(command)
                    print("is supported")
                else:
                    print("is not supported")
        finally:
            self.set_loglevel(loglevel)

        return supported_commands
