# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import time
import logging
from htf.communication.hart import HartFrame, HartApplicationLayer, HartInterface


class HartSlaveSimulator(object):

    def __init__(self, comport, decoder_generator, polling_address=0, device_type=0, device_id=0):

        self._interface = None
        self._application_layer = None
        self._comport = comport
        self._polling_address = polling_address
        self._device_type = device_type
        self._device_id = device_id
        self._device_address = ((self._device_type << 24) + self._device_id) & 0x3FFFFFFFFF
        self._frame_type = decoder_generator
        self._disabled_responses = []
        self._frame_callbacks = []

        self._interface = HartInterface(self._comport)
        self._application_layer = HartApplicationLayer(
            interface=self._interface,
            decoder_generator=HartFrame
        )

        self._application_layer.set_device_address(self._device_address)

    def run_while(self, condition):
        """
        Simulate while the condition is ``True``.

        Args:
            condition (bool or callable): the condition for the run.
        """

        if not callable(condition):
            condition_parameter = condition

            def condition(): return condition_parameter

        while condition():
            self.run(1)

    def _get_command(self, request):
        """
        Gets the command number from a given HART frame

        Args:
            request (HartFrame): the HART frame to get the command from

        Returns:
            int: the command number of the HART frame
        """
        command = request.command.get()
        return command

    def _handle_callbacks(self, frame):
        """
        Calls the registered frame callbacks with a timestamp and
        the HART frame that was processed

        Args:
            frame (HartFrame): the HART frame that was processed
        """
        timestamp = time.time()
        for callback in self._frame_callbacks:
            callback(timestamp, frame)

    def _address_matches(self, frame):
        """
        Checks if the device address in the frame matches the one of the simulator.

        Args:
            frame (HartFrame): the HART frame to check

        Returns:
            bool: True if the address matches, false otherwise
        """
        command = frame.command.get()
        address = frame.address.address.get()

        address_matches = True

        if command != 0 and address != self._device_address:
            logging.info("Skip request due to different device address. Received 0x%010x. Expected 0x%010x."
                         % (address, self._device_address))
            address_matches = False
        return address_matches

    def _get_device_frame(self, frame):
        """
        Decodes a HART frame using the decoder generator set in the simulator

        Args:
            frame (HartFrame): the HART frame to decode

        Returns:
            HartFrame: the device specific HART frame
        """
        device_frame = self._frame_type()
        device_frame.decode(frame.encode())
        return device_frame

    def _response_disabled(self, frame):
        """
        Checks if responses to a HART command have been disabled.

        Args:
            frame (HartFrame): the HART frame to check

        Returns:
            bool: True if responses are disabled, False otherwise
        """
        command = frame.command.get()
        response_disabled = False
        if command in self._disabled_responses or "any" in self._disabled_responses:
            response_disabled = True
        return response_disabled

    def _get_response_from_handler(self, frame):
        """
        Delegates the HART request frame to the command specific handler if
        it has one, or the default handler if it doesn't.

        Args:
            frame (HartFrame): the HART frame to get a response to

        Returns:
            HartFrame or None: the response HART frame, or None
        """
        command = frame.command.get()
        handler = getattr(self, "handle_command{}".format(command), self.handle_default)
        response = handler(frame)

        return response

    def handle_frame(self, frame):
        """
        Checks that the device address matches and whether responses to the HART command
        are disabled. Calls possible frame callbacks and sends a HART response frame if
        there is one.

        Args:
            frame (HartFrame): the HART frame to handle
        """
        if not self._address_matches(frame):
            return

        frame = self._get_device_frame(frame)
        self._handle_callbacks(frame)        
        
        if self._response_disabled(frame):
            return

        response = self._get_response_from_handler(frame)

        if response is not None:
            self._application_layer.write(response)
            self._handle_callbacks(response)

    def set_response_enabled(self, command, enabled=True):
        """
        Enables/disables responses to a given HART command number.
        Responses to all commands are enabled by default.

        Args:
            command (int or str): HART command number or "any"
            enabled=True (bool): enabled state
        """
        if type(command) is not int and command != "any":
            raise ValueError

        if enabled:
            if command in self._disabled_responses:
                self._disabled_responses.remove(command)
        else:
            if command not in self._disabled_responses:
                self._disabled_responses.append(command)

    def register_frame_callback(self, callback):
        """
        Registers a callback for every HART request and response
        that is received or send out. Callbacks are called with
        two arguments: a timestamp and the HART frame.

        Args:
            callback (callable): callback for HART frames
        """
        if callable(callback) and callback not in self._frame_callbacks:
            self._frame_callbacks.append(callback)

    def unregister_frame_callback(self, callback):
        """
        Unregisters a previously registered callback.

        Args:
            callback (callable): callback for HART frames
        """
        if callback in self._frame_callbacks:
            self._frame_callbacks.remove(callback)

    def _create_response(self, request):
        """
        Creates a response HART frame from the command number
        of the request and the decoder generator set in the simulator.

        Args:
            request (HartFrame): the request HART frame

        Returns:
            HartFrame: a HART response to the same command
        """
        command = self._get_command(request)
        response_frame = self._frame_type(command=command)
        response_frame.delimiter.frame_type.set("ack")

        return response_frame

    def handle_default(self, frame):
        """
        Default handler for HART requests that don't have a specialized handler.

        Args:
            frame (HartFrame): the HART frame to handle
        """
        print("Handle default")

    def handle_command0(self, frame):
        """
        Handler for command 0 HART requests.
        Checks if the polling address in the frame matches the one set in the simulator.

        Args:
            frame (HartFrame): the command 0 HART request

        Returns:
            HartFrame: the response containing device type, device id and protocol revision
        """
        if frame.address.address == self._polling_address:
            response = self._create_response(frame)
            response.delimiter.address_type.set("polling")
            response.payload.device_type.set(self._device_type)
            response.payload.device_id.value.set(self._device_id)
            response.payload.protocol_revision.set(7)
            return response
        else:
            print("Wrong polling address")

    def close(self):
        """
        Closes the HART Application Layer of the simulator.
        """
        if self._application_layer is not None:
            self._application_layer.close()

    def __del__(self):
        self.close()

    def handle_exception(self, exc):
        """
        The exception handler that prints the raised exception.

        Args:
            exc (BaseException): the exception that was raised.
        """
        print("An exception was raised:", exc)
        # raise

    def run(self, frame_count=0):
        """
        Run the HART device simulator

        Args:
            frame_count (int): The number of frames to handle. ``0`` means forever.
        """
        if frame_count == 0:  # pragma: no cover
            def condition(x): return True
        else:
            def condition(x): return x < frame_count
        i = 0
        while condition(i):
            frame = None
            try:
                frame = self._application_layer.read()
                if frame is not None:
                    self.handle_frame(frame)
            except KeyboardInterrupt:
                print("Keyboard interrupt was received. Stopping.")
                return
            except BaseException as e:
                self.handle_exception(e)
            i += 1
