# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from oser import (
    ByteStruct,
    BitStruct,
    BitField,
    UBInt8,
    UBInt16,
    UBInt32,
    If,
    Array,
    String,
    BFloat,
    Lazy
)
from .data_types import PackedASCII, Date, Time, ResponseCode, FieldDeviceStatus, _EmptyRequest, UnitAndVariable, \
    HartResponse
import datetime


class Command0_ReadUniqueIdentifierRequest(_EmptyRequest):
    """
    HART command 0 request abstraction.
    """


class Command0_ReadUniqueIdentifierResponse(HartResponse):
    """
    HART command 0 response abstraction.
    """

    def __init__(self):
        super(Command0_ReadUniqueIdentifierResponse, self).__init__()

        self.response_code = ResponseCode()
        self.field_device_status = FieldDeviceStatus()

        self.data_object_id = UBInt8()
        self.device_type = UBInt16()
        self.minimum_preambles_request = UBInt8()
        self.protocol_revision = UBInt8()

        self.device_revision_level = UBInt8()
        self.software_revision_level = UBInt8()

        self.hardware_revision_level_and_physical_signaling_code = UBInt8()
        self.flags = UBInt8()

        self.device_id = BitStruct()
        self.device_id.value = BitField(length=24)

        self.minimum_preambles_response = If(lambda self: self.protocol_revision > 5,
                                             if_true=UBInt8())
        self.maximum_number_of_device_variables = If(lambda self: self.protocol_revision > 5,
                                                     if_true=UBInt8())
        self.configuration_change_counter = If(lambda self: self.protocol_revision > 5,
                                               if_true=UBInt16())
        self.extended_flags = If(lambda self: self.protocol_revision > 5,
                                 if_true=UBInt8())

        self.manufacturer_id = If(lambda self: self.protocol_revision >= 7,
                                  if_true=UBInt16())
        self.private_label_distributor_code = If(lambda self: self.protocol_revision >= 7,
                                                 if_true=UBInt16())
        self.device_profile = If(lambda self: self.protocol_revision >= 7,
                                 if_true=UBInt8())


class Command1_ReadPrimaryVariableRequest(_EmptyRequest):
    """
    HART command 1 request abstraction.
    """


class Command1_ReadPrimaryVariableResponse(HartResponse):
    """
    HART command 1 response abstraction.
    """

    def __init__(self):
        super(Command1_ReadPrimaryVariableResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.units = UBInt8()
        self.primary_variable = BFloat()


class Command2_ReadLoopCurrentAndPercentageOfRangeRequest(_EmptyRequest):
    """
    HART command 2 request abstraction.
    """


class Command2_ReadLoopCurrentAndPercentageOfRangeResponse(HartResponse):
    """
    HART command 2 response abstraction.
    """

    def __init__(self):
        super(Command2_ReadLoopCurrentAndPercentageOfRangeResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_loop_current = BFloat()
        self.primary_variable_percentage_of_range = BFloat()


class Command3_ReadDynamicVariablesAndLoopCurrentRequest(_EmptyRequest):
    """
    HART command 3 request abstraction.
    """


class Command3_ReadDynamicVariablesAndLoopCurrentResponse(HartResponse):
    """
    HART command 3 response abstraction.
    """

    def __init__(self, *args, **kwargs):
        super(Command3_ReadDynamicVariablesAndLoopCurrentResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            warning_update_failure=8,
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_loop_current = BFloat()
        self.units_and_variables = Array(length=lambda self: (self.up().byte_count.get() - 6) // 5,
                                         prototype=UnitAndVariable)


# Command 4: reserved
# Command 5: reserved


class Command6_WritePollingAddressRequest(ByteStruct):
    """
    HART command 6 request abstraction.
    """

    def __init__(self):
        super(Command6_WritePollingAddressRequest, self).__init__()
        self.polling_address = UBInt8()
        self.loop_current_mode = If(lambda self: self.up().byte_count.get() > 1, if_true=UBInt8())


class Command6_WritePollingAddressResponse(HartResponse):
    """
    HART command 6 response abstraction.
    """

    def __init__(self):
        super(Command6_WritePollingAddressResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_polling_address_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_mode_selection=12,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.polling_address = UBInt8()
        self.loop_current_mode = UBInt8()


class Command7_ReadLoopConfigurationRequest(_EmptyRequest):
    """
    HART command 7 request abstraction.
    """


class Command7_ReadLoopConfigurationResponse(HartResponse):
    """
    HART command 7 response abstraction.
    """

    def __init__(self):
        super(Command7_ReadLoopConfigurationResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.polling_address = UBInt8()
        self.loop_current_mode = UBInt8()


class Command8_ReadDynamicVariableClassificationsRequest(_EmptyRequest):
    """
    HART command 8 request abstraction.
    """


class Command8_ReadDynamicVariableClassificationsResponse(HartResponse):
    """
    HART command 8 response abstraction.
    """

    def __init__(self):
        super(Command8_ReadDynamicVariableClassificationsResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16)
        self.field_device_status = FieldDeviceStatus()

        self.primary_variable_classification = UBInt8()
        self.secondary_variable_classification = UBInt8()
        self.tertiary_variable_classification = UBInt8()
        self.quaternary_variable_classification = UBInt8()


# Command9: TBD
class Command9_ReadDeviceVariablesWithStatusRequest(ByteStruct):
    """
    HART command 9 request abstraction.
    """

    def __init__(self):
        super(Command9_ReadDeviceVariablesWithStatusRequest, self).__init__()
        self.device_variable_codes = Array(length=lambda self: self.up().byte_count.get(),
                                           prototype=UBInt8)


class DeviceVariableAndStatus(ByteStruct):
    def __init__(self, *args, **kwargs):
        super(DeviceVariableAndStatus, self).__init__()

        self.device_variable_code = UBInt8()
        self.device_variable_classification = UBInt8()
        self.units_code = UBInt8()
        self.device_variable_value = BFloat()
        self.device_variable_status = UBInt8()


class Command9_ReadDeviceVariablesWithStatusResponse(HartResponse):
    """
    HART command 9 response abstraction.
    """

    def __init__(self):
        super(Command9_ReadDeviceVariablesWithStatusResponse, self).__init__()

        self.response_code = ResponseCode(
            error_invalid_selection=2,
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            warning_update_failures=8,
            warning_dynamic_variables_returned_for_device_variables=14,
            error_access_restricted=16,
            warning_command_response_truncated=30)
        self.field_device_status = FieldDeviceStatus()

        self.extended_field_device_status = UBInt8()
        self.variables_and_statuses = Array(
            length=lambda self: (self.up().byte_count.get() - 7) // 8,
            prototype=DeviceVariableAndStatus)
        self.slot0_data_time_stamp = Time(datetime.time())

# Command10: reserved


class Command11_ReadUniqueIdentifierAssociatedWithTagRequest(ByteStruct):
    """
    HART command 11 request abstraction.
    """

    def __init__(self):
        super(Command11_ReadUniqueIdentifierAssociatedWithTagRequest, self).__init__()
        self.tag = PackedASCII(length=6, value=b"?" * 8)


class Command11_ReadUniqueIdentifierAssociatedWithTagResponse(Command0_ReadUniqueIdentifierResponse):
    """
    HART command 11 response abstraction.
    """


class Command12_ReadMessageRequest(_EmptyRequest):
    """
    HART command 12 request abstraction.
    """


class Command12_ReadMessageResponse(HartResponse):
    """
    HART command 12 response abstraction.
    """

    def __init__(self):
        super(Command12_ReadMessageResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.message = PackedASCII(length=24, value=b"?" * 32)


class Command13_ReadTagDescriptorDateRequest(_EmptyRequest):
    """
    HART command 13 request abstraction.
    """


class Command13_ReadTagDescriptorDateResponse(HartResponse):
    """
    HART command 13 response abstraction.
    """

    def __init__(self):
        super(Command13_ReadTagDescriptorDateResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.tag = PackedASCII(length=6, value=b"?" * 8)
        self.descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.date = Date(datetime.date(1900, 1, 1))


class Command14_ReadPrimaryVariableTransducerInformationRequest(_EmptyRequest):
    """
    HART command 14 request abstraction.
    """


class Command14_ReadPrimaryVariableTransducerInformationResponse(HartResponse):
    """
    HART command 14 response abstraction.
    """

    def __init__(self):
        super(Command14_ReadPrimaryVariableTransducerInformationResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.serial_number = BitStruct()
        self.serial_number.value = BitField(length=24)
        self.units_code = UBInt8()
        self.upper_limit = BFloat()
        self.lower_limit = BFloat()
        self.minimum_span = BFloat()


class Command15_ReadDeviceInformationRequest(_EmptyRequest):
    """
    HART command 15 request abstraction.
    """


class Command15_ReadDeviceInformationResponse(HartResponse):
    """
    HART command 15 response abstraction.
    """

    def __init__(self):
        super(Command15_ReadDeviceInformationResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.alarm_selection_code = UBInt8()
        self.transfer_function_code = UBInt8()
        self.units_code = UBInt8()
        self.upper_range_value = BFloat()
        self.lower_range_value = BFloat()
        self.daming_value = BFloat()
        self.write_protect_code = UBInt8()
        self.reserved = UBInt8(250)
        self.analog_channel_flags = UBInt8()


class Command16_ReadFinalAssemblyNumberRequest(_EmptyRequest):
    """
    HART command 16 request abstraction.
    """


class Command16_ReadFinalAssemblyNumberResponse(HartResponse):
    """
    HART command 16 response abstraction.
    """

    def __init__(self):
        super(Command16_ReadFinalAssemblyNumberResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.final_assembly_number = BitStruct()
        self.final_assembly_number.value = BitField(length=24)


class Command17_WriteMessageRequest(ByteStruct):
    """
    HART command 17 request abstraction.
    """

    def __init__(self):
        super(Command17_WriteMessageRequest, self).__init__()
        self.message = PackedASCII(length=24, value=b"?" * 32)


class Command17_WriteMessageResponse(HartResponse):
    """
    HART command 17 response abstraction.
    """

    def __init__(self):
        super(Command17_WriteMessageResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.message = PackedASCII(length=24, value=b"?" * 32)


class Command18_WriteTageDescriptorDateRequest(ByteStruct):
    """
    HART command 18 response abstraction.
    """

    def __init__(self):
        super(Command18_WriteTageDescriptorDateRequest, self).__init__()
        self.tag = PackedASCII(length=6, value=b"?" * 8)
        self.descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.date = Date(datetime.date(1900, 1, 1))


class Command18_WriteTageDescriptorDateResponse(HartResponse):
    """
    HART command 18 response abstraction.
    """

    def __init__(self):
        super(Command18_WriteTageDescriptorDateResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_invalid_data_code_detected=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.tag = PackedASCII(length=6, value=b"?" * 8)
        self.descriptor = PackedASCII(length=12, value=b"?" * 16)
        self.date = Date(datetime.date(1900, 1, 1))


class Command19_WriteFinalAssemblyNumberRequest(ByteStruct):
    """
    HART command 19 request abstraction.
    """

    def __init__(self):
        super(Command19_WriteFinalAssemblyNumberRequest, self).__init__()
        self.final_assembly_number = BitStruct()
        self.final_assembly_number.value = BitField(length=24)


class Command19_WriteFinalAssemblyNumberResponse(HartResponse):
    """
    HART command 19 response abstraction.
    """

    def __init__(self):
        super(Command19_WriteFinalAssemblyNumberResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.final_assembly_number = BitStruct()
        self.final_assembly_number.value = BitField(length=24)


class Command20_ReadLongTagRequest(_EmptyRequest):
    """
    HART command 20 request abstraction.
    """


class Command20_ReadLongTagResponse(HartResponse):
    """
    HART command 20 response abstraction.
    """

    def __init__(self):
        super(Command20_ReadLongTagResponse, self).__init__()

        self.response_code = ResponseCode(
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.long_tag = String(length=32)


class Command21_ReadUniqueIdentifierAssociatedWithLongTagRequest(ByteStruct):
    """
    HART command 21 request abstraction.
    """

    def __init__(self):
        super(Command21_ReadUniqueIdentifierAssociatedWithLongTagRequest, self).__init__()

        self.long_tag = String(length=32)


class Command21_ReadUniqueIdentifierAssociatedWithLongTagResponse(Command0_ReadUniqueIdentifierResponse):
    """
    HART command 21 response abstraction.
    """


class Command22_WriteLongTagRequest(Command21_ReadUniqueIdentifierAssociatedWithLongTagRequest):
    """
    HART command 22 request abstraction.
    """


class Command22_WriteLongTagResponse(HartResponse):
    """
    HART command 2 response abstraction.
    """

    def __init__(self):
        super(Command22_WriteLongTagResponse, self).__init__()

        self.response_code = ResponseCode(
            error_too_few_data_bytes_received=5,
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_access_restricted=16,
            error_busy=32,
            error_dr_initated=33,
            error_dr_running=34,
            error_dr_dead=35,
            error_dr_conflict=36)
        self.field_device_status = FieldDeviceStatus()

        self.long_tag = String(length=32)


class Command38_ResetConfigurationChangedFlagRequest(ByteStruct):
    """
    HART command 38 request abstraction.
    """

    def __init__(self):
        super(Command38_ResetConfigurationChangedFlagRequest, self).__init__()
        self.configuration_change_counter = If(lambda self: self.root().byte_count >= 2,
                                               if_true=UBInt16())


class Command38_ResetConfigurationChangedFlagResponse(HartResponse):
    """
    HART command 38 response abstraction.
    """

    def __init__(self):
        super(Command38_ResetConfigurationChangedFlagResponse, self).__init__()

        self.response_code = ResponseCode(
            error_device_specific_command_error=6,
            error_in_write_protect_mode=7,
            error_configuration_change_counter_mismatch=9,
            error_access_restricted=16,
            error_busy=32)
        self.field_device_status = FieldDeviceStatus()

        self.configuration_change_counter = UBInt16()


class UnknownRequest(ByteStruct):
    """
    Default request for unknown
    """

    def __init__(self):
        super(UnknownRequest, self).__init__()
        self.data = String(length=lambda self: self.root().byte_count.get())


class UnknownResponse(HartResponse):
    """
    Default response for unknown
    """

    def __init__(self):
        super(UnknownResponse, self).__init__()
        self.response_code = ResponseCode()
        self.field_device_status = FieldDeviceStatus()
        self.data = String(length=lambda self: max(self.root().byte_count.get() - 2, 0))


def requests_generator():
    yield 0, Lazy(Command0_ReadUniqueIdentifierRequest)
    yield 1, Lazy(Command1_ReadPrimaryVariableRequest)
    yield 2, Lazy(Command2_ReadLoopCurrentAndPercentageOfRangeRequest)
    yield 3, Lazy(Command3_ReadDynamicVariablesAndLoopCurrentRequest)
    # 4: reserved
    # 5: reserved
    yield 6, Lazy(Command6_WritePollingAddressRequest)
    yield 7, Lazy(Command7_ReadLoopConfigurationRequest)
    yield 8, Lazy(Command8_ReadDynamicVariableClassificationsRequest)
    yield 9, Lazy(Command9_ReadDeviceVariablesWithStatusRequest)
    # 10: reserved
    yield 11, Lazy(Command11_ReadUniqueIdentifierAssociatedWithTagRequest)
    yield 12, Lazy(Command12_ReadMessageRequest)
    yield 13, Lazy(Command13_ReadTagDescriptorDateRequest)
    yield 14, Lazy(Command14_ReadPrimaryVariableTransducerInformationRequest)
    yield 15, Lazy(Command15_ReadDeviceInformationRequest)
    yield 16, Lazy(Command16_ReadFinalAssemblyNumberRequest)
    yield 17, Lazy(Command17_WriteMessageRequest)
    yield 18, Lazy(Command18_WriteTageDescriptorDateRequest)
    yield 19, Lazy(Command19_WriteFinalAssemblyNumberRequest)
    yield 20, Lazy(Command20_ReadLongTagRequest)
    yield 21, Lazy(Command21_ReadUniqueIdentifierAssociatedWithLongTagRequest)
    yield 22, Lazy(Command22_WriteLongTagRequest)
    yield 38, Lazy(Command38_ResetConfigurationChangedFlagRequest)


def responses_generator():
    yield 0, Lazy(Command0_ReadUniqueIdentifierResponse)
    yield 1, Lazy(Command1_ReadPrimaryVariableResponse)
    yield 2, Lazy(Command2_ReadLoopCurrentAndPercentageOfRangeResponse)
    yield 3, Lazy(Command3_ReadDynamicVariablesAndLoopCurrentResponse)
    # 4: reserved
    # 5: reserved
    yield 6, Lazy(Command6_WritePollingAddressResponse)
    yield 7, Lazy(Command7_ReadLoopConfigurationResponse)
    yield 8, Lazy(Command8_ReadDynamicVariableClassificationsResponse)
    yield 9, Lazy(Command9_ReadDeviceVariablesWithStatusResponse)
    # 10: reserved
    yield 11, Lazy(Command11_ReadUniqueIdentifierAssociatedWithTagResponse)
    yield 12, Lazy(Command12_ReadMessageResponse)
    yield 13, Lazy(Command13_ReadTagDescriptorDateResponse)
    yield 14, Lazy(Command14_ReadPrimaryVariableTransducerInformationResponse)
    yield 15, Lazy(Command15_ReadDeviceInformationResponse)
    yield 16, Lazy(Command16_ReadFinalAssemblyNumberResponse)
    yield 17, Lazy(Command17_WriteMessageResponse)
    yield 18, Lazy(Command18_WriteTageDescriptorDateResponse)
    yield 19, Lazy(Command19_WriteFinalAssemblyNumberResponse)
    yield 20, Lazy(Command20_ReadLongTagResponse)
    yield 21, Lazy(Command21_ReadUniqueIdentifierAssociatedWithLongTagResponse)
    yield 22, Lazy(Command22_WriteLongTagResponse)
    yield 38, Lazy(Command38_ResetConfigurationChangedFlagResponse)
