# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#
# Use with $ complete -f -C 'complete-htf' htf
#
# test in development:
# alias complete-htf="python /home/teichert/hilster_workspace/htf/htf/complete_htf.py"
# alias htf="python -m htf"

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import pkgutil
import os
import re
import inspect

try:  # pragma: no cover
    # Python >= 3.3
    from importlib.machinery import all_suffixes

    _suffixes = all_suffixes()
except ImportError:  # pragma: no cover
    from imp import get_suffixes

    _suffixes = [s[0] for s in get_suffixes()]

import_re = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*?)'
                       r'(?P<package>[/\\]__init__)?'
                       r'(?P<suffix>%s)$' %
                       r'|'.join(re.escape(s) for s in _suffixes))


def module_list(path):
    """
    Return the list containing the names of the modules available in the given
    folder.

    Taken from IPython.
    """
    # sys.path has the cwd as an empty string, but isdir/listdir need it as '.'
    if path == '':
        path = '.'

    # A few local constants to be used in loops below
    pjoin = os.path.join

    if os.path.isdir(path):
        # Build a list of all files in the directory and all files
        # in its subdirectories. For performance reasons, do not
        # recurse more than one level into subdirectories.
        files = []
        for root, dirs, nondirs in os.walk(path, followlinks=True):
            subdir = root[len(path) + 1:]
            if subdir:
                files.extend(pjoin(subdir, f) for f in nondirs)
                dirs[:] = []  # Do not recurse into additional subdirectories.
            else:
                files.extend(nondirs)

    else:
        try:
            files = list(zipimporter(path)._files.keys())
        except:
            files = []

    # Build a list of modules which match the import_re regex.
    modules = []
    for f in files:
        m = import_re.match(f)
        if m:
            modules.append(m.group('name'))

    return list(set(sorted(modules)))


def _to_import_string_list(parts, entries):
    entries = filter(lambda x: not x.startswith("__") and not x.startswith("@"), entries)
    entries = list(sorted(map(lambda x: ".".join(parts) + "." + x, set(entries))))
    return entries


def get_choices(part):
    """
    Return a list of strings
    that describe importable python
    modules and its contents.
    Eg.:
    - tests.test_1
    - tests
    - tests.test_1.TestCase
    - tests.test_1.TestCase.test_method

    Case 1) complete or incomplete root module -> list all root
        modules starting with part
    Case 2) complete root module + "." -> list all sub-modules and entries
    Case 3) complete root module + ".<part>" -> list all sub-modules
        and entries starting with part and
    """

    if not isinstance(part, type("")):
        part = part.decode()

    if "." not in part:
        root_modules = list(map(lambda x: x[1], pkgutil.iter_modules()))
        if part:
            root_modules = list(filter(lambda x: x.startswith(part), root_modules))
        root_modules = root_modules + list(map(lambda x: x + ".", root_modules))
        ret = list(sorted(set(root_modules)))
        return ret

    if "." in part:
        # recursivly import modules
        module = part
        if isinstance(module, type("")):
            parts = module.split(".")

            for i in range(len(parts), 0, -1):
                modules = []
                tmp_module = ".".join(parts[:i])

                try:
                    if tmp_module.endswith(".") or not tmp_module:
                        # on Cygwin it was possible to run __import__("root.")
                        continue
                    m = __import__(tmp_module)
                    for part in tmp_module.split('.')[1:]:
                        if not part:
                            break
                        m = getattr(m, part)

                    sub_modules = []
                    if i == len(parts) - 1:

                        if "__init__" in m.__file__:
                            path = os.path.dirname(m.__file__)

                            sub_modules = module_list(path=path)
                            sub_modules = list(filter(lambda x: x.startswith(parts[i]), sub_modules))
                            sub_modules = sub_modules + list(map(lambda x: x + ".", sub_modules))

                            sub_modules = _to_import_string_list(parts[:i], sub_modules)
                        else:
                            sub_modules = []
                    if i < len(parts):
                        o = m

                        for p in parts[i:]:

                            try:
                                o = getattr(o, p)
                                sub_modules = []
                            except AttributeError as e:
                                matching_entries = list(sorted(filter(lambda x: x.startswith(p), dir(o))))
                                filtered_entries = []
                                for e in matching_entries:
                                    filtered_entries.append(e)

                                    if inspect.isclass(getattr(o, e)):
                                        filtered_entries.append(e + ".")

                                entries = _to_import_string_list(parts[:i], filtered_entries)
                                return list(sorted(set(modules + sub_modules + entries)))
                            i += 1
                        if inspect.isclass(o):
                            return [".".join(parts), ".".join(parts) + "."]
                        else:
                            return [".".join(parts)]
                    else:
                        modules = [tmp_module, tmp_module + "."]

                    # search for parts within the imported module
                    entries = []

                    return list(sorted(set(modules + sub_modules + entries)))
                except ImportError as e:
                    # print("ImportError:", e)
                    pass
    return []


def print_wordlist(words, start):
    for w in sorted(words):
        if w.startswith(start):
            print(w)


def main():
    this, command, current, previous = sys.argv

    if previous == "htf" or not current:
        # this is the first positional argument
        print_wordlist([
            "run",
            "docs",
            "licenses",
            "version",
        ], current)

    # this is not the first argument
    print_wordlist([
        '-T', '--title', '--title=',
        '-H', '--html-report', '--html-report=',
        '-x', '--junit-xml-report', '--junit-xml-report=',
        '-j', '--json-report', '--json-report=',
        '-v', '--verbose',
        # '-b', '--buffer', '--buffer=',
        '-c', '--catch'
              '-f', '--failfast'
                    '-p', '--pattern', '--pattern=',
    ], current)

    choices = get_choices(current)
    print("\n".join(choices))

    exit(0)


if __name__ == "__main__":
    main()
