# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_Volts, DAQmx_Val_Cfg_Default, DAQmx_Val_Rising,\
    DAQmx_Val_FiniteSamps, DAQmx_Val_Diff
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageInputSingleLine(DelegateMixin):
    """
    ``AnalogVoltageInputSingleLine`` is used to access a single analog output line
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and destructor.
    """

    def __init__(self, descriptor, minimumVoltage=-10.0, maximumVoltage=10.0, terminalConfig=DAQmx_Val_Cfg_Default):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`
            minimumVoltage=-10.0 (float): the minimum voltage.
            maximumVoltage=10.0 (float): the maximum voltage.
            terminalConfig=DAQmx_Val_Cfg_Default: the terminal configuration
                for DAQmxCreateAIVoltageChan(). Possible values: DAQmx_Val_RSE,
                DAQmx_Val_NRSE, DAQmx_Val_Diff, DAQmx_Val_PseudoDiff.

        """
        taskName = "AI_" + descriptor.replace("/", "_")
        self.tmpvalue = float64(0.0)
        self.value = 0.0
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateAIVoltageChan(descriptor,
                                      None,
                                      terminalConfig,
                                      float64(minimumVoltage),
                                      float64(maximumVoltage),
                                      DAQmx_Val_Volts,
                                      None)

    def get(self):
        """
        Return the current pin value.
        """
        self.task.ReadAnalogScalarF64(float64(1.0), byref(self.tmpvalue), None)
        self.value = self.tmpvalue.value
        return self.value

    def set(self, state=True):
        raise NotImplementedError(
            "There is no meaningful implementation for "
            "a 'set' method in the context of this class.")

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    ai0 = AnalogVoltageInputSingleLine("Dev1/ai0", minimumVoltage=0.0,
                                       maximumVoltage=10.0, terminalConfig=DAQmx_Val_Diff)

    for ii in range(10000):
        print("ai0:", ai0.get())
