# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import float64, DAQmx_Val_Volts, DAQmx_Val_Cfg_Default, DAQmx_Val_Rising,\
    DAQmx_Val_FiniteSamps, DAQmx_Val_GroupByChannel, int32
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageInputSingleLineAccurate(DelegateMixin):
    """
    ``AnalogVoltageInputSingleLineAccurate`` is used to access a single analog input line
    on a DAQmx device. It acquires some samples and builds the average value of them.

    A DAQmx task is automatically created and deleted in the constructor and destructor.
    """

    def __init__(self, descriptor,
                 minimumVoltage=-10.0,
                 maximumVoltage=10.0,
                 numberOfSamples=2,
                 sampleFrequency=1000.0,
                 timeout=1.0,
                 terminalConfig=DAQmx_Val_Cfg_Default):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg. `cDaq1Mod1/line0`
            minimumVoltage=-10.0 (float): the minimum voltage.
            maximumVoltage=10.0 (float): the maximum voltage.
            numberOfSamples=2 (int): the number of samples to acquire for one average value.
            sampleFrequency=1000.0 (float): the sample frequency in Hertz.
            timeout=1.0 (float): the timeout when acquiring a new value in seconds.
            terminalConfig=DAQmx_Val_Cfg_Default: the terminal configuration for DAQmxCreateAIVoltageChan().
                Possible values: DAQmx_Val_RSE, DAQmx_Val_NRSE, DAQmx_Val_Diff, DAQmx_Val_PseudoDiff.
        """
        taskName = "AI_" + descriptor.replace("/", "_")
        self.numberOfSamples = numberOfSamples
        self.sampleFrequency = sampleFrequency
        self.timeout = timeout
        # self.data = cast((float64 * self.numberOfSamples)(), POINTER(float64))
        self.data = cast((float64 * self.numberOfSamples)(), POINTER(float64))
        self.value = 0.0
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateAIVoltageChan(descriptor,
                                      None,
                                      terminalConfig,
                                      float64(minimumVoltage),
                                      float64(maximumVoltage),
                                      DAQmx_Val_Volts,
                                      None)
        self.task.CfgSampClkTiming(None,
                                   float(self.sampleFrequency),
                                   DAQmx_Val_Rising,
                                   DAQmx_Val_FiniteSamps,
                                   self.numberOfSamples)

    def get(self):
        """
        Return the current pin value.
        """
        self.task.ReadAnalogF64(self.numberOfSamples,
                                timeout=float64(self.timeout),
                                fillMode=DAQmx_Val_GroupByChannel,
                                readArray=self.data,
                                arraySizeInSamps=self.numberOfSamples,
                                sampsPerChanRead=byref(self.read),
                                reserved=None)
        assert self.read.value == self.numberOfSamples

        # build the average value
        s = 0.0
        for ii in range(self.numberOfSamples):
            s += self.data[ii]

        self.value = s / float(self.numberOfSamples)

        return self.value

    def set(self, value=True):
        raise NotImplementedError(
            "There is no meaningful implementation for a 'set' method in the context of this class.")

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time
    ai0 = AnalogVoltageInputSingleLineAccurate(
        "Dev1/ai0", minimumVoltage=0, maximumVoltage=5.0, numberOfSamples=100, timeout=12.0, sampleFrequency=100)

    for ii in range(1000):
        print("ai0:", ai0.get())
        time.sleep(1.0)
