# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import int32,\
    float64, DAQmx_Val_GroupByChannel, DAQmx_Val_FiniteSamps, DAQmx_Val_Rising,\
    DAQmx_Val_Volts, DAQmx_Val_DoNotAllowRegen,\
    DAQmx_Val_OnBrdMemHalfFullOrLess, DAQmx_Val_ContSamps
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageOutputPort(DelegateMixin):
    """
    ``AnalogVoltageOutputPort`` is used to access a single analog output port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(self, descriptor,
                 minimumVoltage=-10.0,
                 maximumVoltage=10.0,
                 numberOfLines=32,
                 numberOfSamples=2,
                 sampleFrequency=1000.0,
                 timeout=1.0,
                 continuousMode=False):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`
            minimumVoltage=-10.0 (float): the minimum voltage.
            maximumVoltage=10.0 (float): the maximum voltage.
            numberOfLines=32 (int): the number of lines defined in the
                descriptor. For example the descriptor "cDaq1Mod3/ai0:31"
                defines 32 lines.
            numberOfSamples=2 (int): the number of samples to write for every
                line.
            sampleFrequency=1000.0 (float): the sample frequency in Hertz.
            timeout=1.0 (float): the timeout when acquiring a new value in
                seconds.
            continuousMode=False: if set to ``True`` continuous sampling is
                done and get() becomes blocking. Simply read samples in a loop.
        """

        taskName = "AO_" + descriptor.replace("/", "_")
        self.numberOfSamples = numberOfSamples
        self.sampleFrequency = sampleFrequency
        self.numberOfLines = numberOfLines
        self.timeout = timeout
        self.continuousMode = continuousMode
        self.data = cast(
            (float64 * self.numberOfSamples * self.numberOfLines)(),
            POINTER(float64))
        self.values = None
        self.written = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateAOVoltageChan(descriptor,
                                      None,
                                      float64(minimumVoltage),
                                      float64(maximumVoltage),
                                      DAQmx_Val_Volts,
                                      None)

        if self.continuousMode is False:
            self.task.CfgSampClkTiming(None,
                                       float(self.sampleFrequency),
                                       DAQmx_Val_Rising,
                                       DAQmx_Val_FiniteSamps,
                                       self.numberOfSamples * self.numberOfLines)
        else:
            self.task.CfgSampClkTiming(None,
                                       float(self.sampleFrequency),
                                       DAQmx_Val_Rising, DAQmx_Val_ContSamps,
                                       self.numberOfSamples * self.numberOfLines)
            self.task.SetWriteRegenMode(DAQmx_Val_DoNotAllowRegen)
            self.task.SetAODataXferReqCond(descriptor,
                                           DAQmx_Val_OnBrdMemHalfFullOrLess)
            self.task.SetBufOutputBufSize(2 * int(self.numberOfSamples))

    def set(self, values):
        """
        Write analog data.

        Args:
            values: samples to be written on the output port.

        ``values`` are expected to have the following layout::

            [
                [line0 sample0, line0 sample1, ..., line0 sampleN-1],
                [line1 sample0, line1 sample1, ..., line1 sampleN-1],
                ...
                [lineM-1 sample0, lineM-1 sample1, ..., lineM-1 sampleN-1]
            ]
        """

        self.values = values

        for line in range(self.numberOfLines):
            for sample in range(self.numberOfSamples):
                self.data[line * self.numberOfSamples + sample] = \
                    values[line][sample]

        self.task.WriteAnalogF64(self.numberOfSamples,
                                 False,
                                 float64(self.timeout),
                                 DAQmx_Val_GroupByChannel,
                                 self.data,
                                 byref(self.written),
                                 None)
        self.task.start()
        self.task.WaitUntilTaskDone(self.timeout)
        self.task.stop()

    def get(self):
        """
        Return the last written analog values.

        Returned data layout::

            [
                [line0 sample0, line0 sample1, ..., line0 sampleN-1],
                [line1 sample0, line1 sample1, ..., line1 sampleN-1],
                ...
                [lineM-1 sample0, lineM-1 sample1, ..., lineM-1 sampleN-1]
            ]
        """
        return self.values

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    ao0 = AnalogVoltageOutputPort("Dev1/ao0", numberOfLines=2)

    read = ao0.get()
    print("read:", read)
    data = [[1.0, 2.0], [3.0, 4.0]]
    ao0.set(data)
    read = ao0.get()
    print("read:", read)
