# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_ChanPerLine, DAQmx_Val_Volts
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageOutputSingleLine(DelegateMixin):
    """
    ``AnalogVoltageOutputSingleLine`` is used to access a single analog output
    line on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(self, descriptor, minimumVoltage=-10.0, maximumVoltage=10.0):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`
            minimumVoltage=-10.0 (float): the minimum voltage.
            maximumVoltage=10.0 (float): the maximum voltage.
        """
        taskName = "AO_" + descriptor.replace("/", "_")
        self.value = None
        self.tmpvalue = float64(0.0)
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateAOVoltageChan(descriptor,
                                      None,
                                      float64(minimumVoltage),
                                      float64(maximumVoltage),
                                      DAQmx_Val_Volts,
                                      None)

    def get(self):
        """
        Return the last written analog value.
        """

        return self.value

    def set(self, value):
        """
        Set the pin value to ``value``.

        Args:
            value (float): new state for output pin.
        """
        self.value = value

        self.task.WriteAnalogScalarF64(True,
                                       float64(1.0),
                                       float64(value),
                                       None)
        self.task.StopTask()

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    ao0 = AnalogVoltageOutputSingleLine("Dev1/ao0", minimumVoltage=0, maximumVoltage=5.0)

    for ii in range(100):
        ao0.set(0.0)
        value = ao0.get()
        print(value)
        # ao1.set(3.0)
        # time.sleep(1.0)
        # ao0.set(3.0)
        # ao1.set(0.0)
        # time.sleep(1.0)
