# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_GroupByChannel
from htf.mixins.delegation import DelegateMixin


class DigitalInputPort(DelegateMixin):
    """
    ``DigitalInputPort`` is used to access a digital input port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/port0`.
        """
        taskName = "DI_" + descriptor.replace("/", "_")
        self.state = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateDIChan(descriptor, None, DAQmx_Val_ChanForAllLines)

    def get(self):
        """
        Return the current port state.
        """
        self.task.ReadDigitalU32(self.N,
                                 float64(1.0),
                                 DAQmx_Val_GroupByChannel,
                                 self.data,
                                 self.N,
                                 byref(self.read),
                                 None)
        return self.data[0]

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time
    p0 = DigitalInputPort("Dev1/port0")

    for ii in range(1000):
        print(p0.get())
        time.sleep(.5)
