# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals


from htf.mixins.delegation import DelegatorMixin
from .core import Request, Response, OutputPin, InputPin, SlipCommunication
import logging
logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)


class BaseBoard(DelegatorMixin):
    VERSION = 2
    BOARDNAME = "not set"
    SETTLE_TIME = 0.0

    def __init__(self, port, timeout=1.0):
        """
        Initialize a board.

        Args:
            port (str): the name of the serial connection
            timeout=1.0 (float): the time-out in seconds
        """
        self._port = port

        self._communication = SlipCommunication(self._port, self.SETTLE_TIME)

        self.request = Request()
        self.response = Response()

        self._initialize_communication()

        self._check_version()
        self._check_board()

    def close(self):
        """
        Close the serial connection.
        """
        self._communication.close()

    def __del__(self):
        """
        Automatically close serial connection on deletion.
        """
        if hasattr(self, "_communication"):
            self.close()

    def _communicate(self, request, retries=2):
        encoded_request = request.encode()
        # print("request:", request)
        # print("bin:", repr(encoded_request))
        # print("frame:", to_hex(encoded_request))
        self._communication.send(encoded_request)
        # try to receive the frame multiple times
        # on Windows with Controllino 1 .. 2 \xf0 are sent when
        # the port is opened. This helps.
        for retry in range(retries+1):
            try:
                # print("try:", retry)
                response = self._communication.receive()
                # print("response:", oser.to_hex(response))
                self.response.decode(response)
                #print(self.response)
                return self.response
            except:
                pass
        return None

    def _initialize_communication(self):
        """
        On Windows 7 and 10 with Controllino the
        device first recieved one or two \xf0 (nobody knows why).
        This leads to a framing error.
        This initialization sequence is a workaround.
        """
        self._communication.send_byte(b"\xc0")
        response = self._communication.receive()

    def digital_input(self, pin):
        """
        Read a digital input pin.

        Args:
            pin (int): the pin to be read

        Returns:
            int: the pin state of ``pin``
        """
        self.request.command.set("digital_input")
        self.request.parameters.pin.set(pin)
        response = self._communicate(self.request)
        return response.result.get()

    def digital_input_pullup(self, pin):
        """
        Read a digital input pin with internal pullup enabled.

        Args:
            pin (int): the pin to be read

        Returns:
            int: the pin state of ``pin``
        """
        self.request.command.set("digital_input_pullup")
        self.request.parameters.pin.set(pin)
        response = self._communicate(self.request)
        return response.result.get()

    def digital_output(self, pin, state):
        """
        Write a digital output pin.

        Args:
            pin (int): the pin to be written
            state (int): the state for ``pin``
        """
        self.request.command.set("digital_output")
        self.request.parameters.pin.set(pin)
        self.request.parameters.state.set(state)
        response = self._communicate(self.request)
        assert response.result.get() == 0

    def analog_input(self, pin):
        """
        Read an analog input pin.

        Args:
            pin (int): the pin to be read

        Returns:
            int: the analog value of ``pin``
        """
        self.request.command.set("analog_input")
        self.request.parameters.pin.set(pin)
        response = self._communicate(self.request)
        return response.result.get()

    def set_analog_reference(self, reference_type):
        """
        Set the analog reference voltage.

        Args:
            reference_type (str): type of reference to use ("DEFAULT",
                "INTERNAL" ATmega8 only, "INTERNAL1V1" Arduino Mega only,
                "INTERNAL2V56", or "EXTERNAL").
        """
        reference_types = {
            "DEFAULT": 0,
            "INTERNAL": 1,  # ATmega8 only,
            "INTERNAL1V1": 2,  # Arduino Mega only,
            "INTERNAL2V56": 3,
            "EXTERNAL": 4,
        }

        if reference_type not in reference_types:
            raise Exception("Unknown analog voltage reference type")

        value = reference_types[reference_type]
        self.request.command.set("analog_reference")
        self.request.parameters.set(value)
        response = self._communicate(self.request)
        if response.result.get() != value:
            raise Exception("Setting analog voltage reference did not work")

    def pwm(self, pin, value):
        """
        Set ``pin`` to pwm-level ``value``.

        Args:
            pin (int): the pin to be used
            value (int): the pwm level (0 .. 255)
        """
        self.request.command.set("pwm")
        self.request.parameters.pin.set(pin)
        self.request.parameters.state.set(value)
        response = self._communicate(self.request)
        assert response.result.get() == 0

    def analog_output(self, pin, value):
        """
        Set ``pin`` to analog level ``value``.
        This is useful if ``pin`` is connected
        to a low-pass filter.

        Args:
            pin (int): the pin to be used
            value (int): the analog output level (0 .. 255)
        """
        self.pwm(pin=pin, value=value)

    def get_version(self):
        """
        Return the version read from the device.
        """
        self.request.command.set("get_version")
        response = self._communicate(self.request)
        return response.result.get()

    def _check_version(self):
        version = self.get_version()
        if version != self.VERSION:
            raise Exception("Wrong protocol version: expected %i but read %i."
                            % (self.VERSION, version))

    def get_board_name(self):
        """
        Return the board name read from the device.
        """
        self.request.command.set("get_board_name")
        response = self._communicate(self.request)
        return response.result.get()

    def _check_board(self):
        boardname = self.get_board_name()
        if boardname != self.BOARDNAME:
            raise Exception("Wrong board: expected %s but read %s."
                            % (repr(self.BOARDNAME), repr(boardname)))

    def restart(self):
        """
        Invoke a software restart.
        """
        raise Exception("restart does not work properly")
        self.request.command.set("restart")
        response = self._communicate(self.request)
        return response.result.get()

    def uptime(self):
        """
        Return the uptime in milliseconds
        """
        self.request.command.set("uptime")
        response = self._communicate(self.request)
        return response.result.get()


# See https://github.com/CONTROLLINO-PLC/CONTROLLINO_Library

class IOMini(BaseBoard):
    BOARDNAME = b"IO Mini"
    SETTLE_TIME = 0.0

    def __init__(self, port, timeout=1.0):
        """
        Initialize an IOMini.

        Args:
            port (str): the name of the serial connection
            timeout=1.0 (float): the time-out in seconds
        """
        super(IOMaxi, self).__init__(port=port, timeout=timeout)

        # define D0 4
        self.D0 = OutputPin(4, self.digital_output)  #: Digital output D0

        # define D1 5
        self.D1 = OutputPin(5, self.digital_output)  #: Digital output D1

        # define D2 6
        self.D2 = OutputPin(6, self.digital_output)  #: Digital output D2

        # define D3 7
        self.D3 = OutputPin(7, self.digital_output)  #: Digital output D3

        # define D4 8
        self.D4 = OutputPin(8, self.digital_output)  #: Digital output D4

        # define D5 9
        self.D5 = OutputPin(9, self.digital_output)  #: Digital output D5

        # define D6 18
        self.D6 = OutputPin(18, self.digital_output)  #: Digital output D6

        # define D7 19
        self.D7 = OutputPin(19, self.digital_output)  #: Digital output D7

        # define A0 14
        self.A0 = InputPin(14, self.analog_input)  #: Analog input A0

        # define A1 15
        self.A1 = InputPin(15, self.analog_input)  #: Analog input A1

        # define A2 16
        self.A2 = InputPin(16, self.analog_input)  #: Analog input A2

        # define A3 17
        self.A3 = InputPin(17, self.analog_input)  #: Analog input A3

        # define A4 20
        self.A4 = InputPin(20, self.analog_input)  #: Analog input A4

        # define A5 21
        self.A5 = InputPin(21, self.analog_input)  #: Analog input A5

        # define IN0 2
        self.IN0 = InputPin(2, self.digital_input)  #: Digital input IN0

        # define IN1 3
        self.IN1 = InputPin(3, self.digital_input)  #: Digital input IN1


class IOMaxi(BaseBoard):
    BOARDNAME = b"IO Maxi"
    SETTLE_TIME = 0.0

    def __init__(self, port, timeout=1.0):
        """
        Initialize an IOMaxi.

        Args:
            port (str): the name of the serial connection
            timeout=1.0 (float): the time-out in seconds
        """
        super(IOMaxiAutomation, self).__init__(port=port, timeout=timeout)

        # define D0 2
        self.D0 = OutputPin(2, self.digital_output)  #: Digital output D0

        # define D1 3
        self.D1 = OutputPin(3, self.digital_output)  #: Digital output D1

        # define D2 4
        self.D2 = OutputPin(4, self.digital_output)  #: Digital output D2

        # define D3 5
        self.D3 = OutputPin(5, self.digital_output)  #: Digital output D3

        # define D4 6
        self.D4 = OutputPin(6, self.digital_output)  #: Digital output D4

        # define D5 7
        self.D5 = OutputPin(7, self.digital_output)  #: Digital output D5

        # define D6 8
        self.D6 = OutputPin(8, self.digital_output)  #: Digital output D6

        # define D7 9
        self.D7 = OutputPin(9, self.digital_output)  #: Digital output D7

        # define D8 10
        self.D8 = OutputPin(10, self.digital_output)  #: Digital output D8

        # define D9 11
        self.D9 = OutputPin(11, self.digital_output)  #: Digital output D9

        # define D10 12
        self.D10 = OutputPin(12, self.digital_output)  #: Digital output D10

        # define D11 13
        self.D11 = OutputPin(13, self.digital_output)  #: Digital output D11

        # define A0 54
        self.A0 = InputPin(54, self.analog_input)  #: Analog input A0

        # define A1 55
        self.A1 = InputPin(55, self.analog_input)  #: Analog input A1

        # define A2 56
        self.A2 = InputPin(56, self.analog_input)  #: Analog input A2

        # define A3 57
        self.A3 = InputPin(57, self.analog_input)  #: Analog input A3

        # define A4 58
        self.A4 = InputPin(58, self.analog_input)  #: Analog input A4

        # define A5 59
        self.A5 = InputPin(59, self.analog_input)  #: Analog input A5

        # define A6 60
        self.A6 = InputPin(60, self.analog_input)  #: Analog input A6

        # define A7 61
        self.A7 = InputPin(61, self.analog_input)  #: Analog input A7

        # define A8 62
        self.A8 = InputPin(62, self.analog_input)  #: Analog input A8

        # define A9 63
        self.A9 = InputPin(63, self.analog_input)  #: Analog input A9

        # define IN0 18
        self.IN0 = InputPin(18, self.digital_input)  #: Digital input IN0

        # define IN1 19
        self.IN1 = InputPin(19, self.digital_input)  #: Digital input IN1

        # define R0 22
        self.R0 = OutputPin(22, self.digital_output)  #: Relay R0

        # define R1 23
        self.R1 = OutputPin(23, self.digital_output)  #: Relay R1

        # define R2 24
        self.R2 = OutputPin(24, self.digital_output)  #: Relay R2

        # define R3 25
        self.R3 = OutputPin(25, self.digital_output)  #: Relay R3

        # define R4 26
        self.R4 = OutputPin(26, self.digital_output)  #: Relay R4

        # define R5 27
        self.R5 = OutputPin(27, self.digital_output)  #: Relay R5

        # define R6 28
        self.R6 = OutputPin(28, self.digital_output)  #: Relay R6

        # define R7 29
        self.R7 = OutputPin(29, self.digital_output)  #: Relay R7

        # define R8 30
        self.R8 = OutputPin(30, self.digital_output)  #: Relay R8

        # define R9 31
        self.R9 = OutputPin(31, self.digital_output)  #: Relay R9


class IOMaxiAutomation(BaseBoard):
    BOARDNAME = b"IO Maxi Automation"
    SETTLE_TIME = 0.0

    def __init__(self, port, timeout=1.0):
        """
        Initialize an IOMaxiAutomation.

        Args:
            port (str): the name of the serial connection
            timeout=1.0 (float): the time-out in seconds
        """
        super(IOMaxiAutomation, self).__init__(port=port, timeout=timeout)

        # define DO0 2
        self.DO0 = OutputPin(2, self.digital_output)  #: Digital output DO0

        # define DO1 3
        self.DO1 = OutputPin(3, self.digital_output)  #: Digital output DO1

        # define DO2 4
        self.DO2 = OutputPin(4, self.digital_output)  #: Digital output DO2

        # define DO3 5
        self.DO3 = OutputPin(5, self.digital_output)  #: Digital output DO3

        # define DO4 6
        self.DO4 = OutputPin(6, self.digital_output)  #: Digital output DO4

        # define DO5 7
        self.DO5 = OutputPin(7, self.digital_output)  #: Digital output DO5

        # define DO6 8
        self.DO6 = OutputPin(8, self.digital_output)  #: Digital output DO6

        # define DO7 9
        self.DO7 = OutputPin(9, self.digital_output)  #: Digital output DO7

        # define DI0 66
        self.DI0 = InputPin(66, self.digital_input)  #: Digital input DI0

        # define DI1 67
        self.DI1 = InputPin(67, self.digital_input)  #: Digital input DI1

        # define DI2 10
        self.DI2 = InputPin(10, self.digital_input)  #: Digital input DI2

        # define DI3 11
        self.DI3 = InputPin(11, self.digital_input)  #: Digital input DI3

        # define AO0 12
        self.AO0 = OutputPin(12, self.pwm)  #: Analog output AO0

        # define AO1 13
        self.AO1 = OutputPin(13, self.pwm)  #: Analog output AO1

        # define AI0 54
        self.AI0 = InputPin(54, self.analog_input)  #: Analog input AI0

        # define AI1 55
        self.AI1 = InputPin(55, self.analog_input)  #: Analog input AI1

        # define AI2 56
        self.AI2 = InputPin(56, self.analog_input)  #: Analog input AI2

        # define AI3 57
        self.AI3 = InputPin(57, self.analog_input)  #: Analog input AI3

        # define AI4 58
        self.AI4 = InputPin(58, self.analog_input)  #: Analog input AI4

        # define AI5 59
        self.AI5 = InputPin(59, self.analog_input)  #: Analog input AI5

        # define AI6 60
        self.AI6 = InputPin(60, self.analog_input)  #: Analog input AI6

        # define AI7 61
        self.AI7 = InputPin(61, self.analog_input)  #: Analog input AI7

        # define AI8 62
        self.AI8 = InputPin(62, self.analog_input)  #: Analog input AI8

        # define AI9 63
        self.AI9 = InputPin(63, self.analog_input)  #: Analog input AI9

        # define AI10 64
        self.AI10 = InputPin(64, self.analog_input)  #: Analog input AI10

        # define AI11 65
        self.AI11 = InputPin(65, self.analog_input)  #: Analog input AI11

        # define AI12 68
        self.AI12 = InputPin(68, self.analog_input)  #: Analog input AI12

        # define AI13 69
        self.AI13 = InputPin(69, self.analog_input)  #: Analog input AI13

        # define IN1 18
        self.IN1 = InputPin(18, self.digital_input)  #: Digital input IN1

        # define IN2 19
        self.IN2 = InputPin(19, self.digital_input)  #: Digital input IN2

        # define R0 22
        self.R0 = OutputPin(22, self.digital_output)  #: Relay R0

        # define R1 23
        self.R1 = OutputPin(23, self.digital_output)  #: Relay R1

        # define R2 24
        self.R2 = OutputPin(24, self.digital_output)  #: Relay R2

        # define R3 25
        self.R3 = OutputPin(25, self.digital_output)  #: Relay R3

        # define R4 26
        self.R4 = OutputPin(26, self.digital_output)  #: Relay R4

        # define R5 27
        self.R5 = OutputPin(27, self.digital_output)  #: Relay R5

        # define R6 28
        self.R6 = OutputPin(28, self.digital_output)  #: Relay R6

        # define R7 29
        self.R7 = OutputPin(29, self.digital_output)  #: Relay R7

        # define R8 30
        self.R8 = OutputPin(30, self.digital_output)  #: Relay R8

        # define R9 31
        self.R9 = OutputPin(31, self.digital_output)  #: Relay R9


class IOMega(BaseBoard):
    BOARDNAME = b"IO Mega"
    SETTLE_TIME = 0.0

    def __init__(self, port, timeout=1.0):
        """
        Initialize an IOMega.

        Args:
            port (str): the name of the serial connection
            timeout=1.0 (float): the time-out in seconds
        """
        super(IOMega, self).__init__(port=port, timeout=timeout)

        # define D0 2
        self.D0 = OutputPin(2, self.digital_output)  #: Digital output D0

        # define D1 3
        self.D1 = OutputPin(3, self.digital_output)  #: Digital output D1

        # define D2 4
        self.D2 = OutputPin(4, self.digital_output)  #: Digital output D2

        # define D3 5
        self.D3 = OutputPin(5, self.digital_output)  #: Digital output D3

        # define D4 6
        self.D4 = OutputPin(6, self.digital_output)  #: Digital output D4

        # define D5 7
        self.D5 = OutputPin(7, self.digital_output)  #: Digital output D5

        # define D6 8
        self.D6 = OutputPin(8, self.digital_output)  #: Digital output D6

        # define D7 9
        self.D7 = OutputPin(9, self.digital_output)  #: Digital output D7

        # define D8 10
        self.D8 = OutputPin(10, self.digital_output)  #: Digital output D8

        # define D9 11
        self.D9 = OutputPin(11, self.digital_output)  #: Digital output D9

        # define D10 12
        self.D10 = OutputPin(12, self.digital_output)  #: Digital output D10

        # define D11 13
        self.D11 = OutputPin(13, self.digital_output)  #: Digital output D11

        # define D12 42
        self.D12 = OutputPin(42, self.digital_output)  #: Digital output D12

        # define D13 43
        self.D13 = OutputPin(43, self.digital_output)  #: Digital output D13

        # define D14 44
        self.D14 = OutputPin(44, self.digital_output)  #: Digital output D14

        # define D15 45
        self.D15 = OutputPin(45, self.digital_output)  #: Digital output D15

        # define D16 46
        self.D16 = OutputPin(46, self.digital_output)  #: Digital output D16

        # define D17 47
        self.D17 = OutputPin(47, self.digital_output)  #: Digital output D17

        # define D18 48
        self.D18 = OutputPin(48, self.digital_output)  #: Digital output D18

        # define D19 49
        self.D19 = OutputPin(49, self.digital_output)  #: Digital output D19

        # define A0 54
        self.A0 = InputPin(54, self.analog_input)  #: Analog input A0

        # define A1 55
        self.A1 = InputPin(55, self.analog_input)  #: Analog input A1

        # define A2 56
        self.A2 = InputPin(56, self.analog_input)  #: Analog input A2

        # define A3 57
        self.A3 = InputPin(57, self.analog_input)  #: Analog input A3

        # define A4 58
        self.A4 = InputPin(58, self.analog_input)  #: Analog input A4

        # define A5 59
        self.A5 = InputPin(59, self.analog_input)  #: Analog input A5

        # define A6 60
        self.A6 = InputPin(60, self.analog_input)  #: Analog input A6

        # define A7 61
        self.A7 = InputPin(61, self.analog_input)  #: Analog input A7

        # define A8 62
        self.A8 = InputPin(62, self.analog_input)  #: Analog input A8

        # define A9 63
        self.A9 = InputPin(63, self.analog_input)  #: Analog input A9

        # define A10 64
        self.A10 = InputPin(64, self.analog_input)  #: Analog input A10

        # define A11 65
        self.A11 = InputPin(65, self.analog_input)  #: Analog input A11

        # define A12 66
        self.A12 = InputPin(66, self.analog_input)  #: Analog input A12

        # define A13 67
        self.A13 = InputPin(67, self.analog_input)  #: Analog input A13

        # define A14 68
        self.A14 = InputPin(68, self.analog_input)  #: Analog input A14

        # define A15 69
        self.A15 = InputPin(69, self.analog_input)  #: Analog input A15

        # define I16 38
        self.I16 = InputPin(38, self.digital_input)  #: Digital input I16

        # define I17 39
        self.I17 = InputPin(39, self.digital_input)  #: Digital input I17

        # define I18 40
        self.I18 = InputPin(40, self.digital_input)  #: Digital input I18

        # define IN0 18
        self.IN0 = InputPin(18, self.digital_input)  #: Digital input IN0

        # define IN1 19
        self.IN1 = InputPin(19, self.digital_input)  #: Digital input IN1

        # define R0 22
        self.R0 = OutputPin(22, self.digital_output)  #: Relay R0

        # define R1 23
        self.R1 = OutputPin(23, self.digital_output)  #: Relay R1

        # define R2 24
        self.R2 = OutputPin(24, self.digital_output)  #: Relay R2

        # define R3 25
        self.R3 = OutputPin(25, self.digital_output)  #: Relay R3

        # define R4 26
        self.R4 = OutputPin(26, self.digital_output)  #: Relay R4

        # define R5 27
        self.R5 = OutputPin(27, self.digital_output)  #: Relay R5

        # define R6 28
        self.R6 = OutputPin(28, self.digital_output)  #: Relay R6

        # define R7 29
        self.R7 = OutputPin(29, self.digital_output)  #: Relay R7

        # define R8 30
        self.R8 = OutputPin(30, self.digital_output)  #: Relay R8

        # define R9 31
        self.R9 = OutputPin(31, self.digital_output)  #: Relay R9

        # define R10 32
        self.R10 = OutputPin(32, self.digital_output)  #: Relay R10

        # define R11 33
        self.R11 = OutputPin(33, self.digital_output)  #: Relay R11

        # define R12 34
        self.R12 = OutputPin(34, self.digital_output)  #: Relay R12

        # define R13 35
        self.R13 = OutputPin(35, self.digital_output)  #: Relay R13

        # define R14 36
        self.R14 = OutputPin(36, self.digital_output)  #: Relay R14

        # define R15 37
        self.R15 = OutputPin(37, self.digital_output)  #: Relay R15
