# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import sys
import argparse
import webbrowser
import os.path
import htf
from .runner import TestRunner
from .suite import TestSuite
import htf.loader as loader
import time
import platform
import re
from unittest.signals import installHandler
import imp


class main(object):
    """
    The htf.main() and ``htf`` command-line utility are used to
    run a set of tests.
    """
    # defaults for testing
    verbosity = 1
    failfast = catchbreak = buffer = progName = warnings = None
    test = tests = testNames = title = None
    html_report = junit_xml_report = json_report = extra_report = None

    def __init__(self,
                 title='Testreport',
                 html_report=None,
                 junit_xml_report=None,
                 json_report=None,
                 extra_report=None,
                 tests='__main__',
                 pattern="test*.py",
                 exit=True,
                 verbosity=1,
                 failfast=None,
                 catchbreak=None,
                 argv=None,
                 parse_args=False):
        """
        Args:
            title (str): the test's title
            html_report=None (str or list of str): HTML-report filename(s)
            junit_xml_report=None (str or list of str): JUnit-XML-report
                filename(s)
            json_report=None (str or list of str): JSON-report
                filename(s)
            extra_report=None (instance or list of instances): Extra reports
                that support instance.render(data). This is useful for
                :class:`htf.DOORSTestReport` for example that needs
                additional parameters.
            tests=None (test-specifier or list of test-specifiers): a
                test-specifier ora list of test-specifiers
                (folder, file, test-case, test-case-method, module)
            pattern="test*.py" (str): the pattern to be used when searching for
                python files in folders
            exit=True (bool): if set to ``True`` the exit() is called at the
                end of the tests
            verbosity=1 (int): the verbosity level (the higher the level the
                more verbose is the output)
            failfast=None (bool): if set to ``True`` the test run ends after
                the first failing test
            catchbreak=None (bool): if set to ``True`` CTRL-C is catched wile
                tests run. If CTRL-C is pressed the test run is stopped after
                the current test.
            argv=None: the arguments to be parsed (useful for testing).
                If set to ``None`` ``sys.argv`` is used instead.
            parse_args=False: if set to ``True`` ``argv`` or ``sys.argv`` is
                parsed using argparse (useful as a command-line utility)
        """
        if argv is None:
            argv = sys.argv

        if html_report is None:
            html_report = []
        self.html_report = html_report

        if junit_xml_report is None:
            junit_xml_report = []
        self.junit_xml_report = junit_xml_report

        if json_report is None:
            json_report = []
        self.json_report = json_report

        self.extra_report = extra_report

        if not isinstance(tests, list):
            tests = [tests]

        self.title = title
        self.exit = exit
        self.failfast = failfast
        self.catchbreak = catchbreak
        self.verbosity = verbosity
        self.parse_args = parse_args
        self.tests = tests
        self.pattern = pattern
        self.testLoader = loader.defaultTestLoader
        self.progName = os.path.basename(argv[0])
        self.parseArgs(argv)

    def parseArgs(self, argv):
        if self.parse_args is True:
            if len(argv) == 1:
                # $ htf == htf run
                argv.append("run")

            if argv[1] not in ["run", "licenses", "docs", "version"]:
                argv = argv[:1] + ["run"] + argv[1:]
            self._initArgParsers()
            self._main_parser.parse_args(argv[1:], self)

            if self.command == "docs":
                path_to_docs = os.path.join(os.path.dirname(htf.__file__),
                                            "doc", "html", "index.html")
                webbrowser.open("file://" + path_to_docs)
                exit(0)

            if self.command == "licenses":
                print(LICENSES)
                exit(0)

            if self.command == "version":
                print("htf-" + htf.__version__)
                exit(0)

            assert self.command == "run"

        self._do_discovery()

        self.runTests()

    def _initArgParsers(self):
        self._main_parser = argparse.ArgumentParser(
            description='htf - HILSTER Testing Framework (HILSTER GmbH, http://hilster.de)')

        self._subparsers = self._main_parser.add_subparsers(dest="command",
                                                            help='the command')
        self._subparsers.required = True

        self._run_parser = self._subparsers.add_parser('run', help='run tests')
        self._addDefaultOptions(self._run_parser)
        self._run_parser.add_argument('-p', '--pattern', default="test*.py",
                                      help="pattern to match test files (default: test*.py)")
        self._run_parser.add_argument('tests', nargs='*', default=["."],
                                      help='a list of any number of test modules, '
                                      'classes and test methods.')

        self._docs_parser = self._subparsers.add_parser('docs', help='open documents')

        self._licenseParser = self._subparsers.add_parser('licenses', help="show open-source licenses")

        self._versionParser = self._subparsers.add_parser('version', help="show version")

    def _addDefaultOptions(self, parser):
        parser.add_argument('-T', '--title', default="Testreport", help="the testreport title")
        parser.add_argument('-H', '--html-report', action="append", help="HTML report filename (may be repeated)")
        parser.add_argument('-x', '--junit-xml-report', action="append",
                            help="Junit-compatible XML report filename (may be repeated)")
        parser.add_argument('-j', '--json-report', action="append", help="JSON report filename (may be repeated)")

        parser.add_argument('-v', '--verbose', dest='verbosity',
                            action='store_const', const=2, default=1,
                            help='verbose output')
        parser.add_argument('-c', '--catch', dest='catchbreak', action="store_true", default=False,
                            help="CTRL-C cancels the current test and shows what was run so far")
        parser.add_argument('-f', '--failfast', action="store_true", default=False,
                            help="stop the test run on the first error or failure.")

    def _collect_tests_from_identifier(self, test):
        loader = self.testLoader
        # if test is a directory, discover all tests within it
        if not isinstance(test, (type(""), bytes)):
            # users can also pass instances, methods, etc.
            try:
                test = loader.loadTestsFromTestCase(test)
                print("loading tests from testcase", test)
            except TypeError:
                try:
                    # try to load it as a method from its __qualname__
                    if hasattr(test, "__qualname__"):  # pragma: no cover
                        # Python 3.5+
                        qualname = test.__module__ + "." + test.__qualname__
                    else:  # pragma: no cover
                        # Python 2.7
                        qualname = test.im_class.__module__ + "." + test.im_class.__name__ + "." + test.__name__
                        # qualname =
                    print("loading test", qualname)
                    test = loader.loadTestsFromName(qualname)
                except AttributeError:
                    # try to load as a module:
                    print("loading tests from module:", test)
                    test = loader.loadTestsFromModule(test)

        elif os.path.isdir(test):
            print("discovering tests from", test)
            test = loader.discover(test, self.pattern)
        elif os.path.isfile(test):
            print("collecting tests from", test)
            name, ext = os.path.splitext((os.path.basename(test)))
            module = imp.load_source(name, test)
            test = loader.loadTestsFromModule(module)
        else:
            print("collecting tests from importable module", test)

            try:
                module = __import__(test)
                for part in test.split('.')[1:]:
                    module = getattr(module, part)
                test = loader.loadTestsFromModule(module)
            except ImportError:
                test = loader.loadTestsFromName(test)
        return test

    def _do_discovery(self):
        """
        Test discovery.

        Sets self.test
        """
        loader = self.testLoader

        self.test = TestSuite()

        for test in self.tests:
            suite = self._collect_tests_from_identifier(test)
            self.test.addTest(suite)

    def _to_filename(self, part):
        """
        Returns the part as a filename part
        """
        ret = re.sub(pattern=r"[^A-Za-z0-9_-]", repl="_", string=part)
        return ret

    def _report_filename(self, template):
        """
        Returns a report filename depending on the user input template.

        Replacements:
            - {{title}} - the title as a filename
            - {{date}} - the current date "%Y-%m-%d_%H-%M-%S"
            - {{hostname}}, {{host}} or {{node}}- the hostname
            - {{htf_version}} or {{version}} - the htf version, eg. htf_1.0.0
            - {{python_version}} - the python version, eg. Python_3.5.1
        """
        current_date = time.strftime("%Y-%m-%d_%H-%M-%S")
        node = platform.node()
        python_version = "Python_" + platform.python_version()
        python_version = self._to_filename(python_version)
        htf_version = "htf_" + htf.__version__
        htf_version = self._to_filename(htf_version)

        title = self._to_filename(self.title)

        template = template.replace("{{title}}", title)

        template = template.replace("{{date}}", current_date)

        template = template.replace("{{hostname}}", node)
        template = template.replace("{{host}}", node)
        template = template.replace("{{node}}", node)

        template = template.replace("{{python_version}}", python_version)

        template = template.replace("{{htf_version}}", htf_version)
        template = template.replace("{{version}}", htf_version)

        return template

    def _get_html_reports(self):
        """
        Returns a list of instances of :class:`htf.HTMLTestReport`
        depending on the options.
        """
        html_reports = []
        if self.html_report is not None:
            if not isinstance(self.html_report, list):
                self.html_report = [self.html_report]
            for filename in self.html_report:
                if not isinstance(filename, type("")):
                    filename = filename.decode()
                # do replacements in filename
                filename = self._report_filename(template=filename)
                # create instance
                filename = htf.HTMLTestReport(filename=filename)
                html_reports.append(filename)
        return html_reports

    def _get_junit_xml_reports(self):
        """
        Returns a list of instances of :class:`htf.JUnitXMLTestReport`
        depending on the options.
        """
        junit_xml_reports = []
        if self.junit_xml_report is not None:
            if not isinstance(self.junit_xml_report, list):
                self.junit_xml_report = [self.junit_xml_report]
            for filename in self.junit_xml_report:
                if not isinstance(filename, type("")):
                    filename = filename.decode()
                # do replacements in filename
                filename = self._report_filename(template=filename)
                # create instance
                filename = htf.JUnitXMLTestReport(filename=filename)
                junit_xml_reports.append(filename)
        return junit_xml_reports

    def _get_json_reports(self):
        """
        Returns a list of instances of :class:`htf.JSONTestReport`
        depending on the options.
        """
        json_reports = []
        if self.json_report is not None:
            if not isinstance(self.json_report, list):
                self.json_report = [self.json_report]
            for filename in self.json_report:
                if not isinstance(filename, type("")):
                    filename = filename.decode()
                # do replacements in filename
                filename = self._report_filename(template=filename)
                # create instance
                filename = htf.JSONTestReport(filename=filename)
                json_reports.append(filename)
        return json_reports

    def _get_extra_report(self):
        if self.extra_report is None:
            return []
        if not isinstance(self.extra_report, list):
            return [self.extra_report]
        return self.extra_report

    def runTests(self):
        if self.catchbreak:
            installHandler()

        reports = self._get_html_reports() \
            + self._get_junit_xml_reports() \
            + self._get_json_reports() \
            + self._get_extra_report()

        # print("verbosity:", self.verbosity)
        testRunner = TestRunner(title=self.title,
                                reports=reports,
                                verbosity=self.verbosity,
                                failfast=self.failfast)

        self.result = testRunner.run(self.test)

        if self.exit:
            sys.exit(int(not self.result.wasSuccessful()))


TestProgram = main

LICENSES = """
Open Source Licenses
====================

The Sodium Crypto library (libsodium-1.0.10, http://libsodium.org)
------------------------------------------------------------------

Copyright (c) 2013-2016
Frank Denis <j at pureftpd dot org>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""
