# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import (absolute_import,
                        division,
                        print_function,
                        unicode_literals)

from htf.core import TestCase as _TestCase


class TestCase(_TestCase):
    """
    ``TestCase`` is the basic test case class for the HILSTER Testing Framework
    that should be used using inheritance.
    """

    def __init__(self, methodName='runTest'):
        """
        Args:
            methodName='runTest' (str): the name of the test method to be run
                by the test runner.
        """
        return super(TestCase, self).__init__(methodName=methodName)

    def setUp(self):
        """
        This method is called from test runner before a test is executed to
        create the test-fixtures.

        If you override this method, make sure to call the original
        implementation::

            def setUp(self):
                super(TestCase, self).setUp()
                 # your code goes here
        """
        return super(TestCase, self).setUp()

    def afterTest(self):
        """
        This method is called after a test ends and before
        :py:func:`htf.unittest.testcase.TestCase.tearDown` is called.
        This method is injected using the :py:func:`htf.decorators.after`
        decorator to be run in the test context.

        If you override this method, make sure to call the original
        implementation::

            class SpecializedTestCase(TestCase):

                def afterTest(self):
                    # your code goes here
                    super(TestCase, self).afterTest()
        """
        return super(TestCase, self).afterTest()

    def tearDown(self):
        """
        This method is called from test runner after a test was executed
        regardless of it's result.

        This method shall destroy all used fixtures.

        If you override this method, make sure to call the original
        implementation::

            def tearDown(self):
                # your code goes here
                super(TestCase, self).tearDown()
        """
        return super(TestCase, self).tearDown()

    def runBackground(self, func, name=None, joinTimeout=1, args=None,
                      kwargs=None):
        """
        With this method you can run another method in background.

        If this method uses a while loop, it must poll
        ``self._runThreadsEnable`` to be able to be terminated.
        If the background thread does not terminate, the foreground thread
        never ends!

        Args:
            func (callable): the method to be run in background.
            name=None (str): a name for the background task.
                Default: ``None``. If `name` is ``None`` `func.__name__` is
                used.
            joinTimeout=1 (float): the timeout to be used when calling ``join``
                on the thread.
            args=None (tuple): a tuple of positional arguments that are passed
                to the background task when it is called.
            kwargs=None (dict): a dictionary of keyword arguments that are
                passed to the background task when it is called.
        """
        return super(TestCase, self).runBackground(
            func=func,
            name=name,
            joinTimeout=joinTimeout,
            args=args,
            kwargs=kwargs
        )

    def runPeriodic(self,
                    func,
                    period,
                    maximumPeriod=None,
                    name=None,
                    raiseException=True,
                    args=None,
                    kwargs=None):
        """
        With this method you can run another method periodically
        in background with a given period.

        Args:
            func (callable): the method to be called periodically in
                background.
            period (float): the period in second.
            maximumPeriod=None (float): if set to a float >= ``period``
                ``func`` may take up to ``maximumPeriod`` time without
                raising an exception. A warning is printed out to stdout
                instead.
            name=None (str): a name for the background task.
            raiseExceptions=True: if set to ``True`` exceptions are put into
                the exception queue to let a test fail if a background thread
                fails.
            args=None (tuple): a tuple of positional arguments that are passed
                to the background task when it is called.
            kwargs=None (dict): a dictionary of keyword arguments that are
                passed to the background task when it is called.

        Returns:
            stop-trigger (callable): a callable object is returned that stops execution when called.
        """
        return super(TestCase, self).runPeriodic(
            func=func,
            period=period,
            maximumPeriod=maximumPeriod,
            name=name,
            raiseException=raiseException,
            args=args,
            kwargs=kwargs
        )

    def stopThread(self, nameOrFunction):
        """
        This method stops a specific thread with the name or
        function described by ``nameOrFunction`` by calling
        :func:`~threading.Thread.join` with join timeout supplied
        by :func:`~htf.TestCase.runPeriodic` or
        :func:`~htf.TestCase.runBackground` on it.
        It does not set a stop condition so the thread must end and must
        not include an endless loop.
        If the thread is still alive after the first join messages are
        printed to inform the user.

        Args:
            nameOrFunction (str, callable): the name or the function
                (reference) of the thread to be stopped.

        Returns:
            int: the number of stopped threads.

        Raises:
            Exception: if the thread was not found
        """
        return super(TestCase, self).stopThread(nameOrFunction=nameOrFunction)

    def stopThreads(self):
        """
        This method stops all background threads by setting
        ``self._runThreadsEnable`` to ``False`` and calling ``thread.join()``
        until it is not alive anymore.
        """
        return super(TestCase, self).stopThreads()

    def step(self, title, callableObj=None, *args, **kwargs):
        """
        Run a test step that is documented in the test report.

        Args:
            title (str): the step's title.
            callableObj=None (callable): a callable object that is called
                within the step or ``None``.
            *args (tuple): a tuple of positional arguments for the step's
                context.
            **kwargs (dict): a dictionary of keyword arguments for the step's
                context.

        Usage:

        .. code-block:: python

            with self.step(title="Description of this test step .."):
                # methodCall()
                self.assertTrue(True "Message in case of an error")

        Or an equivalent approach:

        .. code-block:: python

            self.step(title="Description of this test step comes here ..",
                self.assertTrue, True, "Message in case of an error")
            # method calls are not possible here!
        """
        return super(TestCase, self).step(title, callableObj, *args, **kwargs)

    def log(self, *args):
        """
        Log ``*args`` to stdout and into the result of the given step if run in
        a step context block. This method is not thread-safe!

        Args:
            *args (iterable): a tuple of python objects that must be
                convertible into a string.
                All parameters are concatenated with spaces and converted to
                string before.
        """
        return super(TestCase, self).log(*args)

    def expect(self, expectedValue, observedValue, errorMessage=None):
        """
        Expect that ``expectedValue`` and ``observedValue`` are the same.
        If ``expectedValue`` and ``observedValue`` differ an ``AssertionError``
        is raised.

        Args:
            expectedValue: the expected value.
            observedValue: the observed value that is compared to
                ``expectedValue``.
            errorMessage=None (str): an optional message.
        """
        return super(TestCase, self).expect(
            expectedValue=expectedValue,
            observedValue=observedValue,
            errorMessage=errorMessage
        )

    def expectStep(self, title, expectedValue, observedValue,
                   errorMessage=None):
        """
        Creates a test step in test report and on stdout using ``title``.
        Expect that `expectedValue` and `observedValue` are the same.
        If ``expectedValue`` and ``observedValue`` differ an ``AssertionError``
        is raised.

        Args:
            title (str): the step's title.
            expectedValue: the value that is expected.
            observedValue: the current observed value that is compared to
                ``expectedValue``.
            errorMessage=None (str): an optional message.
        """
        return super(TestCase, self).expectStep(
            title=title,
            expectedValue=expectedValue,
            observedValue=observedValue,
            errorMessage=errorMessage
        )

    def setOutputCaptureEnable(self, enabled=True):
        """
        Enables or disables capturing of stdout and stderr for later
        use in test reports.

        This must be called *before* :func:`htf.TestCase.run` is called for example
        in :func:`htf.TestCase.__init__`.

        Args:
            enabled=True (bool): the new enabled state
        """
        return super(TestCase, self).setOutputCaptureEnable(enabled=enabled)
