# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import oser
from htf.util.prompt import prompt
try:
    from prompt_toolkit.contrib.completers import WordCompleter
except ImportError:
    from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import Validator, ValidationError
import re
import logging
import copy
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class NumberValidator(Validator):
    """ Validator for number inputs

    """

    def validate(self, document):
        """
        Validates that the given document text only contains digits.

         Args:
             document (prompt_toolkit.document.Document): document to validate

         Raises:
             ValidationError: If the document contains non-digit characters
        """
        text = document.text

        if text and not text.isdigit():
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):  # pragma: no cover
                if not c.isdigit():
                    break

            raise ValidationError(
                message='This input contains non-numeric characters',
                cursor_position=i)


class FloatValidator(Validator):
    """ Validator for float inputs

    """

    def validate(self, document):
        """
        Validates that the given document text only contains a float.

         Args:
             document (prompt_toolkit.document.Document): document to validate

         Raises:
             ValidationError: If the document text doesn't match the float regex.
        """
        text = document.text

        if text and not (re.match(r"^\d*(\.\d+){0,1}$", text) or re.match(r"^[nN][aA][nN]$", text)):
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):  # pragma: no cover
                if c != "." and not c.isdigit():
                    break

            raise ValidationError(
                message='This input contains non-numeric characters',
                cursor_position=i)


def _encode_string_cast(s):
    return "{}".format(s).encode()


def _generator_enum_validator(instance):
    prototype_cast = OserPrompt()._get_cast(instance._prototype)

    def _enum_validator(x):
        try:
            x = prototype_cast(x)
        except Exception as e:
            # print("-> enum cast:", e)
            pass
        return x
    return _enum_validator


class OserPrompt(object):
    _structs = (oser.ByteStruct,)
    _enums = (oser.Enum,)
    _strings = (oser.String, oser.Data)
    _floats = (oser.BFloat,
               oser.BDouble,
               oser.LFloat,
               oser.LDouble,
               )
    _conditionals = (oser.If, oser.IfElse, oser.Switch)

    def _prompt(self, message, completer=None, validator=None):
        return prompt(message=message,
                      completer=completer,
                      validator=validator)

    def _get_class_name(self, instance):
        # defaults
        classname = instance.__class__.__name__

        if isinstance(instance, self._enums):
            classname += " of " + instance._prototype.__class__.__name__

        return classname

    def _get_choices(self, instance):
        # defaults
        choices = []

        if isinstance(instance, self._enums):
            for key, value in sorted(instance._values.items()):
                choices += ["{}".format(key), "{}".format(value)]

        return choices

    def _get_validator(self, instance):
        # defaults
        validator = NumberValidator()

        # Enum
        if isinstance(instance, self._enums):
            validator = None

        # Floats
        if isinstance(instance, self._floats):
            validator = FloatValidator()

        # Strings
        if isinstance(instance, self._strings):
            validator = None

        return validator

    def _get_cast(self, instance):
        cast = int

        # Enum
        if isinstance(instance, self._enums):
            cast = _generator_enum_validator(instance)

        # Floats
        if isinstance(instance, self._floats):
            cast = float

        # Strings
        if isinstance(instance, self._strings):
            cast = _encode_string_cast

        return cast

    def _get_wordcompleter(self, choices):
        return WordCompleter(choices)

    def fill_oser_instance(self, instance, prefix="", inputs=None):
        inputs = copy.deepcopy(inputs)
        self._fill_oser_instance(instance, prefix=prefix, inputs=inputs)

    def _fill_oser_instance(self, container, prefix="", inputs=None):
        # print("_fill", container)
        if not prefix:
            prefix = container.__class__.__name__ + "()."
        if hasattr(container, "_items"):
            for name, instance in container._items():
                next_prefix = prefix
                if not isinstance(instance, self._conditionals):
                    next_prefix = prefix + name + "."
                self._fill_oser_instance(instance, prefix=next_prefix, inputs=inputs)
        else:
            # fill a field that is not a struct
            name = prefix[:-1]
            self._fill_oser_field(name, container, "", inputs)

    def _fill_oser_field(self, name, instance, prefix="", inputs=None):
        input_done = False
        while input_done is False:
            classname = self._get_class_name(instance)
            choices = self._get_choices(instance)
            validator = self._get_validator(instance)
            cast = self._get_cast(instance)
            # print("choices:", choices)

            if inputs:
                value = inputs.pop(0)
                # print("using value from inputs:", value)
                if choices:
                    assert value in choices + [""]
            else:
                completer = self._get_wordcompleter(choices)
                value = self._prompt("Enter value for %s (%s) [%s]> " % (
                    prefix + name,
                    classname,
                    repr(instance.get())
                ),
                    completer=completer,
                    validator=validator)
            # print("value:", value, "type:", type(value))
            if value != "":
                value = cast(value)
                # print("casted value:", value, "type:", type(value))
                try:
                    instance.set(value)
                    input_done = True
                except Exception as e:
                    print("Could not set value:", str(e))
                    raise
            else:
                # print("unchanged")
                # no input
                input_done = True


def fill_oser_instance(instance, prefix="", inputs=None):
    """
    Fill an oser instance interactively.

    Args:
        instance: an oser instance to be filled.
        prefix="" (str): the prefix that is printed on interactive prompt.
        inputs=None (list of inputs): a list of inputs to be used instead of
            user inputs.
    """
    oser_prompt = OserPrompt()
    oser_prompt.fill_oser_instance(instance=instance,
                                   prefix=prefix,
                                   inputs=inputs)
