# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, HILSTER GmbH, Germany - http://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError, loadDAQmx
import sys
from ctypes import *
from functools import wraps

# load library
try:
    DAQlib = loadDAQmx()
except OSError as e:
    pass


BUFFER_LENGTH = 2048

# types
int8 = c_byte
uInt8 = c_ubyte
int16 = c_short
uInt16 = c_ushort
int32 = c_int
uInt32 = c_uint
float32 = c_float
float64 = c_double
int64 = c_longlong
uInt64 = c_ulonglong
bool32 = uInt32
TaskHandle = c_void_p

# callback
DAQmxEveryNSamplesEventCallbackPtr = CFUNCTYPE(int32, TaskHandle, int32, uInt32, c_void_p)
DAQmxDoneEventCallbackPtr = CFUNCTYPE(int32, TaskHandle, int32, c_void_p)
DAQmxSignalEventCallbackPtr = CFUNCTYPE(int32, TaskHandle, int32, c_void_p)


def catchDAQError(func):
    @wraps(func)
    def wrapper(*arg):
        error = func(*arg)
        if error < 0:
            errBuff = create_string_buffer(BUFFER_LENGTH)
            DAQmxGetExtendedErrorInfo(errBuff, BUFFER_LENGTH)
            raise DAQError(error, errBuff.value.decode("utf-8"), func.__name__)
        elif error > 0:
            errBuff = create_string_buffer(BUFFER_LENGTH)
            DAQmxGetErrorString(error, errBuff, BUFFER_LENGTH)
            print("WARNING  :", error, "  ", errBuff.value.decode("utf-8"))
            raise DAQError(error, errBuff.value.decode("utf-8"), func.__name__)

        return error
    return wrapper


# constants
DAQmx_Buf_Input_BufSize = 0x186C
DAQmx_Buf_Input_OnbrdBufSize = 0x230A
DAQmx_Buf_Output_BufSize = 0x186D
DAQmx_Buf_Output_OnbrdBufSize = 0x230B
DAQmx_SelfCal_Supported = 0x1860
DAQmx_SelfCal_LastTemp = 0x1864
DAQmx_ExtCal_RecommendedInterval = 0x1868
DAQmx_ExtCal_LastTemp = 0x1867
DAQmx_Cal_UserDefinedInfo = 0x1861
DAQmx_Cal_UserDefinedInfo_MaxSize = 0x191C
DAQmx_Cal_DevTemp = 0x223B
DAQmx_Cal_AccConnectionCount = 0x2FEB
DAQmx_Cal_RecommendedAccConnectionCountLimit = 0x2FEC
DAQmx_AI_Max = 0x17DD
DAQmx_AI_Min = 0x17DE
DAQmx_AI_CustomScaleName = 0x17E0
DAQmx_AI_MeasType = 0x0695
DAQmx_AI_Voltage_Units = 0x1094
DAQmx_AI_Voltage_dBRef = 0x29B0
DAQmx_AI_Voltage_ACRMS_Units = 0x17E2
DAQmx_AI_Temp_Units = 0x1033
DAQmx_AI_Thrmcpl_Type = 0x1050
DAQmx_AI_Thrmcpl_ScaleType = 0x29D0
DAQmx_AI_Thrmcpl_CJCSrc = 0x1035
DAQmx_AI_Thrmcpl_CJCVal = 0x1036
DAQmx_AI_Thrmcpl_CJCChan = 0x1034
DAQmx_AI_RTD_Type = 0x1032
DAQmx_AI_RTD_R0 = 0x1030
DAQmx_AI_RTD_A = 0x1010
DAQmx_AI_RTD_B = 0x1011
DAQmx_AI_RTD_C = 0x1013
DAQmx_AI_Thrmstr_A = 0x18C9
DAQmx_AI_Thrmstr_B = 0x18CB
DAQmx_AI_Thrmstr_C = 0x18CA
DAQmx_AI_Thrmstr_R1 = 0x1061
DAQmx_AI_ForceReadFromChan = 0x18F8
DAQmx_AI_Current_Units = 0x0701
DAQmx_AI_Current_ACRMS_Units = 0x17E3
DAQmx_AI_Strain_Units = 0x0981
DAQmx_AI_StrainGage_ForceReadFromChan = 0x2FFA
DAQmx_AI_StrainGage_GageFactor = 0x0994
DAQmx_AI_StrainGage_PoissonRatio = 0x0998
DAQmx_AI_StrainGage_Cfg = 0x0982
DAQmx_AI_RosetteStrainGage_RosetteType = 0x2FFE
DAQmx_AI_RosetteStrainGage_Orientation = 0x2FFC
DAQmx_AI_RosetteStrainGage_StrainChans = 0x2FFB
DAQmx_AI_RosetteStrainGage_RosetteMeasType = 0x2FFD
DAQmx_AI_Resistance_Units = 0x0955
DAQmx_AI_Freq_Units = 0x0806
DAQmx_AI_Freq_ThreshVoltage = 0x0815
DAQmx_AI_Freq_Hyst = 0x0814
DAQmx_AI_LVDT_Units = 0x0910
DAQmx_AI_LVDT_Sensitivity = 0x0939
DAQmx_AI_LVDT_SensitivityUnits = 0x219A
DAQmx_AI_RVDT_Units = 0x0877
DAQmx_AI_RVDT_Sensitivity = 0x0903
DAQmx_AI_RVDT_SensitivityUnits = 0x219B
DAQmx_AI_EddyCurrentProxProbe_Units = 0x2AC0
DAQmx_AI_EddyCurrentProxProbe_Sensitivity = 0x2ABE
DAQmx_AI_EddyCurrentProxProbe_SensitivityUnits = 0x2ABF
DAQmx_AI_SoundPressure_MaxSoundPressureLvl = 0x223A
DAQmx_AI_SoundPressure_Units = 0x1528
DAQmx_AI_SoundPressure_dBRef = 0x29B1
DAQmx_AI_Microphone_Sensitivity = 0x1536
DAQmx_AI_Accel_Units = 0x0673
DAQmx_AI_Accel_dBRef = 0x29B2
DAQmx_AI_Accel_Sensitivity = 0x0692
DAQmx_AI_Accel_SensitivityUnits = 0x219C
DAQmx_AI_Velocity_Units = 0x2FF4
DAQmx_AI_Velocity_IEPESensor_dBRef = 0x2FF5
DAQmx_AI_Velocity_IEPESensor_Sensitivity = 0x2FF6
DAQmx_AI_Velocity_IEPESensor_SensitivityUnits = 0x2FF7
DAQmx_AI_Force_Units = 0x2F75
DAQmx_AI_Force_IEPESensor_Sensitivity = 0x2F81
DAQmx_AI_Force_IEPESensor_SensitivityUnits = 0x2F82
DAQmx_AI_Pressure_Units = 0x2F76
DAQmx_AI_Torque_Units = 0x2F77
DAQmx_AI_Bridge_Units = 0x2F92
DAQmx_AI_Bridge_ElectricalUnits = 0x2F87
DAQmx_AI_Bridge_PhysicalUnits = 0x2F88
DAQmx_AI_Bridge_ScaleType = 0x2F89
DAQmx_AI_Bridge_TwoPointLin_First_ElectricalVal = 0x2F8A
DAQmx_AI_Bridge_TwoPointLin_First_PhysicalVal = 0x2F8B
DAQmx_AI_Bridge_TwoPointLin_Second_ElectricalVal = 0x2F8C
DAQmx_AI_Bridge_TwoPointLin_Second_PhysicalVal = 0x2F8D
DAQmx_AI_Bridge_Table_ElectricalVals = 0x2F8E
DAQmx_AI_Bridge_Table_PhysicalVals = 0x2F8F
DAQmx_AI_Bridge_Poly_ForwardCoeff = 0x2F90
DAQmx_AI_Bridge_Poly_ReverseCoeff = 0x2F91
DAQmx_AI_Is_TEDS = 0x2983
DAQmx_AI_TEDS_Units = 0x21E0
DAQmx_AI_Coupling = 0x0064
DAQmx_AI_Impedance = 0x0062
DAQmx_AI_TermCfg = 0x1097
DAQmx_AI_InputSrc = 0x2198
DAQmx_AI_ResistanceCfg = 0x1881
DAQmx_AI_LeadWireResistance = 0x17EE
DAQmx_AI_Bridge_Cfg = 0x0087
DAQmx_AI_Bridge_NomResistance = 0x17EC
DAQmx_AI_Bridge_InitialVoltage = 0x17ED
DAQmx_AI_Bridge_InitialRatio = 0x2F86
DAQmx_AI_Bridge_ShuntCal_Enable = 0x0094
DAQmx_AI_Bridge_ShuntCal_Select = 0x21D5
DAQmx_AI_Bridge_ShuntCal_GainAdjust = 0x193F
DAQmx_AI_Bridge_ShuntCal_ShuntCalAResistance = 0x2F78
DAQmx_AI_Bridge_ShuntCal_ShuntCalAActualResistance = 0x2F79
DAQmx_AI_Bridge_Balance_CoarsePot = 0x17F1
DAQmx_AI_Bridge_Balance_FinePot = 0x18F4
DAQmx_AI_CurrentShunt_Loc = 0x17F2
DAQmx_AI_CurrentShunt_Resistance = 0x17F3
DAQmx_AI_Excit_Src = 0x17F4
DAQmx_AI_Excit_Val = 0x17F5
DAQmx_AI_Excit_UseForScaling = 0x17FC
DAQmx_AI_Excit_UseMultiplexed = 0x2180
DAQmx_AI_Excit_ActualVal = 0x1883
DAQmx_AI_Excit_DCorAC = 0x17FB
DAQmx_AI_Excit_VoltageOrCurrent = 0x17F6
DAQmx_AI_ACExcit_Freq = 0x0101
DAQmx_AI_ACExcit_SyncEnable = 0x0102
DAQmx_AI_ACExcit_WireMode = 0x18CD
DAQmx_AI_OpenThrmcplDetectEnable = 0x2F72
DAQmx_AI_Thrmcpl_LeadOffsetVoltage = 0x2FB8
DAQmx_AI_Atten = 0x1801
DAQmx_AI_ProbeAtten = 0x2A88
DAQmx_AI_Lowpass_Enable = 0x1802
DAQmx_AI_Lowpass_CutoffFreq = 0x1803
DAQmx_AI_Lowpass_SwitchCap_ClkSrc = 0x1884
DAQmx_AI_Lowpass_SwitchCap_ExtClkFreq = 0x1885
DAQmx_AI_Lowpass_SwitchCap_ExtClkDiv = 0x1886
DAQmx_AI_Lowpass_SwitchCap_OutClkDiv = 0x1887
DAQmx_AI_FilterDelay = 0x2FED
DAQmx_AI_FilterDelayUnits = 0x3071
DAQmx_AI_RemoveFilterDelay = 0x2FBD
DAQmx_AI_FilterDelayAdjustment = 0x3074
DAQmx_AI_AveragingWinSize = 0x2FEE
DAQmx_AI_ResolutionUnits = 0x1764
DAQmx_AI_Resolution = 0x1765
DAQmx_AI_RawSampSize = 0x22DA
DAQmx_AI_RawSampJustification = 0x0050
DAQmx_AI_ADCTimingMode = 0x29F9
DAQmx_AI_ADCCustomTimingMode = 0x2F6B
DAQmx_AI_Dither_Enable = 0x0068
DAQmx_AI_ChanCal_HasValidCalInfo = 0x2297
DAQmx_AI_ChanCal_EnableCal = 0x2298
DAQmx_AI_ChanCal_ApplyCalIfExp = 0x2299
DAQmx_AI_ChanCal_ScaleType = 0x229C
DAQmx_AI_ChanCal_Table_PreScaledVals = 0x229D
DAQmx_AI_ChanCal_Table_ScaledVals = 0x229E
DAQmx_AI_ChanCal_Poly_ForwardCoeff = 0x229F
DAQmx_AI_ChanCal_Poly_ReverseCoeff = 0x22A0
DAQmx_AI_ChanCal_OperatorName = 0x22A3
DAQmx_AI_ChanCal_Desc = 0x22A4
DAQmx_AI_ChanCal_Verif_RefVals = 0x22A1
DAQmx_AI_ChanCal_Verif_AcqVals = 0x22A2
DAQmx_AI_Rng_High = 0x1815
DAQmx_AI_Rng_Low = 0x1816
DAQmx_AI_DCOffset = 0x2A89
DAQmx_AI_Gain = 0x1818
DAQmx_AI_SampAndHold_Enable = 0x181A
DAQmx_AI_AutoZeroMode = 0x1760
DAQmx_AI_DataXferMech = 0x1821
DAQmx_AI_DataXferReqCond = 0x188B
DAQmx_AI_DataXferCustomThreshold = 0x230C
DAQmx_AI_UsbXferReqSize = 0x2A8E
DAQmx_AI_UsbXferReqCount = 0x3000
DAQmx_AI_MemMapEnable = 0x188C
DAQmx_AI_RawDataCompressionType = 0x22D8
DAQmx_AI_LossyLSBRemoval_CompressedSampSize = 0x22D9
DAQmx_AI_DevScalingCoeff = 0x1930
DAQmx_AI_EnhancedAliasRejectionEnable = 0x2294
DAQmx_AO_Max = 0x1186
DAQmx_AO_Min = 0x1187
DAQmx_AO_CustomScaleName = 0x1188
DAQmx_AO_OutputType = 0x1108
DAQmx_AO_Voltage_Units = 0x1184
DAQmx_AO_Voltage_CurrentLimit = 0x2A1D
DAQmx_AO_Current_Units = 0x1109
DAQmx_AO_FuncGen_Type = 0x2A18
DAQmx_AO_FuncGen_Freq = 0x2A19
DAQmx_AO_FuncGen_Amplitude = 0x2A1A
DAQmx_AO_FuncGen_Offset = 0x2A1B
DAQmx_AO_FuncGen_Square_DutyCycle = 0x2A1C
DAQmx_AO_FuncGen_ModulationType = 0x2A22
DAQmx_AO_FuncGen_FMDeviation = 0x2A23
DAQmx_AO_OutputImpedance = 0x1490
DAQmx_AO_LoadImpedance = 0x0121
DAQmx_AO_IdleOutputBehavior = 0x2240
DAQmx_AO_TermCfg = 0x188E
DAQmx_AO_ResolutionUnits = 0x182B
DAQmx_AO_Resolution = 0x182C
DAQmx_AO_DAC_Rng_High = 0x182E
DAQmx_AO_DAC_Rng_Low = 0x182D
DAQmx_AO_DAC_Ref_ConnToGnd = 0x0130
DAQmx_AO_DAC_Ref_AllowConnToGnd = 0x1830
DAQmx_AO_DAC_Ref_Src = 0x0132
DAQmx_AO_DAC_Ref_ExtSrc = 0x2252
DAQmx_AO_DAC_Ref_Val = 0x1832
DAQmx_AO_DAC_Offset_Src = 0x2253
DAQmx_AO_DAC_Offset_ExtSrc = 0x2254
DAQmx_AO_DAC_Offset_Val = 0x2255
DAQmx_AO_ReglitchEnable = 0x0133
DAQmx_AO_FilterDelay = 0x3075
DAQmx_AO_FilterDelayUnits = 0x3076
DAQmx_AO_FilterDelayAdjustment = 0x3072
DAQmx_AO_Gain = 0x0118
DAQmx_AO_UseOnlyOnBrdMem = 0x183A
DAQmx_AO_DataXferMech = 0x0134
DAQmx_AO_DataXferReqCond = 0x183C
DAQmx_AO_UsbXferReqSize = 0x2A8F
DAQmx_AO_UsbXferReqCount = 0x3001
DAQmx_AO_MemMapEnable = 0x188F
DAQmx_AO_DevScalingCoeff = 0x1931
DAQmx_AO_EnhancedImageRejectionEnable = 0x2241
DAQmx_DI_InvertLines = 0x0793
DAQmx_DI_NumLines = 0x2178
DAQmx_DI_DigFltr_Enable = 0x21D6
DAQmx_DI_DigFltr_MinPulseWidth = 0x21D7
DAQmx_DI_DigFltr_EnableBusMode = 0x2EFE
DAQmx_DI_DigFltr_TimebaseSrc = 0x2ED4
DAQmx_DI_DigFltr_TimebaseRate = 0x2ED5
DAQmx_DI_DigSync_Enable = 0x2ED6
DAQmx_DI_Tristate = 0x1890
DAQmx_DI_LogicFamily = 0x296D
DAQmx_DI_DataXferMech = 0x2263
DAQmx_DI_DataXferReqCond = 0x2264
DAQmx_DI_UsbXferReqSize = 0x2A90
DAQmx_DI_UsbXferReqCount = 0x3002
DAQmx_DI_MemMapEnable = 0x296A
DAQmx_DI_AcquireOn = 0x2966
DAQmx_DO_OutputDriveType = 0x1137
DAQmx_DO_InvertLines = 0x1133
DAQmx_DO_NumLines = 0x2179
DAQmx_DO_Tristate = 0x18F3
DAQmx_DO_LineStates_StartState = 0x2972
DAQmx_DO_LineStates_PausedState = 0x2967
DAQmx_DO_LineStates_DoneState = 0x2968
DAQmx_DO_LogicFamily = 0x296E
DAQmx_DO_Overcurrent_Limit = 0x2A85
DAQmx_DO_Overcurrent_AutoReenable = 0x2A86
DAQmx_DO_Overcurrent_ReenablePeriod = 0x2A87
DAQmx_DO_UseOnlyOnBrdMem = 0x2265
DAQmx_DO_DataXferMech = 0x2266
DAQmx_DO_DataXferReqCond = 0x2267
DAQmx_DO_UsbXferReqSize = 0x2A91
DAQmx_DO_UsbXferReqCount = 0x3003
DAQmx_DO_MemMapEnable = 0x296B
DAQmx_DO_GenerateOn = 0x2969
DAQmx_CI_Max = 0x189C
DAQmx_CI_Min = 0x189D
DAQmx_CI_CustomScaleName = 0x189E
DAQmx_CI_MeasType = 0x18A0
DAQmx_CI_Freq_Units = 0x18A1
DAQmx_CI_Freq_Term = 0x18A2
DAQmx_CI_Freq_StartingEdge = 0x0799
DAQmx_CI_Freq_MeasMeth = 0x0144
DAQmx_CI_Freq_EnableAveraging = 0x2ED0
DAQmx_CI_Freq_MeasTime = 0x0145
DAQmx_CI_Freq_Div = 0x0147
DAQmx_CI_Freq_DigFltr_Enable = 0x21E7
DAQmx_CI_Freq_DigFltr_MinPulseWidth = 0x21E8
DAQmx_CI_Freq_DigFltr_TimebaseSrc = 0x21E9
DAQmx_CI_Freq_DigFltr_TimebaseRate = 0x21EA
DAQmx_CI_Freq_DigSync_Enable = 0x21EB
DAQmx_CI_Period_Units = 0x18A3
DAQmx_CI_Period_Term = 0x18A4
DAQmx_CI_Period_StartingEdge = 0x0852
DAQmx_CI_Period_MeasMeth = 0x192C
DAQmx_CI_Period_EnableAveraging = 0x2ED1
DAQmx_CI_Period_MeasTime = 0x192D
DAQmx_CI_Period_Div = 0x192E
DAQmx_CI_Period_DigFltr_Enable = 0x21EC
DAQmx_CI_Period_DigFltr_MinPulseWidth = 0x21ED
DAQmx_CI_Period_DigFltr_TimebaseSrc = 0x21EE
DAQmx_CI_Period_DigFltr_TimebaseRate = 0x21EF
DAQmx_CI_Period_DigSync_Enable = 0x21F0
DAQmx_CI_CountEdges_Term = 0x18C7
DAQmx_CI_CountEdges_Dir = 0x0696
DAQmx_CI_CountEdges_DirTerm = 0x21E1
DAQmx_CI_CountEdges_CountDir_DigFltr_Enable = 0x21F1
DAQmx_CI_CountEdges_CountDir_DigFltr_MinPulseWidth = 0x21F2
DAQmx_CI_CountEdges_CountDir_DigFltr_TimebaseSrc = 0x21F3
DAQmx_CI_CountEdges_CountDir_DigFltr_TimebaseRate = 0x21F4
DAQmx_CI_CountEdges_CountDir_DigSync_Enable = 0x21F5
DAQmx_CI_CountEdges_InitialCnt = 0x0698
DAQmx_CI_CountEdges_ActiveEdge = 0x0697
DAQmx_CI_CountEdges_CountReset_Enable = 0x2FAF
DAQmx_CI_CountEdges_CountReset_ResetCount = 0x2FB0
DAQmx_CI_CountEdges_CountReset_Term = 0x2FB1
DAQmx_CI_CountEdges_CountReset_ActiveEdge = 0x2FB2
DAQmx_CI_CountEdges_CountReset_DigFltr_Enable = 0x2FB3
DAQmx_CI_CountEdges_CountReset_DigFltr_MinPulseWidth = 0x2FB4
DAQmx_CI_CountEdges_CountReset_DigFltr_TimebaseSrc = 0x2FB5
DAQmx_CI_CountEdges_CountReset_DigFltr_TimebaseRate = 0x2FB6
DAQmx_CI_CountEdges_CountReset_DigSync_Enable = 0x2FB7
DAQmx_CI_CountEdges_DigFltr_Enable = 0x21F6
DAQmx_CI_CountEdges_DigFltr_MinPulseWidth = 0x21F7
DAQmx_CI_CountEdges_DigFltr_TimebaseSrc = 0x21F8
DAQmx_CI_CountEdges_DigFltr_TimebaseRate = 0x21F9
DAQmx_CI_CountEdges_DigSync_Enable = 0x21FA
DAQmx_CI_AngEncoder_Units = 0x18A6
DAQmx_CI_AngEncoder_PulsesPerRev = 0x0875
DAQmx_CI_AngEncoder_InitialAngle = 0x0881
DAQmx_CI_LinEncoder_Units = 0x18A9
DAQmx_CI_LinEncoder_DistPerPulse = 0x0911
DAQmx_CI_LinEncoder_InitialPos = 0x0915
DAQmx_CI_Encoder_DecodingType = 0x21E6
DAQmx_CI_Encoder_AInputTerm = 0x219D
DAQmx_CI_Encoder_AInput_DigFltr_Enable = 0x21FB
DAQmx_CI_Encoder_AInput_DigFltr_MinPulseWidth = 0x21FC
DAQmx_CI_Encoder_AInput_DigFltr_TimebaseSrc = 0x21FD
DAQmx_CI_Encoder_AInput_DigFltr_TimebaseRate = 0x21FE
DAQmx_CI_Encoder_AInput_DigSync_Enable = 0x21FF
DAQmx_CI_Encoder_BInputTerm = 0x219E
DAQmx_CI_Encoder_BInput_DigFltr_Enable = 0x2200
DAQmx_CI_Encoder_BInput_DigFltr_MinPulseWidth = 0x2201
DAQmx_CI_Encoder_BInput_DigFltr_TimebaseSrc = 0x2202
DAQmx_CI_Encoder_BInput_DigFltr_TimebaseRate = 0x2203
DAQmx_CI_Encoder_BInput_DigSync_Enable = 0x2204
DAQmx_CI_Encoder_ZInputTerm = 0x219F
DAQmx_CI_Encoder_ZInput_DigFltr_Enable = 0x2205
DAQmx_CI_Encoder_ZInput_DigFltr_MinPulseWidth = 0x2206
DAQmx_CI_Encoder_ZInput_DigFltr_TimebaseSrc = 0x2207
DAQmx_CI_Encoder_ZInput_DigFltr_TimebaseRate = 0x2208
DAQmx_CI_Encoder_ZInput_DigSync_Enable = 0x2209
DAQmx_CI_Encoder_ZIndexEnable = 0x0890
DAQmx_CI_Encoder_ZIndexVal = 0x0888
DAQmx_CI_Encoder_ZIndexPhase = 0x0889
DAQmx_CI_PulseWidth_Units = 0x0823
DAQmx_CI_PulseWidth_Term = 0x18AA
DAQmx_CI_PulseWidth_StartingEdge = 0x0825
DAQmx_CI_PulseWidth_DigFltr_Enable = 0x220A
DAQmx_CI_PulseWidth_DigFltr_MinPulseWidth = 0x220B
DAQmx_CI_PulseWidth_DigFltr_TimebaseSrc = 0x220C
DAQmx_CI_PulseWidth_DigFltr_TimebaseRate = 0x220D
DAQmx_CI_PulseWidth_DigSync_Enable = 0x220E
DAQmx_CI_Timestamp_Units = 0x22B3
DAQmx_CI_TwoEdgeSep_Units = 0x18AC
DAQmx_CI_TwoEdgeSep_FirstTerm = 0x18AD
DAQmx_CI_TwoEdgeSep_FirstEdge = 0x0833
DAQmx_CI_TwoEdgeSep_First_DigFltr_Enable = 0x220F
DAQmx_CI_TwoEdgeSep_First_DigFltr_MinPulseWidth = 0x2210
DAQmx_CI_TwoEdgeSep_First_DigFltr_TimebaseSrc = 0x2211
DAQmx_CI_TwoEdgeSep_First_DigFltr_TimebaseRate = 0x2212
DAQmx_CI_TwoEdgeSep_First_DigSync_Enable = 0x2213
DAQmx_CI_TwoEdgeSep_SecondTerm = 0x18AE
DAQmx_CI_TwoEdgeSep_SecondEdge = 0x0834
DAQmx_CI_TwoEdgeSep_Second_DigFltr_Enable = 0x2214
DAQmx_CI_TwoEdgeSep_Second_DigFltr_MinPulseWidth = 0x2215
DAQmx_CI_TwoEdgeSep_Second_DigFltr_TimebaseSrc = 0x2216
DAQmx_CI_TwoEdgeSep_Second_DigFltr_TimebaseRate = 0x2217
DAQmx_CI_TwoEdgeSep_Second_DigSync_Enable = 0x2218
DAQmx_CI_SemiPeriod_Units = 0x18AF
DAQmx_CI_SemiPeriod_Term = 0x18B0
DAQmx_CI_SemiPeriod_StartingEdge = 0x22FE
DAQmx_CI_SemiPeriod_DigFltr_Enable = 0x2219
DAQmx_CI_SemiPeriod_DigFltr_MinPulseWidth = 0x221A
DAQmx_CI_SemiPeriod_DigFltr_TimebaseSrc = 0x221B
DAQmx_CI_SemiPeriod_DigFltr_TimebaseRate = 0x221C
DAQmx_CI_SemiPeriod_DigSync_Enable = 0x221D
DAQmx_CI_Pulse_Freq_Units = 0x2F0B
DAQmx_CI_Pulse_Freq_Term = 0x2F04
DAQmx_CI_Pulse_Freq_Start_Edge = 0x2F05
DAQmx_CI_Pulse_Freq_DigFltr_Enable = 0x2F06
DAQmx_CI_Pulse_Freq_DigFltr_MinPulseWidth = 0x2F07
DAQmx_CI_Pulse_Freq_DigFltr_TimebaseSrc = 0x2F08
DAQmx_CI_Pulse_Freq_DigFltr_TimebaseRate = 0x2F09
DAQmx_CI_Pulse_Freq_DigSync_Enable = 0x2F0A
DAQmx_CI_Pulse_Time_Units = 0x2F13
DAQmx_CI_Pulse_Time_Term = 0x2F0C
DAQmx_CI_Pulse_Time_StartEdge = 0x2F0D
DAQmx_CI_Pulse_Time_DigFltr_Enable = 0x2F0E
DAQmx_CI_Pulse_Time_DigFltr_MinPulseWidth = 0x2F0F
DAQmx_CI_Pulse_Time_DigFltr_TimebaseSrc = 0x2F10
DAQmx_CI_Pulse_Time_DigFltr_TimebaseRate = 0x2F11
DAQmx_CI_Pulse_Time_DigSync_Enable = 0x2F12
DAQmx_CI_Pulse_Ticks_Term = 0x2F14
DAQmx_CI_Pulse_Ticks_StartEdge = 0x2F15
DAQmx_CI_Pulse_Ticks_DigFltr_Enable = 0x2F16
DAQmx_CI_Pulse_Ticks_DigFltr_MinPulseWidth = 0x2F17
DAQmx_CI_Pulse_Ticks_DigFltr_TimebaseSrc = 0x2F18
DAQmx_CI_Pulse_Ticks_DigFltr_TimebaseRate = 0x2F19
DAQmx_CI_Pulse_Ticks_DigSync_Enable = 0x2F1A
DAQmx_CI_Timestamp_InitialSeconds = 0x22B4
DAQmx_CI_GPS_SyncMethod = 0x1092
DAQmx_CI_GPS_SyncSrc = 0x1093
DAQmx_CI_CtrTimebaseSrc = 0x0143
DAQmx_CI_CtrTimebaseRate = 0x18B2
DAQmx_CI_CtrTimebaseActiveEdge = 0x0142
DAQmx_CI_CtrTimebase_DigFltr_Enable = 0x2271
DAQmx_CI_CtrTimebase_DigFltr_MinPulseWidth = 0x2272
DAQmx_CI_CtrTimebase_DigFltr_TimebaseSrc = 0x2273
DAQmx_CI_CtrTimebase_DigFltr_TimebaseRate = 0x2274
DAQmx_CI_CtrTimebase_DigSync_Enable = 0x2275
DAQmx_CI_Count = 0x0148
DAQmx_CI_OutputState = 0x0149
DAQmx_CI_TCReached = 0x0150
DAQmx_CI_CtrTimebaseMasterTimebaseDiv = 0x18B3
DAQmx_CI_DataXferMech = 0x0200
DAQmx_CI_DataXferReqCond = 0x2EFB
DAQmx_CI_UsbXferReqSize = 0x2A92
DAQmx_CI_UsbXferReqCount = 0x3004
DAQmx_CI_MemMapEnable = 0x2ED2
DAQmx_CI_NumPossiblyInvalidSamps = 0x193C
DAQmx_CI_DupCountPrevent = 0x21AC
DAQmx_CI_Prescaler = 0x2239
DAQmx_CO_OutputType = 0x18B5
DAQmx_CO_Pulse_IdleState = 0x1170
DAQmx_CO_Pulse_Term = 0x18E1
DAQmx_CO_Pulse_Time_Units = 0x18D6
DAQmx_CO_Pulse_HighTime = 0x18BA
DAQmx_CO_Pulse_LowTime = 0x18BB
DAQmx_CO_Pulse_Time_InitialDelay = 0x18BC
DAQmx_CO_Pulse_DutyCyc = 0x1176
DAQmx_CO_Pulse_Freq_Units = 0x18D5
DAQmx_CO_Pulse_Freq = 0x1178
DAQmx_CO_Pulse_Freq_InitialDelay = 0x0299
DAQmx_CO_Pulse_HighTicks = 0x1169
DAQmx_CO_Pulse_LowTicks = 0x1171
DAQmx_CO_Pulse_Ticks_InitialDelay = 0x0298
DAQmx_CO_CtrTimebaseSrc = 0x0339
DAQmx_CO_CtrTimebaseRate = 0x18C2
DAQmx_CO_CtrTimebaseActiveEdge = 0x0341
DAQmx_CO_CtrTimebase_DigFltr_Enable = 0x2276
DAQmx_CO_CtrTimebase_DigFltr_MinPulseWidth = 0x2277
DAQmx_CO_CtrTimebase_DigFltr_TimebaseSrc = 0x2278
DAQmx_CO_CtrTimebase_DigFltr_TimebaseRate = 0x2279
DAQmx_CO_CtrTimebase_DigSync_Enable = 0x227A
DAQmx_CO_Count = 0x0293
DAQmx_CO_OutputState = 0x0294
DAQmx_CO_AutoIncrCnt = 0x0295
DAQmx_CO_CtrTimebaseMasterTimebaseDiv = 0x18C3
DAQmx_CO_PulseDone = 0x190E
DAQmx_CO_EnableInitialDelayOnRetrigger = 0x2EC9
DAQmx_CO_ConstrainedGenMode = 0x29F2
DAQmx_CO_UseOnlyOnBrdMem = 0x2ECB
DAQmx_CO_DataXferMech = 0x2ECC
DAQmx_CO_DataXferReqCond = 0x2ECD
DAQmx_CO_UsbXferReqSize = 0x2A93
DAQmx_CO_UsbXferReqCount = 0x3005
DAQmx_CO_MemMapEnable = 0x2ED3
DAQmx_CO_Prescaler = 0x226D
DAQmx_CO_RdyForNewVal = 0x22FF
DAQmx_ChanType = 0x187F
DAQmx_PhysicalChanName = 0x18F5
DAQmx_ChanDescr = 0x1926
DAQmx_ChanIsGlobal = 0x2304
DAQmx_Exported_AIConvClk_OutputTerm = 0x1687
DAQmx_Exported_AIConvClk_Pulse_Polarity = 0x1688
DAQmx_Exported_10MHzRefClk_OutputTerm = 0x226E
DAQmx_Exported_20MHzTimebase_OutputTerm = 0x1657
DAQmx_Exported_SampClk_OutputBehavior = 0x186B
DAQmx_Exported_SampClk_OutputTerm = 0x1663
DAQmx_Exported_SampClk_DelayOffset = 0x21C4
DAQmx_Exported_SampClk_Pulse_Polarity = 0x1664
DAQmx_Exported_SampClkTimebase_OutputTerm = 0x18F9
DAQmx_Exported_DividedSampClkTimebase_OutputTerm = 0x21A1
DAQmx_Exported_AdvTrig_OutputTerm = 0x1645
DAQmx_Exported_AdvTrig_Pulse_Polarity = 0x1646
DAQmx_Exported_AdvTrig_Pulse_WidthUnits = 0x1647
DAQmx_Exported_AdvTrig_Pulse_Width = 0x1648
DAQmx_Exported_PauseTrig_OutputTerm = 0x1615
DAQmx_Exported_PauseTrig_Lvl_ActiveLvl = 0x1616
DAQmx_Exported_RefTrig_OutputTerm = 0x0590
DAQmx_Exported_RefTrig_Pulse_Polarity = 0x0591
DAQmx_Exported_StartTrig_OutputTerm = 0x0584
DAQmx_Exported_StartTrig_Pulse_Polarity = 0x0585
DAQmx_Exported_AdvCmpltEvent_OutputTerm = 0x1651
DAQmx_Exported_AdvCmpltEvent_Delay = 0x1757
DAQmx_Exported_AdvCmpltEvent_Pulse_Polarity = 0x1652
DAQmx_Exported_AdvCmpltEvent_Pulse_Width = 0x1654
DAQmx_Exported_AIHoldCmpltEvent_OutputTerm = 0x18ED
DAQmx_Exported_AIHoldCmpltEvent_PulsePolarity = 0x18EE
DAQmx_Exported_ChangeDetectEvent_OutputTerm = 0x2197
DAQmx_Exported_ChangeDetectEvent_Pulse_Polarity = 0x2303
DAQmx_Exported_CtrOutEvent_OutputTerm = 0x1717
DAQmx_Exported_CtrOutEvent_OutputBehavior = 0x174F
DAQmx_Exported_CtrOutEvent_Pulse_Polarity = 0x1718
DAQmx_Exported_CtrOutEvent_Toggle_IdleState = 0x186A
DAQmx_Exported_HshkEvent_OutputTerm = 0x22BA
DAQmx_Exported_HshkEvent_OutputBehavior = 0x22BB
DAQmx_Exported_HshkEvent_Delay = 0x22BC
DAQmx_Exported_HshkEvent_Interlocked_AssertedLvl = 0x22BD
DAQmx_Exported_HshkEvent_Interlocked_AssertOnStart = 0x22BE
DAQmx_Exported_HshkEvent_Interlocked_DeassertDelay = 0x22BF
DAQmx_Exported_HshkEvent_Pulse_Polarity = 0x22C0
DAQmx_Exported_HshkEvent_Pulse_Width = 0x22C1
DAQmx_Exported_RdyForXferEvent_OutputTerm = 0x22B5
DAQmx_Exported_RdyForXferEvent_Lvl_ActiveLvl = 0x22B6
DAQmx_Exported_RdyForXferEvent_DeassertCond = 0x2963
DAQmx_Exported_RdyForXferEvent_DeassertCondCustomThreshold = 0x2964
DAQmx_Exported_DataActiveEvent_OutputTerm = 0x1633
DAQmx_Exported_DataActiveEvent_Lvl_ActiveLvl = 0x1634
DAQmx_Exported_RdyForStartEvent_OutputTerm = 0x1609
DAQmx_Exported_RdyForStartEvent_Lvl_ActiveLvl = 0x1751
DAQmx_Exported_SyncPulseEvent_OutputTerm = 0x223C
DAQmx_Exported_WatchdogExpiredEvent_OutputTerm = 0x21AA
DAQmx_Dev_IsSimulated = 0x22CA
DAQmx_Dev_ProductCategory = 0x29A9
DAQmx_Dev_ProductType = 0x0631
DAQmx_Dev_ProductNum = 0x231D
DAQmx_Dev_SerialNum = 0x0632
DAQmx_Dev_Accessory_ProductTypes = 0x2F6D
DAQmx_Dev_Accessory_ProductNums = 0x2F6E
DAQmx_Dev_Accessory_SerialNums = 0x2F6F
DAQmx_Carrier_SerialNum = 0x2A8A
DAQmx_Dev_Chassis_ModuleDevNames = 0x29B6
DAQmx_Dev_AnlgTrigSupported = 0x2984
DAQmx_Dev_DigTrigSupported = 0x2985
DAQmx_Dev_AI_PhysicalChans = 0x231E
DAQmx_Dev_AI_SupportedMeasTypes = 0x2FD2
DAQmx_Dev_AI_MaxSingleChanRate = 0x298C
DAQmx_Dev_AI_MaxMultiChanRate = 0x298D
DAQmx_Dev_AI_MinRate = 0x298E
DAQmx_Dev_AI_SimultaneousSamplingSupported = 0x298F
DAQmx_Dev_AI_SampModes = 0x2FDC
DAQmx_Dev_AI_TrigUsage = 0x2986
DAQmx_Dev_AI_VoltageRngs = 0x2990
DAQmx_Dev_AI_VoltageIntExcitDiscreteVals = 0x29C9
DAQmx_Dev_AI_VoltageIntExcitRangeVals = 0x29CA
DAQmx_Dev_AI_CurrentRngs = 0x2991
DAQmx_Dev_AI_CurrentIntExcitDiscreteVals = 0x29CB
DAQmx_Dev_AI_BridgeRngs = 0x2FD0
DAQmx_Dev_AI_ResistanceRngs = 0x2A15
DAQmx_Dev_AI_FreqRngs = 0x2992
DAQmx_Dev_AI_Gains = 0x2993
DAQmx_Dev_AI_Couplings = 0x2994
DAQmx_Dev_AI_LowpassCutoffFreqDiscreteVals = 0x2995
DAQmx_Dev_AI_LowpassCutoffFreqRangeVals = 0x29CF
DAQmx_Dev_AO_PhysicalChans = 0x231F
DAQmx_Dev_AO_SupportedOutputTypes = 0x2FD3
DAQmx_Dev_AO_SampClkSupported = 0x2996
DAQmx_Dev_AO_SampModes = 0x2FDD
DAQmx_Dev_AO_MaxRate = 0x2997
DAQmx_Dev_AO_MinRate = 0x2998
DAQmx_Dev_AO_TrigUsage = 0x2987
DAQmx_Dev_AO_VoltageRngs = 0x299B
DAQmx_Dev_AO_CurrentRngs = 0x299C
DAQmx_Dev_AO_Gains = 0x299D
DAQmx_Dev_DI_Lines = 0x2320
DAQmx_Dev_DI_Ports = 0x2321
DAQmx_Dev_DI_MaxRate = 0x2999
DAQmx_Dev_DI_TrigUsage = 0x2988
DAQmx_Dev_DO_Lines = 0x2322
DAQmx_Dev_DO_Ports = 0x2323
DAQmx_Dev_DO_MaxRate = 0x299A
DAQmx_Dev_DO_TrigUsage = 0x2989
DAQmx_Dev_CI_PhysicalChans = 0x2324
DAQmx_Dev_CI_SupportedMeasTypes = 0x2FD4
DAQmx_Dev_CI_TrigUsage = 0x298A
DAQmx_Dev_CI_SampClkSupported = 0x299E
DAQmx_Dev_CI_SampModes = 0x2FDE
DAQmx_Dev_CI_MaxSize = 0x299F
DAQmx_Dev_CI_MaxTimebase = 0x29A0
DAQmx_Dev_CO_PhysicalChans = 0x2325
DAQmx_Dev_CO_SupportedOutputTypes = 0x2FD5
DAQmx_Dev_CO_SampClkSupported = 0x2F5B
DAQmx_Dev_CO_SampModes = 0x2FDF
DAQmx_Dev_CO_TrigUsage = 0x298B
DAQmx_Dev_CO_MaxSize = 0x29A1
DAQmx_Dev_CO_MaxTimebase = 0x29A2
DAQmx_Dev_TEDS_HWTEDSSupported = 0x2FD6
DAQmx_Dev_NumDMAChans = 0x233C
DAQmx_Dev_BusType = 0x2326
DAQmx_Dev_PCI_BusNum = 0x2327
DAQmx_Dev_PCI_DevNum = 0x2328
DAQmx_Dev_PXI_ChassisNum = 0x2329
DAQmx_Dev_PXI_SlotNum = 0x232A
DAQmx_Dev_CompactDAQ_ChassisDevName = 0x29B7
DAQmx_Dev_CompactDAQ_SlotNum = 0x29B8
DAQmx_Dev_TCPIP_Hostname = 0x2A8B
DAQmx_Dev_TCPIP_EthernetIP = 0x2A8C
DAQmx_Dev_TCPIP_WirelessIP = 0x2A8D
DAQmx_Dev_Terminals = 0x2A40
DAQmx_Read_RelativeTo = 0x190A
DAQmx_Read_Offset = 0x190B
DAQmx_Read_ChannelsToRead = 0x1823
DAQmx_Read_ReadAllAvailSamp = 0x1215
DAQmx_Read_AutoStart = 0x1826
DAQmx_Read_OverWrite = 0x1211
DAQmx_Read_CurrReadPos = 0x1221
DAQmx_Read_AvailSampPerChan = 0x1223
DAQmx_Logging_FilePath = 0x2EC4
DAQmx_Logging_Mode = 0x2EC5
DAQmx_Logging_TDMS_GroupName = 0x2EC6
DAQmx_Logging_TDMS_Operation = 0x2EC7
DAQmx_Logging_Pause = 0x2FE3
DAQmx_Logging_SampsPerFile = 0x2FE4
DAQmx_Logging_FileWriteSize = 0x2FC3
DAQmx_Logging_FilePreallocationSize = 0x2FC6
DAQmx_Read_TotalSampPerChanAcquired = 0x192A
DAQmx_Read_CommonModeRangeErrorChansExist = 0x2A98
DAQmx_Read_CommonModeRangeErrorChans = 0x2A99
DAQmx_Read_OvercurrentChansExist = 0x29E6
DAQmx_Read_OvercurrentChans = 0x29E7
DAQmx_Read_OvertemperatureChansExist = 0x3081
DAQmx_Read_OvertemperatureChans = 0x3082
DAQmx_Read_OpenCurrentLoopChansExist = 0x2A09
DAQmx_Read_OpenCurrentLoopChans = 0x2A0A
DAQmx_Read_OpenThrmcplChansExist = 0x2A96
DAQmx_Read_OpenThrmcplChans = 0x2A97
DAQmx_Read_OverloadedChansExist = 0x2174
DAQmx_Read_OverloadedChans = 0x2175
DAQmx_Read_AccessoryInsertionOrRemovalDetected = 0x2F70
DAQmx_Read_DevsWithInsertedOrRemovedAccessories = 0x2F71
DAQmx_Read_ChangeDetect_HasOverflowed = 0x2194
DAQmx_Read_RawDataWidth = 0x217A
DAQmx_Read_NumChans = 0x217B
DAQmx_Read_DigitalLines_BytesPerChan = 0x217C
DAQmx_Read_WaitMode = 0x2232
DAQmx_Read_SleepTime = 0x22B0
DAQmx_RealTime_ConvLateErrorsToWarnings = 0x22EE
DAQmx_RealTime_NumOfWarmupIters = 0x22ED
DAQmx_RealTime_WaitForNextSampClkWaitMode = 0x22EF
DAQmx_RealTime_ReportMissedSamp = 0x2319
DAQmx_RealTime_WriteRecoveryMode = 0x231A
DAQmx_SwitchChan_Usage = 0x18E4
DAQmx_SwitchChan_AnlgBusSharingEnable = 0x2F9E
DAQmx_SwitchChan_MaxACCarryCurrent = 0x0648
DAQmx_SwitchChan_MaxACSwitchCurrent = 0x0646
DAQmx_SwitchChan_MaxACCarryPwr = 0x0642
DAQmx_SwitchChan_MaxACSwitchPwr = 0x0644
DAQmx_SwitchChan_MaxDCCarryCurrent = 0x0647
DAQmx_SwitchChan_MaxDCSwitchCurrent = 0x0645
DAQmx_SwitchChan_MaxDCCarryPwr = 0x0643
DAQmx_SwitchChan_MaxDCSwitchPwr = 0x0649
DAQmx_SwitchChan_MaxACVoltage = 0x0651
DAQmx_SwitchChan_MaxDCVoltage = 0x0650
DAQmx_SwitchChan_WireMode = 0x18E5
DAQmx_SwitchChan_Bandwidth = 0x0640
DAQmx_SwitchChan_Impedance = 0x0641
DAQmx_SwitchDev_SettlingTime = 0x1244
DAQmx_SwitchDev_AutoConnAnlgBus = 0x17DA
DAQmx_SwitchDev_PwrDownLatchRelaysAfterSettling = 0x22DB
DAQmx_SwitchDev_Settled = 0x1243
DAQmx_SwitchDev_RelayList = 0x17DC
DAQmx_SwitchDev_NumRelays = 0x18E6
DAQmx_SwitchDev_SwitchChanList = 0x18E7
DAQmx_SwitchDev_NumSwitchChans = 0x18E8
DAQmx_SwitchDev_NumRows = 0x18E9
DAQmx_SwitchDev_NumColumns = 0x18EA
DAQmx_SwitchDev_Topology = 0x193D
DAQmx_SwitchDev_Temperature = 0x301A
DAQmx_SwitchScan_BreakMode = 0x1247
DAQmx_SwitchScan_RepeatMode = 0x1248
DAQmx_SwitchScan_WaitingForAdv = 0x17D9
DAQmx_Scale_Descr = 0x1226
DAQmx_Scale_ScaledUnits = 0x191B
DAQmx_Scale_PreScaledUnits = 0x18F7
DAQmx_Scale_Type = 0x1929
DAQmx_Scale_Lin_Slope = 0x1227
DAQmx_Scale_Lin_YIntercept = 0x1228
DAQmx_Scale_Map_ScaledMax = 0x1229
DAQmx_Scale_Map_PreScaledMax = 0x1231
DAQmx_Scale_Map_ScaledMin = 0x1230
DAQmx_Scale_Map_PreScaledMin = 0x1232
DAQmx_Scale_Poly_ForwardCoeff = 0x1234
DAQmx_Scale_Poly_ReverseCoeff = 0x1235
DAQmx_Scale_Table_ScaledVals = 0x1236
DAQmx_Scale_Table_PreScaledVals = 0x1237
DAQmx_Sys_GlobalChans = 0x1265
DAQmx_Sys_Scales = 0x1266
DAQmx_Sys_Tasks = 0x1267
DAQmx_Sys_DevNames = 0x193B
DAQmx_Sys_NIDAQMajorVersion = 0x1272
DAQmx_Sys_NIDAQMinorVersion = 0x1923
DAQmx_Sys_NIDAQUpdateVersion = 0x2F22
DAQmx_Task_Name = 0x1276
DAQmx_Task_Channels = 0x1273
DAQmx_Task_NumChans = 0x2181
DAQmx_Task_Devices = 0x230E
DAQmx_Task_NumDevices = 0x29BA
DAQmx_Task_Complete = 0x1274
DAQmx_SampQuant_SampMode = 0x1300
DAQmx_SampQuant_SampPerChan = 0x1310
DAQmx_SampTimingType = 0x1347
DAQmx_SampClk_Rate = 0x1344
DAQmx_SampClk_MaxRate = 0x22C8
DAQmx_SampClk_Src = 0x1852
DAQmx_SampClk_ActiveEdge = 0x1301
DAQmx_SampClk_OverrunBehavior = 0x2EFC
DAQmx_SampClk_UnderflowBehavior = 0x2961
DAQmx_SampClk_TimebaseDiv = 0x18EB
DAQmx_SampClk_Term = 0x2F1B
DAQmx_SampClk_Timebase_Rate = 0x1303
DAQmx_SampClk_Timebase_Src = 0x1308
DAQmx_SampClk_Timebase_ActiveEdge = 0x18EC
DAQmx_SampClk_Timebase_MasterTimebaseDiv = 0x1305
DAQmx_SampClkTimebase_Term = 0x2F1C
DAQmx_SampClk_DigFltr_Enable = 0x221E
DAQmx_SampClk_DigFltr_MinPulseWidth = 0x221F
DAQmx_SampClk_DigFltr_TimebaseSrc = 0x2220
DAQmx_SampClk_DigFltr_TimebaseRate = 0x2221
DAQmx_SampClk_DigSync_Enable = 0x2222
DAQmx_Hshk_DelayAfterXfer = 0x22C2
DAQmx_Hshk_StartCond = 0x22C3
DAQmx_Hshk_SampleInputDataWhen = 0x22C4
DAQmx_ChangeDetect_DI_RisingEdgePhysicalChans = 0x2195
DAQmx_ChangeDetect_DI_FallingEdgePhysicalChans = 0x2196
DAQmx_ChangeDetect_DI_Tristate = 0x2EFA
DAQmx_OnDemand_SimultaneousAOEnable = 0x21A0
DAQmx_Implicit_UnderflowBehavior = 0x2EFD
DAQmx_AIConv_Rate = 0x1848
DAQmx_AIConv_MaxRate = 0x22C9
DAQmx_AIConv_Src = 0x1502
DAQmx_AIConv_ActiveEdge = 0x1853
DAQmx_AIConv_TimebaseDiv = 0x1335
DAQmx_AIConv_Timebase_Src = 0x1339
DAQmx_DelayFromSampClk_DelayUnits = 0x1304
DAQmx_DelayFromSampClk_Delay = 0x1317
DAQmx_AIConv_DigFltr_Enable = 0x2EDC
DAQmx_AIConv_DigFltr_MinPulseWidth = 0x2EDD
DAQmx_AIConv_DigFltr_TimebaseSrc = 0x2EDE
DAQmx_AIConv_DigFltr_TimebaseRate = 0x2EDF
DAQmx_AIConv_DigSync_Enable = 0x2EE0
DAQmx_MasterTimebase_Rate = 0x1495
DAQmx_MasterTimebase_Src = 0x1343
DAQmx_RefClk_Rate = 0x1315
DAQmx_RefClk_Src = 0x1316
DAQmx_SyncPulse_Src = 0x223D
DAQmx_SyncPulse_SyncTime = 0x223E
DAQmx_SyncPulse_MinDelayToStart = 0x223F
DAQmx_SyncPulse_ResetTime = 0x2F7C
DAQmx_SyncPulse_ResetDelay = 0x2F7D
DAQmx_SyncPulse_Term = 0x2F85
DAQmx_SyncClk_Interval = 0x2F7E
DAQmx_SampTimingEngine = 0x2A26
DAQmx_StartTrig_Type = 0x1393
DAQmx_StartTrig_Term = 0x2F1E
DAQmx_DigEdge_StartTrig_Src = 0x1407
DAQmx_DigEdge_StartTrig_Edge = 0x1404
DAQmx_DigEdge_StartTrig_DigFltr_Enable = 0x2223
DAQmx_DigEdge_StartTrig_DigFltr_MinPulseWidth = 0x2224
DAQmx_DigEdge_StartTrig_DigFltr_TimebaseSrc = 0x2225
DAQmx_DigEdge_StartTrig_DigFltr_TimebaseRate = 0x2226
DAQmx_DigEdge_StartTrig_DigSync_Enable = 0x2227
DAQmx_DigPattern_StartTrig_Src = 0x1410
DAQmx_DigPattern_StartTrig_Pattern = 0x2186
DAQmx_DigPattern_StartTrig_When = 0x1411
DAQmx_AnlgEdge_StartTrig_Src = 0x1398
DAQmx_AnlgEdge_StartTrig_Slope = 0x1397
DAQmx_AnlgEdge_StartTrig_Lvl = 0x1396
DAQmx_AnlgEdge_StartTrig_Hyst = 0x1395
DAQmx_AnlgEdge_StartTrig_Coupling = 0x2233
DAQmx_AnlgEdge_StartTrig_DigFltr_Enable = 0x2EE1
DAQmx_AnlgEdge_StartTrig_DigFltr_MinPulseWidth = 0x2EE2
DAQmx_AnlgEdge_StartTrig_DigFltr_TimebaseSrc = 0x2EE3
DAQmx_AnlgEdge_StartTrig_DigFltr_TimebaseRate = 0x2EE4
DAQmx_AnlgEdge_StartTrig_DigSync_Enable = 0x2EE5
DAQmx_AnlgWin_StartTrig_Src = 0x1400
DAQmx_AnlgWin_StartTrig_When = 0x1401
DAQmx_AnlgWin_StartTrig_Top = 0x1403
DAQmx_AnlgWin_StartTrig_Btm = 0x1402
DAQmx_AnlgWin_StartTrig_Coupling = 0x2234
DAQmx_AnlgWin_StartTrig_DigFltr_Enable = 0x2EFF
DAQmx_AnlgWin_StartTrig_DigFltr_MinPulseWidth = 0x2F00
DAQmx_AnlgWin_StartTrig_DigFltr_TimebaseSrc = 0x2F01
DAQmx_AnlgWin_StartTrig_DigFltr_TimebaseRate = 0x2F02
DAQmx_AnlgWin_StartTrig_DigSync_Enable = 0x2F03
DAQmx_StartTrig_Delay = 0x1856
DAQmx_StartTrig_DelayUnits = 0x18C8
DAQmx_StartTrig_Retriggerable = 0x190F
DAQmx_RefTrig_Type = 0x1419
DAQmx_RefTrig_PretrigSamples = 0x1445
DAQmx_RefTrig_Term = 0x2F1F
DAQmx_DigEdge_RefTrig_Src = 0x1434
DAQmx_DigEdge_RefTrig_Edge = 0x1430
DAQmx_DigEdge_RefTrig_DigFltr_Enable = 0x2ED7
DAQmx_DigEdge_RefTrig_DigFltr_MinPulseWidth = 0x2ED8
DAQmx_DigEdge_RefTrig_DigFltr_TimebaseSrc = 0x2ED9
DAQmx_DigEdge_RefTrig_DigFltr_TimebaseRate = 0x2EDA
DAQmx_DigEdge_RefTrig_DigSync_Enable = 0x2EDB
DAQmx_DigPattern_RefTrig_Src = 0x1437
DAQmx_DigPattern_RefTrig_Pattern = 0x2187
DAQmx_DigPattern_RefTrig_When = 0x1438
DAQmx_AnlgEdge_RefTrig_Src = 0x1424
DAQmx_AnlgEdge_RefTrig_Slope = 0x1423
DAQmx_AnlgEdge_RefTrig_Lvl = 0x1422
DAQmx_AnlgEdge_RefTrig_Hyst = 0x1421
DAQmx_AnlgEdge_RefTrig_Coupling = 0x2235
DAQmx_AnlgEdge_RefTrig_DigFltr_Enable = 0x2EE6
DAQmx_AnlgEdge_RefTrig_DigFltr_MinPulseWidth = 0x2EE7
DAQmx_AnlgEdge_RefTrig_DigFltr_TimebaseSrc = 0x2EE8
DAQmx_AnlgEdge_RefTrig_DigFltr_TimebaseRate = 0x2EE9
DAQmx_AnlgEdge_RefTrig_DigSync_Enable = 0x2EEA
DAQmx_AnlgWin_RefTrig_Src = 0x1426
DAQmx_AnlgWin_RefTrig_When = 0x1427
DAQmx_AnlgWin_RefTrig_Top = 0x1429
DAQmx_AnlgWin_RefTrig_Btm = 0x1428
DAQmx_AnlgWin_RefTrig_Coupling = 0x1857
DAQmx_AnlgWin_RefTrig_DigFltr_Enable = 0x2EEB
DAQmx_AnlgWin_RefTrig_DigFltr_MinPulseWidth = 0x2EEC
DAQmx_AnlgWin_RefTrig_DigFltr_TimebaseSrc = 0x2EED
DAQmx_AnlgWin_RefTrig_DigFltr_TimebaseRate = 0x2EEE
DAQmx_AnlgWin_RefTrig_DigSync_Enable = 0x2EEF
DAQmx_RefTrig_AutoTrigEnable = 0x2EC1
DAQmx_RefTrig_AutoTriggered = 0x2EC2
DAQmx_RefTrig_Delay = 0x1483
DAQmx_AdvTrig_Type = 0x1365
DAQmx_DigEdge_AdvTrig_Src = 0x1362
DAQmx_DigEdge_AdvTrig_Edge = 0x1360
DAQmx_DigEdge_AdvTrig_DigFltr_Enable = 0x2238
DAQmx_HshkTrig_Type = 0x22B7
DAQmx_Interlocked_HshkTrig_Src = 0x22B8
DAQmx_Interlocked_HshkTrig_AssertedLvl = 0x22B9
DAQmx_PauseTrig_Type = 0x1366
DAQmx_PauseTrig_Term = 0x2F20
DAQmx_AnlgLvl_PauseTrig_Src = 0x1370
DAQmx_AnlgLvl_PauseTrig_When = 0x1371
DAQmx_AnlgLvl_PauseTrig_Lvl = 0x1369
DAQmx_AnlgLvl_PauseTrig_Hyst = 0x1368
DAQmx_AnlgLvl_PauseTrig_Coupling = 0x2236
DAQmx_AnlgLvl_PauseTrig_DigFltr_Enable = 0x2EF0
DAQmx_AnlgLvl_PauseTrig_DigFltr_MinPulseWidth = 0x2EF1
DAQmx_AnlgLvl_PauseTrig_DigFltr_TimebaseSrc = 0x2EF2
DAQmx_AnlgLvl_PauseTrig_DigFltr_TimebaseRate = 0x2EF3
DAQmx_AnlgLvl_PauseTrig_DigSync_Enable = 0x2EF4
DAQmx_AnlgWin_PauseTrig_Src = 0x1373
DAQmx_AnlgWin_PauseTrig_When = 0x1374
DAQmx_AnlgWin_PauseTrig_Top = 0x1376
DAQmx_AnlgWin_PauseTrig_Btm = 0x1375
DAQmx_AnlgWin_PauseTrig_Coupling = 0x2237
DAQmx_AnlgWin_PauseTrig_DigFltr_Enable = 0x2EF5
DAQmx_AnlgWin_PauseTrig_DigFltr_MinPulseWidth = 0x2EF6
DAQmx_AnlgWin_PauseTrig_DigFltr_TimebaseSrc = 0x2EF7
DAQmx_AnlgWin_PauseTrig_DigFltr_TimebaseRate = 0x2EF8
DAQmx_AnlgWin_PauseTrig_DigSync_Enable = 0x2EF9
DAQmx_DigLvl_PauseTrig_Src = 0x1379
DAQmx_DigLvl_PauseTrig_When = 0x1380
DAQmx_DigLvl_PauseTrig_DigFltr_Enable = 0x2228
DAQmx_DigLvl_PauseTrig_DigFltr_MinPulseWidth = 0x2229
DAQmx_DigLvl_PauseTrig_DigFltr_TimebaseSrc = 0x222A
DAQmx_DigLvl_PauseTrig_DigFltr_TimebaseRate = 0x222B
DAQmx_DigLvl_PauseTrig_DigSync_Enable = 0x222C
DAQmx_DigPattern_PauseTrig_Src = 0x216F
DAQmx_DigPattern_PauseTrig_Pattern = 0x2188
DAQmx_DigPattern_PauseTrig_When = 0x2170
DAQmx_ArmStartTrig_Type = 0x1414
DAQmx_ArmStart_Term = 0x2F7F
DAQmx_DigEdge_ArmStartTrig_Src = 0x1417
DAQmx_DigEdge_ArmStartTrig_Edge = 0x1415
DAQmx_DigEdge_ArmStartTrig_DigFltr_Enable = 0x222D
DAQmx_DigEdge_ArmStartTrig_DigFltr_MinPulseWidth = 0x222E
DAQmx_DigEdge_ArmStartTrig_DigFltr_TimebaseSrc = 0x222F
DAQmx_DigEdge_ArmStartTrig_DigFltr_TimebaseRate = 0x2230
DAQmx_DigEdge_ArmStartTrig_DigSync_Enable = 0x2231
DAQmx_Trigger_SyncType = 0x2F80
DAQmx_Watchdog_Timeout = 0x21A9
DAQmx_WatchdogExpirTrig_Type = 0x21A3
DAQmx_WatchdogExpirTrig_TrigOnNetworkConnLoss = 0x305D
DAQmx_DigEdge_WatchdogExpirTrig_Src = 0x21A4
DAQmx_DigEdge_WatchdogExpirTrig_Edge = 0x21A5
DAQmx_Watchdog_DO_ExpirState = 0x21A7
DAQmx_Watchdog_AO_OutputType = 0x305E
DAQmx_Watchdog_AO_ExpirState = 0x305F
DAQmx_Watchdog_CO_ExpirState = 0x3060
DAQmx_Watchdog_HasExpired = 0x21A8
DAQmx_Write_RelativeTo = 0x190C
DAQmx_Write_Offset = 0x190D
DAQmx_Write_RegenMode = 0x1453
DAQmx_Write_CurrWritePos = 0x1458
DAQmx_Write_OvercurrentChansExist = 0x29E8
DAQmx_Write_OvercurrentChans = 0x29E9
DAQmx_Write_OvertemperatureChansExist = 0x2A84
DAQmx_Write_OvertemperatureChans = 0x3083
DAQmx_Write_OverloadedChansExist = 0x3084
DAQmx_Write_OverloadedChans = 0x3085
DAQmx_Write_OpenCurrentLoopChansExist = 0x29EA
DAQmx_Write_OpenCurrentLoopChans = 0x29EB
DAQmx_Write_PowerSupplyFaultChansExist = 0x29EC
DAQmx_Write_PowerSupplyFaultChans = 0x29ED
DAQmx_Write_SpaceAvail = 0x1460
DAQmx_Write_TotalSampPerChanGenerated = 0x192B
DAQmx_Write_RawDataWidth = 0x217D
DAQmx_Write_NumChans = 0x217E
DAQmx_Write_WaitMode = 0x22B1
DAQmx_Write_SleepTime = 0x22B2
DAQmx_Write_NextWriteIsLast = 0x296C
DAQmx_Write_DigitalLines_BytesPerChan = 0x217F
DAQmx_Write_AccessoryInsertionOrRemovalDetected = 0x3053
DAQmx_Write_DevsWithInsertedOrRemovedAccessories = 0x3054
DAQmx_PhysicalChan_AI_SupportedMeasTypes = 0x2FD7
DAQmx_PhysicalChan_AI_TermCfgs = 0x2342
DAQmx_PhysicalChan_AI_InputSrcs = 0x2FD8
DAQmx_PhysicalChan_AO_SupportedOutputTypes = 0x2FD9
DAQmx_PhysicalChan_AO_SupportedPowerUpOutputTypes = 0x304E
DAQmx_PhysicalChan_AO_TermCfgs = 0x29A3
DAQmx_PhysicalChan_AO_ManualControlEnable = 0x2A1E
DAQmx_PhysicalChan_AO_ManualControl_ShortDetected = 0x2EC3
DAQmx_PhysicalChan_AO_ManualControlAmplitude = 0x2A1F
DAQmx_PhysicalChan_AO_ManualControlFreq = 0x2A20
DAQmx_AO_PowerAmp_ChannelEnable = 0x3062
DAQmx_AO_PowerAmp_ScalingCoeff = 0x3063
DAQmx_AO_PowerAmp_Overcurrent = 0x3064
DAQmx_AO_PowerAmp_Gain = 0x3065
DAQmx_AO_PowerAmp_Offset = 0x3066
DAQmx_PhysicalChan_DI_PortWidth = 0x29A4
DAQmx_PhysicalChan_DI_SampClkSupported = 0x29A5
DAQmx_PhysicalChan_DI_SampModes = 0x2FE0
DAQmx_PhysicalChan_DI_ChangeDetectSupported = 0x29A6
DAQmx_PhysicalChan_DO_PortWidth = 0x29A7
DAQmx_PhysicalChan_DO_SampClkSupported = 0x29A8
DAQmx_PhysicalChan_DO_SampModes = 0x2FE1
DAQmx_PhysicalChan_CI_SupportedMeasTypes = 0x2FDA
DAQmx_PhysicalChan_CO_SupportedOutputTypes = 0x2FDB
DAQmx_PhysicalChan_TEDS_MfgID = 0x21DA
DAQmx_PhysicalChan_TEDS_ModelNum = 0x21DB
DAQmx_PhysicalChan_TEDS_SerialNum = 0x21DC
DAQmx_PhysicalChan_TEDS_VersionNum = 0x21DD
DAQmx_PhysicalChan_TEDS_VersionLetter = 0x21DE
DAQmx_PhysicalChan_TEDS_BitStream = 0x21DF
DAQmx_PhysicalChan_TEDS_TemplateIDs = 0x228F
DAQmx_PersistedTask_Author = 0x22CC
DAQmx_PersistedTask_AllowInteractiveEditing = 0x22CD
DAQmx_PersistedTask_AllowInteractiveDeletion = 0x22CE
DAQmx_PersistedChan_Author = 0x22D0
DAQmx_PersistedChan_AllowInteractiveEditing = 0x22D1
DAQmx_PersistedChan_AllowInteractiveDeletion = 0x22D2
DAQmx_PersistedScale_Author = 0x22D4
DAQmx_PersistedScale_AllowInteractiveEditing = 0x22D5
DAQmx_PersistedScale_AllowInteractiveDeletion = 0x22D6
DAQmx_ReadWaitMode = DAQmx_Read_WaitMode
DAQmx_Val_Task_Start = 0
DAQmx_Val_Task_Stop = 1
DAQmx_Val_Task_Verify = 2
DAQmx_Val_Task_Commit = 3
DAQmx_Val_Task_Reserve = 4
DAQmx_Val_Task_Unreserve = 5
DAQmx_Val_Task_Abort = 6
DAQmx_Val_SynchronousEventCallbacks = (1 << 0)
DAQmx_Val_Acquired_Into_Buffer = 1
DAQmx_Val_Transferred_From_Buffer = 2
DAQmx_Val_ResetTimer = 0
DAQmx_Val_ClearExpiration = 1
DAQmx_Val_ChanPerLine = 0
DAQmx_Val_ChanForAllLines = 1
DAQmx_Val_GroupByChannel = 0
DAQmx_Val_GroupByScanNumber = 1
DAQmx_Val_DoNotInvertPolarity = 0
DAQmx_Val_InvertPolarity = 1
DAQmx_Val_Action_Commit = 0
DAQmx_Val_Action_Cancel = 1
DAQmx_Val_AdvanceTrigger = 12488
DAQmx_Val_Rising = 10280
DAQmx_Val_Falling = 10171
DAQmx_Val_PathStatus_Available = 10431
DAQmx_Val_PathStatus_AlreadyExists = 10432
DAQmx_Val_PathStatus_Unsupported = 10433
DAQmx_Val_PathStatus_ChannelInUse = 10434
DAQmx_Val_PathStatus_SourceChannelConflict = 10435
DAQmx_Val_PathStatus_ChannelReservedForRouting = 10436
DAQmx_Val_DegC = 10143
DAQmx_Val_DegF = 10144
DAQmx_Val_Kelvins = 10325
DAQmx_Val_DegR = 10145
DAQmx_Val_High = 10192
DAQmx_Val_Low = 10214
DAQmx_Val_Tristate = 10310
DAQmx_Val_PullUp = 15950
DAQmx_Val_PullDown = 15951
DAQmx_Val_ChannelVoltage = 0
DAQmx_Val_ChannelCurrent = 1
DAQmx_Val_ChannelHighImpedance = 2
DAQmx_Val_Open = 10437
DAQmx_Val_Closed = 10438
DAQmx_Val_Loopback0 = 0
DAQmx_Val_Loopback180 = 1
DAQmx_Val_Ground = 2
DAQmx_Val_Voltage = 10322
DAQmx_Val_Current = 10134
DAQmx_Val_Cfg_Default = -1
DAQmx_Val_Default = -1
DAQmx_Val_WaitInfinitely = -1.0
DAQmx_Val_Auto = -1
DAQmx_Val_Save_Overwrite = (1 << 0)
DAQmx_Val_Save_AllowInteractiveEditing = (1 << 1)
DAQmx_Val_Save_AllowInteractiveDeletion = (1 << 2)
DAQmx_Val_Bit_TriggerUsageTypes_Advance = (1 << 0)
DAQmx_Val_Bit_TriggerUsageTypes_Pause = (1 << 1)
DAQmx_Val_Bit_TriggerUsageTypes_Reference = (1 << 2)
DAQmx_Val_Bit_TriggerUsageTypes_Start = (1 << 3)
DAQmx_Val_Bit_TriggerUsageTypes_Handshake = (1 << 4)
DAQmx_Val_Bit_TriggerUsageTypes_ArmStart = (1 << 5)
DAQmx_Val_Bit_CouplingTypes_AC = (1 << 0)
DAQmx_Val_Bit_CouplingTypes_DC = (1 << 1)
DAQmx_Val_Bit_CouplingTypes_Ground = (1 << 2)
DAQmx_Val_Bit_CouplingTypes_HFReject = (1 << 3)
DAQmx_Val_Bit_CouplingTypes_LFReject = (1 << 4)
DAQmx_Val_Bit_CouplingTypes_NoiseReject = (1 << 5)
DAQmx_Val_Bit_TermCfg_RSE = (1 << 0)
DAQmx_Val_Bit_TermCfg_NRSE = (1 << 1)
DAQmx_Val_Bit_TermCfg_Diff = (1 << 2)
DAQmx_Val_Bit_TermCfg_PseudoDIFF = (1 << 3)
DAQmx_Val_4Wire = 4
DAQmx_Val_5Wire = 5
DAQmx_Val_HighResolution = 10195
DAQmx_Val_HighSpeed = 14712
DAQmx_Val_Best50HzRejection = 14713
DAQmx_Val_Best60HzRejection = 14714
DAQmx_Val_Custom = 10137
DAQmx_Val_Voltage = 10322
DAQmx_Val_VoltageRMS = 10350
DAQmx_Val_Current = 10134
DAQmx_Val_CurrentRMS = 10351
DAQmx_Val_Voltage_CustomWithExcitation = 10323
DAQmx_Val_Bridge = 15908
DAQmx_Val_Freq_Voltage = 10181
DAQmx_Val_Resistance = 10278
DAQmx_Val_Temp_TC = 10303
DAQmx_Val_Temp_Thrmstr = 10302
DAQmx_Val_Temp_RTD = 10301
DAQmx_Val_Temp_BuiltInSensor = 10311
DAQmx_Val_Strain_Gage = 10300
DAQmx_Val_Rosette_Strain_Gage = 15980
DAQmx_Val_Position_LVDT = 10352
DAQmx_Val_Position_RVDT = 10353
DAQmx_Val_Position_EddyCurrentProximityProbe = 14835
DAQmx_Val_Accelerometer = 10356
DAQmx_Val_Velocity_IEPESensor = 15966
DAQmx_Val_Force_Bridge = 15899
DAQmx_Val_Force_IEPESensor = 15895
DAQmx_Val_Pressure_Bridge = 15902
DAQmx_Val_SoundPressure_Microphone = 10354
DAQmx_Val_Torque_Bridge = 15905
DAQmx_Val_TEDS_Sensor = 12531
DAQmx_Val_ZeroVolts = 12526
DAQmx_Val_HighImpedance = 12527
DAQmx_Val_MaintainExistingValue = 12528
DAQmx_Val_Voltage = 10322
DAQmx_Val_Current = 10134
DAQmx_Val_FuncGen = 14750
DAQmx_Val_mVoltsPerG = 12509
DAQmx_Val_VoltsPerG = 12510
DAQmx_Val_AccelUnit_g = 10186
DAQmx_Val_MetersPerSecondSquared = 12470
DAQmx_Val_InchesPerSecondSquared = 12471
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FiniteSamps = 10178
DAQmx_Val_ContSamps = 10123
DAQmx_Val_HWTimedSinglePoint = 12522
DAQmx_Val_AboveLvl = 10093
DAQmx_Val_BelowLvl = 10107
DAQmx_Val_Degrees = 10146
DAQmx_Val_Radians = 10273
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Degrees = 10146
DAQmx_Val_Radians = 10273
DAQmx_Val_Ticks = 10304
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_None = 10230
DAQmx_Val_Once = 10244
DAQmx_Val_EverySample = 10164
DAQmx_Val_NoAction = 10227
DAQmx_Val_BreakBeforeMake = 10110
DAQmx_Val_FullBridge = 10182
DAQmx_Val_HalfBridge = 10187
DAQmx_Val_QuarterBridge = 10270
DAQmx_Val_NoBridge = 10228
DAQmx_Val_VoltsPerVolt = 15896
DAQmx_Val_mVoltsPerVolt = 15897
DAQmx_Val_Newtons = 15875
DAQmx_Val_Pounds = 15876
DAQmx_Val_KilogramForce = 15877
DAQmx_Val_Pascals = 10081
DAQmx_Val_PoundsPerSquareInch = 15879
DAQmx_Val_Bar = 15880
DAQmx_Val_NewtonMeters = 15881
DAQmx_Val_InchOunces = 15882
DAQmx_Val_InchPounds = 15883
DAQmx_Val_FootPounds = 15884
DAQmx_Val_VoltsPerVolt = 15896
DAQmx_Val_mVoltsPerVolt = 15897
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_PCI = 12582
DAQmx_Val_PCIe = 13612
DAQmx_Val_PXI = 12583
DAQmx_Val_PXIe = 14706
DAQmx_Val_SCXI = 12584
DAQmx_Val_SCC = 14707
DAQmx_Val_PCCard = 12585
DAQmx_Val_USB = 12586
DAQmx_Val_CompactDAQ = 14637
DAQmx_Val_TCPIP = 14828
DAQmx_Val_Unknown = 12588
DAQmx_Val_SwitchBlock = 15870
DAQmx_Val_CountEdges = 10125
DAQmx_Val_Freq = 10179
DAQmx_Val_Period = 10256
DAQmx_Val_PulseWidth = 10359
DAQmx_Val_SemiPeriod = 10289
DAQmx_Val_PulseFrequency = 15864
DAQmx_Val_PulseTime = 15865
DAQmx_Val_PulseTicks = 15866
DAQmx_Val_Position_AngEncoder = 10360
DAQmx_Val_Position_LinEncoder = 10361
DAQmx_Val_TwoEdgeSep = 10267
DAQmx_Val_GPS_Timestamp = 10362
DAQmx_Val_BuiltIn = 10200
DAQmx_Val_ConstVal = 10116
DAQmx_Val_Chan = 10113
DAQmx_Val_Pulse_Time = 10269
DAQmx_Val_Pulse_Freq = 10119
DAQmx_Val_Pulse_Ticks = 10268
DAQmx_Val_AI = 10100
DAQmx_Val_AO = 10102
DAQmx_Val_DI = 10151
DAQmx_Val_DO = 10153
DAQmx_Val_CI = 10131
DAQmx_Val_CO = 10132
DAQmx_Val_Unconstrained = 14708
DAQmx_Val_FixedHighFreq = 14709
DAQmx_Val_FixedLowFreq = 14710
DAQmx_Val_Fixed50PercentDutyCycle = 14711
DAQmx_Val_CountUp = 10128
DAQmx_Val_CountDown = 10124
DAQmx_Val_ExtControlled = 10326
DAQmx_Val_LowFreq1Ctr = 10105
DAQmx_Val_HighFreq2Ctr = 10157
DAQmx_Val_LargeRng2Ctr = 10205
DAQmx_Val_AC = 10045
DAQmx_Val_DC = 10050
DAQmx_Val_GND = 10066
DAQmx_Val_AC = 10045
DAQmx_Val_DC = 10050
DAQmx_Val_Internal = 10200
DAQmx_Val_External = 10167
DAQmx_Val_Amps = 10342
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_Amps = 10342
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_RightJustified = 10279
DAQmx_Val_LeftJustified = 10209
DAQmx_Val_DMA = 10054
DAQmx_Val_Interrupts = 10204
DAQmx_Val_ProgrammedIO = 10264
DAQmx_Val_USBbulk = 12590
DAQmx_Val_OnbrdMemMoreThanHalfFull = 10237
DAQmx_Val_OnbrdMemFull = 10236
DAQmx_Val_OnbrdMemCustomThreshold = 12577
DAQmx_Val_ActiveDrive = 12573
DAQmx_Val_OpenCollector = 12574
DAQmx_Val_High = 10192
DAQmx_Val_Low = 10214
DAQmx_Val_Tristate = 10310
DAQmx_Val_NoChange = 10160
DAQmx_Val_PatternMatches = 10254
DAQmx_Val_PatternDoesNotMatch = 10253
DAQmx_Val_SampClkPeriods = 10286
DAQmx_Val_Seconds = 10364
DAQmx_Val_Ticks = 10304
DAQmx_Val_Seconds = 10364
DAQmx_Val_Ticks = 10304
DAQmx_Val_Seconds = 10364
DAQmx_Val_Seconds = 10364
DAQmx_Val_SampleClkPeriods = 10286
DAQmx_Val_mVoltsPerMil = 14836
DAQmx_Val_VoltsPerMil = 14837
DAQmx_Val_mVoltsPerMillimeter = 14838
DAQmx_Val_VoltsPerMillimeter = 14839
DAQmx_Val_mVoltsPerMicron = 14840
DAQmx_Val_Rising = 10280
DAQmx_Val_Falling = 10171
DAQmx_Val_X1 = 10090
DAQmx_Val_X2 = 10091
DAQmx_Val_X4 = 10092
DAQmx_Val_TwoPulseCounting = 10313
DAQmx_Val_AHighBHigh = 10040
DAQmx_Val_AHighBLow = 10041
DAQmx_Val_ALowBHigh = 10042
DAQmx_Val_ALowBLow = 10043
DAQmx_Val_DC = 10050
DAQmx_Val_AC = 10045
DAQmx_Val_Internal = 10200
DAQmx_Val_External = 10167
DAQmx_Val_None = 10230
DAQmx_Val_Voltage = 10322
DAQmx_Val_Current = 10134
DAQmx_Val_Pulse = 10265
DAQmx_Val_Toggle = 10307
DAQmx_Val_Pulse = 10265
DAQmx_Val_Lvl = 10210
DAQmx_Val_Interlocked = 12549
DAQmx_Val_Pulse = 10265
DAQmx_Val_mVoltsPerNewton = 15891
DAQmx_Val_mVoltsPerPound = 15892
DAQmx_Val_Newtons = 15875
DAQmx_Val_Pounds = 15876
DAQmx_Val_KilogramForce = 15877
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Hz = 10373
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Hz = 10373
DAQmx_Val_Hz = 10373
DAQmx_Val_Ticks = 10304
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Sine = 14751
DAQmx_Val_Triangle = 14752
DAQmx_Val_Square = 14753
DAQmx_Val_Sawtooth = 14754
DAQmx_Val_IRIGB = 10070
DAQmx_Val_PPS = 10080
DAQmx_Val_None = 10230
DAQmx_Val_Immediate = 10198
DAQmx_Val_WaitForHandshakeTriggerAssert = 12550
DAQmx_Val_WaitForHandshakeTriggerDeassert = 12551
DAQmx_Val_OnBrdMemMoreThanHalfFull = 10237
DAQmx_Val_OnBrdMemNotEmpty = 10241
DAQmx_Val_OnbrdMemCustomThreshold = 12577
DAQmx_Val_WhenAcqComplete = 12546
DAQmx_Val_RSE = 10083
DAQmx_Val_NRSE = 10078
DAQmx_Val_Diff = 10106
DAQmx_Val_PseudoDiff = 12529
DAQmx_Val_mVoltsPerVoltPerMillimeter = 12506
DAQmx_Val_mVoltsPerVoltPerMilliInch = 12505
DAQmx_Val_Meters = 10219
DAQmx_Val_Inches = 10379
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Meters = 10219
DAQmx_Val_Inches = 10379
DAQmx_Val_Ticks = 10304
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_High = 10192
DAQmx_Val_Low = 10214
DAQmx_Val_Off = 10231
DAQmx_Val_Log = 15844
DAQmx_Val_LogAndRead = 15842
DAQmx_Val_Open = 10437
DAQmx_Val_OpenOrCreate = 15846
DAQmx_Val_CreateOrReplace = 15847
DAQmx_Val_Create = 15848
DAQmx_Val_2point5V = 14620
DAQmx_Val_3point3V = 14621
DAQmx_Val_5V = 14619
DAQmx_Val_SameAsSampTimebase = 10284
DAQmx_Val_SameAsMasterTimebase = 10282
DAQmx_Val_100MHzTimebase = 15857
DAQmx_Val_80MHzTimebase = 14636
DAQmx_Val_20MHzTimebase = 12537
DAQmx_Val_8MHzTimebase = 16023
DAQmx_Val_AM = 14756
DAQmx_Val_FM = 14757
DAQmx_Val_None = 10230
DAQmx_Val_OnBrdMemEmpty = 10235
DAQmx_Val_OnBrdMemHalfFullOrLess = 10239
DAQmx_Val_OnBrdMemNotFull = 10242
DAQmx_Val_RSE = 10083
DAQmx_Val_Diff = 10106
DAQmx_Val_PseudoDiff = 12529
DAQmx_Val_StopTaskAndError = 15862
DAQmx_Val_IgnoreOverruns = 15863
DAQmx_Val_OverwriteUnreadSamps = 10252
DAQmx_Val_DoNotOverwriteUnreadSamps = 10159
DAQmx_Val_ActiveHigh = 10095
DAQmx_Val_ActiveLow = 10096
DAQmx_Val_Pascals = 10081
DAQmx_Val_PoundsPerSquareInch = 15879
DAQmx_Val_Bar = 15880
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_MSeriesDAQ = 14643
DAQmx_Val_XSeriesDAQ = 15858
DAQmx_Val_ESeriesDAQ = 14642
DAQmx_Val_SSeriesDAQ = 14644
DAQmx_Val_BSeriesDAQ = 14662
DAQmx_Val_SCSeriesDAQ = 14645
DAQmx_Val_USBDAQ = 14646
DAQmx_Val_AOSeries = 14647
DAQmx_Val_DigitalIO = 14648
DAQmx_Val_TIOSeries = 14661
DAQmx_Val_DynamicSignalAcquisition = 14649
DAQmx_Val_Switches = 14650
DAQmx_Val_CompactDAQChassis = 14658
DAQmx_Val_CSeriesModule = 14659
DAQmx_Val_SCXIModule = 14660
DAQmx_Val_SCCConnectorBlock = 14704
DAQmx_Val_SCCModule = 14705
DAQmx_Val_NIELVIS = 14755
DAQmx_Val_NetworkDAQ = 14829
DAQmx_Val_SCExpress = 15886
DAQmx_Val_Unknown = 12588
DAQmx_Val_Pt3750 = 12481
DAQmx_Val_Pt3851 = 10071
DAQmx_Val_Pt3911 = 12482
DAQmx_Val_Pt3916 = 10069
DAQmx_Val_Pt3920 = 10053
DAQmx_Val_Pt3928 = 12483
DAQmx_Val_Custom = 10137
DAQmx_Val_mVoltsPerVoltPerDegree = 12507
DAQmx_Val_mVoltsPerVoltPerRadian = 12508
DAQmx_Val_None = 10230
DAQmx_Val_LosslessPacking = 12555
DAQmx_Val_LossyLSBRemoval = 12556
DAQmx_Val_FirstSample = 10424
DAQmx_Val_CurrReadPos = 10425
DAQmx_Val_RefTrig = 10426
DAQmx_Val_FirstPretrigSamp = 10427
DAQmx_Val_MostRecentSamp = 10428
DAQmx_Val_AllowRegen = 10097
DAQmx_Val_DoNotAllowRegen = 10158
DAQmx_Val_2Wire = 2
DAQmx_Val_3Wire = 3
DAQmx_Val_4Wire = 4
DAQmx_Val_Ohms = 10384
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_Ohms = 10384
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Bits = 10109
DAQmx_Val_SCXI1124Range0to1V = 14629
DAQmx_Val_SCXI1124Range0to5V = 14630
DAQmx_Val_SCXI1124Range0to10V = 14631
DAQmx_Val_SCXI1124RangeNeg1to1V = 14632
DAQmx_Val_SCXI1124RangeNeg5to5V = 14633
DAQmx_Val_SCXI1124RangeNeg10to10V = 14634
DAQmx_Val_SCXI1124Range0to20mA = 14635
DAQmx_Val_SampClkActiveEdge = 14617
DAQmx_Val_SampClkInactiveEdge = 14618
DAQmx_Val_HandshakeTriggerAsserts = 12552
DAQmx_Val_HandshakeTriggerDeasserts = 12553
DAQmx_Val_SampClk = 10388
DAQmx_Val_BurstHandshake = 12548
DAQmx_Val_Handshake = 10389
DAQmx_Val_Implicit = 10451
DAQmx_Val_OnDemand = 10390
DAQmx_Val_ChangeDetection = 12504
DAQmx_Val_PipelinedSampClk = 14668
DAQmx_Val_Linear = 10447
DAQmx_Val_MapRanges = 10448
DAQmx_Val_Polynomial = 10449
DAQmx_Val_Table = 10450
DAQmx_Val_Polynomial = 10449
DAQmx_Val_Table = 10450
DAQmx_Val_Polynomial = 10449
DAQmx_Val_Table = 10450
DAQmx_Val_None = 10230
DAQmx_Val_None = 10230
DAQmx_Val_TwoPointLinear = 15898
DAQmx_Val_Table = 10450
DAQmx_Val_Polynomial = 10449
DAQmx_Val_A = 12513
DAQmx_Val_B = 12514
DAQmx_Val_AandB = 12515
DAQmx_Val_R1 = 12465
DAQmx_Val_R2 = 12466
DAQmx_Val_R3 = 12467
DAQmx_Val_R4 = 14813
DAQmx_Val_None = 10230
DAQmx_Val_AIConvertClock = 12484
DAQmx_Val_10MHzRefClock = 12536
DAQmx_Val_20MHzTimebaseClock = 12486
DAQmx_Val_SampleClock = 12487
DAQmx_Val_AdvanceTrigger = 12488
DAQmx_Val_ReferenceTrigger = 12490
DAQmx_Val_StartTrigger = 12491
DAQmx_Val_AdvCmpltEvent = 12492
DAQmx_Val_AIHoldCmpltEvent = 12493
DAQmx_Val_CounterOutputEvent = 12494
DAQmx_Val_ChangeDetectionEvent = 12511
DAQmx_Val_WDTExpiredEvent = 12512
DAQmx_Val_SampleCompleteEvent = 12530
DAQmx_Val_CounterOutputEvent = 12494
DAQmx_Val_ChangeDetectionEvent = 12511
DAQmx_Val_SampleClock = 12487
DAQmx_Val_RisingSlope = 10280
DAQmx_Val_FallingSlope = 10171
DAQmx_Val_Pascals = 10081
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Internal = 10200
DAQmx_Val_External = 10167
DAQmx_Val_FullBridgeI = 10183
DAQmx_Val_FullBridgeII = 10184
DAQmx_Val_FullBridgeIII = 10185
DAQmx_Val_HalfBridgeI = 10188
DAQmx_Val_HalfBridgeII = 10189
DAQmx_Val_QuarterBridgeI = 10271
DAQmx_Val_QuarterBridgeII = 10272
DAQmx_Val_RectangularRosette = 15968
DAQmx_Val_DeltaRosette = 15969
DAQmx_Val_TeeRosette = 15970
DAQmx_Val_PrincipalStrain1 = 15971
DAQmx_Val_PrincipalStrain2 = 15972
DAQmx_Val_PrincipalStrainAngle = 15973
DAQmx_Val_CartesianStrainX = 15974
DAQmx_Val_CartesianStrainY = 15975
DAQmx_Val_CartesianShearStrainXY = 15976
DAQmx_Val_MaxShearStrain = 15977
DAQmx_Val_MaxShearStrainAngle = 15978
DAQmx_Val_Strain = 10299
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Finite = 10172
DAQmx_Val_Cont = 10117
DAQmx_Val_Source = 10439
DAQmx_Val_Load = 10440
DAQmx_Val_ReservedForRouting = 10441
DAQmx_Val_None = 10230
DAQmx_Val_Master = 15888
DAQmx_Val_Slave = 15889
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_DegC = 10143
DAQmx_Val_DegF = 10144
DAQmx_Val_Kelvins = 10325
DAQmx_Val_DegR = 10145
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_J_Type_TC = 10072
DAQmx_Val_K_Type_TC = 10073
DAQmx_Val_N_Type_TC = 10077
DAQmx_Val_R_Type_TC = 10082
DAQmx_Val_S_Type_TC = 10085
DAQmx_Val_T_Type_TC = 10086
DAQmx_Val_B_Type_TC = 10047
DAQmx_Val_E_Type_TC = 10055
DAQmx_Val_Seconds = 10364
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Seconds = 10364
DAQmx_Val_Seconds = 10364
DAQmx_Val_Ticks = 10304
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_SingleCycle = 14613
DAQmx_Val_Multicycle = 14614
DAQmx_Val_NewtonMeters = 15881
DAQmx_Val_InchOunces = 15882
DAQmx_Val_InchPounds = 15883
DAQmx_Val_FootPounds = 15884
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_DigEdge = 10150
DAQmx_Val_None = 10230
DAQmx_Val_DigEdge = 10150
DAQmx_Val_Software = 10292
DAQmx_Val_None = 10230
DAQmx_Val_AnlgLvl = 10101
DAQmx_Val_AnlgWin = 10103
DAQmx_Val_DigLvl = 10152
DAQmx_Val_DigPattern = 10398
DAQmx_Val_None = 10230
DAQmx_Val_AnlgEdge = 10099
DAQmx_Val_DigEdge = 10150
DAQmx_Val_DigPattern = 10398
DAQmx_Val_AnlgWin = 10103
DAQmx_Val_None = 10230
DAQmx_Val_Interlocked = 12549
DAQmx_Val_None = 10230
DAQmx_Val_AnlgEdge = 10099
DAQmx_Val_DigEdge = 10150
DAQmx_Val_DigPattern = 10398
DAQmx_Val_AnlgWin = 10103
DAQmx_Val_None = 10230
DAQmx_Val_HaltOutputAndError = 14615
DAQmx_Val_PauseUntilDataAvailable = 14616
DAQmx_Val_Volts = 10348
DAQmx_Val_Amps = 10342
DAQmx_Val_DegF = 10144
DAQmx_Val_DegC = 10143
DAQmx_Val_DegR = 10145
DAQmx_Val_Kelvins = 10325
DAQmx_Val_Strain = 10299
DAQmx_Val_Ohms = 10384
DAQmx_Val_Hz = 10373
DAQmx_Val_Seconds = 10364
DAQmx_Val_Meters = 10219
DAQmx_Val_Inches = 10379
DAQmx_Val_Degrees = 10146
DAQmx_Val_Radians = 10273
DAQmx_Val_Ticks = 10304
DAQmx_Val_g = 10186
DAQmx_Val_MetersPerSecondSquared = 12470
DAQmx_Val_Pascals = 10081
DAQmx_Val_Newtons = 15875
DAQmx_Val_Pounds = 15876
DAQmx_Val_KilogramForce = 15877
DAQmx_Val_PoundsPerSquareInch = 15879
DAQmx_Val_Bar = 15880
DAQmx_Val_NewtonMeters = 15881
DAQmx_Val_InchOunces = 15882
DAQmx_Val_InchPounds = 15883
DAQmx_Val_FootPounds = 15884
DAQmx_Val_VoltsPerVolt = 15896
DAQmx_Val_mVoltsPerVolt = 15897
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_MillivoltsPerMillimeterPerSecond = 15963
DAQmx_Val_MilliVoltsPerInchPerSecond = 15964
DAQmx_Val_MetersPerSecond = 15959
DAQmx_Val_InchesPerSecond = 15960
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_Volts = 10348
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_FromTEDS = 12516
DAQmx_Val_Volts = 10348
DAQmx_Val_FromCustomScale = 10065
DAQmx_Val_WaitForInterrupt = 12523
DAQmx_Val_Poll = 12524
DAQmx_Val_Yield = 12525
DAQmx_Val_Sleep = 12547
DAQmx_Val_Poll = 12524
DAQmx_Val_Yield = 12525
DAQmx_Val_Sleep = 12547
DAQmx_Val_WaitForInterrupt = 12523
DAQmx_Val_Poll = 12524
DAQmx_Val_WaitForInterrupt = 12523
DAQmx_Val_Poll = 12524
DAQmx_Val_Voltage = 10322
DAQmx_Val_Current = 10134
DAQmx_Val_NoChange = 10160
DAQmx_Val_Low = 10214
DAQmx_Val_High = 10192
DAQmx_Val_NoChange = 10160
DAQmx_Val_EnteringWin = 10163
DAQmx_Val_LeavingWin = 10208
DAQmx_Val_InsideWin = 10199
DAQmx_Val_OutsideWin = 10251
DAQmx_Val_WriteToEEPROM = 12538
DAQmx_Val_WriteToPROM = 12539
DAQmx_Val_DoNotWrite = 12540
DAQmx_Val_FirstSample = 10424
DAQmx_Val_CurrWritePos = 10430
DAQmx_Val_Switch_Topology_Configured_Topology = "Configured Topology"
DAQmx_Val_Switch_Topology_1127_1_Wire_64x1_Mux = "1127/1-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_1127_2_Wire_32x1_Mux = "1127/2-Wire 32x1 Mux"
DAQmx_Val_Switch_Topology_1127_2_Wire_4x8_Matrix = "1127/2-Wire 4x8 Matrix"
DAQmx_Val_Switch_Topology_1127_4_Wire_16x1_Mux = "1127/4-Wire 16x1 Mux"
DAQmx_Val_Switch_Topology_1127_Independent = "1127/Independent"
DAQmx_Val_Switch_Topology_1128_1_Wire_64x1_Mux = "1128/1-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_1128_2_Wire_32x1_Mux = "1128/2-Wire 32x1 Mux"
DAQmx_Val_Switch_Topology_1128_2_Wire_4x8_Matrix = "1128/2-Wire 4x8 Matrix"
DAQmx_Val_Switch_Topology_1128_4_Wire_16x1_Mux = "1128/4-Wire 16x1 Mux"
DAQmx_Val_Switch_Topology_1128_Independent = "1128/Independent"
DAQmx_Val_Switch_Topology_1129_2_Wire_16x16_Matrix = "1129/2-Wire 16x16 Matrix"
DAQmx_Val_Switch_Topology_1129_2_Wire_8x32_Matrix = "1129/2-Wire 8x32 Matrix"
DAQmx_Val_Switch_Topology_1129_2_Wire_4x64_Matrix = "1129/2-Wire 4x64 Matrix"
DAQmx_Val_Switch_Topology_1129_2_Wire_Dual_8x16_Matrix = "1129/2-Wire Dual 8x16 Matrix"
DAQmx_Val_Switch_Topology_1129_2_Wire_Dual_4x32_Matrix = "1129/2-Wire Dual 4x32 Matrix"
DAQmx_Val_Switch_Topology_1129_2_Wire_Quad_4x16_Matrix = "1129/2-Wire Quad 4x16 Matrix"
DAQmx_Val_Switch_Topology_1130_1_Wire_256x1_Mux = "1130/1-Wire 256x1 Mux"
DAQmx_Val_Switch_Topology_1130_1_Wire_Dual_128x1_Mux = "1130/1-Wire Dual 128x1 Mux"
DAQmx_Val_Switch_Topology_1130_2_Wire_128x1_Mux = "1130/2-Wire 128x1 Mux"
DAQmx_Val_Switch_Topology_1130_4_Wire_64x1_Mux = "1130/4-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_1130_1_Wire_4x64_Matrix = "1130/1-Wire 4x64 Matrix"
DAQmx_Val_Switch_Topology_1130_1_Wire_8x32_Matrix = "1130/1-Wire 8x32 Matrix"
DAQmx_Val_Switch_Topology_1130_1_Wire_Octal_32x1_Mux = "1130/1-Wire Octal 32x1 Mux"
DAQmx_Val_Switch_Topology_1130_1_Wire_Quad_64x1_Mux = "1130/1-Wire Quad 64x1 Mux"
DAQmx_Val_Switch_Topology_1130_1_Wire_Sixteen_16x1_Mux = "1130/1-Wire Sixteen 16x1 Mux"
DAQmx_Val_Switch_Topology_1130_2_Wire_4x32_Matrix = "1130/2-Wire 4x32 Matrix"
DAQmx_Val_Switch_Topology_1130_2_Wire_Octal_16x1_Mux = "1130/2-Wire Octal 16x1 Mux"
DAQmx_Val_Switch_Topology_1130_2_Wire_Quad_32x1_Mux = "1130/2-Wire Quad 32x1 Mux"
DAQmx_Val_Switch_Topology_1130_4_Wire_Quad_16x1_Mux = "1130/4-Wire Quad 16x1 Mux"
DAQmx_Val_Switch_Topology_1130_Independent = "1130/Independent"
DAQmx_Val_Switch_Topology_1160_16_SPDT = "1160/16-SPDT"
DAQmx_Val_Switch_Topology_1161_8_SPDT = "1161/8-SPDT"
DAQmx_Val_Switch_Topology_1163R_Octal_4x1_Mux = "1163R/Octal 4x1 Mux"
DAQmx_Val_Switch_Topology_1166_32_SPDT = "1166/32-SPDT"
DAQmx_Val_Switch_Topology_1166_16_DPDT = "1166/16-DPDT"
DAQmx_Val_Switch_Topology_1167_Independent = "1167/Independent"
DAQmx_Val_Switch_Topology_1169_100_SPST = "1169/100-SPST"
DAQmx_Val_Switch_Topology_1169_50_DPST = "1169/50-DPST"
DAQmx_Val_Switch_Topology_1175_1_Wire_196x1_Mux = "1175/1-Wire 196x1 Mux"
DAQmx_Val_Switch_Topology_1175_2_Wire_98x1_Mux = "1175/2-Wire 98x1 Mux"
DAQmx_Val_Switch_Topology_1175_2_Wire_95x1_Mux = "1175/2-Wire 95x1 Mux"
DAQmx_Val_Switch_Topology_1190_Quad_4x1_Mux = "1190/Quad 4x1 Mux"
DAQmx_Val_Switch_Topology_1191_Quad_4x1_Mux = "1191/Quad 4x1 Mux"
DAQmx_Val_Switch_Topology_1192_8_SPDT = "1192/8-SPDT"
DAQmx_Val_Switch_Topology_1193_32x1_Mux = "1193/32x1 Mux"
DAQmx_Val_Switch_Topology_1193_Dual_16x1_Mux = "1193/Dual 16x1 Mux"
DAQmx_Val_Switch_Topology_1193_Quad_8x1_Mux = "1193/Quad 8x1 Mux"
DAQmx_Val_Switch_Topology_1193_16x1_Terminated_Mux = "1193/16x1 Terminated Mux"
DAQmx_Val_Switch_Topology_1193_Dual_8x1_Terminated_Mux = "1193/Dual 8x1 Terminated Mux"
DAQmx_Val_Switch_Topology_1193_Quad_4x1_Terminated_Mux = "1193/Quad 4x1 Terminated Mux"
DAQmx_Val_Switch_Topology_1193_Independent = "1193/Independent"
DAQmx_Val_Switch_Topology_1194_Quad_4x1_Mux = "1194/Quad 4x1 Mux"
DAQmx_Val_Switch_Topology_1195_Quad_4x1_Mux = "1195/Quad 4x1 Mux"
DAQmx_Val_Switch_Topology_2501_1_Wire_48x1_Mux = "2501/1-Wire 48x1 Mux"
DAQmx_Val_Switch_Topology_2501_1_Wire_48x1_Amplified_Mux = "2501/1-Wire 48x1 Amplified Mux"
DAQmx_Val_Switch_Topology_2501_2_Wire_24x1_Mux = "2501/2-Wire 24x1 Mux"
DAQmx_Val_Switch_Topology_2501_2_Wire_24x1_Amplified_Mux = "2501/2-Wire 24x1 Amplified Mux"
DAQmx_Val_Switch_Topology_2501_2_Wire_Dual_12x1_Mux = "2501/2-Wire Dual 12x1 Mux"
DAQmx_Val_Switch_Topology_2501_2_Wire_Quad_6x1_Mux = "2501/2-Wire Quad 6x1 Mux"
DAQmx_Val_Switch_Topology_2501_2_Wire_4x6_Matrix = "2501/2-Wire 4x6 Matrix"
DAQmx_Val_Switch_Topology_2501_4_Wire_12x1_Mux = "2501/4-Wire 12x1 Mux"
DAQmx_Val_Switch_Topology_2503_1_Wire_48x1_Mux = "2503/1-Wire 48x1 Mux"
DAQmx_Val_Switch_Topology_2503_2_Wire_24x1_Mux = "2503/2-Wire 24x1 Mux"
DAQmx_Val_Switch_Topology_2503_2_Wire_Dual_12x1_Mux = "2503/2-Wire Dual 12x1 Mux"
DAQmx_Val_Switch_Topology_2503_2_Wire_Quad_6x1_Mux = "2503/2-Wire Quad 6x1 Mux"
DAQmx_Val_Switch_Topology_2503_2_Wire_4x6_Matrix = "2503/2-Wire 4x6 Matrix"
DAQmx_Val_Switch_Topology_2503_4_Wire_12x1_Mux = "2503/4-Wire 12x1 Mux"
DAQmx_Val_Switch_Topology_2510_Independent = "2510/Independent"
DAQmx_Val_Switch_Topology_2512_Independent = "2512/Independent"
DAQmx_Val_Switch_Topology_2514_Independent = "2514/Independent"
DAQmx_Val_Switch_Topology_2515_Independent = "2515/Independent"
DAQmx_Val_Switch_Topology_2520_80_SPST = "2520/80-SPST"
DAQmx_Val_Switch_Topology_2521_40_DPST = "2521/40-DPST"
DAQmx_Val_Switch_Topology_2522_53_SPDT = "2522/53-SPDT"
DAQmx_Val_Switch_Topology_2523_26_DPDT = "2523/26-DPDT"
DAQmx_Val_Switch_Topology_2527_1_Wire_64x1_Mux = "2527/1-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_2527_1_Wire_Dual_32x1_Mux = "2527/1-Wire Dual 32x1 Mux"
DAQmx_Val_Switch_Topology_2527_2_Wire_32x1_Mux = "2527/2-Wire 32x1 Mux"
DAQmx_Val_Switch_Topology_2527_2_Wire_Dual_16x1_Mux = "2527/2-Wire Dual 16x1 Mux"
DAQmx_Val_Switch_Topology_2527_4_Wire_16x1_Mux = "2527/4-Wire 16x1 Mux"
DAQmx_Val_Switch_Topology_2527_Independent = "2527/Independent"
DAQmx_Val_Switch_Topology_2529_2_Wire_8x16_Matrix = "2529/2-Wire 8x16 Matrix"
DAQmx_Val_Switch_Topology_2529_2_Wire_4x32_Matrix = "2529/2-Wire 4x32 Matrix"
DAQmx_Val_Switch_Topology_2529_2_Wire_Dual_4x16_Matrix = "2529/2-Wire Dual 4x16 Matrix"
DAQmx_Val_Switch_Topology_2530_1_Wire_128x1_Mux = "2530/1-Wire 128x1 Mux"
DAQmx_Val_Switch_Topology_2530_1_Wire_Dual_64x1_Mux = "2530/1-Wire Dual 64x1 Mux"
DAQmx_Val_Switch_Topology_2530_2_Wire_64x1_Mux = "2530/2-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_2530_4_Wire_32x1_Mux = "2530/4-Wire 32x1 Mux"
DAQmx_Val_Switch_Topology_2530_1_Wire_4x32_Matrix = "2530/1-Wire 4x32 Matrix"
DAQmx_Val_Switch_Topology_2530_1_Wire_8x16_Matrix = "2530/1-Wire 8x16 Matrix"
DAQmx_Val_Switch_Topology_2530_1_Wire_Octal_16x1_Mux = "2530/1-Wire Octal 16x1 Mux"
DAQmx_Val_Switch_Topology_2530_1_Wire_Quad_32x1_Mux = "2530/1-Wire Quad 32x1 Mux"
DAQmx_Val_Switch_Topology_2530_2_Wire_4x16_Matrix = "2530/2-Wire 4x16 Matrix"
DAQmx_Val_Switch_Topology_2530_2_Wire_Dual_32x1_Mux = "2530/2-Wire Dual 32x1 Mux"
DAQmx_Val_Switch_Topology_2530_2_Wire_Quad_16x1_Mux = "2530/2-Wire Quad 16x1 Mux"
DAQmx_Val_Switch_Topology_2530_4_Wire_Dual_16x1_Mux = "2530/4-Wire Dual 16x1 Mux"
DAQmx_Val_Switch_Topology_2530_Independent = "2530/Independent"
DAQmx_Val_Switch_Topology_2531_1_Wire_4x128_Matrix = "2531/1-Wire 4x128 Matrix"
DAQmx_Val_Switch_Topology_2531_1_Wire_8x64_Matrix = "2531/1-Wire 8x64 Matrix"
DAQmx_Val_Switch_Topology_2531_1_Wire_Dual_4x64_Matrix = "2531/1-Wire Dual 4x64 Matrix"
DAQmx_Val_Switch_Topology_2531_1_Wire_Dual_8x32_Matrix = "2531/1-Wire Dual 8x32 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_16x32_Matrix = "2532/1-Wire 16x32 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_4x128_Matrix = "2532/1-Wire 4x128 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_8x64_Matrix = "2532/1-Wire 8x64 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_Dual_16x16_Matrix = "2532/1-Wire Dual 16x16 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_Dual_4x64_Matrix = "2532/1-Wire Dual 4x64 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_Dual_8x32_Matrix = "2532/1-Wire Dual 8x32 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_Quad_4x32_Matrix = "2532/1-Wire Quad 4x32 Matrix"
DAQmx_Val_Switch_Topology_2532_1_Wire_Sixteen_2x16_Matrix = "2532/1-Wire Sixteen 2x16 Matrix"
DAQmx_Val_Switch_Topology_2532_2_Wire_16x16_Matrix = "2532/2-Wire 16x16 Matrix"
DAQmx_Val_Switch_Topology_2532_2_Wire_4x64_Matrix = "2532/2-Wire 4x64 Matrix"
DAQmx_Val_Switch_Topology_2532_2_Wire_8x32_Matrix = "2532/2-Wire 8x32 Matrix"
DAQmx_Val_Switch_Topology_2532_2_Wire_Dual_4x32_Matrix = "2532/2-Wire Dual 4x32 Matrix"
DAQmx_Val_Switch_Topology_2533_1_Wire_4x64_Matrix = "2533/1-Wire 4x64 Matrix"
DAQmx_Val_Switch_Topology_2534_1_Wire_8x32_Matrix = "2534/1-Wire 8x32 Matrix"
DAQmx_Val_Switch_Topology_2535_1_Wire_4x136_Matrix = "2535/1-Wire 4x136 Matrix"
DAQmx_Val_Switch_Topology_2536_1_Wire_8x68_Matrix = "2536/1-Wire 8x68 Matrix"
DAQmx_Val_Switch_Topology_2540_1_Wire_8x9_Matrix = "2540/1-Wire 8x9 Matrix"
DAQmx_Val_Switch_Topology_2541_1_Wire_8x12_Matrix = "2541/1-Wire 8x12 Matrix"
DAQmx_Val_Switch_Topology_2542_Quad_2x1_Terminated_Mux = "2542/Quad 2x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2543_Dual_4x1_Terminated_Mux = "2543/Dual 4x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2544_8x1_Terminated_Mux = "2544/8x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2545_4x1_Terminated_Mux = "2545/4x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2546_Dual_4x1_Mux = "2546/Dual 4x1 Mux"
DAQmx_Val_Switch_Topology_2547_8x1_Mux = "2547/8x1 Mux"
DAQmx_Val_Switch_Topology_2548_4_SPDT = "2548/4-SPDT"
DAQmx_Val_Switch_Topology_2549_Terminated_2_SPDT = "2549/Terminated 2-SPDT"
DAQmx_Val_Switch_Topology_2554_4x1_Mux = "2554/4x1 Mux"
DAQmx_Val_Switch_Topology_2555_4x1_Terminated_Mux = "2555/4x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2556_Dual_4x1_Mux = "2556/Dual 4x1 Mux"
DAQmx_Val_Switch_Topology_2557_8x1_Mux = "2557/8x1 Mux"
DAQmx_Val_Switch_Topology_2558_4_SPDT = "2558/4-SPDT"
DAQmx_Val_Switch_Topology_2559_Terminated_2_SPDT = "2559/Terminated 2-SPDT"
DAQmx_Val_Switch_Topology_2564_16_SPST = "2564/16-SPST"
DAQmx_Val_Switch_Topology_2564_8_DPST = "2564/8-DPST"
DAQmx_Val_Switch_Topology_2565_16_SPST = "2565/16-SPST"
DAQmx_Val_Switch_Topology_2566_16_SPDT = "2566/16-SPDT"
DAQmx_Val_Switch_Topology_2566_8_DPDT = "2566/8-DPDT"
DAQmx_Val_Switch_Topology_2567_Independent = "2567/Independent"
DAQmx_Val_Switch_Topology_2568_31_SPST = "2568/31-SPST"
DAQmx_Val_Switch_Topology_2568_15_DPST = "2568/15-DPST"
DAQmx_Val_Switch_Topology_2569_100_SPST = "2569/100-SPST"
DAQmx_Val_Switch_Topology_2569_50_DPST = "2569/50-DPST"
DAQmx_Val_Switch_Topology_2570_40_SPDT = "2570/40-SPDT"
DAQmx_Val_Switch_Topology_2570_20_DPDT = "2570/20-DPDT"
DAQmx_Val_Switch_Topology_2571_66_SPDT = "2571/66-SPDT"
DAQmx_Val_Switch_Topology_2575_1_Wire_196x1_Mux = "2575/1-Wire 196x1 Mux"
DAQmx_Val_Switch_Topology_2575_2_Wire_98x1_Mux = "2575/2-Wire 98x1 Mux"
DAQmx_Val_Switch_Topology_2575_2_Wire_95x1_Mux = "2575/2-Wire 95x1 Mux"
DAQmx_Val_Switch_Topology_2576_2_Wire_64x1_Mux = "2576/2-Wire 64x1 Mux"
DAQmx_Val_Switch_Topology_2576_2_Wire_Dual_32x1_Mux = "2576/2-Wire Dual 32x1 Mux"
DAQmx_Val_Switch_Topology_2576_2_Wire_Octal_8x1_Mux = "2576/2-Wire Octal 8x1 Mux"
DAQmx_Val_Switch_Topology_2576_2_Wire_Quad_16x1_Mux = "2576/2-Wire Quad 16x1 Mux"
DAQmx_Val_Switch_Topology_2576_2_Wire_Sixteen_4x1_Mux = "2576/2-Wire Sixteen 4x1 Mux"
DAQmx_Val_Switch_Topology_2576_Independent = "2576/Independent"
DAQmx_Val_Switch_Topology_2584_1_Wire_12x1_Mux = "2584/1-Wire 12x1 Mux"
DAQmx_Val_Switch_Topology_2584_1_Wire_Dual_6x1_Mux = "2584/1-Wire Dual 6x1 Mux"
DAQmx_Val_Switch_Topology_2584_2_Wire_6x1_Mux = "2584/2-Wire 6x1 Mux"
DAQmx_Val_Switch_Topology_2584_Independent = "2584/Independent"
DAQmx_Val_Switch_Topology_2585_1_Wire_10x1_Mux = "2585/1-Wire 10x1 Mux"
DAQmx_Val_Switch_Topology_2586_10_SPST = "2586/10-SPST"
DAQmx_Val_Switch_Topology_2586_5_DPST = "2586/5-DPST"
DAQmx_Val_Switch_Topology_2590_4x1_Mux = "2590/4x1 Mux"
DAQmx_Val_Switch_Topology_2591_4x1_Mux = "2591/4x1 Mux"
DAQmx_Val_Switch_Topology_2593_16x1_Mux = "2593/16x1 Mux"
DAQmx_Val_Switch_Topology_2593_Dual_8x1_Mux = "2593/Dual 8x1 Mux"
DAQmx_Val_Switch_Topology_2593_8x1_Terminated_Mux = "2593/8x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2593_Dual_4x1_Terminated_Mux = "2593/Dual 4x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2593_Independent = "2593/Independent"
DAQmx_Val_Switch_Topology_2594_4x1_Mux = "2594/4x1 Mux"
DAQmx_Val_Switch_Topology_2595_4x1_Mux = "2595/4x1 Mux"
DAQmx_Val_Switch_Topology_2596_Dual_6x1_Mux = "2596/Dual 6x1 Mux"
DAQmx_Val_Switch_Topology_2597_6x1_Terminated_Mux = "2597/6x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2598_Dual_Transfer = "2598/Dual Transfer"
DAQmx_Val_Switch_Topology_2599_2_SPDT = "2599/2-SPDT"
DAQmx_Val_Switch_Topology_2720_Independent = "2720/Independent"
DAQmx_Val_Switch_Topology_2722_Independent = "2722/Independent"
DAQmx_Val_Switch_Topology_2725_Independent = "2725/Independent"
DAQmx_Val_Switch_Topology_2727_Independent = "2727/Independent"
DAQmx_Val_Switch_Topology_2790_Independent = "2790/Independent"
DAQmx_Val_Switch_Topology_2796_Dual_6x1_Mux = "2796/Dual 6x1 Mux"
DAQmx_Val_Switch_Topology_2797_6x1_Terminated_Mux = "2797/6x1 Terminated Mux"
DAQmx_Val_Switch_Topology_2798_Dual_Transfer = "2798/Dual Transfer"
DAQmx_Val_Switch_Topology_2799_2_SPDT = "2799/2-SPDT"
DAQmxSuccess = (0)
DAQmxErrorCOCannotKeepUpInHWTimedSinglePoint = (-209805)
DAQmxErrorWaitForNextSampClkDetected3OrMoreSampClks = (-209803)
DAQmxErrorWaitForNextSampClkDetectedMissedSampClk = (-209802)
DAQmxErrorWriteNotCompleteBeforeSampClk = (-209801)
DAQmxErrorReadNotCompleteBeforeSampClk = (-209800)
DAQmxErrorSetupCalibrationNotCalled = (-201486)
DAQmxErrorCalMeasuredDataSizeVsActualDataSizeMismatch = (-201485)
DAQmxErrorCDAQMissingDSAMasterForChanExpansion = (-201484)
DAQmxErrorCDAQMasterNotFoundForChanExpansion = (-201483)
DAQmxErrorAllChansShouldBeProvidedForCalibration = (-201482)
DAQmxErrorMustSpecifyExpirationStateForAllLinesInRange = (-201481)
DAQmxErrorOpenSessionExists = (-201480)
DAQmxErrorCannotQueryTerminalForSWArmStart = (-201479)
DAQmxErrorChassisWatchdogTimerExpired = (-201478)
DAQmxErrorCantReserveWatchdogTaskWhileOtherTasksReserved = (-201477)
DAQmxErrorCantReserveTaskWhileWatchdogTaskReserving = (-201476)
DAQmxErrorAuxPowerSourceRequired = (-201475)
DAQmxErrorDeviceNotSupportedOnLocalSystem = (-201474)
DAQmxErrorOneTimestampChannelRequiredForCombinedNavigationRead = (-201472)
DAQmxErrorMultDevsMultPhysChans = (-201471)
DAQmxErrorInvalidCalAdjustmentPointValues = (-201470)
DAQmxErrorDifferentDigitizerFromCommunicator = (-201469)
DAQmxErrorCDAQSyncMasterClockNotPresent = (-201468)
DAQmxErrorAssociatedChansHaveConflictingProps = (-201467)
DAQmxErrorAutoConfigBetweenMultipleDeviceStatesInvalid = (-201466)
DAQmxErrorAutoConfigOfOfflineDevicesInvalid = (-201465)
DAQmxErrorExternalFIFOFault = (-201464)
DAQmxErrorConnectionsNotReciprocal = (-201463)
DAQmxErrorInvalidOutputToInputCDAQSyncConnection = (-201462)
DAQmxErrorReferenceClockNotPresent = (-201461)
DAQmxErrorBlankStringExpansionFoundNoSupportedCDAQSyncConnectionDevices = (-201460)
DAQmxErrorNoDevicesSupportCDAQSyncConnections = (-201459)
DAQmxErrorInvalidCDAQSyncTimeoutValue = (-201458)
DAQmxErrorCDAQSyncConnectionToSamePort = (-201457)
DAQmxErrorDevsWithoutCommonSyncConnectionStrategy = (-201456)
DAQmxErrorNoCDAQSyncBetweenPhysAndSimulatedDevs = (-201455)
DAQmxErrorUnableToContainCards = (-201454)
DAQmxErrorFindDisconnectedBetweenPhysAndSimDeviceStatesInvalid = (-201453)
DAQmxErrorOperationAborted = (-201452)
DAQmxErrorTwoPortsRequired = (-201451)
DAQmxErrorDeviceDoesNotSupportCDAQSyncConnections = (-201450)
DAQmxErrorInvalidcDAQSyncPortConnectionFormat = (-201449)
DAQmxErrorRosetteMeasurementsNotSpecified = (-201448)
DAQmxErrorInvalidNumOfPhysChansForDeltaRosette = (-201447)
DAQmxErrorInvalidNumOfPhysChansForTeeRosette = (-201446)
DAQmxErrorRosetteStrainChanNamesNeeded = (-201445)
DAQmxErrorMultideviceWithOnDemandTiming = (-201444)
DAQmxErrorFREQOUTCannotProduceDesiredFrequency3 = (-201443)
DAQmxErrorTwoEdgeSeparationSameTerminalSameEdge = (-201442)
DAQmxErrorDontMixSyncPulseAndSampClkTimebaseOn449x = (-201441)
DAQmxErrorNeitherRefClkNorSampClkTimebaseConfiguredForDSASync = (-201440)
DAQmxErrorRetriggeringFiniteCONotAllowed = (-201439)
DAQmxErrorDeviceRebootedFromWDTTimeout = (-201438)
DAQmxErrorTimeoutValueExceedsMaximum = (-201437)
DAQmxErrorSharingDifferentWireModes = (-201436)
DAQmxErrorCantPrimeWithEmptyBuffer = (-201435)
DAQmxErrorConfigFailedBecauseWatchdogExpired = (-201434)
DAQmxErrorWriteFailedBecauseWatchdogChangedLineDirection = (-201433)
DAQmxErrorMultipleSubsytemCalibration = (-201432)
DAQmxErrorIncorrectChannelForOffsetAdjustment = (-201431)
DAQmxErrorInvalidNumRefVoltagesToWrite = (-201430)
DAQmxErrorStartTrigDelayWithDSAModule = (-201429)
DAQmxErrorMoreThanOneSyncPulseDetected = (-201428)
DAQmxErrorDevNotSupportedWithinDAQmxAPI = (-201427)
DAQmxErrorDevsWithoutSyncStrategies = (-201426)
DAQmxErrorDevsWithoutCommonSyncStrategy = (-201425)
DAQmxErrorSyncStrategiesCannotSync = (-201424)
DAQmxErrorChassisCommunicationInterrupted = (-201423)
DAQmxErrorUnknownCardPowerProfileInCarrier = (-201422)
DAQmxErrorAttrNotSupportedOnAccessory = (-201421)
DAQmxErrorNetworkDeviceReservedByAnotherHost = (-201420)
DAQmxErrorIncorrectFirmwareFileUploaded = (-201419)
DAQmxErrorInvalidFirmwareFileUploaded = (-201418)
DAQmxErrorInTimerTimeoutOnArm = (-201417)
DAQmxErrorCantExceedSlotRelayDriveLimit = (-201416)
DAQmxErrorModuleUnsupportedFor9163 = (-201415)
DAQmxErrorConnectionsNotSupported = (-201414)
DAQmxErrorAccessoryNotPresent = (-201413)
DAQmxErrorSpecifiedAccessoryChannelsNotPresentOnDevice = (-201412)
DAQmxErrorConnectionsNotSupportedOnAccessory = (-201411)
DAQmxErrorRateTooFastForHWTSP = (-201410)
DAQmxErrorDelayFromSampleClockOutOfRangeForHWTSP = (-201409)
DAQmxErrorAveragingWhenNotInternalHWTSP = (-201408)
DAQmxErrorAttributeNotSupportedUnlessHWTSP = (-201407)
DAQmxErrorFiveVoltDetectFailed = (-201406)
DAQmxErrorAnalogBusStateInconsistent = (-201405)
DAQmxErrorCardDetectedDoesNotMatchExpectedCard = (-201404)
DAQmxErrorLoggingStartNewFileNotCalled = (-201403)
DAQmxErrorLoggingSampsPerFileNotDivisible = (-201402)
DAQmxErrorRetrievingNetworkDeviceProperties = (-201401)
DAQmxErrorFilePreallocationFailed = (-201400)
DAQmxErrorModuleMismatchInSameTimedTask = (-201399)
DAQmxErrorInvalidAttributeValuePossiblyDueToOtherAttributeValues = (-201398)
DAQmxErrorChangeDetectionStoppedToPreventDeviceHang = (-201397)
DAQmxErrorFilterDelayRemovalNotPosssibleWithAnalogTrigger = (-201396)
DAQmxErrorNonbufferedOrNoChannels = (-201395)
DAQmxErrorTristateLogicLevelNotSpecdForEntirePort = (-201394)
DAQmxErrorTristateLogicLevelNotSupportedOnDigOutChan = (-201393)
DAQmxErrorTristateLogicLevelNotSupported = (-201392)
DAQmxErrorIncompleteGainAndCouplingCalAdjustment = (-201391)
DAQmxErrorNetworkStatusConnectionLost = (-201390)
DAQmxErrorModuleChangeDuringConnectionLoss = (-201389)
DAQmxErrorNetworkDeviceNotReservedByHost = (-201388)
DAQmxErrorDuplicateCalibrationAdjustmentInput = (-201387)
DAQmxErrorSelfCalFailedContactTechSupport = (-201386)
DAQmxErrorSelfCalFailedToConverge = (-201385)
DAQmxErrorUnsupportedSimulatedModuleForSimulatedChassis = (-201384)
DAQmxErrorLoggingWriteSizeTooBig = (-201383)
DAQmxErrorLoggingWriteSizeNotDivisible = (-201382)
DAQmxErrorMyDAQPowerRailFault = (-201381)
DAQmxErrorDeviceDoesNotSupportThisOperation = (-201380)
DAQmxErrorNetworkDevicesNotSupportedOnThisPlatform = (-201379)
DAQmxErrorUnknownFirmwareVersion = (-201378)
DAQmxErrorFirmwareIsUpdating = (-201377)
DAQmxErrorAccessoryEEPROMIsCorrupt = (-201376)
DAQmxErrorThrmcplLeadOffsetNullingCalNotSupported = (-201375)
DAQmxErrorSelfCalFailedTryExtCal = (-201374)
DAQmxErrorOutputP2PNotSupportedWithMultithreadedScripts = (-201373)
DAQmxErrorThrmcplCalibrationChannelsOpen = (-201372)
DAQmxErrorMDNSServiceInstanceAlreadyInUse = (-201371)
DAQmxErrorIPAddressAlreadyInUse = (-201370)
DAQmxErrorHostnameAlreadyInUse = (-201369)
DAQmxErrorInvalidNumberOfCalAdjustmentPoints = (-201368)
DAQmxErrorFilterOrDigitalSyncInternalSignal = (-201367)
DAQmxErrorBadDDSSource = (-201366)
DAQmxErrorOnboardRegenWithMoreThan16Channels = (-201365)
DAQmxErrorTriggerTooFast = (-201364)
DAQmxErrorMinMaxOutsideTableRange = (-201363)
DAQmxErrorChannelExpansionWithInvalidAnalogTriggerDevice = (-201362)
DAQmxErrorSyncPulseSrcInvalidForTask = (-201361)
DAQmxErrorInvalidCarrierSlotNumberSpecd = (-201360)
DAQmxErrorCardsMustBeInSameCarrier = (-201359)
DAQmxErrorCardDevCarrierSimMustMatch = (-201358)
DAQmxErrorDevMustHaveAtLeastOneCard = (-201357)
DAQmxErrorCardTopologyError = (-201356)
DAQmxErrorExceededCarrierPowerLimit = (-201355)
DAQmxErrorCardsIncompatible = (-201354)
DAQmxErrorAnalogBusNotValid = (-201353)
DAQmxErrorReservationConflict = (-201352)
DAQmxErrorMemMappedOnDemandNotSupported = (-201351)
DAQmxErrorChannelExpansionWithDifferentTriggerDevices = (-201349)
DAQmxErrorCounterSyncAndRetriggered = (-201348)
DAQmxErrorNoExternalSyncPulseDetected = (-201347)
DAQmxErrorSlaveAndNoExternalSyncPulse = (-201346)
DAQmxErrorCustomTimingRequiredForAttribute = (-201345)
DAQmxErrorCustomTimingModeNotSet = (-201344)
DAQmxErrorAccessoryPowerTripped = (-201343)
DAQmxErrorUnsupportedAccessory = (-201342)
DAQmxErrorInvalidAccessoryChange = (-201341)
DAQmxErrorFirmwareRequiresUpgrade = (-201340)
DAQmxErrorFastExternalTimebaseNotSupportedForDevice = (-201339)
DAQmxErrorInvalidShuntLocationForCalibration = (-201338)
DAQmxErrorDeviceNameTooLong = (-201337)
DAQmxErrorBridgeScalesUnsupported = (-201336)
DAQmxErrorMismatchedElecPhysValues = (-201335)
DAQmxErrorLinearRequiresUniquePoints = (-201334)
DAQmxErrorMissingRequiredScalingParameter = (-201333)
DAQmxErrorLoggingNotSupportOnOutputTasks = (-201332)
DAQmxErrorMemoryMappedHardwareTimedNonBufferedUnsupported = (-201331)
DAQmxErrorCannotUpdatePulseTrainWithAutoIncrementEnabled = (-201330)
DAQmxErrorHWTimedSinglePointAndDataXferNotDMA = (-201329)
DAQmxErrorSCCSecondStageEmpty = (-201328)
DAQmxErrorSCCInvalidDualStageCombo = (-201327)
DAQmxErrorSCCInvalidSecondStage = (-201326)
DAQmxErrorSCCInvalidFirstStage = (-201325)
DAQmxErrorCounterMultipleSampleClockedChannels = (-201324)
DAQmxError2CounterMeasurementModeAndSampleClocked = (-201323)
DAQmxErrorCantHaveBothMemMappedAndNonMemMappedTasks = (-201322)
DAQmxErrorMemMappedDataReadByAnotherProcess = (-201321)
DAQmxErrorRetriggeringInvalidForGivenSettings = (-201320)
DAQmxErrorAIOverrun = (-201319)
DAQmxErrorCOOverrun = (-201318)
DAQmxErrorCounterMultipleBufferedChannels = (-201317)
DAQmxErrorInvalidTimebaseForCOHWTSP = (-201316)
DAQmxErrorWriteBeforeEvent = (-201315)
DAQmxErrorCIOverrun = (-201314)
DAQmxErrorCounterNonResponsiveAndReset = (-201313)
DAQmxErrorMeasTypeOrChannelNotSupportedForLogging = (-201312)
DAQmxErrorFileAlreadyOpenedForWrite = (-201311)
DAQmxErrorTdmsNotFound = (-201310)
DAQmxErrorGenericFileIO = (-201309)
DAQmxErrorFiniteSTCCounterNotSupportedForLogging = (-201308)
DAQmxErrorMeasurementTypeNotSupportedForLogging = (-201307)
DAQmxErrorFileAlreadyOpened = (-201306)
DAQmxErrorDiskFull = (-201305)
DAQmxErrorFilePathInvalid = (-201304)
DAQmxErrorFileVersionMismatch = (-201303)
DAQmxErrorFileWriteProtected = (-201302)
DAQmxErrorReadNotSupportedForLoggingMode = (-201301)
DAQmxErrorAttributeNotSupportedWhenLogging = (-201300)
DAQmxErrorLoggingModeNotSupportedNonBuffered = (-201299)
DAQmxErrorPropertyNotSupportedWithConflictingProperty = (-201298)
DAQmxErrorParallelSSHOnConnector1 = (-201297)
DAQmxErrorCOOnlyImplicitSampleTimingTypeSupported = (-201296)
DAQmxErrorCalibrationFailedAOOutOfRange = (-201295)
DAQmxErrorCalibrationFailedAIOutOfRange = (-201294)
DAQmxErrorCalPWMLinearityFailed = (-201293)
DAQmxErrorOverrunUnderflowConfigurationCombo = (-201292)
DAQmxErrorCannotWriteToFiniteCOTask = (-201291)
DAQmxErrorNetworkDAQInvalidWEPKeyLength = (-201290)
DAQmxErrorCalInputsShortedNotSupported = (-201289)
DAQmxErrorCannotSetPropertyWhenTaskIsReserved = (-201288)
DAQmxErrorMinus12VFuseBlown = (-201287)
DAQmxErrorPlus12VFuseBlown = (-201286)
DAQmxErrorPlus5VFuseBlown = (-201285)
DAQmxErrorPlus3VFuseBlown = (-201284)
DAQmxErrorDeviceSerialPortError = (-201283)
DAQmxErrorPowerUpStateMachineNotDone = (-201282)
DAQmxErrorTooManyTriggersSpecifiedInTask = (-201281)
DAQmxErrorVerticalOffsetNotSupportedOnDevice = (-201280)
DAQmxErrorInvalidCouplingForMeasurementType = (-201279)
DAQmxErrorDigitalLineUpdateTooFastForDevice = (-201278)
DAQmxErrorCertificateIsTooBigToTransfer = (-201277)
DAQmxErrorOnlyPEMOrDERCertiticatesAccepted = (-201276)
DAQmxErrorCalCouplingNotSupported = (-201275)
DAQmxErrorDeviceNotSupportedIn64Bit = (-201274)
DAQmxErrorNetworkDeviceInUse = (-201273)
DAQmxErrorInvalidIPv4AddressFormat = (-201272)
DAQmxErrorNetworkProductTypeMismatch = (-201271)
DAQmxErrorOnlyPEMCertificatesAccepted = (-201270)
DAQmxErrorCalibrationRequiresPrototypingBoardEnabled = (-201269)
DAQmxErrorAllCurrentLimitingResourcesAlreadyTaken = (-201268)
DAQmxErrorUserDefInfoStringBadLength = (-201267)
DAQmxErrorPropertyNotFound = (-201266)
DAQmxErrorOverVoltageProtectionActivated = (-201265)
DAQmxErrorScaledIQWaveformTooLarge = (-201264)
DAQmxErrorFirmwareFailedToDownload = (-201263)
DAQmxErrorPropertyNotSupportedForBusType = (-201262)
DAQmxErrorChangeRateWhileRunningCouldNotBeCompleted = (-201261)
DAQmxErrorCannotQueryManualControlAttribute = (-201260)
DAQmxErrorInvalidNetworkConfiguration = (-201259)
DAQmxErrorInvalidWirelessConfiguration = (-201258)
DAQmxErrorInvalidWirelessCountryCode = (-201257)
DAQmxErrorInvalidWirelessChannel = (-201256)
DAQmxErrorNetworkEEPROMHasChanged = (-201255)
DAQmxErrorNetworkSerialNumberMismatch = (-201254)
DAQmxErrorNetworkStatusDown = (-201253)
DAQmxErrorNetworkTargetUnreachable = (-201252)
DAQmxErrorNetworkTargetNotFound = (-201251)
DAQmxErrorNetworkStatusTimedOut = (-201250)
DAQmxErrorInvalidWirelessSecuritySelection = (-201249)
DAQmxErrorNetworkDeviceConfigurationLocked = (-201248)
DAQmxErrorNetworkDAQDeviceNotSupported = (-201247)
DAQmxErrorNetworkDAQCannotCreateEmptySleeve = (-201246)
DAQmxErrorModuleTypeDoesNotMatchModuleTypeInDestination = (-201244)
DAQmxErrorInvalidTEDSInterfaceAddress = (-201243)
DAQmxErrorDevDoesNotSupportSCXIComm = (-201242)
DAQmxErrorSCXICommDevConnector0MustBeCabledToModule = (-201241)
DAQmxErrorSCXIModuleDoesNotSupportDigitizationMode = (-201240)
DAQmxErrorDevDoesNotSupportMultiplexedSCXIDigitizationMode = (-201239)
DAQmxErrorDevOrDevPhysChanDoesNotSupportSCXIDigitization = (-201238)
DAQmxErrorInvalidPhysChanName = (-201237)
DAQmxErrorSCXIChassisCommModeInvalid = (-201236)
DAQmxErrorRequiredDependencyNotFound = (-201235)
DAQmxErrorInvalidStorage = (-201234)
DAQmxErrorInvalidObject = (-201233)
DAQmxErrorStorageAlteredPriorToSave = (-201232)
DAQmxErrorTaskDoesNotReferenceLocalChannel = (-201231)
DAQmxErrorReferencedDevSimMustMatchTarget = (-201230)
DAQmxErrorProgrammedIOFailsBecauseOfWatchdogTimer = (-201229)
DAQmxErrorWatchdogTimerFailsBecauseOfProgrammedIO = (-201228)
DAQmxErrorCantUseThisTimingEngineWithAPort = (-201227)
DAQmxErrorProgrammedIOConflict = (-201226)
DAQmxErrorChangeDetectionIncompatibleWithProgrammedIO = (-201225)
DAQmxErrorTristateNotEnoughLines = (-201224)
DAQmxErrorTristateConflict = (-201223)
DAQmxErrorGenerateOrFiniteWaitExpectedBeforeBreakBlock = (-201222)
DAQmxErrorBreakBlockNotAllowedInLoop = (-201221)
DAQmxErrorClearTriggerNotAllowedInBreakBlock = (-201220)
DAQmxErrorNestingNotAllowedInBreakBlock = (-201219)
DAQmxErrorIfElseBlockNotAllowedInBreakBlock = (-201218)
DAQmxErrorRepeatUntilTriggerLoopNotAllowedInBreakBlock = (-201217)
DAQmxErrorWaitUntilTriggerNotAllowedInBreakBlock = (-201216)
DAQmxErrorMarkerPosInvalidInBreakBlock = (-201215)
DAQmxErrorInvalidWaitDurationInBreakBlock = (-201214)
DAQmxErrorInvalidSubsetLengthInBreakBlock = (-201213)
DAQmxErrorInvalidWaveformLengthInBreakBlock = (-201212)
DAQmxErrorInvalidWaitDurationBeforeBreakBlock = (-201211)
DAQmxErrorInvalidSubsetLengthBeforeBreakBlock = (-201210)
DAQmxErrorInvalidWaveformLengthBeforeBreakBlock = (-201209)
DAQmxErrorSampleRateTooHighForADCTimingMode = (-201208)
DAQmxErrorActiveDevNotSupportedWithMultiDevTask = (-201207)
DAQmxErrorRealDevAndSimDevNotSupportedInSameTask = (-201206)
DAQmxErrorRTSISimMustMatchDevSim = (-201205)
DAQmxErrorBridgeShuntCaNotSupported = (-201204)
DAQmxErrorStrainShuntCaNotSupported = (-201203)
DAQmxErrorGainTooLargeForGainCalConst = (-201202)
DAQmxErrorOffsetTooLargeForOffsetCalConst = (-201201)
DAQmxErrorElvisPrototypingBoardRemoved = (-201200)
DAQmxErrorElvis2PowerRailFault = (-201199)
DAQmxErrorElvis2PhysicalChansFault = (-201198)
DAQmxErrorElvis2PhysicalChansThermalEvent = (-201197)
DAQmxErrorRXBitErrorRateLimitExceeded = (-201196)
DAQmxErrorPHYBitErrorRateLimitExceeded = (-201195)
DAQmxErrorTwoPartAttributeCalledOutOfOrder = (-201194)
DAQmxErrorInvalidSCXIChassisAddress = (-201193)
DAQmxErrorCouldNotConnectToRemoteMXS = (-201192)
DAQmxErrorExcitationStateRequiredForAttributes = (-201191)
DAQmxErrorDeviceNotUsableUntilUSBReplug = (-201190)
DAQmxErrorInputFIFOOverflowDuringCalibrationOnFullSpeedUSB = (-201189)
DAQmxErrorInputFIFOOverflowDuringCalibration = (-201188)
DAQmxErrorCJCChanConflictsWithNonThermocoupleChan = (-201187)
DAQmxErrorCommDeviceForPXIBackplaneNotInRightmostSlot = (-201186)
DAQmxErrorCommDeviceForPXIBackplaneNotInSameChassis = (-201185)
DAQmxErrorCommDeviceForPXIBackplaneNotPXI = (-201184)
DAQmxErrorInvalidCalExcitFrequency = (-201183)
DAQmxErrorInvalidCalExcitVoltage = (-201182)
DAQmxErrorInvalidAIInputSrc = (-201181)
DAQmxErrorInvalidCalInputRef = (-201180)
DAQmxErrordBReferenceValueNotGreaterThanZero = (-201179)
DAQmxErrorSampleClockRateIsTooFastForSampleClockTiming = (-201178)
DAQmxErrorDeviceNotUsableUntilColdStart = (-201177)
DAQmxErrorSampleClockRateIsTooFastForBurstTiming = (-201176)
DAQmxErrorDevImportFailedAssociatedResourceIDsNotSupported = (-201175)
DAQmxErrorSCXI1600ImportNotSupported = (-201174)
DAQmxErrorPowerSupplyConfigurationFailed = (-201173)
DAQmxErrorIEPEWithDCNotAllowed = (-201172)
DAQmxErrorMinTempForThermocoupleTypeOutsideAccuracyForPolyScaling = (-201171)
DAQmxErrorDevImportFailedNoDeviceToOverwriteAndSimulationNotSupported = (-201170)
DAQmxErrorDevImportFailedDeviceNotSupportedOnDestination = (-201169)
DAQmxErrorFirmwareIsTooOld = (-201168)
DAQmxErrorFirmwareCouldntUpdate = (-201167)
DAQmxErrorFirmwareIsCorrupt = (-201166)
DAQmxErrorFirmwareTooNew = (-201165)
DAQmxErrorSampClockCannotBeExportedFromExternalSampClockSrc = (-201164)
DAQmxErrorPhysChanReservedForInputWhenDesiredForOutput = (-201163)
DAQmxErrorPhysChanReservedForOutputWhenDesiredForInput = (-201162)
DAQmxErrorSpecifiedCDAQSlotNotEmpty = (-201161)
DAQmxErrorDeviceDoesNotSupportSimulation = (-201160)
DAQmxErrorInvalidCDAQSlotNumberSpecd = (-201159)
DAQmxErrorCSeriesModSimMustMatchCDAQChassisSim = (-201158)
DAQmxErrorSCCCabledDevMustNotBeSimWhenSCCCarrierIsNotSim = (-201157)
DAQmxErrorSCCModSimMustMatchSCCCarrierSim = (-201156)
DAQmxErrorSCXIModuleDoesNotSupportSimulation = (-201155)
DAQmxErrorSCXICableDevMustNotBeSimWhenModIsNotSim = (-201154)
DAQmxErrorSCXIDigitizerSimMustNotBeSimWhenModIsNotSim = (-201153)
DAQmxErrorSCXIModSimMustMatchSCXIChassisSim = (-201152)
DAQmxErrorSimPXIDevReqSlotAndChassisSpecd = (-201151)
DAQmxErrorSimDevConflictWithRealDev = (-201150)
DAQmxErrorInsufficientDataForCalibration = (-201149)
DAQmxErrorTriggerChannelMustBeEnabled = (-201148)
DAQmxErrorCalibrationDataConflictCouldNotBeResolved = (-201147)
DAQmxErrorSoftwareTooNewForSelfCalibrationData = (-201146)
DAQmxErrorSoftwareTooNewForExtCalibrationData = (-201145)
DAQmxErrorSelfCalibrationDataTooNewForSoftware = (-201144)
DAQmxErrorExtCalibrationDataTooNewForSoftware = (-201143)
DAQmxErrorSoftwareTooNewForEEPROM = (-201142)
DAQmxErrorEEPROMTooNewForSoftware = (-201141)
DAQmxErrorSoftwareTooNewForHardware = (-201140)
DAQmxErrorHardwareTooNewForSoftware = (-201139)
DAQmxErrorTaskCannotRestartFirstSampNotAvailToGenerate = (-201138)
DAQmxErrorOnlyUseStartTrigSrcPrptyWithDevDataLines = (-201137)
DAQmxErrorOnlyUsePauseTrigSrcPrptyWithDevDataLines = (-201136)
DAQmxErrorOnlyUseRefTrigSrcPrptyWithDevDataLines = (-201135)
DAQmxErrorPauseTrigDigPatternSizeDoesNotMatchSrcSize = (-201134)
DAQmxErrorLineConflictCDAQ = (-201133)
DAQmxErrorCannotWriteBeyondFinalFiniteSample = (-201132)
DAQmxErrorRefAndStartTriggerSrcCantBeSame = (-201131)
DAQmxErrorMemMappingIncompatibleWithPhysChansInTask = (-201130)
DAQmxErrorOutputDriveTypeMemMappingConflict = (-201129)
DAQmxErrorCAPIDeviceIndexInvalid = (-201128)
DAQmxErrorRatiometricDevicesMustUseExcitationForScaling = (-201127)
DAQmxErrorPropertyRequiresPerDeviceCfg = (-201126)
DAQmxErrorAICouplingAndAIInputSourceConflict = (-201125)
DAQmxErrorOnlyOneTaskCanPerformDOMemoryMappingAtATime = (-201124)
DAQmxErrorTooManyChansForAnalogRefTrigCDAQ = (-201123)
DAQmxErrorSpecdPropertyValueIsIncompatibleWithSampleTimingType = (-201122)
DAQmxErrorCPUNotSupportedRequireSSE = (-201121)
DAQmxErrorSpecdPropertyValueIsIncompatibleWithSampleTimingResponseMode = (-201120)
DAQmxErrorConflictingNextWriteIsLastAndRegenModeProperties = (-201119)
DAQmxErrorMStudioOperationDoesNotSupportDeviceContext = (-201118)
DAQmxErrorPropertyValueInChannelExpansionContextInvalid = (-201117)
DAQmxErrorHWTimedNonBufferedAONotSupported = (-201116)
DAQmxErrorWaveformLengthNotMultOfQuantum = (-201115)
DAQmxErrorDSAExpansionMixedBoardsWrongOrderInPXIChassis = (-201114)
DAQmxErrorPowerLevelTooLowForOOK = (-201113)
DAQmxErrorDeviceComponentTestFailure = (-201112)
DAQmxErrorUserDefinedWfmWithOOKUnsupported = (-201111)
DAQmxErrorInvalidDigitalModulationUserDefinedWaveform = (-201110)
DAQmxErrorBothRefInAndRefOutEnabled = (-201109)
DAQmxErrorBothAnalogAndDigitalModulationEnabled = (-201108)
DAQmxErrorBufferedOpsNotSupportedInSpecdSlotForCDAQ = (-201107)
DAQmxErrorPhysChanNotSupportedInSpecdSlotForCDAQ = (-201106)
DAQmxErrorResourceReservedWithConflictingSettings = (-201105)
DAQmxErrorInconsistentAnalogTrigSettingsCDAQ = (-201104)
DAQmxErrorTooManyChansForAnalogPauseTrigCDAQ = (-201103)
DAQmxErrorAnalogTrigNotFirstInScanListCDAQ = (-201102)
DAQmxErrorTooManyChansGivenTimingType = (-201101)
DAQmxErrorSampClkTimebaseDivWithExtSampClk = (-201100)
DAQmxErrorCantSaveTaskWithPerDeviceTimingProperties = (-201099)
DAQmxErrorConflictingAutoZeroMode = (-201098)
DAQmxErrorSampClkRateNotSupportedWithEAREnabled = (-201097)
DAQmxErrorSampClkTimebaseRateNotSpecd = (-201096)
DAQmxErrorSessionCorruptedByDLLReload = (-201095)
DAQmxErrorActiveDevNotSupportedWithChanExpansion = (-201094)
DAQmxErrorSampClkRateInvalid = (-201093)
DAQmxErrorExtSyncPulseSrcCannotBeExported = (-201092)
DAQmxErrorSyncPulseMinDelayToStartNeededForExtSyncPulseSrc = (-201091)
DAQmxErrorSyncPulseSrcInvalid = (-201090)
DAQmxErrorSampClkTimebaseRateInvalid = (-201089)
DAQmxErrorSampClkTimebaseSrcInvalid = (-201088)
DAQmxErrorSampClkRateMustBeSpecd = (-201087)
DAQmxErrorInvalidAttributeName = (-201086)
DAQmxErrorCJCChanNameMustBeSetWhenCJCSrcIsScannableChan = (-201085)
DAQmxErrorHiddenChanMissingInChansPropertyInCfgFile = (-201084)
DAQmxErrorChanNamesNotSpecdInCfgFile = (-201083)
DAQmxErrorDuplicateHiddenChanNamesInCfgFile = (-201082)
DAQmxErrorDuplicateChanNameInCfgFile = (-201081)
DAQmxErrorInvalidSCCModuleForSlotSpecd = (-201080)
DAQmxErrorInvalidSCCSlotNumberSpecd = (-201079)
DAQmxErrorInvalidSectionIdentifier = (-201078)
DAQmxErrorInvalidSectionName = (-201077)
DAQmxErrorDAQmxVersionNotSupported = (-201076)
DAQmxErrorSWObjectsFoundInFile = (-201075)
DAQmxErrorHWObjectsFoundInFile = (-201074)
DAQmxErrorLocalChannelSpecdWithNoParentTask = (-201073)
DAQmxErrorTaskReferencesMissingLocalChannel = (-201072)
DAQmxErrorTaskReferencesLocalChannelFromOtherTask = (-201071)
DAQmxErrorTaskMissingChannelProperty = (-201070)
DAQmxErrorInvalidLocalChanName = (-201069)
DAQmxErrorInvalidEscapeCharacterInString = (-201068)
DAQmxErrorInvalidTableIdentifier = (-201067)
DAQmxErrorValueFoundInInvalidColumn = (-201066)
DAQmxErrorMissingStartOfTable = (-201065)
DAQmxErrorFileMissingRequiredDAQmxHeader = (-201064)
DAQmxErrorDeviceIDDoesNotMatch = (-201063)
DAQmxErrorBufferedOperationsNotSupportedOnSelectedLines = (-201062)
DAQmxErrorPropertyConflictsWithScale = (-201061)
DAQmxErrorInvalidINIFileSyntax = (-201060)
DAQmxErrorDeviceInfoFailedPXIChassisNotIdentified = (-201059)
DAQmxErrorInvalidHWProductNumber = (-201058)
DAQmxErrorInvalidHWProductType = (-201057)
DAQmxErrorInvalidNumericFormatSpecd = (-201056)
DAQmxErrorDuplicatePropertyInObject = (-201055)
DAQmxErrorInvalidEnumValueSpecd = (-201054)
DAQmxErrorTEDSSensorPhysicalChannelConflict = (-201053)
DAQmxErrorTooManyPhysicalChansForTEDSInterfaceSpecd = (-201052)
DAQmxErrorIncapableTEDSInterfaceControllingDeviceSpecd = (-201051)
DAQmxErrorSCCCarrierSpecdIsMissing = (-201050)
DAQmxErrorIncapableSCCDigitizingDeviceSpecd = (-201049)
DAQmxErrorAccessorySettingNotApplicable = (-201048)
DAQmxErrorDeviceAndConnectorSpecdAlreadyOccupied = (-201047)
DAQmxErrorIllegalAccessoryTypeForDeviceSpecd = (-201046)
DAQmxErrorInvalidDeviceConnectorNumberSpecd = (-201045)
DAQmxErrorInvalidAccessoryName = (-201044)
DAQmxErrorMoreThanOneMatchForSpecdDevice = (-201043)
DAQmxErrorNoMatchForSpecdDevice = (-201042)
DAQmxErrorProductTypeAndProductNumberConflict = (-201041)
DAQmxErrorExtraPropertyDetectedInSpecdObject = (-201040)
DAQmxErrorRequiredPropertyMissing = (-201039)
DAQmxErrorCantSetAuthorForLocalChan = (-201038)
DAQmxErrorInvalidTimeValue = (-201037)
DAQmxErrorInvalidTimeFormat = (-201036)
DAQmxErrorDigDevChansSpecdInModeOtherThanParallel = (-201035)
DAQmxErrorCascadeDigitizationModeNotSupported = (-201034)
DAQmxErrorSpecdSlotAlreadyOccupied = (-201033)
DAQmxErrorInvalidSCXISlotNumberSpecd = (-201032)
DAQmxErrorAddressAlreadyInUse = (-201031)
DAQmxErrorSpecdDeviceDoesNotSupportRTSI = (-201030)
DAQmxErrorSpecdDeviceIsAlreadyOnRTSIBus = (-201029)
DAQmxErrorIdentifierInUse = (-201028)
DAQmxErrorWaitForNextSampleClockOrReadDetected3OrMoreMissedSampClks = (-201027)
DAQmxErrorHWTimedAndDataXferPIO = (-201026)
DAQmxErrorNonBufferedAndHWTimed = (-201025)
DAQmxErrorCTROutSampClkPeriodShorterThanGenPulseTrainPeriodPolled = (-201024)
DAQmxErrorCTROutSampClkPeriodShorterThanGenPulseTrainPeriod2 = (-201023)
DAQmxErrorCOCannotKeepUpInHWTimedSinglePointPolled = (-201022)
DAQmxErrorWriteRecoveryCannotKeepUpInHWTimedSinglePoint = (-201021)
DAQmxErrorNoChangeDetectionOnSelectedLineForDevice = (-201020)
DAQmxErrorSMIOPauseTriggersNotSupportedWithChannelExpansion = (-201019)
DAQmxErrorClockMasterForExternalClockNotLongestPipeline = (-201018)
DAQmxErrorUnsupportedUnicodeByteOrderMarker = (-201017)
DAQmxErrorTooManyInstructionsInLoopInScript = (-201016)
DAQmxErrorPLLNotLocked = (-201015)
DAQmxErrorIfElseBlockNotAllowedInFiniteRepeatLoopInScript = (-201014)
DAQmxErrorIfElseBlockNotAllowedInConditionalRepeatLoopInScript = (-201013)
DAQmxErrorClearIsLastInstructionInIfElseBlockInScript = (-201012)
DAQmxErrorInvalidWaitDurationBeforeIfElseBlockInScript = (-201011)
DAQmxErrorMarkerPosInvalidBeforeIfElseBlockInScript = (-201010)
DAQmxErrorInvalidSubsetLengthBeforeIfElseBlockInScript = (-201009)
DAQmxErrorInvalidWaveformLengthBeforeIfElseBlockInScript = (-201008)
DAQmxErrorGenerateOrFiniteWaitInstructionExpectedBeforeIfElseBlockInScript = (-201007)
DAQmxErrorCalPasswordNotSupported = (-201006)
DAQmxErrorSetupCalNeededBeforeAdjustCal = (-201005)
DAQmxErrorMultipleChansNotSupportedDuringCalSetup = (-201004)
DAQmxErrorDevCannotBeAccessed = (-201003)
DAQmxErrorSampClkRateDoesntMatchSampClkSrc = (-201002)
DAQmxErrorSampClkRateNotSupportedWithEARDisabled = (-201001)
DAQmxErrorLabVIEWVersionDoesntSupportDAQmxEvents = (-201000)
DAQmxErrorCOReadyForNewValNotSupportedWithOnDemand = (-200999)
DAQmxErrorCIHWTimedSinglePointNotSupportedForMeasType = (-200998)
DAQmxErrorOnDemandNotSupportedWithHWTimedSinglePoint = (-200997)
DAQmxErrorHWTimedSinglePointAndDataXferNotProgIO = (-200996)
DAQmxErrorMemMapAndHWTimedSinglePoint = (-200995)
DAQmxErrorCannotSetPropertyWhenHWTimedSinglePointTaskIsRunning = (-200994)
DAQmxErrorCTROutSampClkPeriodShorterThanGenPulseTrainPeriod = (-200993)
DAQmxErrorTooManyEventsGenerated = (-200992)
DAQmxErrorMStudioCppRemoveEventsBeforeStop = (-200991)
DAQmxErrorCAPICannotRegisterSyncEventsFromMultipleThreads = (-200990)
DAQmxErrorReadWaitNextSampClkWaitMismatchTwo = (-200989)
DAQmxErrorReadWaitNextSampClkWaitMismatchOne = (-200988)
DAQmxErrorDAQmxSignalEventTypeNotSupportedByChanTypesOrDevicesInTask = (-200987)
DAQmxErrorCannotUnregisterDAQmxSoftwareEventWhileTaskIsRunning = (-200986)
DAQmxErrorAutoStartWriteNotAllowedEventRegistered = (-200985)
DAQmxErrorAutoStartReadNotAllowedEventRegistered = (-200984)
DAQmxErrorCannotGetPropertyWhenTaskNotReservedCommittedOrRunning = (-200983)
DAQmxErrorSignalEventsNotSupportedByDevice = (-200982)
DAQmxErrorEveryNSamplesAcqIntoBufferEventNotSupportedByDevice = (-200981)
DAQmxErrorEveryNSampsTransferredFromBufferEventNotSupportedByDevice = (-200980)
DAQmxErrorCAPISyncEventsTaskStateChangeNotAllowedFromDifferentThread = (-200979)
DAQmxErrorDAQmxSWEventsWithDifferentCallMechanisms = (-200978)
DAQmxErrorCantSaveChanWithPolyCalScaleAndAllowInteractiveEdit = (-200977)
DAQmxErrorChanDoesNotSupportCJC = (-200976)
DAQmxErrorCOReadyForNewValNotSupportedWithHWTimedSinglePoint = (-200975)
DAQmxErrorDACAllowConnToGndNotSupportedByDevWhenRefSrcExt = (-200974)
DAQmxErrorCantGetPropertyTaskNotRunning = (-200973)
DAQmxErrorCantSetPropertyTaskNotRunning = (-200972)
DAQmxErrorCantSetPropertyTaskNotRunningCommitted = (-200971)
DAQmxErrorAIEveryNSampsEventIntervalNotMultipleOf2 = (-200970)
DAQmxErrorInvalidTEDSPhysChanNotAI = (-200969)
DAQmxErrorCAPICannotPerformTaskOperationInAsyncCallback = (-200968)
DAQmxErrorEveryNSampsTransferredFromBufferEventAlreadyRegistered = (-200967)
DAQmxErrorEveryNSampsAcqIntoBufferEventAlreadyRegistered = (-200966)
DAQmxErrorEveryNSampsTransferredFromBufferNotForInput = (-200965)
DAQmxErrorEveryNSampsAcqIntoBufferNotForOutput = (-200964)
DAQmxErrorAOSampTimingTypeDifferentIn2Tasks = (-200963)
DAQmxErrorCouldNotDownloadFirmwareHWDamaged = (-200962)
DAQmxErrorCouldNotDownloadFirmwareFileMissingOrDamaged = (-200961)
DAQmxErrorCannotRegisterDAQmxSoftwareEventWhileTaskIsRunning = (-200960)
DAQmxErrorDifferentRawDataCompression = (-200959)
DAQmxErrorConfiguredTEDSInterfaceDevNotDetected = (-200958)
DAQmxErrorCompressedSampSizeExceedsResolution = (-200957)
DAQmxErrorChanDoesNotSupportCompression = (-200956)
DAQmxErrorDifferentRawDataFormats = (-200955)
DAQmxErrorSampClkOutputTermIncludesStartTrigSrc = (-200954)
DAQmxErrorStartTrigSrcEqualToSampClkSrc = (-200953)
DAQmxErrorEventOutputTermIncludesTrigSrc = (-200952)
DAQmxErrorCOMultipleWritesBetweenSampClks = (-200951)
DAQmxErrorDoneEventAlreadyRegistered = (-200950)
DAQmxErrorSignalEventAlreadyRegistered = (-200949)
DAQmxErrorCannotHaveTimedLoopAndDAQmxSignalEventsInSameTask = (-200948)
DAQmxErrorNeedLabVIEW711PatchToUseDAQmxEvents = (-200947)
DAQmxErrorStartFailedDueToWriteFailure = (-200946)
DAQmxErrorDataXferCustomThresholdNotDMAXferMethodSpecifiedForDev = (-200945)
DAQmxErrorDataXferRequestConditionNotSpecifiedForCustomThreshold = (-200944)
DAQmxErrorDataXferCustomThresholdNotSpecified = (-200943)
DAQmxErrorCAPISyncCallbackNotSupportedOnThisPlatform = (-200942)
DAQmxErrorCalChanReversePolyCoefNotSpecd = (-200941)
DAQmxErrorCalChanForwardPolyCoefNotSpecd = (-200940)
DAQmxErrorChanCalRepeatedNumberInPreScaledVals = (-200939)
DAQmxErrorChanCalTableNumScaledNotEqualNumPrescaledVals = (-200938)
DAQmxErrorChanCalTableScaledValsNotSpecd = (-200937)
DAQmxErrorChanCalTablePreScaledValsNotSpecd = (-200936)
DAQmxErrorChanCalScaleTypeNotSet = (-200935)
DAQmxErrorChanCalExpired = (-200934)
DAQmxErrorChanCalExpirationDateNotSet = (-200933)
DAQmxError3OutputPortCombinationGivenSampTimingType653x = (-200932)
DAQmxError3InputPortCombinationGivenSampTimingType653x = (-200931)
DAQmxError2OutputPortCombinationGivenSampTimingType653x = (-200930)
DAQmxError2InputPortCombinationGivenSampTimingType653x = (-200929)
DAQmxErrorPatternMatcherMayBeUsedByOneTrigOnly = (-200928)
DAQmxErrorNoChansSpecdForPatternSource = (-200927)
DAQmxErrorChangeDetectionChanNotInTask = (-200926)
DAQmxErrorChangeDetectionChanNotTristated = (-200925)
DAQmxErrorWaitModeValueNotSupportedNonBuffered = (-200924)
DAQmxErrorWaitModePropertyNotSupportedNonBuffered = (-200923)
DAQmxErrorCantSavePerLineConfigDigChanSoInteractiveEditsAllowed = (-200922)
DAQmxErrorCantSaveNonPortMultiLineDigChanSoInteractiveEditsAllowed = (-200921)
DAQmxErrorBufferSizeNotMultipleOfEveryNSampsEventIntervalNoIrqOnDev = (-200920)
DAQmxErrorGlobalTaskNameAlreadyChanName = (-200919)
DAQmxErrorGlobalChanNameAlreadyTaskName = (-200918)
DAQmxErrorAOEveryNSampsEventIntervalNotMultipleOf2 = (-200917)
DAQmxErrorSampleTimebaseDivisorNotSupportedGivenTimingType = (-200916)
DAQmxErrorHandshakeEventOutputTermNotSupportedGivenTimingType = (-200915)
DAQmxErrorChangeDetectionOutputTermNotSupportedGivenTimingType = (-200914)
DAQmxErrorReadyForTransferOutputTermNotSupportedGivenTimingType = (-200913)
DAQmxErrorRefTrigOutputTermNotSupportedGivenTimingType = (-200912)
DAQmxErrorStartTrigOutputTermNotSupportedGivenTimingType = (-200911)
DAQmxErrorSampClockOutputTermNotSupportedGivenTimingType = (-200910)
DAQmxError20MhzTimebaseNotSupportedGivenTimingType = (-200909)
DAQmxErrorSampClockSourceNotSupportedGivenTimingType = (-200908)
DAQmxErrorRefTrigTypeNotSupportedGivenTimingType = (-200907)
DAQmxErrorPauseTrigTypeNotSupportedGivenTimingType = (-200906)
DAQmxErrorHandshakeTrigTypeNotSupportedGivenTimingType = (-200905)
DAQmxErrorStartTrigTypeNotSupportedGivenTimingType = (-200904)
DAQmxErrorRefClkSrcNotSupported = (-200903)
DAQmxErrorDataVoltageLowAndHighIncompatible = (-200902)
DAQmxErrorInvalidCharInDigPatternString = (-200901)
DAQmxErrorCantUsePort3AloneGivenSampTimingTypeOn653x = (-200900)
DAQmxErrorCantUsePort1AloneGivenSampTimingTypeOn653x = (-200899)
DAQmxErrorPartialUseOfPhysicalLinesWithinPortNotSupported653x = (-200898)
DAQmxErrorPhysicalChanNotSupportedGivenSampTimingType653x = (-200897)
DAQmxErrorCanExportOnlyDigEdgeTrigs = (-200896)
DAQmxErrorRefTrigDigPatternSizeDoesNotMatchSourceSize = (-200895)
DAQmxErrorStartTrigDigPatternSizeDoesNotMatchSourceSize = (-200894)
DAQmxErrorChangeDetectionRisingAndFallingEdgeChanDontMatch = (-200893)
DAQmxErrorPhysicalChansForChangeDetectionAndPatternMatch653x = (-200892)
DAQmxErrorCanExportOnlyOnboardSampClk = (-200891)
DAQmxErrorInternalSampClkNotRisingEdge = (-200890)
DAQmxErrorRefTrigDigPatternChanNotInTask = (-200889)
DAQmxErrorRefTrigDigPatternChanNotTristated = (-200888)
DAQmxErrorStartTrigDigPatternChanNotInTask = (-200887)
DAQmxErrorStartTrigDigPatternChanNotTristated = (-200886)
DAQmxErrorPXIStarAndClock10Sync = (-200885)
DAQmxErrorGlobalChanCannotBeSavedSoInteractiveEditsAllowed = (-200884)
DAQmxErrorTaskCannotBeSavedSoInteractiveEditsAllowed = (-200883)
DAQmxErrorInvalidGlobalChan = (-200882)
DAQmxErrorEveryNSampsEventAlreadyRegistered = (-200881)
DAQmxErrorEveryNSampsEventIntervalZeroNotSupported = (-200880)
DAQmxErrorChanSizeTooBigForU16PortWrite = (-200879)
DAQmxErrorChanSizeTooBigForU16PortRead = (-200878)
DAQmxErrorBufferSizeNotMultipleOfEveryNSampsEventIntervalWhenDMA = (-200877)
DAQmxErrorWriteWhenTaskNotRunningCOTicks = (-200876)
DAQmxErrorWriteWhenTaskNotRunningCOFreq = (-200875)
DAQmxErrorWriteWhenTaskNotRunningCOTime = (-200874)
DAQmxErrorAOMinMaxNotSupportedDACRangeTooSmall = (-200873)
DAQmxErrorAOMinMaxNotSupportedGivenDACRange = (-200872)
DAQmxErrorAOMinMaxNotSupportedGivenDACRangeAndOffsetVal = (-200871)
DAQmxErrorAOMinMaxNotSupportedDACOffsetValInappropriate = (-200870)
DAQmxErrorAOMinMaxNotSupportedGivenDACOffsetVal = (-200869)
DAQmxErrorAOMinMaxNotSupportedDACRefValTooSmall = (-200868)
DAQmxErrorAOMinMaxNotSupportedGivenDACRefVal = (-200867)
DAQmxErrorAOMinMaxNotSupportedGivenDACRefAndOffsetVal = (-200866)
DAQmxErrorWhenAcqCompAndNumSampsPerChanExceedsOnBrdBufSize = (-200865)
DAQmxErrorWhenAcqCompAndNoRefTrig = (-200864)
DAQmxErrorWaitForNextSampClkNotSupported = (-200863)
DAQmxErrorDevInUnidentifiedPXIChassis = (-200862)
DAQmxErrorMaxSoundPressureMicSensitivitRelatedAIPropertiesNotSupportedByDev = (-200861)
DAQmxErrorMaxSoundPressureAndMicSensitivityNotSupportedByDev = (-200860)
DAQmxErrorAOBufferSizeZeroForSampClkTimingType = (-200859)
DAQmxErrorAOCallWriteBeforeStartForSampClkTimingType = (-200858)
DAQmxErrorInvalidCalLowPassCutoffFreq = (-200857)
DAQmxErrorSimulationCannotBeDisabledForDevCreatedAsSimulatedDev = (-200856)
DAQmxErrorCannotAddNewDevsAfterTaskConfiguration = (-200855)
DAQmxErrorDifftSyncPulseSrcAndSampClkTimebaseSrcDevMultiDevTask = (-200854)
DAQmxErrorTermWithoutDevInMultiDevTask = (-200853)
DAQmxErrorSyncNoDevSampClkTimebaseOrSyncPulseInPXISlot2 = (-200852)
DAQmxErrorPhysicalChanNotOnThisConnector = (-200851)
DAQmxErrorNumSampsToWaitNotGreaterThanZeroInScript = (-200850)
DAQmxErrorNumSampsToWaitNotMultipleOfAlignmentQuantumInScript = (-200849)
DAQmxErrorEveryNSamplesEventNotSupportedForNonBufferedTasks = (-200848)
DAQmxErrorBufferedAndDataXferPIO = (-200847)
DAQmxErrorCannotWriteWhenAutoStartFalseAndTaskNotRunning = (-200846)
DAQmxErrorNonBufferedAndDataXferInterrupts = (-200845)
DAQmxErrorWriteFailedMultipleCtrsWithFREQOUT = (-200844)
DAQmxErrorReadNotCompleteBefore3SampClkEdges = (-200843)
DAQmxErrorCtrHWTimedSinglePointAndDataXferNotProgIO = (-200842)
DAQmxErrorPrescalerNot1ForInputTerminal = (-200841)
DAQmxErrorPrescalerNot1ForTimebaseSrc = (-200840)
DAQmxErrorSampClkTimingTypeWhenTristateIsFalse = (-200839)
DAQmxErrorOutputBufferSizeNotMultOfXferSize = (-200838)
DAQmxErrorSampPerChanNotMultOfXferSize = (-200837)
DAQmxErrorWriteToTEDSFailed = (-200836)
DAQmxErrorSCXIDevNotUsablePowerTurnedOff = (-200835)
DAQmxErrorCannotReadWhenAutoStartFalseBufSizeZeroAndTaskNotRunning = (-200834)
DAQmxErrorCannotReadWhenAutoStartFalseHWTimedSinglePtAndTaskNotRunning = (-200833)
DAQmxErrorCannotReadWhenAutoStartFalseOnDemandAndTaskNotRunning = (-200832)
DAQmxErrorSimultaneousAOWhenNotOnDemandTiming = (-200831)
DAQmxErrorMemMapAndSimultaneousAO = (-200830)
DAQmxErrorWriteFailedMultipleCOOutputTypes = (-200829)
DAQmxErrorWriteToTEDSNotSupportedOnRT = (-200828)
DAQmxErrorVirtualTEDSDataFileError = (-200827)
DAQmxErrorTEDSSensorDataError = (-200826)
DAQmxErrorDataSizeMoreThanSizeOfEEPROMOnTEDS = (-200825)
DAQmxErrorPROMOnTEDSContainsBasicTEDSData = (-200824)
DAQmxErrorPROMOnTEDSAlreadyWritten = (-200823)
DAQmxErrorTEDSDoesNotContainPROM = (-200822)
DAQmxErrorHWTimedSinglePointNotSupportedAI = (-200821)
DAQmxErrorHWTimedSinglePointOddNumChansInAITask = (-200820)
DAQmxErrorCantUseOnlyOnBoardMemWithProgrammedIO = (-200819)
DAQmxErrorSwitchDevShutDownDueToHighTemp = (-200818)
DAQmxErrorExcitationNotSupportedWhenTermCfgDiff = (-200817)
DAQmxErrorTEDSMinElecValGEMaxElecVal = (-200816)
DAQmxErrorTEDSMinPhysValGEMaxPhysVal = (-200815)
DAQmxErrorCIOnboardClockNotSupportedAsInputTerm = (-200814)
DAQmxErrorInvalidSampModeForPositionMeas = (-200813)
DAQmxErrorTrigWhenAOHWTimedSinglePtSampMode = (-200812)
DAQmxErrorDAQmxCantUseStringDueToUnknownChar = (-200811)
DAQmxErrorDAQmxCantRetrieveStringDueToUnknownChar = (-200810)
DAQmxErrorClearTEDSNotSupportedOnRT = (-200809)
DAQmxErrorCfgTEDSNotSupportedOnRT = (-200808)
DAQmxErrorProgFilterClkCfgdToDifferentMinPulseWidthBySameTask1PerDev = (-200807)
DAQmxErrorProgFilterClkCfgdToDifferentMinPulseWidthByAnotherTask1PerDev = (-200806)
DAQmxErrorNoLastExtCalDateTimeLastExtCalNotDAQmx = (-200804)
DAQmxErrorCannotWriteNotStartedAutoStartFalseNotOnDemandHWTimedSglPt = (-200803)
DAQmxErrorCannotWriteNotStartedAutoStartFalseNotOnDemandBufSizeZero = (-200802)
DAQmxErrorCOInvalidTimingSrcDueToSignal = (-200801)
DAQmxErrorCIInvalidTimingSrcForSampClkDueToSampTimingType = (-200800)
DAQmxErrorCIInvalidTimingSrcForEventCntDueToSampMode = (-200799)
DAQmxErrorNoChangeDetectOnNonInputDigLineForDev = (-200798)
DAQmxErrorEmptyStringTermNameNotSupported = (-200797)
DAQmxErrorMemMapEnabledForHWTimedNonBufferedAO = (-200796)
DAQmxErrorDevOnboardMemOverflowDuringHWTimedNonBufferedGen = (-200795)
DAQmxErrorCODAQmxWriteMultipleChans = (-200794)
DAQmxErrorCantMaintainExistingValueAOSync = (-200793)
DAQmxErrorMStudioMultiplePhysChansNotSupported = (-200792)
DAQmxErrorCantConfigureTEDSForChan = (-200791)
DAQmxErrorWriteDataTypeTooSmall = (-200790)
DAQmxErrorReadDataTypeTooSmall = (-200789)
DAQmxErrorMeasuredBridgeOffsetTooHigh = (-200788)
DAQmxErrorStartTrigConflictWithCOHWTimedSinglePt = (-200787)
DAQmxErrorSampClkRateExtSampClkTimebaseRateMismatch = (-200786)
DAQmxErrorInvalidTimingSrcDueToSampTimingType = (-200785)
DAQmxErrorVirtualTEDSFileNotFound = (-200784)
DAQmxErrorMStudioNoForwardPolyScaleCoeffs = (-200783)
DAQmxErrorMStudioNoReversePolyScaleCoeffs = (-200782)
DAQmxErrorMStudioNoPolyScaleCoeffsUseCalc = (-200781)
DAQmxErrorMStudioNoForwardPolyScaleCoeffsUseCalc = (-200780)
DAQmxErrorMStudioNoReversePolyScaleCoeffsUseCalc = (-200779)
DAQmxErrorCOSampModeSampTimingTypeSampClkConflict = (-200778)
DAQmxErrorDevCannotProduceMinPulseWidth = (-200777)
DAQmxErrorCannotProduceMinPulseWidthGivenPropertyValues = (-200776)
DAQmxErrorTermCfgdToDifferentMinPulseWidthByAnotherTask = (-200775)
DAQmxErrorTermCfgdToDifferentMinPulseWidthByAnotherProperty = (-200774)
DAQmxErrorDigSyncNotAvailableOnTerm = (-200773)
DAQmxErrorDigFilterNotAvailableOnTerm = (-200772)
DAQmxErrorDigFilterEnabledMinPulseWidthNotCfg = (-200771)
DAQmxErrorDigFilterAndSyncBothEnabled = (-200770)
DAQmxErrorHWTimedSinglePointAOAndDataXferNotProgIO = (-200769)
DAQmxErrorNonBufferedAOAndDataXferNotProgIO = (-200768)
DAQmxErrorProgIODataXferForBufferedAO = (-200767)
DAQmxErrorTEDSLegacyTemplateIDInvalidOrUnsupported = (-200766)
DAQmxErrorTEDSMappingMethodInvalidOrUnsupported = (-200765)
DAQmxErrorTEDSLinearMappingSlopeZero = (-200764)
DAQmxErrorAIInputBufferSizeNotMultOfXferSize = (-200763)
DAQmxErrorNoSyncPulseExtSampClkTimebase = (-200762)
DAQmxErrorNoSyncPulseAnotherTaskRunning = (-200761)
DAQmxErrorAOMinMaxNotInGainRange = (-200760)
DAQmxErrorAOMinMaxNotInDACRange = (-200759)
DAQmxErrorDevOnlySupportsSampClkTimingAO = (-200758)
DAQmxErrorDevOnlySupportsSampClkTimingAI = (-200757)
DAQmxErrorTEDSIncompatibleSensorAndMeasType = (-200756)
DAQmxErrorTEDSMultipleCalTemplatesNotSupported = (-200755)
DAQmxErrorTEDSTemplateParametersNotSupported = (-200754)
DAQmxErrorParsingTEDSData = (-200753)
DAQmxErrorMultipleActivePhysChansNotSupported = (-200752)
DAQmxErrorNoChansSpecdForChangeDetect = (-200751)
DAQmxErrorInvalidCalVoltageForGivenGain = (-200750)
DAQmxErrorInvalidCalGain = (-200749)
DAQmxErrorMultipleWritesBetweenSampClks = (-200748)
DAQmxErrorInvalidAcqTypeForFREQOUT = (-200747)
DAQmxErrorSuitableTimebaseNotFoundTimeCombo2 = (-200746)
DAQmxErrorSuitableTimebaseNotFoundFrequencyCombo2 = (-200745)
DAQmxErrorRefClkRateRefClkSrcMismatch = (-200744)
DAQmxErrorNoTEDSTerminalBlock = (-200743)
DAQmxErrorCorruptedTEDSMemory = (-200742)
DAQmxErrorTEDSNotSupported = (-200741)
DAQmxErrorTimingSrcTaskStartedBeforeTimedLoop = (-200740)
DAQmxErrorPropertyNotSupportedForTimingSrc = (-200739)
DAQmxErrorTimingSrcDoesNotExist = (-200738)
DAQmxErrorInputBufferSizeNotEqualSampsPerChanForFiniteSampMode = (-200737)
DAQmxErrorFREQOUTCannotProduceDesiredFrequency2 = (-200736)
DAQmxErrorExtRefClkRateNotSpecified = (-200735)
DAQmxErrorDeviceDoesNotSupportDMADataXferForNonBufferedAcq = (-200734)
DAQmxErrorDigFilterMinPulseWidthSetWhenTristateIsFalse = (-200733)
DAQmxErrorDigFilterEnableSetWhenTristateIsFalse = (-200732)
DAQmxErrorNoHWTimingWithOnDemand = (-200731)
DAQmxErrorCannotDetectChangesWhenTristateIsFalse = (-200730)
DAQmxErrorCannotHandshakeWhenTristateIsFalse = (-200729)
DAQmxErrorLinesUsedForStaticInputNotForHandshakingControl = (-200728)
DAQmxErrorLinesUsedForHandshakingControlNotForStaticInput = (-200727)
DAQmxErrorLinesUsedForStaticInputNotForHandshakingInput = (-200726)
DAQmxErrorLinesUsedForHandshakingInputNotForStaticInput = (-200725)
DAQmxErrorDifferentDITristateValsForChansInTask = (-200724)
DAQmxErrorTimebaseCalFreqVarianceTooLarge = (-200723)
DAQmxErrorTimebaseCalFailedToConverge = (-200722)
DAQmxErrorInadequateResolutionForTimebaseCal = (-200721)
DAQmxErrorInvalidAOGainCalConst = (-200720)
DAQmxErrorInvalidAOOffsetCalConst = (-200719)
DAQmxErrorInvalidAIGainCalConst = (-200718)
DAQmxErrorInvalidAIOffsetCalConst = (-200717)
DAQmxErrorDigOutputOverrun = (-200716)
DAQmxErrorDigInputOverrun = (-200715)
DAQmxErrorAcqStoppedDriverCantXferDataFastEnough = (-200714)
DAQmxErrorChansCantAppearInSameTask = (-200713)
DAQmxErrorInputCfgFailedBecauseWatchdogExpired = (-200712)
DAQmxErrorAnalogTrigChanNotExternal = (-200711)
DAQmxErrorTooManyChansForInternalAIInputSrc = (-200710)
DAQmxErrorTEDSSensorNotDetected = (-200709)
DAQmxErrorPrptyGetSpecdActiveItemFailedDueToDifftValues = (-200708)
DAQmxErrorRoutingDestTermPXIClk10InNotInSlot2 = (-200706)
DAQmxErrorRoutingDestTermPXIStarXNotInSlot2 = (-200705)
DAQmxErrorRoutingSrcTermPXIStarXNotInSlot2 = (-200704)
DAQmxErrorRoutingSrcTermPXIStarInSlot16AndAbove = (-200703)
DAQmxErrorRoutingDestTermPXIStarInSlot16AndAbove = (-200702)
DAQmxErrorRoutingDestTermPXIStarInSlot2 = (-200701)
DAQmxErrorRoutingSrcTermPXIStarInSlot2 = (-200700)
DAQmxErrorRoutingDestTermPXIChassisNotIdentified = (-200699)
DAQmxErrorRoutingSrcTermPXIChassisNotIdentified = (-200698)
DAQmxErrorFailedToAcquireCalData = (-200697)
DAQmxErrorBridgeOffsetNullingCalNotSupported = (-200696)
DAQmxErrorAIMaxNotSpecified = (-200695)
DAQmxErrorAIMinNotSpecified = (-200694)
DAQmxErrorOddTotalBufferSizeToWrite = (-200693)
DAQmxErrorOddTotalNumSampsToWrite = (-200692)
DAQmxErrorBufferWithWaitMode = (-200691)
DAQmxErrorBufferWithHWTimedSinglePointSampMode = (-200690)
DAQmxErrorCOWritePulseLowTicksNotSupported = (-200689)
DAQmxErrorCOWritePulseHighTicksNotSupported = (-200688)
DAQmxErrorCOWritePulseLowTimeOutOfRange = (-200687)
DAQmxErrorCOWritePulseHighTimeOutOfRange = (-200686)
DAQmxErrorCOWriteFreqOutOfRange = (-200685)
DAQmxErrorCOWriteDutyCycleOutOfRange = (-200684)
DAQmxErrorInvalidInstallation = (-200683)
DAQmxErrorRefTrigMasterSessionUnavailable = (-200682)
DAQmxErrorRouteFailedBecauseWatchdogExpired = (-200681)
DAQmxErrorDeviceShutDownDueToHighTemp = (-200680)
DAQmxErrorNoMemMapWhenHWTimedSinglePoint = (-200679)
DAQmxErrorWriteFailedBecauseWatchdogExpired = (-200678)
DAQmxErrorDifftInternalAIInputSrcs = (-200677)
DAQmxErrorDifftAIInputSrcInOneChanGroup = (-200676)
DAQmxErrorInternalAIInputSrcInMultChanGroups = (-200675)
DAQmxErrorSwitchOpFailedDueToPrevError = (-200674)
DAQmxErrorWroteMultiSampsUsingSingleSampWrite = (-200673)
DAQmxErrorMismatchedInputArraySizes = (-200672)
DAQmxErrorCantExceedRelayDriveLimit = (-200671)
DAQmxErrorDACRngLowNotEqualToMinusRefVal = (-200670)
DAQmxErrorCantAllowConnectDACToGnd = (-200669)
DAQmxErrorWatchdogTimeoutOutOfRangeAndNotSpecialVal = (-200668)
DAQmxErrorNoWatchdogOutputOnPortReservedForInput = (-200667)
DAQmxErrorNoInputOnPortCfgdForWatchdogOutput = (-200666)
DAQmxErrorWatchdogExpirationStateNotEqualForLinesInPort = (-200665)
DAQmxErrorCannotPerformOpWhenTaskNotReserved = (-200664)
DAQmxErrorPowerupStateNotSupported = (-200663)
DAQmxErrorWatchdogTimerNotSupported = (-200662)
DAQmxErrorOpNotSupportedWhenRefClkSrcNone = (-200661)
DAQmxErrorSampClkRateUnavailable = (-200660)
DAQmxErrorPrptyGetSpecdSingleActiveChanFailedDueToDifftVals = (-200659)
DAQmxErrorPrptyGetImpliedActiveChanFailedDueToDifftVals = (-200658)
DAQmxErrorPrptyGetSpecdActiveChanFailedDueToDifftVals = (-200657)
DAQmxErrorNoRegenWhenUsingBrdMem = (-200656)
DAQmxErrorNonbufferedReadMoreThanSampsPerChan = (-200655)
DAQmxErrorWatchdogExpirationTristateNotSpecdForEntirePort = (-200654)
DAQmxErrorPowerupTristateNotSpecdForEntirePort = (-200653)
DAQmxErrorPowerupStateNotSpecdForEntirePort = (-200652)
DAQmxErrorCantSetWatchdogExpirationOnDigInChan = (-200651)
DAQmxErrorCantSetPowerupStateOnDigInChan = (-200650)
DAQmxErrorPhysChanNotInTask = (-200649)
DAQmxErrorPhysChanDevNotInTask = (-200648)
DAQmxErrorDigInputNotSupported = (-200647)
DAQmxErrorDigFilterIntervalNotEqualForLines = (-200646)
DAQmxErrorDigFilterIntervalAlreadyCfgd = (-200645)
DAQmxErrorCantResetExpiredWatchdog = (-200644)
DAQmxErrorActiveChanTooManyLinesSpecdWhenGettingPrpty = (-200643)
DAQmxErrorActiveChanNotSpecdWhenGetting1LinePrpty = (-200642)
DAQmxErrorDigPrptyCannotBeSetPerLine = (-200641)
DAQmxErrorSendAdvCmpltAfterWaitForTrigInScanlist = (-200640)
DAQmxErrorDisconnectionRequiredInScanlist = (-200639)
DAQmxErrorTwoWaitForTrigsAfterConnectionInScanlist = (-200638)
DAQmxErrorActionSeparatorRequiredAfterBreakingConnectionInScanlist = (-200637)
DAQmxErrorConnectionInScanlistMustWaitForTrig = (-200636)
DAQmxErrorActionNotSupportedTaskNotWatchdog = (-200635)
DAQmxErrorWfmNameSameAsScriptName = (-200634)
DAQmxErrorScriptNameSameAsWfmName = (-200633)
DAQmxErrorDSFStopClock = (-200632)
DAQmxErrorDSFReadyForStartClock = (-200631)
DAQmxErrorWriteOffsetNotMultOfIncr = (-200630)
DAQmxErrorDifferentPrptyValsNotSupportedOnDev = (-200629)
DAQmxErrorRefAndPauseTrigConfigured = (-200628)
DAQmxErrorFailedToEnableHighSpeedInputClock = (-200627)
DAQmxErrorEmptyPhysChanInPowerUpStatesArray = (-200626)
DAQmxErrorActivePhysChanTooManyLinesSpecdWhenGettingPrpty = (-200625)
DAQmxErrorActivePhysChanNotSpecdWhenGetting1LinePrpty = (-200624)
DAQmxErrorPXIDevTempCausedShutDown = (-200623)
DAQmxErrorInvalidNumSampsToWrite = (-200622)
DAQmxErrorOutputFIFOUnderflow2 = (-200621)
DAQmxErrorRepeatedAIPhysicalChan = (-200620)
DAQmxErrorMultScanOpsInOneChassis = (-200619)
DAQmxErrorInvalidAIChanOrder = (-200618)
DAQmxErrorReversePowerProtectionActivated = (-200617)
DAQmxErrorInvalidAsynOpHandle = (-200616)
DAQmxErrorFailedToEnableHighSpeedOutput = (-200615)
DAQmxErrorCannotReadPastEndOfRecord = (-200614)
DAQmxErrorAcqStoppedToPreventInputBufferOverwriteOneDataXferMech = (-200613)
DAQmxErrorZeroBasedChanIndexInvalid = (-200612)
DAQmxErrorNoChansOfGivenTypeInTask = (-200611)
DAQmxErrorSampClkSrcInvalidForOutputValidForInput = (-200610)
DAQmxErrorOutputBufSizeTooSmallToStartGen = (-200609)
DAQmxErrorInputBufSizeTooSmallToStartAcq = (-200608)
DAQmxErrorExportTwoSignalsOnSameTerminal = (-200607)
DAQmxErrorChanIndexInvalid = (-200606)
DAQmxErrorRangeSyntaxNumberTooBig = (-200605)
DAQmxErrorNULLPtr = (-200604)
DAQmxErrorScaledMinEqualMax = (-200603)
DAQmxErrorPreScaledMinEqualMax = (-200602)
DAQmxErrorPropertyNotSupportedForScaleType = (-200601)
DAQmxErrorChannelNameGenerationNumberTooBig = (-200600)
DAQmxErrorRepeatedNumberInScaledValues = (-200599)
DAQmxErrorRepeatedNumberInPreScaledValues = (-200598)
DAQmxErrorLinesAlreadyReservedForOutput = (-200597)
DAQmxErrorSwitchOperationChansSpanMultipleDevsInList = (-200596)
DAQmxErrorInvalidIDInListAtBeginningOfSwitchOperation = (-200595)
DAQmxErrorMStudioInvalidPolyDirection = (-200594)
DAQmxErrorMStudioPropertyGetWhileTaskNotVerified = (-200593)
DAQmxErrorRangeWithTooManyObjects = (-200592)
DAQmxErrorCppDotNetAPINegativeBufferSize = (-200591)
DAQmxErrorCppCantRemoveInvalidEventHandler = (-200590)
DAQmxErrorCppCantRemoveEventHandlerTwice = (-200589)
DAQmxErrorCppCantRemoveOtherObjectsEventHandler = (-200588)
DAQmxErrorDigLinesReservedOrUnavailable = (-200587)
DAQmxErrorDSFFailedToResetStream = (-200586)
DAQmxErrorDSFReadyForOutputNotAsserted = (-200585)
DAQmxErrorSampToWritePerChanNotMultipleOfIncr = (-200584)
DAQmxErrorAOPropertiesCauseVoltageBelowMin = (-200583)
DAQmxErrorAOPropertiesCauseVoltageOverMax = (-200582)
DAQmxErrorPropertyNotSupportedWhenRefClkSrcNone = (-200581)
DAQmxErrorAIMaxTooSmall = (-200580)
DAQmxErrorAIMaxTooLarge = (-200579)
DAQmxErrorAIMinTooSmall = (-200578)
DAQmxErrorAIMinTooLarge = (-200577)
DAQmxErrorBuiltInCJCSrcNotSupported = (-200576)
DAQmxErrorTooManyPostTrigSampsPerChan = (-200575)
DAQmxErrorTrigLineNotFoundSingleDevRoute = (-200574)
DAQmxErrorDifferentInternalAIInputSources = (-200573)
DAQmxErrorDifferentAIInputSrcInOneChanGroup = (-200572)
DAQmxErrorInternalAIInputSrcInMultipleChanGroups = (-200571)
DAQmxErrorCAPIChanIndexInvalid = (-200570)
DAQmxErrorCollectionDoesNotMatchChanType = (-200569)
DAQmxErrorOutputCantStartChangedRegenerationMode = (-200568)
DAQmxErrorOutputCantStartChangedBufferSize = (-200567)
DAQmxErrorChanSizeTooBigForU32PortWrite = (-200566)
DAQmxErrorChanSizeTooBigForU8PortWrite = (-200565)
DAQmxErrorChanSizeTooBigForU32PortRead = (-200564)
DAQmxErrorChanSizeTooBigForU8PortRead = (-200563)
DAQmxErrorInvalidDigDataWrite = (-200562)
DAQmxErrorInvalidAODataWrite = (-200561)
DAQmxErrorWaitUntilDoneDoesNotIndicateDone = (-200560)
DAQmxErrorMultiChanTypesInTask = (-200559)
DAQmxErrorMultiDevsInTask = (-200558)
DAQmxErrorCannotSetPropertyWhenTaskRunning = (-200557)
DAQmxErrorCannotGetPropertyWhenTaskNotCommittedOrRunning = (-200556)
DAQmxErrorLeadingUnderscoreInString = (-200555)
DAQmxErrorTrailingSpaceInString = (-200554)
DAQmxErrorLeadingSpaceInString = (-200553)
DAQmxErrorInvalidCharInString = (-200552)
DAQmxErrorDLLBecameUnlocked = (-200551)
DAQmxErrorDLLLock = (-200550)
DAQmxErrorSelfCalConstsInvalid = (-200549)
DAQmxErrorInvalidTrigCouplingExceptForExtTrigChan = (-200548)
DAQmxErrorWriteFailsBufferSizeAutoConfigured = (-200547)
DAQmxErrorExtCalAdjustExtRefVoltageFailed = (-200546)
DAQmxErrorSelfCalFailedExtNoiseOrRefVoltageOutOfCal = (-200545)
DAQmxErrorExtCalTemperatureNotDAQmx = (-200544)
DAQmxErrorExtCalDateTimeNotDAQmx = (-200543)
DAQmxErrorSelfCalTemperatureNotDAQmx = (-200542)
DAQmxErrorSelfCalDateTimeNotDAQmx = (-200541)
DAQmxErrorDACRefValNotSet = (-200540)
DAQmxErrorAnalogMultiSampWriteNotSupported = (-200539)
DAQmxErrorInvalidActionInControlTask = (-200538)
DAQmxErrorPolyCoeffsInconsistent = (-200537)
DAQmxErrorSensorValTooLow = (-200536)
DAQmxErrorSensorValTooHigh = (-200535)
DAQmxErrorWaveformNameTooLong = (-200534)
DAQmxErrorIdentifierTooLongInScript = (-200533)
DAQmxErrorUnexpectedIDFollowingSwitchChanName = (-200532)
DAQmxErrorRelayNameNotSpecifiedInList = (-200531)
DAQmxErrorUnexpectedIDFollowingRelayNameInList = (-200530)
DAQmxErrorUnexpectedIDFollowingSwitchOpInList = (-200529)
DAQmxErrorInvalidLineGrouping = (-200528)
DAQmxErrorCtrMinMax = (-200527)
DAQmxErrorWriteChanTypeMismatch = (-200526)
DAQmxErrorReadChanTypeMismatch = (-200525)
DAQmxErrorWriteNumChansMismatch = (-200524)
DAQmxErrorOneChanReadForMultiChanTask = (-200523)
DAQmxErrorCannotSelfCalDuringExtCal = (-200522)
DAQmxErrorMeasCalAdjustOscillatorPhaseDAC = (-200521)
DAQmxErrorInvalidCalConstCalADCAdjustment = (-200520)
DAQmxErrorInvalidCalConstOscillatorFreqDACValue = (-200519)
DAQmxErrorInvalidCalConstOscillatorPhaseDACValue = (-200518)
DAQmxErrorInvalidCalConstOffsetDACValue = (-200517)
DAQmxErrorInvalidCalConstGainDACValue = (-200516)
DAQmxErrorInvalidNumCalADCReadsToAverage = (-200515)
DAQmxErrorInvalidCfgCalAdjustDirectPathOutputImpedance = (-200514)
DAQmxErrorInvalidCfgCalAdjustMainPathOutputImpedance = (-200513)
DAQmxErrorInvalidCfgCalAdjustMainPathPostAmpGainAndOffset = (-200512)
DAQmxErrorInvalidCfgCalAdjustMainPathPreAmpGain = (-200511)
DAQmxErrorInvalidCfgCalAdjustMainPreAmpOffset = (-200510)
DAQmxErrorMeasCalAdjustCalADC = (-200509)
DAQmxErrorMeasCalAdjustOscillatorFrequency = (-200508)
DAQmxErrorMeasCalAdjustDirectPathOutputImpedance = (-200507)
DAQmxErrorMeasCalAdjustMainPathOutputImpedance = (-200506)
DAQmxErrorMeasCalAdjustDirectPathGain = (-200505)
DAQmxErrorMeasCalAdjustMainPathPostAmpGainAndOffset = (-200504)
DAQmxErrorMeasCalAdjustMainPathPreAmpGain = (-200503)
DAQmxErrorMeasCalAdjustMainPathPreAmpOffset = (-200502)
DAQmxErrorInvalidDateTimeInEEPROM = (-200501)
DAQmxErrorUnableToLocateErrorResources = (-200500)
DAQmxErrorDotNetAPINotUnsigned32BitNumber = (-200499)
DAQmxErrorInvalidRangeOfObjectsSyntaxInString = (-200498)
DAQmxErrorAttemptToEnableLineNotPreviouslyDisabled = (-200497)
DAQmxErrorInvalidCharInPattern = (-200496)
DAQmxErrorIntermediateBufferFull = (-200495)
DAQmxErrorLoadTaskFailsBecauseNoTimingOnDev = (-200494)
DAQmxErrorCAPIReservedParamNotNULLNorEmpty = (-200493)
DAQmxErrorCAPIReservedParamNotNULL = (-200492)
DAQmxErrorCAPIReservedParamNotZero = (-200491)
DAQmxErrorSampleValueOutOfRange = (-200490)
DAQmxErrorChanAlreadyInTask = (-200489)
DAQmxErrorVirtualChanDoesNotExist = (-200488)
DAQmxErrorChanNotInTask = (-200486)
DAQmxErrorTaskNotInDataNeighborhood = (-200485)
DAQmxErrorCantSaveTaskWithoutReplace = (-200484)
DAQmxErrorCantSaveChanWithoutReplace = (-200483)
DAQmxErrorDevNotInTask = (-200482)
DAQmxErrorDevAlreadyInTask = (-200481)
DAQmxErrorCanNotPerformOpWhileTaskRunning = (-200479)
DAQmxErrorCanNotPerformOpWhenNoChansInTask = (-200478)
DAQmxErrorCanNotPerformOpWhenNoDevInTask = (-200477)
DAQmxErrorCannotPerformOpWhenTaskNotRunning = (-200475)
DAQmxErrorOperationTimedOut = (-200474)
DAQmxErrorCannotReadWhenAutoStartFalseAndTaskNotRunningOrCommitted = (-200473)
DAQmxErrorCannotWriteWhenAutoStartFalseAndTaskNotRunningOrCommitted = (-200472)
DAQmxErrorTaskVersionNew = (-200470)
DAQmxErrorChanVersionNew = (-200469)
DAQmxErrorEmptyString = (-200467)
DAQmxErrorChannelSizeTooBigForPortReadType = (-200466)
DAQmxErrorChannelSizeTooBigForPortWriteType = (-200465)
DAQmxErrorExpectedNumberOfChannelsVerificationFailed = (-200464)
DAQmxErrorNumLinesMismatchInReadOrWrite = (-200463)
DAQmxErrorOutputBufferEmpty = (-200462)
DAQmxErrorInvalidChanName = (-200461)
DAQmxErrorReadNoInputChansInTask = (-200460)
DAQmxErrorWriteNoOutputChansInTask = (-200459)
DAQmxErrorPropertyNotSupportedNotInputTask = (-200457)
DAQmxErrorPropertyNotSupportedNotOutputTask = (-200456)
DAQmxErrorGetPropertyNotInputBufferedTask = (-200455)
DAQmxErrorGetPropertyNotOutputBufferedTask = (-200454)
DAQmxErrorInvalidTimeoutVal = (-200453)
DAQmxErrorAttributeNotSupportedInTaskContext = (-200452)
DAQmxErrorAttributeNotQueryableUnlessTaskIsCommitted = (-200451)
DAQmxErrorAttributeNotSettableWhenTaskIsRunning = (-200450)
DAQmxErrorDACRngLowNotMinusRefValNorZero = (-200449)
DAQmxErrorDACRngHighNotEqualRefVal = (-200448)
DAQmxErrorUnitsNotFromCustomScale = (-200447)
DAQmxErrorInvalidVoltageReadingDuringExtCal = (-200446)
DAQmxErrorCalFunctionNotSupported = (-200445)
DAQmxErrorInvalidPhysicalChanForCal = (-200444)
DAQmxErrorExtCalNotComplete = (-200443)
DAQmxErrorCantSyncToExtStimulusFreqDuringCal = (-200442)
DAQmxErrorUnableToDetectExtStimulusFreqDuringCal = (-200441)
DAQmxErrorInvalidCloseAction = (-200440)
DAQmxErrorExtCalFunctionOutsideExtCalSession = (-200439)
DAQmxErrorInvalidCalArea = (-200438)
DAQmxErrorExtCalConstsInvalid = (-200437)
DAQmxErrorStartTrigDelayWithExtSampClk = (-200436)
DAQmxErrorDelayFromSampClkWithExtConv = (-200435)
DAQmxErrorFewerThan2PreScaledVals = (-200434)
DAQmxErrorFewerThan2ScaledValues = (-200433)
DAQmxErrorPhysChanOutputType = (-200432)
DAQmxErrorPhysChanMeasType = (-200431)
DAQmxErrorInvalidPhysChanType = (-200430)
DAQmxErrorLabVIEWEmptyTaskOrChans = (-200429)
DAQmxErrorLabVIEWInvalidTaskOrChans = (-200428)
DAQmxErrorInvalidRefClkRate = (-200427)
DAQmxErrorInvalidExtTrigImpedance = (-200426)
DAQmxErrorHystTrigLevelAIMax = (-200425)
DAQmxErrorLineNumIncompatibleWithVideoSignalFormat = (-200424)
DAQmxErrorTrigWindowAIMinAIMaxCombo = (-200423)
DAQmxErrorTrigAIMinAIMax = (-200422)
DAQmxErrorHystTrigLevelAIMin = (-200421)
DAQmxErrorInvalidSampRateConsiderRIS = (-200420)
DAQmxErrorInvalidReadPosDuringRIS = (-200419)
DAQmxErrorImmedTrigDuringRISMode = (-200418)
DAQmxErrorTDCNotEnabledDuringRISMode = (-200417)
DAQmxErrorMultiRecWithRIS = (-200416)
DAQmxErrorInvalidRefClkSrc = (-200415)
DAQmxErrorInvalidSampClkSrc = (-200414)
DAQmxErrorInsufficientOnBoardMemForNumRecsAndSamps = (-200413)
DAQmxErrorInvalidAIAttenuation = (-200412)
DAQmxErrorACCouplingNotAllowedWith50OhmImpedance = (-200411)
DAQmxErrorInvalidRecordNum = (-200410)
DAQmxErrorZeroSlopeLinearScale = (-200409)
DAQmxErrorZeroReversePolyScaleCoeffs = (-200408)
DAQmxErrorZeroForwardPolyScaleCoeffs = (-200407)
DAQmxErrorNoReversePolyScaleCoeffs = (-200406)
DAQmxErrorNoForwardPolyScaleCoeffs = (-200405)
DAQmxErrorNoPolyScaleCoeffs = (-200404)
DAQmxErrorReversePolyOrderLessThanNumPtsToCompute = (-200403)
DAQmxErrorReversePolyOrderNotPositive = (-200402)
DAQmxErrorNumPtsToComputeNotPositive = (-200401)
DAQmxErrorWaveformLengthNotMultipleOfIncr = (-200400)
DAQmxErrorCAPINoExtendedErrorInfoAvailable = (-200399)
DAQmxErrorCVIFunctionNotFoundInDAQmxDLL = (-200398)
DAQmxErrorCVIFailedToLoadDAQmxDLL = (-200397)
DAQmxErrorNoCommonTrigLineForImmedRoute = (-200396)
DAQmxErrorNoCommonTrigLineForTaskRoute = (-200395)
DAQmxErrorF64PrptyValNotUnsignedInt = (-200394)
DAQmxErrorRegisterNotWritable = (-200393)
DAQmxErrorInvalidOutputVoltageAtSampClkRate = (-200392)
DAQmxErrorStrobePhaseShiftDCMBecameUnlocked = (-200391)
DAQmxErrorDrivePhaseShiftDCMBecameUnlocked = (-200390)
DAQmxErrorClkOutPhaseShiftDCMBecameUnlocked = (-200389)
DAQmxErrorOutputBoardClkDCMBecameUnlocked = (-200388)
DAQmxErrorInputBoardClkDCMBecameUnlocked = (-200387)
DAQmxErrorInternalClkDCMBecameUnlocked = (-200386)
DAQmxErrorDCMLock = (-200385)
DAQmxErrorDataLineReservedForDynamicOutput = (-200384)
DAQmxErrorInvalidRefClkSrcGivenSampClkSrc = (-200383)
DAQmxErrorNoPatternMatcherAvailable = (-200382)
DAQmxErrorInvalidDelaySampRateBelowPhaseShiftDCMThresh = (-200381)
DAQmxErrorStrainGageCalibration = (-200380)
DAQmxErrorInvalidExtClockFreqAndDivCombo = (-200379)
DAQmxErrorCustomScaleDoesNotExist = (-200378)
DAQmxErrorOnlyFrontEndChanOpsDuringScan = (-200377)
DAQmxErrorInvalidOptionForDigitalPortChannel = (-200376)
DAQmxErrorUnsupportedSignalTypeExportSignal = (-200375)
DAQmxErrorInvalidSignalTypeExportSignal = (-200374)
DAQmxErrorUnsupportedTrigTypeSendsSWTrig = (-200373)
DAQmxErrorInvalidTrigTypeSendsSWTrig = (-200372)
DAQmxErrorRepeatedPhysicalChan = (-200371)
DAQmxErrorResourcesInUseForRouteInTask = (-200370)
DAQmxErrorResourcesInUseForRoute = (-200369)
DAQmxErrorRouteNotSupportedByHW = (-200368)
DAQmxErrorResourcesInUseForExportSignalPolarity = (-200367)
DAQmxErrorResourcesInUseForInversionInTask = (-200366)
DAQmxErrorResourcesInUseForInversion = (-200365)
DAQmxErrorExportSignalPolarityNotSupportedByHW = (-200364)
DAQmxErrorInversionNotSupportedByHW = (-200363)
DAQmxErrorOverloadedChansExistNotRead = (-200362)
DAQmxErrorInputFIFOOverflow2 = (-200361)
DAQmxErrorCJCChanNotSpecd = (-200360)
DAQmxErrorCtrExportSignalNotPossible = (-200359)
DAQmxErrorRefTrigWhenContinuous = (-200358)
DAQmxErrorIncompatibleSensorOutputAndDeviceInputRanges = (-200357)
DAQmxErrorCustomScaleNameUsed = (-200356)
DAQmxErrorPropertyValNotSupportedByHW = (-200355)
DAQmxErrorPropertyValNotValidTermName = (-200354)
DAQmxErrorResourcesInUseForProperty = (-200353)
DAQmxErrorCJCChanAlreadyUsed = (-200352)
DAQmxErrorForwardPolynomialCoefNotSpecd = (-200351)
DAQmxErrorTableScaleNumPreScaledAndScaledValsNotEqual = (-200350)
DAQmxErrorTableScalePreScaledValsNotSpecd = (-200349)
DAQmxErrorTableScaleScaledValsNotSpecd = (-200348)
DAQmxErrorIntermediateBufferSizeNotMultipleOfIncr = (-200347)
DAQmxErrorEventPulseWidthOutOfRange = (-200346)
DAQmxErrorEventDelayOutOfRange = (-200345)
DAQmxErrorSampPerChanNotMultipleOfIncr = (-200344)
DAQmxErrorCannotCalculateNumSampsTaskNotStarted = (-200343)
DAQmxErrorScriptNotInMem = (-200342)
DAQmxErrorOnboardMemTooSmall = (-200341)
DAQmxErrorReadAllAvailableDataWithoutBuffer = (-200340)
DAQmxErrorPulseActiveAtStart = (-200339)
DAQmxErrorCalTempNotSupported = (-200338)
DAQmxErrorDelayFromSampClkTooLong = (-200337)
DAQmxErrorDelayFromSampClkTooShort = (-200336)
DAQmxErrorAIConvRateTooHigh = (-200335)
DAQmxErrorDelayFromStartTrigTooLong = (-200334)
DAQmxErrorDelayFromStartTrigTooShort = (-200333)
DAQmxErrorSampRateTooHigh = (-200332)
DAQmxErrorSampRateTooLow = (-200331)
DAQmxErrorPFI0UsedForAnalogAndDigitalSrc = (-200330)
DAQmxErrorPrimingCfgFIFO = (-200329)
DAQmxErrorCannotOpenTopologyCfgFile = (-200328)
DAQmxErrorInvalidDTInsideWfmDataType = (-200327)
DAQmxErrorRouteSrcAndDestSame = (-200326)
DAQmxErrorReversePolynomialCoefNotSpecd = (-200325)
DAQmxErrorDevAbsentOrUnavailable = (-200324)
DAQmxErrorNoAdvTrigForMultiDevScan = (-200323)
DAQmxErrorInterruptsInsufficientDataXferMech = (-200322)
DAQmxErrorInvalidAttentuationBasedOnMinMax = (-200321)
DAQmxErrorCabledModuleCannotRouteSSH = (-200320)
DAQmxErrorCabledModuleCannotRouteConvClk = (-200319)
DAQmxErrorInvalidExcitValForScaling = (-200318)
DAQmxErrorNoDevMemForScript = (-200317)
DAQmxErrorScriptDataUnderflow = (-200316)
DAQmxErrorNoDevMemForWaveform = (-200315)
DAQmxErrorStreamDCMBecameUnlocked = (-200314)
DAQmxErrorStreamDCMLock = (-200313)
DAQmxErrorWaveformNotInMem = (-200312)
DAQmxErrorWaveformWriteOutOfBounds = (-200311)
DAQmxErrorWaveformPreviouslyAllocated = (-200310)
DAQmxErrorSampClkTbMasterTbDivNotAppropriateForSampTbSrc = (-200309)
DAQmxErrorSampTbRateSampTbSrcMismatch = (-200308)
DAQmxErrorMasterTbRateMasterTbSrcMismatch = (-200307)
DAQmxErrorSampsPerChanTooBig = (-200306)
DAQmxErrorFinitePulseTrainNotPossible = (-200305)
DAQmxErrorExtMasterTimebaseRateNotSpecified = (-200304)
DAQmxErrorExtSampClkSrcNotSpecified = (-200303)
DAQmxErrorInputSignalSlowerThanMeasTime = (-200302)
DAQmxErrorCannotUpdatePulseGenProperty = (-200301)
DAQmxErrorInvalidTimingType = (-200300)
DAQmxErrorPropertyUnavailWhenUsingOnboardMemory = (-200297)
DAQmxErrorCannotWriteAfterStartWithOnboardMemory = (-200295)
DAQmxErrorNotEnoughSampsWrittenForInitialXferRqstCondition = (-200294)
DAQmxErrorNoMoreSpace = (-200293)
DAQmxErrorSamplesCanNotYetBeWritten = (-200292)
DAQmxErrorGenStoppedToPreventIntermediateBufferRegenOfOldSamples = (-200291)
DAQmxErrorGenStoppedToPreventRegenOfOldSamples = (-200290)
DAQmxErrorSamplesNoLongerWriteable = (-200289)
DAQmxErrorSamplesWillNeverBeGenerated = (-200288)
DAQmxErrorNegativeWriteSampleNumber = (-200287)
DAQmxErrorNoAcqStarted = (-200286)
DAQmxErrorSamplesNotYetAvailable = (-200284)
DAQmxErrorAcqStoppedToPreventIntermediateBufferOverflow = (-200283)
DAQmxErrorNoRefTrigConfigured = (-200282)
DAQmxErrorCannotReadRelativeToRefTrigUntilDone = (-200281)
DAQmxErrorSamplesNoLongerAvailable = (-200279)
DAQmxErrorSamplesWillNeverBeAvailable = (-200278)
DAQmxErrorNegativeReadSampleNumber = (-200277)
DAQmxErrorExternalSampClkAndRefClkThruSameTerm = (-200276)
DAQmxErrorExtSampClkRateTooLowForClkIn = (-200275)
DAQmxErrorExtSampClkRateTooHighForBackplane = (-200274)
DAQmxErrorSampClkRateAndDivCombo = (-200273)
DAQmxErrorSampClkRateTooLowForDivDown = (-200272)
DAQmxErrorProductOfAOMinAndGainTooSmall = (-200271)
DAQmxErrorInterpolationRateNotPossible = (-200270)
DAQmxErrorOffsetTooLarge = (-200269)
DAQmxErrorOffsetTooSmall = (-200268)
DAQmxErrorProductOfAOMaxAndGainTooLarge = (-200267)
DAQmxErrorMinAndMaxNotSymmetric = (-200266)
DAQmxErrorInvalidAnalogTrigSrc = (-200265)
DAQmxErrorTooManyChansForAnalogRefTrig = (-200264)
DAQmxErrorTooManyChansForAnalogPauseTrig = (-200263)
DAQmxErrorTrigWhenOnDemandSampTiming = (-200262)
DAQmxErrorInconsistentAnalogTrigSettings = (-200261)
DAQmxErrorMemMapDataXferModeSampTimingCombo = (-200260)
DAQmxErrorInvalidJumperedAttr = (-200259)
DAQmxErrorInvalidGainBasedOnMinMax = (-200258)
DAQmxErrorInconsistentExcit = (-200257)
DAQmxErrorTopologyNotSupportedByCfgTermBlock = (-200256)
DAQmxErrorBuiltInTempSensorNotSupported = (-200255)
DAQmxErrorInvalidTerm = (-200254)
DAQmxErrorCannotTristateTerm = (-200253)
DAQmxErrorCannotTristateBusyTerm = (-200252)
DAQmxErrorNoDMAChansAvailable = (-200251)
DAQmxErrorInvalidWaveformLengthWithinLoopInScript = (-200250)
DAQmxErrorInvalidSubsetLengthWithinLoopInScript = (-200249)
DAQmxErrorMarkerPosInvalidForLoopInScript = (-200248)
DAQmxErrorIntegerExpectedInScript = (-200247)
DAQmxErrorPLLBecameUnlocked = (-200246)
DAQmxErrorPLLLock = (-200245)
DAQmxErrorDDCClkOutDCMBecameUnlocked = (-200244)
DAQmxErrorDDCClkOutDCMLock = (-200243)
DAQmxErrorClkDoublerDCMBecameUnlocked = (-200242)
DAQmxErrorClkDoublerDCMLock = (-200241)
DAQmxErrorSampClkDCMBecameUnlocked = (-200240)
DAQmxErrorSampClkDCMLock = (-200239)
DAQmxErrorSampClkTimebaseDCMBecameUnlocked = (-200238)
DAQmxErrorSampClkTimebaseDCMLock = (-200237)
DAQmxErrorAttrCannotBeReset = (-200236)
DAQmxErrorExplanationNotFound = (-200235)
DAQmxErrorWriteBufferTooSmall = (-200234)
DAQmxErrorSpecifiedAttrNotValid = (-200233)
DAQmxErrorAttrCannotBeRead = (-200232)
DAQmxErrorAttrCannotBeSet = (-200231)
DAQmxErrorNULLPtrForC_Api = (-200230)
DAQmxErrorReadBufferTooSmall = (-200229)
DAQmxErrorBufferTooSmallForString = (-200228)
DAQmxErrorNoAvailTrigLinesOnDevice = (-200227)
DAQmxErrorTrigBusLineNotAvail = (-200226)
DAQmxErrorCouldNotReserveRequestedTrigLine = (-200225)
DAQmxErrorTrigLineNotFound = (-200224)
DAQmxErrorSCXI1126ThreshHystCombination = (-200223)
DAQmxErrorAcqStoppedToPreventInputBufferOverwrite = (-200222)
DAQmxErrorTimeoutExceeded = (-200221)
DAQmxErrorInvalidDeviceID = (-200220)
DAQmxErrorInvalidAOChanOrder = (-200219)
DAQmxErrorSampleTimingTypeAndDataXferMode = (-200218)
DAQmxErrorBufferWithOnDemandSampTiming = (-200217)
DAQmxErrorBufferAndDataXferMode = (-200216)
DAQmxErrorMemMapAndBuffer = (-200215)
DAQmxErrorNoAnalogTrigHW = (-200214)
DAQmxErrorTooManyPretrigPlusMinPostTrigSamps = (-200213)
DAQmxErrorInconsistentUnitsSpecified = (-200212)
DAQmxErrorMultipleRelaysForSingleRelayOp = (-200211)
DAQmxErrorMultipleDevIDsPerChassisSpecifiedInList = (-200210)
DAQmxErrorDuplicateDevIDInList = (-200209)
DAQmxErrorInvalidRangeStatementCharInList = (-200208)
DAQmxErrorInvalidDeviceIDInList = (-200207)
DAQmxErrorTriggerPolarityConflict = (-200206)
DAQmxErrorCannotScanWithCurrentTopology = (-200205)
DAQmxErrorUnexpectedIdentifierInFullySpecifiedPathInList = (-200204)
DAQmxErrorSwitchCannotDriveMultipleTrigLines = (-200203)
DAQmxErrorInvalidRelayName = (-200202)
DAQmxErrorSwitchScanlistTooBig = (-200201)
DAQmxErrorSwitchChanInUse = (-200200)
DAQmxErrorSwitchNotResetBeforeScan = (-200199)
DAQmxErrorInvalidTopology = (-200198)
DAQmxErrorAttrNotSupported = (-200197)
DAQmxErrorUnexpectedEndOfActionsInList = (-200196)
DAQmxErrorPowerLimitExceeded = (-200195)
DAQmxErrorHWUnexpectedlyPoweredOffAndOn = (-200194)
DAQmxErrorSwitchOperationNotSupported = (-200193)
DAQmxErrorOnlyContinuousScanSupported = (-200192)
DAQmxErrorSwitchDifferentTopologyWhenScanning = (-200191)
DAQmxErrorDisconnectPathNotSameAsExistingPath = (-200190)
DAQmxErrorConnectionNotPermittedOnChanReservedForRouting = (-200189)
DAQmxErrorCannotConnectSrcChans = (-200188)
DAQmxErrorCannotConnectChannelToItself = (-200187)
DAQmxErrorChannelNotReservedForRouting = (-200186)
DAQmxErrorCannotConnectChansDirectly = (-200185)
DAQmxErrorChansAlreadyConnected = (-200184)
DAQmxErrorChanDuplicatedInPath = (-200183)
DAQmxErrorNoPathToDisconnect = (-200182)
DAQmxErrorInvalidSwitchChan = (-200181)
DAQmxErrorNoPathAvailableBetween2SwitchChans = (-200180)
DAQmxErrorExplicitConnectionExists = (-200179)
DAQmxErrorSwitchDifferentSettlingTimeWhenScanning = (-200178)
DAQmxErrorOperationOnlyPermittedWhileScanning = (-200177)
DAQmxErrorOperationNotPermittedWhileScanning = (-200176)
DAQmxErrorHardwareNotResponding = (-200175)
DAQmxErrorInvalidSampAndMasterTimebaseRateCombo = (-200173)
DAQmxErrorNonZeroBufferSizeInProgIOXfer = (-200172)
DAQmxErrorVirtualChanNameUsed = (-200171)
DAQmxErrorPhysicalChanDoesNotExist = (-200170)
DAQmxErrorMemMapOnlyForProgIOXfer = (-200169)
DAQmxErrorTooManyChans = (-200168)
DAQmxErrorCannotHaveCJTempWithOtherChans = (-200167)
DAQmxErrorOutputBufferUnderwrite = (-200166)
DAQmxErrorSensorInvalidCompletionResistance = (-200163)
DAQmxErrorVoltageExcitIncompatibleWith2WireCfg = (-200162)
DAQmxErrorIntExcitSrcNotAvailable = (-200161)
DAQmxErrorCannotCreateChannelAfterTaskVerified = (-200160)
DAQmxErrorLinesReservedForSCXIControl = (-200159)
DAQmxErrorCouldNotReserveLinesForSCXIControl = (-200158)
DAQmxErrorCalibrationFailed = (-200157)
DAQmxErrorReferenceFrequencyInvalid = (-200156)
DAQmxErrorReferenceResistanceInvalid = (-200155)
DAQmxErrorReferenceCurrentInvalid = (-200154)
DAQmxErrorReferenceVoltageInvalid = (-200153)
DAQmxErrorEEPROMDataInvalid = (-200152)
DAQmxErrorCabledModuleNotCapableOfRoutingAI = (-200151)
DAQmxErrorChannelNotAvailableInParallelMode = (-200150)
DAQmxErrorExternalTimebaseRateNotKnownForDelay = (-200149)
DAQmxErrorFREQOUTCannotProduceDesiredFrequency = (-200148)
DAQmxErrorMultipleCounterInputTask = (-200147)
DAQmxErrorCounterStartPauseTriggerConflict = (-200146)
DAQmxErrorCounterInputPauseTriggerAndSampleClockInvalid = (-200145)
DAQmxErrorCounterOutputPauseTriggerInvalid = (-200144)
DAQmxErrorCounterTimebaseRateNotSpecified = (-200143)
DAQmxErrorCounterTimebaseRateNotFound = (-200142)
DAQmxErrorCounterOverflow = (-200141)
DAQmxErrorCounterNoTimebaseEdgesBetweenGates = (-200140)
DAQmxErrorCounterMaxMinRangeFreq = (-200139)
DAQmxErrorCounterMaxMinRangeTime = (-200138)
DAQmxErrorSuitableTimebaseNotFoundTimeCombo = (-200137)
DAQmxErrorSuitableTimebaseNotFoundFrequencyCombo = (-200136)
DAQmxErrorInternalTimebaseSourceDivisorCombo = (-200135)
DAQmxErrorInternalTimebaseSourceRateCombo = (-200134)
DAQmxErrorInternalTimebaseRateDivisorSourceCombo = (-200133)
DAQmxErrorExternalTimebaseRateNotknownForRate = (-200132)
DAQmxErrorAnalogTrigChanNotFirstInScanList = (-200131)
DAQmxErrorNoDivisorForExternalSignal = (-200130)
DAQmxErrorAttributeInconsistentAcrossRepeatedPhysicalChannels = (-200128)
DAQmxErrorCannotHandshakeWithPort0 = (-200127)
DAQmxErrorControlLineConflictOnPortC = (-200126)
DAQmxErrorLines4To7ConfiguredForOutput = (-200125)
DAQmxErrorLines4To7ConfiguredForInput = (-200124)
DAQmxErrorLines0To3ConfiguredForOutput = (-200123)
DAQmxErrorLines0To3ConfiguredForInput = (-200122)
DAQmxErrorPortConfiguredForOutput = (-200121)
DAQmxErrorPortConfiguredForInput = (-200120)
DAQmxErrorPortConfiguredForStaticDigitalOps = (-200119)
DAQmxErrorPortReservedForHandshaking = (-200118)
DAQmxErrorPortDoesNotSupportHandshakingDataIO = (-200117)
DAQmxErrorCannotTristate8255OutputLines = (-200116)
DAQmxErrorTemperatureOutOfRangeForCalibration = (-200113)
DAQmxErrorCalibrationHandleInvalid = (-200112)
DAQmxErrorPasswordRequired = (-200111)
DAQmxErrorIncorrectPassword = (-200110)
DAQmxErrorPasswordTooLong = (-200109)
DAQmxErrorCalibrationSessionAlreadyOpen = (-200108)
DAQmxErrorSCXIModuleIncorrect = (-200107)
DAQmxErrorAttributeInconsistentAcrossChannelsOnDevice = (-200106)
DAQmxErrorSCXI1122ResistanceChanNotSupportedForCfg = (-200105)
DAQmxErrorBracketPairingMismatchInList = (-200104)
DAQmxErrorInconsistentNumSamplesToWrite = (-200103)
DAQmxErrorIncorrectDigitalPattern = (-200102)
DAQmxErrorIncorrectNumChannelsToWrite = (-200101)
DAQmxErrorIncorrectReadFunction = (-200100)
DAQmxErrorPhysicalChannelNotSpecified = (-200099)
DAQmxErrorMoreThanOneTerminal = (-200098)
DAQmxErrorMoreThanOneActiveChannelSpecified = (-200097)
DAQmxErrorInvalidNumberSamplesToRead = (-200096)
DAQmxErrorAnalogWaveformExpected = (-200095)
DAQmxErrorDigitalWaveformExpected = (-200094)
DAQmxErrorActiveChannelNotSpecified = (-200093)
DAQmxErrorFunctionNotSupportedForDeviceTasks = (-200092)
DAQmxErrorFunctionNotInLibrary = (-200091)
DAQmxErrorLibraryNotPresent = (-200090)
DAQmxErrorDuplicateTask = (-200089)
DAQmxErrorInvalidTask = (-200088)
DAQmxErrorInvalidChannel = (-200087)
DAQmxErrorInvalidSyntaxForPhysicalChannelRange = (-200086)
DAQmxErrorMinNotLessThanMax = (-200082)
DAQmxErrorSampleRateNumChansConvertPeriodCombo = (-200081)
DAQmxErrorAODuringCounter1DMAConflict = (-200079)
DAQmxErrorAIDuringCounter0DMAConflict = (-200078)
DAQmxErrorInvalidAttributeValue = (-200077)
DAQmxErrorSuppliedCurrentDataOutsideSpecifiedRange = (-200076)
DAQmxErrorSuppliedVoltageDataOutsideSpecifiedRange = (-200075)
DAQmxErrorCannotStoreCalConst = (-200074)
DAQmxErrorSCXIModuleNotFound = (-200073)
DAQmxErrorDuplicatePhysicalChansNotSupported = (-200072)
DAQmxErrorTooManyPhysicalChansInList = (-200071)
DAQmxErrorInvalidAdvanceEventTriggerType = (-200070)
DAQmxErrorDeviceIsNotAValidSwitch = (-200069)
DAQmxErrorDeviceDoesNotSupportScanning = (-200068)
DAQmxErrorScanListCannotBeTimed = (-200067)
DAQmxErrorConnectOperatorInvalidAtPointInList = (-200066)
DAQmxErrorUnexpectedSwitchActionInList = (-200065)
DAQmxErrorUnexpectedSeparatorInList = (-200064)
DAQmxErrorExpectedTerminatorInList = (-200063)
DAQmxErrorExpectedConnectOperatorInList = (-200062)
DAQmxErrorExpectedSeparatorInList = (-200061)
DAQmxErrorFullySpecifiedPathInListContainsRange = (-200060)
DAQmxErrorConnectionSeparatorAtEndOfList = (-200059)
DAQmxErrorIdentifierInListTooLong = (-200058)
DAQmxErrorDuplicateDeviceIDInListWhenSettling = (-200057)
DAQmxErrorChannelNameNotSpecifiedInList = (-200056)
DAQmxErrorDeviceIDNotSpecifiedInList = (-200055)
DAQmxErrorSemicolonDoesNotFollowRangeInList = (-200054)
DAQmxErrorSwitchActionInListSpansMultipleDevices = (-200053)
DAQmxErrorRangeWithoutAConnectActionInList = (-200052)
DAQmxErrorInvalidIdentifierFollowingSeparatorInList = (-200051)
DAQmxErrorInvalidChannelNameInList = (-200050)
DAQmxErrorInvalidNumberInRepeatStatementInList = (-200049)
DAQmxErrorInvalidTriggerLineInList = (-200048)
DAQmxErrorInvalidIdentifierInListFollowingDeviceID = (-200047)
DAQmxErrorInvalidIdentifierInListAtEndOfSwitchAction = (-200046)
DAQmxErrorDeviceRemoved = (-200045)
DAQmxErrorRoutingPathNotAvailable = (-200044)
DAQmxErrorRoutingHardwareBusy = (-200043)
DAQmxErrorRequestedSignalInversionForRoutingNotPossible = (-200042)
DAQmxErrorInvalidRoutingDestinationTerminalName = (-200041)
DAQmxErrorInvalidRoutingSourceTerminalName = (-200040)
DAQmxErrorRoutingNotSupportedForDevice = (-200039)
DAQmxErrorWaitIsLastInstructionOfLoopInScript = (-200038)
DAQmxErrorClearIsLastInstructionOfLoopInScript = (-200037)
DAQmxErrorInvalidLoopIterationsInScript = (-200036)
DAQmxErrorRepeatLoopNestingTooDeepInScript = (-200035)
DAQmxErrorMarkerPositionOutsideSubsetInScript = (-200034)
DAQmxErrorSubsetStartOffsetNotAlignedInScript = (-200033)
DAQmxErrorInvalidSubsetLengthInScript = (-200032)
DAQmxErrorMarkerPositionNotAlignedInScript = (-200031)
DAQmxErrorSubsetOutsideWaveformInScript = (-200030)
DAQmxErrorMarkerOutsideWaveformInScript = (-200029)
DAQmxErrorWaveformInScriptNotInMem = (-200028)
DAQmxErrorKeywordExpectedInScript = (-200027)
DAQmxErrorBufferNameExpectedInScript = (-200026)
DAQmxErrorProcedureNameExpectedInScript = (-200025)
DAQmxErrorScriptHasInvalidIdentifier = (-200024)
DAQmxErrorScriptHasInvalidCharacter = (-200023)
DAQmxErrorResourceAlreadyReserved = (-200022)
DAQmxErrorSelfTestFailed = (-200020)
DAQmxErrorADCOverrun = (-200019)
DAQmxErrorDACUnderflow = (-200018)
DAQmxErrorInputFIFOUnderflow = (-200017)
DAQmxErrorOutputFIFOUnderflow = (-200016)
DAQmxErrorSCXISerialCommunication = (-200015)
DAQmxErrorDigitalTerminalSpecifiedMoreThanOnce = (-200014)
DAQmxErrorDigitalOutputNotSupported = (-200012)
DAQmxErrorInconsistentChannelDirections = (-200011)
DAQmxErrorInputFIFOOverflow = (-200010)
DAQmxErrorTimeStampOverwritten = (-200009)
DAQmxErrorStopTriggerHasNotOccurred = (-200008)
DAQmxErrorRecordNotAvailable = (-200007)
DAQmxErrorRecordOverwritten = (-200006)
DAQmxErrorDataNotAvailable = (-200005)
DAQmxErrorDataOverwrittenInDeviceMemory = (-200004)
DAQmxErrorDuplicatedChannel = (-200003)
DAQmxWarningTimestampCounterRolledOver = (200003)
DAQmxWarningInputTerminationOverloaded = (200004)
DAQmxWarningADCOverloaded = (200005)
DAQmxWarningPLLUnlocked = (200007)
DAQmxWarningCounter0DMADuringAIConflict = (200008)
DAQmxWarningCounter1DMADuringAOConflict = (200009)
DAQmxWarningStoppedBeforeDone = (200010)
DAQmxWarningRateViolatesSettlingTime = (200011)
DAQmxWarningRateViolatesMaxADCRate = (200012)
DAQmxWarningUserDefInfoStringTooLong = (200013)
DAQmxWarningTooManyInterruptsPerSecond = (200014)
DAQmxWarningPotentialGlitchDuringWrite = (200015)
DAQmxWarningDevNotSelfCalibratedWithDAQmx = (200016)
DAQmxWarningAISampRateTooLow = (200017)
DAQmxWarningAIConvRateTooLow = (200018)
DAQmxWarningReadOffsetCoercion = (200019)
DAQmxWarningPretrigCoercion = (200020)
DAQmxWarningSampValCoercedToMax = (200021)
DAQmxWarningSampValCoercedToMin = (200022)
DAQmxWarningPropertyVersionNew = (200024)
DAQmxWarningUserDefinedInfoTooLong = (200025)
DAQmxWarningCAPIStringTruncatedToFitBuffer = (200026)
DAQmxWarningSampClkRateTooLow = (200027)
DAQmxWarningPossiblyInvalidCTRSampsInFiniteDMAAcq = (200028)
DAQmxWarningRISAcqCompletedSomeBinsNotFilled = (200029)
DAQmxWarningPXIDevTempExceedsMaxOpTemp = (200030)
DAQmxWarningOutputGainTooLowForRFFreq = (200031)
DAQmxWarningOutputGainTooHighForRFFreq = (200032)
DAQmxWarningMultipleWritesBetweenSampClks = (200033)
DAQmxWarningDeviceMayShutDownDueToHighTemp = (200034)
DAQmxWarningRateViolatesMinADCRate = (200035)
DAQmxWarningSampClkRateAboveDevSpecs = (200036)
DAQmxWarningCOPrevDAQmxWriteSettingsOverwrittenForHWTimedSinglePoint = (200037)
DAQmxWarningLowpassFilterSettlingTimeExceedsUserTimeBetween2ADCConversions = (200038)
DAQmxWarningLowpassFilterSettlingTimeExceedsDriverTimeBetween2ADCConversions = (200039)
DAQmxWarningSampClkRateViolatesSettlingTimeForGen = (200040)
DAQmxWarningInvalidCalConstValueForAI = (200041)
DAQmxWarningInvalidCalConstValueForAO = (200042)
DAQmxWarningChanCalExpired = (200043)
DAQmxWarningUnrecognizedEnumValueEncounteredInStorage = (200044)
DAQmxWarningTableCRCNotCorrect = (200045)
DAQmxWarningExternalCRCNotCorrect = (200046)
DAQmxWarningSelfCalCRCNotCorrect = (200047)
DAQmxWarningDeviceSpecExceeded = (200048)
DAQmxWarningOnlyGainCalibrated = (200049)
DAQmxWarningReversePowerProtectionActivated = (200050)
DAQmxWarningOverVoltageProtectionActivated = (200051)
DAQmxWarningBufferSizeNotMultipleOfSectorSize = (200052)
DAQmxWarningSampleRateMayCauseAcqToFail = (200053)
DAQmxWarningUserAreaCRCNotCorrect = (200054)
DAQmxWarningPowerUpInfoCRCNotCorrect = (200055)
DAQmxWarningConnectionCountHasExceededRecommendedLimit = (200056)
DAQmxWarningNetworkDeviceAlreadyAdded = (200057)
DAQmxWarningAccessoryConnectionCountIsInvalid = (200058)
DAQmxWarningUnableToDisconnectPorts = (200059)
DAQmxWarningReadRepeatedData = (200060)
DAQmxWarningPXI5600_NotConfigured = (200061)
DAQmxWarningPXI5661_IncorrectlyConfigured = (200062)
DAQmxWarningPXIe5601_NotConfigured = (200063)
DAQmxWarningPXIe5663_IncorrectlyConfigured = (200064)
DAQmxWarningPXIe5663E_IncorrectlyConfigured = (200065)
DAQmxWarningPXIe5603_NotConfigured = (200066)
DAQmxWarningPXIe5665_5603_IncorrectlyConfigured = (200067)
DAQmxWarningPXIe5667_5603_IncorrectlyConfigured = (200068)
DAQmxWarningPXIe5605_NotConfigured = (200069)
DAQmxWarningPXIe5665_5605_IncorrectlyConfigured = (200070)
DAQmxWarningPXIe5667_5605_IncorrectlyConfigured = (200071)
DAQmxWarningPXIe5606_NotConfigured = (200072)
DAQmxWarningPXIe5665_5606_IncorrectlyConfigured = (200073)
DAQmxWarningPXI5610_NotConfigured = (200074)
DAQmxWarningPXI5610_IncorrectlyConfigured = (200075)
DAQmxWarningPXIe5611_NotConfigured = (200076)
DAQmxWarningPXIe5611_IncorrectlyConfigured = (200077)
DAQmxWarningUSBHotfixForDAQ = (200078)
DAQmxWarningReadNotCompleteBeforeSampClk = (209800)
DAQmxWarningWriteNotCompleteBeforeSampClk = (209801)
DAQmxWarningWaitForNextSampClkDetectedMissedSampClk = (209802)
DAQmxErrorRoutingDestTermPXIDStarXNotInSystemTimingSlot_Routing = (-89167)
DAQmxErrorRoutingSrcTermPXIDStarXNotInSystemTimingSlot_Routing = (-89166)
DAQmxErrorRoutingSrcTermPXIDStarInNonDStarTriggerSlot_Routing = (-89165)
DAQmxErrorRoutingDestTermPXIDStarInNonDStarTriggerSlot_Routing = (-89164)
DAQmxErrorRoutingDestTermPXIClk10InNotInStarTriggerSlot_Routing = (-89162)
DAQmxErrorRoutingDestTermPXIClk10InNotInSystemTimingSlot_Routing = (-89161)
DAQmxErrorRoutingDestTermPXIStarXNotInStarTriggerSlot_Routing = (-89160)
DAQmxErrorRoutingDestTermPXIStarXNotInSystemTimingSlot_Routing = (-89159)
DAQmxErrorRoutingSrcTermPXIStarXNotInStarTriggerSlot_Routing = (-89158)
DAQmxErrorRoutingSrcTermPXIStarXNotInSystemTimingSlot_Routing = (-89157)
DAQmxErrorRoutingSrcTermPXIStarInNonStarTriggerSlot_Routing = (-89156)
DAQmxErrorRoutingDestTermPXIStarInNonStarTriggerSlot_Routing = (-89155)
DAQmxErrorRoutingDestTermPXIStarInStarTriggerSlot_Routing = (-89154)
DAQmxErrorRoutingDestTermPXIStarInSystemTimingSlot_Routing = (-89153)
DAQmxErrorRoutingSrcTermPXIStarInStarTriggerSlot_Routing = (-89152)
DAQmxErrorRoutingSrcTermPXIStarInSystemTimingSlot_Routing = (-89151)
DAQmxErrorInvalidSignalModifier_Routing = (-89150)
DAQmxErrorRoutingDestTermPXIClk10InNotInSlot2_Routing = (-89149)
DAQmxErrorRoutingDestTermPXIStarXNotInSlot2_Routing = (-89148)
DAQmxErrorRoutingSrcTermPXIStarXNotInSlot2_Routing = (-89147)
DAQmxErrorRoutingSrcTermPXIStarInSlot16AndAbove_Routing = (-89146)
DAQmxErrorRoutingDestTermPXIStarInSlot16AndAbove_Routing = (-89145)
DAQmxErrorRoutingDestTermPXIStarInSlot2_Routing = (-89144)
DAQmxErrorRoutingSrcTermPXIStarInSlot2_Routing = (-89143)
DAQmxErrorRoutingDestTermPXIChassisNotIdentified_Routing = (-89142)
DAQmxErrorRoutingSrcTermPXIChassisNotIdentified_Routing = (-89141)
DAQmxErrorTrigLineNotFoundSingleDevRoute_Routing = (-89140)
DAQmxErrorNoCommonTrigLineForRoute_Routing = (-89139)
DAQmxErrorResourcesInUseForRouteInTask_Routing = (-89138)
DAQmxErrorResourcesInUseForRoute_Routing = (-89137)
DAQmxErrorRouteNotSupportedByHW_Routing = (-89136)
DAQmxErrorResourcesInUseForInversionInTask_Routing = (-89135)
DAQmxErrorResourcesInUseForInversion_Routing = (-89134)
DAQmxErrorInversionNotSupportedByHW_Routing = (-89133)
DAQmxErrorResourcesInUseForProperty_Routing = (-89132)
DAQmxErrorRouteSrcAndDestSame_Routing = (-89131)
DAQmxErrorDevAbsentOrUnavailable_Routing = (-89130)
DAQmxErrorInvalidTerm_Routing = (-89129)
DAQmxErrorCannotTristateTerm_Routing = (-89128)
DAQmxErrorCannotTristateBusyTerm_Routing = (-89127)
DAQmxErrorCouldNotReserveRequestedTrigLine_Routing = (-89126)
DAQmxErrorTrigLineNotFound_Routing = (-89125)
DAQmxErrorRoutingPathNotAvailable_Routing = (-89124)
DAQmxErrorRoutingHardwareBusy_Routing = (-89123)
DAQmxErrorRequestedSignalInversionForRoutingNotPossible_Routing = (-89122)
DAQmxErrorInvalidRoutingDestinationTerminalName_Routing = (-89121)
DAQmxErrorInvalidRoutingSourceTerminalName_Routing = (-89120)
DAQmxErrorServiceLocatorNotAvailable_Routing = (-88907)
DAQmxErrorCouldNotConnectToServer_Routing = (-88900)
DAQmxErrorDeviceNameContainsSpacesOrPunctuation_Routing = (-88720)
DAQmxErrorDeviceNameContainsNonprintableCharacters_Routing = (-88719)
DAQmxErrorDeviceNameIsEmpty_Routing = (-88718)
DAQmxErrorDeviceNameNotFound_Routing = (-88717)
DAQmxErrorLocalRemoteDriverVersionMismatch_Routing = (-88716)
DAQmxErrorDuplicateDeviceName_Routing = (-88715)
DAQmxErrorRuntimeAborting_Routing = (-88710)
DAQmxErrorRuntimeAborted_Routing = (-88709)
DAQmxErrorResourceNotInPool_Routing = (-88708)
DAQmxErrorDriverDeviceGUIDNotFound_Routing = (-88705)
DAQmxErrorPALUSBTransactionError = (-50808)
DAQmxErrorPALIsocStreamBufferError = (-50807)
DAQmxErrorPALInvalidAddressComponent = (-50806)
DAQmxErrorPALSharingViolation = (-50805)
DAQmxErrorPALInvalidDeviceState = (-50804)
DAQmxErrorPALConnectionReset = (-50803)
DAQmxErrorPALConnectionAborted = (-50802)
DAQmxErrorPALConnectionRefused = (-50801)
DAQmxErrorPALBusResetOccurred = (-50800)
DAQmxErrorPALWaitInterrupted = (-50700)
DAQmxErrorPALMessageUnderflow = (-50651)
DAQmxErrorPALMessageOverflow = (-50650)
DAQmxErrorPALThreadAlreadyDead = (-50604)
DAQmxErrorPALThreadStackSizeNotSupported = (-50603)
DAQmxErrorPALThreadControllerIsNotThreadCreator = (-50602)
DAQmxErrorPALThreadHasNoThreadObject = (-50601)
DAQmxErrorPALThreadCouldNotRun = (-50600)
DAQmxErrorPALSyncAbandoned = (-50551)
DAQmxErrorPALSyncTimedOut = (-50550)
DAQmxErrorPALReceiverSocketInvalid = (-50503)
DAQmxErrorPALSocketListenerInvalid = (-50502)
DAQmxErrorPALSocketListenerAlreadyRegistered = (-50501)
DAQmxErrorPALDispatcherAlreadyExported = (-50500)
DAQmxErrorPALDMALinkEventMissed = (-50450)
DAQmxErrorPALBusError = (-50413)
DAQmxErrorPALRetryLimitExceeded = (-50412)
DAQmxErrorPALTransferOverread = (-50411)
DAQmxErrorPALTransferOverwritten = (-50410)
DAQmxErrorPALPhysicalBufferFull = (-50409)
DAQmxErrorPALPhysicalBufferEmpty = (-50408)
DAQmxErrorPALLogicalBufferFull = (-50407)
DAQmxErrorPALLogicalBufferEmpty = (-50406)
DAQmxErrorPALTransferAborted = (-50405)
DAQmxErrorPALTransferStopped = (-50404)
DAQmxErrorPALTransferInProgress = (-50403)
DAQmxErrorPALTransferNotInProgress = (-50402)
DAQmxErrorPALCommunicationsFault = (-50401)
DAQmxErrorPALTransferTimedOut = (-50400)
DAQmxErrorPALMemoryHeapNotEmpty = (-50355)
DAQmxErrorPALMemoryBlockCheckFailed = (-50354)
DAQmxErrorPALMemoryPageLockFailed = (-50353)
DAQmxErrorPALMemoryFull = (-50352)
DAQmxErrorPALMemoryAlignmentFault = (-50351)
DAQmxErrorPALMemoryConfigurationFault = (-50350)
DAQmxErrorPALDeviceInitializationFault = (-50303)
DAQmxErrorPALDeviceNotSupported = (-50302)
DAQmxErrorPALDeviceUnknown = (-50301)
DAQmxErrorPALDeviceNotFound = (-50300)
DAQmxErrorPALFeatureDisabled = (-50265)
DAQmxErrorPALComponentBusy = (-50264)
DAQmxErrorPALComponentAlreadyInstalled = (-50263)
DAQmxErrorPALComponentNotUnloadable = (-50262)
DAQmxErrorPALComponentNeverLoaded = (-50261)
DAQmxErrorPALComponentAlreadyLoaded = (-50260)
DAQmxErrorPALComponentCircularDependency = (-50259)
DAQmxErrorPALComponentInitializationFault = (-50258)
DAQmxErrorPALComponentImageCorrupt = (-50257)
DAQmxErrorPALFeatureNotSupported = (-50256)
DAQmxErrorPALFunctionNotFound = (-50255)
DAQmxErrorPALFunctionObsolete = (-50254)
DAQmxErrorPALComponentTooNew = (-50253)
DAQmxErrorPALComponentTooOld = (-50252)
DAQmxErrorPALComponentNotFound = (-50251)
DAQmxErrorPALVersionMismatch = (-50250)
DAQmxErrorPALFileFault = (-50209)
DAQmxErrorPALFileWriteFault = (-50208)
DAQmxErrorPALFileReadFault = (-50207)
DAQmxErrorPALFileSeekFault = (-50206)
DAQmxErrorPALFileCloseFault = (-50205)
DAQmxErrorPALFileOpenFault = (-50204)
DAQmxErrorPALDiskFull = (-50203)
DAQmxErrorPALOSFault = (-50202)
DAQmxErrorPALOSInitializationFault = (-50201)
DAQmxErrorPALOSUnsupported = (-50200)
DAQmxErrorPALCalculationOverflow = (-50175)
DAQmxErrorPALHardwareFault = (-50152)
DAQmxErrorPALFirmwareFault = (-50151)
DAQmxErrorPALSoftwareFault = (-50150)
DAQmxErrorPALMessageQueueFull = (-50108)
DAQmxErrorPALResourceAmbiguous = (-50107)
DAQmxErrorPALResourceBusy = (-50106)
DAQmxErrorPALResourceInitialized = (-50105)
DAQmxErrorPALResourceNotInitialized = (-50104)
DAQmxErrorPALResourceReserved = (-50103)
DAQmxErrorPALResourceNotReserved = (-50102)
DAQmxErrorPALResourceNotAvailable = (-50101)
DAQmxErrorPALResourceOwnedBySystem = (-50100)
DAQmxErrorPALBadToken = (-50020)
DAQmxErrorPALBadThreadMultitask = (-50019)
DAQmxErrorPALBadLibrarySpecifier = (-50018)
DAQmxErrorPALBadAddressSpace = (-50017)
DAQmxErrorPALBadWindowType = (-50016)
DAQmxErrorPALBadAddressClass = (-50015)
DAQmxErrorPALBadWriteCount = (-50014)
DAQmxErrorPALBadWriteOffset = (-50013)
DAQmxErrorPALBadWriteMode = (-50012)
DAQmxErrorPALBadReadCount = (-50011)
DAQmxErrorPALBadReadOffset = (-50010)
DAQmxErrorPALBadReadMode = (-50009)
DAQmxErrorPALBadCount = (-50008)
DAQmxErrorPALBadOffset = (-50007)
DAQmxErrorPALBadMode = (-50006)
DAQmxErrorPALBadDataSize = (-50005)
DAQmxErrorPALBadPointer = (-50004)
DAQmxErrorPALBadSelector = (-50003)
DAQmxErrorPALBadDevice = (-50002)
DAQmxErrorPALIrrelevantAttribute = (-50001)
DAQmxErrorPALValueConflict = (-50000)
DAQmxErrorRetryCall = (-26853)
DAQmxErrorFileDoesNotExist = (-26852)
DAQmxErrorGenerationDisabled = (-26851)
DAQmxErrorAlreadyInitialized = (-26850)
DAQmxErrorPXISystemDescriptionParseError = (-26809)
DAQmxErrorPXISAConfigurationLocked = (-26808)
DAQmxErrorNotActiveResourceManager = (-26807)
DAQmxErrorInvalidAttribute = (-26806)
DAQmxErrorInvalidHandle = (-26805)
DAQmxErrorServiceNotRunning = (-26804)
DAQmxErrorRecursiveCall = (-26803)
DAQmxErrorTimeout = (-26802)
DAQmxErrorUnspecifiedError = (-26801)
DAQmxErrorHandlerNotFound = (-26657)
DAQmxErrorIncorrectDataType = (-26656)
DAQmxErrorInconsistentFileFault = (-26655)
DAQmxErrorChildNotFound = (-26654)
DAQmxErrorAttributeNotFound = (-26653)
DAQmxErrorIOError = (-26652)
DAQmxErrorPxiResmanMxsPxiSystemNotFound = (-26600)
DAQmxErrorPxiResmanPciDescriptionStringParseError = (-26550)
DAQmxErrorPxiResmanModuleParseError = (-26500)
DAQmxErrorPxiResmanSystemDescriptionSpecLimitation = (-26403)
DAQmxErrorPxiResmanSystemDescriptionWriteError = (-26402)
DAQmxErrorPxiResmanSystemDescriptionParseError = (-26401)
DAQmxErrorPxiResmanSystemDescriptionFileNotFound = (-26400)
DAQmxErrorPxiResmanControllerParseError = (-26302)
DAQmxErrorPxiResmanControllerTypeInvalid = (-26301)
DAQmxErrorPxiResmanControllerFileNotFound = (-26300)
DAQmxErrorPxiResmanChassisTriggerParseError = (-26206)
DAQmxErrorPxiResmanChassisBridgeParseError = (-26205)
DAQmxErrorPxiResmanChassisSlotParseError = (-26204)
DAQmxErrorPxiResmanChassisSegmentParseError = (-26203)
DAQmxErrorPxiResmanChassisParseError = (-26202)
DAQmxErrorPxiResmanChassisBridgeNotFound = (-26201)
DAQmxErrorPxiResmanChassisFileNotFound = (-26200)
DAQmxErrorInsufficientBuffer = (-26113)
DAQmxErrorDirCreateFault = (-26112)
DAQmxErrorAccessDenied = (-26111)
DAQmxErrorPathNotFound = (-26110)
DAQmxErrorFileExists = (-26109)
DAQmxErrorFileDeleteFault = (-26107)
DAQmxErrorFileCopyFault = (-26106)
DAQmxErrorFileCloseFault = (-26105)
DAQmxErrorFileNotFound = (-26103)
DAQmxErrorPxiResmanSystemNotInitialized = (-26102)
DAQmxErrorPxiResmanInvalidConfiguration = (-26101)
DAQmxErrorPxiResmanAllocationError = (-26100)
DAQmxErrorTrigmanDisconnected = (-26010)
DAQmxErrorTrigmanInvalidClient = (-26009)
DAQmxErrorTrigmanConflictingRoute = (-26008)
DAQmxErrorTrigmanConflictingRouteNoUnreserve = (-26007)
DAQmxErrorTrigmanConflictingRouteGeneric = (-26006)
DAQmxErrorTrigmanLineAlreadyReserved = (-26005)
DAQmxErrorTrigmanLineNotReservedUnroutable = (-26004)
DAQmxErrorTrigmanLineNotReserved = (-26003)
DAQmxErrorTrigmanInvalidParameter = (-26002)
DAQmxErrorTrigmanUnsupportedOper = (-26001)
DAQmxErrorTrigmanUnknownError = (-26000)
DAQmxWarningTrigmanUnknownError = (26000)
DAQmxWarningTrigmanUnsupportedOper = (26001)
DAQmxWarningTrigmanInvalidParameter = (26002)
DAQmxWarningTrigmanLineNotReserved = (26003)
DAQmxWarningTrigmanLineNotReservedUnroutable = (26004)
DAQmxWarningTrigmanLineAlreadyReserved = (26005)
DAQmxWarningTrigmanConflictingRouteGeneric = (26006)
DAQmxWarningTrigmanConflictingRouteNoUnreserve = (26007)
DAQmxWarningTrigmanConflictingRoute = (26008)
DAQmxWarningTrigmanInvalidClient = (26009)
DAQmxWarningTrigmanDisconnected = (26010)
DAQmxWarningPxiResmanAllocationError = (26100)
DAQmxWarningPxiResmanInvalidConfiguration = (26101)
DAQmxWarningPxiResmanSystemNotInitialized = (26102)
DAQmxWarningFileNotFound = (26103)
DAQmxWarningFileCloseFault = (26105)
DAQmxWarningFileCopyFault = (26106)
DAQmxWarningFileDeleteFault = (26107)
DAQmxWarningFileExists = (26109)
DAQmxWarningPathNotFound = (26110)
DAQmxWarningAccessDenied = (26111)
DAQmxWarningDirCreateFault = (26112)
DAQmxWarningInsufficientBuffer = (26113)
DAQmxWarningPxiResmanChassisFileNotFound = (26200)
DAQmxWarningPxiResmanChassisBridgeNotFound = (26201)
DAQmxWarningPxiResmanChassisParseError = (26202)
DAQmxWarningPxiResmanChassisSegmentParseError = (26203)
DAQmxWarningPxiResmanChassisSlotParseError = (26204)
DAQmxWarningPxiResmanChassisBridgeParseError = (26205)
DAQmxWarningPxiResmanChassisTriggerParseError = (26206)
DAQmxWarningPxiResmanControllerFileNotFound = (26300)
DAQmxWarningPxiResmanControllerTypeInvalid = (26301)
DAQmxWarningPxiResmanControllerParseError = (26302)
DAQmxWarningPxiResmanSystemDescriptionFileNotFound = (26400)
DAQmxWarningPxiResmanSystemDescriptionParseError = (26401)
DAQmxWarningPxiResmanSystemDescriptionWriteError = (26402)
DAQmxWarningPxiResmanSystemDescriptionSpecLimitation = (26403)
DAQmxWarningPxiResmanModuleParseError = (26500)
DAQmxWarningPxiResmanPciDescriptionStringParseError = (26550)
DAQmxWarningPxiResmanMxsPxiSystemNotFound = (26600)
DAQmxWarningIOError = (26652)
DAQmxWarningAttributeNotFound = (26653)
DAQmxWarningChildNotFound = (26654)
DAQmxWarningInconsistentFileFault = (26655)
DAQmxWarningIncorrectDataType = (26656)
DAQmxWarningHandlerNotFound = (26657)
DAQmxWarningUnspecifiedError = (26801)
DAQmxWarningTimeout = (26802)
DAQmxWarningRecursiveCall = (26803)
DAQmxWarningServiceNotRunning = (26804)
DAQmxWarningInvalidHandle = (26805)
DAQmxWarningInvalidAttribute = (26806)
DAQmxWarningNotActiveResourceManager = (26807)
DAQmxWarningPXISAConfigurationLocked = (26808)
DAQmxWarningPXISystemDescriptionParseError = (26809)
DAQmxWarningAlreadyInitialized = (26850)
DAQmxWarningGenerationDisabled = (26851)
DAQmxWarningFileDoesNotExist = (26852)
DAQmxWarningRetryCall = (26853)
DAQmxWarningPALValueConflict = (50000)
DAQmxWarningPALIrrelevantAttribute = (50001)
DAQmxWarningPALBadDevice = (50002)
DAQmxWarningPALBadSelector = (50003)
DAQmxWarningPALBadPointer = (50004)
DAQmxWarningPALBadDataSize = (50005)
DAQmxWarningPALBadMode = (50006)
DAQmxWarningPALBadOffset = (50007)
DAQmxWarningPALBadCount = (50008)
DAQmxWarningPALBadReadMode = (50009)
DAQmxWarningPALBadReadOffset = (50010)
DAQmxWarningPALBadReadCount = (50011)
DAQmxWarningPALBadWriteMode = (50012)
DAQmxWarningPALBadWriteOffset = (50013)
DAQmxWarningPALBadWriteCount = (50014)
DAQmxWarningPALBadAddressClass = (50015)
DAQmxWarningPALBadWindowType = (50016)
DAQmxWarningPALBadThreadMultitask = (50019)
DAQmxWarningPALResourceOwnedBySystem = (50100)
DAQmxWarningPALResourceNotAvailable = (50101)
DAQmxWarningPALResourceNotReserved = (50102)
DAQmxWarningPALResourceReserved = (50103)
DAQmxWarningPALResourceNotInitialized = (50104)
DAQmxWarningPALResourceInitialized = (50105)
DAQmxWarningPALResourceBusy = (50106)
DAQmxWarningPALResourceAmbiguous = (50107)
DAQmxWarningPALFirmwareFault = (50151)
DAQmxWarningPALHardwareFault = (50152)
DAQmxWarningPALOSUnsupported = (50200)
DAQmxWarningPALOSFault = (50202)
DAQmxWarningPALFunctionObsolete = (50254)
DAQmxWarningPALFunctionNotFound = (50255)
DAQmxWarningPALFeatureNotSupported = (50256)
DAQmxWarningPALComponentInitializationFault = (50258)
DAQmxWarningPALComponentAlreadyLoaded = (50260)
DAQmxWarningPALComponentNotUnloadable = (50262)
DAQmxWarningPALMemoryAlignmentFault = (50351)
DAQmxWarningPALMemoryHeapNotEmpty = (50355)
DAQmxWarningPALTransferNotInProgress = (50402)
DAQmxWarningPALTransferInProgress = (50403)
DAQmxWarningPALTransferStopped = (50404)
DAQmxWarningPALTransferAborted = (50405)
DAQmxWarningPALLogicalBufferEmpty = (50406)
DAQmxWarningPALLogicalBufferFull = (50407)
DAQmxWarningPALPhysicalBufferEmpty = (50408)
DAQmxWarningPALPhysicalBufferFull = (50409)
DAQmxWarningPALTransferOverwritten = (50410)
DAQmxWarningPALTransferOverread = (50411)
DAQmxWarningPALDispatcherAlreadyExported = (50500)
DAQmxWarningPALSyncAbandoned = (50551)
# functions


@catchDAQError
def DAQmxLoadTask(taskName, taskHandle):
    """int32 __CFUNC     DAQmxLoadTask                  (const char taskName[], TaskHandle \*taskHandle);"""
    global DAQlib
    DAQlib.DAQmxLoadTask.argtypes = (c_char_p, POINTER(TaskHandle),)
    return DAQlib.DAQmxLoadTask(taskName, taskHandle)


@catchDAQError
def DAQmxCreateTask(taskName, taskHandle):
    """int32 __CFUNC     DAQmxCreateTask                (const char taskName[], TaskHandle \*taskHandle);"""
    global DAQlib
    DAQlib.DAQmxCreateTask.argtypes = (c_char_p, POINTER(TaskHandle),)
    return DAQlib.DAQmxCreateTask(taskName, taskHandle)


@catchDAQError
def DAQmxAddGlobalChansToTask(taskHandle, channelNames):
    """int32 __CFUNC     DAQmxAddGlobalChansToTask      (TaskHandle taskHandle, const char channelNames[]);"""
    global DAQlib
    DAQlib.DAQmxAddGlobalChansToTask.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxAddGlobalChansToTask(taskHandle, channelNames)


@catchDAQError
def DAQmxStartTask(taskHandle):
    """int32 __CFUNC     DAQmxStartTask                 (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxStartTask.argtypes = (TaskHandle,)
    return DAQlib.DAQmxStartTask(taskHandle)


@catchDAQError
def DAQmxStopTask(taskHandle):
    """int32 __CFUNC     DAQmxStopTask                  (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxStopTask.argtypes = (TaskHandle,)
    return DAQlib.DAQmxStopTask(taskHandle)


@catchDAQError
def DAQmxClearTask(taskHandle):
    """int32 __CFUNC     DAQmxClearTask                 (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxClearTask.argtypes = (TaskHandle,)
    return DAQlib.DAQmxClearTask(taskHandle)


@catchDAQError
def DAQmxWaitUntilTaskDone(taskHandle, timeToWait):
    """int32 __CFUNC     DAQmxWaitUntilTaskDone         (TaskHandle taskHandle, float64 timeToWait);"""
    global DAQlib
    DAQlib.DAQmxWaitUntilTaskDone.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxWaitUntilTaskDone(taskHandle, timeToWait)


@catchDAQError
def DAQmxIsTaskDone(taskHandle, isTaskDone):
    """int32 __CFUNC     DAQmxIsTaskDone                (TaskHandle taskHandle, bool32 \*isTaskDone);"""
    global DAQlib
    DAQlib.DAQmxIsTaskDone.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxIsTaskDone(taskHandle, isTaskDone)


@catchDAQError
def DAQmxTaskControl(taskHandle, action):
    """int32 __CFUNC     DAQmxTaskControl               (TaskHandle taskHandle, int32 action);"""
    global DAQlib
    DAQlib.DAQmxTaskControl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxTaskControl(taskHandle, action)


@catchDAQError
def DAQmxGetNthTaskChannel(taskHandle, index, buffer, bufferSize):
    """int32 __CFUNC     DAQmxGetNthTaskChannel
    (TaskHandle taskHandle, uInt32 index, char buffer[], int32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetNthTaskChannel.argtypes = (TaskHandle, uInt32, c_char_p, int32,)
    return DAQlib.DAQmxGetNthTaskChannel(taskHandle, index, buffer, bufferSize)


@catchDAQError
def DAQmxGetNthTaskDevice(taskHandle, index, buffer, bufferSize):
    """int32 __CFUNC     DAQmxGetNthTaskDevice
    (TaskHandle taskHandle, uInt32 index, char buffer[], int32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetNthTaskDevice.argtypes = (TaskHandle, uInt32, c_char_p, int32,)
    return DAQlib.DAQmxGetNthTaskDevice(taskHandle, index, buffer, bufferSize)


@catchDAQError
def DAQmxGetTaskAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetTaskAttribute
    (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetTaskAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetTaskAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxRegisterEveryNSamplesEvent(task, everyNsamplesEventType, nSamples, options, callbackFunction, callbackData):
    """int32 __CFUNC     DAQmxRegisterEveryNSamplesEvent
    (TaskHandle task, int32 everyNsamplesEventType, uInt32 nSamples, uInt32 options,
     DAQmxEveryNSamplesEventCallbackPtr callbackFunction, void \*callbackData);"""
    global DAQlib
    DAQlib.DAQmxRegisterEveryNSamplesEvent.argtypes = (
        TaskHandle, int32, uInt32, uInt32, DAQmxEveryNSamplesEventCallbackPtr, c_void_p,)
    return DAQlib.DAQmxRegisterEveryNSamplesEvent(task, everyNsamplesEventType, nSamples, options, callbackFunction,
                                                  callbackData)


@catchDAQError
def DAQmxRegisterDoneEvent(task, options, callbackFunction, callbackData):
    """int32 __CFUNC     DAQmxRegisterDoneEvent
    (TaskHandle task, uInt32 options, DAQmxDoneEventCallbackPtr callbackFunction, void \*callbackData);"""
    global DAQlib
    DAQlib.DAQmxRegisterDoneEvent.argtypes = (TaskHandle, uInt32, DAQmxDoneEventCallbackPtr, c_void_p,)
    return DAQlib.DAQmxRegisterDoneEvent(task, options, callbackFunction, callbackData)


@catchDAQError
def DAQmxRegisterSignalEvent(task, signalID, options, callbackFunction, callbackData):
    """int32 __CFUNC     DAQmxRegisterSignalEvent
    (TaskHandle task, int32 signalID, uInt32 options, DAQmxSignalEventCallbackPtr callbackFunction,
     void \*callbackData);"""
    global DAQlib
    DAQlib.DAQmxRegisterSignalEvent.argtypes = (TaskHandle, int32, uInt32, DAQmxSignalEventCallbackPtr, c_void_p,)
    return DAQlib.DAQmxRegisterSignalEvent(task, signalID, options, callbackFunction, callbackData)


@catchDAQError
def DAQmxCreateAIVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                             units, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIVoltageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIVoltageChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                           minVal, maxVal, units, customScaleName)


@catchDAQError
def DAQmxCreateAICurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                             units, shuntResistorLoc, extShuntResistorVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateAICurrentChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 shuntResistorLoc, float64 extShuntResistorVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAICurrentChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAICurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                           minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal,
                                           customScaleName)


@catchDAQError
def DAQmxCreateAIVoltageRMSChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                units, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIVoltageRMSChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIVoltageRMSChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIVoltageRMSChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                              minVal, maxVal, units, customScaleName)


@catchDAQError
def DAQmxCreateAICurrentRMSChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                units, shuntResistorLoc, extShuntResistorVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateAICurrentRMSChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 shuntResistorLoc, float64 extShuntResistorVal,
     const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAICurrentRMSChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAICurrentRMSChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                              minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal,
                                              customScaleName)


@catchDAQError
def DAQmxCreateAIThrmcplChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                             thermocoupleType, cjcSource, cjcVal, cjcChannel):
    """int32 __CFUNC     DAQmxCreateAIThrmcplChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 thermocoupleType, int32 cjcSource, float64 cjcVal, const char cjcChannel[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIThrmcplChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIThrmcplChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, thermocoupleType, cjcSource, cjcVal, cjcChannel)


@catchDAQError
def DAQmxCreateAIRTDChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, rtdType,
                         resistanceConfig, currentExcitSource, currentExcitVal, r0):
    """int32 __CFUNC     DAQmxCreateAIRTDChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 rtdType, int32 resistanceConfig, int32 currentExcitSource,
     float64 currentExcitVal, float64 r0);"""
    global DAQlib
    DAQlib.DAQmxCreateAIRTDChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64,
                                            float64, int32, int32, int32, int32, float64, float64,)
    return DAQlib.DAQmxCreateAIRTDChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                       rtdType, resistanceConfig, currentExcitSource, currentExcitVal, r0)


@catchDAQError
def DAQmxCreateAIThrmstrChanIex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                resistanceConfig, currentExcitSource, currentExcitVal, a, b, c):
    """int32 __CFUNC     DAQmxCreateAIThrmstrChanIex
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 resistanceConfig, int32 currentExcitSource, float64 currentExcitVal, float64 a,
     float64 b, float64 c);"""
    global DAQlib
    DAQlib.DAQmxCreateAIThrmstrChanIex.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64, float64,)
    return DAQlib.DAQmxCreateAIThrmstrChanIex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                              units, resistanceConfig, currentExcitSource, currentExcitVal, a, b, c)


@catchDAQError
def DAQmxCreateAIThrmstrChanVex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                resistanceConfig, voltageExcitSource, voltageExcitVal, a, b, c, r1):
    """int32 __CFUNC     DAQmxCreateAIThrmstrChanVex
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
    float64 maxVal, int32 units, int32 resistanceConfig, int32 voltageExcitSource, float64 voltageExcitVal,
    float64 a, float64 b, float64 c, float64 r1);"""
    global DAQlib
    DAQlib.DAQmxCreateAIThrmstrChanVex.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64,
        float64, float64,)
    return DAQlib.DAQmxCreateAIThrmstrChanVex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                              units, resistanceConfig, voltageExcitSource, voltageExcitVal, a, b, c, r1)


@catchDAQError
def DAQmxCreateAIFreqVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                 thresholdLevel, hysteresis, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIFreqVoltageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
    float64 maxVal, int32 units, float64 thresholdLevel,
    float64 hysteresis, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIFreqVoltageChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, float64, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIFreqVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                               maxVal, units, thresholdLevel, hysteresis, customScaleName)


@catchDAQError
def DAQmxCreateAIResistanceChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                resistanceConfig, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIResistanceChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 resistanceConfig, int32 currentExcitSource, float64 currentExcitVal,
     const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIResistanceChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIResistanceChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                              units, resistanceConfig, currentExcitSource, currentExcitVal,
                                              customScaleName)


@catchDAQError
def DAQmxCreateAIStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                strainConfig, voltageExcitSource, voltageExcitVal, gageFactor, initialBridgeVoltage,
                                nominalGageResistance, poissonRatio, leadWireResistance, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIStrainGageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 strainConfig, int32 voltageExcitSource, float64 voltageExcitVal,
     float64 gageFactor, float64 initialBridgeVoltage, float64 nominalGageResistance, float64 poissonRatio,
     float64 leadWireResistance, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIStrainGageChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64, float64,
        float64, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                              units, strainConfig, voltageExcitSource, voltageExcitVal, gageFactor,
                                              initialBridgeVoltage, nominalGageResistance, poissonRatio,
                                              leadWireResistance, customScaleName)


@catchDAQError
def DAQmxCreateAIRosetteStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, rosetteType,
                                       gageOrientation, rosetteMeasTypes, numRosetteMeasTypes, strainConfig,
                                       voltageExcitSource, voltageExcitVal, gageFactor, nominalGageResistance,
                                       poissonRatio, leadWireResistance):
    """int32 __CFUNC     DAQmxCreateAIRosetteStrainGageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 rosetteType, float64 gageOrientation, const int32 rosetteMeasTypes[],
     uInt32 numRosetteMeasTypes, int32 strainConfig, int32 voltageExcitSource, float64 voltageExcitVal,
     float64 gageFactor, float64 nominalGageResistance,float64 poissonRatio,float64 leadWireResistance);"""
    global DAQlib
    DAQlib.DAQmxCreateAIRosetteStrainGageChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,
                                                          float64, POINTER(int32), uInt32, int32, int32, float64,
                                                          float64, float64, float64, float64,)
    return DAQlib.DAQmxCreateAIRosetteStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                     rosetteType, gageOrientation, rosetteMeasTypes,
                                                     numRosetteMeasTypes, strainConfig, voltageExcitSource,
                                                     voltageExcitVal, gageFactor, nominalGageResistance,
                                                     poissonRatio, leadWireResistance)


@catchDAQError
def DAQmxCreateAIForceBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                            bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                            firstElectricalVal, secondElectricalVal, electricalUnits, firstPhysicalVal,
                                            secondPhysicalVal, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIForceBridgeTwoPointLinChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,float64 firstElectricalVal,float64 secondElectricalVal,
    int32 electricalUnits,float64 firstPhysicalVal,float64 secondPhysicalVal,int32 physicalUnits,
    const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIForceBridgeTwoPointLinChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64, float64,
        int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIForceBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel,
                                                          minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                          voltageExcitVal, nominalBridgeResistance, firstElectricalVal,
                                                          secondElectricalVal, electricalUnits, firstPhysicalVal,
                                                          secondPhysicalVal, physicalUnits, customScaleName)


@catchDAQError
def DAQmxCreateAIForceBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                      bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                      electricalVals, numElectricalVals, electricalUnits, physicalVals, numPhysicalVals,
                                      physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIForceBridgeTableChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 electricalVals[],uInt32 numElectricalVals,int32 electricalUnits,
    const float64 physicalVals[],uInt32 numPhysicalVals,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIForceBridgeTableChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,
                                                         int32, int32, float64, float64, POINTER(float64), uInt32,
                                                         int32, POINTER(float64), uInt32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIForceBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                    maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                    nominalBridgeResistance, electricalVals, numElectricalVals,
                                                    electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                    customScaleName)


@catchDAQError
def DAQmxCreateAIForceBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                           nominalBridgeResistance, forwardCoeffs, numForwardCoeffs, reverseCoeffs,
                                           numReverseCoeffs, electricalUnits, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIForceBridgePolynomialChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 forwardCoeffs[],uInt32 numForwardCoeffs,const float64 reverseCoeffs[],
    uInt32 numReverseCoeffs,int32    electricalUnits,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIForceBridgePolynomialChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,
                                                              int32, int32, float64, float64, POINTER(float64), uInt32,
                                                              POINTER(float64), uInt32, int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIForceBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                         maxVal, units, bridgeConfig, voltageExcitSource,
                                                         voltageExcitVal, nominalBridgeResistance, forwardCoeffs,
                                                         numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                                         electricalUnits, physicalUnits, customScaleName)


@catchDAQError
def DAQmxCreateAIPressureBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                               nominalBridgeResistance, firstElectricalVal, secondElectricalVal,
                                               electricalUnits, firstPhysicalVal, secondPhysicalVal, physicalUnits,
                                               customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPressureBridgeTwoPointLinChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,float64 firstElectricalVal,float64 secondElectricalVal,int32 electricalUnits,
    float64 firstPhysicalVal,float64 secondPhysicalVal,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPressureBridgeTwoPointLinChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64, float64,
        int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIPressureBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel,
                                                             minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                             voltageExcitVal, nominalBridgeResistance,
                                                             firstElectricalVal, secondElectricalVal, electricalUnits,
                                                             firstPhysicalVal, secondPhysicalVal, physicalUnits,
                                                             customScaleName)


@catchDAQError
def DAQmxCreateAIPressureBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                         bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                         electricalVals, numElectricalVals, electricalUnits, physicalVals,
                                         numPhysicalVals, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPressureBridgeTableChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 electricalVals[],uInt32 numElectricalVals,int32 electricalUnits,
    const float64 physicalVals[],uInt32 numPhysicalVals,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPressureBridgeTableChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,
                                                            int32, int32, float64, float64, POINTER(float64), uInt32,
                                                            int32, POINTER(float64), uInt32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIPressureBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                       maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                       nominalBridgeResistance, electricalVals, numElectricalVals,
                                                       electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                       customScaleName)


@catchDAQError
def DAQmxCreateAIPressureBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                              bridgeConfig, voltageExcitSource, voltageExcitVal,
                                              nominalBridgeResistance, forwardCoeffs, numForwardCoeffs, reverseCoeffs,
                                              numReverseCoeffs, electricalUnits, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPressureBridgePolynomialChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 forwardCoeffs[],uInt32 numForwardCoeffs,const float64 reverseCoeffs[],
    uInt32 numReverseCoeffs,int32    electricalUnits,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPressureBridgePolynomialChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64,
                                                                 int32, int32, int32, float64, float64,
                                                                 POINTER(float64), uInt32, POINTER(float64), uInt32,
                                                                 int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIPressureBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                            maxVal, units, bridgeConfig, voltageExcitSource,
                                                            voltageExcitVal, nominalBridgeResistance, forwardCoeffs,
                                                            numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                                            electricalUnits, physicalUnits, customScaleName)


@catchDAQError
def DAQmxCreateAITorqueBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                             bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                             firstElectricalVal, secondElectricalVal, electricalUnits, firstPhysicalVal,
                                             secondPhysicalVal, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAITorqueBridgeTwoPointLinChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,float64 firstElectricalVal,float64 secondElectricalVal,int32 electricalUnits,
    float64 firstPhysicalVal,float64 secondPhysicalVal,int32 physicalUnits,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAITorqueBridgeTwoPointLinChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, float64, float64,
        int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAITorqueBridgeTwoPointLinChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                           maxVal, units, bridgeConfig, voltageExcitSource,
                                                           voltageExcitVal, nominalBridgeResistance, firstElectricalVal,
                                                           secondElectricalVal, electricalUnits, firstPhysicalVal,
                                                           secondPhysicalVal, physicalUnits, customScaleName)


@catchDAQError
def DAQmxCreateAITorqueBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                       bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                       electricalVals, numElectricalVals, electricalUnits, physicalVals,
                                       numPhysicalVals, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAITorqueBridgeTableChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 electricalVals[],uInt32 numElectricalVals,
    int32    electricalUnits,const float64 physicalVals[],uInt32 numPhysicalVals,int32 physicalUnits,
    const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAITorqueBridgeTableChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,
                                                          int32, int32, float64, float64, POINTER(float64), uInt32,
                                                          int32, POINTER(float64), uInt32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAITorqueBridgeTableChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                     units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                     nominalBridgeResistance, electricalVals, numElectricalVals,
                                                     electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                     customScaleName)


@catchDAQError
def DAQmxCreateAITorqueBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                            bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                            forwardCoeffs, numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                            electricalUnits, physicalUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAITorqueBridgePolynomialChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const float64 forwardCoeffs[],uInt32 numForwardCoeffs,
    const float64 reverseCoeffs[],uInt32 numReverseCoeffs,int32    electricalUnits,int32 physicalUnits
    ,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAITorqueBridgePolynomialChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64,
                                                               int32, int32, int32, float64, float64,
                                                               POINTER(float64), uInt32, POINTER(float64), uInt32,
                                                               int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateAITorqueBridgePolynomialChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                          maxVal, units, bridgeConfig, voltageExcitSource,
                                                          voltageExcitVal, nominalBridgeResistance, forwardCoeffs,
                                                          numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                                          electricalUnits, physicalUnits, customScaleName)


@catchDAQError
def DAQmxCreateAIBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, bridgeConfig,
                            voltageExcitSource, voltageExcitVal, nominalBridgeResistance, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIBridgeChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 bridgeConfig,int32 voltageExcitSource,float64 voltageExcitVal,
    float64 nominalBridgeResistance,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIBridgeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                          bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                          customScaleName)


@catchDAQError
def DAQmxCreateAIVoltageChanWithExcit(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal,
                                      maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                      useExcitForScaling, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIVoltageChanWithExcit
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 bridgeConfig, int32 voltageExcitSource,
      float64 voltageExcitVal, bool32 useExcitForScaling, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIVoltageChanWithExcit.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, int32, float64, bool32, c_char_p,)
    return DAQlib.DAQmxCreateAIVoltageChanWithExcit(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                                    minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                    voltageExcitVal, useExcitForScaling, customScaleName)


@catchDAQError
def DAQmxCreateAITempBuiltInSensorChan(taskHandle, physicalChannel, nameToAssignToChannel, units):
    """int32 __CFUNC     DAQmxCreateAITempBuiltInSensorChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 units);"""
    global DAQlib
    DAQlib.DAQmxCreateAITempBuiltInSensorChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxCreateAITempBuiltInSensorChan(taskHandle, physicalChannel, nameToAssignToChannel, units)


@catchDAQError
def DAQmxCreateAIAccelChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                           sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIAccelChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, int32 currentExcitSource,
      float64 currentExcitVal, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIAccelChan.argtypes = (TaskHandle, c_char_p, c_char_p,
                                              int32, float64, float64, int32, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIAccelChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal,
                                         maxVal, units, sensitivity, sensitivityUnits, currentExcitSource,
                                         currentExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateAIVelocityIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                  units, sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal,
                                  customScaleName):
    """int32 __CFUNC     DAQmxCreateAIVelocityIEPEChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, int32 currentExcitSource,
     float64 currentExcitVal, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIVelocityIEPEChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIVelocityIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                                minVal, maxVal, units, sensitivity, sensitivityUnits,
                                                currentExcitSource, currentExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateAIForceIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                               units, sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal,
                               customScaleName):
    """int32 __CFUNC     DAQmxCreateAIForceIEPEChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, int32 currentExcitSource,
     float64 currentExcitVal, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIForceIEPEChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIForceIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal,
                                             maxVal, units, sensitivity, sensitivityUnits, currentExcitSource,
                                             currentExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateAIMicrophoneChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, units,
                                micSensitivity, maxSndPressLevel, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIMicrophoneChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     int32 units, float64 micSensitivity, float64 maxSndPressLevel, int32 currentExcitSource, float64 currentExcitVal,
     const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIMicrophoneChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, int32, float64, float64, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateAIMicrophoneChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, units,
                                              micSensitivity, maxSndPressLevel, currentExcitSource, currentExcitVal,
                                              customScaleName)


@catchDAQError
def DAQmxCreateAIPosLVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, sensitivity,
                             sensitivityUnits, voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                             customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPosLVDTChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, int32 voltageExcitSource,
      float64 voltageExcitVal, float64 voltageExcitFreq, int32 ACExcitWireMode, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPosLVDTChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, float64, int32, int32, float64, float64, int32,
        c_char_p,)
    return DAQlib.DAQmxCreateAIPosLVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           sensitivity, sensitivityUnits, voltageExcitSource, voltageExcitVal,
                                           voltageExcitFreq, ACExcitWireMode, customScaleName)


@catchDAQError
def DAQmxCreateAIPosRVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, sensitivity,
                             sensitivityUnits, voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                             customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPosRVDTChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, int32 voltageExcitSource,
     float64 voltageExcitVal, float64 voltageExcitFreq, int32 ACExcitWireMode, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPosRVDTChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, float64, int32, int32, float64, float64, int32,
        c_char_p,)
    return DAQlib.DAQmxCreateAIPosRVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           sensitivity, sensitivityUnits, voltageExcitSource, voltageExcitVal,
                                           voltageExcitFreq, ACExcitWireMode, customScaleName)


@catchDAQError
def DAQmxCreateAIPosEddyCurrProxProbeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                          sensitivity, sensitivityUnits, customScaleName):
    """int32 __CFUNC     DAQmxCreateAIPosEddyCurrProxProbeChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, float64 sensitivity, int32 sensitivityUnits, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAIPosEddyCurrProxProbeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAIPosEddyCurrProxProbeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                        maxVal, units, sensitivity, sensitivityUnits, customScaleName)


@catchDAQError
def DAQmxCreateAIDeviceTempChan(taskHandle, physicalChannel, nameToAssignToChannel, units):
    """int32 __CFUNC     DAQmxCreateAIDeviceTempChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 units);"""
    global DAQlib
    DAQlib.DAQmxCreateAIDeviceTempChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxCreateAIDeviceTempChan(taskHandle, physicalChannel, nameToAssignToChannel, units)


@catchDAQError
def DAQmxCreateTEDSAIVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                 units, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIVoltageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIVoltageChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                               minVal, maxVal, units, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAICurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                 units, shuntResistorLoc, extShuntResistorVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAICurrentChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 shuntResistorLoc, float64 extShuntResistorVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAICurrentChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAICurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                               minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal,
                                               customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIThrmcplChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, cjcSource,
                                 cjcVal, cjcChannel):
    """int32 __CFUNC     DAQmxCreateTEDSAIThrmcplChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 cjcSource, float64 cjcVal, const char cjcChannel[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIThrmcplChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIThrmcplChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, cjcSource, cjcVal, cjcChannel)


@catchDAQError
def DAQmxCreateTEDSAIRTDChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                             resistanceConfig, currentExcitSource, currentExcitVal):
    """int32 __CFUNC     DAQmxCreateTEDSAIRTDChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 resistanceConfig, int32 currentExcitSource, float64 currentExcitVal);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIRTDChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64,)
    return DAQlib.DAQmxCreateTEDSAIRTDChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           resistanceConfig, currentExcitSource, currentExcitVal)


@catchDAQError
def DAQmxCreateTEDSAIThrmstrChanIex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    resistanceConfig, currentExcitSource, currentExcitVal):
    """int32 __CFUNC     DAQmxCreateTEDSAIThrmstrChanIex
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 resistanceConfig, int32 currentExcitSource, float64 currentExcitVal);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIThrmstrChanIex.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64,)
    return DAQlib.DAQmxCreateTEDSAIThrmstrChanIex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                  units, resistanceConfig, currentExcitSource, currentExcitVal)


@catchDAQError
def DAQmxCreateTEDSAIThrmstrChanVex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    resistanceConfig, voltageExcitSource, voltageExcitVal, r1):
    """int32 __CFUNC     DAQmxCreateTEDSAIThrmstrChanVex
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
    float64 maxVal, int32 units, int32 resistanceConfig, int32 voltageExcitSource, float64 voltageExcitVal,
    float64 r1);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIThrmstrChanVex.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, float64,)
    return DAQlib.DAQmxCreateTEDSAIThrmstrChanVex(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                  units, resistanceConfig, voltageExcitSource, voltageExcitVal, r1)


@catchDAQError
def DAQmxCreateTEDSAIResistanceChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    resistanceConfig, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIResistanceChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 resistanceConfig, int32 currentExcitSource, float64 currentExcitVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIResistanceChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIResistanceChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                  units, resistanceConfig, currentExcitSource, currentExcitVal,
                                                  customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    voltageExcitSource, voltageExcitVal, initialBridgeVoltage, leadWireResistance,
                                    customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIStrainGageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 voltageExcitSource, float64 voltageExcitVal, float64 initialBridgeVoltage,
      float64 leadWireResistance, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIStrainGageChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, float64, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIStrainGageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                  units, voltageExcitSource, voltageExcitVal, initialBridgeVoltage,
                                                  leadWireResistance, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIForceBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                     voltageExcitSource, voltageExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIForceBridgeChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 voltageExcitSource,float64 voltageExcitVal,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIForceBridgeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIForceBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                   units, voltageExcitSource, voltageExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIPressureBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        voltageExcitSource, voltageExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIPressureBridgeChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 voltageExcitSource,float64 voltageExcitVal,    const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIPressureBridgeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIPressureBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                      maxVal, units, voltageExcitSource, voltageExcitVal,
                                                      customScaleName)


@catchDAQError
def DAQmxCreateTEDSAITorqueBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                      voltageExcitSource, voltageExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAITorqueBridgeChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 voltageExcitSource,float64 voltageExcitVal,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAITorqueBridgeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAITorqueBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                    units, voltageExcitSource, voltageExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                voltageExcitSource, voltageExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIBridgeChan
    (TaskHandle taskHandle,const char physicalChannel[],const char nameToAssignToChannel[],float64 minVal,
    float64 maxVal,int32 units,int32 voltageExcitSource,float64 voltageExcitVal,const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIBridgeChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIBridgeChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                              units, voltageExcitSource, voltageExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIVoltageChanWithExcit(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal,
                                          maxVal, units, voltageExcitSource, voltageExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIVoltageChanWithExcit
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 voltageExcitSource, float64 voltageExcitVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIVoltageChanWithExcit.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIVoltageChanWithExcit(taskHandle, physicalChannel, nameToAssignToChannel,
                                                        terminalConfig, minVal, maxVal, units, voltageExcitSource,
                                                        voltageExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIAccelChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                               units, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIAccelChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 currentExcitSource, float64 currentExcitVal,
     const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIAccelChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIAccelChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal,
                                             maxVal, units, currentExcitSource, currentExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIForceIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                   units, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIForceIEPEChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     float64 minVal, float64 maxVal, int32 units, int32 currentExcitSource, float64 currentExcitVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIForceIEPEChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, float64, float64, int32, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIForceIEPEChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                                 minVal, maxVal, units, currentExcitSource, currentExcitVal,
                                                 customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIMicrophoneChan(taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig, units,
                                    maxSndPressLevel, currentExcitSource, currentExcitVal, customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIMicrophoneChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig,
     int32 units, float64 maxSndPressLevel, int32 currentExcitSource, float64 currentExcitVal,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIMicrophoneChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, int32, float64, int32, float64, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIMicrophoneChan(taskHandle, physicalChannel, nameToAssignToChannel,
                                                  terminalConfig, units, maxSndPressLevel, currentExcitSource,
                                                  currentExcitVal, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIPosLVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                 voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                                 customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIPosLVDTChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 voltageExcitSource, float64 voltageExcitVal, float64 voltageExcitFreq,
      int32 ACExcitWireMode, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIPosLVDTChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIPosLVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, voltageExcitSource, voltageExcitVal, voltageExcitFreq,
                                               ACExcitWireMode, customScaleName)


@catchDAQError
def DAQmxCreateTEDSAIPosRVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                 voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                                 customScaleName):
    """int32 __CFUNC     DAQmxCreateTEDSAIPosRVDTChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, int32 voltageExcitSource, float64 voltageExcitVal, float64 voltageExcitFreq,
      int32 ACExcitWireMode, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTEDSAIPosRVDTChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateTEDSAIPosRVDTChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, voltageExcitSource, voltageExcitVal, voltageExcitFreq,
                                               ACExcitWireMode, customScaleName)


@catchDAQError
def DAQmxCreateAOVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                             customScaleName):
    """int32 __CFUNC     DAQmxCreateAOVoltageChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAOVoltageChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAOVoltageChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           customScaleName)


@catchDAQError
def DAQmxCreateAOCurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                             customScaleName):
    """int32 __CFUNC     DAQmxCreateAOCurrentChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal,
     float64 maxVal, int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateAOCurrentChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateAOCurrentChan(taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           customScaleName)


@catchDAQError
def DAQmxCreateAOFuncGenChan(taskHandle, physicalChannel, nameToAssignToChannel, type, freq, amplitude, offset):
    """int32 __CFUNC     DAQmxCreateAOFuncGenChan
    (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 type, float64 freq,
     float64 amplitude, float64 offset);"""
    global DAQlib
    DAQlib.DAQmxCreateAOFuncGenChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32, float64, float64, float64,)
    return DAQlib.DAQmxCreateAOFuncGenChan(taskHandle, physicalChannel, nameToAssignToChannel, type, freq, amplitude,
                                           offset)


@catchDAQError
def DAQmxCreateDIChan(taskHandle, lines, nameToAssignToLines, lineGrouping):
    """int32 __CFUNC     DAQmxCreateDIChan
    (TaskHandle taskHandle, const char lines[], const char nameToAssignToLines[], int32 lineGrouping);"""
    global DAQlib
    DAQlib.DAQmxCreateDIChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxCreateDIChan(taskHandle, lines, nameToAssignToLines, lineGrouping)


@catchDAQError
def DAQmxCreateDOChan(taskHandle, lines, nameToAssignToLines, lineGrouping):
    """int32 __CFUNC     DAQmxCreateDOChan
    (TaskHandle taskHandle, const char lines[], const char nameToAssignToLines[], int32 lineGrouping);"""
    global DAQlib
    DAQlib.DAQmxCreateDOChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxCreateDOChan(taskHandle, lines, nameToAssignToLines, lineGrouping)


@catchDAQError
def DAQmxCreateCIFreqChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge, measMethod, measTime,
                          divisor, customScaleName):
    """int32 __CFUNC     DAQmxCreateCIFreqChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units, int32 edge, int32 measMethod, float64 measTime, uInt32 divisor, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCIFreqChan.argtypes = (TaskHandle, c_char_p, c_char_p,
                                             float64, float64, int32, int32, int32, float64, uInt32, c_char_p,)
    return DAQlib.DAQmxCreateCIFreqChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge,
                                        measMethod, measTime, divisor, customScaleName)


@catchDAQError
def DAQmxCreateCIPeriodChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge, measMethod,
                            measTime, divisor, customScaleName):
    """int32 __CFUNC     DAQmxCreateCIPeriodChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units, int32 edge, int32 measMethod, float64 measTime, uInt32 divisor, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCIPeriodChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, float64, uInt32, c_char_p,)
    return DAQlib.DAQmxCreateCIPeriodChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge,
                                          measMethod, measTime, divisor, customScaleName)


@catchDAQError
def DAQmxCreateCICountEdgesChan(taskHandle, counter, nameToAssignToChannel, edge, initialCount, countDirection):
    """int32 __CFUNC     DAQmxCreateCICountEdgesChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 edge, uInt32 initialCount,
     int32 countDirection);"""
    global DAQlib
    DAQlib.DAQmxCreateCICountEdgesChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32, uInt32, int32,)
    return DAQlib.DAQmxCreateCICountEdgesChan(taskHandle, counter, nameToAssignToChannel, edge, initialCount,
                                              countDirection)


@catchDAQError
def DAQmxCreateCIPulseWidthChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, startingEdge,
                                customScaleName):
    """int32 __CFUNC     DAQmxCreateCIPulseWidthChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units, int32 startingEdge, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCIPulseWidthChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateCIPulseWidthChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                              startingEdge, customScaleName)


@catchDAQError
def DAQmxCreateCISemiPeriodChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, customScaleName):
    """int32 __CFUNC     DAQmxCreateCISemiPeriodChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCISemiPeriodChan.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateCISemiPeriodChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                              customScaleName)


@catchDAQError
def DAQmxCreateCITwoEdgeSepChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, firstEdge,
                                secondEdge, customScaleName):
    """int32 __CFUNC     DAQmxCreateCITwoEdgeSepChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units, int32 firstEdge, int32 secondEdge, const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCITwoEdgeSepChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, float64, float64, int32, int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateCITwoEdgeSepChan(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                              firstEdge, secondEdge, customScaleName)


@catchDAQError
def DAQmxCreateCIPulseChanFreq(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units):
    """int32 __CFUNC     DAQmxCreateCIPulseChanFreq
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units);"""
    global DAQlib
    DAQlib.DAQmxCreateCIPulseChanFreq.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,)
    return DAQlib.DAQmxCreateCIPulseChanFreq(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units)


@catchDAQError
def DAQmxCreateCIPulseChanTime(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units):
    """int32 __CFUNC     DAQmxCreateCIPulseChanTime
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal,
     int32 units);"""
    global DAQlib
    DAQlib.DAQmxCreateCIPulseChanTime.argtypes = (TaskHandle, c_char_p, c_char_p, float64, float64, int32,)
    return DAQlib.DAQmxCreateCIPulseChanTime(taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units)


@catchDAQError
def DAQmxCreateCIPulseChanTicks(taskHandle, counter, nameToAssignToChannel, sourceTerminal, minVal, maxVal):
    """int32 __CFUNC     DAQmxCreateCIPulseChanTicks
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], const char sourceTerminal[],
    float64 minVal, float64 maxVal);"""
    global DAQlib
    DAQlib.DAQmxCreateCIPulseChanTicks.argtypes = (TaskHandle, c_char_p, c_char_p, c_char_p, float64, float64,)
    return DAQlib.DAQmxCreateCIPulseChanTicks(taskHandle, counter, nameToAssignToChannel, sourceTerminal, minVal,
                                              maxVal)


@catchDAQError
def DAQmxCreateCILinEncoderChan(taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable, ZidxVal,
                                ZidxPhase, units, distPerPulse, initialPos, customScaleName):
    """int32 __CFUNC     DAQmxCreateCILinEncoderChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 decodingType,
     bool32 ZidxEnable, float64 ZidxVal, int32 ZidxPhase, int32 units, float64 distPerPulse, float64 initialPos,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCILinEncoderChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, bool32, float64, int32, int32, float64, float64, c_char_p,)
    return DAQlib.DAQmxCreateCILinEncoderChan(taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable,
                                              ZidxVal, ZidxPhase, units, distPerPulse, initialPos, customScaleName)


@catchDAQError
def DAQmxCreateCIAngEncoderChan(taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable, ZidxVal,
                                ZidxPhase, units, pulsesPerRev, initialAngle, customScaleName):
    """int32 __CFUNC     DAQmxCreateCIAngEncoderChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 decodingType,
     bool32 ZidxEnable, float64 ZidxVal, int32 ZidxPhase, int32 units, uInt32 pulsesPerRev, float64 initialAngle,
      const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCIAngEncoderChan.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, bool32, float64, int32, int32, uInt32, float64, c_char_p,)
    return DAQlib.DAQmxCreateCIAngEncoderChan(taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable,
                                              ZidxVal, ZidxPhase, units, pulsesPerRev, initialAngle, customScaleName)


@catchDAQError
def DAQmxCreateCIGPSTimestampChan(taskHandle, counter, nameToAssignToChannel, units, syncMethod, customScaleName):
    """int32 __CFUNC     DAQmxCreateCIGPSTimestampChan
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 units, int32 syncMethod,
     const char customScaleName[]);"""
    global DAQlib
    DAQlib.DAQmxCreateCIGPSTimestampChan.argtypes = (TaskHandle, c_char_p, c_char_p, int32, int32, c_char_p,)
    return DAQlib.DAQmxCreateCIGPSTimestampChan(taskHandle, counter, nameToAssignToChannel, units, syncMethod,
                                                customScaleName)


@catchDAQError
def DAQmxCreateCOPulseChanFreq(taskHandle, counter, nameToAssignToChannel, units, idleState, initialDelay, freq,
                               dutyCycle):
    """int32 __CFUNC     DAQmxCreateCOPulseChanFreq
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 units, int32 idleState,
     float64 initialDelay, float64 freq, float64 dutyCycle);"""
    global DAQlib
    DAQlib.DAQmxCreateCOPulseChanFreq.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, int32, float64, float64, float64,)
    return DAQlib.DAQmxCreateCOPulseChanFreq(taskHandle, counter, nameToAssignToChannel, units, idleState, initialDelay,
                                             freq, dutyCycle)


@catchDAQError
def DAQmxCreateCOPulseChanTime(taskHandle, counter, nameToAssignToChannel, units, idleState, initialDelay, lowTime,
                               highTime):
    """int32 __CFUNC     DAQmxCreateCOPulseChanTime
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], int32 units, int32 idleState,
     float64 initialDelay, float64 lowTime, float64 highTime);"""
    global DAQlib
    DAQlib.DAQmxCreateCOPulseChanTime.argtypes = (
        TaskHandle, c_char_p, c_char_p, int32, int32, float64, float64, float64,)
    return DAQlib.DAQmxCreateCOPulseChanTime(taskHandle, counter, nameToAssignToChannel, units, idleState, initialDelay,
                                             lowTime, highTime)


@catchDAQError
def DAQmxCreateCOPulseChanTicks(taskHandle, counter, nameToAssignToChannel, sourceTerminal, idleState, initialDelay,
                                lowTicks, highTicks):
    """int32 __CFUNC     DAQmxCreateCOPulseChanTicks
    (TaskHandle taskHandle, const char counter[], const char nameToAssignToChannel[], const char sourceTerminal[],
     int32 idleState, int32 initialDelay, int32 lowTicks, int32 highTicks);"""
    global DAQlib
    DAQlib.DAQmxCreateCOPulseChanTicks.argtypes = (
        TaskHandle, c_char_p, c_char_p, c_char_p, int32, int32, int32, int32,)
    return DAQlib.DAQmxCreateCOPulseChanTicks(taskHandle, counter, nameToAssignToChannel, sourceTerminal, idleState,
                                              initialDelay, lowTicks, highTicks)


@catchDAQError
def DAQmxGetAIChanCalCalDate(taskHandle, channelName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxGetAIChanCalCalDate
    (TaskHandle taskHandle, const char channelName[], uInt32 \*year, uInt32 \*month, uInt32 \*day, uInt32 \*hour,
     uInt32 \*minute);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalCalDate.argtypes = (TaskHandle, c_char_p, POINTER(
        uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32),)
    return DAQlib.DAQmxGetAIChanCalCalDate(taskHandle, channelName, year, month, day, hour, minute)


@catchDAQError
def DAQmxSetAIChanCalCalDate(taskHandle, channelName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxSetAIChanCalCalDate
    (TaskHandle taskHandle, const char channelName[], uInt32 year, uInt32 month, uInt32 day, uInt32 hour,
     uInt32 minute);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalCalDate.argtypes = (TaskHandle, c_char_p, uInt32, uInt32, uInt32, uInt32, uInt32,)
    return DAQlib.DAQmxSetAIChanCalCalDate(taskHandle, channelName, year, month, day, hour, minute)


@catchDAQError
def DAQmxGetAIChanCalExpDate(taskHandle, channelName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxGetAIChanCalExpDate
    (TaskHandle taskHandle, const char channelName[], uInt32 \*year, uInt32 \*month, uInt32 \*day, uInt32 \*hour,
     uInt32 \*minute);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalExpDate.argtypes = (TaskHandle, c_char_p, POINTER(
        uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32),)
    return DAQlib.DAQmxGetAIChanCalExpDate(taskHandle, channelName, year, month, day, hour, minute)


@catchDAQError
def DAQmxSetAIChanCalExpDate(taskHandle, channelName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxSetAIChanCalExpDate
    (TaskHandle taskHandle, const char channelName[], uInt32 year, uInt32 month, uInt32 day, uInt32 hour,
     uInt32 minute);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalExpDate.argtypes = (TaskHandle, c_char_p, uInt32, uInt32, uInt32, uInt32, uInt32,)
    return DAQlib.DAQmxSetAIChanCalExpDate(taskHandle, channelName, year, month, day, hour, minute)


@catchDAQError
def DAQmxGetChanAttribute(taskHandle, channel, attribute, value):
    """int32 __CFUNC_C   DAQmxGetChanAttribute
    (TaskHandle taskHandle, const char channel[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetChanAttribute.argtypes = (TaskHandle, c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetChanAttribute(taskHandle, channel, attribute, value)


@catchDAQError
def DAQmxSetChanAttribute(taskHandle, channel, attribute):
    """int32 __CFUNC_C   DAQmxSetChanAttribute
    (TaskHandle taskHandle, const char channel[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetChanAttribute.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetChanAttribute(taskHandle, channel, attribute)


@catchDAQError
def DAQmxResetChanAttribute(taskHandle, channel, attribute):
    """int32 __CFUNC     DAQmxResetChanAttribute
    (TaskHandle taskHandle, const char channel[], int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetChanAttribute.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxResetChanAttribute(taskHandle, channel, attribute)


@catchDAQError
def DAQmxCfgSampClkTiming(taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan):
    """int32 __CFUNC     DAQmxCfgSampClkTiming
    (TaskHandle taskHandle, const char source[], float64 rate, int32 activeEdge, int32 sampleMode,
     uInt64 sampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgSampClkTiming.argtypes = (TaskHandle, c_char_p, float64, int32, int32, uInt64,)
    return DAQlib.DAQmxCfgSampClkTiming(taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan)


@catchDAQError
def DAQmxCfgHandshakingTiming(taskHandle, sampleMode, sampsPerChan):
    """int32 __CFUNC     DAQmxCfgHandshakingTiming
    (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgHandshakingTiming.argtypes = (TaskHandle, int32, uInt64,)
    return DAQlib.DAQmxCfgHandshakingTiming(taskHandle, sampleMode, sampsPerChan)


@catchDAQError
def DAQmxCfgBurstHandshakingTimingImportClock(taskHandle, sampleMode, sampsPerChan, sampleClkRate, sampleClkSrc,
                                              sampleClkActiveEdge, pauseWhen, readyEventActiveLevel):
    """int32 __CFUNC     DAQmxCfgBurstHandshakingTimingImportClock
    (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChan, float64 sampleClkRate, const char sampleClkSrc[],
     int32 sampleClkActiveEdge, int32 pauseWhen, int32 readyEventActiveLevel);"""
    global DAQlib
    DAQlib.DAQmxCfgBurstHandshakingTimingImportClock.argtypes = (
        TaskHandle, int32, uInt64, float64, c_char_p, int32, int32, int32,)
    return DAQlib.DAQmxCfgBurstHandshakingTimingImportClock(taskHandle, sampleMode, sampsPerChan, sampleClkRate,
                                                            sampleClkSrc, sampleClkActiveEdge, pauseWhen,
                                                            readyEventActiveLevel)


@catchDAQError
def DAQmxCfgBurstHandshakingTimingExportClock(taskHandle, sampleMode, sampsPerChan, sampleClkRate, sampleClkOutpTerm,
                                              sampleClkPulsePolarity, pauseWhen, readyEventActiveLevel):
    """int32 __CFUNC     DAQmxCfgBurstHandshakingTimingExportClock
    (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChan, float64 sampleClkRate,
     const char sampleClkOutpTerm[], int32 sampleClkPulsePolarity, int32 pauseWhen, int32 readyEventActiveLevel);"""
    global DAQlib
    DAQlib.DAQmxCfgBurstHandshakingTimingExportClock.argtypes = (
        TaskHandle, int32, uInt64, float64, c_char_p, int32, int32, int32,)
    return DAQlib.DAQmxCfgBurstHandshakingTimingExportClock(taskHandle, sampleMode, sampsPerChan, sampleClkRate,
                                                            sampleClkOutpTerm, sampleClkPulsePolarity, pauseWhen,
                                                            readyEventActiveLevel)


@catchDAQError
def DAQmxCfgChangeDetectionTiming(taskHandle, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan):
    """int32 __CFUNC     DAQmxCfgChangeDetectionTiming
    (TaskHandle taskHandle, const char risingEdgeChan[], const char fallingEdgeChan[], int32 sampleMode,
     uInt64 sampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgChangeDetectionTiming.argtypes = (TaskHandle, c_char_p, c_char_p, int32, uInt64,)
    return DAQlib.DAQmxCfgChangeDetectionTiming(taskHandle, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan)


@catchDAQError
def DAQmxCfgImplicitTiming(taskHandle, sampleMode, sampsPerChan):
    """int32 __CFUNC     DAQmxCfgImplicitTiming
    (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgImplicitTiming.argtypes = (TaskHandle, int32, uInt64,)
    return DAQlib.DAQmxCfgImplicitTiming(taskHandle, sampleMode, sampsPerChan)


@catchDAQError
def DAQmxCfgPipelinedSampClkTiming(taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan):
    """int32 __CFUNC     DAQmxCfgPipelinedSampClkTiming
    (TaskHandle taskHandle, const char source[], float64 rate, int32 activeEdge, int32 sampleMode,
     uInt64 sampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgPipelinedSampClkTiming.argtypes = (TaskHandle, c_char_p, float64, int32, int32, uInt64,)
    return DAQlib.DAQmxCfgPipelinedSampClkTiming(taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan)


@catchDAQError
def DAQmxGetTimingAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetTimingAttribute
    (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetTimingAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetTimingAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetTimingAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetTimingAttribute       (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetTimingAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetTimingAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetTimingAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetTimingAttribute      (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetTimingAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetTimingAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxGetTimingAttributeEx(taskHandle, deviceNames, attribute, value):
    """int32 __CFUNC_C   DAQmxGetTimingAttributeEx
    (TaskHandle taskHandle, const char deviceNames[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetTimingAttributeEx.argtypes = (TaskHandle, c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetTimingAttributeEx(taskHandle, deviceNames, attribute, value)


@catchDAQError
def DAQmxSetTimingAttributeEx(taskHandle, deviceNames, attribute):
    """int32 __CFUNC_C   DAQmxSetTimingAttributeEx
    (TaskHandle taskHandle, const char deviceNames[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetTimingAttributeEx.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetTimingAttributeEx(taskHandle, deviceNames, attribute)


@catchDAQError
def DAQmxResetTimingAttributeEx(taskHandle, deviceNames, attribute):
    """int32 __CFUNC     DAQmxResetTimingAttributeEx
    (TaskHandle taskHandle, const char deviceNames[], int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetTimingAttributeEx.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxResetTimingAttributeEx(taskHandle, deviceNames, attribute)


@catchDAQError
def DAQmxDisableStartTrig(taskHandle):
    """int32 __CFUNC     DAQmxDisableStartTrig          (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxDisableStartTrig.argtypes = (TaskHandle,)
    return DAQlib.DAQmxDisableStartTrig(taskHandle)


@catchDAQError
def DAQmxCfgDigEdgeStartTrig(taskHandle, triggerSource, triggerEdge):
    """int32 __CFUNC     DAQmxCfgDigEdgeStartTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerEdge);"""
    global DAQlib
    DAQlib.DAQmxCfgDigEdgeStartTrig.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxCfgDigEdgeStartTrig(taskHandle, triggerSource, triggerEdge)


@catchDAQError
def DAQmxCfgAnlgEdgeStartTrig(taskHandle, triggerSource, triggerSlope, triggerLevel):
    """int32 __CFUNC     DAQmxCfgAnlgEdgeStartTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerSlope, float64 triggerLevel);"""
    global DAQlib
    DAQlib.DAQmxCfgAnlgEdgeStartTrig.argtypes = (TaskHandle, c_char_p, int32, float64,)
    return DAQlib.DAQmxCfgAnlgEdgeStartTrig(taskHandle, triggerSource, triggerSlope, triggerLevel)


@catchDAQError
def DAQmxCfgAnlgWindowStartTrig(taskHandle, triggerSource, triggerWhen, windowTop, windowBottom):
    """int32 __CFUNC     DAQmxCfgAnlgWindowStartTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerWhen, float64 windowTop, float64 windowBottom);"""
    global DAQlib
    DAQlib.DAQmxCfgAnlgWindowStartTrig.argtypes = (TaskHandle, c_char_p, int32, float64, float64,)
    return DAQlib.DAQmxCfgAnlgWindowStartTrig(taskHandle, triggerSource, triggerWhen, windowTop, windowBottom)


@catchDAQError
def DAQmxCfgDigPatternStartTrig(taskHandle, triggerSource, triggerPattern, triggerWhen):
    """int32 __CFUNC     DAQmxCfgDigPatternStartTrig
    (TaskHandle taskHandle, const char triggerSource[], const char triggerPattern[], int32 triggerWhen);"""
    global DAQlib
    DAQlib.DAQmxCfgDigPatternStartTrig.argtypes = (TaskHandle, c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxCfgDigPatternStartTrig(taskHandle, triggerSource, triggerPattern, triggerWhen)


@catchDAQError
def DAQmxDisableRefTrig(taskHandle):
    """int32 __CFUNC     DAQmxDisableRefTrig            (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxDisableRefTrig.argtypes = (TaskHandle,)
    return DAQlib.DAQmxDisableRefTrig(taskHandle)


@catchDAQError
def DAQmxCfgDigEdgeRefTrig(taskHandle, triggerSource, triggerEdge, pretriggerSamples):
    """int32 __CFUNC     DAQmxCfgDigEdgeRefTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerEdge, uInt32 pretriggerSamples);"""
    global DAQlib
    DAQlib.DAQmxCfgDigEdgeRefTrig.argtypes = (TaskHandle, c_char_p, int32, uInt32,)
    return DAQlib.DAQmxCfgDigEdgeRefTrig(taskHandle, triggerSource, triggerEdge, pretriggerSamples)


@catchDAQError
def DAQmxCfgAnlgEdgeRefTrig(taskHandle, triggerSource, triggerSlope, triggerLevel, pretriggerSamples):
    """int32 __CFUNC     DAQmxCfgAnlgEdgeRefTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerSlope, float64 triggerLevel,
     uInt32 pretriggerSamples);"""
    global DAQlib
    DAQlib.DAQmxCfgAnlgEdgeRefTrig.argtypes = (TaskHandle, c_char_p, int32, float64, uInt32,)
    return DAQlib.DAQmxCfgAnlgEdgeRefTrig(taskHandle, triggerSource, triggerSlope, triggerLevel, pretriggerSamples)


@catchDAQError
def DAQmxCfgAnlgWindowRefTrig(taskHandle, triggerSource, triggerWhen, windowTop, windowBottom, pretriggerSamples):
    """int32 __CFUNC     DAQmxCfgAnlgWindowRefTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerWhen, float64 windowTop, float64 windowBottom,
     uInt32 pretriggerSamples);"""
    global DAQlib
    DAQlib.DAQmxCfgAnlgWindowRefTrig.argtypes = (TaskHandle, c_char_p, int32, float64, float64, uInt32,)
    return DAQlib.DAQmxCfgAnlgWindowRefTrig(taskHandle, triggerSource, triggerWhen, windowTop, windowBottom,
                                            pretriggerSamples)


@catchDAQError
def DAQmxCfgDigPatternRefTrig(taskHandle, triggerSource, triggerPattern, triggerWhen, pretriggerSamples):
    """int32 __CFUNC     DAQmxCfgDigPatternRefTrig
    (TaskHandle taskHandle, const char triggerSource[], const char triggerPattern[], int32 triggerWhen,
     uInt32 pretriggerSamples);"""
    global DAQlib
    DAQlib.DAQmxCfgDigPatternRefTrig.argtypes = (TaskHandle, c_char_p, c_char_p, int32, uInt32,)
    return DAQlib.DAQmxCfgDigPatternRefTrig(taskHandle, triggerSource, triggerPattern, triggerWhen, pretriggerSamples)


@catchDAQError
def DAQmxDisableAdvTrig(taskHandle):
    """int32 __CFUNC     DAQmxDisableAdvTrig            (TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxDisableAdvTrig.argtypes = (TaskHandle,)
    return DAQlib.DAQmxDisableAdvTrig(taskHandle)


@catchDAQError
def DAQmxCfgDigEdgeAdvTrig(taskHandle, triggerSource, triggerEdge):
    """int32 __CFUNC     DAQmxCfgDigEdgeAdvTrig
    (TaskHandle taskHandle, const char triggerSource[], int32 triggerEdge);"""
    global DAQlib
    DAQlib.DAQmxCfgDigEdgeAdvTrig.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxCfgDigEdgeAdvTrig(taskHandle, triggerSource, triggerEdge)


@catchDAQError
def DAQmxGetTrigAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetTrigAttribute
    (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetTrigAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetTrigAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetTrigAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetTrigAttribute          (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetTrigAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetTrigAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetTrigAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetTrigAttribute        (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetTrigAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetTrigAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxSendSoftwareTrigger(taskHandle, triggerID):
    """int32 __CFUNC     DAQmxSendSoftwareTrigger       (TaskHandle taskHandle, int32 triggerID);"""
    global DAQlib
    DAQlib.DAQmxSendSoftwareTrigger.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSendSoftwareTrigger(taskHandle, triggerID)


@catchDAQError
def DAQmxReadAnalogF64(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
    """int32 __CFUNC     DAQmxReadAnalogF64
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, float64 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadAnalogF64.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(float64), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadAnalogF64(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadAnalogScalarF64(taskHandle, timeout, value, reserved):
    """int32 __CFUNC     DAQmxReadAnalogScalarF64
    (TaskHandle taskHandle, float64 timeout, float64 \*value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadAnalogScalarF64.argtypes = (TaskHandle, float64, POINTER(float64), POINTER(bool32),)
    return DAQlib.DAQmxReadAnalogScalarF64(taskHandle, timeout, value, reserved)


@catchDAQError
def DAQmxReadBinaryI16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
    """int32 __CFUNC     DAQmxReadBinaryI16
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, int16 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadBinaryI16.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(int16), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadBinaryI16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadBinaryU16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
    """int32 __CFUNC     DAQmxReadBinaryU16
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt16 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadBinaryU16.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(uInt16), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadBinaryU16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadBinaryI32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                       sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadBinaryI32
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, int32 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadBinaryI32.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(int32), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadBinaryI32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadBinaryU32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
    """int32 __CFUNC     DAQmxReadBinaryU32
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt32 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadBinaryU32.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(uInt32), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadBinaryU32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadDigitalU8(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
    """int32 __CFUNC     DAQmxReadDigitalU8
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt8 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadDigitalU8.argtypes = (TaskHandle, int32, float64, bool32,
                                          POINTER(uInt8), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadDigitalU8(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                     sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadDigitalU16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                        reserved):
    """int32 __CFUNC     DAQmxReadDigitalU16
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt16 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadDigitalU16.argtypes = (TaskHandle, int32, float64, bool32,
                                           POINTER(uInt16), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadDigitalU16(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                      sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadDigitalU32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                        reserved):
    """int32 __CFUNC     DAQmxReadDigitalU32
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt32 readArray[],
     uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadDigitalU32.argtypes = (TaskHandle, int32, float64, bool32,
                                           POINTER(uInt32), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadDigitalU32(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                      sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadDigitalScalarU32(taskHandle, timeout, value, reserved):
    """int32 __CFUNC     DAQmxReadDigitalScalarU32
    (TaskHandle taskHandle, float64 timeout, uInt32 \*value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadDigitalScalarU32.argtypes = (TaskHandle, float64, POINTER(uInt32), POINTER(bool32),)
    return DAQlib.DAQmxReadDigitalScalarU32(taskHandle, timeout, value, reserved)


@catchDAQError
def DAQmxReadDigitalLines(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInBytes, sampsPerChanRead,
                          numBytesPerSamp, reserved):
    """int32 __CFUNC     DAQmxReadDigitalLines
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, uInt8 readArray[],
     uInt32 arraySizeInBytes, int32 \*sampsPerChanRead, int32 \*numBytesPerSamp, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadDigitalLines.argtypes = (TaskHandle, int32, float64, bool32, POINTER(
        uInt8), uInt32, POINTER(int32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadDigitalLines(taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInBytes,
                                        sampsPerChanRead, numBytesPerSamp, reserved)


@catchDAQError
def DAQmxReadCounterF64(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps, sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadCounterF64
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, float64 readArray[], uInt32 arraySizeInSamps,
     int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCounterF64.argtypes = (TaskHandle, int32, float64, POINTER(
        float64), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadCounterF64(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps,
                                      sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadCounterU32(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps, sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadCounterU32
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, uInt32 readArray[], uInt32 arraySizeInSamps,
     int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCounterU32.argtypes = (TaskHandle, int32, float64, POINTER(
        uInt32), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadCounterU32(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps,
                                      sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadCounterScalarF64(taskHandle, timeout, value, reserved):
    """int32 __CFUNC     DAQmxReadCounterScalarF64
    (TaskHandle taskHandle, float64 timeout, float64 \*value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCounterScalarF64.argtypes = (TaskHandle, float64, POINTER(float64), POINTER(bool32),)
    return DAQlib.DAQmxReadCounterScalarF64(taskHandle, timeout, value, reserved)


@catchDAQError
def DAQmxReadCounterScalarU32(taskHandle, timeout, value, reserved):
    """int32 __CFUNC     DAQmxReadCounterScalarU32
    (TaskHandle taskHandle, float64 timeout, uInt32 \*value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCounterScalarU32.argtypes = (TaskHandle, float64, POINTER(uInt32), POINTER(bool32),)
    return DAQlib.DAQmxReadCounterScalarU32(taskHandle, timeout, value, reserved)


@catchDAQError
def DAQmxReadCtrFreq(taskHandle, numSampsPerChan, timeout, interleaved, readArrayFrequency, readArrayDutyCycle,
                     arraySizeInSamps, sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadCtrFreq
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 interleaved, float64 readArrayFrequency[],
     float64 readArrayDutyCycle[], uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrFreq.argtypes = (TaskHandle, int32, float64, bool32, POINTER(
        float64), POINTER(float64), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrFreq(taskHandle, numSampsPerChan, timeout, interleaved, readArrayFrequency,
                                   readArrayDutyCycle, arraySizeInSamps, sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadCtrTime(taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTime, readArrayLowTime,
                     arraySizeInSamps, sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadCtrTime
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 interleaved, float64 readArrayHighTime[],
     float64 readArrayLowTime[], uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrTime.argtypes = (TaskHandle, int32, float64, bool32, POINTER(
        float64), POINTER(float64), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrTime(taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTime,
                                   readArrayLowTime, arraySizeInSamps, sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadCtrTicks(taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTicks, readArrayLowTicks,
                      arraySizeInSamps, sampsPerChanRead, reserved):
    """int32 __CFUNC     DAQmxReadCtrTicks
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 interleaved, uInt32 readArrayHighTicks[],
     uInt32 readArrayLowTicks[], uInt32 arraySizeInSamps, int32 \*sampsPerChanRead, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrTicks.argtypes = (TaskHandle, int32, float64, bool32, POINTER(
        uInt32), POINTER(uInt32), uInt32, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrTicks(taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTicks,
                                    readArrayLowTicks, arraySizeInSamps, sampsPerChanRead, reserved)


@catchDAQError
def DAQmxReadCtrFreqScalar(taskHandle, timeout, frequency, dutyCycle, reserved):
    """int32 __CFUNC     DAQmxReadCtrFreqScalar
    (TaskHandle taskHandle, float64 timeout, float64 \*frequency, float64 \*dutyCycle, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrFreqScalar.argtypes = (TaskHandle, float64, POINTER(float64), POINTER(float64), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrFreqScalar(taskHandle, timeout, frequency, dutyCycle, reserved)


@catchDAQError
def DAQmxReadCtrTimeScalar(taskHandle, timeout, highTime, lowTime, reserved):
    """int32 __CFUNC     DAQmxReadCtrTimeScalar
    (TaskHandle taskHandle, float64 timeout, float64 \*highTime, float64 \*lowTime, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrTimeScalar.argtypes = (TaskHandle, float64, POINTER(float64), POINTER(float64), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrTimeScalar(taskHandle, timeout, highTime, lowTime, reserved)


@catchDAQError
def DAQmxReadCtrTicksScalar(taskHandle, timeout, highTicks, lowTicks, reserved):
    """int32 __CFUNC     DAQmxReadCtrTicksScalar
    (TaskHandle taskHandle, float64 timeout, uInt32 \*highTicks, uInt32 \*lowTicks, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadCtrTicksScalar.argtypes = (TaskHandle, float64, POINTER(uInt32), POINTER(uInt32), POINTER(bool32),)
    return DAQlib.DAQmxReadCtrTicksScalar(taskHandle, timeout, highTicks, lowTicks, reserved)


@catchDAQError
def DAQmxReadRaw(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInBytes, sampsRead, numBytesPerSamp,
                 reserved):
    """int32 __CFUNC     DAQmxReadRaw
    (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, void \*readArray, uInt32 arraySizeInBytes,
     int32 \*sampsRead, int32 \*numBytesPerSamp, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxReadRaw.argtypes = (TaskHandle, int32, float64, c_void_p, uInt32,
                                    POINTER(int32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxReadRaw(taskHandle, numSampsPerChan, timeout, readArray, arraySizeInBytes, sampsRead,
                               numBytesPerSamp, reserved)


@catchDAQError
def DAQmxGetNthTaskReadChannel(taskHandle, index, buffer, bufferSize):
    """int32 __CFUNC     DAQmxGetNthTaskReadChannel
    (TaskHandle taskHandle, uInt32 index, char buffer[], int32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetNthTaskReadChannel.argtypes = (TaskHandle, uInt32, c_char_p, int32,)
    return DAQlib.DAQmxGetNthTaskReadChannel(taskHandle, index, buffer, bufferSize)


@catchDAQError
def DAQmxGetReadAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetReadAttribute
    (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetReadAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetReadAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetReadAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetReadAttribute       (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetReadAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetReadAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetReadAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetReadAttribute        (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetReadAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetReadAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxConfigureLogging(taskHandle, filePath, loggingMode, groupName, operation):
    """int32 __CFUNC     DAQmxConfigureLogging
    (TaskHandle taskHandle, const char filePath[], int32 loggingMode, const char groupName[], int32 operation);"""
    global DAQlib
    DAQlib.DAQmxConfigureLogging.argtypes = (TaskHandle, c_char_p, int32, c_char_p, int32,)
    return DAQlib.DAQmxConfigureLogging(taskHandle, filePath, loggingMode, groupName, operation)


@catchDAQError
def DAQmxStartNewFile(taskHandle, filePath):
    """int32 __CFUNC     DAQmxStartNewFile              (TaskHandle taskHandle, const char filePath[]);"""
    global DAQlib
    DAQlib.DAQmxStartNewFile.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxStartNewFile(taskHandle, filePath)


@catchDAQError
def DAQmxWriteAnalogF64(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
    """int32 __CFUNC     DAQmxWriteAnalogF64
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const float64 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteAnalogF64.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(float64), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteAnalogF64(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteAnalogScalarF64(taskHandle, autoStart, timeout, value, reserved):
    """int32 __CFUNC     DAQmxWriteAnalogScalarF64
    (TaskHandle taskHandle, bool32 autoStart, float64 timeout, float64 value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteAnalogScalarF64.argtypes = (TaskHandle, bool32, float64, float64, POINTER(bool32),)
    return DAQlib.DAQmxWriteAnalogScalarF64(taskHandle, autoStart, timeout, value, reserved)


@catchDAQError
def DAQmxWriteBinaryI16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
    """int32 __CFUNC     DAQmxWriteBinaryI16
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const int16 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteBinaryI16.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(int16), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteBinaryI16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteBinaryU16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
    """int32 __CFUNC     DAQmxWriteBinaryU16
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt16 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteBinaryU16.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(uInt16), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteBinaryU16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteBinaryI32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                        sampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteBinaryI32
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const int32 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteBinaryI32.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(int32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteBinaryI32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteBinaryU32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
    """int32 __CFUNC     DAQmxWriteBinaryU32
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt32 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteBinaryU32.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(uInt32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteBinaryU32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteDigitalU8(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
    """int32 __CFUNC     DAQmxWriteDigitalU8
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt8 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteDigitalU8.argtypes = (TaskHandle, int32, bool32, float64,
                                           bool32, POINTER(uInt8), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteDigitalU8(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteDigitalU16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                         reserved):
    """int32 __CFUNC     DAQmxWriteDigitalU16
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt16 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteDigitalU16.argtypes = (TaskHandle, int32, bool32, float64,
                                            bool32, POINTER(uInt16), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteDigitalU16(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                       sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteDigitalU32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                         reserved):
    """int32 __CFUNC     DAQmxWriteDigitalU32
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt32 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteDigitalU32.argtypes = (TaskHandle, int32, bool32, float64,
                                            bool32, POINTER(uInt32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteDigitalU32(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                       sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteDigitalScalarU32(taskHandle, autoStart, timeout, value, reserved):
    """int32 __CFUNC     DAQmxWriteDigitalScalarU32
    (TaskHandle taskHandle, bool32 autoStart, float64 timeout, uInt32 value, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteDigitalScalarU32.argtypes = (TaskHandle, bool32, float64, uInt32, POINTER(bool32),)
    return DAQlib.DAQmxWriteDigitalScalarU32(taskHandle, autoStart, timeout, value, reserved)


@catchDAQError
def DAQmxWriteDigitalLines(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                           sampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteDigitalLines
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt8 writeArray[], int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteDigitalLines.argtypes = (TaskHandle, int32, bool32, float64,
                                              bool32, POINTER(uInt8), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteDigitalLines(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                         sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteCtrFreq(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, frequency, dutyCycle,
                      numSampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteCtrFreq
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const float64 frequency[], const float64 dutyCycle[], int32 \*numSampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrFreq.argtypes = (TaskHandle, int32, bool32, float64, bool32, POINTER(
        float64), POINTER(float64), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrFreq(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, frequency, dutyCycle,
                                    numSampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteCtrFreqScalar(taskHandle, autoStart, timeout, frequency, dutyCycle, reserved):
    """int32 __CFUNC     DAQmxWriteCtrFreqScalar
    (TaskHandle taskHandle, bool32 autoStart, float64 timeout, float64 frequency, float64 dutyCycle,
     bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrFreqScalar.argtypes = (TaskHandle, bool32, float64, float64, float64, POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrFreqScalar(taskHandle, autoStart, timeout, frequency, dutyCycle, reserved)


@catchDAQError
def DAQmxWriteCtrTime(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTime, lowTime,
                      numSampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteCtrTime
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const float64 highTime[], const float64 lowTime[], int32 \*numSampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrTime.argtypes = (TaskHandle, int32, bool32, float64, bool32, POINTER(
        float64), POINTER(float64), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrTime(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTime, lowTime,
                                    numSampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteCtrTimeScalar(taskHandle, autoStart, timeout, highTime, lowTime, reserved):
    """int32 __CFUNC     DAQmxWriteCtrTimeScalar
    (TaskHandle taskHandle, bool32 autoStart, float64 timeout, float64 highTime, float64 lowTime, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrTimeScalar.argtypes = (TaskHandle, bool32, float64, float64, float64, POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrTimeScalar(taskHandle, autoStart, timeout, highTime, lowTime, reserved)


@catchDAQError
def DAQmxWriteCtrTicks(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTicks, lowTicks,
                       numSampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteCtrTicks
    (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout,
     const uInt32 highTicks[], const uInt32 lowTicks[], int32 \*numSampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrTicks.argtypes = (TaskHandle, int32, bool32, float64, bool32, POINTER(
        uInt32), POINTER(uInt32), POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrTicks(taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTicks, lowTicks,
                                     numSampsPerChanWritten, reserved)


@catchDAQError
def DAQmxWriteCtrTicksScalar(taskHandle, autoStart, timeout, highTicks, lowTicks, reserved):
    """int32 __CFUNC     DAQmxWriteCtrTicksScalar
    (TaskHandle taskHandle, bool32 autoStart, float64 timeout, uInt32 highTicks, uInt32 lowTicks, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteCtrTicksScalar.argtypes = (TaskHandle, bool32, float64, uInt32, uInt32, POINTER(bool32),)
    return DAQlib.DAQmxWriteCtrTicksScalar(taskHandle, autoStart, timeout, highTicks, lowTicks, reserved)


@catchDAQError
def DAQmxWriteRaw(taskHandle, numSamps, autoStart, timeout, writeArray, sampsPerChanWritten, reserved):
    """int32 __CFUNC     DAQmxWriteRaw
    (TaskHandle taskHandle, int32 numSamps, bool32 autoStart, float64 timeout, const void \*writeArray,
     int32 \*sampsPerChanWritten, bool32 \*reserved);"""
    global DAQlib
    DAQlib.DAQmxWriteRaw.argtypes = (TaskHandle, int32, bool32, float64, c_void_p, POINTER(int32), POINTER(bool32),)
    return DAQlib.DAQmxWriteRaw(taskHandle, numSamps, autoStart, timeout, writeArray, sampsPerChanWritten, reserved)


@catchDAQError
def DAQmxGetWriteAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetWriteAttribute         (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetWriteAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetWriteAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetWriteAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetWriteAttribute         (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetWriteAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWriteAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetWriteAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetWriteAttribute       (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetWriteAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetWriteAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxExportSignal(taskHandle, signalID, outputTerminal):
    """int32 __CFUNC     DAQmxExportSignal  (TaskHandle taskHandle, int32 signalID, const char outputTerminal[]);"""
    global DAQlib
    DAQlib.DAQmxExportSignal.argtypes = (TaskHandle, int32, c_char_p,)
    return DAQlib.DAQmxExportSignal(taskHandle, signalID, outputTerminal)


@catchDAQError
def DAQmxGetExportedSignalAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetExportedSignalAttribute(TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSignalAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetExportedSignalAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetExportedSignalAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetExportedSignalAttribute(TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSignalAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedSignalAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetExportedSignalAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetExportedSignalAttribute(TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSignalAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetExportedSignalAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxCreateLinScale(name, slope, yIntercept, preScaledUnits, scaledUnits):
    """int32 __CFUNC     DAQmxCreateLinScale
    (const char name[], float64 slope, float64 yIntercept, int32 preScaledUnits, const char scaledUnits[]);"""
    global DAQlib
    DAQlib.DAQmxCreateLinScale.argtypes = (c_char_p, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateLinScale(name, slope, yIntercept, preScaledUnits, scaledUnits)


@catchDAQError
def DAQmxCreateMapScale(name, prescaledMin, prescaledMax, scaledMin, scaledMax, preScaledUnits, scaledUnits):
    """int32 __CFUNC     DAQmxCreateMapScale
    (const char name[], float64 prescaledMin, float64 prescaledMax, float64 scaledMin, float64 scaledMax,
     int32 preScaledUnits, const char scaledUnits[]);"""
    global DAQlib
    DAQlib.DAQmxCreateMapScale.argtypes = (c_char_p, float64, float64, float64, float64, int32, c_char_p,)
    return DAQlib.DAQmxCreateMapScale(name, prescaledMin, prescaledMax, scaledMin, scaledMax, preScaledUnits,
                                      scaledUnits)


@catchDAQError
def DAQmxCreatePolynomialScale(name, forwardCoeffs, numForwardCoeffsIn, reverseCoeffs, numReverseCoeffsIn,
                               preScaledUnits, scaledUnits):
    """int32 __CFUNC     DAQmxCreatePolynomialScale
    (const char name[], const float64 forwardCoeffs[], uInt32 numForwardCoeffsIn, const float64 reverseCoeffs[],
     uInt32 numReverseCoeffsIn, int32 preScaledUnits, const char scaledUnits[]);"""
    global DAQlib
    DAQlib.DAQmxCreatePolynomialScale.argtypes = (c_char_p, POINTER(
        float64), uInt32, POINTER(float64), uInt32, int32, c_char_p,)
    return DAQlib.DAQmxCreatePolynomialScale(name, forwardCoeffs, numForwardCoeffsIn, reverseCoeffs, numReverseCoeffsIn,
                                             preScaledUnits, scaledUnits)


@catchDAQError
def DAQmxCreateTableScale(name, prescaledVals, numPrescaledValsIn, scaledVals, numScaledValsIn, preScaledUnits,
                          scaledUnits):
    """int32 __CFUNC     DAQmxCreateTableScale
    (const char name[], const float64 prescaledVals[], uInt32 numPrescaledValsIn, const float64 scaledVals[],
     uInt32 numScaledValsIn, int32 preScaledUnits, const char scaledUnits[]);"""
    global DAQlib
    DAQlib.DAQmxCreateTableScale.argtypes = (c_char_p, POINTER(
        float64), uInt32, POINTER(float64), uInt32, int32, c_char_p,)
    return DAQlib.DAQmxCreateTableScale(name, prescaledVals, numPrescaledValsIn, scaledVals, numScaledValsIn,
                                        preScaledUnits, scaledUnits)


@catchDAQError
def DAQmxCalculateReversePolyCoeff(forwardCoeffs, numForwardCoeffsIn, minValX, maxValX, numPointsToCompute,
                                   reversePolyOrder, reverseCoeffs):
    """int32 __CFUNC     DAQmxCalculateReversePolyCoeff
    (const float64 forwardCoeffs[], uInt32 numForwardCoeffsIn, float64 minValX, float64 maxValX,
     int32 numPointsToCompute, int32 reversePolyOrder, float64 reverseCoeffs[]);"""
    global DAQlib
    DAQlib.DAQmxCalculateReversePolyCoeff.argtypes = (
        POINTER(float64), uInt32, float64, float64, int32, int32, POINTER(float64),)
    return DAQlib.DAQmxCalculateReversePolyCoeff(forwardCoeffs, numForwardCoeffsIn, minValX, maxValX,
                                                 numPointsToCompute, reversePolyOrder, reverseCoeffs)


@catchDAQError
def DAQmxGetScaleAttribute(scaleName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetScaleAttribute         (const char scaleName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetScaleAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetScaleAttribute(scaleName, attribute, value)


@catchDAQError
def DAQmxSetScaleAttribute(scaleName, attribute):
    """int32 __CFUNC_C   DAQmxSetScaleAttribute         (const char scaleName[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetScaleAttribute.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetScaleAttribute(scaleName, attribute)


@catchDAQError
def DAQmxCfgInputBuffer(taskHandle, numSampsPerChan):
    """int32 __CFUNC     DAQmxCfgInputBuffer            (TaskHandle taskHandle, uInt32 numSampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgInputBuffer.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxCfgInputBuffer(taskHandle, numSampsPerChan)


@catchDAQError
def DAQmxCfgOutputBuffer(taskHandle, numSampsPerChan):
    """int32 __CFUNC     DAQmxCfgOutputBuffer           (TaskHandle taskHandle, uInt32 numSampsPerChan);"""
    global DAQlib
    DAQlib.DAQmxCfgOutputBuffer.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxCfgOutputBuffer(taskHandle, numSampsPerChan)


@catchDAQError
def DAQmxGetBufferAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetBufferAttribute        (TaskHandle taskHandle, int32 attribute, void \*value);"""
    global DAQlib
    DAQlib.DAQmxGetBufferAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetBufferAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetBufferAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetBufferAttribute        (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetBufferAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetBufferAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetBufferAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetBufferAttribute      (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetBufferAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetBufferAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxSwitchCreateScanList(scanList, taskHandle):
    """int32 __CFUNC     DAQmxSwitchCreateScanList      (const char scanList[], TaskHandle \*taskHandle);"""
    global DAQlib
    DAQlib.DAQmxSwitchCreateScanList.argtypes = (c_char_p, POINTER(TaskHandle),)
    return DAQlib.DAQmxSwitchCreateScanList(scanList, taskHandle)


@catchDAQError
def DAQmxSwitchConnect(switchChannel1, switchChannel2, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchConnect
    (const char switchChannel1[], const char switchChannel2[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchConnect.argtypes = (c_char_p, c_char_p, bool32,)
    return DAQlib.DAQmxSwitchConnect(switchChannel1, switchChannel2, waitForSettling)


@catchDAQError
def DAQmxSwitchConnectMulti(connectionList, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchConnectMulti        (const char connectionList[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchConnectMulti.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSwitchConnectMulti(connectionList, waitForSettling)


@catchDAQError
def DAQmxSwitchDisconnect(switchChannel1, switchChannel2, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchDisconnect
    (const char switchChannel1[], const char switchChannel2[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchDisconnect.argtypes = (c_char_p, c_char_p, bool32,)
    return DAQlib.DAQmxSwitchDisconnect(switchChannel1, switchChannel2, waitForSettling)


@catchDAQError
def DAQmxSwitchDisconnectMulti(connectionList, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchDisconnectMulti     (const char connectionList[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchDisconnectMulti.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSwitchDisconnectMulti(connectionList, waitForSettling)


@catchDAQError
def DAQmxSwitchDisconnectAll(deviceName, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchDisconnectAll       (const char deviceName[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchDisconnectAll.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSwitchDisconnectAll(deviceName, waitForSettling)


@catchDAQError
def DAQmxSwitchSetTopologyAndReset(deviceName, newTopology):
    """int32 __CFUNC     DAQmxSwitchSetTopologyAndReset (const char deviceName[], const char newTopology[]);"""
    global DAQlib
    DAQlib.DAQmxSwitchSetTopologyAndReset.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxSwitchSetTopologyAndReset(deviceName, newTopology)


@catchDAQError
def DAQmxSwitchFindPath(switchChannel1, switchChannel2, path, pathBufferSize, pathStatus):
    """int32 __CFUNC     DAQmxSwitchFindPath
    (const char switchChannel1[], const char switchChannel2[], char path[], uInt32 pathBufferSize,
     int32 \*pathStatus);"""
    global DAQlib
    DAQlib.DAQmxSwitchFindPath.argtypes = (c_char_p, c_char_p, c_char_p, uInt32, POINTER(int32),)
    return DAQlib.DAQmxSwitchFindPath(switchChannel1, switchChannel2, path, pathBufferSize, pathStatus)


@catchDAQError
def DAQmxSwitchOpenRelays(relayList, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchOpenRelays          (const char relayList[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchOpenRelays.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSwitchOpenRelays(relayList, waitForSettling)


@catchDAQError
def DAQmxSwitchCloseRelays(relayList, waitForSettling):
    """int32 __CFUNC     DAQmxSwitchCloseRelays         (const char relayList[], bool32 waitForSettling);"""
    global DAQlib
    DAQlib.DAQmxSwitchCloseRelays.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSwitchCloseRelays(relayList, waitForSettling)


@catchDAQError
def DAQmxSwitchGetSingleRelayCount(relayName, count):
    """int32 __CFUNC     DAQmxSwitchGetSingleRelayCount (const char relayName[], uInt32 \*count);"""
    global DAQlib
    DAQlib.DAQmxSwitchGetSingleRelayCount.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxSwitchGetSingleRelayCount(relayName, count)


@catchDAQError
def DAQmxSwitchGetMultiRelayCount(relayList, count, countArraySize, numRelayCountsRead):
    """int32 __CFUNC     DAQmxSwitchGetMultiRelayCount
    (const char relayList[], uInt32 count[], uInt32 countArraySize, uInt32 \*numRelayCountsRead);"""
    global DAQlib
    DAQlib.DAQmxSwitchGetMultiRelayCount.argtypes = (c_char_p, POINTER(uInt32), uInt32, POINTER(uInt32),)
    return DAQlib.DAQmxSwitchGetMultiRelayCount(relayList, count, countArraySize, numRelayCountsRead)


@catchDAQError
def DAQmxSwitchGetSingleRelayPos(relayName, relayPos):
    """int32 __CFUNC     DAQmxSwitchGetSingleRelayPos   (const char relayName[], uInt32 \*relayPos);"""
    global DAQlib
    DAQlib.DAQmxSwitchGetSingleRelayPos.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxSwitchGetSingleRelayPos(relayName, relayPos)


@catchDAQError
def DAQmxSwitchGetMultiRelayPos(relayList, relayPos, relayPosArraySize, numRelayPossRead):
    """int32 __CFUNC     DAQmxSwitchGetMultiRelayPos
    (const char relayList[], uInt32 relayPos[], uInt32 relayPosArraySize, uInt32 \*numRelayPossRead);"""
    global DAQlib
    DAQlib.DAQmxSwitchGetMultiRelayPos.argtypes = (c_char_p, POINTER(uInt32), uInt32, POINTER(uInt32),)
    return DAQlib.DAQmxSwitchGetMultiRelayPos(relayList, relayPos, relayPosArraySize, numRelayPossRead)


@catchDAQError
def DAQmxSwitchWaitForSettling(deviceName):
    """int32 __CFUNC     DAQmxSwitchWaitForSettling     (const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxSwitchWaitForSettling.argtypes = (c_char_p,)
    return DAQlib.DAQmxSwitchWaitForSettling(deviceName)


@catchDAQError
def DAQmxGetSwitchChanAttribute(switchChannelName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetSwitchChanAttribute
    (const char switchChannelName[], int32 attribute, void \*value);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetSwitchChanAttribute(switchChannelName, attribute, value)


@catchDAQError
def DAQmxSetSwitchChanAttribute(switchChannelName, attribute):
    """int32 __CFUNC_C   DAQmxSetSwitchChanAttribute    (const char switchChannelName[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchChanAttribute.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetSwitchChanAttribute(switchChannelName, attribute)


@catchDAQError
def DAQmxGetSwitchDeviceAttribute(deviceName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetSwitchDeviceAttribute
    (const char deviceName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDeviceAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetSwitchDeviceAttribute(deviceName, attribute, value)


@catchDAQError
def DAQmxSetSwitchDeviceAttribute(deviceName, attribute):
    """int32 __CFUNC_C   DAQmxSetSwitchDeviceAttribute  (const char deviceName[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchDeviceAttribute.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetSwitchDeviceAttribute(deviceName, attribute)


@catchDAQError
def DAQmxGetSwitchScanAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetSwitchScanAttribute    (TaskHandle taskHandle, int32 attribute, void \*value);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchScanAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetSwitchScanAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetSwitchScanAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetSwitchScanAttribute    (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchScanAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSwitchScanAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetSwitchScanAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetSwitchScanAttribute  (TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetSwitchScanAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetSwitchScanAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxConnectTerms(sourceTerminal, destinationTerminal, signalModifiers):
    """int32 __CFUNC     DAQmxConnectTerms
    (const char sourceTerminal[], const char destinationTerminal[], int32 signalModifiers);"""
    global DAQlib
    DAQlib.DAQmxConnectTerms.argtypes = (c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxConnectTerms(sourceTerminal, destinationTerminal, signalModifiers)


@catchDAQError
def DAQmxDisconnectTerms(sourceTerminal, destinationTerminal):
    """int32 __CFUNC     DAQmxDisconnectTerms
    (const char sourceTerminal[], const char destinationTerminal[]);"""
    global DAQlib
    DAQlib.DAQmxDisconnectTerms.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxDisconnectTerms(sourceTerminal, destinationTerminal)


@catchDAQError
def DAQmxTristateOutputTerm(outputTerminal):
    """int32 __CFUNC     DAQmxTristateOutputTerm        (const char outputTerminal[]);"""
    global DAQlib
    DAQlib.DAQmxTristateOutputTerm.argtypes = (c_char_p,)
    return DAQlib.DAQmxTristateOutputTerm(outputTerminal)


@catchDAQError
def DAQmxResetDevice(deviceName):
    """int32 __CFUNC     DAQmxResetDevice               (const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxResetDevice.argtypes = (c_char_p,)
    return DAQlib.DAQmxResetDevice(deviceName)


@catchDAQError
def DAQmxSelfTestDevice(deviceName):
    """int32 __CFUNC     DAQmxSelfTestDevice            (const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxSelfTestDevice.argtypes = (c_char_p,)
    return DAQlib.DAQmxSelfTestDevice(deviceName)


@catchDAQError
def DAQmxGetDeviceAttribute(deviceName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetDeviceAttribute  (const char deviceName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetDeviceAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetDeviceAttribute(deviceName, attribute, value)


@catchDAQError
def DAQmxCreateWatchdogTimerTask(deviceName, taskName, taskHandle, timeout, lines, expState):
    """int32 __CFUNC_C   DAQmxCreateWatchdogTimerTask
    (const char deviceName[], const char taskName[], TaskHandle \*taskHandle, float64 timeout, const char lines[],
     int32 expState, ...);"""
    global DAQlib
    DAQlib.DAQmxCreateWatchdogTimerTask.argtypes = (c_char_p, c_char_p, POINTER(TaskHandle), float64, c_char_p, int32,)
    return DAQlib.DAQmxCreateWatchdogTimerTask(deviceName, taskName, taskHandle, timeout, lines, expState)


@catchDAQError
def DAQmxCreateWatchdogTimerTaskEx(deviceName, taskName, taskHandle, timeout):
    """int32 __CFUNC     DAQmxCreateWatchdogTimerTaskEx
    (const char deviceName[], const char taskName[], TaskHandle \*taskHandle, float64 timeout);"""
    global DAQlib
    DAQlib.DAQmxCreateWatchdogTimerTaskEx.argtypes = (c_char_p, c_char_p, POINTER(TaskHandle), float64,)
    return DAQlib.DAQmxCreateWatchdogTimerTaskEx(deviceName, taskName, taskHandle, timeout)


@catchDAQError
def DAQmxControlWatchdogTask(taskHandle, action):
    """int32 __CFUNC     DAQmxControlWatchdogTask       (TaskHandle taskHandle, int32 action);"""
    global DAQlib
    DAQlib.DAQmxControlWatchdogTask.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxControlWatchdogTask(taskHandle, action)


@catchDAQError
def DAQmxCfgWatchdogAOExpirStates(taskHandle, channelNames, expirStateArray, outputTypeArray, arraySize):
    """int32 __CFUNC DAQmxCfgWatchdogAOExpirStates
    (TaskHandle taskHandle, const char channelNames[], const float64 expirStateArray[], const int32 outputTypeArray[],
     uInt32 arraySize);"""
    global DAQlib
    DAQlib.DAQmxCfgWatchdogAOExpirStates.argtypes = (TaskHandle, c_char_p, POINTER(float64), POINTER(int32), uInt32,)
    return DAQlib.DAQmxCfgWatchdogAOExpirStates(taskHandle, channelNames, expirStateArray, outputTypeArray, arraySize)


@catchDAQError
def DAQmxCfgWatchdogCOExpirStates(taskHandle, channelNames, expirStateArray, arraySize):
    """int32 __CFUNC DAQmxCfgWatchdogCOExpirStates
    (TaskHandle taskHandle, const char channelNames[], const int32 expirStateArray[], uInt32 arraySize);"""
    global DAQlib
    DAQlib.DAQmxCfgWatchdogCOExpirStates.argtypes = (TaskHandle, c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxCfgWatchdogCOExpirStates(taskHandle, channelNames, expirStateArray, arraySize)


@catchDAQError
def DAQmxCfgWatchdogDOExpirStates(taskHandle, channelNames, expirStateArray, arraySize):
    """int32 __CFUNC DAQmxCfgWatchdogDOExpirStates
    (TaskHandle taskHandle, const char channelNames[], const int32 expirStateArray[], uInt32 arraySize);"""
    global DAQlib
    DAQlib.DAQmxCfgWatchdogDOExpirStates.argtypes = (TaskHandle, c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxCfgWatchdogDOExpirStates(taskHandle, channelNames, expirStateArray, arraySize)


@catchDAQError
def DAQmxGetWatchdogAttribute(taskHandle, lines, attribute, value):
    """int32 __CFUNC_C   DAQmxGetWatchdogAttribute
    (TaskHandle taskHandle, const char lines[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogAttribute.argtypes = (TaskHandle, c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetWatchdogAttribute(taskHandle, lines, attribute, value)


@catchDAQError
def DAQmxSetWatchdogAttribute(taskHandle, lines, attribute):
    """int32 __CFUNC_C   DAQmxSetWatchdogAttribute
    (TaskHandle taskHandle, const char lines[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogAttribute.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetWatchdogAttribute(taskHandle, lines, attribute)


@catchDAQError
def DAQmxResetWatchdogAttribute(taskHandle, lines, attribute):
    """int32 __CFUNC     DAQmxResetWatchdogAttribute (TaskHandle taskHandle, const char lines[], int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogAttribute.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxResetWatchdogAttribute(taskHandle, lines, attribute)


@catchDAQError
def DAQmxSelfCal(deviceName):
    """int32 __CFUNC     DAQmxSelfCal                   (const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxSelfCal.argtypes = (c_char_p,)
    return DAQlib.DAQmxSelfCal(deviceName)


@catchDAQError
def DAQmxPerformBridgeOffsetNullingCal(taskHandle, channel):
    """int32 __CFUNC     DAQmxPerformBridgeOffsetNullingCal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxPerformBridgeOffsetNullingCal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxPerformBridgeOffsetNullingCal(taskHandle, channel)


@catchDAQError
def DAQmxPerformBridgeOffsetNullingCalEx(taskHandle, channel, skipUnsupportedChannels):
    """int32 __CFUNC     DAQmxPerformBridgeOffsetNullingCalEx
    (TaskHandle taskHandle, const char channel[], bool32 skipUnsupportedChannels);"""
    global DAQlib
    DAQlib.DAQmxPerformBridgeOffsetNullingCalEx.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxPerformBridgeOffsetNullingCalEx(taskHandle, channel, skipUnsupportedChannels)


@catchDAQError
def DAQmxPerformThrmcplLeadOffsetNullingCal(taskHandle, channel, skipUnsupportedChannels):
    """int32 __CFUNC     DAQmxPerformThrmcplLeadOffsetNullingCal
    (TaskHandle taskHandle, const char channel[], bool32 skipUnsupportedChannels);"""
    global DAQlib
    DAQlib.DAQmxPerformThrmcplLeadOffsetNullingCal.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxPerformThrmcplLeadOffsetNullingCal(taskHandle, channel, skipUnsupportedChannels)


@catchDAQError
def DAQmxPerformStrainShuntCal(taskHandle, channel, shuntResistorValue, shuntResistorLocation, skipUnsupportedChannels):
    """int32 __CFUNC     DAQmxPerformStrainShuntCal
    (TaskHandle taskHandle, const char channel[], float64 shuntResistorValue, int32 shuntResistorLocation,
     bool32 skipUnsupportedChannels);"""
    global DAQlib
    DAQlib.DAQmxPerformStrainShuntCal.argtypes = (TaskHandle, c_char_p, float64, int32, bool32,)
    return DAQlib.DAQmxPerformStrainShuntCal(taskHandle, channel, shuntResistorValue, shuntResistorLocation,
                                             skipUnsupportedChannels)


@catchDAQError
def DAQmxPerformBridgeShuntCal(taskHandle, channel, shuntResistorValue, shuntResistorLocation, bridgeResistance,
                               skipUnsupportedChannels):
    """int32 __CFUNC     DAQmxPerformBridgeShuntCal
    (TaskHandle taskHandle, const char channel[], float64 shuntResistorValue, int32 shuntResistorLocation,
     float64 bridgeResistance, bool32 skipUnsupportedChannels);"""
    global DAQlib
    DAQlib.DAQmxPerformBridgeShuntCal.argtypes = (TaskHandle, c_char_p, float64, int32, float64, bool32,)
    return DAQlib.DAQmxPerformBridgeShuntCal(taskHandle, channel, shuntResistorValue, shuntResistorLocation,
                                             bridgeResistance, skipUnsupportedChannels)


@catchDAQError
def DAQmxGetSelfCalLastDateAndTime(deviceName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxGetSelfCalLastDateAndTime
    (const char deviceName[], uInt32 \*year, uInt32 \*month, uInt32 \*day, uInt32 \*hour, uInt32 \*minute);"""
    global DAQlib
    DAQlib.DAQmxGetSelfCalLastDateAndTime.argtypes = (c_char_p, POINTER(
        uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32),)
    return DAQlib.DAQmxGetSelfCalLastDateAndTime(deviceName, year, month, day, hour, minute)


@catchDAQError
def DAQmxGetExtCalLastDateAndTime(deviceName, year, month, day, hour, minute):
    """int32 __CFUNC     DAQmxGetExtCalLastDateAndTime
    (const char deviceName[], uInt32 \*year, uInt32 \*month, uInt32 \*day, uInt32 \*hour, uInt32 \*minute);"""
    global DAQlib
    DAQlib.DAQmxGetExtCalLastDateAndTime.argtypes = (c_char_p, POINTER(
        uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32), POINTER(uInt32),)
    return DAQlib.DAQmxGetExtCalLastDateAndTime(deviceName, year, month, day, hour, minute)


@catchDAQError
def DAQmxRestoreLastExtCalConst(deviceName):
    """int32 __CFUNC     DAQmxRestoreLastExtCalConst    (const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxRestoreLastExtCalConst.argtypes = (c_char_p,)
    return DAQlib.DAQmxRestoreLastExtCalConst(deviceName)


@catchDAQError
def DAQmxESeriesCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxESeriesCalAdjust          (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxESeriesCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxESeriesCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxMSeriesCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxMSeriesCalAdjust          (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxMSeriesCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxMSeriesCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxSSeriesCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxSSeriesCalAdjust          (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxSSeriesCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxSSeriesCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxSCBaseboardCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxSCBaseboardCalAdjust      (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxSCBaseboardCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxSCBaseboardCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxAOSeriesCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxAOSeriesCalAdjust         (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxAOSeriesCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxAOSeriesCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxXSeriesCalAdjust(referenceVoltage):
    """int32 __CFUNC     DAQmxXSeriesCalAdjust          (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxXSeriesCalAdjust.argtypes = (float64,)
    return DAQlib.DAQmxXSeriesCalAdjust(referenceVoltage)


@catchDAQError
def DAQmxDeviceSupportsCal(deviceName, calSupported):
    """int32 __CFUNC     DAQmxDeviceSupportsCal         (const char deviceName[], bool32 \*calSupported);"""
    global DAQlib
    DAQlib.DAQmxDeviceSupportsCal.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxDeviceSupportsCal(deviceName, calSupported)


@catchDAQError
def DAQmxGetCalInfoAttribute(deviceName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetCalInfoAttribute
    (const char deviceName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetCalInfoAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetCalInfoAttribute(deviceName, attribute, value)


@catchDAQError
def DAQmxSetCalInfoAttribute(deviceName, attribute):
    """int32 __CFUNC_C   DAQmxSetCalInfoAttribute       (const char deviceName[], int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetCalInfoAttribute.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetCalInfoAttribute(deviceName, attribute)


@catchDAQError
def DAQmxInitExtCal(deviceName, password):
    """int32 __CFUNC     DAQmxInitExtCal (const char deviceName[], const char password[], CalHandle \*calHandle);"""
    global DAQlib
    DAQlib.DAQmxInitExtCal.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxInitExtCal(deviceName, password)


@catchDAQError
def DAQmxCloseExtCal(action):
    """int32 __CFUNC     DAQmxCloseExtCal               (CalHandle calHandle, int32 action);"""
    global DAQlib
    DAQlib.DAQmxCloseExtCal.argtypes = (int32,)
    return DAQlib.DAQmxCloseExtCal(action)


@catchDAQError
def DAQmxChangeExtCalPassword(deviceName, password, newPassword):
    """int32 __CFUNC     DAQmxChangeExtCalPassword
    (const char deviceName[], const char password[], const char newPassword[]);"""
    global DAQlib
    DAQlib.DAQmxChangeExtCalPassword.argtypes = (c_char_p, c_char_p, c_char_p,)
    return DAQlib.DAQmxChangeExtCalPassword(deviceName, password, newPassword)


@catchDAQError
def DAQmxDSASetCalTemp(temperature):
    """int32 __CFUNC     DAQmxDSASetCalTemp             (CalHandle calHandle, float64 temperature);"""
    global DAQlib
    DAQlib.DAQmxDSASetCalTemp.argtypes = (float64,)
    return DAQlib.DAQmxDSASetCalTemp(temperature)


@catchDAQError
def DAQmxAdjustDSAAICal(referenceVoltage):
    """int32 __CFUNC     DAQmxAdjustDSAAICal            (CalHandle calHandle, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSAAICal.argtypes = (float64,)
    return DAQlib.DAQmxAdjustDSAAICal(referenceVoltage)


@catchDAQError
def DAQmxAdjustDSAAICalEx(referenceVoltage, inputsShorted):
    """int32 __CFUNC     DAQmxAdjustDSAAICalEx (CalHandle calHandle, float64 referenceVoltage, bool32 inputsShorted);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSAAICalEx.argtypes = (float64, bool32,)
    return DAQlib.DAQmxAdjustDSAAICalEx(referenceVoltage, inputsShorted)


@catchDAQError
def DAQmxAdjustDSAAICalWithGainAndCoupling(coupling, gain, referenceVoltage):
    """int32 __CFUNC     DAQmxAdjustDSAAICalWithGainAndCoupling
    (CalHandle calHandle, int32 coupling, float64 gain, float64 referenceVoltage);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSAAICalWithGainAndCoupling.argtypes = (int32, float64, float64,)
    return DAQlib.DAQmxAdjustDSAAICalWithGainAndCoupling(coupling, gain, referenceVoltage)


@catchDAQError
def DAQmxAdjustDSAAOCal(channel, requestedLowVoltage, actualLowVoltage, requestedHighVoltage, actualHighVoltage,
                        gainSetting):
    """int32 __CFUNC     DAQmxAdjustDSAAOCal
    (CalHandle calHandle, uInt32 channel, float64 requestedLowVoltage, float64 actualLowVoltage,
     float64 requestedHighVoltage, float64 actualHighVoltage, float64 gainSetting);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSAAOCal.argtypes = (uInt32, float64, float64, float64, float64, float64,)
    return DAQlib.DAQmxAdjustDSAAOCal(channel, requestedLowVoltage, actualLowVoltage, requestedHighVoltage,
                                      actualHighVoltage, gainSetting)


@catchDAQError
def DAQmxAdjust4610Cal(channelName, gain, offset):
    """int32 __CFUNC     DAQmxAdjust4610Cal
    (CalHandle calHandle, const char channelName[], float64 gain, float64 offset);"""
    global DAQlib
    DAQlib.DAQmxAdjust4610Cal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxAdjust4610Cal(channelName, gain, offset)


@catchDAQError
def DAQmxAdjustDSATimebaseCal(referenceFrequency):
    """int32 __CFUNC     DAQmxAdjustDSATimebaseCal (CalHandle calHandle, float64 referenceFrequency);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSATimebaseCal.argtypes = (float64,)
    return DAQlib.DAQmxAdjustDSATimebaseCal(referenceFrequency)


@catchDAQError
def DAQmxAdjustDSAAOTimebaseCal(measuredFrequency,  calComplete):
    """int32 __CFUNC     DAQmxAdjustDSAAOTimebaseCal
    (CalHandle calHandle, float64 measuredFrequency, bool32* calComplete);"""
    global DAQlib
    DAQlib.DAQmxAdjustDSAAOTimebaseCal.argtypes = (float64, POINTER(bool32),)
    return DAQlib.DAQmxAdjustDSAAOTimebaseCal(measuredFrequency,  calComplete)


@catchDAQError
def DAQmxSetupDSAAOTimebaseCal(expectedFrequency):
    """int32 __CFUNC     DAQmxSetupDSAAOTimebaseCal      (CalHandle calHandle, float64* expectedFrequency);"""
    global DAQlib
    DAQlib.DAQmxSetupDSAAOTimebaseCal.argtypes = (POINTER(float64),)
    return DAQlib.DAQmxSetupDSAAOTimebaseCal(expectedFrequency)


@catchDAQError
def DAQmxAdjustTIOTimebaseCal(referenceFrequency):
    """int32 __CFUNC     DAQmxAdjustTIOTimebaseCal (CalHandle calHandle, float64 referenceFrequency);"""
    global DAQlib
    DAQlib.DAQmxAdjustTIOTimebaseCal.argtypes = (float64,)
    return DAQlib.DAQmxAdjustTIOTimebaseCal(referenceFrequency)


@catchDAQError
def DAQmxAdjust4204Cal(channelNames, lowPassFreq, trackHoldEnabled, inputVal):
    """int32 __CFUNC     DAQmxAdjust4204Cal
    (CalHandle calHandle, const char channelNames[], float64 lowPassFreq, bool32 trackHoldEnabled, float64 inputVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4204Cal.argtypes = (c_char_p, float64, bool32, float64,)
    return DAQlib.DAQmxAdjust4204Cal(channelNames, lowPassFreq, trackHoldEnabled, inputVal)


@catchDAQError
def DAQmxAdjust4220Cal(channelNames, gain, inputVal):
    """int32 __CFUNC     DAQmxAdjust4220Cal
    (CalHandle calHandle, const char channelNames[], float64 gain, float64 inputVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4220Cal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxAdjust4220Cal(channelNames, gain, inputVal)


@catchDAQError
def DAQmxAdjust4224Cal(channelNames, gain, inputVal):
    """int32 __CFUNC     DAQmxAdjust4224Cal
    (CalHandle calHandle, const char channelNames[], float64 gain, float64 inputVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4224Cal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxAdjust4224Cal(channelNames, gain, inputVal)


@catchDAQError
def DAQmxAdjust4225Cal(channelNames, gain, inputVal):
    """int32 __CFUNC     DAQmxAdjust4225Cal
    (CalHandle calHandle, const char channelNames[], float64 gain, float64 inputVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4225Cal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxAdjust4225Cal(channelNames, gain, inputVal)


@catchDAQError
def DAQmxSetup433xCal(channelNames, excitationVoltage):
    """int32 __CFUNC     DAQmxSetup433xCal
    (CalHandle calHandle, const char channelNames[], float64 excitationVoltage);"""
    global DAQlib
    DAQlib.DAQmxSetup433xCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup433xCal(channelNames, excitationVoltage)


@catchDAQError
def DAQmxAdjust433xCal(refVoltage, refExcitation, shuntLocation):
    """int32 __CFUNC     DAQmxAdjust433xCal
     (CalHandle calHandle, float64 refVoltage, float64 refExcitation, int32 shuntLocation);"""
    global DAQlib
    DAQlib.DAQmxAdjust433xCal.argtypes = (float64, float64, int32,)
    return DAQlib.DAQmxAdjust433xCal(refVoltage, refExcitation, shuntLocation)


@catchDAQError
def DAQmxAdjust4300Cal(refVoltage):
    """int32 __CFUNC     DAQmxAdjust4300Cal
    (CalHandle calHandle, float64 refVoltage);"""
    global DAQlib
    DAQlib.DAQmxAdjust4300Cal.argtypes = (float64,)
    return DAQlib.DAQmxAdjust4300Cal(refVoltage)


@catchDAQError
def DAQmxAdjust4353Cal(channelNames, refVal):
    """int32 __CFUNC     DAQmxAdjust4353Cal
    (CalHandle calHandle, const char channelNames[], float64 refVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4353Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust4353Cal(channelNames, refVal)


@catchDAQError
def DAQmxAdjust4357Cal(channelNames, refVals, numRefVals):
    """int32 __CFUNC     DAQmxAdjust4357Cal
    (CalHandle calHandle, const char channelNames[], const float64 refVals[], int32 numRefVals);"""
    global DAQlib
    DAQlib.DAQmxAdjust4357Cal.argtypes = (c_char_p, POINTER(float64), int32,)
    return DAQlib.DAQmxAdjust4357Cal(channelNames, refVals, numRefVals)


@catchDAQError
def DAQmxSetup4322Cal(channelNames, outputType, outputVal):
    """int32 __CFUNC     DAQmxSetup4322Cal
    (CalHandle calHandle, const char channelNames[], int32 outputType, float64 outputVal);"""
    global DAQlib
    DAQlib.DAQmxSetup4322Cal.argtypes = (c_char_p, int32, float64,)
    return DAQlib.DAQmxSetup4322Cal(channelNames, outputType, outputVal)


@catchDAQError
def DAQmxAdjust4322Cal(channelNames, refVal):
    """int32 __CFUNC     DAQmxAdjust4322Cal   (CalHandle calHandle, const char channelNames[], float64 refVal);"""
    global DAQlib
    DAQlib.DAQmxAdjust4322Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust4322Cal(channelNames, refVal)


@catchDAQError
def DAQmxGet4322CalAdjustPoints(outputType,  adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet4322CalAdjustPoints
    (CalHandle calHandle, int32 outputType, float64* adjustmentPoints, uInt32 bufferSize );"""
    global DAQlib
    DAQlib.DAQmxGet4322CalAdjustPoints.argtypes = (int32, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet4322CalAdjustPoints(outputType,  adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxConnectSCExpressCalAccChans(channelNames, connection):
    """int32 __CFUNC     DAQmxConnectSCExpressCalAccChans
    (CalHandle calHandle, const char channelNames[], const char connection[]);"""
    global DAQlib
    DAQlib.DAQmxConnectSCExpressCalAccChans.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxConnectSCExpressCalAccChans(channelNames, connection)


@catchDAQError
def DAQmxDisconnectSCExpressCalAccChans():
    """int32 __CFUNC     DAQmxDisconnectSCExpressCalAccChans          (CalHandle calHandle);"""
    global DAQlib
    DAQlib.DAQmxDisconnectSCExpressCalAccChans.argtypes = ()
    return DAQlib.DAQmxDisconnectSCExpressCalAccChans()


@catchDAQError
def DAQmxGetPossibleSCExpressCalAccConnections(deviceName, channelNames, connections, connectionsBufferSize):
    """int32 __CFUNC     DAQmxGetPossibleSCExpressCalAccConnections
    (const char deviceName[], const char channelNames[], char \*connections, uInt32 connectionsBufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPossibleSCExpressCalAccConnections.argtypes = (c_char_p, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPossibleSCExpressCalAccConnections(deviceName, channelNames, connections,
                                                             connectionsBufferSize)


@catchDAQError
def DAQmxSetSCExpressCalAccBridgeOutput(voltsPerVolt):
    """int32 __CFUNC     DAQmxSetSCExpressCalAccBridgeOutput          (CalHandle calHandle, float64 voltsPerVolt);"""
    global DAQlib
    DAQlib.DAQmxSetSCExpressCalAccBridgeOutput.argtypes = (float64,)
    return DAQlib.DAQmxSetSCExpressCalAccBridgeOutput(voltsPerVolt)


@catchDAQError
def DAQmxGet9201CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9201CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9201CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9201CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxCSeriesSetCalTemp(temperature):
    """int32 __CFUNC     DAQmxCSeriesSetCalTemp(CalHandle calHandle, float64 temperature);"""
    global DAQlib
    DAQlib.DAQmxCSeriesSetCalTemp.argtypes = (float64,)
    return DAQlib.DAQmxCSeriesSetCalTemp(temperature)


@catchDAQError
def DAQmxAdjust9201Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9201Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9201Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9201Cal(channelNames, value)


@catchDAQError
def DAQmxGet9203CalAdjustPoints(rangeMin, rangeMax,  adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9203CalAdjustPoints
    (CalHandle calHandle, float64 rangeMin, float64 rangeMax, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9203CalAdjustPoints.argtypes = (float64, float64, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9203CalAdjustPoints(rangeMin, rangeMax,  adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9203GainCal(channelNames, rangeMin, rangeMax, value):
    """int32 __CFUNC     DAQmxAdjust9203GainCal
    (CalHandle calHandle, const char channelNames[], float64 rangeMin, float64 rangeMax, float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9203GainCal.argtypes = (c_char_p, float64, float64, float64,)
    return DAQlib.DAQmxAdjust9203GainCal(channelNames, rangeMin, rangeMax, value)


@catchDAQError
def DAQmxAdjust9203OffsetCal(channelNames, rangeMin, rangeMax):
    """int32 __CFUNC     DAQmxAdjust9203OffsetCal
    (CalHandle calHandle, const char channelNames[], float64 rangeMin, float64 rangeMax);"""
    global DAQlib
    DAQlib.DAQmxAdjust9203OffsetCal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxAdjust9203OffsetCal(channelNames, rangeMin, rangeMax)


@catchDAQError
def DAQmxAdjust9205Cal(value):
    """int32 __CFUNC     DAQmxAdjust9205Cal(CalHandle calHandle, float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9205Cal.argtypes = (float64,)
    return DAQlib.DAQmxAdjust9205Cal(value)


@catchDAQError
def DAQmxAdjust9206Cal(value):
    """int32 __CFUNC     DAQmxAdjust9206Cal(CalHandle calHandle, float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9206Cal.argtypes = (float64,)
    return DAQlib.DAQmxAdjust9206Cal(value)


@catchDAQError
def DAQmxGet9207CalAdjustPoints(channelNames,  adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9207CalAdjustPoints
    (CalHandle calHandle, const char channelNames[], float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9207CalAdjustPoints.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9207CalAdjustPoints(channelNames,  adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9207GainCal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9207GainCal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9207GainCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9207GainCal(channelNames, value)


@catchDAQError
def DAQmxAdjust9207OffsetCal(channelNames):
    """int32 __CFUNC     DAQmxAdjust9207OffsetCal(CalHandle calHandle, const char channelNames[]);"""
    global DAQlib
    DAQlib.DAQmxAdjust9207OffsetCal.argtypes = (c_char_p,)
    return DAQlib.DAQmxAdjust9207OffsetCal(channelNames)


@catchDAQError
def DAQmxGet9208CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9208CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9208CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9208CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9208GainCal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9208GainCal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9208GainCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9208GainCal(channelNames, value)


@catchDAQError
def DAQmxAdjust9208OffsetCal(channelNames):
    """int32 __CFUNC     DAQmxAdjust9208OffsetCal(CalHandle calHandle, const char channelNames[]);"""
    global DAQlib
    DAQlib.DAQmxAdjust9208OffsetCal.argtypes = (c_char_p,)
    return DAQlib.DAQmxAdjust9208OffsetCal(channelNames)


@catchDAQError
def DAQmxAdjust9211Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9211Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9211Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9211Cal(channelNames, value)


@catchDAQError
def DAQmxGet9213CalAdjustPoints(rangeMin, rangeMax,  adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9213CalAdjustPoints
    (CalHandle calHandle, float64 rangeMin, float64 rangeMax, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9213CalAdjustPoints.argtypes = (float64, float64, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9213CalAdjustPoints(rangeMin, rangeMax,  adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9213Cal(channelNames, rangeMin, rangeMax, value):
    """int32 __CFUNC     DAQmxAdjust9213Cal
    (CalHandle calHandle, const char channelNames[], float64 rangeMin, float64 rangeMax, float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9213Cal.argtypes = (c_char_p, float64, float64, float64,)
    return DAQlib.DAQmxAdjust9213Cal(channelNames, rangeMin, rangeMax, value)


@catchDAQError
def DAQmxGet9214CalAdjustPoints(channelNames, adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9214CalAdjustPoints
    (CalHandle calHandle, const char channelNames[], float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9214CalAdjustPoints.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9214CalAdjustPoints(channelNames,  adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9214Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9214Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9214Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9214Cal(channelNames, value)


@catchDAQError
def DAQmxGet9215CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9215CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9215CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9215CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9215Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9215Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9215Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9215Cal(channelNames, value)


@catchDAQError
def DAQmxGet9217CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9217CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9217CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9217CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9217Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9217Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9217Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9217Cal(channelNames, value)


@catchDAQError
def DAQmxSetup9219Cal(channelNames, rangeMin, rangeMax, measType, bridgeConfig):
    """int32 __CFUNC     DAQmxSetup9219Cal
    (CalHandle calHandle, const char channelNames[], float64 rangeMin, float64 rangeMax, int32 measType,
     int32 bridgeConfig);"""
    global DAQlib
    DAQlib.DAQmxSetup9219Cal.argtypes = (c_char_p, float64, float64, int32, int32,)
    return DAQlib.DAQmxSetup9219Cal(channelNames, rangeMin, rangeMax, measType, bridgeConfig)


@catchDAQError
def DAQmxGet9219CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9219CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9219CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9219CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9219Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9219Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9219Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9219Cal(channelNames, value)


@catchDAQError
def DAQmxGet9220CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9220CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9220CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9220CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9220Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9220Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9220Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9220Cal(channelNames, value)


@catchDAQError
def DAQmxGet9221CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9221CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9221CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9221CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9221Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9221Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9221Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9221Cal(channelNames, value)


@catchDAQError
def DAQmxGet9222CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9222CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9222CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9222CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9222Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9222Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9222Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9222Cal(channelNames, value)


@catchDAQError
def DAQmxGet9223CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9223CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9223CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9223CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9223Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9223Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9223Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9223Cal(channelNames, value)


@catchDAQError
def DAQmxGet9225CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9225CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9225CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9225CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9225Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9225Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9225Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9225Cal(channelNames, value)


@catchDAQError
def DAQmxGet9227CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9227CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9227CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9227CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9227Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9227Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9227Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9227Cal(channelNames, value)


@catchDAQError
def DAQmxGet9229CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9229CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9229CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9229CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9229Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9229Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9229Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9229Cal(channelNames, value)


@catchDAQError
def DAQmxGet9232CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9232CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9232CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9232CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9232Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9232Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9232Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9232Cal(channelNames, value)


@catchDAQError
def DAQmxGet9234CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9234CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9234CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9234CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9234GainCal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9234GainCal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9234GainCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9234GainCal(channelNames, value)


@catchDAQError
def DAQmxAdjust9234OffsetCal(channelNames):
    """int32 __CFUNC     DAQmxAdjust9234OffsetCal(CalHandle calHandle, const char channelNames[]);"""
    global DAQlib
    DAQlib.DAQmxAdjust9234OffsetCal.argtypes = (c_char_p,)
    return DAQlib.DAQmxAdjust9234OffsetCal(channelNames)


@catchDAQError
def DAQmxGet9239CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9239CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9239CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9239CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxAdjust9239Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9239Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9239Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9239Cal(channelNames, value)


@catchDAQError
def DAQmxGet9242CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9242CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9242CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9242CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9242Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9242Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9242Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup9242Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9242Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9242Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9242Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9242Cal(channelNames, value)


@catchDAQError
def DAQmxGet9244CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9244CalAdjustPoints
    (CalHandle calHandle, float64* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9244CalAdjustPoints.argtypes = (POINTER(float64), uInt32,)
    return DAQlib.DAQmxGet9244CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9244Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9244Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9244Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup9244Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9244Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9244Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9244Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9244Cal(channelNames, value)


@catchDAQError
def DAQmxGet9263CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9263CalAdjustPoints
    (CalHandle calHandle, int32* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9263CalAdjustPoints.argtypes = (POINTER(int32), uInt32,)
    return DAQlib.DAQmxGet9263CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9263Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9263Cal(CalHandle calHandle, const char channelNames[], int32 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9263Cal.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetup9263Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9263Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9263Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9263Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9263Cal(channelNames, value)


@catchDAQError
def DAQmxGet9264CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9264CalAdjustPoints
    (CalHandle calHandle, int32* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9264CalAdjustPoints.argtypes = (POINTER(int32), uInt32,)
    return DAQlib.DAQmxGet9264CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9264Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9264Cal(CalHandle calHandle, const char channelNames[], int32 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9264Cal.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetup9264Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9264Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9264Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9264Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9264Cal(channelNames, value)


@catchDAQError
def DAQmxGet9265CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9265CalAdjustPoints
    (CalHandle calHandle, int32* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9265CalAdjustPoints.argtypes = (POINTER(int32), uInt32,)
    return DAQlib.DAQmxGet9265CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9265Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9265Cal(CalHandle calHandle, const char channelNames[], int32 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9265Cal.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetup9265Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9265Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9265Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9265Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9265Cal(channelNames, value)


@catchDAQError
def DAQmxGet9269CalAdjustPoints(adjustmentPoints, bufferSize):
    """int32 __CFUNC     DAQmxGet9269CalAdjustPoints
    (CalHandle calHandle, int32* adjustmentPoints, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGet9269CalAdjustPoints.argtypes = (POINTER(int32), uInt32,)
    return DAQlib.DAQmxGet9269CalAdjustPoints(adjustmentPoints, bufferSize)


@catchDAQError
def DAQmxSetup9269Cal(channelNames, value):
    """int32 __CFUNC     DAQmxSetup9269Cal(CalHandle calHandle, const char channelNames[], int32 value);"""
    global DAQlib
    DAQlib.DAQmxSetup9269Cal.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetup9269Cal(channelNames, value)


@catchDAQError
def DAQmxAdjust9269Cal(channelNames, value):
    """int32 __CFUNC     DAQmxAdjust9269Cal(CalHandle calHandle, const char channelNames[], float64 value);"""
    global DAQlib
    DAQlib.DAQmxAdjust9269Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust9269Cal(channelNames, value)


@catchDAQError
def DAQmxSetup1102Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1102Cal (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1102Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1102Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1102Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1102Cal (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1102Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1102Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1104Cal(channelName):
    """int32 __CFUNC     DAQmxSetup1104Cal              (CalHandle calHandle, const char channelName[]);"""
    global DAQlib
    DAQlib.DAQmxSetup1104Cal.argtypes = (c_char_p,)
    return DAQlib.DAQmxSetup1104Cal(channelName)


@catchDAQError
def DAQmxAdjust1104Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1104Cal         (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1104Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1104Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1112Cal(channelName):
    """int32 __CFUNC     DAQmxSetup1112Cal              (CalHandle calHandle, const char channelName[]);"""
    global DAQlib
    DAQlib.DAQmxSetup1112Cal.argtypes = (c_char_p,)
    return DAQlib.DAQmxSetup1112Cal(channelName)


@catchDAQError
def DAQmxAdjust1112Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1112Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1112Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1112Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1122Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1122Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1122Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1122Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1122Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1122Cal           (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1122Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1122Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1124Cal(channelName, range, dacValue):
    """int32 __CFUNC     DAQmxSetup1124Cal
    (CalHandle calHandle, const char channelName[], int32 range, uInt32 dacValue);"""
    global DAQlib
    DAQlib.DAQmxSetup1124Cal.argtypes = (c_char_p, int32, uInt32,)
    return DAQlib.DAQmxSetup1124Cal(channelName, range, dacValue)


@catchDAQError
def DAQmxAdjust1124Cal(measOutput):
    """int32 __CFUNC     DAQmxAdjust1124Cal             (CalHandle calHandle, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1124Cal.argtypes = (float64,)
    return DAQlib.DAQmxAdjust1124Cal(measOutput)


@catchDAQError
def DAQmxSetup1125Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1125Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1125Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1125Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1125Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1125Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1125Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1125Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1126Cal(channelName, upperFreqLimit):
    """int32 __CFUNC     DAQmxSetup1126Cal (CalHandle calHandle, const char channelName[], float64 upperFreqLimit);"""
    global DAQlib
    DAQlib.DAQmxSetup1126Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1126Cal(channelName, upperFreqLimit)


@catchDAQError
def DAQmxAdjust1126Cal(refFreq, measOutput):
    """int32 __CFUNC     DAQmxAdjust1126Cal             (CalHandle calHandle, float64 refFreq, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1126Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1126Cal(refFreq, measOutput)


@catchDAQError
def DAQmxSetup1141Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1141Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1141Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1141Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1141Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1141Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1141Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1141Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1142Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1142Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1142Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1142Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1142Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1142Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1142Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1142Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1143Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1143Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1143Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1143Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1143Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1143Cal         (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1143Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1143Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1502Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1502Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1502Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1502Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1502Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1502Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1502Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1502Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1503Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1503Cal           (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1503Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1503Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1503Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1503Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1503Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1503Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxAdjust1503CurrentCal(channelName, measCurrent):
    """int32 __CFUNC     DAQmxAdjust1503CurrentCal
    (CalHandle calHandle, const char channelName[], float64 measCurrent);"""
    global DAQlib
    DAQlib.DAQmxAdjust1503CurrentCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAdjust1503CurrentCal(channelName, measCurrent)


@catchDAQError
def DAQmxSetup1520Cal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup1520Cal          (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup1520Cal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup1520Cal(channelName, gain)


@catchDAQError
def DAQmxAdjust1520Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1520Cal          (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1520Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1520Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1521Cal(channelName):
    """int32 __CFUNC     DAQmxSetup1521Cal              (CalHandle calHandle, const char channelName[]);"""
    global DAQlib
    DAQlib.DAQmxSetup1521Cal.argtypes = (c_char_p,)
    return DAQlib.DAQmxSetup1521Cal(channelName)


@catchDAQError
def DAQmxAdjust1521Cal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust1521Cal         (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust1521Cal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust1521Cal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup153xCal(channelName, gain):
    """int32 __CFUNC     DAQmxSetup153xCal          (CalHandle calHandle, const char channelName[], float64 gain);"""
    global DAQlib
    DAQlib.DAQmxSetup153xCal.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetup153xCal(channelName, gain)


@catchDAQError
def DAQmxAdjust153xCal(refVoltage, measOutput):
    """int32 __CFUNC     DAQmxAdjust153xCal       (CalHandle calHandle, float64 refVoltage, float64 measOutput);"""
    global DAQlib
    DAQlib.DAQmxAdjust153xCal.argtypes = (float64, float64,)
    return DAQlib.DAQmxAdjust153xCal(refVoltage, measOutput)


@catchDAQError
def DAQmxSetup1540Cal(channelName, excitationVoltage, excitationFreq):
    """int32 __CFUNC     DAQmxSetup1540Cal
    (CalHandle calHandle, const char channelName[], float64 excitationVoltage, float64 excitationFreq);"""
    global DAQlib
    DAQlib.DAQmxSetup1540Cal.argtypes = (c_char_p, float64, float64,)
    return DAQlib.DAQmxSetup1540Cal(channelName, excitationVoltage, excitationFreq)


@catchDAQError
def DAQmxAdjust1540Cal(refVoltage, measOutput, inputCalSource):
    """int32 __CFUNC     DAQmxAdjust1540Cal
    (CalHandle calHandle, float64 refVoltage, float64 measOutput, int32 inputCalSource);"""
    global DAQlib
    DAQlib.DAQmxAdjust1540Cal.argtypes = (float64, float64, int32,)
    return DAQlib.DAQmxAdjust1540Cal(refVoltage, measOutput, inputCalSource)


@catchDAQError
def DAQmxConfigureTEDS(physicalChannel, filePath):
    """int32 __CFUNC     DAQmxConfigureTEDS (const char physicalChannel[], const char filePath[]);"""
    global DAQlib
    DAQlib.DAQmxConfigureTEDS.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxConfigureTEDS(physicalChannel, filePath)


@catchDAQError
def DAQmxClearTEDS(physicalChannel):
    """int32 __CFUNC     DAQmxClearTEDS                 (const char physicalChannel[]);"""
    global DAQlib
    DAQlib.DAQmxClearTEDS.argtypes = (c_char_p,)
    return DAQlib.DAQmxClearTEDS(physicalChannel)


@catchDAQError
def DAQmxWriteToTEDSFromArray(physicalChannel, bitStream, arraySize, basicTEDSOptions):
    """int32 __CFUNC     DAQmxWriteToTEDSFromArray
    (const char physicalChannel[], const uInt8 bitStream[], uInt32 arraySize, int32 basicTEDSOptions);"""
    global DAQlib
    DAQlib.DAQmxWriteToTEDSFromArray.argtypes = (c_char_p, POINTER(uInt8), uInt32, int32,)
    return DAQlib.DAQmxWriteToTEDSFromArray(physicalChannel, bitStream, arraySize, basicTEDSOptions)


@catchDAQError
def DAQmxWriteToTEDSFromFile(physicalChannel, filePath, basicTEDSOptions):
    """int32 __CFUNC     DAQmxWriteToTEDSFromFile
    const char physicalChannel[], const char filePath[], int32 basicTEDSOptions);"""
    global DAQlib
    DAQlib.DAQmxWriteToTEDSFromFile.argtypes = (c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxWriteToTEDSFromFile(physicalChannel, filePath, basicTEDSOptions)


@catchDAQError
def DAQmxGetPhysicalChanAttribute(physicalChannel, attribute, value):
    """int32 __CFUNC_C   DAQmxGetPhysicalChanAttribute
    (const char physicalChannel[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetPhysicalChanAttribute(physicalChannel, attribute, value)


@catchDAQError
def DAQmxWaitForNextSampleClock(taskHandle, timeout, isLate):
    """int32 __CFUNC     DAQmxWaitForNextSampleClock(TaskHandle taskHandle, float64 timeout, bool32 \*isLate);"""
    global DAQlib
    DAQlib.DAQmxWaitForNextSampleClock.argtypes = (TaskHandle, float64, POINTER(bool32),)
    return DAQlib.DAQmxWaitForNextSampleClock(taskHandle, timeout, isLate)


@catchDAQError
def DAQmxGetRealTimeAttribute(taskHandle, attribute, value):
    """int32 __CFUNC_C   DAQmxGetRealTimeAttribute  (TaskHandle taskHandle, int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeAttribute.argtypes = (TaskHandle, int32, c_void_p,)
    return DAQlib.DAQmxGetRealTimeAttribute(taskHandle, attribute, value)


@catchDAQError
def DAQmxSetRealTimeAttribute(taskHandle, attribute):
    """int32 __CFUNC_C   DAQmxSetRealTimeAttribute  (TaskHandle taskHandle, int32 attribute, ...);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetRealTimeAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxResetRealTimeAttribute(taskHandle, attribute):
    """int32 __CFUNC     DAQmxResetRealTimeAttribute(TaskHandle taskHandle, int32 attribute);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeAttribute.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxResetRealTimeAttribute(taskHandle, attribute)


@catchDAQError
def DAQmxIsReadOrWriteLate(errorCode):
    """bool32 __CFUNC    DAQmxIsReadOrWriteLate     (int32 errorCode);"""
    global DAQlib
    DAQlib.DAQmxIsReadOrWriteLate.argtypes = (int32,)
    return DAQlib.DAQmxIsReadOrWriteLate(errorCode)


@catchDAQError
def DAQmxSaveTask(taskHandle, saveAs, author, options):
    """int32 __CFUNC     DAQmxSaveTask
    (TaskHandle taskHandle, const char saveAs[], const char author[], uInt32 options);"""
    global DAQlib
    DAQlib.DAQmxSaveTask.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxSaveTask(taskHandle, saveAs, author, options)


@catchDAQError
def DAQmxSaveGlobalChan(taskHandle, channelName, saveAs, author, options):
    """int32 __CFUNC     DAQmxSaveGlobalChan
    (TaskHandle taskHandle, const char channelName[], const char saveAs[], const char author[], uInt32 options);"""
    global DAQlib
    DAQlib.DAQmxSaveGlobalChan.argtypes = (TaskHandle, c_char_p, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxSaveGlobalChan(taskHandle, channelName, saveAs, author, options)


@catchDAQError
def DAQmxSaveScale(scaleName, saveAs, author, options):
    """int32 __CFUNC     DAQmxSaveScale
    (const char scaleName[], const char saveAs[], const char author[], uInt32 options);"""
    global DAQlib
    DAQlib.DAQmxSaveScale.argtypes = (c_char_p, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxSaveScale(scaleName, saveAs, author, options)


@catchDAQError
def DAQmxDeleteSavedTask(taskName):
    """int32 __CFUNC     DAQmxDeleteSavedTask (const char taskName[]);"""
    global DAQlib
    DAQlib.DAQmxDeleteSavedTask.argtypes = (c_char_p,)
    return DAQlib.DAQmxDeleteSavedTask(taskName)


@catchDAQError
def DAQmxDeleteSavedGlobalChan(channelName):
    """int32 __CFUNC     DAQmxDeleteSavedGlobalChan     (const char channelName[]);"""
    global DAQlib
    DAQlib.DAQmxDeleteSavedGlobalChan.argtypes = (c_char_p,)
    return DAQlib.DAQmxDeleteSavedGlobalChan(channelName)


@catchDAQError
def DAQmxDeleteSavedScale(scaleName):
    """int32 __CFUNC     DAQmxDeleteSavedScale          (const char scaleName[]);"""
    global DAQlib
    DAQlib.DAQmxDeleteSavedScale.argtypes = (c_char_p,)
    return DAQlib.DAQmxDeleteSavedScale(scaleName)


@catchDAQError
def DAQmxGetPersistedTaskAttribute(taskName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetPersistedTaskAttribute (const char taskName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedTaskAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetPersistedTaskAttribute(taskName, attribute, value)


@catchDAQError
def DAQmxGetPersistedChanAttribute(channel, attribute, value):
    """int32 __CFUNC_C   DAQmxGetPersistedChanAttribute (const char channel[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedChanAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetPersistedChanAttribute(channel, attribute, value)


@catchDAQError
def DAQmxGetPersistedScaleAttribute(scaleName, attribute, value):
    """int32 __CFUNC_C   DAQmxGetPersistedScaleAttribute(const char scaleName[], int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedScaleAttribute.argtypes = (c_char_p, int32, c_void_p,)
    return DAQlib.DAQmxGetPersistedScaleAttribute(scaleName, attribute, value)


@catchDAQError
def DAQmxGetSystemInfoAttribute(attribute, value):
    """int32 __CFUNC_C   DAQmxGetSystemInfoAttribute    (int32 attribute, void \*value, ...);"""
    global DAQlib
    DAQlib.DAQmxGetSystemInfoAttribute.argtypes = (int32, c_void_p,)
    return DAQlib.DAQmxGetSystemInfoAttribute(attribute, value)


@catchDAQError
def DAQmxSetDigitalPowerUpStates(deviceName, channelNames, state):
    """int32 __CFUNC_C   DAQmxSetDigitalPowerUpStates
    (const char deviceName[], const char channelNames[], int32 state, ...);"""
    global DAQlib
    DAQlib.DAQmxSetDigitalPowerUpStates.argtypes = (c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxSetDigitalPowerUpStates(deviceName, channelNames, state)


@catchDAQError
def DAQmxGetDigitalPowerUpStates(deviceName, channelName,  state):
    """int32 __CFUNC_C   DAQmxGetDigitalPowerUpStates
    (const char deviceName[], const char channelName[], int32* state, ...);"""
    global DAQlib
    DAQlib.DAQmxGetDigitalPowerUpStates.argtypes = (c_char_p, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDigitalPowerUpStates(deviceName, channelName,  state)


@catchDAQError
def DAQmxSetDigitalPullUpPullDownStates(deviceName, channelName, state):
    """int32 __CFUNC_C   DAQmxSetDigitalPullUpPullDownStates
    (const char deviceName[], const char channelName[], int32 state, ...);"""
    global DAQlib
    DAQlib.DAQmxSetDigitalPullUpPullDownStates.argtypes = (c_char_p, c_char_p, int32,)
    return DAQlib.DAQmxSetDigitalPullUpPullDownStates(deviceName, channelName, state)


@catchDAQError
def DAQmxGetDigitalPullUpPullDownStates(deviceName, channelName,  state):
    """int32 __CFUNC_C   DAQmxGetDigitalPullUpPullDownStates
    (const char deviceName[], const char channelName[], int32* state, ...);"""
    global DAQlib
    DAQlib.DAQmxGetDigitalPullUpPullDownStates.argtypes = (c_char_p, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDigitalPullUpPullDownStates(deviceName, channelName,  state)


@catchDAQError
def DAQmxSetAnalogPowerUpStates(deviceName, channelNames, state, channelType):
    """int32 __CFUNC_C   DAQmxSetAnalogPowerUpStates
    (const char deviceName[], const char channelNames[], float64 state, int32 channelType, ...);"""
    global DAQlib
    DAQlib.DAQmxSetAnalogPowerUpStates.argtypes = (c_char_p, c_char_p, float64, int32,)
    return DAQlib.DAQmxSetAnalogPowerUpStates(deviceName, channelNames, state, channelType)


@catchDAQError
def DAQmxSetAnalogPowerUpStatesWithOutputType(channelNames, stateArray, channelTypeArray, arraySize):
    """int32 __CFUNC_C   DAQmxSetAnalogPowerUpStatesWithOutputType
    (const char channelNames[], const float64 stateArray[], const int32 channelTypeArray[], uInt32 arraySize);"""
    global DAQlib
    DAQlib.DAQmxSetAnalogPowerUpStatesWithOutputType.argtypes = (c_char_p, POINTER(float64), POINTER(int32), uInt32,)
    return DAQlib.DAQmxSetAnalogPowerUpStatesWithOutputType(channelNames, stateArray, channelTypeArray, arraySize)


@catchDAQError
def DAQmxGetAnalogPowerUpStates(deviceName, channelName,  state, channelType):
    """int32 __CFUNC_C   DAQmxGetAnalogPowerUpStates
    (const char deviceName[], const char channelName[], float64* state, int32 channelType, ...);"""
    global DAQlib
    DAQlib.DAQmxGetAnalogPowerUpStates.argtypes = (c_char_p, c_char_p, POINTER(float64), int32,)
    return DAQlib.DAQmxGetAnalogPowerUpStates(deviceName, channelName,  state, channelType)


@catchDAQError
def DAQmxGetAnalogPowerUpStatesWithOutputType(channelNames, stateArray, channelTypeArray, arraySizePtr):
    """int32 __CFUNC_C DAQmxGetAnalogPowerUpStatesWithOutputType
    (const char channelNames[], float64 stateArray[], int32 channelTypeArray[], uInt32 \*arraySizePtr);"""
    global DAQlib
    DAQlib.DAQmxGetAnalogPowerUpStatesWithOutputType.argtypes = (
        c_char_p, POINTER(float64), POINTER(int32), POINTER(uInt32),)
    return DAQlib.DAQmxGetAnalogPowerUpStatesWithOutputType(channelNames, stateArray, channelTypeArray, arraySizePtr)


@catchDAQError
def DAQmxSetDigitalLogicFamilyPowerUpState(deviceName, logicFamily):
    """int32 __CFUNC     DAQmxSetDigitalLogicFamilyPowerUpState(const char deviceName[], int32 logicFamily);"""
    global DAQlib
    DAQlib.DAQmxSetDigitalLogicFamilyPowerUpState.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetDigitalLogicFamilyPowerUpState(deviceName, logicFamily)


@catchDAQError
def DAQmxGetDigitalLogicFamilyPowerUpState(deviceName,  logicFamily):
    """int32 __CFUNC     DAQmxGetDigitalLogicFamilyPowerUpState(const char deviceName[], int32* logicFamily);"""
    global DAQlib
    DAQlib.DAQmxGetDigitalLogicFamilyPowerUpState.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDigitalLogicFamilyPowerUpState(deviceName,  logicFamily)


@catchDAQError
def DAQmxAddNetworkDevice(IPAddress, deviceName, attemptReservation, timeout, deviceNameOut, deviceNameOutBufferSize):
    """int32 __CFUNC     DAQmxAddNetworkDevice
    (const char IPAddress[],const char deviceName[], bool32 attemptReservation, float64 timeout, char deviceNameOut[],
     uInt32 deviceNameOutBufferSize);"""
    global DAQlib
    DAQlib.DAQmxAddNetworkDevice.argtypes = (c_char_p, c_char_p, bool32, float64, c_char_p, uInt32,)
    return DAQlib.DAQmxAddNetworkDevice(IPAddress, deviceName, attemptReservation, timeout, deviceNameOut,
                                        deviceNameOutBufferSize)


@catchDAQError
def DAQmxDeleteNetworkDevice(deviceName):
    """int32 __CFUNC     DAQmxDeleteNetworkDevice(const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxDeleteNetworkDevice.argtypes = (c_char_p,)
    return DAQlib.DAQmxDeleteNetworkDevice(deviceName)


@catchDAQError
def DAQmxReserveNetworkDevice(deviceName, overrideReservation):
    """int32 __CFUNC     DAQmxReserveNetworkDevice(const char deviceName[], bool32 overrideReservation);"""
    global DAQlib
    DAQlib.DAQmxReserveNetworkDevice.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxReserveNetworkDevice(deviceName, overrideReservation)


@catchDAQError
def DAQmxUnreserveNetworkDevice(deviceName):
    """int32 __CFUNC     DAQmxUnreserveNetworkDevice(const char deviceName[]);"""
    global DAQlib
    DAQlib.DAQmxUnreserveNetworkDevice.argtypes = (c_char_p,)
    return DAQlib.DAQmxUnreserveNetworkDevice(deviceName)


@catchDAQError
def DAQmxAutoConfigureCDAQSyncConnections(chassisDevicesPorts, timeout):
    """int32 __CFUNC DAQmxAutoConfigureCDAQSyncConnections(const char chassisDevicesPorts[], float64 timeout);"""
    global DAQlib
    DAQlib.DAQmxAutoConfigureCDAQSyncConnections.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxAutoConfigureCDAQSyncConnections(chassisDevicesPorts, timeout)


@catchDAQError
def DAQmxGetAutoConfiguredCDAQSyncConnections(portList, portListSize):
    """int32 __CFUNC DAQmxGetAutoConfiguredCDAQSyncConnections(char portList[], uInt32 portListSize);"""
    global DAQlib
    DAQlib.DAQmxGetAutoConfiguredCDAQSyncConnections.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetAutoConfiguredCDAQSyncConnections(portList, portListSize)


@catchDAQError
def DAQmxAreConfiguredCDAQSyncPortsDisconnected(chassisDevicesPorts, timeout,  disconnectedPortsExist):
    """int32 __CFUNC DAQmxAreConfiguredCDAQSyncPortsDisconnected
    (const char chassisDevicesPorts[], float64 timeout, bool32* disconnectedPortsExist);"""
    global DAQlib
    DAQlib.DAQmxAreConfiguredCDAQSyncPortsDisconnected.argtypes = (c_char_p, float64, POINTER(bool32),)
    return DAQlib.DAQmxAreConfiguredCDAQSyncPortsDisconnected(chassisDevicesPorts, timeout,  disconnectedPortsExist)


@catchDAQError
def DAQmxGetDisconnectedCDAQSyncPorts(portList, portListSize):
    """int32 __CFUNC DAQmxGetDisconnectedCDAQSyncPorts(char portList[], uInt32 portListSize);"""
    global DAQlib
    DAQlib.DAQmxGetDisconnectedCDAQSyncPorts.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetDisconnectedCDAQSyncPorts(portList, portListSize)


@catchDAQError
def DAQmxAddCDAQSyncConnection(portList):
    """int32 __CFUNC DAQmxAddCDAQSyncConnection(const char portList[]);"""
    global DAQlib
    DAQlib.DAQmxAddCDAQSyncConnection.argtypes = (c_char_p,)
    return DAQlib.DAQmxAddCDAQSyncConnection(portList)


@catchDAQError
def DAQmxRemoveCDAQSyncConnection(portList):
    """int32 __CFUNC DAQmxRemoveCDAQSyncConnection(const char portList[]);"""
    global DAQlib
    DAQlib.DAQmxRemoveCDAQSyncConnection.argtypes = (c_char_p,)
    return DAQlib.DAQmxRemoveCDAQSyncConnection(portList)


def DAQmxGetErrorString(errorCode, errorString, bufferSize):
    """int32 __CFUNC     DAQmxGetErrorString            (int32 errorCode, char errorString[], uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetErrorString.argtypes = (int32, c_char_p, uInt32,)
    return DAQlib.DAQmxGetErrorString(errorCode, errorString, bufferSize)


def DAQmxGetExtendedErrorInfo(errorString, bufferSize):
    """int32 __CFUNC     DAQmxGetExtendedErrorInfo      (char errorString[], uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExtendedErrorInfo.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetExtendedErrorInfo(errorString, bufferSize)


@catchDAQError
def DAQmxGetBufInputBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetBufInputBufSize(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetBufInputBufSize.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetBufInputBufSize(taskHandle, data)


@catchDAQError
def DAQmxSetBufInputBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxSetBufInputBufSize(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetBufInputBufSize.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetBufInputBufSize(taskHandle, data)


@catchDAQError
def DAQmxResetBufInputBufSize(taskHandle):
    """int32 __CFUNC DAQmxResetBufInputBufSize(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetBufInputBufSize.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetBufInputBufSize(taskHandle)


@catchDAQError
def DAQmxGetBufInputOnbrdBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetBufInputOnbrdBufSize(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetBufInputOnbrdBufSize.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetBufInputOnbrdBufSize(taskHandle, data)


@catchDAQError
def DAQmxGetBufOutputBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetBufOutputBufSize(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetBufOutputBufSize.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetBufOutputBufSize(taskHandle, data)


@catchDAQError
def DAQmxSetBufOutputBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxSetBufOutputBufSize(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetBufOutputBufSize.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetBufOutputBufSize(taskHandle, data)


@catchDAQError
def DAQmxResetBufOutputBufSize(taskHandle):
    """int32 __CFUNC DAQmxResetBufOutputBufSize(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetBufOutputBufSize.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetBufOutputBufSize(taskHandle)


@catchDAQError
def DAQmxGetBufOutputOnbrdBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetBufOutputOnbrdBufSize(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetBufOutputOnbrdBufSize.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetBufOutputOnbrdBufSize(taskHandle, data)


@catchDAQError
def DAQmxSetBufOutputOnbrdBufSize(taskHandle, data):
    """int32 __CFUNC DAQmxSetBufOutputOnbrdBufSize(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetBufOutputOnbrdBufSize.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetBufOutputOnbrdBufSize(taskHandle, data)


@catchDAQError
def DAQmxResetBufOutputOnbrdBufSize(taskHandle):
    """int32 __CFUNC DAQmxResetBufOutputOnbrdBufSize(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetBufOutputOnbrdBufSize.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetBufOutputOnbrdBufSize(taskHandle)


@catchDAQError
def DAQmxGetSelfCalSupported(deviceName, data):
    """int32 __CFUNC DAQmxGetSelfCalSupported(const char deviceName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSelfCalSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetSelfCalSupported(deviceName, data)


@catchDAQError
def DAQmxGetSelfCalLastTemp(deviceName, data):
    """int32 __CFUNC DAQmxGetSelfCalLastTemp(const char deviceName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSelfCalLastTemp.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSelfCalLastTemp(deviceName, data)


@catchDAQError
def DAQmxGetExtCalRecommendedInterval(deviceName, data):
    """int32 __CFUNC DAQmxGetExtCalRecommendedInterval(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExtCalRecommendedInterval.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetExtCalRecommendedInterval(deviceName, data)


@catchDAQError
def DAQmxGetExtCalLastTemp(deviceName, data):
    """int32 __CFUNC DAQmxGetExtCalLastTemp(const char deviceName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExtCalLastTemp.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetExtCalLastTemp(deviceName, data)


@catchDAQError
def DAQmxGetCalUserDefinedInfo(deviceName, data, bufferSize):
    """int32 __CFUNC DAQmxGetCalUserDefinedInfo(const char deviceName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCalUserDefinedInfo.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCalUserDefinedInfo(deviceName, data, bufferSize)


@catchDAQError
def DAQmxSetCalUserDefinedInfo(deviceName, data):
    """int32 __CFUNC DAQmxSetCalUserDefinedInfo(const char deviceName[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCalUserDefinedInfo.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCalUserDefinedInfo(deviceName, data)


@catchDAQError
def DAQmxGetCalUserDefinedInfoMaxSize(deviceName, data):
    """int32 __CFUNC DAQmxGetCalUserDefinedInfoMaxSize(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCalUserDefinedInfoMaxSize.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCalUserDefinedInfoMaxSize(deviceName, data)


@catchDAQError
def DAQmxGetCalDevTemp(deviceName, data):
    """int32 __CFUNC DAQmxGetCalDevTemp(const char deviceName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCalDevTemp.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCalDevTemp(deviceName, data)


@catchDAQError
def DAQmxGetCalAccConnectionCount(deviceName, data):
    """int32 __CFUNC DAQmxGetCalAccConnectionCount(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCalAccConnectionCount.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCalAccConnectionCount(deviceName, data)


@catchDAQError
def DAQmxSetCalAccConnectionCount(deviceName, data):
    """int32 __CFUNC DAQmxSetCalAccConnectionCount(const char deviceName[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCalAccConnectionCount.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxSetCalAccConnectionCount(deviceName, data)


@catchDAQError
def DAQmxGetCalRecommendedAccConnectionCountLimit(deviceName, data):
    """int32 __CFUNC DAQmxGetCalRecommendedAccConnectionCountLimit(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCalRecommendedAccConnectionCountLimit.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCalRecommendedAccConnectionCountLimit(deviceName, data)


@catchDAQError
def DAQmxGetAIMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIMax(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIMax.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIMax(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIMax(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIMax.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIMax(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIMax(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIMax(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIMax.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIMax(taskHandle, channel)


@catchDAQError
def DAQmxGetAIMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIMin(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIMin.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIMin(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIMin(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIMin.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIMin(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIMin(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIMin(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIMin.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIMin(taskHandle, channel)


@catchDAQError
def DAQmxGetAICustomScaleName(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAICustomScaleName
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAICustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAICustomScaleName(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAICustomScaleName(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICustomScaleName(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAICustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAICustomScaleName(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICustomScaleName(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICustomScaleName(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICustomScaleName.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICustomScaleName(taskHandle, channel)


@catchDAQError
def DAQmxGetAIMeasType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIMeasType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIMeasType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIMeasType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIVoltageUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVoltageUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVoltageUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIVoltageUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVoltageUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVoltageUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVoltageUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIVoltageUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVoltageUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVoltageUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVoltageUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVoltageUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVoltagedBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVoltagedBRef(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVoltagedBRef.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIVoltagedBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVoltagedBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVoltagedBRef(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVoltagedBRef.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIVoltagedBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVoltagedBRef(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVoltagedBRef(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVoltagedBRef.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVoltagedBRef(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVoltageACRMSUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVoltageACRMSUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVoltageACRMSUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIVoltageACRMSUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVoltageACRMSUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVoltageACRMSUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVoltageACRMSUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIVoltageACRMSUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVoltageACRMSUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVoltageACRMSUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVoltageACRMSUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVoltageACRMSUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAITempUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAITempUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAITempUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAITempUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAITempUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAITempUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAITempUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAITempUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAITempUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAITempUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAITempUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAITempUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmcplType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmcplType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIThrmcplType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmcplType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmcplType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmcplType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIThrmcplType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmcplType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmcplType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmcplType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmcplType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmcplScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmcplScaleType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplScaleType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIThrmcplScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmcplScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmcplScaleType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmcplScaleType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIThrmcplScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmcplScaleType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmcplScaleType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmcplScaleType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmcplScaleType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmcplCJCSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmcplCJCSrc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplCJCSrc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIThrmcplCJCSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIThrmcplCJCVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmcplCJCVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplCJCVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmcplCJCVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmcplCJCVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmcplCJCVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmcplCJCVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmcplCJCVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmcplCJCVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmcplCJCVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmcplCJCVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmcplCJCVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmcplCJCChan(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIThrmcplCJCChan
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplCJCChan.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIThrmcplCJCChan(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxGetAIRTDType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRTDType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRTDType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRTDType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRTDType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRTDType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRTDType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIRTDType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRTDType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRTDType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRTDType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRTDType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRTDR0(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRTDR0(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRTDR0.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRTDR0(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRTDR0(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRTDR0(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRTDR0.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRTDR0(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRTDR0(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRTDR0(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRTDR0.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRTDR0(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRTDA(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRTDA(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRTDA.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRTDA(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRTDA(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRTDA(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRTDA.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRTDA(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRTDA(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRTDA(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRTDA.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRTDA(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRTDB(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRTDB(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRTDB.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRTDB(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRTDB(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRTDB(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRTDB.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRTDB(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRTDB(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRTDB(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRTDB.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRTDB(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRTDC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRTDC(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRTDC.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRTDC(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRTDC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRTDC(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRTDC.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRTDC(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRTDC(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRTDC(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRTDC.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRTDC(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmstrA(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmstrA(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmstrA.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmstrA(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmstrA(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmstrA(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmstrA.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmstrA(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmstrA(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmstrA(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmstrA.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmstrA(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmstrB(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmstrB(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmstrB.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmstrB(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmstrB(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmstrB(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmstrB.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmstrB(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmstrB(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmstrB(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmstrB.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmstrB(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmstrC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmstrC(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmstrC.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmstrC(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmstrC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmstrC(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmstrC.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmstrC(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmstrC(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmstrC(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmstrC.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmstrC(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmstrR1(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmstrR1(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmstrR1.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmstrR1(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmstrR1(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmstrR1(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmstrR1.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmstrR1(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmstrR1(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmstrR1(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmstrR1.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmstrR1(taskHandle, channel)


@catchDAQError
def DAQmxGetAIForceReadFromChan(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIForceReadFromChan(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIForceReadFromChan.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIForceReadFromChan(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIForceReadFromChan(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIForceReadFromChan(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIForceReadFromChan.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIForceReadFromChan(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIForceReadFromChan(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIForceReadFromChan(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIForceReadFromChan.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIForceReadFromChan(taskHandle, channel)


@catchDAQError
def DAQmxGetAICurrentUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAICurrentUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAICurrentUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAICurrentUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAICurrentUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICurrentUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAICurrentUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAICurrentUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICurrentUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICurrentUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICurrentUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICurrentUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAICurrentACRMSUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAICurrentACRMSUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAICurrentACRMSUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAICurrentACRMSUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAICurrentACRMSUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICurrentACRMSUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAICurrentACRMSUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAICurrentACRMSUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICurrentACRMSUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICurrentACRMSUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICurrentACRMSUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICurrentACRMSUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIStrainUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIStrainUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIStrainUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIStrainUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIStrainUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIStrainUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIStrainUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIStrainUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIStrainUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIStrainUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIStrainUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIStrainUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIStrainGageForceReadFromChan(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIStrainGageForceReadFromChan
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIStrainGageForceReadFromChan.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIStrainGageForceReadFromChan(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIStrainGageForceReadFromChan(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIStrainGageForceReadFromChan
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIStrainGageForceReadFromChan.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIStrainGageForceReadFromChan(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIStrainGageForceReadFromChan(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIStrainGageForceReadFromChan(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIStrainGageForceReadFromChan.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIStrainGageForceReadFromChan(taskHandle, channel)


@catchDAQError
def DAQmxGetAIStrainGageGageFactor(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIStrainGageGageFactor(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIStrainGageGageFactor.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIStrainGageGageFactor(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIStrainGageGageFactor(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIStrainGageGageFactor(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIStrainGageGageFactor.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIStrainGageGageFactor(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIStrainGageGageFactor(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIStrainGageGageFactor(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIStrainGageGageFactor.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIStrainGageGageFactor(taskHandle, channel)


@catchDAQError
def DAQmxGetAIStrainGagePoissonRatio(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIStrainGagePoissonRatio(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIStrainGagePoissonRatio.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIStrainGagePoissonRatio(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIStrainGagePoissonRatio(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIStrainGagePoissonRatio(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIStrainGagePoissonRatio.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIStrainGagePoissonRatio(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIStrainGagePoissonRatio(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIStrainGagePoissonRatio(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIStrainGagePoissonRatio.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIStrainGagePoissonRatio(taskHandle, channel)


@catchDAQError
def DAQmxGetAIStrainGageCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIStrainGageCfg(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIStrainGageCfg.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIStrainGageCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIStrainGageCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIStrainGageCfg(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIStrainGageCfg.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIStrainGageCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIStrainGageCfg(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIStrainGageCfg(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIStrainGageCfg.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIStrainGageCfg(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRosetteStrainGageRosetteType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRosetteStrainGageRosetteType
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRosetteStrainGageRosetteType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRosetteStrainGageRosetteType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIRosetteStrainGageOrientation(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRosetteStrainGageOrientation
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRosetteStrainGageOrientation.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRosetteStrainGageOrientation(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRosetteStrainGageOrientation(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRosetteStrainGageOrientation
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRosetteStrainGageOrientation.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRosetteStrainGageOrientation(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRosetteStrainGageOrientation(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRosetteStrainGageOrientation(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRosetteStrainGageOrientation.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRosetteStrainGageOrientation(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRosetteStrainGageStrainChans(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIRosetteStrainGageStrainChans
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIRosetteStrainGageStrainChans.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIRosetteStrainGageStrainChans(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxGetAIRosetteStrainGageRosetteMeasType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRosetteStrainGageRosetteMeasType
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRosetteStrainGageRosetteMeasType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRosetteStrainGageRosetteMeasType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRosetteStrainGageRosetteMeasType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRosetteStrainGageRosetteMeasType
    (TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRosetteStrainGageRosetteMeasType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIRosetteStrainGageRosetteMeasType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRosetteStrainGageRosetteMeasType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRosetteStrainGageRosetteMeasType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRosetteStrainGageRosetteMeasType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRosetteStrainGageRosetteMeasType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIResistanceUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIResistanceUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIResistanceUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIResistanceUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIResistanceUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIResistanceUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIResistanceUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIResistanceUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIResistanceUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIResistanceUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIResistanceUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIResistanceUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFreqUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFreqUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIFreqUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIFreqUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIFreqUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIFreqUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIFreqUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIFreqUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIFreqThreshVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFreqThreshVoltage(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFreqThreshVoltage.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIFreqThreshVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIFreqThreshVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIFreqThreshVoltage(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIFreqThreshVoltage.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIFreqThreshVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIFreqThreshVoltage(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIFreqThreshVoltage(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIFreqThreshVoltage.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIFreqThreshVoltage(taskHandle, channel)


@catchDAQError
def DAQmxGetAIFreqHyst(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFreqHyst(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFreqHyst.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIFreqHyst(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIFreqHyst(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIFreqHyst(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIFreqHyst.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIFreqHyst(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIFreqHyst(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIFreqHyst(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIFreqHyst.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIFreqHyst(taskHandle, channel)


@catchDAQError
def DAQmxGetAILVDTUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILVDTUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILVDTUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAILVDTUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILVDTUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILVDTUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILVDTUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAILVDTUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILVDTUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILVDTUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILVDTUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILVDTUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAILVDTSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILVDTSensitivity(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILVDTSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAILVDTSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILVDTSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILVDTSensitivity(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILVDTSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAILVDTSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILVDTSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILVDTSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILVDTSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILVDTSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAILVDTSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILVDTSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAILVDTSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILVDTSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILVDTSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAILVDTSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILVDTSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILVDTSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILVDTSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRVDTUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRVDTUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRVDTUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRVDTUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRVDTUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRVDTUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRVDTUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIRVDTUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRVDTUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRVDTUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRVDTUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRVDTUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRVDTSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRVDTSensitivity(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRVDTSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRVDTSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRVDTSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRVDTSensitivity(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRVDTSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRVDTSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRVDTSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRVDTSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRVDTSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRVDTSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRVDTSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRVDTSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRVDTSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRVDTSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRVDTSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIRVDTSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRVDTSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRVDTSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRVDTSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRVDTSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIEddyCurrentProxProbeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIEddyCurrentProxProbeUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIEddyCurrentProxProbeUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIEddyCurrentProxProbeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIEddyCurrentProxProbeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIEddyCurrentProxProbeUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIEddyCurrentProxProbeUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIEddyCurrentProxProbeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIEddyCurrentProxProbeUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIEddyCurrentProxProbeUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIEddyCurrentProxProbeUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIEddyCurrentProxProbeUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIEddyCurrentProxProbeSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIEddyCurrentProxProbeSensitivity
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIEddyCurrentProxProbeSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIEddyCurrentProxProbeSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIEddyCurrentProxProbeSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIEddyCurrentProxProbeSensitivity
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIEddyCurrentProxProbeSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIEddyCurrentProxProbeSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIEddyCurrentProxProbeSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIEddyCurrentProxProbeSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIEddyCurrentProxProbeSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIEddyCurrentProxProbeSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIEddyCurrentProxProbeSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIEddyCurrentProxProbeSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIEddyCurrentProxProbeSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIEddyCurrentProxProbeSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIEddyCurrentProxProbeSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIEddyCurrentProxProbeSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIEddyCurrentProxProbeSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAISoundPressureMaxSoundPressureLvl(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAISoundPressureMaxSoundPressureLvl
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAISoundPressureMaxSoundPressureLvl.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAISoundPressureMaxSoundPressureLvl(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAISoundPressureMaxSoundPressureLvl(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAISoundPressureMaxSoundPressureLvl
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAISoundPressureMaxSoundPressureLvl.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAISoundPressureMaxSoundPressureLvl(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAISoundPressureMaxSoundPressureLvl(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAISoundPressureMaxSoundPressureLvl(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAISoundPressureMaxSoundPressureLvl.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAISoundPressureMaxSoundPressureLvl(taskHandle, channel)


@catchDAQError
def DAQmxGetAISoundPressureUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAISoundPressureUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAISoundPressureUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAISoundPressureUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAISoundPressureUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAISoundPressureUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAISoundPressureUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAISoundPressureUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAISoundPressureUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAISoundPressureUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAISoundPressureUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAISoundPressureUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAISoundPressuredBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAISoundPressuredBRef(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAISoundPressuredBRef.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAISoundPressuredBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAISoundPressuredBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAISoundPressuredBRef(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAISoundPressuredBRef.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAISoundPressuredBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAISoundPressuredBRef(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAISoundPressuredBRef(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAISoundPressuredBRef.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAISoundPressuredBRef(taskHandle, channel)


@catchDAQError
def DAQmxGetAIMicrophoneSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIMicrophoneSensitivity(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIMicrophoneSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIMicrophoneSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIMicrophoneSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIMicrophoneSensitivity(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIMicrophoneSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIMicrophoneSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIMicrophoneSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIMicrophoneSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIMicrophoneSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIMicrophoneSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAccelUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAccelUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAccelUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIAccelUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAccelUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAccelUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAccelUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIAccelUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAccelUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAccelUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAccelUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAccelUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAcceldBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAcceldBRef(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAcceldBRef.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIAcceldBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAcceldBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAcceldBRef(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAcceldBRef.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIAcceldBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAcceldBRef(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAcceldBRef(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAcceldBRef.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAcceldBRef(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAccelSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAccelSensitivity(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAccelSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIAccelSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAccelSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAccelSensitivity(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAccelSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIAccelSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAccelSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAccelSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAccelSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAccelSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAccelSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAccelSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAccelSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIAccelSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAccelSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAccelSensitivityUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAccelSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIAccelSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAccelSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAccelSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAccelSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAccelSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVelocityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVelocityUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVelocityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIVelocityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVelocityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVelocityUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVelocityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIVelocityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVelocityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVelocityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVelocityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVelocityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVelocityIEPESensordBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVelocityIEPESensordBRef(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVelocityIEPESensordBRef.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIVelocityIEPESensordBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVelocityIEPESensordBRef(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVelocityIEPESensordBRef(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVelocityIEPESensordBRef.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIVelocityIEPESensordBRef(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVelocityIEPESensordBRef(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVelocityIEPESensordBRef(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVelocityIEPESensordBRef.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVelocityIEPESensordBRef(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVelocityIEPESensorSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVelocityIEPESensorSensitivity
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVelocityIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIVelocityIEPESensorSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVelocityIEPESensorSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVelocityIEPESensorSensitivity
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVelocityIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIVelocityIEPESensorSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVelocityIEPESensorSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVelocityIEPESensorSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVelocityIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVelocityIEPESensorSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIVelocityIEPESensorSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIVelocityIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIVelocityIEPESensorSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIVelocityIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIVelocityIEPESensorSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIVelocityIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIVelocityIEPESensorSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIForceUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIForceUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIForceUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIForceUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIForceUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIForceUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIForceUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIForceUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIForceUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIForceUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIForceUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIForceUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIForceIEPESensorSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIForceIEPESensorSensitivity
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIForceIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIForceIEPESensorSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIForceIEPESensorSensitivity(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIForceIEPESensorSensitivity(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIForceIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIForceIEPESensorSensitivity(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIForceIEPESensorSensitivity(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIForceIEPESensorSensitivity(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIForceIEPESensorSensitivity.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIForceIEPESensorSensitivity(taskHandle, channel)


@catchDAQError
def DAQmxGetAIForceIEPESensorSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIForceIEPESensorSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIForceIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIForceIEPESensorSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIForceIEPESensorSensitivityUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIForceIEPESensorSensitivityUnits
    (TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIForceIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIForceIEPESensorSensitivityUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIForceIEPESensorSensitivityUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIForceIEPESensorSensitivityUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIForceIEPESensorSensitivityUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIForceIEPESensorSensitivityUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIPressureUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIPressureUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIPressureUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIPressureUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIPressureUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIPressureUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIPressureUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIPressureUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIPressureUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIPressureUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIPressureUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIPressureUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAITorqueUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAITorqueUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAITorqueUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAITorqueUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAITorqueUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAITorqueUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAITorqueUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAITorqueUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAITorqueUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAITorqueUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAITorqueUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAITorqueUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeElectricalUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeElectricalUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeElectricalUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeElectricalUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeElectricalUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeElectricalUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeElectricalUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeElectricalUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeElectricalUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeElectricalUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeElectricalUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeElectricalUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgePhysicalUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgePhysicalUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgePhysicalUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgePhysicalUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgePhysicalUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgePhysicalUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgePhysicalUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgePhysicalUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgePhysicalUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgePhysicalUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgePhysicalUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgePhysicalUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeScaleType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeScaleType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeScaleType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeScaleType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeScaleType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeScaleType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeScaleType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeScaleType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeTwoPointLinFirstElectricalVal
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTwoPointLinFirstElectricalVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeTwoPointLinFirstElectricalVal
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTwoPointLinFirstElectricalVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTwoPointLinFirstElectricalVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTwoPointLinFirstElectricalVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTwoPointLinFirstElectricalVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeTwoPointLinSecondElectricalVal
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTwoPointLinSecondElectricalVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeTwoPointLinSecondElectricalVal
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTwoPointLinSecondElectricalVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTwoPointLinSecondElectricalVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTwoPointLinSecondElectricalVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTwoPointLinSecondElectricalVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTableElectricalVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIBridgeTableElectricalVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTableElectricalVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIBridgeTableElectricalVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIBridgeTableElectricalVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIBridgeTableElectricalVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTableElectricalVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIBridgeTableElectricalVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIBridgeTableElectricalVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTableElectricalVals(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTableElectricalVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTableElectricalVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeTablePhysicalVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIBridgeTablePhysicalVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeTablePhysicalVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIBridgeTablePhysicalVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIBridgeTablePhysicalVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIBridgeTablePhysicalVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeTablePhysicalVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIBridgeTablePhysicalVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIBridgeTablePhysicalVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeTablePhysicalVals(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeTablePhysicalVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeTablePhysicalVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgePolyForwardCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIBridgePolyForwardCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgePolyForwardCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIBridgePolyForwardCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIBridgePolyForwardCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIBridgePolyForwardCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgePolyForwardCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIBridgePolyForwardCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIBridgePolyForwardCoeff(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgePolyForwardCoeff(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgePolyForwardCoeff.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgePolyForwardCoeff(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgePolyReverseCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIBridgePolyReverseCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgePolyReverseCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIBridgePolyReverseCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIBridgePolyReverseCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIBridgePolyReverseCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgePolyReverseCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIBridgePolyReverseCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIBridgePolyReverseCoeff(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgePolyReverseCoeff(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgePolyReverseCoeff.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgePolyReverseCoeff(taskHandle, channel)


@catchDAQError
def DAQmxGetAIIsTEDS(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIIsTEDS(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIIsTEDS.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIIsTEDS(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAITEDSUnits(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAITEDSUnits
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAITEDSUnits.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAITEDSUnits(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxGetAICoupling(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAICoupling(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAICoupling.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAICoupling(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAICoupling(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICoupling(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAICoupling.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAICoupling(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICoupling(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICoupling(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICoupling.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICoupling(taskHandle, channel)


@catchDAQError
def DAQmxGetAIImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIImpedance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIImpedance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIImpedance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIImpedance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIImpedance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIImpedance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIImpedance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIImpedance(taskHandle, channel)


@catchDAQError
def DAQmxGetAITermCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAITermCfg(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAITermCfg.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAITermCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAITermCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAITermCfg(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAITermCfg.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAITermCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAITermCfg(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAITermCfg(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAITermCfg.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAITermCfg(taskHandle, channel)


@catchDAQError
def DAQmxGetAIInputSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIInputSrc(TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIInputSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIInputSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAIInputSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIInputSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIInputSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAIInputSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIInputSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIInputSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIInputSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIInputSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAIResistanceCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIResistanceCfg(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIResistanceCfg.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIResistanceCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIResistanceCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIResistanceCfg(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIResistanceCfg.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIResistanceCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIResistanceCfg(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIResistanceCfg(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIResistanceCfg.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIResistanceCfg(taskHandle, channel)


@catchDAQError
def DAQmxGetAILeadWireResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILeadWireResistance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILeadWireResistance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAILeadWireResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILeadWireResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILeadWireResistance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILeadWireResistance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAILeadWireResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILeadWireResistance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILeadWireResistance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILeadWireResistance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILeadWireResistance(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeCfg(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeCfg.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeCfg(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeCfg.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeCfg(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeCfg(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeCfg.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeCfg(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeNomResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeNomResistance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeNomResistance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeNomResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeNomResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeNomResistance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeNomResistance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeNomResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeNomResistance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeNomResistance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeNomResistance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeNomResistance(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeInitialVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeInitialVoltage(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeInitialVoltage.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeInitialVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeInitialVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeInitialVoltage(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeInitialVoltage.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeInitialVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeInitialVoltage(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeInitialVoltage(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeInitialVoltage.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeInitialVoltage(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeInitialRatio(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeInitialRatio(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeInitialRatio.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeInitialRatio(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeInitialRatio(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeInitialRatio(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeInitialRatio.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeInitialRatio(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeInitialRatio(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeInitialRatio(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeInitialRatio.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeInitialRatio(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeShuntCalEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeShuntCalEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeShuntCalEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIBridgeShuntCalEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeShuntCalEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeShuntCalEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeShuntCalEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIBridgeShuntCalEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeShuntCalEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeShuntCalEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeShuntCalEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeShuntCalEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeShuntCalSelect(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeShuntCalSelect(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeShuntCalSelect.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeShuntCalSelect(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeShuntCalSelect(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeShuntCalSelect(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeShuntCalSelect.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeShuntCalSelect(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeShuntCalSelect(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeShuntCalSelect(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeShuntCalSelect.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeShuntCalSelect(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeShuntCalGainAdjust(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeShuntCalGainAdjust(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeShuntCalGainAdjust.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeShuntCalGainAdjust(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeShuntCalGainAdjust(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeShuntCalGainAdjust(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeShuntCalGainAdjust.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeShuntCalGainAdjust(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeShuntCalGainAdjust(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeShuntCalGainAdjust(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeShuntCalGainAdjust.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeShuntCalGainAdjust(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeShuntCalShuntCalAResistance
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeShuntCalShuntCalAResistance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeShuntCalShuntCalAResistance
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeShuntCalShuntCalAResistance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeShuntCalShuntCalAResistance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeShuntCalShuntCalAResistance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeShuntCalShuntCalAResistance(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeShuntCalShuntCalAActualResistance
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeShuntCalShuntCalAActualResistance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeShuntCalShuntCalAActualResistance
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeShuntCalShuntCalAActualResistance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeShuntCalShuntCalAActualResistance
    (TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeShuntCalShuntCalAActualResistance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeShuntCalShuntCalAActualResistance(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeBalanceCoarsePot(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeBalanceCoarsePot(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeBalanceCoarsePot.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeBalanceCoarsePot(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeBalanceCoarsePot(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeBalanceCoarsePot(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeBalanceCoarsePot.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeBalanceCoarsePot(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeBalanceCoarsePot(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeBalanceCoarsePot(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeBalanceCoarsePot.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeBalanceCoarsePot(taskHandle, channel)


@catchDAQError
def DAQmxGetAIBridgeBalanceFinePot(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIBridgeBalanceFinePot(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIBridgeBalanceFinePot.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIBridgeBalanceFinePot(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIBridgeBalanceFinePot(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIBridgeBalanceFinePot(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIBridgeBalanceFinePot.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIBridgeBalanceFinePot(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIBridgeBalanceFinePot(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIBridgeBalanceFinePot(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIBridgeBalanceFinePot.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIBridgeBalanceFinePot(taskHandle, channel)


@catchDAQError
def DAQmxGetAICurrentShuntLoc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAICurrentShuntLoc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAICurrentShuntLoc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAICurrentShuntLoc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAICurrentShuntLoc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICurrentShuntLoc(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAICurrentShuntLoc.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAICurrentShuntLoc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICurrentShuntLoc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICurrentShuntLoc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICurrentShuntLoc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICurrentShuntLoc(taskHandle, channel)


@catchDAQError
def DAQmxGetAICurrentShuntResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAICurrentShuntResistance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAICurrentShuntResistance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAICurrentShuntResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAICurrentShuntResistance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAICurrentShuntResistance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAICurrentShuntResistance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAICurrentShuntResistance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAICurrentShuntResistance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAICurrentShuntResistance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAICurrentShuntResistance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAICurrentShuntResistance(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitSrc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitSrc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIExcitSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitSrc(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitSrc.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIExcitSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIExcitVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIExcitVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitUseForScaling(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitUseForScaling(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitUseForScaling.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIExcitUseForScaling(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitUseForScaling(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitUseForScaling(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitUseForScaling.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIExcitUseForScaling(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitUseForScaling(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitUseForScaling(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitUseForScaling.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitUseForScaling(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitUseMultiplexed(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitUseMultiplexed(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitUseMultiplexed.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIExcitUseMultiplexed(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitUseMultiplexed(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitUseMultiplexed(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitUseMultiplexed.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIExcitUseMultiplexed(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitUseMultiplexed(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitUseMultiplexed(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitUseMultiplexed.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitUseMultiplexed(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitActualVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitActualVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitActualVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIExcitActualVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitActualVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitActualVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitActualVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIExcitActualVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitActualVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitActualVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitActualVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitActualVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitDCorAC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitDCorAC(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitDCorAC.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIExcitDCorAC(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitDCorAC(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitDCorAC(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitDCorAC.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIExcitDCorAC(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitDCorAC(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitDCorAC(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitDCorAC.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitDCorAC(taskHandle, channel)


@catchDAQError
def DAQmxGetAIExcitVoltageOrCurrent(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIExcitVoltageOrCurrent(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIExcitVoltageOrCurrent.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIExcitVoltageOrCurrent(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIExcitVoltageOrCurrent(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIExcitVoltageOrCurrent(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIExcitVoltageOrCurrent.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIExcitVoltageOrCurrent(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIExcitVoltageOrCurrent(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIExcitVoltageOrCurrent(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIExcitVoltageOrCurrent.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIExcitVoltageOrCurrent(taskHandle, channel)


@catchDAQError
def DAQmxGetAIACExcitFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIACExcitFreq(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIACExcitFreq.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIACExcitFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIACExcitFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIACExcitFreq(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIACExcitFreq.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIACExcitFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIACExcitFreq(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIACExcitFreq(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIACExcitFreq.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIACExcitFreq(taskHandle, channel)


@catchDAQError
def DAQmxGetAIACExcitSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIACExcitSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIACExcitSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIACExcitSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIACExcitSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIACExcitSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIACExcitSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIACExcitSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIACExcitSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIACExcitSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIACExcitSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIACExcitSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIACExcitWireMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIACExcitWireMode(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIACExcitWireMode.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIACExcitWireMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIACExcitWireMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIACExcitWireMode(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIACExcitWireMode.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIACExcitWireMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIACExcitWireMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIACExcitWireMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIACExcitWireMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIACExcitWireMode(taskHandle, channel)


@catchDAQError
def DAQmxGetAIOpenThrmcplDetectEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIOpenThrmcplDetectEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIOpenThrmcplDetectEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIOpenThrmcplDetectEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIOpenThrmcplDetectEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIOpenThrmcplDetectEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIOpenThrmcplDetectEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIOpenThrmcplDetectEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIOpenThrmcplDetectEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIOpenThrmcplDetectEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIOpenThrmcplDetectEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIOpenThrmcplDetectEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIThrmcplLeadOffsetVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIThrmcplLeadOffsetVoltage(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIThrmcplLeadOffsetVoltage.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIThrmcplLeadOffsetVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIThrmcplLeadOffsetVoltage(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIThrmcplLeadOffsetVoltage(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIThrmcplLeadOffsetVoltage.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIThrmcplLeadOffsetVoltage(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIThrmcplLeadOffsetVoltage(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIThrmcplLeadOffsetVoltage(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIThrmcplLeadOffsetVoltage.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIThrmcplLeadOffsetVoltage(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAtten(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAtten(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAtten.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIAtten(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAtten(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAtten(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAtten.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIAtten(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAtten(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAtten(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAtten.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAtten(taskHandle, channel)


@catchDAQError
def DAQmxGetAIProbeAtten(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIProbeAtten(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIProbeAtten.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIProbeAtten(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIProbeAtten(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIProbeAtten(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIProbeAtten.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIProbeAtten(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIProbeAtten(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIProbeAtten(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIProbeAtten.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIProbeAtten(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAILowpassEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAILowpassEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassCutoffFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassCutoffFreq(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassCutoffFreq.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAILowpassCutoffFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassCutoffFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassCutoffFreq(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassCutoffFreq.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAILowpassCutoffFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassCutoffFreq(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassCutoffFreq(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassCutoffFreq.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassCutoffFreq(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassSwitchCapClkSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassSwitchCapClkSrc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassSwitchCapClkSrc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAILowpassSwitchCapClkSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassSwitchCapClkSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassSwitchCapClkSrc(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassSwitchCapClkSrc.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAILowpassSwitchCapClkSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassSwitchCapClkSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassSwitchCapClkSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassSwitchCapClkSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassSwitchCapClkSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassSwitchCapExtClkFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassSwitchCapExtClkFreq
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassSwitchCapExtClkFreq.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAILowpassSwitchCapExtClkFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassSwitchCapExtClkFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassSwitchCapExtClkFreq(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassSwitchCapExtClkFreq.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAILowpassSwitchCapExtClkFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassSwitchCapExtClkFreq(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassSwitchCapExtClkFreq(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassSwitchCapExtClkFreq.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassSwitchCapExtClkFreq(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassSwitchCapExtClkDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassSwitchCapExtClkDiv(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassSwitchCapExtClkDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAILowpassSwitchCapExtClkDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassSwitchCapExtClkDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassSwitchCapExtClkDiv(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassSwitchCapExtClkDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAILowpassSwitchCapExtClkDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassSwitchCapExtClkDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassSwitchCapExtClkDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassSwitchCapExtClkDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassSwitchCapExtClkDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetAILowpassSwitchCapOutClkDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILowpassSwitchCapOutClkDiv(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILowpassSwitchCapOutClkDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAILowpassSwitchCapOutClkDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILowpassSwitchCapOutClkDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILowpassSwitchCapOutClkDiv(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILowpassSwitchCapOutClkDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAILowpassSwitchCapOutClkDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILowpassSwitchCapOutClkDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILowpassSwitchCapOutClkDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILowpassSwitchCapOutClkDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILowpassSwitchCapOutClkDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetAIFilterDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFilterDelay(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFilterDelay.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIFilterDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIFilterDelayUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFilterDelayUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFilterDelayUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIFilterDelayUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIFilterDelayUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIFilterDelayUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIFilterDelayUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIFilterDelayUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIFilterDelayUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIFilterDelayUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIFilterDelayUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIFilterDelayUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRemoveFilterDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRemoveFilterDelay(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRemoveFilterDelay.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIRemoveFilterDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRemoveFilterDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRemoveFilterDelay(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRemoveFilterDelay.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIRemoveFilterDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRemoveFilterDelay(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRemoveFilterDelay(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRemoveFilterDelay.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRemoveFilterDelay(taskHandle, channel)


@catchDAQError
def DAQmxGetAIFilterDelayAdjustment(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIFilterDelayAdjustment(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIFilterDelayAdjustment(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIFilterDelayAdjustment(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIFilterDelayAdjustment(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIFilterDelayAdjustment(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIFilterDelayAdjustment(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIFilterDelayAdjustment(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIFilterDelayAdjustment(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAveragingWinSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAveragingWinSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAveragingWinSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIAveragingWinSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAveragingWinSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAveragingWinSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAveragingWinSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIAveragingWinSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAveragingWinSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAveragingWinSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAveragingWinSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAveragingWinSize(taskHandle, channel)


@catchDAQError
def DAQmxGetAIResolutionUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIResolutionUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIResolutionUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIResolutionUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIResolution(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIResolution(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIResolution.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIResolution(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIRawSampSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRawSampSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRawSampSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIRawSampSize(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIRawSampJustification(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRawSampJustification(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRawSampJustification.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRawSampJustification(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIADCTimingMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIADCTimingMode(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIADCTimingMode.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIADCTimingMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIADCTimingMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIADCTimingMode(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIADCTimingMode.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIADCTimingMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIADCTimingMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIADCTimingMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIADCTimingMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIADCTimingMode(taskHandle, channel)


@catchDAQError
def DAQmxGetAIADCCustomTimingMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIADCCustomTimingMode(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIADCCustomTimingMode.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIADCCustomTimingMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIADCCustomTimingMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIADCCustomTimingMode(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIADCCustomTimingMode.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIADCCustomTimingMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIADCCustomTimingMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIADCCustomTimingMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIADCCustomTimingMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIADCCustomTimingMode(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDitherEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIDitherEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIDitherEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIDitherEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIDitherEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIDitherEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIDitherEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIDitherEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIDitherEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIDitherEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIDitherEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIDitherEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalHasValidCalInfo(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIChanCalHasValidCalInfo(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalHasValidCalInfo.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIChanCalHasValidCalInfo(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAIChanCalEnableCal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIChanCalEnableCal(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalEnableCal.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIChanCalEnableCal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIChanCalEnableCal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIChanCalEnableCal(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalEnableCal.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIChanCalEnableCal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIChanCalEnableCal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalEnableCal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalEnableCal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalEnableCal(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalApplyCalIfExp(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIChanCalApplyCalIfExp(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalApplyCalIfExp.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIChanCalApplyCalIfExp(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIChanCalApplyCalIfExp(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIChanCalApplyCalIfExp(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalApplyCalIfExp.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIChanCalApplyCalIfExp(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIChanCalApplyCalIfExp(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalApplyCalIfExp(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalApplyCalIfExp.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalApplyCalIfExp(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIChanCalScaleType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalScaleType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIChanCalScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIChanCalScaleType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIChanCalScaleType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalScaleType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIChanCalScaleType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIChanCalScaleType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalScaleType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalScaleType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalScaleType(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalTablePreScaledVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalTablePreScaledVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalTablePreScaledVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalTablePreScaledVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalTablePreScaledVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalTablePreScaledVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalTablePreScaledVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalTablePreScaledVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalTablePreScaledVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalTablePreScaledVals(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalTablePreScaledVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalTablePreScaledVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalTableScaledVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalTableScaledVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalTableScaledVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalTableScaledVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalTableScaledVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalTableScaledVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalTableScaledVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalTableScaledVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalTableScaledVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalTableScaledVals
    (TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalTableScaledVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalTableScaledVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalPolyForwardCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalPolyForwardCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalPolyForwardCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalPolyForwardCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalPolyForwardCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalPolyForwardCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalPolyForwardCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalPolyForwardCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalPolyForwardCoeff(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalPolyForwardCoeff(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalPolyForwardCoeff.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalPolyForwardCoeff(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalPolyReverseCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalPolyReverseCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalPolyReverseCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalPolyReverseCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalPolyReverseCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalPolyReverseCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalPolyReverseCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalPolyReverseCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalPolyReverseCoeff(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalPolyReverseCoeff(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalPolyReverseCoeff.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalPolyReverseCoeff(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalOperatorName(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIChanCalOperatorName
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalOperatorName.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIChanCalOperatorName(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAIChanCalOperatorName(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIChanCalOperatorName(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalOperatorName.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAIChanCalOperatorName(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIChanCalOperatorName(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalOperatorName(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalOperatorName.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalOperatorName(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalDesc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIChanCalDesc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalDesc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIChanCalDesc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAIChanCalDesc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIChanCalDesc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalDesc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAIChanCalDesc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIChanCalDesc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalDesc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalDesc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalDesc(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalVerifRefVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalVerifRefVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalVerifRefVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalVerifRefVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalVerifRefVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalVerifRefVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalVerifRefVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalVerifRefVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalVerifRefVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalVerifRefVals(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalVerifRefVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalVerifRefVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIChanCalVerifAcqVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIChanCalVerifAcqVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIChanCalVerifAcqVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIChanCalVerifAcqVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxSetAIChanCalVerifAcqVals(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetAIChanCalVerifAcqVals
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetAIChanCalVerifAcqVals.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetAIChanCalVerifAcqVals(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxResetAIChanCalVerifAcqVals(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIChanCalVerifAcqVals(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIChanCalVerifAcqVals.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIChanCalVerifAcqVals(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRngHigh(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRngHigh(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRngHigh.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRngHigh(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRngHigh(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRngHigh(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRngHigh.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRngHigh(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRngHigh(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRngHigh(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRngHigh.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRngHigh(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRngLow(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRngLow(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRngLow.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIRngLow(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRngLow(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRngLow(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRngLow.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIRngLow(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRngLow(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRngLow(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRngLow.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRngLow(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDCOffset(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIDCOffset(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIDCOffset.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIDCOffset(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIDCOffset(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIDCOffset(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIDCOffset.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIDCOffset(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIDCOffset(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIDCOffset(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIDCOffset.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIDCOffset(taskHandle, channel)


@catchDAQError
def DAQmxGetAIGain(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIGain(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIGain.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIGain(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIGain(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIGain(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIGain.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIGain(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIGain(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIGain(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIGain.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIGain(taskHandle, channel)


@catchDAQError
def DAQmxGetAISampAndHoldEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAISampAndHoldEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAISampAndHoldEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAISampAndHoldEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAISampAndHoldEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAISampAndHoldEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAISampAndHoldEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAISampAndHoldEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAISampAndHoldEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAISampAndHoldEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAISampAndHoldEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAISampAndHoldEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIAutoZeroMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIAutoZeroMode(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIAutoZeroMode.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIAutoZeroMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIAutoZeroMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIAutoZeroMode(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIAutoZeroMode.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIAutoZeroMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIAutoZeroMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIAutoZeroMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIAutoZeroMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIAutoZeroMode(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIDataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIDataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIDataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIDataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIDataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIDataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIDataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIDataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIDataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIDataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIDataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIDataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIDataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIDataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDataXferCustomThreshold(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIDataXferCustomThreshold(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIDataXferCustomThreshold.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIDataXferCustomThreshold(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIDataXferCustomThreshold(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIDataXferCustomThreshold(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIDataXferCustomThreshold.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIDataXferCustomThreshold(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIDataXferCustomThreshold(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIDataXferCustomThreshold(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIDataXferCustomThreshold.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIDataXferCustomThreshold(taskHandle, channel)


@catchDAQError
def DAQmxGetAIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetAIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetAIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAIRawDataCompressionType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIRawDataCompressionType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIRawDataCompressionType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIRawDataCompressionType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIRawDataCompressionType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIRawDataCompressionType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIRawDataCompressionType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIRawDataCompressionType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIRawDataCompressionType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIRawDataCompressionType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIRawDataCompressionType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIRawDataCompressionType(taskHandle, channel)


@catchDAQError
def DAQmxGetAILossyLSBRemovalCompressedSampSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAILossyLSBRemovalCompressedSampSize
    (TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAILossyLSBRemovalCompressedSampSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAILossyLSBRemovalCompressedSampSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAILossyLSBRemovalCompressedSampSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAILossyLSBRemovalCompressedSampSize
    (TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAILossyLSBRemovalCompressedSampSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAILossyLSBRemovalCompressedSampSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAILossyLSBRemovalCompressedSampSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAILossyLSBRemovalCompressedSampSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAILossyLSBRemovalCompressedSampSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAILossyLSBRemovalCompressedSampSize(taskHandle, channel)


@catchDAQError
def DAQmxGetAIDevScalingCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAIDevScalingCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAIDevScalingCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAIDevScalingCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetAIEnhancedAliasRejectionEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAIEnhancedAliasRejectionEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIEnhancedAliasRejectionEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIEnhancedAliasRejectionEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAIEnhancedAliasRejectionEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAIEnhancedAliasRejectionEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIEnhancedAliasRejectionEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIEnhancedAliasRejectionEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAIEnhancedAliasRejectionEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAIEnhancedAliasRejectionEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIEnhancedAliasRejectionEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIEnhancedAliasRejectionEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAOMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOMax(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOMax.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOMax(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOMax(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOMax.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOMax(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOMax(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOMax(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOMax.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOMax(taskHandle, channel)


@catchDAQError
def DAQmxGetAOMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOMin(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOMin.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOMin(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOMin(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOMin.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOMin(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOMin(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOMin(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOMin.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOMin(taskHandle, channel)


@catchDAQError
def DAQmxGetAOCustomScaleName(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAOCustomScaleName
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAOCustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAOCustomScaleName(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAOCustomScaleName(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOCustomScaleName(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAOCustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAOCustomScaleName(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOCustomScaleName(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOCustomScaleName(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOCustomScaleName.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOCustomScaleName(taskHandle, channel)


@catchDAQError
def DAQmxGetAOOutputType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOOutputType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOOutputType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOOutputType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAOVoltageUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOVoltageUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOVoltageUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOVoltageUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOVoltageUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOVoltageUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOVoltageUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOVoltageUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOVoltageUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOVoltageUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOVoltageUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOVoltageUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAOVoltageCurrentLimit(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOVoltageCurrentLimit(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOVoltageCurrentLimit.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOVoltageCurrentLimit(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOVoltageCurrentLimit(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOVoltageCurrentLimit(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOVoltageCurrentLimit.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOVoltageCurrentLimit(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOVoltageCurrentLimit(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOVoltageCurrentLimit(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOVoltageCurrentLimit.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOVoltageCurrentLimit(taskHandle, channel)


@catchDAQError
def DAQmxGetAOCurrentUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOCurrentUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOCurrentUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOCurrentUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOCurrentUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOCurrentUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOCurrentUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOCurrentUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOCurrentUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOCurrentUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOCurrentUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOCurrentUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOFuncGenType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOFuncGenType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenType(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenFreq(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenFreq.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFuncGenFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenFreq(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenFreq.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFuncGenFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenFreq(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenFreq(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenFreq.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenFreq(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenAmplitude(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenAmplitude(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenAmplitude.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFuncGenAmplitude(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenAmplitude(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenAmplitude(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenAmplitude.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFuncGenAmplitude(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenAmplitude(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenAmplitude(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenAmplitude.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenAmplitude(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenOffset(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenOffset(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenOffset.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFuncGenOffset(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenOffset(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenOffset(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenOffset.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFuncGenOffset(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenOffset(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenOffset(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenOffset.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenOffset(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenSquareDutyCycle(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenSquareDutyCycle(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenSquareDutyCycle.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFuncGenSquareDutyCycle(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenSquareDutyCycle(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenSquareDutyCycle(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenSquareDutyCycle.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFuncGenSquareDutyCycle(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenSquareDutyCycle(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenSquareDutyCycle(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenSquareDutyCycle.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenSquareDutyCycle(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenModulationType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenModulationType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenModulationType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOFuncGenModulationType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenModulationType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenModulationType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenModulationType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOFuncGenModulationType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenModulationType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenModulationType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenModulationType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenModulationType(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFuncGenFMDeviation(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFuncGenFMDeviation(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFuncGenFMDeviation.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFuncGenFMDeviation(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFuncGenFMDeviation(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFuncGenFMDeviation(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFuncGenFMDeviation.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFuncGenFMDeviation(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFuncGenFMDeviation(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFuncGenFMDeviation(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFuncGenFMDeviation.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFuncGenFMDeviation(taskHandle, channel)


@catchDAQError
def DAQmxGetAOOutputImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOOutputImpedance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOOutputImpedance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOOutputImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOOutputImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOOutputImpedance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOOutputImpedance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOOutputImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOOutputImpedance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOOutputImpedance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOOutputImpedance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOOutputImpedance(taskHandle, channel)


@catchDAQError
def DAQmxGetAOLoadImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOLoadImpedance(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOLoadImpedance.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOLoadImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOLoadImpedance(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOLoadImpedance(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOLoadImpedance.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOLoadImpedance(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOLoadImpedance(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOLoadImpedance(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOLoadImpedance.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOLoadImpedance(taskHandle, channel)


@catchDAQError
def DAQmxGetAOIdleOutputBehavior(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOIdleOutputBehavior(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOIdleOutputBehavior.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOIdleOutputBehavior(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOIdleOutputBehavior(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOIdleOutputBehavior(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOIdleOutputBehavior.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOIdleOutputBehavior(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOIdleOutputBehavior(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOIdleOutputBehavior(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOIdleOutputBehavior.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOIdleOutputBehavior(taskHandle, channel)


@catchDAQError
def DAQmxGetAOTermCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOTermCfg(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOTermCfg.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOTermCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOTermCfg(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOTermCfg(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOTermCfg.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOTermCfg(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOTermCfg(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOTermCfg(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOTermCfg.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOTermCfg(taskHandle, channel)


@catchDAQError
def DAQmxGetAOResolutionUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOResolutionUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOResolutionUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOResolutionUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOResolutionUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOResolutionUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOResolutionUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOResolutionUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOResolutionUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOResolutionUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOResolutionUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOResolutionUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAOResolution(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOResolution(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOResolution.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOResolution(taskHandle, channel, data)


@catchDAQError
def DAQmxGetAODACRngHigh(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRngHigh(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRngHigh.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAODACRngHigh(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRngHigh(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRngHigh(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRngHigh.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAODACRngHigh(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRngHigh(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRngHigh(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRngHigh.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRngHigh(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRngLow(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRngLow(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRngLow.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAODACRngLow(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRngLow(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRngLow(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRngLow.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAODACRngLow(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRngLow(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRngLow(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRngLow.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRngLow(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRefConnToGnd(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRefConnToGnd(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRefConnToGnd.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAODACRefConnToGnd(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRefConnToGnd(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRefConnToGnd(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRefConnToGnd.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAODACRefConnToGnd(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRefConnToGnd(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRefConnToGnd(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRefConnToGnd.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRefConnToGnd(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRefAllowConnToGnd(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRefAllowConnToGnd(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRefAllowConnToGnd.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAODACRefAllowConnToGnd(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRefAllowConnToGnd(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRefAllowConnToGnd(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRefAllowConnToGnd.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAODACRefAllowConnToGnd(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRefAllowConnToGnd(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRefAllowConnToGnd(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRefAllowConnToGnd.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRefAllowConnToGnd(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRefSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRefSrc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRefSrc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAODACRefSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRefSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRefSrc(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRefSrc.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAODACRefSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRefSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRefSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRefSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRefSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRefExtSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAODACRefExtSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRefExtSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAODACRefExtSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAODACRefExtSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRefExtSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRefExtSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAODACRefExtSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRefExtSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRefExtSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRefExtSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRefExtSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACRefVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACRefVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACRefVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAODACRefVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACRefVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACRefVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACRefVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAODACRefVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACRefVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACRefVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACRefVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACRefVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACOffsetSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACOffsetSrc(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACOffsetSrc.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAODACOffsetSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACOffsetSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACOffsetSrc(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACOffsetSrc.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAODACOffsetSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACOffsetSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACOffsetSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACOffsetSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACOffsetSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACOffsetExtSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetAODACOffsetExtSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAODACOffsetExtSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAODACOffsetExtSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetAODACOffsetExtSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACOffsetExtSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACOffsetExtSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAODACOffsetExtSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACOffsetExtSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACOffsetExtSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACOffsetExtSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACOffsetExtSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetAODACOffsetVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODACOffsetVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODACOffsetVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAODACOffsetVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODACOffsetVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODACOffsetVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODACOffsetVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAODACOffsetVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODACOffsetVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODACOffsetVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODACOffsetVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODACOffsetVal(taskHandle, channel)


@catchDAQError
def DAQmxGetAOReglitchEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOReglitchEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOReglitchEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOReglitchEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOReglitchEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOReglitchEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOReglitchEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAOReglitchEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOReglitchEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOReglitchEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOReglitchEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOReglitchEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFilterDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFilterDelay(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFilterDelay.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFilterDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFilterDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFilterDelay(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFilterDelay.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFilterDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFilterDelay(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFilterDelay(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFilterDelay.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFilterDelay(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFilterDelayUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFilterDelayUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFilterDelayUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAOFilterDelayUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFilterDelayUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFilterDelayUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFilterDelayUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAOFilterDelayUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFilterDelayUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFilterDelayUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFilterDelayUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFilterDelayUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetAOFilterDelayAdjustment(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOFilterDelayAdjustment(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOFilterDelayAdjustment(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOFilterDelayAdjustment(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOFilterDelayAdjustment(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOFilterDelayAdjustment(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOFilterDelayAdjustment(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOFilterDelayAdjustment(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOFilterDelayAdjustment.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOFilterDelayAdjustment(taskHandle, channel)


@catchDAQError
def DAQmxGetAOGain(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOGain(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOGain.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOGain(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOGain(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOGain(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOGain.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAOGain(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOGain(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOGain(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOGain.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOGain(taskHandle, channel)


@catchDAQError
def DAQmxGetAOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOUseOnlyOnBrdMem(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOUseOnlyOnBrdMem(taskHandle, channel)


@catchDAQError
def DAQmxGetAODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetAODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAODataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAODataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAODataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAODataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAODataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAODataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetAOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetAOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetAOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetAODevScalingCoeff(taskHandle, channel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAODevScalingCoeff
    (TaskHandle taskHandle, const char channel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAODevScalingCoeff.argtypes = (TaskHandle, c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAODevScalingCoeff(taskHandle, channel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetAOEnhancedImageRejectionEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetAOEnhancedImageRejectionEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOEnhancedImageRejectionEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOEnhancedImageRejectionEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetAOEnhancedImageRejectionEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetAOEnhancedImageRejectionEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOEnhancedImageRejectionEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAOEnhancedImageRejectionEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetAOEnhancedImageRejectionEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetAOEnhancedImageRejectionEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOEnhancedImageRejectionEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAOEnhancedImageRejectionEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetDIInvertLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIInvertLines(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIInvertLines.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDIInvertLines(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIInvertLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIInvertLines(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIInvertLines.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDIInvertLines(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIInvertLines(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIInvertLines(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIInvertLines.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIInvertLines(taskHandle, channel)


@catchDAQError
def DAQmxGetDINumLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDINumLines(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDINumLines.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDINumLines(taskHandle, channel, data)


@catchDAQError
def DAQmxGetDIDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDIDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDIDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDIDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetDIDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDigFltrEnableBusMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDigFltrEnableBusMode(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigFltrEnableBusMode.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDIDigFltrEnableBusMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDigFltrEnableBusMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigFltrEnableBusMode(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigFltrEnableBusMode.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDIDigFltrEnableBusMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigFltrEnableBusMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigFltrEnableBusMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigFltrEnableBusMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigFltrEnableBusMode(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetDIDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDIDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetDIDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetDIDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDIDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetDIDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDIDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDIDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetDITristate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDITristate(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDITristate.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDITristate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDITristate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDITristate(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDITristate.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDITristate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDITristate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDITristate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDITristate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDITristate(taskHandle, channel)


@catchDAQError
def DAQmxGetDILogicFamily(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDILogicFamily(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDILogicFamily.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDILogicFamily(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDILogicFamily(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDILogicFamily(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDILogicFamily.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDILogicFamily(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDILogicFamily(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDILogicFamily(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDILogicFamily.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDILogicFamily(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetDIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIDataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIDataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIDataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIDataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIDataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIDataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetDIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetDIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetDIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetDIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetDIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetDIAcquireOn(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDIAcquireOn(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDIAcquireOn.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDIAcquireOn(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDIAcquireOn(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDIAcquireOn(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDIAcquireOn.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDIAcquireOn(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDIAcquireOn(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDIAcquireOn(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDIAcquireOn.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDIAcquireOn(taskHandle, channel)


@catchDAQError
def DAQmxGetDOOutputDriveType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOOutputDriveType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOOutputDriveType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOOutputDriveType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOOutputDriveType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOOutputDriveType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOOutputDriveType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOOutputDriveType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOOutputDriveType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOOutputDriveType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOOutputDriveType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOOutputDriveType(taskHandle, channel)


@catchDAQError
def DAQmxGetDOInvertLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOInvertLines(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOInvertLines.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDOInvertLines(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOInvertLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOInvertLines(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOInvertLines.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDOInvertLines(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOInvertLines(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOInvertLines(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOInvertLines.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOInvertLines(taskHandle, channel)


@catchDAQError
def DAQmxGetDONumLines(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDONumLines(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDONumLines.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDONumLines(taskHandle, channel, data)


@catchDAQError
def DAQmxGetDOTristate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOTristate(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOTristate.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDOTristate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOTristate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOTristate(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOTristate.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDOTristate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOTristate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOTristate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOTristate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOTristate(taskHandle, channel)


@catchDAQError
def DAQmxGetDOLineStatesStartState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOLineStatesStartState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOLineStatesStartState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOLineStatesStartState(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOLineStatesStartState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOLineStatesStartState(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOLineStatesStartState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOLineStatesStartState(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOLineStatesStartState(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOLineStatesStartState(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOLineStatesStartState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOLineStatesStartState(taskHandle, channel)


@catchDAQError
def DAQmxGetDOLineStatesPausedState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOLineStatesPausedState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOLineStatesPausedState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOLineStatesPausedState(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOLineStatesPausedState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOLineStatesPausedState(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOLineStatesPausedState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOLineStatesPausedState(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOLineStatesPausedState(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOLineStatesPausedState(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOLineStatesPausedState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOLineStatesPausedState(taskHandle, channel)


@catchDAQError
def DAQmxGetDOLineStatesDoneState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOLineStatesDoneState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOLineStatesDoneState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOLineStatesDoneState(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOLineStatesDoneState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOLineStatesDoneState(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOLineStatesDoneState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOLineStatesDoneState(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOLineStatesDoneState(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOLineStatesDoneState(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOLineStatesDoneState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOLineStatesDoneState(taskHandle, channel)


@catchDAQError
def DAQmxGetDOLogicFamily(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOLogicFamily(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOLogicFamily.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOLogicFamily(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOLogicFamily(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOLogicFamily(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOLogicFamily.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOLogicFamily(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOLogicFamily(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOLogicFamily(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOLogicFamily.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOLogicFamily(taskHandle, channel)


@catchDAQError
def DAQmxGetDOOvercurrentLimit(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOOvercurrentLimit(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOOvercurrentLimit.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDOOvercurrentLimit(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOOvercurrentLimit(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOOvercurrentLimit(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOOvercurrentLimit.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetDOOvercurrentLimit(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOOvercurrentLimit(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOOvercurrentLimit(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOOvercurrentLimit.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOOvercurrentLimit(taskHandle, channel)


@catchDAQError
def DAQmxGetDOOvercurrentAutoReenable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOOvercurrentAutoReenable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOOvercurrentAutoReenable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDOOvercurrentAutoReenable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOOvercurrentAutoReenable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOOvercurrentAutoReenable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOOvercurrentAutoReenable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDOOvercurrentAutoReenable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOOvercurrentAutoReenable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOOvercurrentAutoReenable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOOvercurrentAutoReenable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOOvercurrentAutoReenable(taskHandle, channel)


@catchDAQError
def DAQmxGetDOOvercurrentReenablePeriod(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOOvercurrentReenablePeriod
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOOvercurrentReenablePeriod.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDOOvercurrentReenablePeriod(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOOvercurrentReenablePeriod(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOOvercurrentReenablePeriod(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOOvercurrentReenablePeriod.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetDOOvercurrentReenablePeriod(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOOvercurrentReenablePeriod(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOOvercurrentReenablePeriod(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOOvercurrentReenablePeriod.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOOvercurrentReenablePeriod(taskHandle, channel)


@catchDAQError
def DAQmxGetDOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOUseOnlyOnBrdMem(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOUseOnlyOnBrdMem(taskHandle, channel)


@catchDAQError
def DAQmxGetDODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDODataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDODataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDODataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDODataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDODataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDODataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDODataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDODataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetDODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDODataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDODataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDODataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDODataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDODataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDODataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetDOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetDOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetDOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetDOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetDOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetDOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetDOGenerateOn(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetDOGenerateOn(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDOGenerateOn.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDOGenerateOn(taskHandle, channel, data)


@catchDAQError
def DAQmxSetDOGenerateOn(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetDOGenerateOn(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDOGenerateOn.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDOGenerateOn(taskHandle, channel, data)


@catchDAQError
def DAQmxResetDOGenerateOn(taskHandle, channel):
    """int32 __CFUNC DAQmxResetDOGenerateOn(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetDOGenerateOn.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDOGenerateOn(taskHandle, channel)


@catchDAQError
def DAQmxGetCIMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIMax(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIMax.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIMax(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIMax(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIMax(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIMax.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIMax(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIMax(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIMax(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIMax.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIMax(taskHandle, channel)


@catchDAQError
def DAQmxGetCIMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIMin(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIMin.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIMin(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIMin(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIMin(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIMin.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIMin(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIMin(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIMin(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIMin.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIMin(taskHandle, channel)


@catchDAQError
def DAQmxGetCICustomScaleName(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICustomScaleName
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICustomScaleName(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICustomScaleName(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICustomScaleName(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICustomScaleName.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICustomScaleName(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICustomScaleName(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICustomScaleName(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICustomScaleName.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICustomScaleName(taskHandle, channel)


@catchDAQError
def DAQmxGetCIMeasType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIMeasType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIMeasType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIMeasType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCIFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIFreqTerm(TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIFreqTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIFreqTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIFreqTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqStartingEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqStartingEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIFreqStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqStartingEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqStartingEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIFreqStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqStartingEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqStartingEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqStartingEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqStartingEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqMeasMeth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqMeasMeth(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqMeasMeth.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIFreqMeasMeth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqMeasMeth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqMeasMeth(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqMeasMeth.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIFreqMeasMeth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqMeasMeth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqMeasMeth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqMeasMeth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqMeasMeth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqEnableAveraging(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqEnableAveraging(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqEnableAveraging.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIFreqEnableAveraging(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqEnableAveraging(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqEnableAveraging(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqEnableAveraging.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIFreqEnableAveraging(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqEnableAveraging(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqEnableAveraging(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqEnableAveraging.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqEnableAveraging(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqMeasTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqMeasTime(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqMeasTime.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIFreqMeasTime(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqMeasTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqMeasTime(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqMeasTime.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIFreqMeasTime(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqMeasTime(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqMeasTime(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqMeasTime.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqMeasTime(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqDiv(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIFreqDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDiv(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIFreqDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIFreqDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIFreqDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIFreqDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIFreqDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIFreqDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIFreqDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIFreqDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIFreqDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIFreqDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIFreqDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIFreqDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIFreqDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIFreqDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIFreqDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIFreqDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIFreqDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIFreqDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIFreqDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIFreqDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPeriodUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPeriodUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPeriodTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPeriodTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPeriodTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPeriodTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodStartingEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodStartingEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPeriodStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodStartingEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodStartingEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPeriodStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodStartingEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodStartingEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodStartingEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodStartingEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodMeasMeth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodMeasMeth(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodMeasMeth.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPeriodMeasMeth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodMeasMeth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodMeasMeth(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodMeasMeth.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPeriodMeasMeth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodMeasMeth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodMeasMeth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodMeasMeth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodMeasMeth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodEnableAveraging(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodEnableAveraging(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodEnableAveraging.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPeriodEnableAveraging(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodEnableAveraging(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodEnableAveraging(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodEnableAveraging.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPeriodEnableAveraging(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodEnableAveraging(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodEnableAveraging(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodEnableAveraging.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodEnableAveraging(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodMeasTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodMeasTime(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodMeasTime.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPeriodMeasTime(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodMeasTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodMeasTime(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodMeasTime.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPeriodMeasTime(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodMeasTime(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodMeasTime(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodMeasTime.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodMeasTime(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodDiv(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIPeriodDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDiv(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIPeriodDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPeriodDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPeriodDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPeriodDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPeriodDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPeriodDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPeriodDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPeriodDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPeriodDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPeriodDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPeriodDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPeriodDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPeriodDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPeriodDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPeriodDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPeriodDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPeriodDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDir(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesDir(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDir.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCICountEdgesDir(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesDir(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDir(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDir.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCICountEdgesDir(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDir(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDir(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDir.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDir(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDirTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesDirTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDirTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesDirTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesDirTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDirTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDirTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesDirTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDirTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDirTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDirTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDirTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountDirDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountDirDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountDirDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesCountDirDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountDirDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountDirDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountDirDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesCountDirDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountDirDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountDirDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountDirDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountDirDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountDirDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountDirDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountDirDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesCountDirDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountDirDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountDirDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountDirDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesCountDirDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountDirDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountDirDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountDirDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountDirDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesInitialCnt(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesInitialCnt(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesInitialCnt.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCICountEdgesInitialCnt(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesInitialCnt(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesInitialCnt(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesInitialCnt.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCICountEdgesInitialCnt(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesInitialCnt(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesInitialCnt(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesInitialCnt.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesInitialCnt(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesActiveEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesActiveEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCICountEdgesActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesActiveEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesActiveEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCICountEdgesActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesActiveEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesActiveEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesActiveEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesActiveEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesCountResetEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesCountResetEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetResetCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetResetCount
    (TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetResetCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCICountEdgesCountResetResetCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetResetCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetResetCount
    (TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetResetCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCICountEdgesCountResetResetCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetResetCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetResetCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetResetCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetResetCount(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesCountResetTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesCountResetTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetTerm
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesCountResetTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetActiveEdge
    (TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetActiveEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCICountEdgesCountResetActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetActiveEdge
    (TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetActiveEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCICountEdgesCountResetActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetActiveEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetActiveEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetActiveEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetActiveEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesCountResetDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesCountResetDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesCountResetDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesCountResetDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesCountResetDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesCountResetDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesCountResetDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesCountResetDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesCountResetDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesCountResetDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesCountResetDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesCountResetDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesCountResetDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesCountResetDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICountEdgesDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICountEdgesDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICountEdgesDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCICountEdgesDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICountEdgesDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICountEdgesDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICountEdgesDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICountEdgesDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICountEdgesDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICountEdgesDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICountEdgesDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICountEdgesDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICountEdgesDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICountEdgesDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICountEdgesDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIAngEncoderUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIAngEncoderUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIAngEncoderUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIAngEncoderUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIAngEncoderUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIAngEncoderUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIAngEncoderUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIAngEncoderUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIAngEncoderUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIAngEncoderUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIAngEncoderUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIAngEncoderUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIAngEncoderPulsesPerRev(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIAngEncoderPulsesPerRev(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIAngEncoderPulsesPerRev.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIAngEncoderPulsesPerRev(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIAngEncoderPulsesPerRev(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIAngEncoderPulsesPerRev(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIAngEncoderPulsesPerRev.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIAngEncoderPulsesPerRev(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIAngEncoderPulsesPerRev(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIAngEncoderPulsesPerRev(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIAngEncoderPulsesPerRev.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIAngEncoderPulsesPerRev(taskHandle, channel)


@catchDAQError
def DAQmxGetCIAngEncoderInitialAngle(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIAngEncoderInitialAngle(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIAngEncoderInitialAngle.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIAngEncoderInitialAngle(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIAngEncoderInitialAngle(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIAngEncoderInitialAngle(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIAngEncoderInitialAngle.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIAngEncoderInitialAngle(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIAngEncoderInitialAngle(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIAngEncoderInitialAngle(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIAngEncoderInitialAngle.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIAngEncoderInitialAngle(taskHandle, channel)


@catchDAQError
def DAQmxGetCILinEncoderUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCILinEncoderUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCILinEncoderUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCILinEncoderUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCILinEncoderUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCILinEncoderUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCILinEncoderUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCILinEncoderUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCILinEncoderUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCILinEncoderUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCILinEncoderUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCILinEncoderUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCILinEncoderDistPerPulse(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCILinEncoderDistPerPulse(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCILinEncoderDistPerPulse.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCILinEncoderDistPerPulse(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCILinEncoderDistPerPulse(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCILinEncoderDistPerPulse(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCILinEncoderDistPerPulse.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCILinEncoderDistPerPulse(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCILinEncoderDistPerPulse(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCILinEncoderDistPerPulse(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCILinEncoderDistPerPulse.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCILinEncoderDistPerPulse(taskHandle, channel)


@catchDAQError
def DAQmxGetCILinEncoderInitialPos(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCILinEncoderInitialPos(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCILinEncoderInitialPos.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCILinEncoderInitialPos(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCILinEncoderInitialPos(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCILinEncoderInitialPos(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCILinEncoderInitialPos.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCILinEncoderInitialPos(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCILinEncoderInitialPos(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCILinEncoderInitialPos(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCILinEncoderInitialPos.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCILinEncoderInitialPos(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderDecodingType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderDecodingType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderDecodingType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIEncoderDecodingType(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderDecodingType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderDecodingType(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderDecodingType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIEncoderDecodingType(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderDecodingType(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderDecodingType(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderDecodingType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderDecodingType(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderAInputTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderAInputTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderAInputTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderAInputTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderAInputDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderAInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderAInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderAInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderAInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderAInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderAInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderAInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderAInputDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderAInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderAInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderAInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderAInputDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderAInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderAInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderAInputDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderAInputDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderAInputDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderAInputDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderBInputTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderBInputTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderBInputTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderBInputTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderBInputDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderBInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderBInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderBInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderBInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderBInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderBInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderBInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderBInputDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderBInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderBInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderBInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderBInputDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderBInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderBInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderBInputDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderBInputDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderBInputDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderBInputDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderZInputTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderZInputTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderZInputTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderZInputTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZInputDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderZInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZInputDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderZInputDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIEncoderZInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZInputDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderZInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZInputDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZInputDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZInputDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderZInputDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZInputDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZInputDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZInputDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZInputDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZIndexEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZIndexEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZIndexEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIEncoderZIndexEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZIndexEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZIndexEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZIndexEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIEncoderZIndexEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZIndexEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZIndexEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZIndexEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZIndexEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZIndexVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZIndexVal(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZIndexVal.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIEncoderZIndexVal(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZIndexVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZIndexVal(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZIndexVal.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIEncoderZIndexVal(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZIndexVal(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZIndexVal(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZIndexVal.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZIndexVal(taskHandle, channel)


@catchDAQError
def DAQmxGetCIEncoderZIndexPhase(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIEncoderZIndexPhase(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIEncoderZIndexPhase.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIEncoderZIndexPhase(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIEncoderZIndexPhase(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIEncoderZIndexPhase(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIEncoderZIndexPhase.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIEncoderZIndexPhase(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIEncoderZIndexPhase(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIEncoderZIndexPhase(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIEncoderZIndexPhase.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIEncoderZIndexPhase(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseWidthUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseWidthUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseWidthTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseWidthTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseWidthTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseWidthTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthStartingEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthStartingEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseWidthStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthStartingEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthStartingEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseWidthStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthStartingEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthStartingEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthStartingEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthStartingEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseWidthDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseWidthDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseWidthDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseWidthDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseWidthDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseWidthDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseWidthDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseWidthDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseWidthDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseWidthDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseWidthDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseWidthDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseWidthDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseWidthDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseWidthDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCITimestampUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITimestampUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITimestampUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCITimestampUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITimestampUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITimestampUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITimestampUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCITimestampUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITimestampUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITimestampUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITimestampUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITimestampUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCITwoEdgeSepUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCITwoEdgeSepUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepFirstDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepFirstDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepFirstDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepFirstDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepFirstDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepFirstDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepFirstDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondDigFltrEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITwoEdgeSepSecondDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITwoEdgeSepSecondDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITwoEdgeSepSecondDigSyncEnable
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITwoEdgeSepSecondDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITwoEdgeSepSecondDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITwoEdgeSepSecondDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITwoEdgeSepSecondDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCISemiPeriodUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCISemiPeriodUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCISemiPeriodTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCISemiPeriodTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCISemiPeriodTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCISemiPeriodTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodStartingEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodStartingEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCISemiPeriodStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodStartingEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodStartingEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodStartingEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCISemiPeriodStartingEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodStartingEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodStartingEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodStartingEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodStartingEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCISemiPeriodDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCISemiPeriodDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCISemiPeriodDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCISemiPeriodDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCISemiPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCISemiPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCISemiPeriodDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCISemiPeriodDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCISemiPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCISemiPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCISemiPeriodDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCISemiPeriodDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCISemiPeriodDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCISemiPeriodDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCISemiPeriodDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseFreqTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseFreqTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseFreqTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseFreqTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqStartEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqStartEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseFreqStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqStartEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqStartEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseFreqStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqStartEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqStartEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqStartEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqStartEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseFreqDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseFreqDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseFreqDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseFreqDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseFreqDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseFreqDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseFreqDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseFreqDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseFreqDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseFreqDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseFreqDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseFreqDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseFreqDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseTimeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseTimeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseTimeTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseTimeTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseTimeTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseTimeTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeStartEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeStartEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseTimeStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeStartEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeStartEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseTimeStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeStartEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeStartEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeStartEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeStartEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseTimeDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseTimeDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseTimeDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTimeDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTimeDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTimeDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseTimeDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTimeDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTimeDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTimeDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseTimeDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTimeDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTimeDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTimeDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTimeDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseTicksTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseTicksTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseTicksTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseTicksTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTicksStartEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksStartEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIPulseTicksStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTicksStartEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksStartEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksStartEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIPulseTicksStartEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksStartEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksStartEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksStartEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksStartEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTicksDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseTicksDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTicksDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseTicksDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTicksDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIPulseTicksDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTicksDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPulseTicksDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPulseTicksDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPulseTicksDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIPulseTicksDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPulseTicksDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPulseTicksDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPulseTicksDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIPulseTicksDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPulseTicksDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPulseTicksDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPulseTicksDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPulseTicksDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCITimestampInitialSeconds(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITimestampInitialSeconds(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITimestampInitialSeconds.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCITimestampInitialSeconds(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCITimestampInitialSeconds(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCITimestampInitialSeconds(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCITimestampInitialSeconds.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCITimestampInitialSeconds(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCITimestampInitialSeconds(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCITimestampInitialSeconds(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCITimestampInitialSeconds.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCITimestampInitialSeconds(taskHandle, channel)


@catchDAQError
def DAQmxGetCIGPSSyncMethod(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIGPSSyncMethod(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIGPSSyncMethod.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIGPSSyncMethod(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIGPSSyncMethod(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIGPSSyncMethod(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIGPSSyncMethod.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIGPSSyncMethod(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIGPSSyncMethod(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIGPSSyncMethod(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIGPSSyncMethod.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIGPSSyncMethod(taskHandle, channel)


@catchDAQError
def DAQmxGetCIGPSSyncSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCIGPSSyncSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCIGPSSyncSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCIGPSSyncSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCIGPSSyncSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIGPSSyncSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCIGPSSyncSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCIGPSSyncSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIGPSSyncSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIGPSSyncSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIGPSSyncSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIGPSSyncSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICtrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICtrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICtrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICtrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICtrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICtrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCICtrTimebaseActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCICtrTimebaseActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseActiveEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseActiveEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICtrTimebaseDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICtrTimebaseDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCICtrTimebaseDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCICtrTimebaseDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCICtrTimebaseDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCICtrTimebaseDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCICount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCICount(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCIOutputState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIOutputState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIOutputState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIOutputState(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCITCReached(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCITCReached(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCITCReached.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCITCReached(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCICtrTimebaseMasterTimebaseDiv
    (TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCICtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCICtrTimebaseMasterTimebaseDiv
    (TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCICtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCICtrTimebaseMasterTimebaseDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCICtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCICtrTimebaseMasterTimebaseDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetCIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIDataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIDataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIDataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIDataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIDataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIDataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIDataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIDataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIDataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIDataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetCIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIDataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIDataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIDataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIDataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCIDataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIDataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIDataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIDataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIDataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetCIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetCIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetCIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCINumPossiblyInvalidSamps(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCINumPossiblyInvalidSamps(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCINumPossiblyInvalidSamps.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCINumPossiblyInvalidSamps(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCIDupCountPrevent(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIDupCountPrevent(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIDupCountPrevent.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCIDupCountPrevent(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIDupCountPrevent(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIDupCountPrevent(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIDupCountPrevent.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCIDupCountPrevent(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIDupCountPrevent(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIDupCountPrevent(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIDupCountPrevent.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIDupCountPrevent(taskHandle, channel)


@catchDAQError
def DAQmxGetCIPrescaler(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCIPrescaler(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCIPrescaler.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCIPrescaler(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCIPrescaler(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCIPrescaler(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCIPrescaler.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCIPrescaler(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCIPrescaler(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCIPrescaler(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCIPrescaler.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCIPrescaler(taskHandle, channel)


@catchDAQError
def DAQmxGetCOOutputType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOOutputType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOOutputType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOOutputType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCOPulseIdleState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseIdleState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseIdleState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOPulseIdleState(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseIdleState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseIdleState(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseIdleState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCOPulseIdleState(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseIdleState(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseIdleState(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseIdleState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseIdleState(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseTerm(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCOPulseTerm
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseTerm.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCOPulseTerm(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCOPulseTerm(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseTerm(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseTerm.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCOPulseTerm(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseTerm(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseTerm(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseTerm(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseTimeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseTimeUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseTimeUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOPulseTimeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseTimeUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseTimeUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseTimeUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCOPulseTimeUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseTimeUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseTimeUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseTimeUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseTimeUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseHighTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseHighTime(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseHighTime.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseHighTime(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseHighTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseHighTime(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseHighTime.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseHighTime(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseHighTime(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseHighTime(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseHighTime.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseHighTime(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseLowTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseLowTime(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseLowTime.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseLowTime(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseLowTime(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseLowTime(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseLowTime.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseLowTime(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseLowTime(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseLowTime(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseLowTime.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseLowTime(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseTimeInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseTimeInitialDelay(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseTimeInitialDelay.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseTimeInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseTimeInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseTimeInitialDelay(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseTimeInitialDelay.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseTimeInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseTimeInitialDelay(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseTimeInitialDelay(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseTimeInitialDelay.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseTimeInitialDelay(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseDutyCyc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseDutyCyc(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseDutyCyc.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseDutyCyc(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseDutyCyc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseDutyCyc(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseDutyCyc.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseDutyCyc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseDutyCyc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseDutyCyc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseDutyCyc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseDutyCyc(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseFreqUnits(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseFreqUnits.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOPulseFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseFreqUnits(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseFreqUnits(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseFreqUnits.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCOPulseFreqUnits(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseFreqUnits(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseFreqUnits(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseFreqUnits.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseFreqUnits(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseFreq(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseFreq.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseFreq(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseFreq(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseFreq.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseFreq(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseFreq(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseFreq(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseFreq.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseFreq(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseFreqInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseFreqInitialDelay(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseFreqInitialDelay.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOPulseFreqInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseFreqInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseFreqInitialDelay(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseFreqInitialDelay.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOPulseFreqInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseFreqInitialDelay(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseFreqInitialDelay(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseFreqInitialDelay.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseFreqInitialDelay(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseHighTicks(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseHighTicks(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseHighTicks.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOPulseHighTicks(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseHighTicks(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseHighTicks(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseHighTicks.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOPulseHighTicks(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseHighTicks(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseHighTicks(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseHighTicks.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseHighTicks(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseLowTicks(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseLowTicks(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseLowTicks.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOPulseLowTicks(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseLowTicks(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseLowTicks(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseLowTicks.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOPulseLowTicks(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseLowTicks(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseLowTicks(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseLowTicks.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseLowTicks(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseTicksInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseTicksInitialDelay(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseTicksInitialDelay.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOPulseTicksInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPulseTicksInitialDelay(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPulseTicksInitialDelay(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPulseTicksInitialDelay.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOPulseTicksInitialDelay(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPulseTicksInitialDelay(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPulseTicksInitialDelay(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPulseTicksInitialDelay.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPulseTicksInitialDelay(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCOCtrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCOCtrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseSrc(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCOCtrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOCtrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseRate(TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOCtrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOCtrTimebaseActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseActiveEdge(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCOCtrTimebaseActiveEdge(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseActiveEdge(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseActiveEdge(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseActiveEdge.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseActiveEdge(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOCtrTimebaseDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseDigFltrEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCOCtrTimebaseDigFltrEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseDigFltrEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseDigFltrEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseDigFltrEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseDigFltrEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc
    (TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseDigFltrTimebaseRate
    (TaskHandle taskHandle, const char channel[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseDigFltrTimebaseRate(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseDigFltrTimebaseRate.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseDigFltrTimebaseRate(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOCtrTimebaseDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseDigSyncEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCOCtrTimebaseDigSyncEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseDigSyncEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseDigSyncEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseDigSyncEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseDigSyncEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOCount(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCOOutputState(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOOutputState(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOOutputState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOOutputState(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCOAutoIncrCnt(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOAutoIncrCnt(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOAutoIncrCnt.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOAutoIncrCnt(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOAutoIncrCnt(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOAutoIncrCnt(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOAutoIncrCnt.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOAutoIncrCnt(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOAutoIncrCnt(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOAutoIncrCnt(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOAutoIncrCnt.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOAutoIncrCnt(taskHandle, channel)


@catchDAQError
def DAQmxGetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOCtrTimebaseMasterTimebaseDiv
    (TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOCtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOCtrTimebaseMasterTimebaseDiv
    (TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOCtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOCtrTimebaseMasterTimebaseDiv(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOCtrTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOCtrTimebaseMasterTimebaseDiv(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPulseDone(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPulseDone(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPulseDone.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOPulseDone(taskHandle, channel, data)


@catchDAQError
def DAQmxGetCOEnableInitialDelayOnRetrigger(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOEnableInitialDelayOnRetrigger
    (TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOEnableInitialDelayOnRetrigger.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOEnableInitialDelayOnRetrigger(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOEnableInitialDelayOnRetrigger(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOEnableInitialDelayOnRetrigger
    (TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOEnableInitialDelayOnRetrigger.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCOEnableInitialDelayOnRetrigger(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOEnableInitialDelayOnRetrigger(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOEnableInitialDelayOnRetrigger(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOEnableInitialDelayOnRetrigger.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOEnableInitialDelayOnRetrigger(taskHandle, channel)


@catchDAQError
def DAQmxGetCOConstrainedGenMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOConstrainedGenMode(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOConstrainedGenMode.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCOConstrainedGenMode(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOConstrainedGenMode(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOConstrainedGenMode(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOConstrainedGenMode.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCOConstrainedGenMode(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOConstrainedGenMode(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOConstrainedGenMode(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOConstrainedGenMode.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOConstrainedGenMode(taskHandle, channel)


@catchDAQError
def DAQmxGetCOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOUseOnlyOnBrdMem(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCOUseOnlyOnBrdMem(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOUseOnlyOnBrdMem(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOUseOnlyOnBrdMem(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOUseOnlyOnBrdMem.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOUseOnlyOnBrdMem(taskHandle, channel)


@catchDAQError
def DAQmxGetCODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCODataXferMech(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCODataXferMech.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCODataXferMech(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCODataXferMech(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCODataXferMech.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCODataXferMech(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCODataXferMech(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCODataXferMech(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCODataXferMech.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCODataXferMech(taskHandle, channel)


@catchDAQError
def DAQmxGetCODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCODataXferReqCond.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetCODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCODataXferReqCond(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCODataXferReqCond(TaskHandle taskHandle, const char channel[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCODataXferReqCond.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetCODataXferReqCond(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCODataXferReqCond(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCODataXferReqCond(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCODataXferReqCond.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCODataXferReqCond(taskHandle, channel)


@catchDAQError
def DAQmxGetCOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOUsbXferReqSize(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOUsbXferReqSize(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOUsbXferReqSize.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOUsbXferReqSize(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOUsbXferReqSize(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOUsbXferReqSize(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOUsbXferReqSize.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOUsbXferReqSize(taskHandle, channel)


@catchDAQError
def DAQmxGetCOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOUsbXferReqCount(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOUsbXferReqCount(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOUsbXferReqCount.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOUsbXferReqCount(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOUsbXferReqCount(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOUsbXferReqCount(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOUsbXferReqCount.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOUsbXferReqCount(taskHandle, channel)


@catchDAQError
def DAQmxGetCOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOMemMapEnable.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOMemMapEnable(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOMemMapEnable(TaskHandle taskHandle, const char channel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOMemMapEnable.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetCOMemMapEnable(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOMemMapEnable(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOMemMapEnable(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOMemMapEnable.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOMemMapEnable(taskHandle, channel)


@catchDAQError
def DAQmxGetCOPrescaler(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCOPrescaler(TaskHandle taskHandle, const char channel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCOPrescaler.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCOPrescaler(taskHandle, channel, data)


@catchDAQError
def DAQmxSetCOPrescaler(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetCOPrescaler(TaskHandle taskHandle, const char channel[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetCOPrescaler.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetCOPrescaler(taskHandle, channel, data)


@catchDAQError
def DAQmxResetCOPrescaler(taskHandle, channel):
    """int32 __CFUNC DAQmxResetCOPrescaler(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetCOPrescaler.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetCOPrescaler(taskHandle, channel)


@catchDAQError
def DAQmxGetCORdyForNewVal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetCORdyForNewVal(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCORdyForNewVal.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetCORdyForNewVal(taskHandle, channel, data)


@catchDAQError
def DAQmxGetChanType(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetChanType(TaskHandle taskHandle, const char channel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetChanType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetChanType(taskHandle, channel, data)


@catchDAQError
def DAQmxGetPhysicalChanName(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetPhysicalChanName
    (TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanName.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPhysicalChanName(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetPhysicalChanName(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetPhysicalChanName(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetPhysicalChanName.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetPhysicalChanName(taskHandle, channel, data)


@catchDAQError
def DAQmxGetChanDescr(taskHandle, channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetChanDescr(TaskHandle taskHandle, const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetChanDescr.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetChanDescr(taskHandle, channel, data, bufferSize)


@catchDAQError
def DAQmxSetChanDescr(taskHandle, channel, data):
    """int32 __CFUNC DAQmxSetChanDescr(TaskHandle taskHandle, const char channel[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetChanDescr.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetChanDescr(taskHandle, channel, data)


@catchDAQError
def DAQmxResetChanDescr(taskHandle, channel):
    """int32 __CFUNC DAQmxResetChanDescr(TaskHandle taskHandle, const char channel[]);"""
    global DAQlib
    DAQlib.DAQmxResetChanDescr.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetChanDescr(taskHandle, channel)


@catchDAQError
def DAQmxGetChanIsGlobal(taskHandle, channel, data):
    """int32 __CFUNC DAQmxGetChanIsGlobal(TaskHandle taskHandle, const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetChanIsGlobal.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetChanIsGlobal(taskHandle, channel, data)


@catchDAQError
def DAQmxGetExportedAIConvClkOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedAIConvClkOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAIConvClkOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedAIConvClkOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedAIConvClkOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAIConvClkOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAIConvClkOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedAIConvClkOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAIConvClkOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAIConvClkOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAIConvClkOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAIConvClkOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedAIConvClkPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAIConvClkPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAIConvClkPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedAIConvClkPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxGetExported10MHzRefClkOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExported10MHzRefClkOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExported10MHzRefClkOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExported10MHzRefClkOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExported10MHzRefClkOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExported10MHzRefClkOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExported10MHzRefClkOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExported10MHzRefClkOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExported10MHzRefClkOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExported10MHzRefClkOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExported10MHzRefClkOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExported10MHzRefClkOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExported20MHzTimebaseOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExported20MHzTimebaseOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExported20MHzTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExported20MHzTimebaseOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExported20MHzTimebaseOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExported20MHzTimebaseOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExported20MHzTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExported20MHzTimebaseOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExported20MHzTimebaseOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExported20MHzTimebaseOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExported20MHzTimebaseOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExported20MHzTimebaseOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedSampClkOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedSampClkOutputBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSampClkOutputBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedSampClkOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetExportedSampClkOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSampClkOutputBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSampClkOutputBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedSampClkOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSampClkOutputBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSampClkOutputBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSampClkOutputBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSampClkOutputBehavior(taskHandle)


@catchDAQError
def DAQmxGetExportedSampClkOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedSampClkOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSampClkOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedSampClkOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedSampClkOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSampClkOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSampClkOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedSampClkOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSampClkOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSampClkOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSampClkOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSampClkOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedSampClkDelayOffset(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedSampClkDelayOffset(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSampClkDelayOffset.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedSampClkDelayOffset(taskHandle, data)


@catchDAQError
def DAQmxSetExportedSampClkDelayOffset(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSampClkDelayOffset(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSampClkDelayOffset.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedSampClkDelayOffset(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSampClkDelayOffset(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSampClkDelayOffset(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSampClkDelayOffset.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSampClkDelayOffset(taskHandle)


@catchDAQError
def DAQmxGetExportedSampClkPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedSampClkPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSampClkPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedSampClkPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedSampClkPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSampClkPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSampClkPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedSampClkPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSampClkPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSampClkPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSampClkPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSampClkPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedSampClkTimebaseOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedSampClkTimebaseOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSampClkTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedSampClkTimebaseOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedSampClkTimebaseOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSampClkTimebaseOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSampClkTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedSampClkTimebaseOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSampClkTimebaseOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSampClkTimebaseOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSampClkTimebaseOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSampClkTimebaseOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedDividedSampClkTimebaseOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedDividedSampClkTimebaseOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedDividedSampClkTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedDividedSampClkTimebaseOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedDividedSampClkTimebaseOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedDividedSampClkTimebaseOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedDividedSampClkTimebaseOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedDividedSampClkTimebaseOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedDividedSampClkTimebaseOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedDividedSampClkTimebaseOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedDividedSampClkTimebaseOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedDividedSampClkTimebaseOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvTrigOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedAdvTrigOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvTrigOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedAdvTrigOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedAdvTrigOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvTrigOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvTrigOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedAdvTrigOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvTrigOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvTrigOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvTrigOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvTrigOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvTrigPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvTrigPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvTrigPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedAdvTrigPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxGetExportedAdvTrigPulseWidthUnits(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvTrigPulseWidthUnits(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvTrigPulseWidthUnits.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedAdvTrigPulseWidthUnits(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAdvTrigPulseWidthUnits(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvTrigPulseWidthUnits(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvTrigPulseWidthUnits.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedAdvTrigPulseWidthUnits(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvTrigPulseWidthUnits(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvTrigPulseWidthUnits(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvTrigPulseWidthUnits.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvTrigPulseWidthUnits(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvTrigPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvTrigPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvTrigPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedAdvTrigPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAdvTrigPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvTrigPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvTrigPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedAdvTrigPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvTrigPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvTrigPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvTrigPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvTrigPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetExportedPauseTrigOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedPauseTrigOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedPauseTrigOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedPauseTrigOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedPauseTrigOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedPauseTrigOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedPauseTrigOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedPauseTrigOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedPauseTrigOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedPauseTrigOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedPauseTrigOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedPauseTrigOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedPauseTrigLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedPauseTrigLvlActiveLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedPauseTrigLvlActiveLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedPauseTrigLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxSetExportedPauseTrigLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedPauseTrigLvlActiveLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedPauseTrigLvlActiveLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedPauseTrigLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxResetExportedPauseTrigLvlActiveLvl(taskHandle):
    """int32 __CFUNC DAQmxResetExportedPauseTrigLvlActiveLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedPauseTrigLvlActiveLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedPauseTrigLvlActiveLvl(taskHandle)


@catchDAQError
def DAQmxGetExportedRefTrigOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedRefTrigOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRefTrigOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedRefTrigOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedRefTrigOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRefTrigOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRefTrigOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedRefTrigOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRefTrigOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRefTrigOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRefTrigOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRefTrigOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedRefTrigPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedRefTrigPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRefTrigPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedRefTrigPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedRefTrigPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRefTrigPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRefTrigPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedRefTrigPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRefTrigPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRefTrigPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRefTrigPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRefTrigPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedStartTrigOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedStartTrigOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedStartTrigOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedStartTrigOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedStartTrigOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedStartTrigOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedStartTrigOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedStartTrigOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedStartTrigOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedStartTrigOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedStartTrigOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedStartTrigOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedStartTrigPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedStartTrigPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedStartTrigPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedStartTrigPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedStartTrigPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedStartTrigPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedStartTrigPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedStartTrigPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedStartTrigPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedStartTrigPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedStartTrigPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedStartTrigPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvCmpltEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedAdvCmpltEventOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvCmpltEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedAdvCmpltEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedAdvCmpltEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvCmpltEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvCmpltEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedAdvCmpltEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvCmpltEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvCmpltEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvCmpltEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvCmpltEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvCmpltEventDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvCmpltEventDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvCmpltEventDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedAdvCmpltEventDelay(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAdvCmpltEventDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvCmpltEventDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvCmpltEventDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedAdvCmpltEventDelay(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvCmpltEventDelay(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvCmpltEventDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvCmpltEventDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvCmpltEventDelay(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvCmpltEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvCmpltEventPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvCmpltEventPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedAdvCmpltEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAdvCmpltEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvCmpltEventPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvCmpltEventPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedAdvCmpltEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvCmpltEventPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvCmpltEventPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvCmpltEventPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvCmpltEventPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedAdvCmpltEventPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAdvCmpltEventPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAdvCmpltEventPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedAdvCmpltEventPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAdvCmpltEventPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAdvCmpltEventPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAdvCmpltEventPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedAdvCmpltEventPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAdvCmpltEventPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAdvCmpltEventPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAdvCmpltEventPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAdvCmpltEventPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetExportedAIHoldCmpltEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedAIHoldCmpltEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAIHoldCmpltEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedAIHoldCmpltEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedAIHoldCmpltEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAIHoldCmpltEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAIHoldCmpltEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedAIHoldCmpltEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAIHoldCmpltEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAIHoldCmpltEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAIHoldCmpltEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAIHoldCmpltEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedAIHoldCmpltEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedAIHoldCmpltEventPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedAIHoldCmpltEventPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedAIHoldCmpltEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedAIHoldCmpltEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedAIHoldCmpltEventPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedAIHoldCmpltEventPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedAIHoldCmpltEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedAIHoldCmpltEventPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedAIHoldCmpltEventPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedAIHoldCmpltEventPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedAIHoldCmpltEventPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedChangeDetectEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedChangeDetectEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedChangeDetectEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedChangeDetectEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedChangeDetectEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedChangeDetectEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedChangeDetectEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedChangeDetectEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedChangeDetectEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedChangeDetectEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedChangeDetectEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedChangeDetectEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedChangeDetectEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedChangeDetectEventPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedChangeDetectEventPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedChangeDetectEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedChangeDetectEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedChangeDetectEventPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedChangeDetectEventPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedChangeDetectEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedChangeDetectEventPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedChangeDetectEventPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedChangeDetectEventPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedChangeDetectEventPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedCtrOutEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedCtrOutEventOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedCtrOutEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedCtrOutEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedCtrOutEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedCtrOutEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedCtrOutEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedCtrOutEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedCtrOutEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedCtrOutEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedCtrOutEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedCtrOutEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedCtrOutEventOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedCtrOutEventOutputBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedCtrOutEventOutputBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedCtrOutEventOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetExportedCtrOutEventOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedCtrOutEventOutputBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedCtrOutEventOutputBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedCtrOutEventOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetExportedCtrOutEventOutputBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetExportedCtrOutEventOutputBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedCtrOutEventOutputBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedCtrOutEventOutputBehavior(taskHandle)


@catchDAQError
def DAQmxGetExportedCtrOutEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedCtrOutEventPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedCtrOutEventPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedCtrOutEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedCtrOutEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedCtrOutEventPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedCtrOutEventPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedCtrOutEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedCtrOutEventPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedCtrOutEventPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedCtrOutEventPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedCtrOutEventPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedCtrOutEventToggleIdleState(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedCtrOutEventToggleIdleState(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedCtrOutEventToggleIdleState.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedCtrOutEventToggleIdleState(taskHandle, data)


@catchDAQError
def DAQmxSetExportedCtrOutEventToggleIdleState(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedCtrOutEventToggleIdleState(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedCtrOutEventToggleIdleState.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedCtrOutEventToggleIdleState(taskHandle, data)


@catchDAQError
def DAQmxResetExportedCtrOutEventToggleIdleState(taskHandle):
    """int32 __CFUNC DAQmxResetExportedCtrOutEventToggleIdleState(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedCtrOutEventToggleIdleState.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedCtrOutEventToggleIdleState(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedHshkEventOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedHshkEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedHshkEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedHshkEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventOutputBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventOutputBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedHshkEventOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventOutputBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventOutputBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventOutputBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedHshkEventOutputBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventOutputBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventOutputBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventOutputBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventOutputBehavior(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedHshkEventDelay(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedHshkEventDelay(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventDelay(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventDelay(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventInterlockedAssertedLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventInterlockedAssertedLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventInterlockedAssertedLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedHshkEventInterlockedAssertedLvl(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventInterlockedAssertedLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventInterlockedAssertedLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventInterlockedAssertedLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedHshkEventInterlockedAssertedLvl(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventInterlockedAssertedLvl(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventInterlockedAssertedLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventInterlockedAssertedLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventInterlockedAssertedLvl(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventInterlockedAssertOnStart(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventInterlockedAssertOnStart(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventInterlockedAssertOnStart.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetExportedHshkEventInterlockedAssertOnStart(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventInterlockedAssertOnStart(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventInterlockedAssertOnStart(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventInterlockedAssertOnStart.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetExportedHshkEventInterlockedAssertOnStart(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventInterlockedAssertOnStart(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventInterlockedAssertOnStart(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventInterlockedAssertOnStart.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventInterlockedAssertOnStart(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventInterlockedDeassertDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventInterlockedDeassertDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventInterlockedDeassertDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedHshkEventInterlockedDeassertDelay(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventInterlockedDeassertDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventInterlockedDeassertDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventInterlockedDeassertDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedHshkEventInterlockedDeassertDelay(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventInterlockedDeassertDelay(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventInterlockedDeassertDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventInterlockedDeassertDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventInterlockedDeassertDelay(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventPulsePolarity(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventPulsePolarity.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedHshkEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventPulsePolarity(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventPulsePolarity(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventPulsePolarity.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedHshkEventPulsePolarity(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventPulsePolarity(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventPulsePolarity(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventPulsePolarity.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventPulsePolarity(taskHandle)


@catchDAQError
def DAQmxGetExportedHshkEventPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedHshkEventPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedHshkEventPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetExportedHshkEventPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetExportedHshkEventPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedHshkEventPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedHshkEventPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetExportedHshkEventPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetExportedHshkEventPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetExportedHshkEventPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedHshkEventPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedHshkEventPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForXferEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedRdyForXferEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForXferEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedRdyForXferEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedRdyForXferEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForXferEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForXferEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedRdyForXferEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForXferEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForXferEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForXferEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForXferEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForXferEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedRdyForXferEventLvlActiveLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForXferEventLvlActiveLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedRdyForXferEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxSetExportedRdyForXferEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForXferEventLvlActiveLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForXferEventLvlActiveLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedRdyForXferEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForXferEventLvlActiveLvl(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForXferEventLvlActiveLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForXferEventLvlActiveLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForXferEventLvlActiveLvl(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForXferEventDeassertCond(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedRdyForXferEventDeassertCond(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForXferEventDeassertCond.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedRdyForXferEventDeassertCond(taskHandle, data)


@catchDAQError
def DAQmxSetExportedRdyForXferEventDeassertCond(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForXferEventDeassertCond(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForXferEventDeassertCond.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedRdyForXferEventDeassertCond(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForXferEventDeassertCond(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForXferEventDeassertCond(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForXferEventDeassertCond.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForXferEventDeassertCond(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold
    (TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle, data)


@catchDAQError
def DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold(taskHandle)


@catchDAQError
def DAQmxGetExportedDataActiveEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedDataActiveEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedDataActiveEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedDataActiveEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedDataActiveEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedDataActiveEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedDataActiveEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedDataActiveEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedDataActiveEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedDataActiveEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedDataActiveEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedDataActiveEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedDataActiveEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedDataActiveEventLvlActiveLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedDataActiveEventLvlActiveLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedDataActiveEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxSetExportedDataActiveEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedDataActiveEventLvlActiveLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedDataActiveEventLvlActiveLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedDataActiveEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxResetExportedDataActiveEventLvlActiveLvl(taskHandle):
    """int32 __CFUNC DAQmxResetExportedDataActiveEventLvlActiveLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedDataActiveEventLvlActiveLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedDataActiveEventLvlActiveLvl(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForStartEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedRdyForStartEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForStartEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedRdyForStartEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedRdyForStartEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForStartEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForStartEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedRdyForStartEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForStartEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForStartEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForStartEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForStartEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedRdyForStartEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetExportedRdyForStartEventLvlActiveLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetExportedRdyForStartEventLvlActiveLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetExportedRdyForStartEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxSetExportedRdyForStartEventLvlActiveLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedRdyForStartEventLvlActiveLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedRdyForStartEventLvlActiveLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetExportedRdyForStartEventLvlActiveLvl(taskHandle, data)


@catchDAQError
def DAQmxResetExportedRdyForStartEventLvlActiveLvl(taskHandle):
    """int32 __CFUNC DAQmxResetExportedRdyForStartEventLvlActiveLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedRdyForStartEventLvlActiveLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedRdyForStartEventLvlActiveLvl(taskHandle)


@catchDAQError
def DAQmxGetExportedSyncPulseEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedSyncPulseEventOutputTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedSyncPulseEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedSyncPulseEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedSyncPulseEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedSyncPulseEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedSyncPulseEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedSyncPulseEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedSyncPulseEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedSyncPulseEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedSyncPulseEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedSyncPulseEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetExportedWatchdogExpiredEventOutputTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetExportedWatchdogExpiredEventOutputTerm
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetExportedWatchdogExpiredEventOutputTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetExportedWatchdogExpiredEventOutputTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetExportedWatchdogExpiredEventOutputTerm(taskHandle, data):
    """int32 __CFUNC DAQmxSetExportedWatchdogExpiredEventOutputTerm(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetExportedWatchdogExpiredEventOutputTerm.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetExportedWatchdogExpiredEventOutputTerm(taskHandle, data)


@catchDAQError
def DAQmxResetExportedWatchdogExpiredEventOutputTerm(taskHandle):
    """int32 __CFUNC DAQmxResetExportedWatchdogExpiredEventOutputTerm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetExportedWatchdogExpiredEventOutputTerm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetExportedWatchdogExpiredEventOutputTerm(taskHandle)


@catchDAQError
def DAQmxGetDevIsSimulated(device, data):
    """int32 __CFUNC DAQmxGetDevIsSimulated(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevIsSimulated.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevIsSimulated(device, data)


@catchDAQError
def DAQmxGetDevProductCategory(device, data):
    """int32 __CFUNC DAQmxGetDevProductCategory(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevProductCategory.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevProductCategory(device, data)


@catchDAQError
def DAQmxGetDevProductType(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevProductType(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevProductType.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevProductType(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevProductNum(device, data):
    """int32 __CFUNC DAQmxGetDevProductNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevProductNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevProductNum(device, data)


@catchDAQError
def DAQmxGetDevSerialNum(device, data):
    """int32 __CFUNC DAQmxGetDevSerialNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevSerialNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevSerialNum(device, data)


@catchDAQError
def DAQmxGetDevAccessoryProductTypes(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevAccessoryProductTypes(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevAccessoryProductTypes.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevAccessoryProductTypes(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevAccessoryProductNums(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAccessoryProductNums(const char device[], uInt32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAccessoryProductNums.argtypes = (c_char_p, POINTER(uInt32), uInt32,)
    return DAQlib.DAQmxGetDevAccessoryProductNums(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAccessorySerialNums(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAccessorySerialNums(const char device[], uInt32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAccessorySerialNums.argtypes = (c_char_p, POINTER(uInt32), uInt32,)
    return DAQlib.DAQmxGetDevAccessorySerialNums(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetCarrierSerialNum(device, data):
    """int32 __CFUNC DAQmxGetCarrierSerialNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetCarrierSerialNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetCarrierSerialNum(device, data)


@catchDAQError
def DAQmxGetDevChassisModuleDevNames(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevChassisModuleDevNames(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevChassisModuleDevNames.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevChassisModuleDevNames(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevAnlgTrigSupported(device, data):
    """int32 __CFUNC DAQmxGetDevAnlgTrigSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAnlgTrigSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevAnlgTrigSupported(device, data)


@catchDAQError
def DAQmxGetDevDigTrigSupported(device, data):
    """int32 __CFUNC DAQmxGetDevDigTrigSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevDigTrigSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevDigTrigSupported(device, data)


@catchDAQError
def DAQmxGetDevAIPhysicalChans(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevAIPhysicalChans(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIPhysicalChans.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevAIPhysicalChans(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevAISupportedMeasTypes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAISupportedMeasTypes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAISupportedMeasTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevAISupportedMeasTypes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIMaxSingleChanRate(device, data):
    """int32 __CFUNC DAQmxGetDevAIMaxSingleChanRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIMaxSingleChanRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevAIMaxSingleChanRate(device, data)


@catchDAQError
def DAQmxGetDevAIMaxMultiChanRate(device, data):
    """int32 __CFUNC DAQmxGetDevAIMaxMultiChanRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIMaxMultiChanRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevAIMaxMultiChanRate(device, data)


@catchDAQError
def DAQmxGetDevAIMinRate(device, data):
    """int32 __CFUNC DAQmxGetDevAIMinRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIMinRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevAIMinRate(device, data)


@catchDAQError
def DAQmxGetDevAISimultaneousSamplingSupported(device, data):
    """int32 __CFUNC DAQmxGetDevAISimultaneousSamplingSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAISimultaneousSamplingSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevAISimultaneousSamplingSupported(device, data)


@catchDAQError
def DAQmxGetDevAISampModes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAISampModes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAISampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevAISampModes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAITrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevAITrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAITrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevAITrigUsage(device, data)


@catchDAQError
def DAQmxGetDevAIVoltageRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIVoltageRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIVoltageRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIVoltageRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIVoltageIntExcitDiscreteVals(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIVoltageIntExcitDiscreteVals
    (const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIVoltageIntExcitDiscreteVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIVoltageIntExcitDiscreteVals(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIVoltageIntExcitRangeVals(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIVoltageIntExcitRangeVals
    (const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIVoltageIntExcitRangeVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIVoltageIntExcitRangeVals(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAICurrentRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAICurrentRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAICurrentRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAICurrentRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAICurrentIntExcitDiscreteVals(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAICurrentIntExcitDiscreteVals
    (const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAICurrentIntExcitDiscreteVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAICurrentIntExcitDiscreteVals(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIBridgeRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIBridgeRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIBridgeRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIBridgeRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIResistanceRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIResistanceRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIResistanceRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIResistanceRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIFreqRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIFreqRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIFreqRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIFreqRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAIGains(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAIGains(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAIGains.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAIGains(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAICouplings(device, data):
    """int32 __CFUNC DAQmxGetDevAICouplings(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAICouplings.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevAICouplings(device, data)


@catchDAQError
def DAQmxGetDevAILowpassCutoffFreqDiscreteVals(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAILowpassCutoffFreqDiscreteVals
    (const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAILowpassCutoffFreqDiscreteVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAILowpassCutoffFreqDiscreteVals(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAILowpassCutoffFreqRangeVals(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAILowpassCutoffFreqRangeVals
    (const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAILowpassCutoffFreqRangeVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAILowpassCutoffFreqRangeVals(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAOPhysicalChans(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevAOPhysicalChans(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOPhysicalChans.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevAOPhysicalChans(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevAOSupportedOutputTypes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAOSupportedOutputTypes
    (const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOSupportedOutputTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevAOSupportedOutputTypes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAOSampClkSupported(device, data):
    """int32 __CFUNC DAQmxGetDevAOSampClkSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOSampClkSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevAOSampClkSupported(device, data)


@catchDAQError
def DAQmxGetDevAOSampModes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAOSampModes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOSampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevAOSampModes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAOMaxRate(device, data):
    """int32 __CFUNC DAQmxGetDevAOMaxRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOMaxRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevAOMaxRate(device, data)


@catchDAQError
def DAQmxGetDevAOMinRate(device, data):
    """int32 __CFUNC DAQmxGetDevAOMinRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOMinRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevAOMinRate(device, data)


@catchDAQError
def DAQmxGetDevAOTrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevAOTrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOTrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevAOTrigUsage(device, data)


@catchDAQError
def DAQmxGetDevAOVoltageRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAOVoltageRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOVoltageRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAOVoltageRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAOCurrentRngs(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAOCurrentRngs(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOCurrentRngs.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAOCurrentRngs(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevAOGains(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevAOGains(const char device[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevAOGains.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetDevAOGains(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevDILines(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevDILines(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevDILines.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevDILines(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevDIPorts(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevDIPorts(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevDIPorts.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevDIPorts(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevDIMaxRate(device, data):
    """int32 __CFUNC DAQmxGetDevDIMaxRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevDIMaxRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevDIMaxRate(device, data)


@catchDAQError
def DAQmxGetDevDITrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevDITrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevDITrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevDITrigUsage(device, data)


@catchDAQError
def DAQmxGetDevDOLines(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevDOLines(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevDOLines.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevDOLines(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevDOPorts(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevDOPorts(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevDOPorts.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevDOPorts(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevDOMaxRate(device, data):
    """int32 __CFUNC DAQmxGetDevDOMaxRate(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevDOMaxRate.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevDOMaxRate(device, data)


@catchDAQError
def DAQmxGetDevDOTrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevDOTrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevDOTrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevDOTrigUsage(device, data)


@catchDAQError
def DAQmxGetDevCIPhysicalChans(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevCIPhysicalChans(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevCIPhysicalChans.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevCIPhysicalChans(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevCISupportedMeasTypes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevCISupportedMeasTypes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevCISupportedMeasTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevCISupportedMeasTypes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevCITrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevCITrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCITrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevCITrigUsage(device, data)


@catchDAQError
def DAQmxGetDevCISampClkSupported(device, data):
    """int32 __CFUNC DAQmxGetDevCISampClkSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCISampClkSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevCISampClkSupported(device, data)


@catchDAQError
def DAQmxGetDevCISampModes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevCISampModes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevCISampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevCISampModes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevCIMaxSize(device, data):
    """int32 __CFUNC DAQmxGetDevCIMaxSize(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCIMaxSize.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevCIMaxSize(device, data)


@catchDAQError
def DAQmxGetDevCIMaxTimebase(device, data):
    """int32 __CFUNC DAQmxGetDevCIMaxTimebase(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCIMaxTimebase.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevCIMaxTimebase(device, data)


@catchDAQError
def DAQmxGetDevCOPhysicalChans(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevCOPhysicalChans(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOPhysicalChans.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevCOPhysicalChans(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevCOSupportedOutputTypes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevCOSupportedOutputTypes
    (const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOSupportedOutputTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevCOSupportedOutputTypes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevCOSampClkSupported(device, data):
    """int32 __CFUNC DAQmxGetDevCOSampClkSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOSampClkSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevCOSampClkSupported(device, data)


@catchDAQError
def DAQmxGetDevCOSampModes(device, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetDevCOSampModes(const char device[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOSampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetDevCOSampModes(device, data, arraySizeInElements)


@catchDAQError
def DAQmxGetDevCOTrigUsage(device, data):
    """int32 __CFUNC DAQmxGetDevCOTrigUsage(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOTrigUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevCOTrigUsage(device, data)


@catchDAQError
def DAQmxGetDevCOMaxSize(device, data):
    """int32 __CFUNC DAQmxGetDevCOMaxSize(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOMaxSize.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevCOMaxSize(device, data)


@catchDAQError
def DAQmxGetDevCOMaxTimebase(device, data):
    """int32 __CFUNC DAQmxGetDevCOMaxTimebase(const char device[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCOMaxTimebase.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDevCOMaxTimebase(device, data)


@catchDAQError
def DAQmxGetDevTEDSHWTEDSSupported(device, data):
    """int32 __CFUNC DAQmxGetDevTEDSHWTEDSSupported(const char device[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevTEDSHWTEDSSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetDevTEDSHWTEDSSupported(device, data)


@catchDAQError
def DAQmxGetDevNumDMAChans(device, data):
    """int32 __CFUNC DAQmxGetDevNumDMAChans(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevNumDMAChans.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevNumDMAChans(device, data)


@catchDAQError
def DAQmxGetDevBusType(device, data):
    """int32 __CFUNC DAQmxGetDevBusType(const char device[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevBusType.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDevBusType(device, data)


@catchDAQError
def DAQmxGetDevPCIBusNum(device, data):
    """int32 __CFUNC DAQmxGetDevPCIBusNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevPCIBusNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevPCIBusNum(device, data)


@catchDAQError
def DAQmxGetDevPCIDevNum(device, data):
    """int32 __CFUNC DAQmxGetDevPCIDevNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevPCIDevNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevPCIDevNum(device, data)


@catchDAQError
def DAQmxGetDevPXIChassisNum(device, data):
    """int32 __CFUNC DAQmxGetDevPXIChassisNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevPXIChassisNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevPXIChassisNum(device, data)


@catchDAQError
def DAQmxGetDevPXISlotNum(device, data):
    """int32 __CFUNC DAQmxGetDevPXISlotNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevPXISlotNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevPXISlotNum(device, data)


@catchDAQError
def DAQmxGetDevCompactDAQChassisDevName(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevCompactDAQChassisDevName(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevCompactDAQChassisDevName.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevCompactDAQChassisDevName(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevCompactDAQSlotNum(device, data):
    """int32 __CFUNC DAQmxGetDevCompactDAQSlotNum(const char device[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDevCompactDAQSlotNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetDevCompactDAQSlotNum(device, data)


@catchDAQError
def DAQmxGetDevTCPIPHostname(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevTCPIPHostname(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevTCPIPHostname.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevTCPIPHostname(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevTCPIPEthernetIP(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevTCPIPEthernetIP(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevTCPIPEthernetIP.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevTCPIPEthernetIP(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevTCPIPWirelessIP(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevTCPIPWirelessIP(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevTCPIPWirelessIP.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevTCPIPWirelessIP(device, data, bufferSize)


@catchDAQError
def DAQmxGetDevTerminals(device, data, bufferSize):
    """int32 __CFUNC DAQmxGetDevTerminals(const char device[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDevTerminals.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDevTerminals(device, data, bufferSize)


@catchDAQError
def DAQmxGetReadRelativeTo(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadRelativeTo(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadRelativeTo.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetReadRelativeTo(taskHandle, data)


@catchDAQError
def DAQmxSetReadRelativeTo(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadRelativeTo(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadRelativeTo.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetReadRelativeTo(taskHandle, data)


@catchDAQError
def DAQmxResetReadRelativeTo(taskHandle):
    """int32 __CFUNC DAQmxResetReadRelativeTo(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadRelativeTo.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadRelativeTo(taskHandle)


@catchDAQError
def DAQmxGetReadOffset(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOffset(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOffset.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetReadOffset(taskHandle, data)


@catchDAQError
def DAQmxSetReadOffset(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadOffset(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadOffset.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetReadOffset(taskHandle, data)


@catchDAQError
def DAQmxResetReadOffset(taskHandle):
    """int32 __CFUNC DAQmxResetReadOffset(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadOffset.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadOffset(taskHandle)


@catchDAQError
def DAQmxGetReadChannelsToRead(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadChannelsToRead(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadChannelsToRead.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadChannelsToRead(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetReadChannelsToRead(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadChannelsToRead(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetReadChannelsToRead.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetReadChannelsToRead(taskHandle, data)


@catchDAQError
def DAQmxResetReadChannelsToRead(taskHandle):
    """int32 __CFUNC DAQmxResetReadChannelsToRead(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadChannelsToRead.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadChannelsToRead(taskHandle)


@catchDAQError
def DAQmxGetReadReadAllAvailSamp(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadReadAllAvailSamp(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadReadAllAvailSamp.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadReadAllAvailSamp(taskHandle, data)


@catchDAQError
def DAQmxSetReadReadAllAvailSamp(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadReadAllAvailSamp(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadReadAllAvailSamp.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetReadReadAllAvailSamp(taskHandle, data)


@catchDAQError
def DAQmxResetReadReadAllAvailSamp(taskHandle):
    """int32 __CFUNC DAQmxResetReadReadAllAvailSamp(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadReadAllAvailSamp.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadReadAllAvailSamp(taskHandle)


@catchDAQError
def DAQmxGetReadAutoStart(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadAutoStart(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadAutoStart.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadAutoStart(taskHandle, data)


@catchDAQError
def DAQmxSetReadAutoStart(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadAutoStart(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadAutoStart.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetReadAutoStart(taskHandle, data)


@catchDAQError
def DAQmxResetReadAutoStart(taskHandle):
    """int32 __CFUNC DAQmxResetReadAutoStart(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadAutoStart.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadAutoStart(taskHandle)


@catchDAQError
def DAQmxGetReadOverWrite(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOverWrite(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOverWrite.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetReadOverWrite(taskHandle, data)


@catchDAQError
def DAQmxSetReadOverWrite(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadOverWrite(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadOverWrite.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetReadOverWrite(taskHandle, data)


@catchDAQError
def DAQmxResetReadOverWrite(taskHandle):
    """int32 __CFUNC DAQmxResetReadOverWrite(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadOverWrite.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadOverWrite(taskHandle)


@catchDAQError
def DAQmxGetReadCurrReadPos(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadCurrReadPos(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadCurrReadPos.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetReadCurrReadPos(taskHandle, data)


@catchDAQError
def DAQmxGetReadAvailSampPerChan(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadAvailSampPerChan(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadAvailSampPerChan.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetReadAvailSampPerChan(taskHandle, data)


@catchDAQError
def DAQmxGetLoggingFilePath(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetLoggingFilePath(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingFilePath.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetLoggingFilePath(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetLoggingFilePath(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingFilePath(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingFilePath.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetLoggingFilePath(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingFilePath(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingFilePath(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingFilePath.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingFilePath(taskHandle)


@catchDAQError
def DAQmxGetLoggingMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetLoggingMode(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetLoggingMode(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingMode(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingMode(taskHandle)


@catchDAQError
def DAQmxGetLoggingTDMSGroupName(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetLoggingTDMSGroupName(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingTDMSGroupName.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetLoggingTDMSGroupName(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetLoggingTDMSGroupName(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingTDMSGroupName(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingTDMSGroupName.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetLoggingTDMSGroupName(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingTDMSGroupName(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingTDMSGroupName(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingTDMSGroupName.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingTDMSGroupName(taskHandle)


@catchDAQError
def DAQmxGetLoggingTDMSOperation(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingTDMSOperation(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingTDMSOperation.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetLoggingTDMSOperation(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingTDMSOperation(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingTDMSOperation(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingTDMSOperation.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetLoggingTDMSOperation(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingTDMSOperation(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingTDMSOperation(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingTDMSOperation.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingTDMSOperation(taskHandle)


@catchDAQError
def DAQmxGetLoggingPause(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingPause(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingPause.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetLoggingPause(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingPause(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingPause(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingPause.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetLoggingPause(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingPause(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingPause(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingPause.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingPause(taskHandle)


@catchDAQError
def DAQmxGetLoggingSampsPerFile(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingSampsPerFile(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingSampsPerFile.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetLoggingSampsPerFile(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingSampsPerFile(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingSampsPerFile(TaskHandle taskHandle, uInt64 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingSampsPerFile.argtypes = (TaskHandle, uInt64,)
    return DAQlib.DAQmxSetLoggingSampsPerFile(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingSampsPerFile(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingSampsPerFile(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingSampsPerFile.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingSampsPerFile(taskHandle)


@catchDAQError
def DAQmxGetLoggingFileWriteSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingFileWriteSize(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingFileWriteSize.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetLoggingFileWriteSize(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingFileWriteSize(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingFileWriteSize(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingFileWriteSize.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetLoggingFileWriteSize(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingFileWriteSize(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingFileWriteSize(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingFileWriteSize.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingFileWriteSize(taskHandle)


@catchDAQError
def DAQmxGetLoggingFilePreallocationSize(taskHandle, data):
    """int32 __CFUNC DAQmxGetLoggingFilePreallocationSize(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetLoggingFilePreallocationSize.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetLoggingFilePreallocationSize(taskHandle, data)


@catchDAQError
def DAQmxSetLoggingFilePreallocationSize(taskHandle, data):
    """int32 __CFUNC DAQmxSetLoggingFilePreallocationSize(TaskHandle taskHandle, uInt64 data);"""
    global DAQlib
    DAQlib.DAQmxSetLoggingFilePreallocationSize.argtypes = (TaskHandle, uInt64,)
    return DAQlib.DAQmxSetLoggingFilePreallocationSize(taskHandle, data)


@catchDAQError
def DAQmxResetLoggingFilePreallocationSize(taskHandle):
    """int32 __CFUNC DAQmxResetLoggingFilePreallocationSize(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetLoggingFilePreallocationSize.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetLoggingFilePreallocationSize(taskHandle)


@catchDAQError
def DAQmxGetReadTotalSampPerChanAcquired(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadTotalSampPerChanAcquired(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadTotalSampPerChanAcquired.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetReadTotalSampPerChanAcquired(taskHandle, data)


@catchDAQError
def DAQmxGetReadCommonModeRangeErrorChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadCommonModeRangeErrorChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadCommonModeRangeErrorChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadCommonModeRangeErrorChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadCommonModeRangeErrorChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadCommonModeRangeErrorChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadCommonModeRangeErrorChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadCommonModeRangeErrorChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadOvercurrentChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOvercurrentChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOvercurrentChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadOvercurrentChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadOvercurrentChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadOvercurrentChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadOvercurrentChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadOvercurrentChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadOvertemperatureChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOvertemperatureChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOvertemperatureChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadOvertemperatureChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadOvertemperatureChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadOvertemperatureChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadOvertemperatureChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadOvertemperatureChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadOpenCurrentLoopChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOpenCurrentLoopChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOpenCurrentLoopChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadOpenCurrentLoopChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadOpenCurrentLoopChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadOpenCurrentLoopChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadOpenCurrentLoopChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadOpenCurrentLoopChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadOpenThrmcplChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOpenThrmcplChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOpenThrmcplChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadOpenThrmcplChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadOpenThrmcplChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadOpenThrmcplChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadOpenThrmcplChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadOpenThrmcplChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadOverloadedChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadOverloadedChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadOverloadedChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadOverloadedChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetReadOverloadedChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadOverloadedChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadOverloadedChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadOverloadedChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadAccessoryInsertionOrRemovalDetected(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadAccessoryInsertionOrRemovalDetected(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadAccessoryInsertionOrRemovalDetected.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadAccessoryInsertionOrRemovalDetected(taskHandle, data)


@catchDAQError
def DAQmxGetReadDevsWithInsertedOrRemovedAccessories(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetReadDevsWithInsertedOrRemovedAccessories
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetReadDevsWithInsertedOrRemovedAccessories.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetReadDevsWithInsertedOrRemovedAccessories(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetReadChangeDetectHasOverflowed(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadChangeDetectHasOverflowed(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadChangeDetectHasOverflowed.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetReadChangeDetectHasOverflowed(taskHandle, data)


@catchDAQError
def DAQmxGetReadRawDataWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadRawDataWidth(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadRawDataWidth.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetReadRawDataWidth(taskHandle, data)


@catchDAQError
def DAQmxGetReadNumChans(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadNumChans(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadNumChans.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetReadNumChans(taskHandle, data)


@catchDAQError
def DAQmxGetReadDigitalLinesBytesPerChan(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadDigitalLinesBytesPerChan(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadDigitalLinesBytesPerChan.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetReadDigitalLinesBytesPerChan(taskHandle, data)


@catchDAQError
def DAQmxGetReadWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadWaitMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadWaitMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetReadWaitMode(taskHandle, data)


@catchDAQError
def DAQmxSetReadWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadWaitMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadWaitMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetReadWaitMode(taskHandle, data)


@catchDAQError
def DAQmxResetReadWaitMode(taskHandle):
    """int32 __CFUNC DAQmxResetReadWaitMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadWaitMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadWaitMode(taskHandle)


@catchDAQError
def DAQmxGetReadSleepTime(taskHandle, data):
    """int32 __CFUNC DAQmxGetReadSleepTime(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetReadSleepTime.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetReadSleepTime(taskHandle, data)


@catchDAQError
def DAQmxSetReadSleepTime(taskHandle, data):
    """int32 __CFUNC DAQmxSetReadSleepTime(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetReadSleepTime.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetReadSleepTime(taskHandle, data)


@catchDAQError
def DAQmxResetReadSleepTime(taskHandle):
    """int32 __CFUNC DAQmxResetReadSleepTime(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetReadSleepTime.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetReadSleepTime(taskHandle)


@catchDAQError
def DAQmxGetRealTimeConvLateErrorsToWarnings(taskHandle, data):
    """int32 __CFUNC DAQmxGetRealTimeConvLateErrorsToWarnings(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeConvLateErrorsToWarnings.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetRealTimeConvLateErrorsToWarnings(taskHandle, data)


@catchDAQError
def DAQmxSetRealTimeConvLateErrorsToWarnings(taskHandle, data):
    """int32 __CFUNC DAQmxSetRealTimeConvLateErrorsToWarnings(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeConvLateErrorsToWarnings.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetRealTimeConvLateErrorsToWarnings(taskHandle, data)


@catchDAQError
def DAQmxResetRealTimeConvLateErrorsToWarnings(taskHandle):
    """int32 __CFUNC DAQmxResetRealTimeConvLateErrorsToWarnings(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeConvLateErrorsToWarnings.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRealTimeConvLateErrorsToWarnings(taskHandle)


@catchDAQError
def DAQmxGetRealTimeNumOfWarmupIters(taskHandle, data):
    """int32 __CFUNC DAQmxGetRealTimeNumOfWarmupIters(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeNumOfWarmupIters.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetRealTimeNumOfWarmupIters(taskHandle, data)


@catchDAQError
def DAQmxSetRealTimeNumOfWarmupIters(taskHandle, data):
    """int32 __CFUNC DAQmxSetRealTimeNumOfWarmupIters(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeNumOfWarmupIters.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetRealTimeNumOfWarmupIters(taskHandle, data)


@catchDAQError
def DAQmxResetRealTimeNumOfWarmupIters(taskHandle):
    """int32 __CFUNC DAQmxResetRealTimeNumOfWarmupIters(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeNumOfWarmupIters.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRealTimeNumOfWarmupIters(taskHandle)


@catchDAQError
def DAQmxGetRealTimeWaitForNextSampClkWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetRealTimeWaitForNextSampClkWaitMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeWaitForNextSampClkWaitMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetRealTimeWaitForNextSampClkWaitMode(taskHandle, data)


@catchDAQError
def DAQmxSetRealTimeWaitForNextSampClkWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetRealTimeWaitForNextSampClkWaitMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeWaitForNextSampClkWaitMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetRealTimeWaitForNextSampClkWaitMode(taskHandle, data)


@catchDAQError
def DAQmxResetRealTimeWaitForNextSampClkWaitMode(taskHandle):
    """int32 __CFUNC DAQmxResetRealTimeWaitForNextSampClkWaitMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeWaitForNextSampClkWaitMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRealTimeWaitForNextSampClkWaitMode(taskHandle)


@catchDAQError
def DAQmxGetRealTimeReportMissedSamp(taskHandle, data):
    """int32 __CFUNC DAQmxGetRealTimeReportMissedSamp(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeReportMissedSamp.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetRealTimeReportMissedSamp(taskHandle, data)


@catchDAQError
def DAQmxSetRealTimeReportMissedSamp(taskHandle, data):
    """int32 __CFUNC DAQmxSetRealTimeReportMissedSamp(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeReportMissedSamp.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetRealTimeReportMissedSamp(taskHandle, data)


@catchDAQError
def DAQmxResetRealTimeReportMissedSamp(taskHandle):
    """int32 __CFUNC DAQmxResetRealTimeReportMissedSamp(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeReportMissedSamp.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRealTimeReportMissedSamp(taskHandle)


@catchDAQError
def DAQmxGetRealTimeWriteRecoveryMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetRealTimeWriteRecoveryMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRealTimeWriteRecoveryMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetRealTimeWriteRecoveryMode(taskHandle, data)


@catchDAQError
def DAQmxSetRealTimeWriteRecoveryMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetRealTimeWriteRecoveryMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRealTimeWriteRecoveryMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetRealTimeWriteRecoveryMode(taskHandle, data)


@catchDAQError
def DAQmxResetRealTimeWriteRecoveryMode(taskHandle):
    """int32 __CFUNC DAQmxResetRealTimeWriteRecoveryMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRealTimeWriteRecoveryMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRealTimeWriteRecoveryMode(taskHandle)


@catchDAQError
def DAQmxGetSwitchChanUsage(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanUsage(const char switchChannelName[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanUsage.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetSwitchChanUsage(switchChannelName, data)


@catchDAQError
def DAQmxSetSwitchChanUsage(switchChannelName, data):
    """int32 __CFUNC DAQmxSetSwitchChanUsage(const char switchChannelName[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchChanUsage.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetSwitchChanUsage(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanAnlgBusSharingEnable(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanAnlgBusSharingEnable(const char switchChannelName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanAnlgBusSharingEnable.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetSwitchChanAnlgBusSharingEnable(switchChannelName, data)


@catchDAQError
def DAQmxSetSwitchChanAnlgBusSharingEnable(switchChannelName, data):
    """int32 __CFUNC DAQmxSetSwitchChanAnlgBusSharingEnable(const char switchChannelName[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchChanAnlgBusSharingEnable.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSetSwitchChanAnlgBusSharingEnable(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxACCarryCurrent(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxACCarryCurrent(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxACCarryCurrent.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxACCarryCurrent(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxACSwitchCurrent(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxACSwitchCurrent(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxACSwitchCurrent.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxACSwitchCurrent(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxACCarryPwr(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxACCarryPwr(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxACCarryPwr.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxACCarryPwr(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxACSwitchPwr(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxACSwitchPwr(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxACSwitchPwr.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxACSwitchPwr(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxDCCarryCurrent(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxDCCarryCurrent(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxDCCarryCurrent.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxDCCarryCurrent(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxDCSwitchCurrent(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxDCSwitchCurrent(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxDCSwitchCurrent.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxDCSwitchCurrent(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxDCCarryPwr(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxDCCarryPwr(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxDCCarryPwr.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxDCCarryPwr(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxDCSwitchPwr(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxDCSwitchPwr(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxDCSwitchPwr.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxDCSwitchPwr(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxACVoltage(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxACVoltage(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxACVoltage.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxACVoltage(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanMaxDCVoltage(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanMaxDCVoltage(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanMaxDCVoltage.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanMaxDCVoltage(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanWireMode(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanWireMode(const char switchChannelName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanWireMode.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetSwitchChanWireMode(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanBandwidth(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanBandwidth(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanBandwidth.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanBandwidth(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchChanImpedance(switchChannelName, data):
    """int32 __CFUNC DAQmxGetSwitchChanImpedance(const char switchChannelName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchChanImpedance.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchChanImpedance(switchChannelName, data)


@catchDAQError
def DAQmxGetSwitchDevSettlingTime(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevSettlingTime(const char deviceName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevSettlingTime.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchDevSettlingTime(deviceName, data)


@catchDAQError
def DAQmxSetSwitchDevSettlingTime(deviceName, data):
    """int32 __CFUNC DAQmxSetSwitchDevSettlingTime(const char deviceName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchDevSettlingTime.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetSwitchDevSettlingTime(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevAutoConnAnlgBus(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevAutoConnAnlgBus(const char deviceName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevAutoConnAnlgBus.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetSwitchDevAutoConnAnlgBus(deviceName, data)


@catchDAQError
def DAQmxSetSwitchDevAutoConnAnlgBus(deviceName, data):
    """int32 __CFUNC DAQmxSetSwitchDevAutoConnAnlgBus(const char deviceName[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchDevAutoConnAnlgBus.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSetSwitchDevAutoConnAnlgBus(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevPwrDownLatchRelaysAfterSettling(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevPwrDownLatchRelaysAfterSettling(const char deviceName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevPwrDownLatchRelaysAfterSettling.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetSwitchDevPwrDownLatchRelaysAfterSettling(deviceName, data)


@catchDAQError
def DAQmxSetSwitchDevPwrDownLatchRelaysAfterSettling(deviceName, data):
    """int32 __CFUNC DAQmxSetSwitchDevPwrDownLatchRelaysAfterSettling(const char deviceName[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchDevPwrDownLatchRelaysAfterSettling.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSetSwitchDevPwrDownLatchRelaysAfterSettling(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevSettled(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevSettled(const char deviceName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevSettled.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetSwitchDevSettled(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevRelayList(deviceName, data, bufferSize):
    """int32 __CFUNC DAQmxGetSwitchDevRelayList(const char deviceName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevRelayList.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSwitchDevRelayList(deviceName, data, bufferSize)


@catchDAQError
def DAQmxGetSwitchDevNumRelays(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevNumRelays(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevNumRelays.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetSwitchDevNumRelays(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevSwitchChanList(deviceName, data, bufferSize):
    """int32 __CFUNC DAQmxGetSwitchDevSwitchChanList(const char deviceName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevSwitchChanList.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSwitchDevSwitchChanList(deviceName, data, bufferSize)


@catchDAQError
def DAQmxGetSwitchDevNumSwitchChans(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevNumSwitchChans(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevNumSwitchChans.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetSwitchDevNumSwitchChans(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevNumRows(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevNumRows(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevNumRows.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetSwitchDevNumRows(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevNumColumns(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevNumColumns(const char deviceName[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevNumColumns.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetSwitchDevNumColumns(deviceName, data)


@catchDAQError
def DAQmxGetSwitchDevTopology(deviceName, data, bufferSize):
    """int32 __CFUNC DAQmxGetSwitchDevTopology(const char deviceName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevTopology.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSwitchDevTopology(deviceName, data, bufferSize)


@catchDAQError
def DAQmxGetSwitchDevTemperature(deviceName, data):
    """int32 __CFUNC DAQmxGetSwitchDevTemperature(const char deviceName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchDevTemperature.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetSwitchDevTemperature(deviceName, data)


@catchDAQError
def DAQmxGetSwitchScanBreakMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetSwitchScanBreakMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchScanBreakMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSwitchScanBreakMode(taskHandle, data)


@catchDAQError
def DAQmxSetSwitchScanBreakMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetSwitchScanBreakMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchScanBreakMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSwitchScanBreakMode(taskHandle, data)


@catchDAQError
def DAQmxResetSwitchScanBreakMode(taskHandle):
    """int32 __CFUNC DAQmxResetSwitchScanBreakMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSwitchScanBreakMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSwitchScanBreakMode(taskHandle)


@catchDAQError
def DAQmxGetSwitchScanRepeatMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetSwitchScanRepeatMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchScanRepeatMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSwitchScanRepeatMode(taskHandle, data)


@catchDAQError
def DAQmxSetSwitchScanRepeatMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetSwitchScanRepeatMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSwitchScanRepeatMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSwitchScanRepeatMode(taskHandle, data)


@catchDAQError
def DAQmxResetSwitchScanRepeatMode(taskHandle):
    """int32 __CFUNC DAQmxResetSwitchScanRepeatMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSwitchScanRepeatMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSwitchScanRepeatMode(taskHandle)


@catchDAQError
def DAQmxGetSwitchScanWaitingForAdv(taskHandle, data):
    """int32 __CFUNC DAQmxGetSwitchScanWaitingForAdv(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSwitchScanWaitingForAdv.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetSwitchScanWaitingForAdv(taskHandle, data)


@catchDAQError
def DAQmxGetScaleDescr(scaleName, data, bufferSize):
    """int32 __CFUNC DAQmxGetScaleDescr(const char scaleName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetScaleDescr.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetScaleDescr(scaleName, data, bufferSize)


@catchDAQError
def DAQmxSetScaleDescr(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleDescr(const char scaleName[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleDescr.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxSetScaleDescr(scaleName, data)


@catchDAQError
def DAQmxGetScaleScaledUnits(scaleName, data, bufferSize):
    """int32 __CFUNC DAQmxGetScaleScaledUnits(const char scaleName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetScaleScaledUnits.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetScaleScaledUnits(scaleName, data, bufferSize)


@catchDAQError
def DAQmxSetScaleScaledUnits(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleScaledUnits(const char scaleName[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleScaledUnits.argtypes = (c_char_p, c_char_p,)
    return DAQlib.DAQmxSetScaleScaledUnits(scaleName, data)


@catchDAQError
def DAQmxGetScalePreScaledUnits(scaleName, data):
    """int32 __CFUNC DAQmxGetScalePreScaledUnits(const char scaleName[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScalePreScaledUnits.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetScalePreScaledUnits(scaleName, data)


@catchDAQError
def DAQmxSetScalePreScaledUnits(scaleName, data):
    """int32 __CFUNC DAQmxSetScalePreScaledUnits(const char scaleName[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetScalePreScaledUnits.argtypes = (c_char_p, int32,)
    return DAQlib.DAQmxSetScalePreScaledUnits(scaleName, data)


@catchDAQError
def DAQmxGetScaleType(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleType(const char scaleName[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleType.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetScaleType(scaleName, data)


@catchDAQError
def DAQmxGetScaleLinSlope(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleLinSlope(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleLinSlope.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleLinSlope(scaleName, data)


@catchDAQError
def DAQmxSetScaleLinSlope(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleLinSlope(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleLinSlope.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleLinSlope(scaleName, data)


@catchDAQError
def DAQmxGetScaleLinYIntercept(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleLinYIntercept(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleLinYIntercept.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleLinYIntercept(scaleName, data)


@catchDAQError
def DAQmxSetScaleLinYIntercept(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleLinYIntercept(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleLinYIntercept.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleLinYIntercept(scaleName, data)


@catchDAQError
def DAQmxGetScaleMapScaledMax(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleMapScaledMax(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleMapScaledMax.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleMapScaledMax(scaleName, data)


@catchDAQError
def DAQmxSetScaleMapScaledMax(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleMapScaledMax(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleMapScaledMax.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleMapScaledMax(scaleName, data)


@catchDAQError
def DAQmxGetScaleMapPreScaledMax(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleMapPreScaledMax(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleMapPreScaledMax.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleMapPreScaledMax(scaleName, data)


@catchDAQError
def DAQmxSetScaleMapPreScaledMax(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleMapPreScaledMax(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleMapPreScaledMax.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleMapPreScaledMax(scaleName, data)


@catchDAQError
def DAQmxGetScaleMapScaledMin(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleMapScaledMin(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleMapScaledMin.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleMapScaledMin(scaleName, data)


@catchDAQError
def DAQmxSetScaleMapScaledMin(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleMapScaledMin(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleMapScaledMin.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleMapScaledMin(scaleName, data)


@catchDAQError
def DAQmxGetScaleMapPreScaledMin(scaleName, data):
    """int32 __CFUNC DAQmxGetScaleMapPreScaledMin(const char scaleName[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetScaleMapPreScaledMin.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetScaleMapPreScaledMin(scaleName, data)


@catchDAQError
def DAQmxSetScaleMapPreScaledMin(scaleName, data):
    """int32 __CFUNC DAQmxSetScaleMapPreScaledMin(const char scaleName[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetScaleMapPreScaledMin.argtypes = (c_char_p, float64,)
    return DAQlib.DAQmxSetScaleMapPreScaledMin(scaleName, data)


@catchDAQError
def DAQmxGetScalePolyForwardCoeff(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetScalePolyForwardCoeff
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetScalePolyForwardCoeff.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetScalePolyForwardCoeff(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxSetScalePolyForwardCoeff(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetScalePolyForwardCoeff
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetScalePolyForwardCoeff.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetScalePolyForwardCoeff(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxGetScalePolyReverseCoeff(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetScalePolyReverseCoeff
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetScalePolyReverseCoeff.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetScalePolyReverseCoeff(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxSetScalePolyReverseCoeff(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetScalePolyReverseCoeff
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetScalePolyReverseCoeff.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetScalePolyReverseCoeff(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxGetScaleTableScaledVals(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetScaleTableScaledVals
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetScaleTableScaledVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetScaleTableScaledVals(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxSetScaleTableScaledVals(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetScaleTableScaledVals
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetScaleTableScaledVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetScaleTableScaledVals(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxGetScaleTablePreScaledVals(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetScaleTablePreScaledVals
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetScaleTablePreScaledVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetScaleTablePreScaledVals(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxSetScaleTablePreScaledVals(scaleName, data, arraySizeInElements):
    """int32 __CFUNC DAQmxSetScaleTablePreScaledVals
    (const char scaleName[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxSetScaleTablePreScaledVals.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxSetScaleTablePreScaledVals(scaleName, data, arraySizeInElements)


@catchDAQError
def DAQmxGetSysGlobalChans(data, bufferSize):
    """int32 __CFUNC DAQmxGetSysGlobalChans(char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSysGlobalChans.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetSysGlobalChans(data, bufferSize)


@catchDAQError
def DAQmxGetSysScales(data, bufferSize):
    """int32 __CFUNC DAQmxGetSysScales(char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSysScales.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetSysScales(data, bufferSize)


@catchDAQError
def DAQmxGetSysTasks(data, bufferSize):
    """int32 __CFUNC DAQmxGetSysTasks(char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSysTasks.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetSysTasks(data, bufferSize)


@catchDAQError
def DAQmxGetSysDevNames(data, bufferSize):
    """int32 __CFUNC DAQmxGetSysDevNames(char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSysDevNames.argtypes = (c_char_p, uInt32,)
    return DAQlib.DAQmxGetSysDevNames(data, bufferSize)


@catchDAQError
def DAQmxGetSysNIDAQMajorVersion(data):
    """int32 __CFUNC DAQmxGetSysNIDAQMajorVersion(uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSysNIDAQMajorVersion.argtypes = (POINTER(uInt32),)
    return DAQlib.DAQmxGetSysNIDAQMajorVersion(data)


@catchDAQError
def DAQmxGetSysNIDAQMinorVersion(data):
    """int32 __CFUNC DAQmxGetSysNIDAQMinorVersion(uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSysNIDAQMinorVersion.argtypes = (POINTER(uInt32),)
    return DAQlib.DAQmxGetSysNIDAQMinorVersion(data)


@catchDAQError
def DAQmxGetSysNIDAQUpdateVersion(data):
    """int32 __CFUNC DAQmxGetSysNIDAQUpdateVersion(uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSysNIDAQUpdateVersion.argtypes = (POINTER(uInt32),)
    return DAQlib.DAQmxGetSysNIDAQUpdateVersion(data)


@catchDAQError
def DAQmxGetTaskName(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetTaskName(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetTaskName.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetTaskName(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetTaskChannels(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetTaskChannels(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetTaskChannels.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetTaskChannels(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetTaskNumChans(taskHandle, data):
    """int32 __CFUNC DAQmxGetTaskNumChans(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetTaskNumChans.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetTaskNumChans(taskHandle, data)


@catchDAQError
def DAQmxGetTaskDevices(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetTaskDevices(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetTaskDevices.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetTaskDevices(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetTaskNumDevices(taskHandle, data):
    """int32 __CFUNC DAQmxGetTaskNumDevices(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetTaskNumDevices.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetTaskNumDevices(taskHandle, data)


@catchDAQError
def DAQmxGetTaskComplete(taskHandle, data):
    """int32 __CFUNC DAQmxGetTaskComplete(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetTaskComplete.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetTaskComplete(taskHandle, data)


@catchDAQError
def DAQmxGetSampQuantSampMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampQuantSampMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampQuantSampMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampQuantSampMode(taskHandle, data)


@catchDAQError
def DAQmxSetSampQuantSampMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampQuantSampMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampQuantSampMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampQuantSampMode(taskHandle, data)


@catchDAQError
def DAQmxResetSampQuantSampMode(taskHandle):
    """int32 __CFUNC DAQmxResetSampQuantSampMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampQuantSampMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampQuantSampMode(taskHandle)


@catchDAQError
def DAQmxGetSampQuantSampPerChan(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampQuantSampPerChan(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampQuantSampPerChan.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetSampQuantSampPerChan(taskHandle, data)


@catchDAQError
def DAQmxSetSampQuantSampPerChan(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampQuantSampPerChan(TaskHandle taskHandle, uInt64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampQuantSampPerChan.argtypes = (TaskHandle, uInt64,)
    return DAQlib.DAQmxSetSampQuantSampPerChan(taskHandle, data)


@catchDAQError
def DAQmxResetSampQuantSampPerChan(taskHandle):
    """int32 __CFUNC DAQmxResetSampQuantSampPerChan(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampQuantSampPerChan.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampQuantSampPerChan(taskHandle)


@catchDAQError
def DAQmxGetSampTimingType(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampTimingType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampTimingType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampTimingType(taskHandle, data)


@catchDAQError
def DAQmxSetSampTimingType(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampTimingType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampTimingType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampTimingType(taskHandle, data)


@catchDAQError
def DAQmxResetSampTimingType(taskHandle):
    """int32 __CFUNC DAQmxResetSampTimingType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampTimingType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampTimingType(taskHandle)


@catchDAQError
def DAQmxGetSampClkRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSampClkRate(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSampClkRate(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkRate(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkRate(taskHandle)


@catchDAQError
def DAQmxGetSampClkMaxRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkMaxRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkMaxRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSampClkMaxRate(taskHandle, data)


@catchDAQError
def DAQmxGetSampClkSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSampClkSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSampClkSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetSampClkSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetSampClkSrc(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkSrc(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkSrc(taskHandle)


@catchDAQError
def DAQmxGetSampClkActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkActiveEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkActiveEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampClkActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkActiveEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkActiveEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampClkActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkActiveEdge(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkActiveEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkActiveEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkActiveEdge(taskHandle)


@catchDAQError
def DAQmxGetSampClkOverrunBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkOverrunBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkOverrunBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampClkOverrunBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkOverrunBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkOverrunBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkOverrunBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampClkOverrunBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkOverrunBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkOverrunBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkOverrunBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkOverrunBehavior(taskHandle)


@catchDAQError
def DAQmxGetSampClkUnderflowBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkUnderflowBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkUnderflowBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampClkUnderflowBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkUnderflowBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkUnderflowBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkUnderflowBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampClkUnderflowBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkUnderflowBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkUnderflowBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkUnderflowBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkUnderflowBehavior(taskHandle)


@catchDAQError
def DAQmxGetSampClkTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkTimebaseDiv(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseDiv.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetSampClkTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimebaseDiv(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimebaseDiv.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetSampClkTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimebaseDiv(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimebaseDiv(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimebaseDiv.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimebaseDiv(taskHandle)


@catchDAQError
def DAQmxGetSampClkTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSampClkTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSampClkTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetSampClkTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSampClkTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSampClkTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetSampClkTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSampClkTimebaseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSampClkTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetSampClkTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetSampClkTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetSampClkTimebaseActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkTimebaseActiveEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseActiveEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampClkTimebaseActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkTimebaseActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimebaseActiveEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimebaseActiveEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampClkTimebaseActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimebaseActiveEdge(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimebaseActiveEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimebaseActiveEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimebaseActiveEdge(taskHandle)


@catchDAQError
def DAQmxGetSampClkTimebaseMasterTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkTimebaseMasterTimebaseDiv(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetSampClkTimebaseMasterTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkTimebaseMasterTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimebaseMasterTimebaseDiv(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimebaseMasterTimebaseDiv.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetSampClkTimebaseMasterTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimebaseMasterTimebaseDiv(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimebaseMasterTimebaseDiv(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimebaseMasterTimebaseDiv.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimebaseMasterTimebaseDiv(taskHandle)


@catchDAQError
def DAQmxGetSampClkTimebaseTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSampClkTimebaseTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimebaseTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSampClkTimebaseTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetSampClkDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetSampClkDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetSampClkDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetSampClkDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSampClkDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSampClkDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetSampClkDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSampClkDigFltrTimebaseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSampClkDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetSampClkDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetSampClkDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetSampClkDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSampClkDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSampClkDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetSampClkDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetSampClkDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetSampClkDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetHshkDelayAfterXfer(taskHandle, data):
    """int32 __CFUNC DAQmxGetHshkDelayAfterXfer(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetHshkDelayAfterXfer.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetHshkDelayAfterXfer(taskHandle, data)


@catchDAQError
def DAQmxSetHshkDelayAfterXfer(taskHandle, data):
    """int32 __CFUNC DAQmxSetHshkDelayAfterXfer(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetHshkDelayAfterXfer.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetHshkDelayAfterXfer(taskHandle, data)


@catchDAQError
def DAQmxResetHshkDelayAfterXfer(taskHandle):
    """int32 __CFUNC DAQmxResetHshkDelayAfterXfer(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetHshkDelayAfterXfer.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetHshkDelayAfterXfer(taskHandle)


@catchDAQError
def DAQmxGetHshkStartCond(taskHandle, data):
    """int32 __CFUNC DAQmxGetHshkStartCond(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetHshkStartCond.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetHshkStartCond(taskHandle, data)


@catchDAQError
def DAQmxSetHshkStartCond(taskHandle, data):
    """int32 __CFUNC DAQmxSetHshkStartCond(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetHshkStartCond.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetHshkStartCond(taskHandle, data)


@catchDAQError
def DAQmxResetHshkStartCond(taskHandle):
    """int32 __CFUNC DAQmxResetHshkStartCond(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetHshkStartCond.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetHshkStartCond(taskHandle)


@catchDAQError
def DAQmxGetHshkSampleInputDataWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetHshkSampleInputDataWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetHshkSampleInputDataWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetHshkSampleInputDataWhen(taskHandle, data)


@catchDAQError
def DAQmxSetHshkSampleInputDataWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetHshkSampleInputDataWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetHshkSampleInputDataWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetHshkSampleInputDataWhen(taskHandle, data)


@catchDAQError
def DAQmxResetHshkSampleInputDataWhen(taskHandle):
    """int32 __CFUNC DAQmxResetHshkSampleInputDataWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetHshkSampleInputDataWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetHshkSampleInputDataWhen(taskHandle)


@catchDAQError
def DAQmxGetChangeDetectDIRisingEdgePhysicalChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetChangeDetectDIRisingEdgePhysicalChans
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetChangeDetectDIRisingEdgePhysicalChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetChangeDetectDIRisingEdgePhysicalChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetChangeDetectDIRisingEdgePhysicalChans(taskHandle, data):
    """int32 __CFUNC DAQmxSetChangeDetectDIRisingEdgePhysicalChans(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetChangeDetectDIRisingEdgePhysicalChans.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetChangeDetectDIRisingEdgePhysicalChans(taskHandle, data)


@catchDAQError
def DAQmxResetChangeDetectDIRisingEdgePhysicalChans(taskHandle):
    """int32 __CFUNC DAQmxResetChangeDetectDIRisingEdgePhysicalChans(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetChangeDetectDIRisingEdgePhysicalChans.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetChangeDetectDIRisingEdgePhysicalChans(taskHandle)


@catchDAQError
def DAQmxGetChangeDetectDIFallingEdgePhysicalChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetChangeDetectDIFallingEdgePhysicalChans
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetChangeDetectDIFallingEdgePhysicalChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetChangeDetectDIFallingEdgePhysicalChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetChangeDetectDIFallingEdgePhysicalChans(taskHandle, data):
    """int32 __CFUNC DAQmxSetChangeDetectDIFallingEdgePhysicalChans(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetChangeDetectDIFallingEdgePhysicalChans.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetChangeDetectDIFallingEdgePhysicalChans(taskHandle, data)


@catchDAQError
def DAQmxResetChangeDetectDIFallingEdgePhysicalChans(taskHandle):
    """int32 __CFUNC DAQmxResetChangeDetectDIFallingEdgePhysicalChans(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetChangeDetectDIFallingEdgePhysicalChans.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetChangeDetectDIFallingEdgePhysicalChans(taskHandle)


@catchDAQError
def DAQmxGetChangeDetectDITristate(taskHandle, data):
    """int32 __CFUNC DAQmxGetChangeDetectDITristate(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetChangeDetectDITristate.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetChangeDetectDITristate(taskHandle, data)


@catchDAQError
def DAQmxSetChangeDetectDITristate(taskHandle, data):
    """int32 __CFUNC DAQmxSetChangeDetectDITristate(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetChangeDetectDITristate.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetChangeDetectDITristate(taskHandle, data)


@catchDAQError
def DAQmxResetChangeDetectDITristate(taskHandle):
    """int32 __CFUNC DAQmxResetChangeDetectDITristate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetChangeDetectDITristate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetChangeDetectDITristate(taskHandle)


@catchDAQError
def DAQmxGetOnDemandSimultaneousAOEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetOnDemandSimultaneousAOEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetOnDemandSimultaneousAOEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetOnDemandSimultaneousAOEnable(taskHandle, data)


@catchDAQError
def DAQmxSetOnDemandSimultaneousAOEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetOnDemandSimultaneousAOEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetOnDemandSimultaneousAOEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetOnDemandSimultaneousAOEnable(taskHandle, data)


@catchDAQError
def DAQmxResetOnDemandSimultaneousAOEnable(taskHandle):
    """int32 __CFUNC DAQmxResetOnDemandSimultaneousAOEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetOnDemandSimultaneousAOEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetOnDemandSimultaneousAOEnable(taskHandle)


@catchDAQError
def DAQmxGetImplicitUnderflowBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxGetImplicitUnderflowBehavior(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetImplicitUnderflowBehavior.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetImplicitUnderflowBehavior(taskHandle, data)


@catchDAQError
def DAQmxSetImplicitUnderflowBehavior(taskHandle, data):
    """int32 __CFUNC DAQmxSetImplicitUnderflowBehavior(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetImplicitUnderflowBehavior.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetImplicitUnderflowBehavior(taskHandle, data)


@catchDAQError
def DAQmxResetImplicitUnderflowBehavior(taskHandle):
    """int32 __CFUNC DAQmxResetImplicitUnderflowBehavior(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetImplicitUnderflowBehavior.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetImplicitUnderflowBehavior(taskHandle)


@catchDAQError
def DAQmxGetAIConvRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvRate(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAIConvRate(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvRate(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvRate(taskHandle)


@catchDAQError
def DAQmxGetAIConvRateEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvRateEx(TaskHandle taskHandle, const char deviceNames[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvRateEx.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvRateEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvRateEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvRateEx(TaskHandle taskHandle, const char deviceNames[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvRateEx.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIConvRateEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvRateEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvRateEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvRateEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvRateEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvMaxRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvMaxRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvMaxRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvMaxRate(taskHandle, data)


@catchDAQError
def DAQmxGetAIConvMaxRateEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvMaxRateEx(TaskHandle taskHandle, const char deviceNames[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvMaxRateEx.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvMaxRateEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxGetAIConvSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIConvSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIConvSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAIConvSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAIConvSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvSrc(taskHandle)


@catchDAQError
def DAQmxGetAIConvSrcEx(taskHandle, deviceNames, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIConvSrcEx
    (TaskHandle taskHandle, const char deviceNames[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvSrcEx.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIConvSrcEx(taskHandle, deviceNames, data, bufferSize)


@catchDAQError
def DAQmxSetAIConvSrcEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvSrcEx(TaskHandle taskHandle, const char deviceNames[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvSrcEx.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAIConvSrcEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvSrcEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvSrcEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvSrcEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvSrcEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvActiveEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvActiveEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAIConvActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvActiveEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvActiveEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvActiveEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAIConvActiveEdge(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvActiveEdge(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvActiveEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvActiveEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvActiveEdge(taskHandle)


@catchDAQError
def DAQmxGetAIConvActiveEdgeEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvActiveEdgeEx(TaskHandle taskHandle, const char deviceNames[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvActiveEdgeEx.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIConvActiveEdgeEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvActiveEdgeEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvActiveEdgeEx(TaskHandle taskHandle, const char deviceNames[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvActiveEdgeEx.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIConvActiveEdgeEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvActiveEdgeEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvActiveEdgeEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvActiveEdgeEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvActiveEdgeEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvTimebaseDiv(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvTimebaseDiv.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIConvTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvTimebaseDiv(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvTimebaseDiv(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvTimebaseDiv.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetAIConvTimebaseDiv(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvTimebaseDiv(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvTimebaseDiv(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvTimebaseDiv.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvTimebaseDiv(taskHandle)


@catchDAQError
def DAQmxGetAIConvTimebaseDivEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvTimebaseDivEx(TaskHandle taskHandle, const char deviceNames[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvTimebaseDivEx.argtypes = (TaskHandle, c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetAIConvTimebaseDivEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvTimebaseDivEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvTimebaseDivEx(TaskHandle taskHandle, const char deviceNames[], uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvTimebaseDivEx.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxSetAIConvTimebaseDivEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvTimebaseDivEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvTimebaseDivEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvTimebaseDivEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvTimebaseDivEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvTimebaseSrc(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvTimebaseSrc.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAIConvTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvTimebaseSrc(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvTimebaseSrc.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAIConvTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAIConvTimebaseSrcEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvTimebaseSrcEx(TaskHandle taskHandle, const char deviceNames[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvTimebaseSrcEx.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetAIConvTimebaseSrcEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvTimebaseSrcEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvTimebaseSrcEx(TaskHandle taskHandle, const char deviceNames[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvTimebaseSrcEx.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetAIConvTimebaseSrcEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvTimebaseSrcEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvTimebaseSrcEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvTimebaseSrcEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvTimebaseSrcEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetDelayFromSampClkDelayUnits(taskHandle, data):
    """int32 __CFUNC DAQmxGetDelayFromSampClkDelayUnits(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDelayFromSampClkDelayUnits.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDelayFromSampClkDelayUnits(taskHandle, data)


@catchDAQError
def DAQmxSetDelayFromSampClkDelayUnits(taskHandle, data):
    """int32 __CFUNC DAQmxSetDelayFromSampClkDelayUnits(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDelayFromSampClkDelayUnits.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDelayFromSampClkDelayUnits(taskHandle, data)


@catchDAQError
def DAQmxResetDelayFromSampClkDelayUnits(taskHandle):
    """int32 __CFUNC DAQmxResetDelayFromSampClkDelayUnits(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDelayFromSampClkDelayUnits.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDelayFromSampClkDelayUnits(taskHandle)


@catchDAQError
def DAQmxGetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetDelayFromSampClkDelayUnitsEx
    (TaskHandle taskHandle, const char deviceNames[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDelayFromSampClkDelayUnitsEx.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetDelayFromSampClkDelayUnitsEx
    (TaskHandle taskHandle, const char deviceNames[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDelayFromSampClkDelayUnitsEx.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetDelayFromSampClkDelayUnitsEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetDelayFromSampClkDelayUnitsEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDelayFromSampClkDelayUnitsEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetDelayFromSampClkDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetDelayFromSampClkDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDelayFromSampClkDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDelayFromSampClkDelay(taskHandle, data)


@catchDAQError
def DAQmxSetDelayFromSampClkDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetDelayFromSampClkDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDelayFromSampClkDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDelayFromSampClkDelay(taskHandle, data)


@catchDAQError
def DAQmxResetDelayFromSampClkDelay(taskHandle):
    """int32 __CFUNC DAQmxResetDelayFromSampClkDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDelayFromSampClkDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDelayFromSampClkDelay(taskHandle)


@catchDAQError
def DAQmxGetDelayFromSampClkDelayEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetDelayFromSampClkDelayEx
    (TaskHandle taskHandle, const char deviceNames[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDelayFromSampClkDelayEx.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetDelayFromSampClkDelayEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetDelayFromSampClkDelayEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetDelayFromSampClkDelayEx(TaskHandle taskHandle, const char deviceNames[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDelayFromSampClkDelayEx.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetDelayFromSampClkDelayEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetDelayFromSampClkDelayEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetDelayFromSampClkDelayEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetDelayFromSampClkDelayEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetDelayFromSampClkDelayEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAIConvDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAIConvDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAIConvDigFltrEnableEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrEnableEx(TaskHandle taskHandle, const char deviceNames[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrEnableEx.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIConvDigFltrEnableEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvDigFltrEnableEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrEnableEx(TaskHandle taskHandle, const char deviceNames[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrEnableEx.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIConvDigFltrEnableEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvDigFltrEnableEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvDigFltrEnableEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrEnableEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvDigFltrEnableEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAIConvDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrMinPulseWidthEx
    (TaskHandle taskHandle, const char deviceNames[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrMinPulseWidthEx.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrMinPulseWidthEx
    (TaskHandle taskHandle, const char deviceNames[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrMinPulseWidthEx.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvDigFltrMinPulseWidthEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrMinPulseWidthEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvDigFltrMinPulseWidthEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIConvDigFltrTimebaseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIConvDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAIConvDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAIConvDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames, data, bufferSize):
    """int32 __CFUNC DAQmxGetAIConvDigFltrTimebaseSrcEx
    (TaskHandle taskHandle, const char deviceNames[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrTimebaseSrcEx.argtypes = (TaskHandle, c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames, data, bufferSize)


@catchDAQError
def DAQmxSetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrTimebaseSrcEx
    (TaskHandle taskHandle, const char deviceNames[], const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrTimebaseSrcEx.argtypes = (TaskHandle, c_char_p, c_char_p,)
    return DAQlib.DAQmxSetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvDigFltrTimebaseSrcEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrTimebaseSrcEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvDigFltrTimebaseSrcEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAIConvDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvDigFltrTimebaseRateEx
    (TaskHandle taskHandle, const char deviceNames[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigFltrTimebaseRateEx.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvDigFltrTimebaseRateEx
    (TaskHandle taskHandle, const char deviceNames[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigFltrTimebaseRateEx.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvDigFltrTimebaseRateEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigFltrTimebaseRateEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvDigFltrTimebaseRateEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetAIConvDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAIConvDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAIConvDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAIConvDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAIConvDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAIConvDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAIConvDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAIConvDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAIConvDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetAIConvDigSyncEnableEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxGetAIConvDigSyncEnableEx(TaskHandle taskHandle, const char deviceNames[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAIConvDigSyncEnableEx.argtypes = (TaskHandle, c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAIConvDigSyncEnableEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxSetAIConvDigSyncEnableEx(taskHandle, deviceNames, data):
    """int32 __CFUNC DAQmxSetAIConvDigSyncEnableEx(TaskHandle taskHandle, const char deviceNames[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAIConvDigSyncEnableEx.argtypes = (TaskHandle, c_char_p, bool32,)
    return DAQlib.DAQmxSetAIConvDigSyncEnableEx(taskHandle, deviceNames, data)


@catchDAQError
def DAQmxResetAIConvDigSyncEnableEx(taskHandle, deviceNames):
    """int32 __CFUNC DAQmxResetAIConvDigSyncEnableEx(TaskHandle taskHandle, const char deviceNames[]);"""
    global DAQlib
    DAQlib.DAQmxResetAIConvDigSyncEnableEx.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetAIConvDigSyncEnableEx(taskHandle, deviceNames)


@catchDAQError
def DAQmxGetMasterTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetMasterTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetMasterTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetMasterTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetMasterTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetMasterTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetMasterTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetMasterTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetMasterTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetMasterTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetMasterTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetMasterTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetMasterTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetMasterTimebaseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetMasterTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetMasterTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetMasterTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetMasterTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetMasterTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetMasterTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetMasterTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetMasterTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetMasterTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetMasterTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetRefClkRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefClkRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefClkRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetRefClkRate(taskHandle, data)


@catchDAQError
def DAQmxSetRefClkRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefClkRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetRefClkRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetRefClkRate(taskHandle, data)


@catchDAQError
def DAQmxResetRefClkRate(taskHandle):
    """int32 __CFUNC DAQmxResetRefClkRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefClkRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefClkRate(taskHandle)


@catchDAQError
def DAQmxGetRefClkSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetRefClkSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetRefClkSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetRefClkSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetRefClkSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefClkSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetRefClkSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetRefClkSrc(taskHandle, data)


@catchDAQError
def DAQmxResetRefClkSrc(taskHandle):
    """int32 __CFUNC DAQmxResetRefClkSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefClkSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefClkSrc(taskHandle)


@catchDAQError
def DAQmxGetSyncPulseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSyncPulseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSyncPulseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetSyncPulseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetSyncPulseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetSyncPulseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetSyncPulseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetSyncPulseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetSyncPulseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSyncPulseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSyncPulseSrc(taskHandle)


@catchDAQError
def DAQmxGetSyncPulseSyncTime(taskHandle, data):
    """int32 __CFUNC DAQmxGetSyncPulseSyncTime(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseSyncTime.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSyncPulseSyncTime(taskHandle, data)


@catchDAQError
def DAQmxGetSyncPulseMinDelayToStart(taskHandle, data):
    """int32 __CFUNC DAQmxGetSyncPulseMinDelayToStart(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseMinDelayToStart.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSyncPulseMinDelayToStart(taskHandle, data)


@catchDAQError
def DAQmxSetSyncPulseMinDelayToStart(taskHandle, data):
    """int32 __CFUNC DAQmxSetSyncPulseMinDelayToStart(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSyncPulseMinDelayToStart.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSyncPulseMinDelayToStart(taskHandle, data)


@catchDAQError
def DAQmxResetSyncPulseMinDelayToStart(taskHandle):
    """int32 __CFUNC DAQmxResetSyncPulseMinDelayToStart(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSyncPulseMinDelayToStart.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSyncPulseMinDelayToStart(taskHandle)


@catchDAQError
def DAQmxGetSyncPulseResetTime(taskHandle, data):
    """int32 __CFUNC DAQmxGetSyncPulseResetTime(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseResetTime.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSyncPulseResetTime(taskHandle, data)


@catchDAQError
def DAQmxGetSyncPulseResetDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetSyncPulseResetDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseResetDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetSyncPulseResetDelay(taskHandle, data)


@catchDAQError
def DAQmxSetSyncPulseResetDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetSyncPulseResetDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetSyncPulseResetDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetSyncPulseResetDelay(taskHandle, data)


@catchDAQError
def DAQmxResetSyncPulseResetDelay(taskHandle):
    """int32 __CFUNC DAQmxResetSyncPulseResetDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSyncPulseResetDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSyncPulseResetDelay(taskHandle)


@catchDAQError
def DAQmxGetSyncPulseTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetSyncPulseTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetSyncPulseTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetSyncPulseTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetSyncClkInterval(taskHandle, data):
    """int32 __CFUNC DAQmxGetSyncClkInterval(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSyncClkInterval.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetSyncClkInterval(taskHandle, data)


@catchDAQError
def DAQmxSetSyncClkInterval(taskHandle, data):
    """int32 __CFUNC DAQmxSetSyncClkInterval(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSyncClkInterval.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetSyncClkInterval(taskHandle, data)


@catchDAQError
def DAQmxResetSyncClkInterval(taskHandle):
    """int32 __CFUNC DAQmxResetSyncClkInterval(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSyncClkInterval.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSyncClkInterval(taskHandle)


@catchDAQError
def DAQmxGetSampTimingEngine(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampTimingEngine(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampTimingEngine.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetSampTimingEngine(taskHandle, data)


@catchDAQError
def DAQmxSetSampTimingEngine(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampTimingEngine(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampTimingEngine.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetSampTimingEngine(taskHandle, data)


@catchDAQError
def DAQmxResetSampTimingEngine(taskHandle):
    """int32 __CFUNC DAQmxResetSampTimingEngine(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampTimingEngine.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampTimingEngine(taskHandle)


@catchDAQError
def DAQmxGetStartTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetStartTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetStartTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetStartTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetStartTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetStartTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetStartTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetStartTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetStartTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetStartTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetStartTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetStartTrigType(taskHandle)


@catchDAQError
def DAQmxGetStartTrigTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetStartTrigTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetStartTrigTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetStartTrigTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetDigEdgeStartTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeStartTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeStartTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeStartTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigEdgeStartTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeStartTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigEdgeStartTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigEdge(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigEdge(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeStartTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeStartTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeStartTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetDigPatternStartTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternStartTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternStartTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternStartTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternStartTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternStartTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternStartTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternStartTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternStartTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternStartTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternStartTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternStartTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigPatternStartTrigPattern(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternStartTrigPattern(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternStartTrigPattern.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternStartTrigPattern(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternStartTrigPattern(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternStartTrigPattern(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternStartTrigPattern.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternStartTrigPattern(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternStartTrigPattern(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternStartTrigPattern(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternStartTrigPattern.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternStartTrigPattern(taskHandle)


@catchDAQError
def DAQmxGetDigPatternStartTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigPatternStartTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternStartTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigPatternStartTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetDigPatternStartTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternStartTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternStartTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigPatternStartTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternStartTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternStartTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternStartTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternStartTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigSlope(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigSlope(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigSlope.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigSlope(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigSlope(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigSlope(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigSlope.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigSlope(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigSlope(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigSlope(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigSlope.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigSlope(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigLvl(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigLvl.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigLvl(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigLvl.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigLvl(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigLvl(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigHyst(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigHyst.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigHyst(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigHyst.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigHyst(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigHyst(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigHyst.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigHyst(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgEdgeStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgEdgeStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeStartTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeStartTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeStartTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeStartTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinStartTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinStartTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinStartTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinStartTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinStartTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigTop(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigTop.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinStartTrigTop(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigTop(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigTop.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinStartTrigTop(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigTop(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigTop(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigTop.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigTop(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigBtm(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigBtm.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinStartTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigBtm(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigBtm.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinStartTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigBtm(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigBtm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigBtm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigBtm(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinStartTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinStartTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinStartTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinStartTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinStartTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinStartTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinStartTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinStartTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetStartTrigDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetStartTrigDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetStartTrigDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetStartTrigDelay(taskHandle, data)


@catchDAQError
def DAQmxSetStartTrigDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetStartTrigDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetStartTrigDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetStartTrigDelay(taskHandle, data)


@catchDAQError
def DAQmxResetStartTrigDelay(taskHandle):
    """int32 __CFUNC DAQmxResetStartTrigDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetStartTrigDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetStartTrigDelay(taskHandle)


@catchDAQError
def DAQmxGetStartTrigDelayUnits(taskHandle, data):
    """int32 __CFUNC DAQmxGetStartTrigDelayUnits(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetStartTrigDelayUnits.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetStartTrigDelayUnits(taskHandle, data)


@catchDAQError
def DAQmxSetStartTrigDelayUnits(taskHandle, data):
    """int32 __CFUNC DAQmxSetStartTrigDelayUnits(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetStartTrigDelayUnits.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetStartTrigDelayUnits(taskHandle, data)


@catchDAQError
def DAQmxResetStartTrigDelayUnits(taskHandle):
    """int32 __CFUNC DAQmxResetStartTrigDelayUnits(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetStartTrigDelayUnits.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetStartTrigDelayUnits(taskHandle)


@catchDAQError
def DAQmxGetStartTrigRetriggerable(taskHandle, data):
    """int32 __CFUNC DAQmxGetStartTrigRetriggerable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetStartTrigRetriggerable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetStartTrigRetriggerable(taskHandle, data)


@catchDAQError
def DAQmxSetStartTrigRetriggerable(taskHandle, data):
    """int32 __CFUNC DAQmxSetStartTrigRetriggerable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetStartTrigRetriggerable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetStartTrigRetriggerable(taskHandle, data)


@catchDAQError
def DAQmxResetStartTrigRetriggerable(taskHandle):
    """int32 __CFUNC DAQmxResetStartTrigRetriggerable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetStartTrigRetriggerable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetStartTrigRetriggerable(taskHandle)


@catchDAQError
def DAQmxGetRefTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetRefTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetRefTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRefTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetRefTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetRefTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetRefTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefTrigType(taskHandle)


@catchDAQError
def DAQmxGetRefTrigPretrigSamples(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefTrigPretrigSamples(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigPretrigSamples.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetRefTrigPretrigSamples(taskHandle, data)


@catchDAQError
def DAQmxSetRefTrigPretrigSamples(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefTrigPretrigSamples(TaskHandle taskHandle, uInt32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRefTrigPretrigSamples.argtypes = (TaskHandle, uInt32,)
    return DAQlib.DAQmxSetRefTrigPretrigSamples(taskHandle, data)


@catchDAQError
def DAQmxResetRefTrigPretrigSamples(taskHandle):
    """int32 __CFUNC DAQmxResetRefTrigPretrigSamples(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefTrigPretrigSamples.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefTrigPretrigSamples(taskHandle)


@catchDAQError
def DAQmxGetRefTrigTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetRefTrigTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetRefTrigTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetDigEdgeRefTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeRefTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeRefTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeRefTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigEdgeRefTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeRefTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigEdgeRefTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigEdge(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigEdge(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeRefTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeRefTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeRefTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetDigPatternRefTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternRefTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternRefTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternRefTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternRefTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternRefTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternRefTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternRefTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternRefTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternRefTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternRefTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternRefTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigPatternRefTrigPattern(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternRefTrigPattern(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternRefTrigPattern.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternRefTrigPattern(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternRefTrigPattern(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternRefTrigPattern(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternRefTrigPattern.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternRefTrigPattern(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternRefTrigPattern(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternRefTrigPattern(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternRefTrigPattern.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternRefTrigPattern(taskHandle)


@catchDAQError
def DAQmxGetDigPatternRefTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigPatternRefTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternRefTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigPatternRefTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetDigPatternRefTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternRefTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternRefTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigPatternRefTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternRefTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternRefTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternRefTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternRefTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigSlope(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigSlope(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigSlope.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigSlope(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigSlope(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigSlope(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigSlope.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigSlope(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigSlope(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigSlope(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigSlope.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigSlope(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigLvl(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigLvl.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigLvl(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigLvl.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigLvl(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigLvl(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigHyst(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigHyst.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigHyst(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigHyst.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigHyst(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigHyst(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigHyst.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigHyst(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgEdgeRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgEdgeRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgEdgeRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgEdgeRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgEdgeRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgEdgeRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgEdgeRefTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgEdgeRefTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgEdgeRefTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgEdgeRefTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinRefTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinRefTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinRefTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinRefTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinRefTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigTop(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigTop.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinRefTrigTop(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigTop(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigTop.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinRefTrigTop(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigTop(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigTop(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigTop.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigTop(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigBtm(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigBtm.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinRefTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigBtm(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigBtm.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinRefTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigBtm(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigBtm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigBtm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigBtm(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinRefTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinRefTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinRefTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinRefTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinRefTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinRefTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinRefTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinRefTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinRefTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinRefTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinRefTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinRefTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetRefTrigAutoTrigEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefTrigAutoTrigEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigAutoTrigEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetRefTrigAutoTrigEnable(taskHandle, data)


@catchDAQError
def DAQmxSetRefTrigAutoTrigEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefTrigAutoTrigEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetRefTrigAutoTrigEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetRefTrigAutoTrigEnable(taskHandle, data)


@catchDAQError
def DAQmxResetRefTrigAutoTrigEnable(taskHandle):
    """int32 __CFUNC DAQmxResetRefTrigAutoTrigEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefTrigAutoTrigEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefTrigAutoTrigEnable(taskHandle)


@catchDAQError
def DAQmxGetRefTrigAutoTriggered(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefTrigAutoTriggered(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigAutoTriggered.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetRefTrigAutoTriggered(taskHandle, data)


@catchDAQError
def DAQmxGetRefTrigDelay(taskHandle, data):
    """int32 __CFUNC DAQmxGetRefTrigDelay(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetRefTrigDelay.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetRefTrigDelay(taskHandle, data)


@catchDAQError
def DAQmxSetRefTrigDelay(taskHandle, data):
    """int32 __CFUNC DAQmxSetRefTrigDelay(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetRefTrigDelay.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetRefTrigDelay(taskHandle, data)


@catchDAQError
def DAQmxResetRefTrigDelay(taskHandle):
    """int32 __CFUNC DAQmxResetRefTrigDelay(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetRefTrigDelay.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetRefTrigDelay(taskHandle)


@catchDAQError
def DAQmxGetAdvTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetAdvTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAdvTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAdvTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetAdvTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetAdvTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAdvTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAdvTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetAdvTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetAdvTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAdvTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAdvTrigType(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeAdvTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeAdvTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeAdvTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeAdvTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeAdvTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeAdvTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeAdvTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeAdvTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeAdvTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeAdvTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeAdvTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeAdvTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeAdvTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeAdvTrigEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeAdvTrigEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigEdgeAdvTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeAdvTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeAdvTrigEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeAdvTrigEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigEdgeAdvTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeAdvTrigEdge(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeAdvTrigEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeAdvTrigEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeAdvTrigEdge(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeAdvTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeAdvTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeAdvTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeAdvTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeAdvTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeAdvTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeAdvTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeAdvTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeAdvTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeAdvTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeAdvTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeAdvTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetHshkTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetHshkTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetHshkTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetHshkTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetHshkTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetHshkTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetHshkTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetHshkTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetHshkTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetHshkTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetHshkTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetHshkTrigType(taskHandle)


@catchDAQError
def DAQmxGetInterlockedHshkTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetInterlockedHshkTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetInterlockedHshkTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetInterlockedHshkTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetInterlockedHshkTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetInterlockedHshkTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetInterlockedHshkTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetInterlockedHshkTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetInterlockedHshkTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetInterlockedHshkTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetInterlockedHshkTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetInterlockedHshkTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetInterlockedHshkTrigAssertedLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetInterlockedHshkTrigAssertedLvl(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetInterlockedHshkTrigAssertedLvl.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetInterlockedHshkTrigAssertedLvl(taskHandle, data)


@catchDAQError
def DAQmxSetInterlockedHshkTrigAssertedLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetInterlockedHshkTrigAssertedLvl(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetInterlockedHshkTrigAssertedLvl.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetInterlockedHshkTrigAssertedLvl(taskHandle, data)


@catchDAQError
def DAQmxResetInterlockedHshkTrigAssertedLvl(taskHandle):
    """int32 __CFUNC DAQmxResetInterlockedHshkTrigAssertedLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetInterlockedHshkTrigAssertedLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetInterlockedHshkTrigAssertedLvl(taskHandle)


@catchDAQError
def DAQmxGetPauseTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetPauseTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPauseTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetPauseTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetPauseTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetPauseTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetPauseTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetPauseTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetPauseTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetPauseTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetPauseTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetPauseTrigType(taskHandle)


@catchDAQError
def DAQmxGetPauseTrigTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetPauseTrigTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPauseTrigTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPauseTrigTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigLvl(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigLvl.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigLvl(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigLvl(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigLvl.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigLvl(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigLvl(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigLvl(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigLvl.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigLvl(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigHyst(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigHyst.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigHyst(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigHyst(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigHyst.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigHyst(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigHyst(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigHyst(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigHyst.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigHyst(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgLvlPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgLvlPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgLvlPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgLvlPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgLvlPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgLvlPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgLvlPauseTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgLvlPauseTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgLvlPauseTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinPauseTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigTop(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigTop.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigTop(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigTop(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigTop(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigTop.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigTop(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigTop(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigTop(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigTop.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigTop(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigBtm(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigBtm.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigBtm(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigBtm(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigBtm.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigBtm(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigBtm(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigBtm(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigBtm.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigBtm(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigCoupling(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigCoupling.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigCoupling(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigCoupling(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigCoupling.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigCoupling(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigCoupling(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigCoupling(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigCoupling.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigCoupling(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetAnlgWinPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetAnlgWinPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAnlgWinPauseTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetAnlgWinPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetAnlgWinPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetAnlgWinPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAnlgWinPauseTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetAnlgWinPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetAnlgWinPauseTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetAnlgWinPauseTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetAnlgWinPauseTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetAnlgWinPauseTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigLvlPauseTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigLvlPauseTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigLvlPauseTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigLvlPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetDigLvlPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigLvlPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigLvlPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigLvlPauseTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigLvlPauseTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetDigLvlPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigLvlPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigLvlPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigLvlPauseTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigLvlPauseTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigLvlPauseTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigLvlPauseTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigLvlPauseTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigLvlPauseTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigLvlPauseTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetDigPatternPauseTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternPauseTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternPauseTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternPauseTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternPauseTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternPauseTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternPauseTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternPauseTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternPauseTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternPauseTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternPauseTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternPauseTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigPatternPauseTrigPattern(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigPatternPauseTrigPattern(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternPauseTrigPattern.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigPatternPauseTrigPattern(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigPatternPauseTrigPattern(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternPauseTrigPattern(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternPauseTrigPattern.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigPatternPauseTrigPattern(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternPauseTrigPattern(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternPauseTrigPattern(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternPauseTrigPattern.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternPauseTrigPattern(taskHandle)


@catchDAQError
def DAQmxGetDigPatternPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigPatternPauseTrigWhen(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigPatternPauseTrigWhen.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigPatternPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxSetDigPatternPauseTrigWhen(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigPatternPauseTrigWhen(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigPatternPauseTrigWhen.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigPatternPauseTrigWhen(taskHandle, data)


@catchDAQError
def DAQmxResetDigPatternPauseTrigWhen(taskHandle):
    """int32 __CFUNC DAQmxResetDigPatternPauseTrigWhen(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigPatternPauseTrigWhen.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigPatternPauseTrigWhen(taskHandle)


@catchDAQError
def DAQmxGetArmStartTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetArmStartTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetArmStartTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetArmStartTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetArmStartTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetArmStartTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetArmStartTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetArmStartTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetArmStartTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetArmStartTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetArmStartTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetArmStartTrigType(taskHandle)


@catchDAQError
def DAQmxGetArmStartTerm(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetArmStartTerm(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetArmStartTerm.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetArmStartTerm(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigEdge(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigEdge(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigDigFltrEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigDigFltrEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigDigFltrEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigDigFltrEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrEnable(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeArmStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeArmStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeArmStartTrigDigSyncEnable.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetDigEdgeArmStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeArmStartTrigDigSyncEnable(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeArmStartTrigDigSyncEnable(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeArmStartTrigDigSyncEnable.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetDigEdgeArmStartTrigDigSyncEnable(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeArmStartTrigDigSyncEnable(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeArmStartTrigDigSyncEnable(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeArmStartTrigDigSyncEnable.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeArmStartTrigDigSyncEnable(taskHandle)


@catchDAQError
def DAQmxGetTriggerSyncType(taskHandle, data):
    """int32 __CFUNC DAQmxGetTriggerSyncType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetTriggerSyncType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetTriggerSyncType(taskHandle, data)


@catchDAQError
def DAQmxSetTriggerSyncType(taskHandle, data):
    """int32 __CFUNC DAQmxSetTriggerSyncType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetTriggerSyncType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetTriggerSyncType(taskHandle, data)


@catchDAQError
def DAQmxResetTriggerSyncType(taskHandle):
    """int32 __CFUNC DAQmxResetTriggerSyncType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetTriggerSyncType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetTriggerSyncType(taskHandle)


@catchDAQError
def DAQmxGetWatchdogTimeout(taskHandle, data):
    """int32 __CFUNC DAQmxGetWatchdogTimeout(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogTimeout.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetWatchdogTimeout(taskHandle, data)


@catchDAQError
def DAQmxSetWatchdogTimeout(taskHandle, data):
    """int32 __CFUNC DAQmxSetWatchdogTimeout(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogTimeout.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetWatchdogTimeout(taskHandle, data)


@catchDAQError
def DAQmxResetWatchdogTimeout(taskHandle):
    """int32 __CFUNC DAQmxResetWatchdogTimeout(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogTimeout.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWatchdogTimeout(taskHandle)


@catchDAQError
def DAQmxGetWatchdogExpirTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxGetWatchdogExpirTrigType(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogExpirTrigType.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetWatchdogExpirTrigType(taskHandle, data)


@catchDAQError
def DAQmxSetWatchdogExpirTrigType(taskHandle, data):
    """int32 __CFUNC DAQmxSetWatchdogExpirTrigType(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogExpirTrigType.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWatchdogExpirTrigType(taskHandle, data)


@catchDAQError
def DAQmxResetWatchdogExpirTrigType(taskHandle):
    """int32 __CFUNC DAQmxResetWatchdogExpirTrigType(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogExpirTrigType.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWatchdogExpirTrigType(taskHandle)


@catchDAQError
def DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle, data):
    """int32 __CFUNC DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle, data)


@catchDAQError
def DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle, data):
    """int32 __CFUNC DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle, data)


@catchDAQError
def DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle):
    """int32 __CFUNC DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeWatchdogExpirTrigSrc(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetDigEdgeWatchdogExpirTrigSrc(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeWatchdogExpirTrigSrc.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetDigEdgeWatchdogExpirTrigSrc(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxSetDigEdgeWatchdogExpirTrigSrc(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeWatchdogExpirTrigSrc(TaskHandle taskHandle, const char \*data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeWatchdogExpirTrigSrc.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxSetDigEdgeWatchdogExpirTrigSrc(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeWatchdogExpirTrigSrc(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeWatchdogExpirTrigSrc(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeWatchdogExpirTrigSrc.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeWatchdogExpirTrigSrc(taskHandle)


@catchDAQError
def DAQmxGetDigEdgeWatchdogExpirTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxGetDigEdgeWatchdogExpirTrigEdge(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetDigEdgeWatchdogExpirTrigEdge.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetDigEdgeWatchdogExpirTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxSetDigEdgeWatchdogExpirTrigEdge(taskHandle, data):
    """int32 __CFUNC DAQmxSetDigEdgeWatchdogExpirTrigEdge(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetDigEdgeWatchdogExpirTrigEdge.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetDigEdgeWatchdogExpirTrigEdge(taskHandle, data)


@catchDAQError
def DAQmxResetDigEdgeWatchdogExpirTrigEdge(taskHandle):
    """int32 __CFUNC DAQmxResetDigEdgeWatchdogExpirTrigEdge(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetDigEdgeWatchdogExpirTrigEdge.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetDigEdgeWatchdogExpirTrigEdge(taskHandle)


@catchDAQError
def DAQmxGetWatchdogDOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxGetWatchdogDOExpirState(TaskHandle taskHandle, const char lines[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogDOExpirState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetWatchdogDOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxSetWatchdogDOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxSetWatchdogDOExpirState(TaskHandle taskHandle, const char lines[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogDOExpirState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetWatchdogDOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxResetWatchdogDOExpirState(taskHandle, lines):
    """int32 __CFUNC DAQmxResetWatchdogDOExpirState(TaskHandle taskHandle, const char lines[]);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogDOExpirState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetWatchdogDOExpirState(taskHandle, lines)


@catchDAQError
def DAQmxGetWatchdogAOOutputType(taskHandle, lines, data):
    """int32 __CFUNC DAQmxGetWatchdogAOOutputType(TaskHandle taskHandle, const char lines[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogAOOutputType.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetWatchdogAOOutputType(taskHandle, lines, data)


@catchDAQError
def DAQmxSetWatchdogAOOutputType(taskHandle, lines, data):
    """int32 __CFUNC DAQmxSetWatchdogAOOutputType(TaskHandle taskHandle, const char lines[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogAOOutputType.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetWatchdogAOOutputType(taskHandle, lines, data)


@catchDAQError
def DAQmxResetWatchdogAOOutputType(taskHandle, lines):
    """int32 __CFUNC DAQmxResetWatchdogAOOutputType(TaskHandle taskHandle, const char lines[]);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogAOOutputType.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetWatchdogAOOutputType(taskHandle, lines)


@catchDAQError
def DAQmxGetWatchdogAOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxGetWatchdogAOExpirState(TaskHandle taskHandle, const char lines[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogAOExpirState.argtypes = (TaskHandle, c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetWatchdogAOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxSetWatchdogAOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxSetWatchdogAOExpirState(TaskHandle taskHandle, const char lines[], float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogAOExpirState.argtypes = (TaskHandle, c_char_p, float64,)
    return DAQlib.DAQmxSetWatchdogAOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxResetWatchdogAOExpirState(taskHandle, lines):
    """int32 __CFUNC DAQmxResetWatchdogAOExpirState(TaskHandle taskHandle, const char lines[]);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogAOExpirState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetWatchdogAOExpirState(taskHandle, lines)


@catchDAQError
def DAQmxGetWatchdogCOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxGetWatchdogCOExpirState(TaskHandle taskHandle, const char lines[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogCOExpirState.argtypes = (TaskHandle, c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetWatchdogCOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxSetWatchdogCOExpirState(taskHandle, lines, data):
    """int32 __CFUNC DAQmxSetWatchdogCOExpirState(TaskHandle taskHandle, const char lines[], int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWatchdogCOExpirState.argtypes = (TaskHandle, c_char_p, int32,)
    return DAQlib.DAQmxSetWatchdogCOExpirState(taskHandle, lines, data)


@catchDAQError
def DAQmxResetWatchdogCOExpirState(taskHandle, lines):
    """int32 __CFUNC DAQmxResetWatchdogCOExpirState(TaskHandle taskHandle, const char lines[]);"""
    global DAQlib
    DAQlib.DAQmxResetWatchdogCOExpirState.argtypes = (TaskHandle, c_char_p,)
    return DAQlib.DAQmxResetWatchdogCOExpirState(taskHandle, lines)


@catchDAQError
def DAQmxGetWatchdogHasExpired(taskHandle, data):
    """int32 __CFUNC DAQmxGetWatchdogHasExpired(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWatchdogHasExpired.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWatchdogHasExpired(taskHandle, data)


@catchDAQError
def DAQmxGetWriteRelativeTo(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteRelativeTo(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteRelativeTo.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetWriteRelativeTo(taskHandle, data)


@catchDAQError
def DAQmxSetWriteRelativeTo(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteRelativeTo(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteRelativeTo.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWriteRelativeTo(taskHandle, data)


@catchDAQError
def DAQmxResetWriteRelativeTo(taskHandle):
    """int32 __CFUNC DAQmxResetWriteRelativeTo(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteRelativeTo.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteRelativeTo(taskHandle)


@catchDAQError
def DAQmxGetWriteOffset(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteOffset(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOffset.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetWriteOffset(taskHandle, data)


@catchDAQError
def DAQmxSetWriteOffset(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteOffset(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteOffset.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWriteOffset(taskHandle, data)


@catchDAQError
def DAQmxResetWriteOffset(taskHandle):
    """int32 __CFUNC DAQmxResetWriteOffset(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteOffset.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteOffset(taskHandle)


@catchDAQError
def DAQmxGetWriteRegenMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteRegenMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteRegenMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetWriteRegenMode(taskHandle, data)


@catchDAQError
def DAQmxSetWriteRegenMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteRegenMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteRegenMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWriteRegenMode(taskHandle, data)


@catchDAQError
def DAQmxResetWriteRegenMode(taskHandle):
    """int32 __CFUNC DAQmxResetWriteRegenMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteRegenMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteRegenMode(taskHandle)


@catchDAQError
def DAQmxGetWriteCurrWritePos(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteCurrWritePos(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteCurrWritePos.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetWriteCurrWritePos(taskHandle, data)


@catchDAQError
def DAQmxGetWriteOvercurrentChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteOvercurrentChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOvercurrentChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteOvercurrentChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetWriteOvercurrentChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWriteOvercurrentChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOvercurrentChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWriteOvercurrentChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetWriteOvertemperatureChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteOvertemperatureChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOvertemperatureChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteOvertemperatureChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetWriteOvertemperatureChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWriteOvertemperatureChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOvertemperatureChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWriteOvertemperatureChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetWriteOverloadedChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteOverloadedChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOverloadedChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteOverloadedChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetWriteOverloadedChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWriteOverloadedChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOverloadedChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWriteOverloadedChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetWriteOpenCurrentLoopChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteOpenCurrentLoopChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOpenCurrentLoopChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteOpenCurrentLoopChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetWriteOpenCurrentLoopChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWriteOpenCurrentLoopChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWriteOpenCurrentLoopChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWriteOpenCurrentLoopChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetWritePowerSupplyFaultChansExist(taskHandle, data):
    """int32 __CFUNC DAQmxGetWritePowerSupplyFaultChansExist(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWritePowerSupplyFaultChansExist.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWritePowerSupplyFaultChansExist(taskHandle, data)


@catchDAQError
def DAQmxGetWritePowerSupplyFaultChans(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWritePowerSupplyFaultChans(TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWritePowerSupplyFaultChans.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWritePowerSupplyFaultChans(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetWriteSpaceAvail(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteSpaceAvail(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteSpaceAvail.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetWriteSpaceAvail(taskHandle, data)


@catchDAQError
def DAQmxGetWriteTotalSampPerChanGenerated(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteTotalSampPerChanGenerated(TaskHandle taskHandle, uInt64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteTotalSampPerChanGenerated.argtypes = (TaskHandle, POINTER(uInt64),)
    return DAQlib.DAQmxGetWriteTotalSampPerChanGenerated(taskHandle, data)


@catchDAQError
def DAQmxGetWriteRawDataWidth(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteRawDataWidth(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteRawDataWidth.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetWriteRawDataWidth(taskHandle, data)


@catchDAQError
def DAQmxGetWriteNumChans(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteNumChans(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteNumChans.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetWriteNumChans(taskHandle, data)


@catchDAQError
def DAQmxGetWriteWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteWaitMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteWaitMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetWriteWaitMode(taskHandle, data)


@catchDAQError
def DAQmxSetWriteWaitMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteWaitMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteWaitMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetWriteWaitMode(taskHandle, data)


@catchDAQError
def DAQmxResetWriteWaitMode(taskHandle):
    """int32 __CFUNC DAQmxResetWriteWaitMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteWaitMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteWaitMode(taskHandle)


@catchDAQError
def DAQmxGetWriteSleepTime(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteSleepTime(TaskHandle taskHandle, float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteSleepTime.argtypes = (TaskHandle, POINTER(float64),)
    return DAQlib.DAQmxGetWriteSleepTime(taskHandle, data)


@catchDAQError
def DAQmxSetWriteSleepTime(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteSleepTime(TaskHandle taskHandle, float64 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteSleepTime.argtypes = (TaskHandle, float64,)
    return DAQlib.DAQmxSetWriteSleepTime(taskHandle, data)


@catchDAQError
def DAQmxResetWriteSleepTime(taskHandle):
    """int32 __CFUNC DAQmxResetWriteSleepTime(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteSleepTime.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteSleepTime(taskHandle)


@catchDAQError
def DAQmxGetWriteNextWriteIsLast(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteNextWriteIsLast(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteNextWriteIsLast.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteNextWriteIsLast(taskHandle, data)


@catchDAQError
def DAQmxSetWriteNextWriteIsLast(taskHandle, data):
    """int32 __CFUNC DAQmxSetWriteNextWriteIsLast(TaskHandle taskHandle, bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetWriteNextWriteIsLast.argtypes = (TaskHandle, bool32,)
    return DAQlib.DAQmxSetWriteNextWriteIsLast(taskHandle, data)


@catchDAQError
def DAQmxResetWriteNextWriteIsLast(taskHandle):
    """int32 __CFUNC DAQmxResetWriteNextWriteIsLast(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetWriteNextWriteIsLast.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetWriteNextWriteIsLast(taskHandle)


@catchDAQError
def DAQmxGetWriteDigitalLinesBytesPerChan(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteDigitalLinesBytesPerChan(TaskHandle taskHandle, uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteDigitalLinesBytesPerChan.argtypes = (TaskHandle, POINTER(uInt32),)
    return DAQlib.DAQmxGetWriteDigitalLinesBytesPerChan(taskHandle, data)


@catchDAQError
def DAQmxGetWriteAccessoryInsertionOrRemovalDetected(taskHandle, data):
    """int32 __CFUNC DAQmxGetWriteAccessoryInsertionOrRemovalDetected(TaskHandle taskHandle, bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetWriteAccessoryInsertionOrRemovalDetected.argtypes = (TaskHandle, POINTER(bool32),)
    return DAQlib.DAQmxGetWriteAccessoryInsertionOrRemovalDetected(taskHandle, data)


@catchDAQError
def DAQmxGetWriteDevsWithInsertedOrRemovedAccessories(taskHandle, data, bufferSize):
    """int32 __CFUNC DAQmxGetWriteDevsWithInsertedOrRemovedAccessories
    (TaskHandle taskHandle, char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetWriteDevsWithInsertedOrRemovedAccessories.argtypes = (TaskHandle, c_char_p, uInt32,)
    return DAQlib.DAQmxGetWriteDevsWithInsertedOrRemovedAccessories(taskHandle, data, bufferSize)


@catchDAQError
def DAQmxGetPhysicalChanAISupportedMeasTypes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanAISupportedMeasTypes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAISupportedMeasTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanAISupportedMeasTypes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanAITermCfgs(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAITermCfgs(const char physicalChannel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAITermCfgs.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetPhysicalChanAITermCfgs(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanAIInputSrcs(physicalChannel, data, bufferSize):
    """int32 __CFUNC DAQmxGetPhysicalChanAIInputSrcs(const char physicalChannel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAIInputSrcs.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPhysicalChanAIInputSrcs(physicalChannel, data, bufferSize)


@catchDAQError
def DAQmxGetPhysicalChanAOSupportedOutputTypes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanAOSupportedOutputTypes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOSupportedOutputTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanAOSupportedOutputTypes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanAOSupportedPowerUpOutputTypes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanAOSupportedPowerUpOutputTypes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOSupportedPowerUpOutputTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanAOSupportedPowerUpOutputTypes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanAOTermCfgs(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAOTermCfgs(const char physicalChannel[], int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOTermCfgs.argtypes = (c_char_p, POINTER(int32),)
    return DAQlib.DAQmxGetPhysicalChanAOTermCfgs(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanAOManualControlEnable(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAOManualControlEnable(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOManualControlEnable.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPhysicalChanAOManualControlEnable(physicalChannel, data)


@catchDAQError
def DAQmxSetPhysicalChanAOManualControlEnable(physicalChannel, data):
    """int32 __CFUNC DAQmxSetPhysicalChanAOManualControlEnable(const char physicalChannel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetPhysicalChanAOManualControlEnable.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSetPhysicalChanAOManualControlEnable(physicalChannel, data)


@catchDAQError
def DAQmxResetPhysicalChanAOManualControlEnable(physicalChannel):
    """int32 __CFUNC DAQmxResetPhysicalChanAOManualControlEnable(const char physicalChannel[]);"""
    global DAQlib
    DAQlib.DAQmxResetPhysicalChanAOManualControlEnable.argtypes = (c_char_p,)
    return DAQlib.DAQmxResetPhysicalChanAOManualControlEnable(physicalChannel)


@catchDAQError
def DAQmxGetPhysicalChanAOManualControlShortDetected(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAOManualControlShortDetected(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOManualControlShortDetected.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPhysicalChanAOManualControlShortDetected(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanAOManualControlAmplitude(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAOManualControlAmplitude(const char physicalChannel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOManualControlAmplitude.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetPhysicalChanAOManualControlAmplitude(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanAOManualControlFreq(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanAOManualControlFreq(const char physicalChannel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanAOManualControlFreq.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetPhysicalChanAOManualControlFreq(physicalChannel, data)


@catchDAQError
def DAQmxGetAOPowerAmpChannelEnable(physicalChannel, data):
    """int32 __CFUNC DAQmxGetAOPowerAmpChannelEnable(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOPowerAmpChannelEnable.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOPowerAmpChannelEnable(physicalChannel, data)


@catchDAQError
def DAQmxSetAOPowerAmpChannelEnable(physicalChannel, data):
    """int32 __CFUNC DAQmxSetAOPowerAmpChannelEnable(const char physicalChannel[], bool32 data);"""
    global DAQlib
    DAQlib.DAQmxSetAOPowerAmpChannelEnable.argtypes = (c_char_p, bool32,)
    return DAQlib.DAQmxSetAOPowerAmpChannelEnable(physicalChannel, data)


@catchDAQError
def DAQmxResetAOPowerAmpChannelEnable(physicalChannel):
    """int32 __CFUNC DAQmxResetAOPowerAmpChannelEnable(const char physicalChannel[]);"""
    global DAQlib
    DAQlib.DAQmxResetAOPowerAmpChannelEnable.argtypes = (c_char_p,)
    return DAQlib.DAQmxResetAOPowerAmpChannelEnable(physicalChannel)


@catchDAQError
def DAQmxGetAOPowerAmpScalingCoeff(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetAOPowerAmpScalingCoeff
    (const char physicalChannel[], float64 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetAOPowerAmpScalingCoeff.argtypes = (c_char_p, POINTER(float64), uInt32,)
    return DAQlib.DAQmxGetAOPowerAmpScalingCoeff(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetAOPowerAmpOvercurrent(physicalChannel, data):
    """int32 __CFUNC DAQmxGetAOPowerAmpOvercurrent(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOPowerAmpOvercurrent.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetAOPowerAmpOvercurrent(physicalChannel, data)


@catchDAQError
def DAQmxGetAOPowerAmpGain(physicalChannel, data):
    """int32 __CFUNC DAQmxGetAOPowerAmpGain(const char physicalChannel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOPowerAmpGain.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOPowerAmpGain(physicalChannel, data)


@catchDAQError
def DAQmxGetAOPowerAmpOffset(physicalChannel, data):
    """int32 __CFUNC DAQmxGetAOPowerAmpOffset(const char physicalChannel[], float64 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetAOPowerAmpOffset.argtypes = (c_char_p, POINTER(float64),)
    return DAQlib.DAQmxGetAOPowerAmpOffset(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDIPortWidth(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanDIPortWidth(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDIPortWidth.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanDIPortWidth(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDISampClkSupported(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanDISampClkSupported(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDISampClkSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPhysicalChanDISampClkSupported(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDISampModes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanDISampModes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDISampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanDISampModes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanDIChangeDetectSupported(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanDIChangeDetectSupported(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDIChangeDetectSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPhysicalChanDIChangeDetectSupported(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDOPortWidth(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanDOPortWidth(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDOPortWidth.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanDOPortWidth(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDOSampClkSupported(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanDOSampClkSupported(const char physicalChannel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDOSampClkSupported.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPhysicalChanDOSampClkSupported(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanDOSampModes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanDOSampModes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanDOSampModes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanDOSampModes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanCISupportedMeasTypes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanCISupportedMeasTypes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanCISupportedMeasTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanCISupportedMeasTypes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanCOSupportedOutputTypes(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanCOSupportedOutputTypes
    (const char physicalChannel[], int32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanCOSupportedOutputTypes.argtypes = (c_char_p, POINTER(int32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanCOSupportedOutputTypes(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanTEDSMfgID(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSMfgID(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSMfgID.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanTEDSMfgID(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanTEDSModelNum(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSModelNum(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSModelNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanTEDSModelNum(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanTEDSSerialNum(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSSerialNum(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSSerialNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanTEDSSerialNum(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanTEDSVersionNum(physicalChannel, data):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSVersionNum(const char physicalChannel[], uInt32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSVersionNum.argtypes = (c_char_p, POINTER(uInt32),)
    return DAQlib.DAQmxGetPhysicalChanTEDSVersionNum(physicalChannel, data)


@catchDAQError
def DAQmxGetPhysicalChanTEDSVersionLetter(physicalChannel, data, bufferSize):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSVersionLetter
    (const char physicalChannel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSVersionLetter.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPhysicalChanTEDSVersionLetter(physicalChannel, data, bufferSize)


@catchDAQError
def DAQmxGetPhysicalChanTEDSBitStream(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSBitStream
    (const char physicalChannel[], uInt8 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSBitStream.argtypes = (c_char_p, POINTER(uInt8), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanTEDSBitStream(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPhysicalChanTEDSTemplateIDs(physicalChannel, data, arraySizeInElements):
    """int32 __CFUNC DAQmxGetPhysicalChanTEDSTemplateIDs
    (const char physicalChannel[], uInt32 \*data, uInt32 arraySizeInElements);"""
    global DAQlib
    DAQlib.DAQmxGetPhysicalChanTEDSTemplateIDs.argtypes = (c_char_p, POINTER(uInt32), uInt32,)
    return DAQlib.DAQmxGetPhysicalChanTEDSTemplateIDs(physicalChannel, data, arraySizeInElements)


@catchDAQError
def DAQmxGetPersistedTaskAuthor(taskName, data, bufferSize):
    """int32 __CFUNC DAQmxGetPersistedTaskAuthor(const char taskName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedTaskAuthor.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPersistedTaskAuthor(taskName, data, bufferSize)


@catchDAQError
def DAQmxGetPersistedTaskAllowInteractiveEditing(taskName, data):
    """int32 __CFUNC DAQmxGetPersistedTaskAllowInteractiveEditing(const char taskName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedTaskAllowInteractiveEditing.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedTaskAllowInteractiveEditing(taskName, data)


@catchDAQError
def DAQmxGetPersistedTaskAllowInteractiveDeletion(taskName, data):
    """int32 __CFUNC DAQmxGetPersistedTaskAllowInteractiveDeletion(const char taskName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedTaskAllowInteractiveDeletion.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedTaskAllowInteractiveDeletion(taskName, data)


@catchDAQError
def DAQmxGetPersistedChanAuthor(channel, data, bufferSize):
    """int32 __CFUNC DAQmxGetPersistedChanAuthor(const char channel[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedChanAuthor.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPersistedChanAuthor(channel, data, bufferSize)


@catchDAQError
def DAQmxGetPersistedChanAllowInteractiveEditing(channel, data):
    """int32 __CFUNC DAQmxGetPersistedChanAllowInteractiveEditing(const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedChanAllowInteractiveEditing.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedChanAllowInteractiveEditing(channel, data)


@catchDAQError
def DAQmxGetPersistedChanAllowInteractiveDeletion(channel, data):
    """int32 __CFUNC DAQmxGetPersistedChanAllowInteractiveDeletion(const char channel[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedChanAllowInteractiveDeletion.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedChanAllowInteractiveDeletion(channel, data)


@catchDAQError
def DAQmxGetPersistedScaleAuthor(scaleName, data, bufferSize):
    """int32 __CFUNC DAQmxGetPersistedScaleAuthor(const char scaleName[], char \*data, uInt32 bufferSize);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedScaleAuthor.argtypes = (c_char_p, c_char_p, uInt32,)
    return DAQlib.DAQmxGetPersistedScaleAuthor(scaleName, data, bufferSize)


@catchDAQError
def DAQmxGetPersistedScaleAllowInteractiveEditing(scaleName, data):
    """int32 __CFUNC DAQmxGetPersistedScaleAllowInteractiveEditing(const char scaleName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedScaleAllowInteractiveEditing.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedScaleAllowInteractiveEditing(scaleName, data)


@catchDAQError
def DAQmxGetPersistedScaleAllowInteractiveDeletion(scaleName, data):
    """int32 __CFUNC DAQmxGetPersistedScaleAllowInteractiveDeletion(const char scaleName[], bool32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetPersistedScaleAllowInteractiveDeletion.argtypes = (c_char_p, POINTER(bool32),)
    return DAQlib.DAQmxGetPersistedScaleAllowInteractiveDeletion(scaleName, data)


@catchDAQError
def DAQmxGetSampClkTimingResponseMode(taskHandle, data):
    """int32 __CFUNC DAQmxGetSampClkTimingResponseMode(TaskHandle taskHandle, int32 \*data);"""
    global DAQlib
    DAQlib.DAQmxGetSampClkTimingResponseMode.argtypes = (TaskHandle, POINTER(int32),)
    return DAQlib.DAQmxGetSampClkTimingResponseMode(taskHandle, data)


@catchDAQError
def DAQmxSetSampClkTimingResponseMode(taskHandle, data):
    """int32 __CFUNC DAQmxSetSampClkTimingResponseMode(TaskHandle taskHandle, int32 data);"""
    global DAQlib
    DAQlib.DAQmxSetSampClkTimingResponseMode.argtypes = (TaskHandle, int32,)
    return DAQlib.DAQmxSetSampClkTimingResponseMode(taskHandle, data)


@catchDAQError
def DAQmxResetSampClkTimingResponseMode(taskHandle):
    """int32 __CFUNC DAQmxResetSampClkTimingResponseMode(TaskHandle taskHandle);"""
    global DAQlib
    DAQlib.DAQmxResetSampClkTimingResponseMode.argtypes = (TaskHandle,)
    return DAQlib.DAQmxResetSampClkTimingResponseMode(taskHandle)


if __name__ == "__main__":
    import unittest

    class TestDAQmx(unittest.TestCase):

        def test_1(self):
            print(DAQmxCreateTask.__doc__)
            taskHandle = c_uint32(0)
            print(DAQmxCreateTask("task", byref(taskHandle)))
            print("taskHandle:", taskHandle)
            print(DAQmxCreateTask("task2", byref(taskHandle)))
            # print(DAQmxCreateTask("task", taskHandle))
            print("taskHandle:", taskHandle)
    unittest.main(verbosity=2)
