# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from .core import *
import random
import string


class Task(object):
    """
    ``Task`` automatically creates a DAQmx task.

    Every method that is called does not need the prefix ``DAQmx`` and you must not
    supply the first argument if it is a taskhandle.

    The task is automatically cleared when the destructor is called.
    """

    def __init__(self, name):
        self.name = name.replace(":", "_") + "_" + ''.join(random.choice(string.ascii_uppercase +
                                                                         string.ascii_lowercase + string.digits)
                                                           for _ in range(10))
        if not isinstance(self.name, (bytes)):
            self.name = self.name.encode()
        self.taskHandle = TaskHandle()
        DAQmxCreateTask(self.name, byref(self.taskHandle))

    def AddGlobalChansToTask(self, channelNames):
        return DAQmxAddGlobalChansToTask(self.taskHandle, channelNames)

    def StartTask(self):
        return DAQmxStartTask(self.taskHandle)

    def StopTask(self):
        return DAQmxStopTask(self.taskHandle)

    def ClearTask(self):
        return DAQmxClearTask(self.taskHandle)

    def WaitUntilTaskDone(self, timeToWait):
        return DAQmxWaitUntilTaskDone(self.taskHandle, timeToWait)

    def IsTaskDone(self, isTaskDone):
        return DAQmxIsTaskDone(self.taskHandle, isTaskDone)

    def TaskControl(self, action):
        return DAQmxTaskControl(self.taskHandle, action)

    def GetNthTaskChannel(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskChannel(self.taskHandle, index, buffer, bufferSize)

    def GetNthTaskDevice(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskDevice(self.taskHandle, index, buffer, bufferSize)

    def GetTaskAttribute(self, attribute, value):
        return DAQmxGetTaskAttribute(self.taskHandle, attribute, value)

    def CreateAIVoltageChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                            customScaleName):
        return DAQmxCreateAIVoltageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                        minVal, maxVal, units, customScaleName)

    def CreateAICurrentChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                            shuntResistorLoc, extShuntResistorVal, customScaleName):
        return DAQmxCreateAICurrentChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                        minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal, customScaleName)

    def CreateAIVoltageRMSChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                               customScaleName):
        return DAQmxCreateAIVoltageRMSChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                           minVal, maxVal, units, customScaleName)

    def CreateAICurrentRMSChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                               shuntResistorLoc, extShuntResistorVal, customScaleName):
        return DAQmxCreateAICurrentRMSChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                           minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal,
                                           customScaleName)

    def CreateAIThrmcplChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, thermocoupleType,
                            cjcSource, cjcVal, cjcChannel):
        return DAQmxCreateAIThrmcplChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        thermocoupleType, cjcSource, cjcVal, cjcChannel)

    def CreateAIRTDChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, rtdType, resistanceConfig,
                        currentExcitSource, currentExcitVal, r0):
        return DAQmxCreateAIRTDChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    rtdType, resistanceConfig, currentExcitSource, currentExcitVal, r0)

    def CreateAIThrmstrChanIex(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, resistanceConfig,
                               currentExcitSource, currentExcitVal, a, b, c):
        return DAQmxCreateAIThrmstrChanIex(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, resistanceConfig, currentExcitSource, currentExcitVal, a, b, c)

    def CreateAIThrmstrChanVex(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, resistanceConfig,
                               voltageExcitSource, voltageExcitVal, a, b, c, r1):
        return DAQmxCreateAIThrmstrChanVex(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, resistanceConfig, voltageExcitSource, voltageExcitVal, a, b, c, r1)

    def CreateAIFreqVoltageChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, thresholdLevel,
                                hysteresis, customScaleName):
        return DAQmxCreateAIFreqVoltageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                            units, thresholdLevel, hysteresis, customScaleName)

    def CreateAIResistanceChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, resistanceConfig,
                               currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateAIResistanceChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, resistanceConfig, currentExcitSource, currentExcitVal,
                                           customScaleName)

    def CreateAIStrainGageChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, strainConfig,
                               voltageExcitSource, voltageExcitVal, gageFactor, initialBridgeVoltage,
                               nominalGageResistance, poissonRatio, leadWireResistance, customScaleName):
        return DAQmxCreateAIStrainGageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, strainConfig, voltageExcitSource, voltageExcitVal, gageFactor,
                                           initialBridgeVoltage, nominalGageResistance, poissonRatio,
                                           leadWireResistance, customScaleName)

    def CreateAIRosetteStrainGageChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, rosetteType,
                                      gageOrientation, rosetteMeasTypes, numRosetteMeasTypes, strainConfig,
                                      voltageExcitSource, voltageExcitVal, gageFactor, nominalGageResistance,
                                      poissonRatio, leadWireResistance):
        return DAQmxCreateAIRosetteStrainGageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                  maxVal, rosetteType, gageOrientation, rosetteMeasTypes,
                                                  numRosetteMeasTypes, strainConfig, voltageExcitSource,
                                                  voltageExcitVal, gageFactor, nominalGageResistance, poissonRatio,
                                                  leadWireResistance)

    def CreateAIForceBridgeTwoPointLinChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                           firstElectricalVal, secondElectricalVal, electricalUnits, firstPhysicalVal,
                                           secondPhysicalVal, physicalUnits, customScaleName):
        return DAQmxCreateAIForceBridgeTwoPointLinChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                       maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                       nominalBridgeResistance, firstElectricalVal, secondElectricalVal,
                                                       electricalUnits, firstPhysicalVal, secondPhysicalVal,
                                                       physicalUnits, customScaleName)

    def CreateAIForceBridgeTableChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, bridgeConfig,
                                     voltageExcitSource, voltageExcitVal, nominalBridgeResistance, electricalVals,
                                     numElectricalVals, electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                     customScaleName):
        return DAQmxCreateAIForceBridgeTableChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                 maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                 nominalBridgeResistance, electricalVals, numElectricalVals,
                                                 electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                 customScaleName)

    def CreateAIForceBridgePolynomialChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                          bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                          forwardCoeffs, numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                          electricalUnits, physicalUnits, customScaleName):
        return DAQmxCreateAIForceBridgePolynomialChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                      maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                      nominalBridgeResistance, forwardCoeffs, numForwardCoeffs,
                                                      reverseCoeffs, numReverseCoeffs, electricalUnits, physicalUnits,
                                                      customScaleName)

    def CreateAIPressureBridgeTwoPointLinChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                              bridgeConfig, voltageExcitSource, voltageExcitVal,
                                              nominalBridgeResistance, firstElectricalVal, secondElectricalVal,
                                              electricalUnits, firstPhysicalVal, secondPhysicalVal, physicalUnits,
                                              customScaleName):
        return DAQmxCreateAIPressureBridgeTwoPointLinChan(self.taskHandle, physicalChannel, nameToAssignToChannel,
                                                          minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                          voltageExcitVal, nominalBridgeResistance, firstElectricalVal,
                                                          secondElectricalVal, electricalUnits, firstPhysicalVal,
                                                          secondPhysicalVal, physicalUnits, customScaleName)

    def CreateAIPressureBridgeTableChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                        electricalVals, numElectricalVals, electricalUnits, physicalVals,
                                        numPhysicalVals, physicalUnits, customScaleName):
        return DAQmxCreateAIPressureBridgeTableChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                    maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                    nominalBridgeResistance, electricalVals, numElectricalVals,
                                                    electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                    customScaleName)

    def CreateAIPressureBridgePolynomialChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                             bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                             forwardCoeffs, numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                             electricalUnits, physicalUnits, customScaleName):
        return DAQmxCreateAIPressureBridgePolynomialChan(self.taskHandle, physicalChannel, nameToAssignToChannel,
                                                         minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                         voltageExcitVal, nominalBridgeResistance, forwardCoeffs,
                                                         numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                                         electricalUnits, physicalUnits, customScaleName)

    def CreateAITorqueBridgeTwoPointLinChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                            bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                            firstElectricalVal, secondElectricalVal, electricalUnits, firstPhysicalVal,
                                            secondPhysicalVal, physicalUnits, customScaleName):
        return DAQmxCreateAITorqueBridgeTwoPointLinChan(self.taskHandle, physicalChannel, nameToAssignToChannel,
                                                        minVal, maxVal, units, bridgeConfig, voltageExcitSource,
                                                        voltageExcitVal, nominalBridgeResistance, firstElectricalVal,
                                                        secondElectricalVal, electricalUnits, firstPhysicalVal,
                                                        secondPhysicalVal, physicalUnits, customScaleName)

    def CreateAITorqueBridgeTableChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, bridgeConfig,
                                      voltageExcitSource, voltageExcitVal, nominalBridgeResistance, electricalVals,
                                      numElectricalVals, electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                      customScaleName):
        return DAQmxCreateAITorqueBridgeTableChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                  maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                  nominalBridgeResistance, electricalVals, numElectricalVals,
                                                  electricalUnits, physicalVals, numPhysicalVals, physicalUnits,
                                                  customScaleName)

    def CreateAITorqueBridgePolynomialChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                           bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                           forwardCoeffs, numForwardCoeffs, reverseCoeffs, numReverseCoeffs,
                                           electricalUnits, physicalUnits, customScaleName):
        return DAQmxCreateAITorqueBridgePolynomialChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                       maxVal, units, bridgeConfig, voltageExcitSource, voltageExcitVal,
                                                       nominalBridgeResistance, forwardCoeffs, numForwardCoeffs,
                                                       reverseCoeffs, numReverseCoeffs, electricalUnits, physicalUnits,
                                                       customScaleName)

    def CreateAIBridgeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, bridgeConfig,
                           voltageExcitSource, voltageExcitVal, nominalBridgeResistance, customScaleName):
        return DAQmxCreateAIBridgeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                       bridgeConfig, voltageExcitSource, voltageExcitVal, nominalBridgeResistance,
                                       customScaleName)

    def CreateAIVoltageChanWithExcit(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                     units, bridgeConfig, voltageExcitSource, voltageExcitVal, useExcitForScaling,
                                     customScaleName):
        return DAQmxCreateAIVoltageChanWithExcit(self.taskHandle, physicalChannel, nameToAssignToChannel,
                                                 terminalConfig, minVal, maxVal, units, bridgeConfig,
                                                 voltageExcitSource, voltageExcitVal, useExcitForScaling,
                                                 customScaleName)

    def CreateAITempBuiltInSensorChan(self, physicalChannel, nameToAssignToChannel, units):
        return DAQmxCreateAITempBuiltInSensorChan(self.taskHandle, physicalChannel, nameToAssignToChannel, units)

    def CreateAIAccelChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                          sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateAIAccelChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                      minVal, maxVal, units, sensitivity, sensitivityUnits, currentExcitSource,
                                      currentExcitVal, customScaleName)

    def CreateAIVelocityIEPEChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                                 sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateAIVelocityIEPEChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                             minVal, maxVal, units, sensitivity, sensitivityUnits, currentExcitSource,
                                             currentExcitVal, customScaleName)

    def CreateAIForceIEPEChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                              sensitivity, sensitivityUnits, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateAIForceIEPEChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                          minVal, maxVal, units, sensitivity, sensitivityUnits, currentExcitSource,
                                          currentExcitVal, customScaleName)

    def CreateAIMicrophoneChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, units, micSensitivity,
                               maxSndPressLevel, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateAIMicrophoneChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                           units, micSensitivity, maxSndPressLevel, currentExcitSource, currentExcitVal,
                                           customScaleName)

    def CreateAIPosLVDTChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, sensitivity,
                            sensitivityUnits, voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                            customScaleName):
        return DAQmxCreateAIPosLVDTChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        sensitivity, sensitivityUnits, voltageExcitSource, voltageExcitVal,
                                        voltageExcitFreq, ACExcitWireMode, customScaleName)

    def CreateAIPosRVDTChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, sensitivity,
                            sensitivityUnits, voltageExcitSource, voltageExcitVal, voltageExcitFreq, ACExcitWireMode,
                            customScaleName):
        return DAQmxCreateAIPosRVDTChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        sensitivity, sensitivityUnits, voltageExcitSource, voltageExcitVal,
                                        voltageExcitFreq, ACExcitWireMode, customScaleName)

    def CreateAIPosEddyCurrProxProbeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                         sensitivity, sensitivityUnits, customScaleName):
        return DAQmxCreateAIPosEddyCurrProxProbeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                     maxVal, units, sensitivity, sensitivityUnits, customScaleName)

    def CreateAIDeviceTempChan(self, physicalChannel, nameToAssignToChannel, units):
        return DAQmxCreateAIDeviceTempChan(self.taskHandle, physicalChannel, nameToAssignToChannel, units)

    def CreateTEDSAIVoltageChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                                customScaleName):
        return DAQmxCreateTEDSAIVoltageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                            minVal, maxVal, units, customScaleName)

    def CreateTEDSAICurrentChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                                shuntResistorLoc, extShuntResistorVal, customScaleName):
        return DAQmxCreateTEDSAICurrentChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                            minVal, maxVal, units, shuntResistorLoc, extShuntResistorVal,
                                            customScaleName)

    def CreateTEDSAIThrmcplChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, cjcSource, cjcVal,
                                cjcChannel):
        return DAQmxCreateTEDSAIThrmcplChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                            units, cjcSource, cjcVal, cjcChannel)

    def CreateTEDSAIRTDChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, resistanceConfig,
                            currentExcitSource, currentExcitVal):
        return DAQmxCreateTEDSAIRTDChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        resistanceConfig, currentExcitSource, currentExcitVal)

    def CreateTEDSAIThrmstrChanIex(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                   resistanceConfig, currentExcitSource, currentExcitVal):
        return DAQmxCreateTEDSAIThrmstrChanIex(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, resistanceConfig, currentExcitSource, currentExcitVal)

    def CreateTEDSAIThrmstrChanVex(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                   resistanceConfig, voltageExcitSource, voltageExcitVal, r1):
        return DAQmxCreateTEDSAIThrmstrChanVex(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, resistanceConfig, voltageExcitSource, voltageExcitVal, r1)

    def CreateTEDSAIResistanceChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                   resistanceConfig, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateTEDSAIResistanceChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, resistanceConfig, currentExcitSource, currentExcitVal,
                                               customScaleName)

    def CreateTEDSAIStrainGageChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                   voltageExcitSource, voltageExcitVal, initialBridgeVoltage, leadWireResistance,
                                   customScaleName):
        return DAQmxCreateTEDSAIStrainGageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                               units, voltageExcitSource, voltageExcitVal, initialBridgeVoltage,
                                               leadWireResistance, customScaleName)

    def CreateTEDSAIForceBridgeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                    voltageExcitSource, voltageExcitVal, customScaleName):
        return DAQmxCreateTEDSAIForceBridgeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                                units, voltageExcitSource, voltageExcitVal, customScaleName)

    def CreateTEDSAIPressureBridgeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                       voltageExcitSource, voltageExcitVal, customScaleName):
        return DAQmxCreateTEDSAIPressureBridgeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                   maxVal, units, voltageExcitSource, voltageExcitVal, customScaleName)

    def CreateTEDSAITorqueBridgeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                     voltageExcitSource, voltageExcitVal, customScaleName):
        return DAQmxCreateTEDSAITorqueBridgeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal,
                                                 maxVal, units, voltageExcitSource, voltageExcitVal, customScaleName)

    def CreateTEDSAIBridgeChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, voltageExcitSource,
                               voltageExcitVal, customScaleName):
        return DAQmxCreateTEDSAIBridgeChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                           units, voltageExcitSource, voltageExcitVal, customScaleName)

    def CreateTEDSAIVoltageChanWithExcit(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal,
                                         units, voltageExcitSource, voltageExcitVal, customScaleName):
        return DAQmxCreateTEDSAIVoltageChanWithExcit(self.taskHandle, physicalChannel, nameToAssignToChannel,
                                                     terminalConfig, minVal, maxVal, units, voltageExcitSource,
                                                     voltageExcitVal, customScaleName)

    def CreateTEDSAIAccelChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                              currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateTEDSAIAccelChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                          minVal, maxVal, units, currentExcitSource, currentExcitVal, customScaleName)

    def CreateTEDSAIForceIEPEChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, minVal, maxVal, units,
                                  currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateTEDSAIForceIEPEChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                              minVal, maxVal, units, currentExcitSource, currentExcitVal,
                                              customScaleName)

    def CreateTEDSAIMicrophoneChan(self, physicalChannel, nameToAssignToChannel, terminalConfig, units,
                                   maxSndPressLevel, currentExcitSource, currentExcitVal, customScaleName):
        return DAQmxCreateTEDSAIMicrophoneChan(self.taskHandle, physicalChannel, nameToAssignToChannel, terminalConfig,
                                               units, maxSndPressLevel, currentExcitSource, currentExcitVal,
                                               customScaleName)

    def CreateTEDSAIPosLVDTChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, voltageExcitSource,
                                voltageExcitVal, voltageExcitFreq, ACExcitWireMode, customScaleName):
        return DAQmxCreateTEDSAIPosLVDTChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                            units, voltageExcitSource, voltageExcitVal, voltageExcitFreq,
                                            ACExcitWireMode, customScaleName)

    def CreateTEDSAIPosRVDTChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, voltageExcitSource,
                                voltageExcitVal, voltageExcitFreq, ACExcitWireMode, customScaleName):
        return DAQmxCreateTEDSAIPosRVDTChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal,
                                            units, voltageExcitSource, voltageExcitVal, voltageExcitFreq,
                                            ACExcitWireMode, customScaleName)

    def CreateAOVoltageChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, customScaleName):
        return DAQmxCreateAOVoltageChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        customScaleName)

    def CreateAOCurrentChan(self, physicalChannel, nameToAssignToChannel, minVal, maxVal, units, customScaleName):
        return DAQmxCreateAOCurrentChan(self.taskHandle, physicalChannel, nameToAssignToChannel, minVal, maxVal, units,
                                        customScaleName)

    def CreateAOFuncGenChan(self, physicalChannel, nameToAssignToChannel, type, freq, amplitude, offset):
        return DAQmxCreateAOFuncGenChan(self.taskHandle, physicalChannel, nameToAssignToChannel, type, freq, amplitude,
                                        offset)

    def CreateDIChan(self, lines, nameToAssignToLines, lineGrouping):
        return DAQmxCreateDIChan(self.taskHandle, lines, nameToAssignToLines, lineGrouping)

    def CreateDOChan(self, lines, nameToAssignToLines, lineGrouping):
        return DAQmxCreateDOChan(self.taskHandle, lines, nameToAssignToLines, lineGrouping)

    def CreateCIFreqChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, edge, measMethod, measTime,
                         divisor, customScaleName):
        return DAQmxCreateCIFreqChan(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge,
                                     measMethod, measTime, divisor, customScaleName)

    def CreateCIPeriodChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, edge, measMethod, measTime,
                           divisor, customScaleName):
        return DAQmxCreateCIPeriodChan(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units, edge,
                                       measMethod, measTime, divisor, customScaleName)

    def CreateCICountEdgesChan(self, counter, nameToAssignToChannel, edge, initialCount, countDirection):
        return DAQmxCreateCICountEdgesChan(self.taskHandle, counter, nameToAssignToChannel, edge, initialCount,
                                           countDirection)

    def CreateCIPulseWidthChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, startingEdge,
                               customScaleName):
        return DAQmxCreateCIPulseWidthChan(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                           startingEdge, customScaleName)

    def CreateCISemiPeriodChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, customScaleName):
        return DAQmxCreateCISemiPeriodChan(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                           customScaleName)

    def CreateCITwoEdgeSepChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, firstEdge, secondEdge,
                               customScaleName):
        return DAQmxCreateCITwoEdgeSepChan(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units,
                                           firstEdge, secondEdge, customScaleName)

    def CreateCIPulseChanFreq(self, counter, nameToAssignToChannel, minVal, maxVal, units):
        return DAQmxCreateCIPulseChanFreq(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units)

    def CreateCIPulseChanTime(self, counter, nameToAssignToChannel, minVal, maxVal, units):
        return DAQmxCreateCIPulseChanTime(self.taskHandle, counter, nameToAssignToChannel, minVal, maxVal, units)

    def CreateCIPulseChanTicks(self, counter, nameToAssignToChannel, sourceTerminal, minVal, maxVal):
        return DAQmxCreateCIPulseChanTicks(self.taskHandle, counter, nameToAssignToChannel, sourceTerminal, minVal,
                                           maxVal)

    def CreateCILinEncoderChan(self, counter, nameToAssignToChannel, decodingType, ZidxEnable, ZidxVal, ZidxPhase,
                               units, distPerPulse, initialPos, customScaleName):
        return DAQmxCreateCILinEncoderChan(self.taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable,
                                           ZidxVal, ZidxPhase, units, distPerPulse, initialPos, customScaleName)

    def CreateCIAngEncoderChan(self, counter, nameToAssignToChannel, decodingType, ZidxEnable, ZidxVal, ZidxPhase,
                               units, pulsesPerRev, initialAngle, customScaleName):
        return DAQmxCreateCIAngEncoderChan(self.taskHandle, counter, nameToAssignToChannel, decodingType, ZidxEnable,
                                           ZidxVal, ZidxPhase, units, pulsesPerRev, initialAngle, customScaleName)

    def CreateCIGPSTimestampChan(self, counter, nameToAssignToChannel, units, syncMethod, customScaleName):
        return DAQmxCreateCIGPSTimestampChan(self.taskHandle, counter, nameToAssignToChannel, units, syncMethod,
                                             customScaleName)

    def CreateCOPulseChanFreq(self, counter, nameToAssignToChannel, units, idleState, initialDelay, freq, dutyCycle):
        return DAQmxCreateCOPulseChanFreq(self.taskHandle, counter, nameToAssignToChannel, units, idleState,
                                          initialDelay, freq, dutyCycle)

    def CreateCOPulseChanTime(self, counter, nameToAssignToChannel, units, idleState, initialDelay, lowTime, highTime):
        return DAQmxCreateCOPulseChanTime(self.taskHandle, counter, nameToAssignToChannel, units, idleState,
                                          initialDelay, lowTime, highTime)

    def CreateCOPulseChanTicks(self, counter, nameToAssignToChannel, sourceTerminal, idleState, initialDelay, lowTicks,
                               highTicks):
        return DAQmxCreateCOPulseChanTicks(self.taskHandle, counter, nameToAssignToChannel, sourceTerminal, idleState,
                                           initialDelay, lowTicks, highTicks)

    def GetAIChanCalCalDate(self, channelName, year, month, day, hour, minute):
        return DAQmxGetAIChanCalCalDate(self.taskHandle, channelName, year, month, day, hour, minute)

    def SetAIChanCalCalDate(self, channelName, year, month, day, hour, minute):
        return DAQmxSetAIChanCalCalDate(self.taskHandle, channelName, year, month, day, hour, minute)

    def GetAIChanCalExpDate(self, channelName, year, month, day, hour, minute):
        return DAQmxGetAIChanCalExpDate(self.taskHandle, channelName, year, month, day, hour, minute)

    def SetAIChanCalExpDate(self, channelName, year, month, day, hour, minute):
        return DAQmxSetAIChanCalExpDate(self.taskHandle, channelName, year, month, day, hour, minute)

    def GetChanAttribute(self, channel, attribute, value):
        return DAQmxGetChanAttribute(self.taskHandle, channel, attribute, value)

    def SetChanAttribute(self, channel, attribute):
        return DAQmxSetChanAttribute(self.taskHandle, channel, attribute)

    def ResetChanAttribute(self, channel, attribute):
        return DAQmxResetChanAttribute(self.taskHandle, channel, attribute)

    def CfgSampClkTiming(self, source, rate, activeEdge, sampleMode, sampsPerChan):
        return DAQmxCfgSampClkTiming(self.taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan)

    def CfgHandshakingTiming(self, sampleMode, sampsPerChan):
        return DAQmxCfgHandshakingTiming(self.taskHandle, sampleMode, sampsPerChan)

    def CfgBurstHandshakingTimingImportClock(self, sampleMode, sampsPerChan, sampleClkRate, sampleClkSrc,
                                             sampleClkActiveEdge, pauseWhen, readyEventActiveLevel):
        return DAQmxCfgBurstHandshakingTimingImportClock(self.taskHandle, sampleMode, sampsPerChan, sampleClkRate,
                                                         sampleClkSrc, sampleClkActiveEdge, pauseWhen,
                                                         readyEventActiveLevel)

    def CfgBurstHandshakingTimingExportClock(self, sampleMode, sampsPerChan, sampleClkRate, sampleClkOutpTerm,
                                             sampleClkPulsePolarity, pauseWhen, readyEventActiveLevel):
        return DAQmxCfgBurstHandshakingTimingExportClock(self.taskHandle, sampleMode, sampsPerChan, sampleClkRate,
                                                         sampleClkOutpTerm, sampleClkPulsePolarity, pauseWhen,
                                                         readyEventActiveLevel)

    def CfgChangeDetectionTiming(self, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan):
        return DAQmxCfgChangeDetectionTiming(self.taskHandle, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan)

    def CfgImplicitTiming(self, sampleMode, sampsPerChan):
        return DAQmxCfgImplicitTiming(self.taskHandle, sampleMode, sampsPerChan)

    def CfgPipelinedSampClkTiming(self, source, rate, activeEdge, sampleMode, sampsPerChan):
        return DAQmxCfgPipelinedSampClkTiming(self.taskHandle, source, rate, activeEdge, sampleMode, sampsPerChan)

    def GetTimingAttribute(self, attribute, value):
        return DAQmxGetTimingAttribute(self.taskHandle, attribute, value)

    def SetTimingAttribute(self, attribute):
        return DAQmxSetTimingAttribute(self.taskHandle, attribute)

    def ResetTimingAttribute(self, attribute):
        return DAQmxResetTimingAttribute(self.taskHandle, attribute)

    def GetTimingAttributeEx(self, deviceNames, attribute, value):
        return DAQmxGetTimingAttributeEx(self.taskHandle, deviceNames, attribute, value)

    def SetTimingAttributeEx(self, deviceNames, attribute):
        return DAQmxSetTimingAttributeEx(self.taskHandle, deviceNames, attribute)

    def ResetTimingAttributeEx(self, deviceNames, attribute):
        return DAQmxResetTimingAttributeEx(self.taskHandle, deviceNames, attribute)

    def DisableStartTrig(self):
        return DAQmxDisableStartTrig(self.taskHandle)

    def CfgDigEdgeStartTrig(self, triggerSource, triggerEdge):
        return DAQmxCfgDigEdgeStartTrig(self.taskHandle, triggerSource, triggerEdge)

    def CfgAnlgEdgeStartTrig(self, triggerSource, triggerSlope, triggerLevel):
        return DAQmxCfgAnlgEdgeStartTrig(self.taskHandle, triggerSource, triggerSlope, triggerLevel)

    def CfgAnlgWindowStartTrig(self, triggerSource, triggerWhen, windowTop, windowBottom):
        return DAQmxCfgAnlgWindowStartTrig(self.taskHandle, triggerSource, triggerWhen, windowTop, windowBottom)

    def CfgDigPatternStartTrig(self, triggerSource, triggerPattern, triggerWhen):
        return DAQmxCfgDigPatternStartTrig(self.taskHandle, triggerSource, triggerPattern, triggerWhen)

    def DisableRefTrig(self):
        return DAQmxDisableRefTrig(self.taskHandle)

    def CfgDigEdgeRefTrig(self, triggerSource, triggerEdge, pretriggerSamples):
        return DAQmxCfgDigEdgeRefTrig(self.taskHandle, triggerSource, triggerEdge, pretriggerSamples)

    def CfgAnlgEdgeRefTrig(self, triggerSource, triggerSlope, triggerLevel, pretriggerSamples):
        return DAQmxCfgAnlgEdgeRefTrig(self.taskHandle, triggerSource, triggerSlope, triggerLevel, pretriggerSamples)

    def CfgAnlgWindowRefTrig(self, triggerSource, triggerWhen, windowTop, windowBottom, pretriggerSamples):
        return DAQmxCfgAnlgWindowRefTrig(self.taskHandle, triggerSource, triggerWhen, windowTop, windowBottom,
                                         pretriggerSamples)

    def CfgDigPatternRefTrig(self, triggerSource, triggerPattern, triggerWhen, pretriggerSamples):
        return DAQmxCfgDigPatternRefTrig(self.taskHandle, triggerSource, triggerPattern, triggerWhen, pretriggerSamples)

    def DisableAdvTrig(self):
        return DAQmxDisableAdvTrig(self.taskHandle)

    def CfgDigEdgeAdvTrig(self, triggerSource, triggerEdge):
        return DAQmxCfgDigEdgeAdvTrig(self.taskHandle, triggerSource, triggerEdge)

    def GetTrigAttribute(self, attribute, value):
        return DAQmxGetTrigAttribute(self.taskHandle, attribute, value)

    def SetTrigAttribute(self, attribute):
        return DAQmxSetTrigAttribute(self.taskHandle, attribute)

    def ResetTrigAttribute(self, attribute):
        return DAQmxResetTrigAttribute(self.taskHandle, attribute)

    def SendSoftwareTrigger(self, triggerID):
        return DAQmxSendSoftwareTrigger(self.taskHandle, triggerID)

    def ReadAnalogF64(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadAnalogF64(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadAnalogScalarF64(self, timeout, value, reserved):
        return DAQmxReadAnalogScalarF64(self.taskHandle, timeout, value, reserved)

    def ReadBinaryI16(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadBinaryI16(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadBinaryU16(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadBinaryU16(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadBinaryI32(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadBinaryI32(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadBinaryU32(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadBinaryU32(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadDigitalU8(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                      reserved):
        return DAQmxReadDigitalU8(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                  sampsPerChanRead, reserved)

    def ReadDigitalU16(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
        return DAQmxReadDigitalU16(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                   sampsPerChanRead, reserved)

    def ReadDigitalU32(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps, sampsPerChanRead,
                       reserved):
        return DAQmxReadDigitalU32(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInSamps,
                                   sampsPerChanRead, reserved)

    def ReadDigitalScalarU32(self, timeout, value, reserved):
        return DAQmxReadDigitalScalarU32(self.taskHandle, timeout, value, reserved)

    def ReadDigitalLines(self, numSampsPerChan, timeout, fillMode, readArray, arraySizeInBytes, sampsPerChanRead,
                         numBytesPerSamp, reserved):
        return DAQmxReadDigitalLines(self.taskHandle, numSampsPerChan, timeout, fillMode, readArray, arraySizeInBytes,
                                     sampsPerChanRead, numBytesPerSamp, reserved)

    def ReadCounterF64(self, numSampsPerChan, timeout, readArray, arraySizeInSamps, sampsPerChanRead, reserved):
        return DAQmxReadCounterF64(self.taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps,
                                   sampsPerChanRead, reserved)

    def ReadCounterU32(self, numSampsPerChan, timeout, readArray, arraySizeInSamps, sampsPerChanRead, reserved):
        return DAQmxReadCounterU32(self.taskHandle, numSampsPerChan, timeout, readArray, arraySizeInSamps,
                                   sampsPerChanRead, reserved)

    def ReadCounterScalarF64(self, timeout, value, reserved):
        return DAQmxReadCounterScalarF64(self.taskHandle, timeout, value, reserved)

    def ReadCounterScalarU32(self, timeout, value, reserved):
        return DAQmxReadCounterScalarU32(self.taskHandle, timeout, value, reserved)

    def ReadCtrFreq(self, numSampsPerChan, timeout, interleaved, readArrayFrequency, readArrayDutyCycle,
                    arraySizeInSamps, sampsPerChanRead, reserved):
        return DAQmxReadCtrFreq(self.taskHandle, numSampsPerChan, timeout, interleaved, readArrayFrequency,
                                readArrayDutyCycle, arraySizeInSamps, sampsPerChanRead, reserved)

    def ReadCtrTime(self, numSampsPerChan, timeout, interleaved, readArrayHighTime, readArrayLowTime, arraySizeInSamps,
                    sampsPerChanRead, reserved):
        return DAQmxReadCtrTime(self.taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTime,
                                readArrayLowTime, arraySizeInSamps, sampsPerChanRead, reserved)

    def ReadCtrTicks(self, numSampsPerChan, timeout, interleaved, readArrayHighTicks, readArrayLowTicks,
                     arraySizeInSamps, sampsPerChanRead, reserved):
        return DAQmxReadCtrTicks(self.taskHandle, numSampsPerChan, timeout, interleaved, readArrayHighTicks,
                                 readArrayLowTicks, arraySizeInSamps, sampsPerChanRead, reserved)

    def ReadCtrFreqScalar(self, timeout, frequency, dutyCycle, reserved):
        return DAQmxReadCtrFreqScalar(self.taskHandle, timeout, frequency, dutyCycle, reserved)

    def ReadCtrTimeScalar(self, timeout, highTime, lowTime, reserved):
        return DAQmxReadCtrTimeScalar(self.taskHandle, timeout, highTime, lowTime, reserved)

    def ReadCtrTicksScalar(self, timeout, highTicks, lowTicks, reserved):
        return DAQmxReadCtrTicksScalar(self.taskHandle, timeout, highTicks, lowTicks, reserved)

    def ReadRaw(self, numSampsPerChan, timeout, readArray, arraySizeInBytes, sampsRead, numBytesPerSamp, reserved):
        return DAQmxReadRaw(self.taskHandle, numSampsPerChan, timeout, readArray, arraySizeInBytes, sampsRead,
                            numBytesPerSamp, reserved)

    def GetNthTaskReadChannel(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskReadChannel(self.taskHandle, index, buffer, bufferSize)

    def GetReadAttribute(self, attribute, value):
        return DAQmxGetReadAttribute(self.taskHandle, attribute, value)

    def SetReadAttribute(self, attribute):
        return DAQmxSetReadAttribute(self.taskHandle, attribute)

    def ResetReadAttribute(self, attribute):
        return DAQmxResetReadAttribute(self.taskHandle, attribute)

    def ConfigureLogging(self, filePath, loggingMode, groupName, operation):
        return DAQmxConfigureLogging(self.taskHandle, filePath, loggingMode, groupName, operation)

    def StartNewFile(self, filePath):
        return DAQmxStartNewFile(self.taskHandle, filePath)

    def WriteAnalogF64(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteAnalogF64(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteAnalogScalarF64(self, autoStart, timeout, value, reserved):
        return DAQmxWriteAnalogScalarF64(self.taskHandle, autoStart, timeout, value, reserved)

    def WriteBinaryI16(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteBinaryI16(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteBinaryU16(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteBinaryU16(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteBinaryI32(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteBinaryI32(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteBinaryU32(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteBinaryU32(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteDigitalU8(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                       reserved):
        return DAQmxWriteDigitalU8(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                   sampsPerChanWritten, reserved)

    def WriteDigitalU16(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
        return DAQmxWriteDigitalU16(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                    sampsPerChanWritten, reserved)

    def WriteDigitalU32(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                        reserved):
        return DAQmxWriteDigitalU32(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                    sampsPerChanWritten, reserved)

    def WriteDigitalScalarU32(self, autoStart, timeout, value, reserved):
        return DAQmxWriteDigitalScalarU32(self.taskHandle, autoStart, timeout, value, reserved)

    def WriteDigitalLines(self, numSampsPerChan, autoStart, timeout, dataLayout, writeArray, sampsPerChanWritten,
                          reserved):
        return DAQmxWriteDigitalLines(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, writeArray,
                                      sampsPerChanWritten, reserved)

    def WriteCtrFreq(self, numSampsPerChan, autoStart, timeout, dataLayout, frequency, dutyCycle,
                     numSampsPerChanWritten, reserved):
        return DAQmxWriteCtrFreq(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, frequency, dutyCycle,
                                 numSampsPerChanWritten, reserved)

    def WriteCtrFreqScalar(self, autoStart, timeout, frequency, dutyCycle, reserved):
        return DAQmxWriteCtrFreqScalar(self.taskHandle, autoStart, timeout, frequency, dutyCycle, reserved)

    def WriteCtrTime(self, numSampsPerChan, autoStart, timeout, dataLayout, highTime, lowTime, numSampsPerChanWritten,
                     reserved):
        return DAQmxWriteCtrTime(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTime, lowTime,
                                 numSampsPerChanWritten, reserved)

    def WriteCtrTimeScalar(self, autoStart, timeout, highTime, lowTime, reserved):
        return DAQmxWriteCtrTimeScalar(self.taskHandle, autoStart, timeout, highTime, lowTime, reserved)

    def WriteCtrTicks(self, numSampsPerChan, autoStart, timeout, dataLayout, highTicks, lowTicks,
                      numSampsPerChanWritten, reserved):
        return DAQmxWriteCtrTicks(self.taskHandle, numSampsPerChan, autoStart, timeout, dataLayout, highTicks, lowTicks,
                                  numSampsPerChanWritten, reserved)

    def WriteCtrTicksScalar(self, autoStart, timeout, highTicks, lowTicks, reserved):
        return DAQmxWriteCtrTicksScalar(self.taskHandle, autoStart, timeout, highTicks, lowTicks, reserved)

    def WriteRaw(self, numSamps, autoStart, timeout, writeArray, sampsPerChanWritten, reserved):
        return DAQmxWriteRaw(self.taskHandle, numSamps, autoStart, timeout, writeArray, sampsPerChanWritten, reserved)

    def GetWriteAttribute(self, attribute, value):
        return DAQmxGetWriteAttribute(self.taskHandle, attribute, value)

    def SetWriteAttribute(self, attribute):
        return DAQmxSetWriteAttribute(self.taskHandle, attribute)

    def ResetWriteAttribute(self, attribute):
        return DAQmxResetWriteAttribute(self.taskHandle, attribute)

    def ExportSignal(self, signalID, outputTerminal):
        return DAQmxExportSignal(self.taskHandle, signalID, outputTerminal)

    def GetExportedSignalAttribute(self, attribute, value):
        return DAQmxGetExportedSignalAttribute(self.taskHandle, attribute, value)

    def SetExportedSignalAttribute(self, attribute):
        return DAQmxSetExportedSignalAttribute(self.taskHandle, attribute)

    def ResetExportedSignalAttribute(self, attribute):
        return DAQmxResetExportedSignalAttribute(self.taskHandle, attribute)

    def CfgInputBuffer(self, numSampsPerChan):
        return DAQmxCfgInputBuffer(self.taskHandle, numSampsPerChan)

    def CfgOutputBuffer(self, numSampsPerChan):
        return DAQmxCfgOutputBuffer(self.taskHandle, numSampsPerChan)

    def GetBufferAttribute(self, attribute, value):
        return DAQmxGetBufferAttribute(self.taskHandle, attribute, value)

    def SetBufferAttribute(self, attribute):
        return DAQmxSetBufferAttribute(self.taskHandle, attribute)

    def ResetBufferAttribute(self, attribute):
        return DAQmxResetBufferAttribute(self.taskHandle, attribute)

    def GetSwitchScanAttribute(self, attribute, value):
        return DAQmxGetSwitchScanAttribute(self.taskHandle, attribute, value)

    def SetSwitchScanAttribute(self, attribute):
        return DAQmxSetSwitchScanAttribute(self.taskHandle, attribute)

    def ResetSwitchScanAttribute(self, attribute):
        return DAQmxResetSwitchScanAttribute(self.taskHandle, attribute)

    def ControlWatchdogTask(self, action):
        return DAQmxControlWatchdogTask(self.taskHandle, action)

    def CfgWatchdogAOExpirStates(self, channelNames, expirStateArray, outputTypeArray, arraySize):
        return DAQmxCfgWatchdogAOExpirStates(self.taskHandle, channelNames, expirStateArray, outputTypeArray, arraySize)

    def CfgWatchdogCOExpirStates(self, channelNames, expirStateArray, arraySize):
        return DAQmxCfgWatchdogCOExpirStates(self.taskHandle, channelNames, expirStateArray, arraySize)

    def CfgWatchdogDOExpirStates(self, channelNames, expirStateArray, arraySize):
        return DAQmxCfgWatchdogDOExpirStates(self.taskHandle, channelNames, expirStateArray, arraySize)

    def GetWatchdogAttribute(self, lines, attribute, value):
        return DAQmxGetWatchdogAttribute(self.taskHandle, lines, attribute, value)

    def SetWatchdogAttribute(self, lines, attribute):
        return DAQmxSetWatchdogAttribute(self.taskHandle, lines, attribute)

    def ResetWatchdogAttribute(self, lines, attribute):
        return DAQmxResetWatchdogAttribute(self.taskHandle, lines, attribute)

    def PerformBridgeOffsetNullingCal(self, channel):
        return DAQmxPerformBridgeOffsetNullingCal(self.taskHandle, channel)

    def PerformBridgeOffsetNullingCalEx(self, channel, skipUnsupportedChannels):
        return DAQmxPerformBridgeOffsetNullingCalEx(self.taskHandle, channel, skipUnsupportedChannels)

    def PerformThrmcplLeadOffsetNullingCal(self, channel, skipUnsupportedChannels):
        return DAQmxPerformThrmcplLeadOffsetNullingCal(self.taskHandle, channel, skipUnsupportedChannels)

    def PerformStrainShuntCal(self, channel, shuntResistorValue, shuntResistorLocation, skipUnsupportedChannels):
        return DAQmxPerformStrainShuntCal(self.taskHandle, channel, shuntResistorValue, shuntResistorLocation,
                                          skipUnsupportedChannels)

    def PerformBridgeShuntCal(self, channel, shuntResistorValue, shuntResistorLocation, bridgeResistance,
                              skipUnsupportedChannels):
        return DAQmxPerformBridgeShuntCal(self.taskHandle, channel, shuntResistorValue, shuntResistorLocation,
                                          bridgeResistance, skipUnsupportedChannels)

    def WaitForNextSampleClock(self, timeout, isLate):
        return DAQmxWaitForNextSampleClock(self.taskHandle, timeout, isLate)

    def GetRealTimeAttribute(self, attribute, value):
        return DAQmxGetRealTimeAttribute(self.taskHandle, attribute, value)

    def SetRealTimeAttribute(self, attribute):
        return DAQmxSetRealTimeAttribute(self.taskHandle, attribute)

    def ResetRealTimeAttribute(self, attribute):
        return DAQmxResetRealTimeAttribute(self.taskHandle, attribute)

    def SaveTask(self, saveAs, author, options):
        return DAQmxSaveTask(self.taskHandle, saveAs, author, options)

    def SaveGlobalChan(self, channelName, saveAs, author, options):
        return DAQmxSaveGlobalChan(self.taskHandle, channelName, saveAs, author, options)

    def GetBufInputBufSize(self, data):
        return DAQmxGetBufInputBufSize(self.taskHandle, data)

    def SetBufInputBufSize(self, data):
        return DAQmxSetBufInputBufSize(self.taskHandle, data)

    def ResetBufInputBufSize(self):
        return DAQmxResetBufInputBufSize(self.taskHandle)

    def GetBufInputOnbrdBufSize(self, data):
        return DAQmxGetBufInputOnbrdBufSize(self.taskHandle, data)

    def GetBufOutputBufSize(self, data):
        return DAQmxGetBufOutputBufSize(self.taskHandle, data)

    def SetBufOutputBufSize(self, data):
        return DAQmxSetBufOutputBufSize(self.taskHandle, data)

    def ResetBufOutputBufSize(self):
        return DAQmxResetBufOutputBufSize(self.taskHandle)

    def GetBufOutputOnbrdBufSize(self, data):
        return DAQmxGetBufOutputOnbrdBufSize(self.taskHandle, data)

    def SetBufOutputOnbrdBufSize(self, data):
        return DAQmxSetBufOutputOnbrdBufSize(self.taskHandle, data)

    def ResetBufOutputOnbrdBufSize(self):
        return DAQmxResetBufOutputOnbrdBufSize(self.taskHandle)

    def GetAIMax(self, channel, data):
        return DAQmxGetAIMax(self.taskHandle, channel, data)

    def SetAIMax(self, channel, data):
        return DAQmxSetAIMax(self.taskHandle, channel, data)

    def ResetAIMax(self, channel):
        return DAQmxResetAIMax(self.taskHandle, channel)

    def GetAIMin(self, channel, data):
        return DAQmxGetAIMin(self.taskHandle, channel, data)

    def SetAIMin(self, channel, data):
        return DAQmxSetAIMin(self.taskHandle, channel, data)

    def ResetAIMin(self, channel):
        return DAQmxResetAIMin(self.taskHandle, channel)

    def GetAICustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetAICustomScaleName(self.taskHandle, channel, data, bufferSize)

    def SetAICustomScaleName(self, channel, data):
        return DAQmxSetAICustomScaleName(self.taskHandle, channel, data)

    def ResetAICustomScaleName(self, channel):
        return DAQmxResetAICustomScaleName(self.taskHandle, channel)

    def GetAIMeasType(self, channel, data):
        return DAQmxGetAIMeasType(self.taskHandle, channel, data)

    def GetAIVoltageUnits(self, channel, data):
        return DAQmxGetAIVoltageUnits(self.taskHandle, channel, data)

    def SetAIVoltageUnits(self, channel, data):
        return DAQmxSetAIVoltageUnits(self.taskHandle, channel, data)

    def ResetAIVoltageUnits(self, channel):
        return DAQmxResetAIVoltageUnits(self.taskHandle, channel)

    def GetAIVoltagedBRef(self, channel, data):
        return DAQmxGetAIVoltagedBRef(self.taskHandle, channel, data)

    def SetAIVoltagedBRef(self, channel, data):
        return DAQmxSetAIVoltagedBRef(self.taskHandle, channel, data)

    def ResetAIVoltagedBRef(self, channel):
        return DAQmxResetAIVoltagedBRef(self.taskHandle, channel)

    def GetAIVoltageACRMSUnits(self, channel, data):
        return DAQmxGetAIVoltageACRMSUnits(self.taskHandle, channel, data)

    def SetAIVoltageACRMSUnits(self, channel, data):
        return DAQmxSetAIVoltageACRMSUnits(self.taskHandle, channel, data)

    def ResetAIVoltageACRMSUnits(self, channel):
        return DAQmxResetAIVoltageACRMSUnits(self.taskHandle, channel)

    def GetAITempUnits(self, channel, data):
        return DAQmxGetAITempUnits(self.taskHandle, channel, data)

    def SetAITempUnits(self, channel, data):
        return DAQmxSetAITempUnits(self.taskHandle, channel, data)

    def ResetAITempUnits(self, channel):
        return DAQmxResetAITempUnits(self.taskHandle, channel)

    def GetAIThrmcplType(self, channel, data):
        return DAQmxGetAIThrmcplType(self.taskHandle, channel, data)

    def SetAIThrmcplType(self, channel, data):
        return DAQmxSetAIThrmcplType(self.taskHandle, channel, data)

    def ResetAIThrmcplType(self, channel):
        return DAQmxResetAIThrmcplType(self.taskHandle, channel)

    def GetAIThrmcplScaleType(self, channel, data):
        return DAQmxGetAIThrmcplScaleType(self.taskHandle, channel, data)

    def SetAIThrmcplScaleType(self, channel, data):
        return DAQmxSetAIThrmcplScaleType(self.taskHandle, channel, data)

    def ResetAIThrmcplScaleType(self, channel):
        return DAQmxResetAIThrmcplScaleType(self.taskHandle, channel)

    def GetAIThrmcplCJCSrc(self, channel, data):
        return DAQmxGetAIThrmcplCJCSrc(self.taskHandle, channel, data)

    def GetAIThrmcplCJCVal(self, channel, data):
        return DAQmxGetAIThrmcplCJCVal(self.taskHandle, channel, data)

    def SetAIThrmcplCJCVal(self, channel, data):
        return DAQmxSetAIThrmcplCJCVal(self.taskHandle, channel, data)

    def ResetAIThrmcplCJCVal(self, channel):
        return DAQmxResetAIThrmcplCJCVal(self.taskHandle, channel)

    def GetAIThrmcplCJCChan(self, channel, data, bufferSize):
        return DAQmxGetAIThrmcplCJCChan(self.taskHandle, channel, data, bufferSize)

    def GetAIRTDType(self, channel, data):
        return DAQmxGetAIRTDType(self.taskHandle, channel, data)

    def SetAIRTDType(self, channel, data):
        return DAQmxSetAIRTDType(self.taskHandle, channel, data)

    def ResetAIRTDType(self, channel):
        return DAQmxResetAIRTDType(self.taskHandle, channel)

    def GetAIRTDR0(self, channel, data):
        return DAQmxGetAIRTDR0(self.taskHandle, channel, data)

    def SetAIRTDR0(self, channel, data):
        return DAQmxSetAIRTDR0(self.taskHandle, channel, data)

    def ResetAIRTDR0(self, channel):
        return DAQmxResetAIRTDR0(self.taskHandle, channel)

    def GetAIRTDA(self, channel, data):
        return DAQmxGetAIRTDA(self.taskHandle, channel, data)

    def SetAIRTDA(self, channel, data):
        return DAQmxSetAIRTDA(self.taskHandle, channel, data)

    def ResetAIRTDA(self, channel):
        return DAQmxResetAIRTDA(self.taskHandle, channel)

    def GetAIRTDB(self, channel, data):
        return DAQmxGetAIRTDB(self.taskHandle, channel, data)

    def SetAIRTDB(self, channel, data):
        return DAQmxSetAIRTDB(self.taskHandle, channel, data)

    def ResetAIRTDB(self, channel):
        return DAQmxResetAIRTDB(self.taskHandle, channel)

    def GetAIRTDC(self, channel, data):
        return DAQmxGetAIRTDC(self.taskHandle, channel, data)

    def SetAIRTDC(self, channel, data):
        return DAQmxSetAIRTDC(self.taskHandle, channel, data)

    def ResetAIRTDC(self, channel):
        return DAQmxResetAIRTDC(self.taskHandle, channel)

    def GetAIThrmstrA(self, channel, data):
        return DAQmxGetAIThrmstrA(self.taskHandle, channel, data)

    def SetAIThrmstrA(self, channel, data):
        return DAQmxSetAIThrmstrA(self.taskHandle, channel, data)

    def ResetAIThrmstrA(self, channel):
        return DAQmxResetAIThrmstrA(self.taskHandle, channel)

    def GetAIThrmstrB(self, channel, data):
        return DAQmxGetAIThrmstrB(self.taskHandle, channel, data)

    def SetAIThrmstrB(self, channel, data):
        return DAQmxSetAIThrmstrB(self.taskHandle, channel, data)

    def ResetAIThrmstrB(self, channel):
        return DAQmxResetAIThrmstrB(self.taskHandle, channel)

    def GetAIThrmstrC(self, channel, data):
        return DAQmxGetAIThrmstrC(self.taskHandle, channel, data)

    def SetAIThrmstrC(self, channel, data):
        return DAQmxSetAIThrmstrC(self.taskHandle, channel, data)

    def ResetAIThrmstrC(self, channel):
        return DAQmxResetAIThrmstrC(self.taskHandle, channel)

    def GetAIThrmstrR1(self, channel, data):
        return DAQmxGetAIThrmstrR1(self.taskHandle, channel, data)

    def SetAIThrmstrR1(self, channel, data):
        return DAQmxSetAIThrmstrR1(self.taskHandle, channel, data)

    def ResetAIThrmstrR1(self, channel):
        return DAQmxResetAIThrmstrR1(self.taskHandle, channel)

    def GetAIForceReadFromChan(self, channel, data):
        return DAQmxGetAIForceReadFromChan(self.taskHandle, channel, data)

    def SetAIForceReadFromChan(self, channel, data):
        return DAQmxSetAIForceReadFromChan(self.taskHandle, channel, data)

    def ResetAIForceReadFromChan(self, channel):
        return DAQmxResetAIForceReadFromChan(self.taskHandle, channel)

    def GetAICurrentUnits(self, channel, data):
        return DAQmxGetAICurrentUnits(self.taskHandle, channel, data)

    def SetAICurrentUnits(self, channel, data):
        return DAQmxSetAICurrentUnits(self.taskHandle, channel, data)

    def ResetAICurrentUnits(self, channel):
        return DAQmxResetAICurrentUnits(self.taskHandle, channel)

    def GetAICurrentACRMSUnits(self, channel, data):
        return DAQmxGetAICurrentACRMSUnits(self.taskHandle, channel, data)

    def SetAICurrentACRMSUnits(self, channel, data):
        return DAQmxSetAICurrentACRMSUnits(self.taskHandle, channel, data)

    def ResetAICurrentACRMSUnits(self, channel):
        return DAQmxResetAICurrentACRMSUnits(self.taskHandle, channel)

    def GetAIStrainUnits(self, channel, data):
        return DAQmxGetAIStrainUnits(self.taskHandle, channel, data)

    def SetAIStrainUnits(self, channel, data):
        return DAQmxSetAIStrainUnits(self.taskHandle, channel, data)

    def ResetAIStrainUnits(self, channel):
        return DAQmxResetAIStrainUnits(self.taskHandle, channel)

    def GetAIStrainGageForceReadFromChan(self, channel, data):
        return DAQmxGetAIStrainGageForceReadFromChan(self.taskHandle, channel, data)

    def SetAIStrainGageForceReadFromChan(self, channel, data):
        return DAQmxSetAIStrainGageForceReadFromChan(self.taskHandle, channel, data)

    def ResetAIStrainGageForceReadFromChan(self, channel):
        return DAQmxResetAIStrainGageForceReadFromChan(self.taskHandle, channel)

    def GetAIStrainGageGageFactor(self, channel, data):
        return DAQmxGetAIStrainGageGageFactor(self.taskHandle, channel, data)

    def SetAIStrainGageGageFactor(self, channel, data):
        return DAQmxSetAIStrainGageGageFactor(self.taskHandle, channel, data)

    def ResetAIStrainGageGageFactor(self, channel):
        return DAQmxResetAIStrainGageGageFactor(self.taskHandle, channel)

    def GetAIStrainGagePoissonRatio(self, channel, data):
        return DAQmxGetAIStrainGagePoissonRatio(self.taskHandle, channel, data)

    def SetAIStrainGagePoissonRatio(self, channel, data):
        return DAQmxSetAIStrainGagePoissonRatio(self.taskHandle, channel, data)

    def ResetAIStrainGagePoissonRatio(self, channel):
        return DAQmxResetAIStrainGagePoissonRatio(self.taskHandle, channel)

    def GetAIStrainGageCfg(self, channel, data):
        return DAQmxGetAIStrainGageCfg(self.taskHandle, channel, data)

    def SetAIStrainGageCfg(self, channel, data):
        return DAQmxSetAIStrainGageCfg(self.taskHandle, channel, data)

    def ResetAIStrainGageCfg(self, channel):
        return DAQmxResetAIStrainGageCfg(self.taskHandle, channel)

    def GetAIRosetteStrainGageRosetteType(self, channel, data):
        return DAQmxGetAIRosetteStrainGageRosetteType(self.taskHandle, channel, data)

    def GetAIRosetteStrainGageOrientation(self, channel, data):
        return DAQmxGetAIRosetteStrainGageOrientation(self.taskHandle, channel, data)

    def SetAIRosetteStrainGageOrientation(self, channel, data):
        return DAQmxSetAIRosetteStrainGageOrientation(self.taskHandle, channel, data)

    def ResetAIRosetteStrainGageOrientation(self, channel):
        return DAQmxResetAIRosetteStrainGageOrientation(self.taskHandle, channel)

    def GetAIRosetteStrainGageStrainChans(self, channel, data, bufferSize):
        return DAQmxGetAIRosetteStrainGageStrainChans(self.taskHandle, channel, data, bufferSize)

    def GetAIRosetteStrainGageRosetteMeasType(self, channel, data):
        return DAQmxGetAIRosetteStrainGageRosetteMeasType(self.taskHandle, channel, data)

    def SetAIRosetteStrainGageRosetteMeasType(self, channel, data):
        return DAQmxSetAIRosetteStrainGageRosetteMeasType(self.taskHandle, channel, data)

    def ResetAIRosetteStrainGageRosetteMeasType(self, channel):
        return DAQmxResetAIRosetteStrainGageRosetteMeasType(self.taskHandle, channel)

    def GetAIResistanceUnits(self, channel, data):
        return DAQmxGetAIResistanceUnits(self.taskHandle, channel, data)

    def SetAIResistanceUnits(self, channel, data):
        return DAQmxSetAIResistanceUnits(self.taskHandle, channel, data)

    def ResetAIResistanceUnits(self, channel):
        return DAQmxResetAIResistanceUnits(self.taskHandle, channel)

    def GetAIFreqUnits(self, channel, data):
        return DAQmxGetAIFreqUnits(self.taskHandle, channel, data)

    def SetAIFreqUnits(self, channel, data):
        return DAQmxSetAIFreqUnits(self.taskHandle, channel, data)

    def ResetAIFreqUnits(self, channel):
        return DAQmxResetAIFreqUnits(self.taskHandle, channel)

    def GetAIFreqThreshVoltage(self, channel, data):
        return DAQmxGetAIFreqThreshVoltage(self.taskHandle, channel, data)

    def SetAIFreqThreshVoltage(self, channel, data):
        return DAQmxSetAIFreqThreshVoltage(self.taskHandle, channel, data)

    def ResetAIFreqThreshVoltage(self, channel):
        return DAQmxResetAIFreqThreshVoltage(self.taskHandle, channel)

    def GetAIFreqHyst(self, channel, data):
        return DAQmxGetAIFreqHyst(self.taskHandle, channel, data)

    def SetAIFreqHyst(self, channel, data):
        return DAQmxSetAIFreqHyst(self.taskHandle, channel, data)

    def ResetAIFreqHyst(self, channel):
        return DAQmxResetAIFreqHyst(self.taskHandle, channel)

    def GetAILVDTUnits(self, channel, data):
        return DAQmxGetAILVDTUnits(self.taskHandle, channel, data)

    def SetAILVDTUnits(self, channel, data):
        return DAQmxSetAILVDTUnits(self.taskHandle, channel, data)

    def ResetAILVDTUnits(self, channel):
        return DAQmxResetAILVDTUnits(self.taskHandle, channel)

    def GetAILVDTSensitivity(self, channel, data):
        return DAQmxGetAILVDTSensitivity(self.taskHandle, channel, data)

    def SetAILVDTSensitivity(self, channel, data):
        return DAQmxSetAILVDTSensitivity(self.taskHandle, channel, data)

    def ResetAILVDTSensitivity(self, channel):
        return DAQmxResetAILVDTSensitivity(self.taskHandle, channel)

    def GetAILVDTSensitivityUnits(self, channel, data):
        return DAQmxGetAILVDTSensitivityUnits(self.taskHandle, channel, data)

    def SetAILVDTSensitivityUnits(self, channel, data):
        return DAQmxSetAILVDTSensitivityUnits(self.taskHandle, channel, data)

    def ResetAILVDTSensitivityUnits(self, channel):
        return DAQmxResetAILVDTSensitivityUnits(self.taskHandle, channel)

    def GetAIRVDTUnits(self, channel, data):
        return DAQmxGetAIRVDTUnits(self.taskHandle, channel, data)

    def SetAIRVDTUnits(self, channel, data):
        return DAQmxSetAIRVDTUnits(self.taskHandle, channel, data)

    def ResetAIRVDTUnits(self, channel):
        return DAQmxResetAIRVDTUnits(self.taskHandle, channel)

    def GetAIRVDTSensitivity(self, channel, data):
        return DAQmxGetAIRVDTSensitivity(self.taskHandle, channel, data)

    def SetAIRVDTSensitivity(self, channel, data):
        return DAQmxSetAIRVDTSensitivity(self.taskHandle, channel, data)

    def ResetAIRVDTSensitivity(self, channel):
        return DAQmxResetAIRVDTSensitivity(self.taskHandle, channel)

    def GetAIRVDTSensitivityUnits(self, channel, data):
        return DAQmxGetAIRVDTSensitivityUnits(self.taskHandle, channel, data)

    def SetAIRVDTSensitivityUnits(self, channel, data):
        return DAQmxSetAIRVDTSensitivityUnits(self.taskHandle, channel, data)

    def ResetAIRVDTSensitivityUnits(self, channel):
        return DAQmxResetAIRVDTSensitivityUnits(self.taskHandle, channel)

    def GetAIEddyCurrentProxProbeUnits(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeUnits(self.taskHandle, channel, data)

    def SetAIEddyCurrentProxProbeUnits(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeUnits(self.taskHandle, channel, data)

    def ResetAIEddyCurrentProxProbeUnits(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeUnits(self.taskHandle, channel)

    def GetAIEddyCurrentProxProbeSensitivity(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeSensitivity(self.taskHandle, channel, data)

    def SetAIEddyCurrentProxProbeSensitivity(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeSensitivity(self.taskHandle, channel, data)

    def ResetAIEddyCurrentProxProbeSensitivity(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeSensitivity(self.taskHandle, channel)

    def GetAIEddyCurrentProxProbeSensitivityUnits(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeSensitivityUnits(self.taskHandle, channel, data)

    def SetAIEddyCurrentProxProbeSensitivityUnits(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeSensitivityUnits(self.taskHandle, channel, data)

    def ResetAIEddyCurrentProxProbeSensitivityUnits(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeSensitivityUnits(self.taskHandle, channel)

    def GetAISoundPressureMaxSoundPressureLvl(self, channel, data):
        return DAQmxGetAISoundPressureMaxSoundPressureLvl(self.taskHandle, channel, data)

    def SetAISoundPressureMaxSoundPressureLvl(self, channel, data):
        return DAQmxSetAISoundPressureMaxSoundPressureLvl(self.taskHandle, channel, data)

    def ResetAISoundPressureMaxSoundPressureLvl(self, channel):
        return DAQmxResetAISoundPressureMaxSoundPressureLvl(self.taskHandle, channel)

    def GetAISoundPressureUnits(self, channel, data):
        return DAQmxGetAISoundPressureUnits(self.taskHandle, channel, data)

    def SetAISoundPressureUnits(self, channel, data):
        return DAQmxSetAISoundPressureUnits(self.taskHandle, channel, data)

    def ResetAISoundPressureUnits(self, channel):
        return DAQmxResetAISoundPressureUnits(self.taskHandle, channel)

    def GetAISoundPressuredBRef(self, channel, data):
        return DAQmxGetAISoundPressuredBRef(self.taskHandle, channel, data)

    def SetAISoundPressuredBRef(self, channel, data):
        return DAQmxSetAISoundPressuredBRef(self.taskHandle, channel, data)

    def ResetAISoundPressuredBRef(self, channel):
        return DAQmxResetAISoundPressuredBRef(self.taskHandle, channel)

    def GetAIMicrophoneSensitivity(self, channel, data):
        return DAQmxGetAIMicrophoneSensitivity(self.taskHandle, channel, data)

    def SetAIMicrophoneSensitivity(self, channel, data):
        return DAQmxSetAIMicrophoneSensitivity(self.taskHandle, channel, data)

    def ResetAIMicrophoneSensitivity(self, channel):
        return DAQmxResetAIMicrophoneSensitivity(self.taskHandle, channel)

    def GetAIAccelUnits(self, channel, data):
        return DAQmxGetAIAccelUnits(self.taskHandle, channel, data)

    def SetAIAccelUnits(self, channel, data):
        return DAQmxSetAIAccelUnits(self.taskHandle, channel, data)

    def ResetAIAccelUnits(self, channel):
        return DAQmxResetAIAccelUnits(self.taskHandle, channel)

    def GetAIAcceldBRef(self, channel, data):
        return DAQmxGetAIAcceldBRef(self.taskHandle, channel, data)

    def SetAIAcceldBRef(self, channel, data):
        return DAQmxSetAIAcceldBRef(self.taskHandle, channel, data)

    def ResetAIAcceldBRef(self, channel):
        return DAQmxResetAIAcceldBRef(self.taskHandle, channel)

    def GetAIAccelSensitivity(self, channel, data):
        return DAQmxGetAIAccelSensitivity(self.taskHandle, channel, data)

    def SetAIAccelSensitivity(self, channel, data):
        return DAQmxSetAIAccelSensitivity(self.taskHandle, channel, data)

    def ResetAIAccelSensitivity(self, channel):
        return DAQmxResetAIAccelSensitivity(self.taskHandle, channel)

    def GetAIAccelSensitivityUnits(self, channel, data):
        return DAQmxGetAIAccelSensitivityUnits(self.taskHandle, channel, data)

    def SetAIAccelSensitivityUnits(self, channel, data):
        return DAQmxSetAIAccelSensitivityUnits(self.taskHandle, channel, data)

    def ResetAIAccelSensitivityUnits(self, channel):
        return DAQmxResetAIAccelSensitivityUnits(self.taskHandle, channel)

    def GetAIVelocityUnits(self, channel, data):
        return DAQmxGetAIVelocityUnits(self.taskHandle, channel, data)

    def SetAIVelocityUnits(self, channel, data):
        return DAQmxSetAIVelocityUnits(self.taskHandle, channel, data)

    def ResetAIVelocityUnits(self, channel):
        return DAQmxResetAIVelocityUnits(self.taskHandle, channel)

    def GetAIVelocityIEPESensordBRef(self, channel, data):
        return DAQmxGetAIVelocityIEPESensordBRef(self.taskHandle, channel, data)

    def SetAIVelocityIEPESensordBRef(self, channel, data):
        return DAQmxSetAIVelocityIEPESensordBRef(self.taskHandle, channel, data)

    def ResetAIVelocityIEPESensordBRef(self, channel):
        return DAQmxResetAIVelocityIEPESensordBRef(self.taskHandle, channel)

    def GetAIVelocityIEPESensorSensitivity(self, channel, data):
        return DAQmxGetAIVelocityIEPESensorSensitivity(self.taskHandle, channel, data)

    def SetAIVelocityIEPESensorSensitivity(self, channel, data):
        return DAQmxSetAIVelocityIEPESensorSensitivity(self.taskHandle, channel, data)

    def ResetAIVelocityIEPESensorSensitivity(self, channel):
        return DAQmxResetAIVelocityIEPESensorSensitivity(self.taskHandle, channel)

    def GetAIVelocityIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxGetAIVelocityIEPESensorSensitivityUnits(self.taskHandle, channel, data)

    def SetAIVelocityIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxSetAIVelocityIEPESensorSensitivityUnits(self.taskHandle, channel, data)

    def ResetAIVelocityIEPESensorSensitivityUnits(self, channel):
        return DAQmxResetAIVelocityIEPESensorSensitivityUnits(self.taskHandle, channel)

    def GetAIForceUnits(self, channel, data):
        return DAQmxGetAIForceUnits(self.taskHandle, channel, data)

    def SetAIForceUnits(self, channel, data):
        return DAQmxSetAIForceUnits(self.taskHandle, channel, data)

    def ResetAIForceUnits(self, channel):
        return DAQmxResetAIForceUnits(self.taskHandle, channel)

    def GetAIForceIEPESensorSensitivity(self, channel, data):
        return DAQmxGetAIForceIEPESensorSensitivity(self.taskHandle, channel, data)

    def SetAIForceIEPESensorSensitivity(self, channel, data):
        return DAQmxSetAIForceIEPESensorSensitivity(self.taskHandle, channel, data)

    def ResetAIForceIEPESensorSensitivity(self, channel):
        return DAQmxResetAIForceIEPESensorSensitivity(self.taskHandle, channel)

    def GetAIForceIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxGetAIForceIEPESensorSensitivityUnits(self.taskHandle, channel, data)

    def SetAIForceIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxSetAIForceIEPESensorSensitivityUnits(self.taskHandle, channel, data)

    def ResetAIForceIEPESensorSensitivityUnits(self, channel):
        return DAQmxResetAIForceIEPESensorSensitivityUnits(self.taskHandle, channel)

    def GetAIPressureUnits(self, channel, data):
        return DAQmxGetAIPressureUnits(self.taskHandle, channel, data)

    def SetAIPressureUnits(self, channel, data):
        return DAQmxSetAIPressureUnits(self.taskHandle, channel, data)

    def ResetAIPressureUnits(self, channel):
        return DAQmxResetAIPressureUnits(self.taskHandle, channel)

    def GetAITorqueUnits(self, channel, data):
        return DAQmxGetAITorqueUnits(self.taskHandle, channel, data)

    def SetAITorqueUnits(self, channel, data):
        return DAQmxSetAITorqueUnits(self.taskHandle, channel, data)

    def ResetAITorqueUnits(self, channel):
        return DAQmxResetAITorqueUnits(self.taskHandle, channel)

    def GetAIBridgeUnits(self, channel, data):
        return DAQmxGetAIBridgeUnits(self.taskHandle, channel, data)

    def SetAIBridgeUnits(self, channel, data):
        return DAQmxSetAIBridgeUnits(self.taskHandle, channel, data)

    def ResetAIBridgeUnits(self, channel):
        return DAQmxResetAIBridgeUnits(self.taskHandle, channel)

    def GetAIBridgeElectricalUnits(self, channel, data):
        return DAQmxGetAIBridgeElectricalUnits(self.taskHandle, channel, data)

    def SetAIBridgeElectricalUnits(self, channel, data):
        return DAQmxSetAIBridgeElectricalUnits(self.taskHandle, channel, data)

    def ResetAIBridgeElectricalUnits(self, channel):
        return DAQmxResetAIBridgeElectricalUnits(self.taskHandle, channel)

    def GetAIBridgePhysicalUnits(self, channel, data):
        return DAQmxGetAIBridgePhysicalUnits(self.taskHandle, channel, data)

    def SetAIBridgePhysicalUnits(self, channel, data):
        return DAQmxSetAIBridgePhysicalUnits(self.taskHandle, channel, data)

    def ResetAIBridgePhysicalUnits(self, channel):
        return DAQmxResetAIBridgePhysicalUnits(self.taskHandle, channel)

    def GetAIBridgeScaleType(self, channel, data):
        return DAQmxGetAIBridgeScaleType(self.taskHandle, channel, data)

    def SetAIBridgeScaleType(self, channel, data):
        return DAQmxSetAIBridgeScaleType(self.taskHandle, channel, data)

    def ResetAIBridgeScaleType(self, channel):
        return DAQmxResetAIBridgeScaleType(self.taskHandle, channel)

    def GetAIBridgeTwoPointLinFirstElectricalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinFirstElectricalVal(self.taskHandle, channel, data)

    def SetAIBridgeTwoPointLinFirstElectricalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinFirstElectricalVal(self.taskHandle, channel, data)

    def ResetAIBridgeTwoPointLinFirstElectricalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinFirstElectricalVal(self.taskHandle, channel)

    def GetAIBridgeTwoPointLinFirstPhysicalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal(self.taskHandle, channel, data)

    def SetAIBridgeTwoPointLinFirstPhysicalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal(self.taskHandle, channel, data)

    def ResetAIBridgeTwoPointLinFirstPhysicalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal(self.taskHandle, channel)

    def GetAIBridgeTwoPointLinSecondElectricalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinSecondElectricalVal(self.taskHandle, channel, data)

    def SetAIBridgeTwoPointLinSecondElectricalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinSecondElectricalVal(self.taskHandle, channel, data)

    def ResetAIBridgeTwoPointLinSecondElectricalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinSecondElectricalVal(self.taskHandle, channel)

    def GetAIBridgeTwoPointLinSecondPhysicalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal(self.taskHandle, channel, data)

    def SetAIBridgeTwoPointLinSecondPhysicalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal(self.taskHandle, channel, data)

    def ResetAIBridgeTwoPointLinSecondPhysicalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal(self.taskHandle, channel)

    def GetAIBridgeTableElectricalVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgeTableElectricalVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIBridgeTableElectricalVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgeTableElectricalVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIBridgeTableElectricalVals(self, channel):
        return DAQmxResetAIBridgeTableElectricalVals(self.taskHandle, channel)

    def GetAIBridgeTablePhysicalVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgeTablePhysicalVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIBridgeTablePhysicalVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgeTablePhysicalVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIBridgeTablePhysicalVals(self, channel):
        return DAQmxResetAIBridgeTablePhysicalVals(self.taskHandle, channel)

    def GetAIBridgePolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgePolyForwardCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIBridgePolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgePolyForwardCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIBridgePolyForwardCoeff(self, channel):
        return DAQmxResetAIBridgePolyForwardCoeff(self.taskHandle, channel)

    def GetAIBridgePolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgePolyReverseCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIBridgePolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgePolyReverseCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIBridgePolyReverseCoeff(self, channel):
        return DAQmxResetAIBridgePolyReverseCoeff(self.taskHandle, channel)

    def GetAIIsTEDS(self, channel, data):
        return DAQmxGetAIIsTEDS(self.taskHandle, channel, data)

    def GetAITEDSUnits(self, channel, data, bufferSize):
        return DAQmxGetAITEDSUnits(self.taskHandle, channel, data, bufferSize)

    def GetAICoupling(self, channel, data):
        return DAQmxGetAICoupling(self.taskHandle, channel, data)

    def SetAICoupling(self, channel, data):
        return DAQmxSetAICoupling(self.taskHandle, channel, data)

    def ResetAICoupling(self, channel):
        return DAQmxResetAICoupling(self.taskHandle, channel)

    def GetAIImpedance(self, channel, data):
        return DAQmxGetAIImpedance(self.taskHandle, channel, data)

    def SetAIImpedance(self, channel, data):
        return DAQmxSetAIImpedance(self.taskHandle, channel, data)

    def ResetAIImpedance(self, channel):
        return DAQmxResetAIImpedance(self.taskHandle, channel)

    def GetAITermCfg(self, channel, data):
        return DAQmxGetAITermCfg(self.taskHandle, channel, data)

    def SetAITermCfg(self, channel, data):
        return DAQmxSetAITermCfg(self.taskHandle, channel, data)

    def ResetAITermCfg(self, channel):
        return DAQmxResetAITermCfg(self.taskHandle, channel)

    def GetAIInputSrc(self, channel, data, bufferSize):
        return DAQmxGetAIInputSrc(self.taskHandle, channel, data, bufferSize)

    def SetAIInputSrc(self, channel, data):
        return DAQmxSetAIInputSrc(self.taskHandle, channel, data)

    def ResetAIInputSrc(self, channel):
        return DAQmxResetAIInputSrc(self.taskHandle, channel)

    def GetAIResistanceCfg(self, channel, data):
        return DAQmxGetAIResistanceCfg(self.taskHandle, channel, data)

    def SetAIResistanceCfg(self, channel, data):
        return DAQmxSetAIResistanceCfg(self.taskHandle, channel, data)

    def ResetAIResistanceCfg(self, channel):
        return DAQmxResetAIResistanceCfg(self.taskHandle, channel)

    def GetAILeadWireResistance(self, channel, data):
        return DAQmxGetAILeadWireResistance(self.taskHandle, channel, data)

    def SetAILeadWireResistance(self, channel, data):
        return DAQmxSetAILeadWireResistance(self.taskHandle, channel, data)

    def ResetAILeadWireResistance(self, channel):
        return DAQmxResetAILeadWireResistance(self.taskHandle, channel)

    def GetAIBridgeCfg(self, channel, data):
        return DAQmxGetAIBridgeCfg(self.taskHandle, channel, data)

    def SetAIBridgeCfg(self, channel, data):
        return DAQmxSetAIBridgeCfg(self.taskHandle, channel, data)

    def ResetAIBridgeCfg(self, channel):
        return DAQmxResetAIBridgeCfg(self.taskHandle, channel)

    def GetAIBridgeNomResistance(self, channel, data):
        return DAQmxGetAIBridgeNomResistance(self.taskHandle, channel, data)

    def SetAIBridgeNomResistance(self, channel, data):
        return DAQmxSetAIBridgeNomResistance(self.taskHandle, channel, data)

    def ResetAIBridgeNomResistance(self, channel):
        return DAQmxResetAIBridgeNomResistance(self.taskHandle, channel)

    def GetAIBridgeInitialVoltage(self, channel, data):
        return DAQmxGetAIBridgeInitialVoltage(self.taskHandle, channel, data)

    def SetAIBridgeInitialVoltage(self, channel, data):
        return DAQmxSetAIBridgeInitialVoltage(self.taskHandle, channel, data)

    def ResetAIBridgeInitialVoltage(self, channel):
        return DAQmxResetAIBridgeInitialVoltage(self.taskHandle, channel)

    def GetAIBridgeInitialRatio(self, channel, data):
        return DAQmxGetAIBridgeInitialRatio(self.taskHandle, channel, data)

    def SetAIBridgeInitialRatio(self, channel, data):
        return DAQmxSetAIBridgeInitialRatio(self.taskHandle, channel, data)

    def ResetAIBridgeInitialRatio(self, channel):
        return DAQmxResetAIBridgeInitialRatio(self.taskHandle, channel)

    def GetAIBridgeShuntCalEnable(self, channel, data):
        return DAQmxGetAIBridgeShuntCalEnable(self.taskHandle, channel, data)

    def SetAIBridgeShuntCalEnable(self, channel, data):
        return DAQmxSetAIBridgeShuntCalEnable(self.taskHandle, channel, data)

    def ResetAIBridgeShuntCalEnable(self, channel):
        return DAQmxResetAIBridgeShuntCalEnable(self.taskHandle, channel)

    def GetAIBridgeShuntCalSelect(self, channel, data):
        return DAQmxGetAIBridgeShuntCalSelect(self.taskHandle, channel, data)

    def SetAIBridgeShuntCalSelect(self, channel, data):
        return DAQmxSetAIBridgeShuntCalSelect(self.taskHandle, channel, data)

    def ResetAIBridgeShuntCalSelect(self, channel):
        return DAQmxResetAIBridgeShuntCalSelect(self.taskHandle, channel)

    def GetAIBridgeShuntCalGainAdjust(self, channel, data):
        return DAQmxGetAIBridgeShuntCalGainAdjust(self.taskHandle, channel, data)

    def SetAIBridgeShuntCalGainAdjust(self, channel, data):
        return DAQmxSetAIBridgeShuntCalGainAdjust(self.taskHandle, channel, data)

    def ResetAIBridgeShuntCalGainAdjust(self, channel):
        return DAQmxResetAIBridgeShuntCalGainAdjust(self.taskHandle, channel)

    def GetAIBridgeShuntCalShuntCalAResistance(self, channel, data):
        return DAQmxGetAIBridgeShuntCalShuntCalAResistance(self.taskHandle, channel, data)

    def SetAIBridgeShuntCalShuntCalAResistance(self, channel, data):
        return DAQmxSetAIBridgeShuntCalShuntCalAResistance(self.taskHandle, channel, data)

    def ResetAIBridgeShuntCalShuntCalAResistance(self, channel):
        return DAQmxResetAIBridgeShuntCalShuntCalAResistance(self.taskHandle, channel)

    def GetAIBridgeShuntCalShuntCalAActualResistance(self, channel, data):
        return DAQmxGetAIBridgeShuntCalShuntCalAActualResistance(self.taskHandle, channel, data)

    def SetAIBridgeShuntCalShuntCalAActualResistance(self, channel, data):
        return DAQmxSetAIBridgeShuntCalShuntCalAActualResistance(self.taskHandle, channel, data)

    def ResetAIBridgeShuntCalShuntCalAActualResistance(self, channel):
        return DAQmxResetAIBridgeShuntCalShuntCalAActualResistance(self.taskHandle, channel)

    def GetAIBridgeBalanceCoarsePot(self, channel, data):
        return DAQmxGetAIBridgeBalanceCoarsePot(self.taskHandle, channel, data)

    def SetAIBridgeBalanceCoarsePot(self, channel, data):
        return DAQmxSetAIBridgeBalanceCoarsePot(self.taskHandle, channel, data)

    def ResetAIBridgeBalanceCoarsePot(self, channel):
        return DAQmxResetAIBridgeBalanceCoarsePot(self.taskHandle, channel)

    def GetAIBridgeBalanceFinePot(self, channel, data):
        return DAQmxGetAIBridgeBalanceFinePot(self.taskHandle, channel, data)

    def SetAIBridgeBalanceFinePot(self, channel, data):
        return DAQmxSetAIBridgeBalanceFinePot(self.taskHandle, channel, data)

    def ResetAIBridgeBalanceFinePot(self, channel):
        return DAQmxResetAIBridgeBalanceFinePot(self.taskHandle, channel)

    def GetAICurrentShuntLoc(self, channel, data):
        return DAQmxGetAICurrentShuntLoc(self.taskHandle, channel, data)

    def SetAICurrentShuntLoc(self, channel, data):
        return DAQmxSetAICurrentShuntLoc(self.taskHandle, channel, data)

    def ResetAICurrentShuntLoc(self, channel):
        return DAQmxResetAICurrentShuntLoc(self.taskHandle, channel)

    def GetAICurrentShuntResistance(self, channel, data):
        return DAQmxGetAICurrentShuntResistance(self.taskHandle, channel, data)

    def SetAICurrentShuntResistance(self, channel, data):
        return DAQmxSetAICurrentShuntResistance(self.taskHandle, channel, data)

    def ResetAICurrentShuntResistance(self, channel):
        return DAQmxResetAICurrentShuntResistance(self.taskHandle, channel)

    def GetAIExcitSrc(self, channel, data):
        return DAQmxGetAIExcitSrc(self.taskHandle, channel, data)

    def SetAIExcitSrc(self, channel, data):
        return DAQmxSetAIExcitSrc(self.taskHandle, channel, data)

    def ResetAIExcitSrc(self, channel):
        return DAQmxResetAIExcitSrc(self.taskHandle, channel)

    def GetAIExcitVal(self, channel, data):
        return DAQmxGetAIExcitVal(self.taskHandle, channel, data)

    def SetAIExcitVal(self, channel, data):
        return DAQmxSetAIExcitVal(self.taskHandle, channel, data)

    def ResetAIExcitVal(self, channel):
        return DAQmxResetAIExcitVal(self.taskHandle, channel)

    def GetAIExcitUseForScaling(self, channel, data):
        return DAQmxGetAIExcitUseForScaling(self.taskHandle, channel, data)

    def SetAIExcitUseForScaling(self, channel, data):
        return DAQmxSetAIExcitUseForScaling(self.taskHandle, channel, data)

    def ResetAIExcitUseForScaling(self, channel):
        return DAQmxResetAIExcitUseForScaling(self.taskHandle, channel)

    def GetAIExcitUseMultiplexed(self, channel, data):
        return DAQmxGetAIExcitUseMultiplexed(self.taskHandle, channel, data)

    def SetAIExcitUseMultiplexed(self, channel, data):
        return DAQmxSetAIExcitUseMultiplexed(self.taskHandle, channel, data)

    def ResetAIExcitUseMultiplexed(self, channel):
        return DAQmxResetAIExcitUseMultiplexed(self.taskHandle, channel)

    def GetAIExcitActualVal(self, channel, data):
        return DAQmxGetAIExcitActualVal(self.taskHandle, channel, data)

    def SetAIExcitActualVal(self, channel, data):
        return DAQmxSetAIExcitActualVal(self.taskHandle, channel, data)

    def ResetAIExcitActualVal(self, channel):
        return DAQmxResetAIExcitActualVal(self.taskHandle, channel)

    def GetAIExcitDCorAC(self, channel, data):
        return DAQmxGetAIExcitDCorAC(self.taskHandle, channel, data)

    def SetAIExcitDCorAC(self, channel, data):
        return DAQmxSetAIExcitDCorAC(self.taskHandle, channel, data)

    def ResetAIExcitDCorAC(self, channel):
        return DAQmxResetAIExcitDCorAC(self.taskHandle, channel)

    def GetAIExcitVoltageOrCurrent(self, channel, data):
        return DAQmxGetAIExcitVoltageOrCurrent(self.taskHandle, channel, data)

    def SetAIExcitVoltageOrCurrent(self, channel, data):
        return DAQmxSetAIExcitVoltageOrCurrent(self.taskHandle, channel, data)

    def ResetAIExcitVoltageOrCurrent(self, channel):
        return DAQmxResetAIExcitVoltageOrCurrent(self.taskHandle, channel)

    def GetAIACExcitFreq(self, channel, data):
        return DAQmxGetAIACExcitFreq(self.taskHandle, channel, data)

    def SetAIACExcitFreq(self, channel, data):
        return DAQmxSetAIACExcitFreq(self.taskHandle, channel, data)

    def ResetAIACExcitFreq(self, channel):
        return DAQmxResetAIACExcitFreq(self.taskHandle, channel)

    def GetAIACExcitSyncEnable(self, channel, data):
        return DAQmxGetAIACExcitSyncEnable(self.taskHandle, channel, data)

    def SetAIACExcitSyncEnable(self, channel, data):
        return DAQmxSetAIACExcitSyncEnable(self.taskHandle, channel, data)

    def ResetAIACExcitSyncEnable(self, channel):
        return DAQmxResetAIACExcitSyncEnable(self.taskHandle, channel)

    def GetAIACExcitWireMode(self, channel, data):
        return DAQmxGetAIACExcitWireMode(self.taskHandle, channel, data)

    def SetAIACExcitWireMode(self, channel, data):
        return DAQmxSetAIACExcitWireMode(self.taskHandle, channel, data)

    def ResetAIACExcitWireMode(self, channel):
        return DAQmxResetAIACExcitWireMode(self.taskHandle, channel)

    def GetAIOpenThrmcplDetectEnable(self, channel, data):
        return DAQmxGetAIOpenThrmcplDetectEnable(self.taskHandle, channel, data)

    def SetAIOpenThrmcplDetectEnable(self, channel, data):
        return DAQmxSetAIOpenThrmcplDetectEnable(self.taskHandle, channel, data)

    def ResetAIOpenThrmcplDetectEnable(self, channel):
        return DAQmxResetAIOpenThrmcplDetectEnable(self.taskHandle, channel)

    def GetAIThrmcplLeadOffsetVoltage(self, channel, data):
        return DAQmxGetAIThrmcplLeadOffsetVoltage(self.taskHandle, channel, data)

    def SetAIThrmcplLeadOffsetVoltage(self, channel, data):
        return DAQmxSetAIThrmcplLeadOffsetVoltage(self.taskHandle, channel, data)

    def ResetAIThrmcplLeadOffsetVoltage(self, channel):
        return DAQmxResetAIThrmcplLeadOffsetVoltage(self.taskHandle, channel)

    def GetAIAtten(self, channel, data):
        return DAQmxGetAIAtten(self.taskHandle, channel, data)

    def SetAIAtten(self, channel, data):
        return DAQmxSetAIAtten(self.taskHandle, channel, data)

    def ResetAIAtten(self, channel):
        return DAQmxResetAIAtten(self.taskHandle, channel)

    def GetAIProbeAtten(self, channel, data):
        return DAQmxGetAIProbeAtten(self.taskHandle, channel, data)

    def SetAIProbeAtten(self, channel, data):
        return DAQmxSetAIProbeAtten(self.taskHandle, channel, data)

    def ResetAIProbeAtten(self, channel):
        return DAQmxResetAIProbeAtten(self.taskHandle, channel)

    def GetAILowpassEnable(self, channel, data):
        return DAQmxGetAILowpassEnable(self.taskHandle, channel, data)

    def SetAILowpassEnable(self, channel, data):
        return DAQmxSetAILowpassEnable(self.taskHandle, channel, data)

    def ResetAILowpassEnable(self, channel):
        return DAQmxResetAILowpassEnable(self.taskHandle, channel)

    def GetAILowpassCutoffFreq(self, channel, data):
        return DAQmxGetAILowpassCutoffFreq(self.taskHandle, channel, data)

    def SetAILowpassCutoffFreq(self, channel, data):
        return DAQmxSetAILowpassCutoffFreq(self.taskHandle, channel, data)

    def ResetAILowpassCutoffFreq(self, channel):
        return DAQmxResetAILowpassCutoffFreq(self.taskHandle, channel)

    def GetAILowpassSwitchCapClkSrc(self, channel, data):
        return DAQmxGetAILowpassSwitchCapClkSrc(self.taskHandle, channel, data)

    def SetAILowpassSwitchCapClkSrc(self, channel, data):
        return DAQmxSetAILowpassSwitchCapClkSrc(self.taskHandle, channel, data)

    def ResetAILowpassSwitchCapClkSrc(self, channel):
        return DAQmxResetAILowpassSwitchCapClkSrc(self.taskHandle, channel)

    def GetAILowpassSwitchCapExtClkFreq(self, channel, data):
        return DAQmxGetAILowpassSwitchCapExtClkFreq(self.taskHandle, channel, data)

    def SetAILowpassSwitchCapExtClkFreq(self, channel, data):
        return DAQmxSetAILowpassSwitchCapExtClkFreq(self.taskHandle, channel, data)

    def ResetAILowpassSwitchCapExtClkFreq(self, channel):
        return DAQmxResetAILowpassSwitchCapExtClkFreq(self.taskHandle, channel)

    def GetAILowpassSwitchCapExtClkDiv(self, channel, data):
        return DAQmxGetAILowpassSwitchCapExtClkDiv(self.taskHandle, channel, data)

    def SetAILowpassSwitchCapExtClkDiv(self, channel, data):
        return DAQmxSetAILowpassSwitchCapExtClkDiv(self.taskHandle, channel, data)

    def ResetAILowpassSwitchCapExtClkDiv(self, channel):
        return DAQmxResetAILowpassSwitchCapExtClkDiv(self.taskHandle, channel)

    def GetAILowpassSwitchCapOutClkDiv(self, channel, data):
        return DAQmxGetAILowpassSwitchCapOutClkDiv(self.taskHandle, channel, data)

    def SetAILowpassSwitchCapOutClkDiv(self, channel, data):
        return DAQmxSetAILowpassSwitchCapOutClkDiv(self.taskHandle, channel, data)

    def ResetAILowpassSwitchCapOutClkDiv(self, channel):
        return DAQmxResetAILowpassSwitchCapOutClkDiv(self.taskHandle, channel)

    def GetAIFilterDelay(self, channel, data):
        return DAQmxGetAIFilterDelay(self.taskHandle, channel, data)

    def GetAIRemoveFilterDelay(self, channel, data):
        return DAQmxGetAIRemoveFilterDelay(self.taskHandle, channel, data)

    def SetAIRemoveFilterDelay(self, channel, data):
        return DAQmxSetAIRemoveFilterDelay(self.taskHandle, channel, data)

    def ResetAIRemoveFilterDelay(self, channel):
        return DAQmxResetAIRemoveFilterDelay(self.taskHandle, channel)

    def GetAIAveragingWinSize(self, channel, data):
        return DAQmxGetAIAveragingWinSize(self.taskHandle, channel, data)

    def SetAIAveragingWinSize(self, channel, data):
        return DAQmxSetAIAveragingWinSize(self.taskHandle, channel, data)

    def ResetAIAveragingWinSize(self, channel):
        return DAQmxResetAIAveragingWinSize(self.taskHandle, channel)

    def GetAIResolutionUnits(self, channel, data):
        return DAQmxGetAIResolutionUnits(self.taskHandle, channel, data)

    def GetAIResolution(self, channel, data):
        return DAQmxGetAIResolution(self.taskHandle, channel, data)

    def GetAIRawSampSize(self, channel, data):
        return DAQmxGetAIRawSampSize(self.taskHandle, channel, data)

    def GetAIRawSampJustification(self, channel, data):
        return DAQmxGetAIRawSampJustification(self.taskHandle, channel, data)

    def GetAIADCTimingMode(self, channel, data):
        return DAQmxGetAIADCTimingMode(self.taskHandle, channel, data)

    def SetAIADCTimingMode(self, channel, data):
        return DAQmxSetAIADCTimingMode(self.taskHandle, channel, data)

    def ResetAIADCTimingMode(self, channel):
        return DAQmxResetAIADCTimingMode(self.taskHandle, channel)

    def GetAIADCCustomTimingMode(self, channel, data):
        return DAQmxGetAIADCCustomTimingMode(self.taskHandle, channel, data)

    def SetAIADCCustomTimingMode(self, channel, data):
        return DAQmxSetAIADCCustomTimingMode(self.taskHandle, channel, data)

    def ResetAIADCCustomTimingMode(self, channel):
        return DAQmxResetAIADCCustomTimingMode(self.taskHandle, channel)

    def GetAIDitherEnable(self, channel, data):
        return DAQmxGetAIDitherEnable(self.taskHandle, channel, data)

    def SetAIDitherEnable(self, channel, data):
        return DAQmxSetAIDitherEnable(self.taskHandle, channel, data)

    def ResetAIDitherEnable(self, channel):
        return DAQmxResetAIDitherEnable(self.taskHandle, channel)

    def GetAIChanCalHasValidCalInfo(self, channel, data):
        return DAQmxGetAIChanCalHasValidCalInfo(self.taskHandle, channel, data)

    def GetAIChanCalEnableCal(self, channel, data):
        return DAQmxGetAIChanCalEnableCal(self.taskHandle, channel, data)

    def SetAIChanCalEnableCal(self, channel, data):
        return DAQmxSetAIChanCalEnableCal(self.taskHandle, channel, data)

    def ResetAIChanCalEnableCal(self, channel):
        return DAQmxResetAIChanCalEnableCal(self.taskHandle, channel)

    def GetAIChanCalApplyCalIfExp(self, channel, data):
        return DAQmxGetAIChanCalApplyCalIfExp(self.taskHandle, channel, data)

    def SetAIChanCalApplyCalIfExp(self, channel, data):
        return DAQmxSetAIChanCalApplyCalIfExp(self.taskHandle, channel, data)

    def ResetAIChanCalApplyCalIfExp(self, channel):
        return DAQmxResetAIChanCalApplyCalIfExp(self.taskHandle, channel)

    def GetAIChanCalScaleType(self, channel, data):
        return DAQmxGetAIChanCalScaleType(self.taskHandle, channel, data)

    def SetAIChanCalScaleType(self, channel, data):
        return DAQmxSetAIChanCalScaleType(self.taskHandle, channel, data)

    def ResetAIChanCalScaleType(self, channel):
        return DAQmxResetAIChanCalScaleType(self.taskHandle, channel)

    def GetAIChanCalTablePreScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalTablePreScaledVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalTablePreScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalTablePreScaledVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalTablePreScaledVals(self, channel):
        return DAQmxResetAIChanCalTablePreScaledVals(self.taskHandle, channel)

    def GetAIChanCalTableScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalTableScaledVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalTableScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalTableScaledVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalTableScaledVals(self, channel):
        return DAQmxResetAIChanCalTableScaledVals(self.taskHandle, channel)

    def GetAIChanCalPolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalPolyForwardCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalPolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalPolyForwardCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalPolyForwardCoeff(self, channel):
        return DAQmxResetAIChanCalPolyForwardCoeff(self.taskHandle, channel)

    def GetAIChanCalPolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalPolyReverseCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalPolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalPolyReverseCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalPolyReverseCoeff(self, channel):
        return DAQmxResetAIChanCalPolyReverseCoeff(self.taskHandle, channel)

    def GetAIChanCalOperatorName(self, channel, data, bufferSize):
        return DAQmxGetAIChanCalOperatorName(self.taskHandle, channel, data, bufferSize)

    def SetAIChanCalOperatorName(self, channel, data):
        return DAQmxSetAIChanCalOperatorName(self.taskHandle, channel, data)

    def ResetAIChanCalOperatorName(self, channel):
        return DAQmxResetAIChanCalOperatorName(self.taskHandle, channel)

    def GetAIChanCalDesc(self, channel, data, bufferSize):
        return DAQmxGetAIChanCalDesc(self.taskHandle, channel, data, bufferSize)

    def SetAIChanCalDesc(self, channel, data):
        return DAQmxSetAIChanCalDesc(self.taskHandle, channel, data)

    def ResetAIChanCalDesc(self, channel):
        return DAQmxResetAIChanCalDesc(self.taskHandle, channel)

    def GetAIChanCalVerifRefVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalVerifRefVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalVerifRefVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalVerifRefVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalVerifRefVals(self, channel):
        return DAQmxResetAIChanCalVerifRefVals(self.taskHandle, channel)

    def GetAIChanCalVerifAcqVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalVerifAcqVals(self.taskHandle, channel, data, arraySizeInElements)

    def SetAIChanCalVerifAcqVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalVerifAcqVals(self.taskHandle, channel, data, arraySizeInElements)

    def ResetAIChanCalVerifAcqVals(self, channel):
        return DAQmxResetAIChanCalVerifAcqVals(self.taskHandle, channel)

    def GetAIRngHigh(self, channel, data):
        return DAQmxGetAIRngHigh(self.taskHandle, channel, data)

    def SetAIRngHigh(self, channel, data):
        return DAQmxSetAIRngHigh(self.taskHandle, channel, data)

    def ResetAIRngHigh(self, channel):
        return DAQmxResetAIRngHigh(self.taskHandle, channel)

    def GetAIRngLow(self, channel, data):
        return DAQmxGetAIRngLow(self.taskHandle, channel, data)

    def SetAIRngLow(self, channel, data):
        return DAQmxSetAIRngLow(self.taskHandle, channel, data)

    def ResetAIRngLow(self, channel):
        return DAQmxResetAIRngLow(self.taskHandle, channel)

    def GetAIDCOffset(self, channel, data):
        return DAQmxGetAIDCOffset(self.taskHandle, channel, data)

    def SetAIDCOffset(self, channel, data):
        return DAQmxSetAIDCOffset(self.taskHandle, channel, data)

    def ResetAIDCOffset(self, channel):
        return DAQmxResetAIDCOffset(self.taskHandle, channel)

    def GetAIGain(self, channel, data):
        return DAQmxGetAIGain(self.taskHandle, channel, data)

    def SetAIGain(self, channel, data):
        return DAQmxSetAIGain(self.taskHandle, channel, data)

    def ResetAIGain(self, channel):
        return DAQmxResetAIGain(self.taskHandle, channel)

    def GetAISampAndHoldEnable(self, channel, data):
        return DAQmxGetAISampAndHoldEnable(self.taskHandle, channel, data)

    def SetAISampAndHoldEnable(self, channel, data):
        return DAQmxSetAISampAndHoldEnable(self.taskHandle, channel, data)

    def ResetAISampAndHoldEnable(self, channel):
        return DAQmxResetAISampAndHoldEnable(self.taskHandle, channel)

    def GetAIAutoZeroMode(self, channel, data):
        return DAQmxGetAIAutoZeroMode(self.taskHandle, channel, data)

    def SetAIAutoZeroMode(self, channel, data):
        return DAQmxSetAIAutoZeroMode(self.taskHandle, channel, data)

    def ResetAIAutoZeroMode(self, channel):
        return DAQmxResetAIAutoZeroMode(self.taskHandle, channel)

    def GetAIDataXferMech(self, channel, data):
        return DAQmxGetAIDataXferMech(self.taskHandle, channel, data)

    def SetAIDataXferMech(self, channel, data):
        return DAQmxSetAIDataXferMech(self.taskHandle, channel, data)

    def ResetAIDataXferMech(self, channel):
        return DAQmxResetAIDataXferMech(self.taskHandle, channel)

    def GetAIDataXferReqCond(self, channel, data):
        return DAQmxGetAIDataXferReqCond(self.taskHandle, channel, data)

    def SetAIDataXferReqCond(self, channel, data):
        return DAQmxSetAIDataXferReqCond(self.taskHandle, channel, data)

    def ResetAIDataXferReqCond(self, channel):
        return DAQmxResetAIDataXferReqCond(self.taskHandle, channel)

    def GetAIDataXferCustomThreshold(self, channel, data):
        return DAQmxGetAIDataXferCustomThreshold(self.taskHandle, channel, data)

    def SetAIDataXferCustomThreshold(self, channel, data):
        return DAQmxSetAIDataXferCustomThreshold(self.taskHandle, channel, data)

    def ResetAIDataXferCustomThreshold(self, channel):
        return DAQmxResetAIDataXferCustomThreshold(self.taskHandle, channel)

    def GetAIUsbXferReqSize(self, channel, data):
        return DAQmxGetAIUsbXferReqSize(self.taskHandle, channel, data)

    def SetAIUsbXferReqSize(self, channel, data):
        return DAQmxSetAIUsbXferReqSize(self.taskHandle, channel, data)

    def ResetAIUsbXferReqSize(self, channel):
        return DAQmxResetAIUsbXferReqSize(self.taskHandle, channel)

    def GetAIUsbXferReqCount(self, channel, data):
        return DAQmxGetAIUsbXferReqCount(self.taskHandle, channel, data)

    def SetAIUsbXferReqCount(self, channel, data):
        return DAQmxSetAIUsbXferReqCount(self.taskHandle, channel, data)

    def ResetAIUsbXferReqCount(self, channel):
        return DAQmxResetAIUsbXferReqCount(self.taskHandle, channel)

    def GetAIMemMapEnable(self, channel, data):
        return DAQmxGetAIMemMapEnable(self.taskHandle, channel, data)

    def SetAIMemMapEnable(self, channel, data):
        return DAQmxSetAIMemMapEnable(self.taskHandle, channel, data)

    def ResetAIMemMapEnable(self, channel):
        return DAQmxResetAIMemMapEnable(self.taskHandle, channel)

    def GetAIRawDataCompressionType(self, channel, data):
        return DAQmxGetAIRawDataCompressionType(self.taskHandle, channel, data)

    def SetAIRawDataCompressionType(self, channel, data):
        return DAQmxSetAIRawDataCompressionType(self.taskHandle, channel, data)

    def ResetAIRawDataCompressionType(self, channel):
        return DAQmxResetAIRawDataCompressionType(self.taskHandle, channel)

    def GetAILossyLSBRemovalCompressedSampSize(self, channel, data):
        return DAQmxGetAILossyLSBRemovalCompressedSampSize(self.taskHandle, channel, data)

    def SetAILossyLSBRemovalCompressedSampSize(self, channel, data):
        return DAQmxSetAILossyLSBRemovalCompressedSampSize(self.taskHandle, channel, data)

    def ResetAILossyLSBRemovalCompressedSampSize(self, channel):
        return DAQmxResetAILossyLSBRemovalCompressedSampSize(self.taskHandle, channel)

    def GetAIDevScalingCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIDevScalingCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def GetAIEnhancedAliasRejectionEnable(self, channel, data):
        return DAQmxGetAIEnhancedAliasRejectionEnable(self.taskHandle, channel, data)

    def SetAIEnhancedAliasRejectionEnable(self, channel, data):
        return DAQmxSetAIEnhancedAliasRejectionEnable(self.taskHandle, channel, data)

    def ResetAIEnhancedAliasRejectionEnable(self, channel):
        return DAQmxResetAIEnhancedAliasRejectionEnable(self.taskHandle, channel)

    def GetAOMax(self, channel, data):
        return DAQmxGetAOMax(self.taskHandle, channel, data)

    def SetAOMax(self, channel, data):
        return DAQmxSetAOMax(self.taskHandle, channel, data)

    def ResetAOMax(self, channel):
        return DAQmxResetAOMax(self.taskHandle, channel)

    def GetAOMin(self, channel, data):
        return DAQmxGetAOMin(self.taskHandle, channel, data)

    def SetAOMin(self, channel, data):
        return DAQmxSetAOMin(self.taskHandle, channel, data)

    def ResetAOMin(self, channel):
        return DAQmxResetAOMin(self.taskHandle, channel)

    def GetAOCustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetAOCustomScaleName(self.taskHandle, channel, data, bufferSize)

    def SetAOCustomScaleName(self, channel, data):
        return DAQmxSetAOCustomScaleName(self.taskHandle, channel, data)

    def ResetAOCustomScaleName(self, channel):
        return DAQmxResetAOCustomScaleName(self.taskHandle, channel)

    def GetAOOutputType(self, channel, data):
        return DAQmxGetAOOutputType(self.taskHandle, channel, data)

    def GetAOVoltageUnits(self, channel, data):
        return DAQmxGetAOVoltageUnits(self.taskHandle, channel, data)

    def SetAOVoltageUnits(self, channel, data):
        return DAQmxSetAOVoltageUnits(self.taskHandle, channel, data)

    def ResetAOVoltageUnits(self, channel):
        return DAQmxResetAOVoltageUnits(self.taskHandle, channel)

    def GetAOVoltageCurrentLimit(self, channel, data):
        return DAQmxGetAOVoltageCurrentLimit(self.taskHandle, channel, data)

    def SetAOVoltageCurrentLimit(self, channel, data):
        return DAQmxSetAOVoltageCurrentLimit(self.taskHandle, channel, data)

    def ResetAOVoltageCurrentLimit(self, channel):
        return DAQmxResetAOVoltageCurrentLimit(self.taskHandle, channel)

    def GetAOCurrentUnits(self, channel, data):
        return DAQmxGetAOCurrentUnits(self.taskHandle, channel, data)

    def SetAOCurrentUnits(self, channel, data):
        return DAQmxSetAOCurrentUnits(self.taskHandle, channel, data)

    def ResetAOCurrentUnits(self, channel):
        return DAQmxResetAOCurrentUnits(self.taskHandle, channel)

    def GetAOFuncGenType(self, channel, data):
        return DAQmxGetAOFuncGenType(self.taskHandle, channel, data)

    def SetAOFuncGenType(self, channel, data):
        return DAQmxSetAOFuncGenType(self.taskHandle, channel, data)

    def ResetAOFuncGenType(self, channel):
        return DAQmxResetAOFuncGenType(self.taskHandle, channel)

    def GetAOFuncGenFreq(self, channel, data):
        return DAQmxGetAOFuncGenFreq(self.taskHandle, channel, data)

    def SetAOFuncGenFreq(self, channel, data):
        return DAQmxSetAOFuncGenFreq(self.taskHandle, channel, data)

    def ResetAOFuncGenFreq(self, channel):
        return DAQmxResetAOFuncGenFreq(self.taskHandle, channel)

    def GetAOFuncGenAmplitude(self, channel, data):
        return DAQmxGetAOFuncGenAmplitude(self.taskHandle, channel, data)

    def SetAOFuncGenAmplitude(self, channel, data):
        return DAQmxSetAOFuncGenAmplitude(self.taskHandle, channel, data)

    def ResetAOFuncGenAmplitude(self, channel):
        return DAQmxResetAOFuncGenAmplitude(self.taskHandle, channel)

    def GetAOFuncGenOffset(self, channel, data):
        return DAQmxGetAOFuncGenOffset(self.taskHandle, channel, data)

    def SetAOFuncGenOffset(self, channel, data):
        return DAQmxSetAOFuncGenOffset(self.taskHandle, channel, data)

    def ResetAOFuncGenOffset(self, channel):
        return DAQmxResetAOFuncGenOffset(self.taskHandle, channel)

    def GetAOFuncGenSquareDutyCycle(self, channel, data):
        return DAQmxGetAOFuncGenSquareDutyCycle(self.taskHandle, channel, data)

    def SetAOFuncGenSquareDutyCycle(self, channel, data):
        return DAQmxSetAOFuncGenSquareDutyCycle(self.taskHandle, channel, data)

    def ResetAOFuncGenSquareDutyCycle(self, channel):
        return DAQmxResetAOFuncGenSquareDutyCycle(self.taskHandle, channel)

    def GetAOFuncGenModulationType(self, channel, data):
        return DAQmxGetAOFuncGenModulationType(self.taskHandle, channel, data)

    def SetAOFuncGenModulationType(self, channel, data):
        return DAQmxSetAOFuncGenModulationType(self.taskHandle, channel, data)

    def ResetAOFuncGenModulationType(self, channel):
        return DAQmxResetAOFuncGenModulationType(self.taskHandle, channel)

    def GetAOFuncGenFMDeviation(self, channel, data):
        return DAQmxGetAOFuncGenFMDeviation(self.taskHandle, channel, data)

    def SetAOFuncGenFMDeviation(self, channel, data):
        return DAQmxSetAOFuncGenFMDeviation(self.taskHandle, channel, data)

    def ResetAOFuncGenFMDeviation(self, channel):
        return DAQmxResetAOFuncGenFMDeviation(self.taskHandle, channel)

    def GetAOOutputImpedance(self, channel, data):
        return DAQmxGetAOOutputImpedance(self.taskHandle, channel, data)

    def SetAOOutputImpedance(self, channel, data):
        return DAQmxSetAOOutputImpedance(self.taskHandle, channel, data)

    def ResetAOOutputImpedance(self, channel):
        return DAQmxResetAOOutputImpedance(self.taskHandle, channel)

    def GetAOLoadImpedance(self, channel, data):
        return DAQmxGetAOLoadImpedance(self.taskHandle, channel, data)

    def SetAOLoadImpedance(self, channel, data):
        return DAQmxSetAOLoadImpedance(self.taskHandle, channel, data)

    def ResetAOLoadImpedance(self, channel):
        return DAQmxResetAOLoadImpedance(self.taskHandle, channel)

    def GetAOIdleOutputBehavior(self, channel, data):
        return DAQmxGetAOIdleOutputBehavior(self.taskHandle, channel, data)

    def SetAOIdleOutputBehavior(self, channel, data):
        return DAQmxSetAOIdleOutputBehavior(self.taskHandle, channel, data)

    def ResetAOIdleOutputBehavior(self, channel):
        return DAQmxResetAOIdleOutputBehavior(self.taskHandle, channel)

    def GetAOTermCfg(self, channel, data):
        return DAQmxGetAOTermCfg(self.taskHandle, channel, data)

    def SetAOTermCfg(self, channel, data):
        return DAQmxSetAOTermCfg(self.taskHandle, channel, data)

    def ResetAOTermCfg(self, channel):
        return DAQmxResetAOTermCfg(self.taskHandle, channel)

    def GetAOResolutionUnits(self, channel, data):
        return DAQmxGetAOResolutionUnits(self.taskHandle, channel, data)

    def SetAOResolutionUnits(self, channel, data):
        return DAQmxSetAOResolutionUnits(self.taskHandle, channel, data)

    def ResetAOResolutionUnits(self, channel):
        return DAQmxResetAOResolutionUnits(self.taskHandle, channel)

    def GetAOResolution(self, channel, data):
        return DAQmxGetAOResolution(self.taskHandle, channel, data)

    def GetAODACRngHigh(self, channel, data):
        return DAQmxGetAODACRngHigh(self.taskHandle, channel, data)

    def SetAODACRngHigh(self, channel, data):
        return DAQmxSetAODACRngHigh(self.taskHandle, channel, data)

    def ResetAODACRngHigh(self, channel):
        return DAQmxResetAODACRngHigh(self.taskHandle, channel)

    def GetAODACRngLow(self, channel, data):
        return DAQmxGetAODACRngLow(self.taskHandle, channel, data)

    def SetAODACRngLow(self, channel, data):
        return DAQmxSetAODACRngLow(self.taskHandle, channel, data)

    def ResetAODACRngLow(self, channel):
        return DAQmxResetAODACRngLow(self.taskHandle, channel)

    def GetAODACRefConnToGnd(self, channel, data):
        return DAQmxGetAODACRefConnToGnd(self.taskHandle, channel, data)

    def SetAODACRefConnToGnd(self, channel, data):
        return DAQmxSetAODACRefConnToGnd(self.taskHandle, channel, data)

    def ResetAODACRefConnToGnd(self, channel):
        return DAQmxResetAODACRefConnToGnd(self.taskHandle, channel)

    def GetAODACRefAllowConnToGnd(self, channel, data):
        return DAQmxGetAODACRefAllowConnToGnd(self.taskHandle, channel, data)

    def SetAODACRefAllowConnToGnd(self, channel, data):
        return DAQmxSetAODACRefAllowConnToGnd(self.taskHandle, channel, data)

    def ResetAODACRefAllowConnToGnd(self, channel):
        return DAQmxResetAODACRefAllowConnToGnd(self.taskHandle, channel)

    def GetAODACRefSrc(self, channel, data):
        return DAQmxGetAODACRefSrc(self.taskHandle, channel, data)

    def SetAODACRefSrc(self, channel, data):
        return DAQmxSetAODACRefSrc(self.taskHandle, channel, data)

    def ResetAODACRefSrc(self, channel):
        return DAQmxResetAODACRefSrc(self.taskHandle, channel)

    def GetAODACRefExtSrc(self, channel, data, bufferSize):
        return DAQmxGetAODACRefExtSrc(self.taskHandle, channel, data, bufferSize)

    def SetAODACRefExtSrc(self, channel, data):
        return DAQmxSetAODACRefExtSrc(self.taskHandle, channel, data)

    def ResetAODACRefExtSrc(self, channel):
        return DAQmxResetAODACRefExtSrc(self.taskHandle, channel)

    def GetAODACRefVal(self, channel, data):
        return DAQmxGetAODACRefVal(self.taskHandle, channel, data)

    def SetAODACRefVal(self, channel, data):
        return DAQmxSetAODACRefVal(self.taskHandle, channel, data)

    def ResetAODACRefVal(self, channel):
        return DAQmxResetAODACRefVal(self.taskHandle, channel)

    def GetAODACOffsetSrc(self, channel, data):
        return DAQmxGetAODACOffsetSrc(self.taskHandle, channel, data)

    def SetAODACOffsetSrc(self, channel, data):
        return DAQmxSetAODACOffsetSrc(self.taskHandle, channel, data)

    def ResetAODACOffsetSrc(self, channel):
        return DAQmxResetAODACOffsetSrc(self.taskHandle, channel)

    def GetAODACOffsetExtSrc(self, channel, data, bufferSize):
        return DAQmxGetAODACOffsetExtSrc(self.taskHandle, channel, data, bufferSize)

    def SetAODACOffsetExtSrc(self, channel, data):
        return DAQmxSetAODACOffsetExtSrc(self.taskHandle, channel, data)

    def ResetAODACOffsetExtSrc(self, channel):
        return DAQmxResetAODACOffsetExtSrc(self.taskHandle, channel)

    def GetAODACOffsetVal(self, channel, data):
        return DAQmxGetAODACOffsetVal(self.taskHandle, channel, data)

    def SetAODACOffsetVal(self, channel, data):
        return DAQmxSetAODACOffsetVal(self.taskHandle, channel, data)

    def ResetAODACOffsetVal(self, channel):
        return DAQmxResetAODACOffsetVal(self.taskHandle, channel)

    def GetAOReglitchEnable(self, channel, data):
        return DAQmxGetAOReglitchEnable(self.taskHandle, channel, data)

    def SetAOReglitchEnable(self, channel, data):
        return DAQmxSetAOReglitchEnable(self.taskHandle, channel, data)

    def ResetAOReglitchEnable(self, channel):
        return DAQmxResetAOReglitchEnable(self.taskHandle, channel)

    def GetAOGain(self, channel, data):
        return DAQmxGetAOGain(self.taskHandle, channel, data)

    def SetAOGain(self, channel, data):
        return DAQmxSetAOGain(self.taskHandle, channel, data)

    def ResetAOGain(self, channel):
        return DAQmxResetAOGain(self.taskHandle, channel)

    def GetAOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetAOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def SetAOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetAOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def ResetAOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetAOUseOnlyOnBrdMem(self.taskHandle, channel)

    def GetAODataXferMech(self, channel, data):
        return DAQmxGetAODataXferMech(self.taskHandle, channel, data)

    def SetAODataXferMech(self, channel, data):
        return DAQmxSetAODataXferMech(self.taskHandle, channel, data)

    def ResetAODataXferMech(self, channel):
        return DAQmxResetAODataXferMech(self.taskHandle, channel)

    def GetAODataXferReqCond(self, channel, data):
        return DAQmxGetAODataXferReqCond(self.taskHandle, channel, data)

    def SetAODataXferReqCond(self, channel, data):
        return DAQmxSetAODataXferReqCond(self.taskHandle, channel, data)

    def ResetAODataXferReqCond(self, channel):
        return DAQmxResetAODataXferReqCond(self.taskHandle, channel)

    def GetAOUsbXferReqSize(self, channel, data):
        return DAQmxGetAOUsbXferReqSize(self.taskHandle, channel, data)

    def SetAOUsbXferReqSize(self, channel, data):
        return DAQmxSetAOUsbXferReqSize(self.taskHandle, channel, data)

    def ResetAOUsbXferReqSize(self, channel):
        return DAQmxResetAOUsbXferReqSize(self.taskHandle, channel)

    def GetAOUsbXferReqCount(self, channel, data):
        return DAQmxGetAOUsbXferReqCount(self.taskHandle, channel, data)

    def SetAOUsbXferReqCount(self, channel, data):
        return DAQmxSetAOUsbXferReqCount(self.taskHandle, channel, data)

    def ResetAOUsbXferReqCount(self, channel):
        return DAQmxResetAOUsbXferReqCount(self.taskHandle, channel)

    def GetAOMemMapEnable(self, channel, data):
        return DAQmxGetAOMemMapEnable(self.taskHandle, channel, data)

    def SetAOMemMapEnable(self, channel, data):
        return DAQmxSetAOMemMapEnable(self.taskHandle, channel, data)

    def ResetAOMemMapEnable(self, channel):
        return DAQmxResetAOMemMapEnable(self.taskHandle, channel)

    def GetAODevScalingCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAODevScalingCoeff(self.taskHandle, channel, data, arraySizeInElements)

    def GetAOEnhancedImageRejectionEnable(self, channel, data):
        return DAQmxGetAOEnhancedImageRejectionEnable(self.taskHandle, channel, data)

    def SetAOEnhancedImageRejectionEnable(self, channel, data):
        return DAQmxSetAOEnhancedImageRejectionEnable(self.taskHandle, channel, data)

    def ResetAOEnhancedImageRejectionEnable(self, channel):
        return DAQmxResetAOEnhancedImageRejectionEnable(self.taskHandle, channel)

    def GetDIInvertLines(self, channel, data):
        return DAQmxGetDIInvertLines(self.taskHandle, channel, data)

    def SetDIInvertLines(self, channel, data):
        return DAQmxSetDIInvertLines(self.taskHandle, channel, data)

    def ResetDIInvertLines(self, channel):
        return DAQmxResetDIInvertLines(self.taskHandle, channel)

    def GetDINumLines(self, channel, data):
        return DAQmxGetDINumLines(self.taskHandle, channel, data)

    def GetDIDigFltrEnable(self, channel, data):
        return DAQmxGetDIDigFltrEnable(self.taskHandle, channel, data)

    def SetDIDigFltrEnable(self, channel, data):
        return DAQmxSetDIDigFltrEnable(self.taskHandle, channel, data)

    def ResetDIDigFltrEnable(self, channel):
        return DAQmxResetDIDigFltrEnable(self.taskHandle, channel)

    def GetDIDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetDIDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetDIDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetDIDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetDIDigFltrMinPulseWidth(self, channel):
        return DAQmxResetDIDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetDIDigFltrEnableBusMode(self, channel, data):
        return DAQmxGetDIDigFltrEnableBusMode(self.taskHandle, channel, data)

    def SetDIDigFltrEnableBusMode(self, channel, data):
        return DAQmxSetDIDigFltrEnableBusMode(self.taskHandle, channel, data)

    def ResetDIDigFltrEnableBusMode(self, channel):
        return DAQmxResetDIDigFltrEnableBusMode(self.taskHandle, channel)

    def GetDIDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetDIDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetDIDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetDIDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetDIDigFltrTimebaseSrc(self, channel):
        return DAQmxResetDIDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetDIDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetDIDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetDIDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetDIDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetDIDigFltrTimebaseRate(self, channel):
        return DAQmxResetDIDigFltrTimebaseRate(self.taskHandle, channel)

    def GetDIDigSyncEnable(self, channel, data):
        return DAQmxGetDIDigSyncEnable(self.taskHandle, channel, data)

    def SetDIDigSyncEnable(self, channel, data):
        return DAQmxSetDIDigSyncEnable(self.taskHandle, channel, data)

    def ResetDIDigSyncEnable(self, channel):
        return DAQmxResetDIDigSyncEnable(self.taskHandle, channel)

    def GetDITristate(self, channel, data):
        return DAQmxGetDITristate(self.taskHandle, channel, data)

    def SetDITristate(self, channel, data):
        return DAQmxSetDITristate(self.taskHandle, channel, data)

    def ResetDITristate(self, channel):
        return DAQmxResetDITristate(self.taskHandle, channel)

    def GetDILogicFamily(self, channel, data):
        return DAQmxGetDILogicFamily(self.taskHandle, channel, data)

    def SetDILogicFamily(self, channel, data):
        return DAQmxSetDILogicFamily(self.taskHandle, channel, data)

    def ResetDILogicFamily(self, channel):
        return DAQmxResetDILogicFamily(self.taskHandle, channel)

    def GetDIDataXferMech(self, channel, data):
        return DAQmxGetDIDataXferMech(self.taskHandle, channel, data)

    def SetDIDataXferMech(self, channel, data):
        return DAQmxSetDIDataXferMech(self.taskHandle, channel, data)

    def ResetDIDataXferMech(self, channel):
        return DAQmxResetDIDataXferMech(self.taskHandle, channel)

    def GetDIDataXferReqCond(self, channel, data):
        return DAQmxGetDIDataXferReqCond(self.taskHandle, channel, data)

    def SetDIDataXferReqCond(self, channel, data):
        return DAQmxSetDIDataXferReqCond(self.taskHandle, channel, data)

    def ResetDIDataXferReqCond(self, channel):
        return DAQmxResetDIDataXferReqCond(self.taskHandle, channel)

    def GetDIUsbXferReqSize(self, channel, data):
        return DAQmxGetDIUsbXferReqSize(self.taskHandle, channel, data)

    def SetDIUsbXferReqSize(self, channel, data):
        return DAQmxSetDIUsbXferReqSize(self.taskHandle, channel, data)

    def ResetDIUsbXferReqSize(self, channel):
        return DAQmxResetDIUsbXferReqSize(self.taskHandle, channel)

    def GetDIUsbXferReqCount(self, channel, data):
        return DAQmxGetDIUsbXferReqCount(self.taskHandle, channel, data)

    def SetDIUsbXferReqCount(self, channel, data):
        return DAQmxSetDIUsbXferReqCount(self.taskHandle, channel, data)

    def ResetDIUsbXferReqCount(self, channel):
        return DAQmxResetDIUsbXferReqCount(self.taskHandle, channel)

    def GetDIMemMapEnable(self, channel, data):
        return DAQmxGetDIMemMapEnable(self.taskHandle, channel, data)

    def SetDIMemMapEnable(self, channel, data):
        return DAQmxSetDIMemMapEnable(self.taskHandle, channel, data)

    def ResetDIMemMapEnable(self, channel):
        return DAQmxResetDIMemMapEnable(self.taskHandle, channel)

    def GetDIAcquireOn(self, channel, data):
        return DAQmxGetDIAcquireOn(self.taskHandle, channel, data)

    def SetDIAcquireOn(self, channel, data):
        return DAQmxSetDIAcquireOn(self.taskHandle, channel, data)

    def ResetDIAcquireOn(self, channel):
        return DAQmxResetDIAcquireOn(self.taskHandle, channel)

    def GetDOOutputDriveType(self, channel, data):
        return DAQmxGetDOOutputDriveType(self.taskHandle, channel, data)

    def SetDOOutputDriveType(self, channel, data):
        return DAQmxSetDOOutputDriveType(self.taskHandle, channel, data)

    def ResetDOOutputDriveType(self, channel):
        return DAQmxResetDOOutputDriveType(self.taskHandle, channel)

    def GetDOInvertLines(self, channel, data):
        return DAQmxGetDOInvertLines(self.taskHandle, channel, data)

    def SetDOInvertLines(self, channel, data):
        return DAQmxSetDOInvertLines(self.taskHandle, channel, data)

    def ResetDOInvertLines(self, channel):
        return DAQmxResetDOInvertLines(self.taskHandle, channel)

    def GetDONumLines(self, channel, data):
        return DAQmxGetDONumLines(self.taskHandle, channel, data)

    def GetDOTristate(self, channel, data):
        return DAQmxGetDOTristate(self.taskHandle, channel, data)

    def SetDOTristate(self, channel, data):
        return DAQmxSetDOTristate(self.taskHandle, channel, data)

    def ResetDOTristate(self, channel):
        return DAQmxResetDOTristate(self.taskHandle, channel)

    def GetDOLineStatesStartState(self, channel, data):
        return DAQmxGetDOLineStatesStartState(self.taskHandle, channel, data)

    def SetDOLineStatesStartState(self, channel, data):
        return DAQmxSetDOLineStatesStartState(self.taskHandle, channel, data)

    def ResetDOLineStatesStartState(self, channel):
        return DAQmxResetDOLineStatesStartState(self.taskHandle, channel)

    def GetDOLineStatesPausedState(self, channel, data):
        return DAQmxGetDOLineStatesPausedState(self.taskHandle, channel, data)

    def SetDOLineStatesPausedState(self, channel, data):
        return DAQmxSetDOLineStatesPausedState(self.taskHandle, channel, data)

    def ResetDOLineStatesPausedState(self, channel):
        return DAQmxResetDOLineStatesPausedState(self.taskHandle, channel)

    def GetDOLineStatesDoneState(self, channel, data):
        return DAQmxGetDOLineStatesDoneState(self.taskHandle, channel, data)

    def SetDOLineStatesDoneState(self, channel, data):
        return DAQmxSetDOLineStatesDoneState(self.taskHandle, channel, data)

    def ResetDOLineStatesDoneState(self, channel):
        return DAQmxResetDOLineStatesDoneState(self.taskHandle, channel)

    def GetDOLogicFamily(self, channel, data):
        return DAQmxGetDOLogicFamily(self.taskHandle, channel, data)

    def SetDOLogicFamily(self, channel, data):
        return DAQmxSetDOLogicFamily(self.taskHandle, channel, data)

    def ResetDOLogicFamily(self, channel):
        return DAQmxResetDOLogicFamily(self.taskHandle, channel)

    def GetDOOvercurrentLimit(self, channel, data):
        return DAQmxGetDOOvercurrentLimit(self.taskHandle, channel, data)

    def SetDOOvercurrentLimit(self, channel, data):
        return DAQmxSetDOOvercurrentLimit(self.taskHandle, channel, data)

    def ResetDOOvercurrentLimit(self, channel):
        return DAQmxResetDOOvercurrentLimit(self.taskHandle, channel)

    def GetDOOvercurrentAutoReenable(self, channel, data):
        return DAQmxGetDOOvercurrentAutoReenable(self.taskHandle, channel, data)

    def SetDOOvercurrentAutoReenable(self, channel, data):
        return DAQmxSetDOOvercurrentAutoReenable(self.taskHandle, channel, data)

    def ResetDOOvercurrentAutoReenable(self, channel):
        return DAQmxResetDOOvercurrentAutoReenable(self.taskHandle, channel)

    def GetDOOvercurrentReenablePeriod(self, channel, data):
        return DAQmxGetDOOvercurrentReenablePeriod(self.taskHandle, channel, data)

    def SetDOOvercurrentReenablePeriod(self, channel, data):
        return DAQmxSetDOOvercurrentReenablePeriod(self.taskHandle, channel, data)

    def ResetDOOvercurrentReenablePeriod(self, channel):
        return DAQmxResetDOOvercurrentReenablePeriod(self.taskHandle, channel)

    def GetDOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetDOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def SetDOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetDOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def ResetDOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetDOUseOnlyOnBrdMem(self.taskHandle, channel)

    def GetDODataXferMech(self, channel, data):
        return DAQmxGetDODataXferMech(self.taskHandle, channel, data)

    def SetDODataXferMech(self, channel, data):
        return DAQmxSetDODataXferMech(self.taskHandle, channel, data)

    def ResetDODataXferMech(self, channel):
        return DAQmxResetDODataXferMech(self.taskHandle, channel)

    def GetDODataXferReqCond(self, channel, data):
        return DAQmxGetDODataXferReqCond(self.taskHandle, channel, data)

    def SetDODataXferReqCond(self, channel, data):
        return DAQmxSetDODataXferReqCond(self.taskHandle, channel, data)

    def ResetDODataXferReqCond(self, channel):
        return DAQmxResetDODataXferReqCond(self.taskHandle, channel)

    def GetDOUsbXferReqSize(self, channel, data):
        return DAQmxGetDOUsbXferReqSize(self.taskHandle, channel, data)

    def SetDOUsbXferReqSize(self, channel, data):
        return DAQmxSetDOUsbXferReqSize(self.taskHandle, channel, data)

    def ResetDOUsbXferReqSize(self, channel):
        return DAQmxResetDOUsbXferReqSize(self.taskHandle, channel)

    def GetDOUsbXferReqCount(self, channel, data):
        return DAQmxGetDOUsbXferReqCount(self.taskHandle, channel, data)

    def SetDOUsbXferReqCount(self, channel, data):
        return DAQmxSetDOUsbXferReqCount(self.taskHandle, channel, data)

    def ResetDOUsbXferReqCount(self, channel):
        return DAQmxResetDOUsbXferReqCount(self.taskHandle, channel)

    def GetDOMemMapEnable(self, channel, data):
        return DAQmxGetDOMemMapEnable(self.taskHandle, channel, data)

    def SetDOMemMapEnable(self, channel, data):
        return DAQmxSetDOMemMapEnable(self.taskHandle, channel, data)

    def ResetDOMemMapEnable(self, channel):
        return DAQmxResetDOMemMapEnable(self.taskHandle, channel)

    def GetDOGenerateOn(self, channel, data):
        return DAQmxGetDOGenerateOn(self.taskHandle, channel, data)

    def SetDOGenerateOn(self, channel, data):
        return DAQmxSetDOGenerateOn(self.taskHandle, channel, data)

    def ResetDOGenerateOn(self, channel):
        return DAQmxResetDOGenerateOn(self.taskHandle, channel)

    def GetCIMax(self, channel, data):
        return DAQmxGetCIMax(self.taskHandle, channel, data)

    def SetCIMax(self, channel, data):
        return DAQmxSetCIMax(self.taskHandle, channel, data)

    def ResetCIMax(self, channel):
        return DAQmxResetCIMax(self.taskHandle, channel)

    def GetCIMin(self, channel, data):
        return DAQmxGetCIMin(self.taskHandle, channel, data)

    def SetCIMin(self, channel, data):
        return DAQmxSetCIMin(self.taskHandle, channel, data)

    def ResetCIMin(self, channel):
        return DAQmxResetCIMin(self.taskHandle, channel)

    def GetCICustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetCICustomScaleName(self.taskHandle, channel, data, bufferSize)

    def SetCICustomScaleName(self, channel, data):
        return DAQmxSetCICustomScaleName(self.taskHandle, channel, data)

    def ResetCICustomScaleName(self, channel):
        return DAQmxResetCICustomScaleName(self.taskHandle, channel)

    def GetCIMeasType(self, channel, data):
        return DAQmxGetCIMeasType(self.taskHandle, channel, data)

    def GetCIFreqUnits(self, channel, data):
        return DAQmxGetCIFreqUnits(self.taskHandle, channel, data)

    def SetCIFreqUnits(self, channel, data):
        return DAQmxSetCIFreqUnits(self.taskHandle, channel, data)

    def ResetCIFreqUnits(self, channel):
        return DAQmxResetCIFreqUnits(self.taskHandle, channel)

    def GetCIFreqTerm(self, channel, data, bufferSize):
        return DAQmxGetCIFreqTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIFreqTerm(self, channel, data):
        return DAQmxSetCIFreqTerm(self.taskHandle, channel, data)

    def ResetCIFreqTerm(self, channel):
        return DAQmxResetCIFreqTerm(self.taskHandle, channel)

    def GetCIFreqStartingEdge(self, channel, data):
        return DAQmxGetCIFreqStartingEdge(self.taskHandle, channel, data)

    def SetCIFreqStartingEdge(self, channel, data):
        return DAQmxSetCIFreqStartingEdge(self.taskHandle, channel, data)

    def ResetCIFreqStartingEdge(self, channel):
        return DAQmxResetCIFreqStartingEdge(self.taskHandle, channel)

    def GetCIFreqMeasMeth(self, channel, data):
        return DAQmxGetCIFreqMeasMeth(self.taskHandle, channel, data)

    def SetCIFreqMeasMeth(self, channel, data):
        return DAQmxSetCIFreqMeasMeth(self.taskHandle, channel, data)

    def ResetCIFreqMeasMeth(self, channel):
        return DAQmxResetCIFreqMeasMeth(self.taskHandle, channel)

    def GetCIFreqEnableAveraging(self, channel, data):
        return DAQmxGetCIFreqEnableAveraging(self.taskHandle, channel, data)

    def SetCIFreqEnableAveraging(self, channel, data):
        return DAQmxSetCIFreqEnableAveraging(self.taskHandle, channel, data)

    def ResetCIFreqEnableAveraging(self, channel):
        return DAQmxResetCIFreqEnableAveraging(self.taskHandle, channel)

    def GetCIFreqMeasTime(self, channel, data):
        return DAQmxGetCIFreqMeasTime(self.taskHandle, channel, data)

    def SetCIFreqMeasTime(self, channel, data):
        return DAQmxSetCIFreqMeasTime(self.taskHandle, channel, data)

    def ResetCIFreqMeasTime(self, channel):
        return DAQmxResetCIFreqMeasTime(self.taskHandle, channel)

    def GetCIFreqDiv(self, channel, data):
        return DAQmxGetCIFreqDiv(self.taskHandle, channel, data)

    def SetCIFreqDiv(self, channel, data):
        return DAQmxSetCIFreqDiv(self.taskHandle, channel, data)

    def ResetCIFreqDiv(self, channel):
        return DAQmxResetCIFreqDiv(self.taskHandle, channel)

    def GetCIFreqDigFltrEnable(self, channel, data):
        return DAQmxGetCIFreqDigFltrEnable(self.taskHandle, channel, data)

    def SetCIFreqDigFltrEnable(self, channel, data):
        return DAQmxSetCIFreqDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIFreqDigFltrEnable(self, channel):
        return DAQmxResetCIFreqDigFltrEnable(self.taskHandle, channel)

    def GetCIFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIFreqDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIFreqDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIFreqDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIFreqDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIFreqDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIFreqDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIFreqDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIFreqDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIFreqDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIFreqDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIFreqDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIFreqDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIFreqDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIFreqDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIFreqDigSyncEnable(self, channel, data):
        return DAQmxGetCIFreqDigSyncEnable(self.taskHandle, channel, data)

    def SetCIFreqDigSyncEnable(self, channel, data):
        return DAQmxSetCIFreqDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIFreqDigSyncEnable(self, channel):
        return DAQmxResetCIFreqDigSyncEnable(self.taskHandle, channel)

    def GetCIPeriodUnits(self, channel, data):
        return DAQmxGetCIPeriodUnits(self.taskHandle, channel, data)

    def SetCIPeriodUnits(self, channel, data):
        return DAQmxSetCIPeriodUnits(self.taskHandle, channel, data)

    def ResetCIPeriodUnits(self, channel):
        return DAQmxResetCIPeriodUnits(self.taskHandle, channel)

    def GetCIPeriodTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPeriodTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIPeriodTerm(self, channel, data):
        return DAQmxSetCIPeriodTerm(self.taskHandle, channel, data)

    def ResetCIPeriodTerm(self, channel):
        return DAQmxResetCIPeriodTerm(self.taskHandle, channel)

    def GetCIPeriodStartingEdge(self, channel, data):
        return DAQmxGetCIPeriodStartingEdge(self.taskHandle, channel, data)

    def SetCIPeriodStartingEdge(self, channel, data):
        return DAQmxSetCIPeriodStartingEdge(self.taskHandle, channel, data)

    def ResetCIPeriodStartingEdge(self, channel):
        return DAQmxResetCIPeriodStartingEdge(self.taskHandle, channel)

    def GetCIPeriodMeasMeth(self, channel, data):
        return DAQmxGetCIPeriodMeasMeth(self.taskHandle, channel, data)

    def SetCIPeriodMeasMeth(self, channel, data):
        return DAQmxSetCIPeriodMeasMeth(self.taskHandle, channel, data)

    def ResetCIPeriodMeasMeth(self, channel):
        return DAQmxResetCIPeriodMeasMeth(self.taskHandle, channel)

    def GetCIPeriodEnableAveraging(self, channel, data):
        return DAQmxGetCIPeriodEnableAveraging(self.taskHandle, channel, data)

    def SetCIPeriodEnableAveraging(self, channel, data):
        return DAQmxSetCIPeriodEnableAveraging(self.taskHandle, channel, data)

    def ResetCIPeriodEnableAveraging(self, channel):
        return DAQmxResetCIPeriodEnableAveraging(self.taskHandle, channel)

    def GetCIPeriodMeasTime(self, channel, data):
        return DAQmxGetCIPeriodMeasTime(self.taskHandle, channel, data)

    def SetCIPeriodMeasTime(self, channel, data):
        return DAQmxSetCIPeriodMeasTime(self.taskHandle, channel, data)

    def ResetCIPeriodMeasTime(self, channel):
        return DAQmxResetCIPeriodMeasTime(self.taskHandle, channel)

    def GetCIPeriodDiv(self, channel, data):
        return DAQmxGetCIPeriodDiv(self.taskHandle, channel, data)

    def SetCIPeriodDiv(self, channel, data):
        return DAQmxSetCIPeriodDiv(self.taskHandle, channel, data)

    def ResetCIPeriodDiv(self, channel):
        return DAQmxResetCIPeriodDiv(self.taskHandle, channel)

    def GetCIPeriodDigFltrEnable(self, channel, data):
        return DAQmxGetCIPeriodDigFltrEnable(self.taskHandle, channel, data)

    def SetCIPeriodDigFltrEnable(self, channel, data):
        return DAQmxSetCIPeriodDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIPeriodDigFltrEnable(self, channel):
        return DAQmxResetCIPeriodDigFltrEnable(self.taskHandle, channel)

    def GetCIPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPeriodDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPeriodDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIPeriodDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPeriodDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIPeriodDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPeriodDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIPeriodDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPeriodDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIPeriodDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPeriodDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPeriodDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPeriodDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIPeriodDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPeriodDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIPeriodDigSyncEnable(self, channel, data):
        return DAQmxGetCIPeriodDigSyncEnable(self.taskHandle, channel, data)

    def SetCIPeriodDigSyncEnable(self, channel, data):
        return DAQmxSetCIPeriodDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIPeriodDigSyncEnable(self, channel):
        return DAQmxResetCIPeriodDigSyncEnable(self.taskHandle, channel)

    def GetCICountEdgesTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesTerm(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesTerm(self, channel, data):
        return DAQmxSetCICountEdgesTerm(self.taskHandle, channel, data)

    def ResetCICountEdgesTerm(self, channel):
        return DAQmxResetCICountEdgesTerm(self.taskHandle, channel)

    def GetCICountEdgesDir(self, channel, data):
        return DAQmxGetCICountEdgesDir(self.taskHandle, channel, data)

    def SetCICountEdgesDir(self, channel, data):
        return DAQmxSetCICountEdgesDir(self.taskHandle, channel, data)

    def ResetCICountEdgesDir(self, channel):
        return DAQmxResetCICountEdgesDir(self.taskHandle, channel)

    def GetCICountEdgesDirTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesDirTerm(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesDirTerm(self, channel, data):
        return DAQmxSetCICountEdgesDirTerm(self.taskHandle, channel, data)

    def ResetCICountEdgesDirTerm(self, channel):
        return DAQmxResetCICountEdgesDirTerm(self.taskHandle, channel)

    def GetCICountEdgesCountDirDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrEnable(self.taskHandle, channel, data)

    def SetCICountEdgesCountDirDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesCountDirDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrEnable(self.taskHandle, channel)

    def GetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCICountEdgesCountDirDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCICountEdgesCountDirDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCICountEdgesCountDirDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCICountEdgesCountDirDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigSyncEnable(self.taskHandle, channel, data)

    def SetCICountEdgesCountDirDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigSyncEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesCountDirDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesCountDirDigSyncEnable(self.taskHandle, channel)

    def GetCICountEdgesInitialCnt(self, channel, data):
        return DAQmxGetCICountEdgesInitialCnt(self.taskHandle, channel, data)

    def SetCICountEdgesInitialCnt(self, channel, data):
        return DAQmxSetCICountEdgesInitialCnt(self.taskHandle, channel, data)

    def ResetCICountEdgesInitialCnt(self, channel):
        return DAQmxResetCICountEdgesInitialCnt(self.taskHandle, channel)

    def GetCICountEdgesActiveEdge(self, channel, data):
        return DAQmxGetCICountEdgesActiveEdge(self.taskHandle, channel, data)

    def SetCICountEdgesActiveEdge(self, channel, data):
        return DAQmxSetCICountEdgesActiveEdge(self.taskHandle, channel, data)

    def ResetCICountEdgesActiveEdge(self, channel):
        return DAQmxResetCICountEdgesActiveEdge(self.taskHandle, channel)

    def GetCICountEdgesCountResetEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetEnable(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetEnable(self.taskHandle, channel)

    def GetCICountEdgesCountResetResetCount(self, channel, data):
        return DAQmxGetCICountEdgesCountResetResetCount(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetResetCount(self, channel, data):
        return DAQmxSetCICountEdgesCountResetResetCount(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetResetCount(self, channel):
        return DAQmxResetCICountEdgesCountResetResetCount(self.taskHandle, channel)

    def GetCICountEdgesCountResetTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountResetTerm(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesCountResetTerm(self, channel, data):
        return DAQmxSetCICountEdgesCountResetTerm(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetTerm(self, channel):
        return DAQmxResetCICountEdgesCountResetTerm(self.taskHandle, channel)

    def GetCICountEdgesCountResetActiveEdge(self, channel, data):
        return DAQmxGetCICountEdgesCountResetActiveEdge(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetActiveEdge(self, channel, data):
        return DAQmxSetCICountEdgesCountResetActiveEdge(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetActiveEdge(self, channel):
        return DAQmxResetCICountEdgesCountResetActiveEdge(self.taskHandle, channel)

    def GetCICountEdgesCountResetDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrEnable(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrEnable(self.taskHandle, channel)

    def GetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCICountEdgesCountResetDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCICountEdgesCountResetDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigSyncEnable(self.taskHandle, channel, data)

    def SetCICountEdgesCountResetDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigSyncEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesCountResetDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetDigSyncEnable(self.taskHandle, channel)

    def GetCICountEdgesDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrEnable(self.taskHandle, channel, data)

    def SetCICountEdgesDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesDigFltrEnable(self.taskHandle, channel)

    def GetCICountEdgesDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCICountEdgesDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCICountEdgesDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCICountEdgesDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCICountEdgesDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCICountEdgesDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCICountEdgesDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCICountEdgesDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCICountEdgesDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCICountEdgesDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesDigSyncEnable(self.taskHandle, channel, data)

    def SetCICountEdgesDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesDigSyncEnable(self.taskHandle, channel, data)

    def ResetCICountEdgesDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesDigSyncEnable(self.taskHandle, channel)

    def GetCIAngEncoderUnits(self, channel, data):
        return DAQmxGetCIAngEncoderUnits(self.taskHandle, channel, data)

    def SetCIAngEncoderUnits(self, channel, data):
        return DAQmxSetCIAngEncoderUnits(self.taskHandle, channel, data)

    def ResetCIAngEncoderUnits(self, channel):
        return DAQmxResetCIAngEncoderUnits(self.taskHandle, channel)

    def GetCIAngEncoderPulsesPerRev(self, channel, data):
        return DAQmxGetCIAngEncoderPulsesPerRev(self.taskHandle, channel, data)

    def SetCIAngEncoderPulsesPerRev(self, channel, data):
        return DAQmxSetCIAngEncoderPulsesPerRev(self.taskHandle, channel, data)

    def ResetCIAngEncoderPulsesPerRev(self, channel):
        return DAQmxResetCIAngEncoderPulsesPerRev(self.taskHandle, channel)

    def GetCIAngEncoderInitialAngle(self, channel, data):
        return DAQmxGetCIAngEncoderInitialAngle(self.taskHandle, channel, data)

    def SetCIAngEncoderInitialAngle(self, channel, data):
        return DAQmxSetCIAngEncoderInitialAngle(self.taskHandle, channel, data)

    def ResetCIAngEncoderInitialAngle(self, channel):
        return DAQmxResetCIAngEncoderInitialAngle(self.taskHandle, channel)

    def GetCILinEncoderUnits(self, channel, data):
        return DAQmxGetCILinEncoderUnits(self.taskHandle, channel, data)

    def SetCILinEncoderUnits(self, channel, data):
        return DAQmxSetCILinEncoderUnits(self.taskHandle, channel, data)

    def ResetCILinEncoderUnits(self, channel):
        return DAQmxResetCILinEncoderUnits(self.taskHandle, channel)

    def GetCILinEncoderDistPerPulse(self, channel, data):
        return DAQmxGetCILinEncoderDistPerPulse(self.taskHandle, channel, data)

    def SetCILinEncoderDistPerPulse(self, channel, data):
        return DAQmxSetCILinEncoderDistPerPulse(self.taskHandle, channel, data)

    def ResetCILinEncoderDistPerPulse(self, channel):
        return DAQmxResetCILinEncoderDistPerPulse(self.taskHandle, channel)

    def GetCILinEncoderInitialPos(self, channel, data):
        return DAQmxGetCILinEncoderInitialPos(self.taskHandle, channel, data)

    def SetCILinEncoderInitialPos(self, channel, data):
        return DAQmxSetCILinEncoderInitialPos(self.taskHandle, channel, data)

    def ResetCILinEncoderInitialPos(self, channel):
        return DAQmxResetCILinEncoderInitialPos(self.taskHandle, channel)

    def GetCIEncoderDecodingType(self, channel, data):
        return DAQmxGetCIEncoderDecodingType(self.taskHandle, channel, data)

    def SetCIEncoderDecodingType(self, channel, data):
        return DAQmxSetCIEncoderDecodingType(self.taskHandle, channel, data)

    def ResetCIEncoderDecodingType(self, channel):
        return DAQmxResetCIEncoderDecodingType(self.taskHandle, channel)

    def GetCIEncoderAInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderAInputTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderAInputTerm(self, channel, data):
        return DAQmxSetCIEncoderAInputTerm(self.taskHandle, channel, data)

    def ResetCIEncoderAInputTerm(self, channel):
        return DAQmxResetCIEncoderAInputTerm(self.taskHandle, channel)

    def GetCIEncoderAInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrEnable(self.taskHandle, channel, data)

    def SetCIEncoderAInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIEncoderAInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrEnable(self.taskHandle, channel)

    def GetCIEncoderAInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIEncoderAInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIEncoderAInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIEncoderAInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderAInputDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderAInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIEncoderAInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIEncoderAInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIEncoderAInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIEncoderAInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIEncoderAInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderAInputDigSyncEnable(self.taskHandle, channel, data)

    def SetCIEncoderAInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderAInputDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIEncoderAInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderAInputDigSyncEnable(self.taskHandle, channel)

    def GetCIEncoderBInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderBInputTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderBInputTerm(self, channel, data):
        return DAQmxSetCIEncoderBInputTerm(self.taskHandle, channel, data)

    def ResetCIEncoderBInputTerm(self, channel):
        return DAQmxResetCIEncoderBInputTerm(self.taskHandle, channel)

    def GetCIEncoderBInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrEnable(self.taskHandle, channel, data)

    def SetCIEncoderBInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIEncoderBInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrEnable(self.taskHandle, channel)

    def GetCIEncoderBInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIEncoderBInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIEncoderBInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIEncoderBInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderBInputDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderBInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIEncoderBInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIEncoderBInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIEncoderBInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIEncoderBInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIEncoderBInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderBInputDigSyncEnable(self.taskHandle, channel, data)

    def SetCIEncoderBInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderBInputDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIEncoderBInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderBInputDigSyncEnable(self.taskHandle, channel)

    def GetCIEncoderZInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderZInputTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderZInputTerm(self, channel, data):
        return DAQmxSetCIEncoderZInputTerm(self.taskHandle, channel, data)

    def ResetCIEncoderZInputTerm(self, channel):
        return DAQmxResetCIEncoderZInputTerm(self.taskHandle, channel)

    def GetCIEncoderZInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrEnable(self.taskHandle, channel, data)

    def SetCIEncoderZInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIEncoderZInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrEnable(self.taskHandle, channel)

    def GetCIEncoderZInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIEncoderZInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIEncoderZInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIEncoderZInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderZInputDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIEncoderZInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIEncoderZInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIEncoderZInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIEncoderZInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIEncoderZInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIEncoderZInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderZInputDigSyncEnable(self.taskHandle, channel, data)

    def SetCIEncoderZInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderZInputDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIEncoderZInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderZInputDigSyncEnable(self.taskHandle, channel)

    def GetCIEncoderZIndexEnable(self, channel, data):
        return DAQmxGetCIEncoderZIndexEnable(self.taskHandle, channel, data)

    def SetCIEncoderZIndexEnable(self, channel, data):
        return DAQmxSetCIEncoderZIndexEnable(self.taskHandle, channel, data)

    def ResetCIEncoderZIndexEnable(self, channel):
        return DAQmxResetCIEncoderZIndexEnable(self.taskHandle, channel)

    def GetCIEncoderZIndexVal(self, channel, data):
        return DAQmxGetCIEncoderZIndexVal(self.taskHandle, channel, data)

    def SetCIEncoderZIndexVal(self, channel, data):
        return DAQmxSetCIEncoderZIndexVal(self.taskHandle, channel, data)

    def ResetCIEncoderZIndexVal(self, channel):
        return DAQmxResetCIEncoderZIndexVal(self.taskHandle, channel)

    def GetCIEncoderZIndexPhase(self, channel, data):
        return DAQmxGetCIEncoderZIndexPhase(self.taskHandle, channel, data)

    def SetCIEncoderZIndexPhase(self, channel, data):
        return DAQmxSetCIEncoderZIndexPhase(self.taskHandle, channel, data)

    def ResetCIEncoderZIndexPhase(self, channel):
        return DAQmxResetCIEncoderZIndexPhase(self.taskHandle, channel)

    def GetCIPulseWidthUnits(self, channel, data):
        return DAQmxGetCIPulseWidthUnits(self.taskHandle, channel, data)

    def SetCIPulseWidthUnits(self, channel, data):
        return DAQmxSetCIPulseWidthUnits(self.taskHandle, channel, data)

    def ResetCIPulseWidthUnits(self, channel):
        return DAQmxResetCIPulseWidthUnits(self.taskHandle, channel)

    def GetCIPulseWidthTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseWidthTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseWidthTerm(self, channel, data):
        return DAQmxSetCIPulseWidthTerm(self.taskHandle, channel, data)

    def ResetCIPulseWidthTerm(self, channel):
        return DAQmxResetCIPulseWidthTerm(self.taskHandle, channel)

    def GetCIPulseWidthStartingEdge(self, channel, data):
        return DAQmxGetCIPulseWidthStartingEdge(self.taskHandle, channel, data)

    def SetCIPulseWidthStartingEdge(self, channel, data):
        return DAQmxSetCIPulseWidthStartingEdge(self.taskHandle, channel, data)

    def ResetCIPulseWidthStartingEdge(self, channel):
        return DAQmxResetCIPulseWidthStartingEdge(self.taskHandle, channel)

    def GetCIPulseWidthDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrEnable(self.taskHandle, channel, data)

    def SetCIPulseWidthDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIPulseWidthDigFltrEnable(self, channel):
        return DAQmxResetCIPulseWidthDigFltrEnable(self.taskHandle, channel)

    def GetCIPulseWidthDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIPulseWidthDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIPulseWidthDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseWidthDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIPulseWidthDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseWidthDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseWidthDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIPulseWidthDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseWidthDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIPulseWidthDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIPulseWidthDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIPulseWidthDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseWidthDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIPulseWidthDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseWidthDigSyncEnable(self.taskHandle, channel, data)

    def SetCIPulseWidthDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseWidthDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIPulseWidthDigSyncEnable(self, channel):
        return DAQmxResetCIPulseWidthDigSyncEnable(self.taskHandle, channel)

    def GetCITimestampUnits(self, channel, data):
        return DAQmxGetCITimestampUnits(self.taskHandle, channel, data)

    def SetCITimestampUnits(self, channel, data):
        return DAQmxSetCITimestampUnits(self.taskHandle, channel, data)

    def ResetCITimestampUnits(self, channel):
        return DAQmxResetCITimestampUnits(self.taskHandle, channel)

    def GetCITwoEdgeSepUnits(self, channel, data):
        return DAQmxGetCITwoEdgeSepUnits(self.taskHandle, channel, data)

    def SetCITwoEdgeSepUnits(self, channel, data):
        return DAQmxSetCITwoEdgeSepUnits(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepUnits(self, channel):
        return DAQmxResetCITwoEdgeSepUnits(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstTerm(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepFirstTerm(self.taskHandle, channel, data, bufferSize)

    def SetCITwoEdgeSepFirstTerm(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstTerm(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstTerm(self, channel):
        return DAQmxResetCITwoEdgeSepFirstTerm(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstEdge(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstEdge(self.taskHandle, channel, data)

    def SetCITwoEdgeSepFirstEdge(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstEdge(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstEdge(self, channel):
        return DAQmxResetCITwoEdgeSepFirstEdge(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstDigFltrEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrEnable(self.taskHandle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrEnable(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrEnable(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrEnable(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCITwoEdgeSepFirstDigSyncEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigSyncEnable(self.taskHandle, channel, data)

    def SetCITwoEdgeSepFirstDigSyncEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigSyncEnable(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepFirstDigSyncEnable(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigSyncEnable(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondTerm(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepSecondTerm(self.taskHandle, channel, data, bufferSize)

    def SetCITwoEdgeSepSecondTerm(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondTerm(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondTerm(self, channel):
        return DAQmxResetCITwoEdgeSepSecondTerm(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondEdge(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondEdge(self.taskHandle, channel, data)

    def SetCITwoEdgeSepSecondEdge(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondEdge(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondEdge(self, channel):
        return DAQmxResetCITwoEdgeSepSecondEdge(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondDigFltrEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrEnable(self.taskHandle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrEnable(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrEnable(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrEnable(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCITwoEdgeSepSecondDigSyncEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigSyncEnable(self.taskHandle, channel, data)

    def SetCITwoEdgeSepSecondDigSyncEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigSyncEnable(self.taskHandle, channel, data)

    def ResetCITwoEdgeSepSecondDigSyncEnable(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigSyncEnable(self.taskHandle, channel)

    def GetCISemiPeriodUnits(self, channel, data):
        return DAQmxGetCISemiPeriodUnits(self.taskHandle, channel, data)

    def SetCISemiPeriodUnits(self, channel, data):
        return DAQmxSetCISemiPeriodUnits(self.taskHandle, channel, data)

    def ResetCISemiPeriodUnits(self, channel):
        return DAQmxResetCISemiPeriodUnits(self.taskHandle, channel)

    def GetCISemiPeriodTerm(self, channel, data, bufferSize):
        return DAQmxGetCISemiPeriodTerm(self.taskHandle, channel, data, bufferSize)

    def SetCISemiPeriodTerm(self, channel, data):
        return DAQmxSetCISemiPeriodTerm(self.taskHandle, channel, data)

    def ResetCISemiPeriodTerm(self, channel):
        return DAQmxResetCISemiPeriodTerm(self.taskHandle, channel)

    def GetCISemiPeriodStartingEdge(self, channel, data):
        return DAQmxGetCISemiPeriodStartingEdge(self.taskHandle, channel, data)

    def SetCISemiPeriodStartingEdge(self, channel, data):
        return DAQmxSetCISemiPeriodStartingEdge(self.taskHandle, channel, data)

    def ResetCISemiPeriodStartingEdge(self, channel):
        return DAQmxResetCISemiPeriodStartingEdge(self.taskHandle, channel)

    def GetCISemiPeriodDigFltrEnable(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrEnable(self.taskHandle, channel, data)

    def SetCISemiPeriodDigFltrEnable(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrEnable(self.taskHandle, channel, data)

    def ResetCISemiPeriodDigFltrEnable(self, channel):
        return DAQmxResetCISemiPeriodDigFltrEnable(self.taskHandle, channel)

    def GetCISemiPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCISemiPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCISemiPeriodDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCISemiPeriodDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCISemiPeriodDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCISemiPeriodDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCISemiPeriodDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCISemiPeriodDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCISemiPeriodDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCISemiPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCISemiPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCISemiPeriodDigFltrTimebaseRate(self, channel):
        return DAQmxResetCISemiPeriodDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCISemiPeriodDigSyncEnable(self, channel, data):
        return DAQmxGetCISemiPeriodDigSyncEnable(self.taskHandle, channel, data)

    def SetCISemiPeriodDigSyncEnable(self, channel, data):
        return DAQmxSetCISemiPeriodDigSyncEnable(self.taskHandle, channel, data)

    def ResetCISemiPeriodDigSyncEnable(self, channel):
        return DAQmxResetCISemiPeriodDigSyncEnable(self.taskHandle, channel)

    def GetCIPulseFreqUnits(self, channel, data):
        return DAQmxGetCIPulseFreqUnits(self.taskHandle, channel, data)

    def SetCIPulseFreqUnits(self, channel, data):
        return DAQmxSetCIPulseFreqUnits(self.taskHandle, channel, data)

    def ResetCIPulseFreqUnits(self, channel):
        return DAQmxResetCIPulseFreqUnits(self.taskHandle, channel)

    def GetCIPulseFreqTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseFreqTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseFreqTerm(self, channel, data):
        return DAQmxSetCIPulseFreqTerm(self.taskHandle, channel, data)

    def ResetCIPulseFreqTerm(self, channel):
        return DAQmxResetCIPulseFreqTerm(self.taskHandle, channel)

    def GetCIPulseFreqStartEdge(self, channel, data):
        return DAQmxGetCIPulseFreqStartEdge(self.taskHandle, channel, data)

    def SetCIPulseFreqStartEdge(self, channel, data):
        return DAQmxSetCIPulseFreqStartEdge(self.taskHandle, channel, data)

    def ResetCIPulseFreqStartEdge(self, channel):
        return DAQmxResetCIPulseFreqStartEdge(self.taskHandle, channel)

    def GetCIPulseFreqDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrEnable(self.taskHandle, channel, data)

    def SetCIPulseFreqDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIPulseFreqDigFltrEnable(self, channel):
        return DAQmxResetCIPulseFreqDigFltrEnable(self.taskHandle, channel)

    def GetCIPulseFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIPulseFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIPulseFreqDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseFreqDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIPulseFreqDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseFreqDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseFreqDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIPulseFreqDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseFreqDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIPulseFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIPulseFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIPulseFreqDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseFreqDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIPulseFreqDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseFreqDigSyncEnable(self.taskHandle, channel, data)

    def SetCIPulseFreqDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseFreqDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIPulseFreqDigSyncEnable(self, channel):
        return DAQmxResetCIPulseFreqDigSyncEnable(self.taskHandle, channel)

    def GetCIPulseTimeUnits(self, channel, data):
        return DAQmxGetCIPulseTimeUnits(self.taskHandle, channel, data)

    def SetCIPulseTimeUnits(self, channel, data):
        return DAQmxSetCIPulseTimeUnits(self.taskHandle, channel, data)

    def ResetCIPulseTimeUnits(self, channel):
        return DAQmxResetCIPulseTimeUnits(self.taskHandle, channel)

    def GetCIPulseTimeTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTimeTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseTimeTerm(self, channel, data):
        return DAQmxSetCIPulseTimeTerm(self.taskHandle, channel, data)

    def ResetCIPulseTimeTerm(self, channel):
        return DAQmxResetCIPulseTimeTerm(self.taskHandle, channel)

    def GetCIPulseTimeStartEdge(self, channel, data):
        return DAQmxGetCIPulseTimeStartEdge(self.taskHandle, channel, data)

    def SetCIPulseTimeStartEdge(self, channel, data):
        return DAQmxSetCIPulseTimeStartEdge(self.taskHandle, channel, data)

    def ResetCIPulseTimeStartEdge(self, channel):
        return DAQmxResetCIPulseTimeStartEdge(self.taskHandle, channel)

    def GetCIPulseTimeDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrEnable(self.taskHandle, channel, data)

    def SetCIPulseTimeDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIPulseTimeDigFltrEnable(self, channel):
        return DAQmxResetCIPulseTimeDigFltrEnable(self.taskHandle, channel)

    def GetCIPulseTimeDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIPulseTimeDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIPulseTimeDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseTimeDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIPulseTimeDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTimeDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseTimeDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIPulseTimeDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseTimeDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIPulseTimeDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIPulseTimeDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIPulseTimeDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseTimeDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIPulseTimeDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseTimeDigSyncEnable(self.taskHandle, channel, data)

    def SetCIPulseTimeDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseTimeDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIPulseTimeDigSyncEnable(self, channel):
        return DAQmxResetCIPulseTimeDigSyncEnable(self.taskHandle, channel)

    def GetCIPulseTicksTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTicksTerm(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseTicksTerm(self, channel, data):
        return DAQmxSetCIPulseTicksTerm(self.taskHandle, channel, data)

    def ResetCIPulseTicksTerm(self, channel):
        return DAQmxResetCIPulseTicksTerm(self.taskHandle, channel)

    def GetCIPulseTicksStartEdge(self, channel, data):
        return DAQmxGetCIPulseTicksStartEdge(self.taskHandle, channel, data)

    def SetCIPulseTicksStartEdge(self, channel, data):
        return DAQmxSetCIPulseTicksStartEdge(self.taskHandle, channel, data)

    def ResetCIPulseTicksStartEdge(self, channel):
        return DAQmxResetCIPulseTicksStartEdge(self.taskHandle, channel)

    def GetCIPulseTicksDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrEnable(self.taskHandle, channel, data)

    def SetCIPulseTicksDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrEnable(self.taskHandle, channel, data)

    def ResetCIPulseTicksDigFltrEnable(self, channel):
        return DAQmxResetCIPulseTicksDigFltrEnable(self.taskHandle, channel)

    def GetCIPulseTicksDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCIPulseTicksDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCIPulseTicksDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseTicksDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCIPulseTicksDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTicksDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIPulseTicksDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCIPulseTicksDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseTicksDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCIPulseTicksDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCIPulseTicksDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCIPulseTicksDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseTicksDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCIPulseTicksDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseTicksDigSyncEnable(self.taskHandle, channel, data)

    def SetCIPulseTicksDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseTicksDigSyncEnable(self.taskHandle, channel, data)

    def ResetCIPulseTicksDigSyncEnable(self, channel):
        return DAQmxResetCIPulseTicksDigSyncEnable(self.taskHandle, channel)

    def GetCITimestampInitialSeconds(self, channel, data):
        return DAQmxGetCITimestampInitialSeconds(self.taskHandle, channel, data)

    def SetCITimestampInitialSeconds(self, channel, data):
        return DAQmxSetCITimestampInitialSeconds(self.taskHandle, channel, data)

    def ResetCITimestampInitialSeconds(self, channel):
        return DAQmxResetCITimestampInitialSeconds(self.taskHandle, channel)

    def GetCIGPSSyncMethod(self, channel, data):
        return DAQmxGetCIGPSSyncMethod(self.taskHandle, channel, data)

    def SetCIGPSSyncMethod(self, channel, data):
        return DAQmxSetCIGPSSyncMethod(self.taskHandle, channel, data)

    def ResetCIGPSSyncMethod(self, channel):
        return DAQmxResetCIGPSSyncMethod(self.taskHandle, channel)

    def GetCIGPSSyncSrc(self, channel, data, bufferSize):
        return DAQmxGetCIGPSSyncSrc(self.taskHandle, channel, data, bufferSize)

    def SetCIGPSSyncSrc(self, channel, data):
        return DAQmxSetCIGPSSyncSrc(self.taskHandle, channel, data)

    def ResetCIGPSSyncSrc(self, channel):
        return DAQmxResetCIGPSSyncSrc(self.taskHandle, channel)

    def GetCICtrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICtrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCICtrTimebaseSrc(self, channel, data):
        return DAQmxSetCICtrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCICtrTimebaseSrc(self, channel):
        return DAQmxResetCICtrTimebaseSrc(self.taskHandle, channel)

    def GetCICtrTimebaseRate(self, channel, data):
        return DAQmxGetCICtrTimebaseRate(self.taskHandle, channel, data)

    def SetCICtrTimebaseRate(self, channel, data):
        return DAQmxSetCICtrTimebaseRate(self.taskHandle, channel, data)

    def ResetCICtrTimebaseRate(self, channel):
        return DAQmxResetCICtrTimebaseRate(self.taskHandle, channel)

    def GetCICtrTimebaseActiveEdge(self, channel, data):
        return DAQmxGetCICtrTimebaseActiveEdge(self.taskHandle, channel, data)

    def SetCICtrTimebaseActiveEdge(self, channel, data):
        return DAQmxSetCICtrTimebaseActiveEdge(self.taskHandle, channel, data)

    def ResetCICtrTimebaseActiveEdge(self, channel):
        return DAQmxResetCICtrTimebaseActiveEdge(self.taskHandle, channel)

    def GetCICtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrEnable(self.taskHandle, channel, data)

    def SetCICtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrEnable(self.taskHandle, channel, data)

    def ResetCICtrTimebaseDigFltrEnable(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrEnable(self.taskHandle, channel)

    def GetCICtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCICtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCICtrTimebaseDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCICtrTimebaseDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCICtrTimebaseDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCICtrTimebaseDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCICtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCICtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCICtrTimebaseDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCICtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxGetCICtrTimebaseDigSyncEnable(self.taskHandle, channel, data)

    def SetCICtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxSetCICtrTimebaseDigSyncEnable(self.taskHandle, channel, data)

    def ResetCICtrTimebaseDigSyncEnable(self, channel):
        return DAQmxResetCICtrTimebaseDigSyncEnable(self.taskHandle, channel)

    def GetCICount(self, channel, data):
        return DAQmxGetCICount(self.taskHandle, channel, data)

    def GetCIOutputState(self, channel, data):
        return DAQmxGetCIOutputState(self.taskHandle, channel, data)

    def GetCITCReached(self, channel, data):
        return DAQmxGetCITCReached(self.taskHandle, channel, data)

    def GetCICtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxGetCICtrTimebaseMasterTimebaseDiv(self.taskHandle, channel, data)

    def SetCICtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxSetCICtrTimebaseMasterTimebaseDiv(self.taskHandle, channel, data)

    def ResetCICtrTimebaseMasterTimebaseDiv(self, channel):
        return DAQmxResetCICtrTimebaseMasterTimebaseDiv(self.taskHandle, channel)

    def GetCIDataXferMech(self, channel, data):
        return DAQmxGetCIDataXferMech(self.taskHandle, channel, data)

    def SetCIDataXferMech(self, channel, data):
        return DAQmxSetCIDataXferMech(self.taskHandle, channel, data)

    def ResetCIDataXferMech(self, channel):
        return DAQmxResetCIDataXferMech(self.taskHandle, channel)

    def GetCIDataXferReqCond(self, channel, data):
        return DAQmxGetCIDataXferReqCond(self.taskHandle, channel, data)

    def SetCIDataXferReqCond(self, channel, data):
        return DAQmxSetCIDataXferReqCond(self.taskHandle, channel, data)

    def ResetCIDataXferReqCond(self, channel):
        return DAQmxResetCIDataXferReqCond(self.taskHandle, channel)

    def GetCIUsbXferReqSize(self, channel, data):
        return DAQmxGetCIUsbXferReqSize(self.taskHandle, channel, data)

    def SetCIUsbXferReqSize(self, channel, data):
        return DAQmxSetCIUsbXferReqSize(self.taskHandle, channel, data)

    def ResetCIUsbXferReqSize(self, channel):
        return DAQmxResetCIUsbXferReqSize(self.taskHandle, channel)

    def GetCIUsbXferReqCount(self, channel, data):
        return DAQmxGetCIUsbXferReqCount(self.taskHandle, channel, data)

    def SetCIUsbXferReqCount(self, channel, data):
        return DAQmxSetCIUsbXferReqCount(self.taskHandle, channel, data)

    def ResetCIUsbXferReqCount(self, channel):
        return DAQmxResetCIUsbXferReqCount(self.taskHandle, channel)

    def GetCIMemMapEnable(self, channel, data):
        return DAQmxGetCIMemMapEnable(self.taskHandle, channel, data)

    def SetCIMemMapEnable(self, channel, data):
        return DAQmxSetCIMemMapEnable(self.taskHandle, channel, data)

    def ResetCIMemMapEnable(self, channel):
        return DAQmxResetCIMemMapEnable(self.taskHandle, channel)

    def GetCINumPossiblyInvalidSamps(self, channel, data):
        return DAQmxGetCINumPossiblyInvalidSamps(self.taskHandle, channel, data)

    def GetCIDupCountPrevent(self, channel, data):
        return DAQmxGetCIDupCountPrevent(self.taskHandle, channel, data)

    def SetCIDupCountPrevent(self, channel, data):
        return DAQmxSetCIDupCountPrevent(self.taskHandle, channel, data)

    def ResetCIDupCountPrevent(self, channel):
        return DAQmxResetCIDupCountPrevent(self.taskHandle, channel)

    def GetCIPrescaler(self, channel, data):
        return DAQmxGetCIPrescaler(self.taskHandle, channel, data)

    def SetCIPrescaler(self, channel, data):
        return DAQmxSetCIPrescaler(self.taskHandle, channel, data)

    def ResetCIPrescaler(self, channel):
        return DAQmxResetCIPrescaler(self.taskHandle, channel)

    def GetCOOutputType(self, channel, data):
        return DAQmxGetCOOutputType(self.taskHandle, channel, data)

    def GetCOPulseIdleState(self, channel, data):
        return DAQmxGetCOPulseIdleState(self.taskHandle, channel, data)

    def SetCOPulseIdleState(self, channel, data):
        return DAQmxSetCOPulseIdleState(self.taskHandle, channel, data)

    def ResetCOPulseIdleState(self, channel):
        return DAQmxResetCOPulseIdleState(self.taskHandle, channel)

    def GetCOPulseTerm(self, channel, data, bufferSize):
        return DAQmxGetCOPulseTerm(self.taskHandle, channel, data, bufferSize)

    def SetCOPulseTerm(self, channel, data):
        return DAQmxSetCOPulseTerm(self.taskHandle, channel, data)

    def ResetCOPulseTerm(self, channel):
        return DAQmxResetCOPulseTerm(self.taskHandle, channel)

    def GetCOPulseTimeUnits(self, channel, data):
        return DAQmxGetCOPulseTimeUnits(self.taskHandle, channel, data)

    def SetCOPulseTimeUnits(self, channel, data):
        return DAQmxSetCOPulseTimeUnits(self.taskHandle, channel, data)

    def ResetCOPulseTimeUnits(self, channel):
        return DAQmxResetCOPulseTimeUnits(self.taskHandle, channel)

    def GetCOPulseHighTime(self, channel, data):
        return DAQmxGetCOPulseHighTime(self.taskHandle, channel, data)

    def SetCOPulseHighTime(self, channel, data):
        return DAQmxSetCOPulseHighTime(self.taskHandle, channel, data)

    def ResetCOPulseHighTime(self, channel):
        return DAQmxResetCOPulseHighTime(self.taskHandle, channel)

    def GetCOPulseLowTime(self, channel, data):
        return DAQmxGetCOPulseLowTime(self.taskHandle, channel, data)

    def SetCOPulseLowTime(self, channel, data):
        return DAQmxSetCOPulseLowTime(self.taskHandle, channel, data)

    def ResetCOPulseLowTime(self, channel):
        return DAQmxResetCOPulseLowTime(self.taskHandle, channel)

    def GetCOPulseTimeInitialDelay(self, channel, data):
        return DAQmxGetCOPulseTimeInitialDelay(self.taskHandle, channel, data)

    def SetCOPulseTimeInitialDelay(self, channel, data):
        return DAQmxSetCOPulseTimeInitialDelay(self.taskHandle, channel, data)

    def ResetCOPulseTimeInitialDelay(self, channel):
        return DAQmxResetCOPulseTimeInitialDelay(self.taskHandle, channel)

    def GetCOPulseDutyCyc(self, channel, data):
        return DAQmxGetCOPulseDutyCyc(self.taskHandle, channel, data)

    def SetCOPulseDutyCyc(self, channel, data):
        return DAQmxSetCOPulseDutyCyc(self.taskHandle, channel, data)

    def ResetCOPulseDutyCyc(self, channel):
        return DAQmxResetCOPulseDutyCyc(self.taskHandle, channel)

    def GetCOPulseFreqUnits(self, channel, data):
        return DAQmxGetCOPulseFreqUnits(self.taskHandle, channel, data)

    def SetCOPulseFreqUnits(self, channel, data):
        return DAQmxSetCOPulseFreqUnits(self.taskHandle, channel, data)

    def ResetCOPulseFreqUnits(self, channel):
        return DAQmxResetCOPulseFreqUnits(self.taskHandle, channel)

    def GetCOPulseFreq(self, channel, data):
        return DAQmxGetCOPulseFreq(self.taskHandle, channel, data)

    def SetCOPulseFreq(self, channel, data):
        return DAQmxSetCOPulseFreq(self.taskHandle, channel, data)

    def ResetCOPulseFreq(self, channel):
        return DAQmxResetCOPulseFreq(self.taskHandle, channel)

    def GetCOPulseFreqInitialDelay(self, channel, data):
        return DAQmxGetCOPulseFreqInitialDelay(self.taskHandle, channel, data)

    def SetCOPulseFreqInitialDelay(self, channel, data):
        return DAQmxSetCOPulseFreqInitialDelay(self.taskHandle, channel, data)

    def ResetCOPulseFreqInitialDelay(self, channel):
        return DAQmxResetCOPulseFreqInitialDelay(self.taskHandle, channel)

    def GetCOPulseHighTicks(self, channel, data):
        return DAQmxGetCOPulseHighTicks(self.taskHandle, channel, data)

    def SetCOPulseHighTicks(self, channel, data):
        return DAQmxSetCOPulseHighTicks(self.taskHandle, channel, data)

    def ResetCOPulseHighTicks(self, channel):
        return DAQmxResetCOPulseHighTicks(self.taskHandle, channel)

    def GetCOPulseLowTicks(self, channel, data):
        return DAQmxGetCOPulseLowTicks(self.taskHandle, channel, data)

    def SetCOPulseLowTicks(self, channel, data):
        return DAQmxSetCOPulseLowTicks(self.taskHandle, channel, data)

    def ResetCOPulseLowTicks(self, channel):
        return DAQmxResetCOPulseLowTicks(self.taskHandle, channel)

    def GetCOPulseTicksInitialDelay(self, channel, data):
        return DAQmxGetCOPulseTicksInitialDelay(self.taskHandle, channel, data)

    def SetCOPulseTicksInitialDelay(self, channel, data):
        return DAQmxSetCOPulseTicksInitialDelay(self.taskHandle, channel, data)

    def ResetCOPulseTicksInitialDelay(self, channel):
        return DAQmxResetCOPulseTicksInitialDelay(self.taskHandle, channel)

    def GetCOCtrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCOCtrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCOCtrTimebaseSrc(self, channel, data):
        return DAQmxSetCOCtrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseSrc(self, channel):
        return DAQmxResetCOCtrTimebaseSrc(self.taskHandle, channel)

    def GetCOCtrTimebaseRate(self, channel, data):
        return DAQmxGetCOCtrTimebaseRate(self.taskHandle, channel, data)

    def SetCOCtrTimebaseRate(self, channel, data):
        return DAQmxSetCOCtrTimebaseRate(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseRate(self, channel):
        return DAQmxResetCOCtrTimebaseRate(self.taskHandle, channel)

    def GetCOCtrTimebaseActiveEdge(self, channel, data):
        return DAQmxGetCOCtrTimebaseActiveEdge(self.taskHandle, channel, data)

    def SetCOCtrTimebaseActiveEdge(self, channel, data):
        return DAQmxSetCOCtrTimebaseActiveEdge(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseActiveEdge(self, channel):
        return DAQmxResetCOCtrTimebaseActiveEdge(self.taskHandle, channel)

    def GetCOCtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrEnable(self.taskHandle, channel, data)

    def SetCOCtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrEnable(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseDigFltrEnable(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrEnable(self.taskHandle, channel)

    def GetCOCtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def SetCOCtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth(self.taskHandle, channel)

    def GetCOCtrTimebaseDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel, data, bufferSize)

    def SetCOCtrTimebaseDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc(self.taskHandle, channel)

    def GetCOCtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel, data)

    def SetCOCtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseDigFltrTimebaseRate(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrTimebaseRate(self.taskHandle, channel)

    def GetCOCtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigSyncEnable(self.taskHandle, channel, data)

    def SetCOCtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigSyncEnable(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseDigSyncEnable(self, channel):
        return DAQmxResetCOCtrTimebaseDigSyncEnable(self.taskHandle, channel)

    def GetCOCount(self, channel, data):
        return DAQmxGetCOCount(self.taskHandle, channel, data)

    def GetCOOutputState(self, channel, data):
        return DAQmxGetCOOutputState(self.taskHandle, channel, data)

    def GetCOAutoIncrCnt(self, channel, data):
        return DAQmxGetCOAutoIncrCnt(self.taskHandle, channel, data)

    def SetCOAutoIncrCnt(self, channel, data):
        return DAQmxSetCOAutoIncrCnt(self.taskHandle, channel, data)

    def ResetCOAutoIncrCnt(self, channel):
        return DAQmxResetCOAutoIncrCnt(self.taskHandle, channel)

    def GetCOCtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxGetCOCtrTimebaseMasterTimebaseDiv(self.taskHandle, channel, data)

    def SetCOCtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxSetCOCtrTimebaseMasterTimebaseDiv(self.taskHandle, channel, data)

    def ResetCOCtrTimebaseMasterTimebaseDiv(self, channel):
        return DAQmxResetCOCtrTimebaseMasterTimebaseDiv(self.taskHandle, channel)

    def GetCOPulseDone(self, channel, data):
        return DAQmxGetCOPulseDone(self.taskHandle, channel, data)

    def GetCOEnableInitialDelayOnRetrigger(self, channel, data):
        return DAQmxGetCOEnableInitialDelayOnRetrigger(self.taskHandle, channel, data)

    def SetCOEnableInitialDelayOnRetrigger(self, channel, data):
        return DAQmxSetCOEnableInitialDelayOnRetrigger(self.taskHandle, channel, data)

    def ResetCOEnableInitialDelayOnRetrigger(self, channel):
        return DAQmxResetCOEnableInitialDelayOnRetrigger(self.taskHandle, channel)

    def GetCOConstrainedGenMode(self, channel, data):
        return DAQmxGetCOConstrainedGenMode(self.taskHandle, channel, data)

    def SetCOConstrainedGenMode(self, channel, data):
        return DAQmxSetCOConstrainedGenMode(self.taskHandle, channel, data)

    def ResetCOConstrainedGenMode(self, channel):
        return DAQmxResetCOConstrainedGenMode(self.taskHandle, channel)

    def GetCOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetCOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def SetCOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetCOUseOnlyOnBrdMem(self.taskHandle, channel, data)

    def ResetCOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetCOUseOnlyOnBrdMem(self.taskHandle, channel)

    def GetCODataXferMech(self, channel, data):
        return DAQmxGetCODataXferMech(self.taskHandle, channel, data)

    def SetCODataXferMech(self, channel, data):
        return DAQmxSetCODataXferMech(self.taskHandle, channel, data)

    def ResetCODataXferMech(self, channel):
        return DAQmxResetCODataXferMech(self.taskHandle, channel)

    def GetCODataXferReqCond(self, channel, data):
        return DAQmxGetCODataXferReqCond(self.taskHandle, channel, data)

    def SetCODataXferReqCond(self, channel, data):
        return DAQmxSetCODataXferReqCond(self.taskHandle, channel, data)

    def ResetCODataXferReqCond(self, channel):
        return DAQmxResetCODataXferReqCond(self.taskHandle, channel)

    def GetCOUsbXferReqSize(self, channel, data):
        return DAQmxGetCOUsbXferReqSize(self.taskHandle, channel, data)

    def SetCOUsbXferReqSize(self, channel, data):
        return DAQmxSetCOUsbXferReqSize(self.taskHandle, channel, data)

    def ResetCOUsbXferReqSize(self, channel):
        return DAQmxResetCOUsbXferReqSize(self.taskHandle, channel)

    def GetCOUsbXferReqCount(self, channel, data):
        return DAQmxGetCOUsbXferReqCount(self.taskHandle, channel, data)

    def SetCOUsbXferReqCount(self, channel, data):
        return DAQmxSetCOUsbXferReqCount(self.taskHandle, channel, data)

    def ResetCOUsbXferReqCount(self, channel):
        return DAQmxResetCOUsbXferReqCount(self.taskHandle, channel)

    def GetCOMemMapEnable(self, channel, data):
        return DAQmxGetCOMemMapEnable(self.taskHandle, channel, data)

    def SetCOMemMapEnable(self, channel, data):
        return DAQmxSetCOMemMapEnable(self.taskHandle, channel, data)

    def ResetCOMemMapEnable(self, channel):
        return DAQmxResetCOMemMapEnable(self.taskHandle, channel)

    def GetCOPrescaler(self, channel, data):
        return DAQmxGetCOPrescaler(self.taskHandle, channel, data)

    def SetCOPrescaler(self, channel, data):
        return DAQmxSetCOPrescaler(self.taskHandle, channel, data)

    def ResetCOPrescaler(self, channel):
        return DAQmxResetCOPrescaler(self.taskHandle, channel)

    def GetCORdyForNewVal(self, channel, data):
        return DAQmxGetCORdyForNewVal(self.taskHandle, channel, data)

    def GetChanType(self, channel, data):
        return DAQmxGetChanType(self.taskHandle, channel, data)

    def GetPhysicalChanName(self, channel, data, bufferSize):
        return DAQmxGetPhysicalChanName(self.taskHandle, channel, data, bufferSize)

    def SetPhysicalChanName(self, channel, data):
        return DAQmxSetPhysicalChanName(self.taskHandle, channel, data)

    def GetChanDescr(self, channel, data, bufferSize):
        return DAQmxGetChanDescr(self.taskHandle, channel, data, bufferSize)

    def SetChanDescr(self, channel, data):
        return DAQmxSetChanDescr(self.taskHandle, channel, data)

    def ResetChanDescr(self, channel):
        return DAQmxResetChanDescr(self.taskHandle, channel)

    def GetChanIsGlobal(self, channel, data):
        return DAQmxGetChanIsGlobal(self.taskHandle, channel, data)

    def GetExportedAIConvClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAIConvClkOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedAIConvClkOutputTerm(self, data):
        return DAQmxSetExportedAIConvClkOutputTerm(self.taskHandle, data)

    def ResetExportedAIConvClkOutputTerm(self):
        return DAQmxResetExportedAIConvClkOutputTerm(self.taskHandle)

    def GetExportedAIConvClkPulsePolarity(self, data):
        return DAQmxGetExportedAIConvClkPulsePolarity(self.taskHandle, data)

    def GetExported10MHzRefClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExported10MHzRefClkOutputTerm(self.taskHandle, data, bufferSize)

    def SetExported10MHzRefClkOutputTerm(self, data):
        return DAQmxSetExported10MHzRefClkOutputTerm(self.taskHandle, data)

    def ResetExported10MHzRefClkOutputTerm(self):
        return DAQmxResetExported10MHzRefClkOutputTerm(self.taskHandle)

    def GetExported20MHzTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExported20MHzTimebaseOutputTerm(self.taskHandle, data, bufferSize)

    def SetExported20MHzTimebaseOutputTerm(self, data):
        return DAQmxSetExported20MHzTimebaseOutputTerm(self.taskHandle, data)

    def ResetExported20MHzTimebaseOutputTerm(self):
        return DAQmxResetExported20MHzTimebaseOutputTerm(self.taskHandle)

    def GetExportedSampClkOutputBehavior(self, data):
        return DAQmxGetExportedSampClkOutputBehavior(self.taskHandle, data)

    def SetExportedSampClkOutputBehavior(self, data):
        return DAQmxSetExportedSampClkOutputBehavior(self.taskHandle, data)

    def ResetExportedSampClkOutputBehavior(self):
        return DAQmxResetExportedSampClkOutputBehavior(self.taskHandle)

    def GetExportedSampClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSampClkOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedSampClkOutputTerm(self, data):
        return DAQmxSetExportedSampClkOutputTerm(self.taskHandle, data)

    def ResetExportedSampClkOutputTerm(self):
        return DAQmxResetExportedSampClkOutputTerm(self.taskHandle)

    def GetExportedSampClkDelayOffset(self, data):
        return DAQmxGetExportedSampClkDelayOffset(self.taskHandle, data)

    def SetExportedSampClkDelayOffset(self, data):
        return DAQmxSetExportedSampClkDelayOffset(self.taskHandle, data)

    def ResetExportedSampClkDelayOffset(self):
        return DAQmxResetExportedSampClkDelayOffset(self.taskHandle)

    def GetExportedSampClkPulsePolarity(self, data):
        return DAQmxGetExportedSampClkPulsePolarity(self.taskHandle, data)

    def SetExportedSampClkPulsePolarity(self, data):
        return DAQmxSetExportedSampClkPulsePolarity(self.taskHandle, data)

    def ResetExportedSampClkPulsePolarity(self):
        return DAQmxResetExportedSampClkPulsePolarity(self.taskHandle)

    def GetExportedSampClkTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSampClkTimebaseOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedSampClkTimebaseOutputTerm(self, data):
        return DAQmxSetExportedSampClkTimebaseOutputTerm(self.taskHandle, data)

    def ResetExportedSampClkTimebaseOutputTerm(self):
        return DAQmxResetExportedSampClkTimebaseOutputTerm(self.taskHandle)

    def GetExportedDividedSampClkTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedDividedSampClkTimebaseOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedDividedSampClkTimebaseOutputTerm(self, data):
        return DAQmxSetExportedDividedSampClkTimebaseOutputTerm(self.taskHandle, data)

    def ResetExportedDividedSampClkTimebaseOutputTerm(self):
        return DAQmxResetExportedDividedSampClkTimebaseOutputTerm(self.taskHandle)

    def GetExportedAdvTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAdvTrigOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedAdvTrigOutputTerm(self, data):
        return DAQmxSetExportedAdvTrigOutputTerm(self.taskHandle, data)

    def ResetExportedAdvTrigOutputTerm(self):
        return DAQmxResetExportedAdvTrigOutputTerm(self.taskHandle)

    def GetExportedAdvTrigPulsePolarity(self, data):
        return DAQmxGetExportedAdvTrigPulsePolarity(self.taskHandle, data)

    def GetExportedAdvTrigPulseWidthUnits(self, data):
        return DAQmxGetExportedAdvTrigPulseWidthUnits(self.taskHandle, data)

    def SetExportedAdvTrigPulseWidthUnits(self, data):
        return DAQmxSetExportedAdvTrigPulseWidthUnits(self.taskHandle, data)

    def ResetExportedAdvTrigPulseWidthUnits(self):
        return DAQmxResetExportedAdvTrigPulseWidthUnits(self.taskHandle)

    def GetExportedAdvTrigPulseWidth(self, data):
        return DAQmxGetExportedAdvTrigPulseWidth(self.taskHandle, data)

    def SetExportedAdvTrigPulseWidth(self, data):
        return DAQmxSetExportedAdvTrigPulseWidth(self.taskHandle, data)

    def ResetExportedAdvTrigPulseWidth(self):
        return DAQmxResetExportedAdvTrigPulseWidth(self.taskHandle)

    def GetExportedPauseTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedPauseTrigOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedPauseTrigOutputTerm(self, data):
        return DAQmxSetExportedPauseTrigOutputTerm(self.taskHandle, data)

    def ResetExportedPauseTrigOutputTerm(self):
        return DAQmxResetExportedPauseTrigOutputTerm(self.taskHandle)

    def GetExportedPauseTrigLvlActiveLvl(self, data):
        return DAQmxGetExportedPauseTrigLvlActiveLvl(self.taskHandle, data)

    def SetExportedPauseTrigLvlActiveLvl(self, data):
        return DAQmxSetExportedPauseTrigLvlActiveLvl(self.taskHandle, data)

    def ResetExportedPauseTrigLvlActiveLvl(self):
        return DAQmxResetExportedPauseTrigLvlActiveLvl(self.taskHandle)

    def GetExportedRefTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRefTrigOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedRefTrigOutputTerm(self, data):
        return DAQmxSetExportedRefTrigOutputTerm(self.taskHandle, data)

    def ResetExportedRefTrigOutputTerm(self):
        return DAQmxResetExportedRefTrigOutputTerm(self.taskHandle)

    def GetExportedRefTrigPulsePolarity(self, data):
        return DAQmxGetExportedRefTrigPulsePolarity(self.taskHandle, data)

    def SetExportedRefTrigPulsePolarity(self, data):
        return DAQmxSetExportedRefTrigPulsePolarity(self.taskHandle, data)

    def ResetExportedRefTrigPulsePolarity(self):
        return DAQmxResetExportedRefTrigPulsePolarity(self.taskHandle)

    def GetExportedStartTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedStartTrigOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedStartTrigOutputTerm(self, data):
        return DAQmxSetExportedStartTrigOutputTerm(self.taskHandle, data)

    def ResetExportedStartTrigOutputTerm(self):
        return DAQmxResetExportedStartTrigOutputTerm(self.taskHandle)

    def GetExportedStartTrigPulsePolarity(self, data):
        return DAQmxGetExportedStartTrigPulsePolarity(self.taskHandle, data)

    def SetExportedStartTrigPulsePolarity(self, data):
        return DAQmxSetExportedStartTrigPulsePolarity(self.taskHandle, data)

    def ResetExportedStartTrigPulsePolarity(self):
        return DAQmxResetExportedStartTrigPulsePolarity(self.taskHandle)

    def GetExportedAdvCmpltEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAdvCmpltEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedAdvCmpltEventOutputTerm(self, data):
        return DAQmxSetExportedAdvCmpltEventOutputTerm(self.taskHandle, data)

    def ResetExportedAdvCmpltEventOutputTerm(self):
        return DAQmxResetExportedAdvCmpltEventOutputTerm(self.taskHandle)

    def GetExportedAdvCmpltEventDelay(self, data):
        return DAQmxGetExportedAdvCmpltEventDelay(self.taskHandle, data)

    def SetExportedAdvCmpltEventDelay(self, data):
        return DAQmxSetExportedAdvCmpltEventDelay(self.taskHandle, data)

    def ResetExportedAdvCmpltEventDelay(self):
        return DAQmxResetExportedAdvCmpltEventDelay(self.taskHandle)

    def GetExportedAdvCmpltEventPulsePolarity(self, data):
        return DAQmxGetExportedAdvCmpltEventPulsePolarity(self.taskHandle, data)

    def SetExportedAdvCmpltEventPulsePolarity(self, data):
        return DAQmxSetExportedAdvCmpltEventPulsePolarity(self.taskHandle, data)

    def ResetExportedAdvCmpltEventPulsePolarity(self):
        return DAQmxResetExportedAdvCmpltEventPulsePolarity(self.taskHandle)

    def GetExportedAdvCmpltEventPulseWidth(self, data):
        return DAQmxGetExportedAdvCmpltEventPulseWidth(self.taskHandle, data)

    def SetExportedAdvCmpltEventPulseWidth(self, data):
        return DAQmxSetExportedAdvCmpltEventPulseWidth(self.taskHandle, data)

    def ResetExportedAdvCmpltEventPulseWidth(self):
        return DAQmxResetExportedAdvCmpltEventPulseWidth(self.taskHandle)

    def GetExportedAIHoldCmpltEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAIHoldCmpltEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedAIHoldCmpltEventOutputTerm(self, data):
        return DAQmxSetExportedAIHoldCmpltEventOutputTerm(self.taskHandle, data)

    def ResetExportedAIHoldCmpltEventOutputTerm(self):
        return DAQmxResetExportedAIHoldCmpltEventOutputTerm(self.taskHandle)

    def GetExportedAIHoldCmpltEventPulsePolarity(self, data):
        return DAQmxGetExportedAIHoldCmpltEventPulsePolarity(self.taskHandle, data)

    def SetExportedAIHoldCmpltEventPulsePolarity(self, data):
        return DAQmxSetExportedAIHoldCmpltEventPulsePolarity(self.taskHandle, data)

    def ResetExportedAIHoldCmpltEventPulsePolarity(self):
        return DAQmxResetExportedAIHoldCmpltEventPulsePolarity(self.taskHandle)

    def GetExportedChangeDetectEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedChangeDetectEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedChangeDetectEventOutputTerm(self, data):
        return DAQmxSetExportedChangeDetectEventOutputTerm(self.taskHandle, data)

    def ResetExportedChangeDetectEventOutputTerm(self):
        return DAQmxResetExportedChangeDetectEventOutputTerm(self.taskHandle)

    def GetExportedChangeDetectEventPulsePolarity(self, data):
        return DAQmxGetExportedChangeDetectEventPulsePolarity(self.taskHandle, data)

    def SetExportedChangeDetectEventPulsePolarity(self, data):
        return DAQmxSetExportedChangeDetectEventPulsePolarity(self.taskHandle, data)

    def ResetExportedChangeDetectEventPulsePolarity(self):
        return DAQmxResetExportedChangeDetectEventPulsePolarity(self.taskHandle)

    def GetExportedCtrOutEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedCtrOutEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedCtrOutEventOutputTerm(self, data):
        return DAQmxSetExportedCtrOutEventOutputTerm(self.taskHandle, data)

    def ResetExportedCtrOutEventOutputTerm(self):
        return DAQmxResetExportedCtrOutEventOutputTerm(self.taskHandle)

    def GetExportedCtrOutEventOutputBehavior(self, data):
        return DAQmxGetExportedCtrOutEventOutputBehavior(self.taskHandle, data)

    def SetExportedCtrOutEventOutputBehavior(self, data):
        return DAQmxSetExportedCtrOutEventOutputBehavior(self.taskHandle, data)

    def ResetExportedCtrOutEventOutputBehavior(self):
        return DAQmxResetExportedCtrOutEventOutputBehavior(self.taskHandle)

    def GetExportedCtrOutEventPulsePolarity(self, data):
        return DAQmxGetExportedCtrOutEventPulsePolarity(self.taskHandle, data)

    def SetExportedCtrOutEventPulsePolarity(self, data):
        return DAQmxSetExportedCtrOutEventPulsePolarity(self.taskHandle, data)

    def ResetExportedCtrOutEventPulsePolarity(self):
        return DAQmxResetExportedCtrOutEventPulsePolarity(self.taskHandle)

    def GetExportedCtrOutEventToggleIdleState(self, data):
        return DAQmxGetExportedCtrOutEventToggleIdleState(self.taskHandle, data)

    def SetExportedCtrOutEventToggleIdleState(self, data):
        return DAQmxSetExportedCtrOutEventToggleIdleState(self.taskHandle, data)

    def ResetExportedCtrOutEventToggleIdleState(self):
        return DAQmxResetExportedCtrOutEventToggleIdleState(self.taskHandle)

    def GetExportedHshkEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedHshkEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedHshkEventOutputTerm(self, data):
        return DAQmxSetExportedHshkEventOutputTerm(self.taskHandle, data)

    def ResetExportedHshkEventOutputTerm(self):
        return DAQmxResetExportedHshkEventOutputTerm(self.taskHandle)

    def GetExportedHshkEventOutputBehavior(self, data):
        return DAQmxGetExportedHshkEventOutputBehavior(self.taskHandle, data)

    def SetExportedHshkEventOutputBehavior(self, data):
        return DAQmxSetExportedHshkEventOutputBehavior(self.taskHandle, data)

    def ResetExportedHshkEventOutputBehavior(self):
        return DAQmxResetExportedHshkEventOutputBehavior(self.taskHandle)

    def GetExportedHshkEventDelay(self, data):
        return DAQmxGetExportedHshkEventDelay(self.taskHandle, data)

    def SetExportedHshkEventDelay(self, data):
        return DAQmxSetExportedHshkEventDelay(self.taskHandle, data)

    def ResetExportedHshkEventDelay(self):
        return DAQmxResetExportedHshkEventDelay(self.taskHandle)

    def GetExportedHshkEventInterlockedAssertedLvl(self, data):
        return DAQmxGetExportedHshkEventInterlockedAssertedLvl(self.taskHandle, data)

    def SetExportedHshkEventInterlockedAssertedLvl(self, data):
        return DAQmxSetExportedHshkEventInterlockedAssertedLvl(self.taskHandle, data)

    def ResetExportedHshkEventInterlockedAssertedLvl(self):
        return DAQmxResetExportedHshkEventInterlockedAssertedLvl(self.taskHandle)

    def GetExportedHshkEventInterlockedAssertOnStart(self, data):
        return DAQmxGetExportedHshkEventInterlockedAssertOnStart(self.taskHandle, data)

    def SetExportedHshkEventInterlockedAssertOnStart(self, data):
        return DAQmxSetExportedHshkEventInterlockedAssertOnStart(self.taskHandle, data)

    def ResetExportedHshkEventInterlockedAssertOnStart(self):
        return DAQmxResetExportedHshkEventInterlockedAssertOnStart(self.taskHandle)

    def GetExportedHshkEventInterlockedDeassertDelay(self, data):
        return DAQmxGetExportedHshkEventInterlockedDeassertDelay(self.taskHandle, data)

    def SetExportedHshkEventInterlockedDeassertDelay(self, data):
        return DAQmxSetExportedHshkEventInterlockedDeassertDelay(self.taskHandle, data)

    def ResetExportedHshkEventInterlockedDeassertDelay(self):
        return DAQmxResetExportedHshkEventInterlockedDeassertDelay(self.taskHandle)

    def GetExportedHshkEventPulsePolarity(self, data):
        return DAQmxGetExportedHshkEventPulsePolarity(self.taskHandle, data)

    def SetExportedHshkEventPulsePolarity(self, data):
        return DAQmxSetExportedHshkEventPulsePolarity(self.taskHandle, data)

    def ResetExportedHshkEventPulsePolarity(self):
        return DAQmxResetExportedHshkEventPulsePolarity(self.taskHandle)

    def GetExportedHshkEventPulseWidth(self, data):
        return DAQmxGetExportedHshkEventPulseWidth(self.taskHandle, data)

    def SetExportedHshkEventPulseWidth(self, data):
        return DAQmxSetExportedHshkEventPulseWidth(self.taskHandle, data)

    def ResetExportedHshkEventPulseWidth(self):
        return DAQmxResetExportedHshkEventPulseWidth(self.taskHandle)

    def GetExportedRdyForXferEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRdyForXferEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedRdyForXferEventOutputTerm(self, data):
        return DAQmxSetExportedRdyForXferEventOutputTerm(self.taskHandle, data)

    def ResetExportedRdyForXferEventOutputTerm(self):
        return DAQmxResetExportedRdyForXferEventOutputTerm(self.taskHandle)

    def GetExportedRdyForXferEventLvlActiveLvl(self, data):
        return DAQmxGetExportedRdyForXferEventLvlActiveLvl(self.taskHandle, data)

    def SetExportedRdyForXferEventLvlActiveLvl(self, data):
        return DAQmxSetExportedRdyForXferEventLvlActiveLvl(self.taskHandle, data)

    def ResetExportedRdyForXferEventLvlActiveLvl(self):
        return DAQmxResetExportedRdyForXferEventLvlActiveLvl(self.taskHandle)

    def GetExportedRdyForXferEventDeassertCond(self, data):
        return DAQmxGetExportedRdyForXferEventDeassertCond(self.taskHandle, data)

    def SetExportedRdyForXferEventDeassertCond(self, data):
        return DAQmxSetExportedRdyForXferEventDeassertCond(self.taskHandle, data)

    def ResetExportedRdyForXferEventDeassertCond(self):
        return DAQmxResetExportedRdyForXferEventDeassertCond(self.taskHandle)

    def GetExportedRdyForXferEventDeassertCondCustomThreshold(self, data):
        return DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold(self.taskHandle, data)

    def SetExportedRdyForXferEventDeassertCondCustomThreshold(self, data):
        return DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold(self.taskHandle, data)

    def ResetExportedRdyForXferEventDeassertCondCustomThreshold(self):
        return DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold(self.taskHandle)

    def GetExportedDataActiveEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedDataActiveEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedDataActiveEventOutputTerm(self, data):
        return DAQmxSetExportedDataActiveEventOutputTerm(self.taskHandle, data)

    def ResetExportedDataActiveEventOutputTerm(self):
        return DAQmxResetExportedDataActiveEventOutputTerm(self.taskHandle)

    def GetExportedDataActiveEventLvlActiveLvl(self, data):
        return DAQmxGetExportedDataActiveEventLvlActiveLvl(self.taskHandle, data)

    def SetExportedDataActiveEventLvlActiveLvl(self, data):
        return DAQmxSetExportedDataActiveEventLvlActiveLvl(self.taskHandle, data)

    def ResetExportedDataActiveEventLvlActiveLvl(self):
        return DAQmxResetExportedDataActiveEventLvlActiveLvl(self.taskHandle)

    def GetExportedRdyForStartEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRdyForStartEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedRdyForStartEventOutputTerm(self, data):
        return DAQmxSetExportedRdyForStartEventOutputTerm(self.taskHandle, data)

    def ResetExportedRdyForStartEventOutputTerm(self):
        return DAQmxResetExportedRdyForStartEventOutputTerm(self.taskHandle)

    def GetExportedRdyForStartEventLvlActiveLvl(self, data):
        return DAQmxGetExportedRdyForStartEventLvlActiveLvl(self.taskHandle, data)

    def SetExportedRdyForStartEventLvlActiveLvl(self, data):
        return DAQmxSetExportedRdyForStartEventLvlActiveLvl(self.taskHandle, data)

    def ResetExportedRdyForStartEventLvlActiveLvl(self):
        return DAQmxResetExportedRdyForStartEventLvlActiveLvl(self.taskHandle)

    def GetExportedSyncPulseEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSyncPulseEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedSyncPulseEventOutputTerm(self, data):
        return DAQmxSetExportedSyncPulseEventOutputTerm(self.taskHandle, data)

    def ResetExportedSyncPulseEventOutputTerm(self):
        return DAQmxResetExportedSyncPulseEventOutputTerm(self.taskHandle)

    def GetExportedWatchdogExpiredEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedWatchdogExpiredEventOutputTerm(self.taskHandle, data, bufferSize)

    def SetExportedWatchdogExpiredEventOutputTerm(self, data):
        return DAQmxSetExportedWatchdogExpiredEventOutputTerm(self.taskHandle, data)

    def ResetExportedWatchdogExpiredEventOutputTerm(self):
        return DAQmxResetExportedWatchdogExpiredEventOutputTerm(self.taskHandle)

    def GetReadRelativeTo(self, data):
        return DAQmxGetReadRelativeTo(self.taskHandle, data)

    def SetReadRelativeTo(self, data):
        return DAQmxSetReadRelativeTo(self.taskHandle, data)

    def ResetReadRelativeTo(self):
        return DAQmxResetReadRelativeTo(self.taskHandle)

    def GetReadOffset(self, data):
        return DAQmxGetReadOffset(self.taskHandle, data)

    def SetReadOffset(self, data):
        return DAQmxSetReadOffset(self.taskHandle, data)

    def ResetReadOffset(self):
        return DAQmxResetReadOffset(self.taskHandle)

    def GetReadChannelsToRead(self, data, bufferSize):
        return DAQmxGetReadChannelsToRead(self.taskHandle, data, bufferSize)

    def SetReadChannelsToRead(self, data):
        return DAQmxSetReadChannelsToRead(self.taskHandle, data)

    def ResetReadChannelsToRead(self):
        return DAQmxResetReadChannelsToRead(self.taskHandle)

    def GetReadReadAllAvailSamp(self, data):
        return DAQmxGetReadReadAllAvailSamp(self.taskHandle, data)

    def SetReadReadAllAvailSamp(self, data):
        return DAQmxSetReadReadAllAvailSamp(self.taskHandle, data)

    def ResetReadReadAllAvailSamp(self):
        return DAQmxResetReadReadAllAvailSamp(self.taskHandle)

    def GetReadAutoStart(self, data):
        return DAQmxGetReadAutoStart(self.taskHandle, data)

    def SetReadAutoStart(self, data):
        return DAQmxSetReadAutoStart(self.taskHandle, data)

    def ResetReadAutoStart(self):
        return DAQmxResetReadAutoStart(self.taskHandle)

    def GetReadOverWrite(self, data):
        return DAQmxGetReadOverWrite(self.taskHandle, data)

    def SetReadOverWrite(self, data):
        return DAQmxSetReadOverWrite(self.taskHandle, data)

    def ResetReadOverWrite(self):
        return DAQmxResetReadOverWrite(self.taskHandle)

    def GetReadCurrReadPos(self, data):
        return DAQmxGetReadCurrReadPos(self.taskHandle, data)

    def GetReadAvailSampPerChan(self, data):
        return DAQmxGetReadAvailSampPerChan(self.taskHandle, data)

    def GetLoggingFilePath(self, data, bufferSize):
        return DAQmxGetLoggingFilePath(self.taskHandle, data, bufferSize)

    def SetLoggingFilePath(self, data):
        return DAQmxSetLoggingFilePath(self.taskHandle, data)

    def ResetLoggingFilePath(self):
        return DAQmxResetLoggingFilePath(self.taskHandle)

    def GetLoggingMode(self, data):
        return DAQmxGetLoggingMode(self.taskHandle, data)

    def SetLoggingMode(self, data):
        return DAQmxSetLoggingMode(self.taskHandle, data)

    def ResetLoggingMode(self):
        return DAQmxResetLoggingMode(self.taskHandle)

    def GetLoggingTDMSGroupName(self, data, bufferSize):
        return DAQmxGetLoggingTDMSGroupName(self.taskHandle, data, bufferSize)

    def SetLoggingTDMSGroupName(self, data):
        return DAQmxSetLoggingTDMSGroupName(self.taskHandle, data)

    def ResetLoggingTDMSGroupName(self):
        return DAQmxResetLoggingTDMSGroupName(self.taskHandle)

    def GetLoggingTDMSOperation(self, data):
        return DAQmxGetLoggingTDMSOperation(self.taskHandle, data)

    def SetLoggingTDMSOperation(self, data):
        return DAQmxSetLoggingTDMSOperation(self.taskHandle, data)

    def ResetLoggingTDMSOperation(self):
        return DAQmxResetLoggingTDMSOperation(self.taskHandle)

    def GetLoggingPause(self, data):
        return DAQmxGetLoggingPause(self.taskHandle, data)

    def SetLoggingPause(self, data):
        return DAQmxSetLoggingPause(self.taskHandle, data)

    def ResetLoggingPause(self):
        return DAQmxResetLoggingPause(self.taskHandle)

    def GetLoggingSampsPerFile(self, data):
        return DAQmxGetLoggingSampsPerFile(self.taskHandle, data)

    def SetLoggingSampsPerFile(self, data):
        return DAQmxSetLoggingSampsPerFile(self.taskHandle, data)

    def ResetLoggingSampsPerFile(self):
        return DAQmxResetLoggingSampsPerFile(self.taskHandle)

    def GetLoggingFileWriteSize(self, data):
        return DAQmxGetLoggingFileWriteSize(self.taskHandle, data)

    def SetLoggingFileWriteSize(self, data):
        return DAQmxSetLoggingFileWriteSize(self.taskHandle, data)

    def ResetLoggingFileWriteSize(self):
        return DAQmxResetLoggingFileWriteSize(self.taskHandle)

    def GetLoggingFilePreallocationSize(self, data):
        return DAQmxGetLoggingFilePreallocationSize(self.taskHandle, data)

    def SetLoggingFilePreallocationSize(self, data):
        return DAQmxSetLoggingFilePreallocationSize(self.taskHandle, data)

    def ResetLoggingFilePreallocationSize(self):
        return DAQmxResetLoggingFilePreallocationSize(self.taskHandle)

    def GetReadTotalSampPerChanAcquired(self, data):
        return DAQmxGetReadTotalSampPerChanAcquired(self.taskHandle, data)

    def GetReadCommonModeRangeErrorChansExist(self, data):
        return DAQmxGetReadCommonModeRangeErrorChansExist(self.taskHandle, data)

    def GetReadCommonModeRangeErrorChans(self, data, bufferSize):
        return DAQmxGetReadCommonModeRangeErrorChans(self.taskHandle, data, bufferSize)

    def GetReadOvercurrentChansExist(self, data):
        return DAQmxGetReadOvercurrentChansExist(self.taskHandle, data)

    def GetReadOvercurrentChans(self, data, bufferSize):
        return DAQmxGetReadOvercurrentChans(self.taskHandle, data, bufferSize)

    def GetReadOpenCurrentLoopChansExist(self, data):
        return DAQmxGetReadOpenCurrentLoopChansExist(self.taskHandle, data)

    def GetReadOpenCurrentLoopChans(self, data, bufferSize):
        return DAQmxGetReadOpenCurrentLoopChans(self.taskHandle, data, bufferSize)

    def GetReadOpenThrmcplChansExist(self, data):
        return DAQmxGetReadOpenThrmcplChansExist(self.taskHandle, data)

    def GetReadOpenThrmcplChans(self, data, bufferSize):
        return DAQmxGetReadOpenThrmcplChans(self.taskHandle, data, bufferSize)

    def GetReadOverloadedChansExist(self, data):
        return DAQmxGetReadOverloadedChansExist(self.taskHandle, data)

    def GetReadOverloadedChans(self, data, bufferSize):
        return DAQmxGetReadOverloadedChans(self.taskHandle, data, bufferSize)

    def GetReadAccessoryInsertionOrRemovalDetected(self, data):
        return DAQmxGetReadAccessoryInsertionOrRemovalDetected(self.taskHandle, data)

    def GetReadDevsWithInsertedOrRemovedAccessories(self, data, bufferSize):
        return DAQmxGetReadDevsWithInsertedOrRemovedAccessories(self.taskHandle, data, bufferSize)

    def GetReadChangeDetectHasOverflowed(self, data):
        return DAQmxGetReadChangeDetectHasOverflowed(self.taskHandle, data)

    def GetReadRawDataWidth(self, data):
        return DAQmxGetReadRawDataWidth(self.taskHandle, data)

    def GetReadNumChans(self, data):
        return DAQmxGetReadNumChans(self.taskHandle, data)

    def GetReadDigitalLinesBytesPerChan(self, data):
        return DAQmxGetReadDigitalLinesBytesPerChan(self.taskHandle, data)

    def GetReadWaitMode(self, data):
        return DAQmxGetReadWaitMode(self.taskHandle, data)

    def SetReadWaitMode(self, data):
        return DAQmxSetReadWaitMode(self.taskHandle, data)

    def ResetReadWaitMode(self):
        return DAQmxResetReadWaitMode(self.taskHandle)

    def GetReadSleepTime(self, data):
        return DAQmxGetReadSleepTime(self.taskHandle, data)

    def SetReadSleepTime(self, data):
        return DAQmxSetReadSleepTime(self.taskHandle, data)

    def ResetReadSleepTime(self):
        return DAQmxResetReadSleepTime(self.taskHandle)

    def GetRealTimeConvLateErrorsToWarnings(self, data):
        return DAQmxGetRealTimeConvLateErrorsToWarnings(self.taskHandle, data)

    def SetRealTimeConvLateErrorsToWarnings(self, data):
        return DAQmxSetRealTimeConvLateErrorsToWarnings(self.taskHandle, data)

    def ResetRealTimeConvLateErrorsToWarnings(self):
        return DAQmxResetRealTimeConvLateErrorsToWarnings(self.taskHandle)

    def GetRealTimeNumOfWarmupIters(self, data):
        return DAQmxGetRealTimeNumOfWarmupIters(self.taskHandle, data)

    def SetRealTimeNumOfWarmupIters(self, data):
        return DAQmxSetRealTimeNumOfWarmupIters(self.taskHandle, data)

    def ResetRealTimeNumOfWarmupIters(self):
        return DAQmxResetRealTimeNumOfWarmupIters(self.taskHandle)

    def GetRealTimeWaitForNextSampClkWaitMode(self, data):
        return DAQmxGetRealTimeWaitForNextSampClkWaitMode(self.taskHandle, data)

    def SetRealTimeWaitForNextSampClkWaitMode(self, data):
        return DAQmxSetRealTimeWaitForNextSampClkWaitMode(self.taskHandle, data)

    def ResetRealTimeWaitForNextSampClkWaitMode(self):
        return DAQmxResetRealTimeWaitForNextSampClkWaitMode(self.taskHandle)

    def GetRealTimeReportMissedSamp(self, data):
        return DAQmxGetRealTimeReportMissedSamp(self.taskHandle, data)

    def SetRealTimeReportMissedSamp(self, data):
        return DAQmxSetRealTimeReportMissedSamp(self.taskHandle, data)

    def ResetRealTimeReportMissedSamp(self):
        return DAQmxResetRealTimeReportMissedSamp(self.taskHandle)

    def GetRealTimeWriteRecoveryMode(self, data):
        return DAQmxGetRealTimeWriteRecoveryMode(self.taskHandle, data)

    def SetRealTimeWriteRecoveryMode(self, data):
        return DAQmxSetRealTimeWriteRecoveryMode(self.taskHandle, data)

    def ResetRealTimeWriteRecoveryMode(self):
        return DAQmxResetRealTimeWriteRecoveryMode(self.taskHandle)

    def GetSwitchScanBreakMode(self, data):
        return DAQmxGetSwitchScanBreakMode(self.taskHandle, data)

    def SetSwitchScanBreakMode(self, data):
        return DAQmxSetSwitchScanBreakMode(self.taskHandle, data)

    def ResetSwitchScanBreakMode(self):
        return DAQmxResetSwitchScanBreakMode(self.taskHandle)

    def GetSwitchScanRepeatMode(self, data):
        return DAQmxGetSwitchScanRepeatMode(self.taskHandle, data)

    def SetSwitchScanRepeatMode(self, data):
        return DAQmxSetSwitchScanRepeatMode(self.taskHandle, data)

    def ResetSwitchScanRepeatMode(self):
        return DAQmxResetSwitchScanRepeatMode(self.taskHandle)

    def GetSwitchScanWaitingForAdv(self, data):
        return DAQmxGetSwitchScanWaitingForAdv(self.taskHandle, data)

    def GetTaskName(self, data, bufferSize):
        return DAQmxGetTaskName(self.taskHandle, data, bufferSize)

    def GetTaskChannels(self, data, bufferSize):
        return DAQmxGetTaskChannels(self.taskHandle, data, bufferSize)

    def GetTaskNumChans(self, data):
        return DAQmxGetTaskNumChans(self.taskHandle, data)

    def GetTaskDevices(self, data, bufferSize):
        return DAQmxGetTaskDevices(self.taskHandle, data, bufferSize)

    def GetTaskNumDevices(self, data):
        return DAQmxGetTaskNumDevices(self.taskHandle, data)

    def GetTaskComplete(self, data):
        return DAQmxGetTaskComplete(self.taskHandle, data)

    def GetSampQuantSampMode(self, data):
        return DAQmxGetSampQuantSampMode(self.taskHandle, data)

    def SetSampQuantSampMode(self, data):
        return DAQmxSetSampQuantSampMode(self.taskHandle, data)

    def ResetSampQuantSampMode(self):
        return DAQmxResetSampQuantSampMode(self.taskHandle)

    def GetSampQuantSampPerChan(self, data):
        return DAQmxGetSampQuantSampPerChan(self.taskHandle, data)

    def SetSampQuantSampPerChan(self, data):
        return DAQmxSetSampQuantSampPerChan(self.taskHandle, data)

    def ResetSampQuantSampPerChan(self):
        return DAQmxResetSampQuantSampPerChan(self.taskHandle)

    def GetSampTimingType(self, data):
        return DAQmxGetSampTimingType(self.taskHandle, data)

    def SetSampTimingType(self, data):
        return DAQmxSetSampTimingType(self.taskHandle, data)

    def ResetSampTimingType(self):
        return DAQmxResetSampTimingType(self.taskHandle)

    def GetSampClkRate(self, data):
        return DAQmxGetSampClkRate(self.taskHandle, data)

    def SetSampClkRate(self, data):
        return DAQmxSetSampClkRate(self.taskHandle, data)

    def ResetSampClkRate(self):
        return DAQmxResetSampClkRate(self.taskHandle)

    def GetSampClkMaxRate(self, data):
        return DAQmxGetSampClkMaxRate(self.taskHandle, data)

    def GetSampClkSrc(self, data, bufferSize):
        return DAQmxGetSampClkSrc(self.taskHandle, data, bufferSize)

    def SetSampClkSrc(self, data):
        return DAQmxSetSampClkSrc(self.taskHandle, data)

    def ResetSampClkSrc(self):
        return DAQmxResetSampClkSrc(self.taskHandle)

    def GetSampClkActiveEdge(self, data):
        return DAQmxGetSampClkActiveEdge(self.taskHandle, data)

    def SetSampClkActiveEdge(self, data):
        return DAQmxSetSampClkActiveEdge(self.taskHandle, data)

    def ResetSampClkActiveEdge(self):
        return DAQmxResetSampClkActiveEdge(self.taskHandle)

    def GetSampClkOverrunBehavior(self, data):
        return DAQmxGetSampClkOverrunBehavior(self.taskHandle, data)

    def SetSampClkOverrunBehavior(self, data):
        return DAQmxSetSampClkOverrunBehavior(self.taskHandle, data)

    def ResetSampClkOverrunBehavior(self):
        return DAQmxResetSampClkOverrunBehavior(self.taskHandle)

    def GetSampClkUnderflowBehavior(self, data):
        return DAQmxGetSampClkUnderflowBehavior(self.taskHandle, data)

    def SetSampClkUnderflowBehavior(self, data):
        return DAQmxSetSampClkUnderflowBehavior(self.taskHandle, data)

    def ResetSampClkUnderflowBehavior(self):
        return DAQmxResetSampClkUnderflowBehavior(self.taskHandle)

    def GetSampClkTimebaseDiv(self, data):
        return DAQmxGetSampClkTimebaseDiv(self.taskHandle, data)

    def SetSampClkTimebaseDiv(self, data):
        return DAQmxSetSampClkTimebaseDiv(self.taskHandle, data)

    def ResetSampClkTimebaseDiv(self):
        return DAQmxResetSampClkTimebaseDiv(self.taskHandle)

    def GetSampClkTerm(self, data, bufferSize):
        return DAQmxGetSampClkTerm(self.taskHandle, data, bufferSize)

    def GetSampClkTimebaseRate(self, data):
        return DAQmxGetSampClkTimebaseRate(self.taskHandle, data)

    def SetSampClkTimebaseRate(self, data):
        return DAQmxSetSampClkTimebaseRate(self.taskHandle, data)

    def ResetSampClkTimebaseRate(self):
        return DAQmxResetSampClkTimebaseRate(self.taskHandle)

    def GetSampClkTimebaseSrc(self, data, bufferSize):
        return DAQmxGetSampClkTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetSampClkTimebaseSrc(self, data):
        return DAQmxSetSampClkTimebaseSrc(self.taskHandle, data)

    def ResetSampClkTimebaseSrc(self):
        return DAQmxResetSampClkTimebaseSrc(self.taskHandle)

    def GetSampClkTimebaseActiveEdge(self, data):
        return DAQmxGetSampClkTimebaseActiveEdge(self.taskHandle, data)

    def SetSampClkTimebaseActiveEdge(self, data):
        return DAQmxSetSampClkTimebaseActiveEdge(self.taskHandle, data)

    def ResetSampClkTimebaseActiveEdge(self):
        return DAQmxResetSampClkTimebaseActiveEdge(self.taskHandle)

    def GetSampClkTimebaseMasterTimebaseDiv(self, data):
        return DAQmxGetSampClkTimebaseMasterTimebaseDiv(self.taskHandle, data)

    def SetSampClkTimebaseMasterTimebaseDiv(self, data):
        return DAQmxSetSampClkTimebaseMasterTimebaseDiv(self.taskHandle, data)

    def ResetSampClkTimebaseMasterTimebaseDiv(self):
        return DAQmxResetSampClkTimebaseMasterTimebaseDiv(self.taskHandle)

    def GetSampClkTimebaseTerm(self, data, bufferSize):
        return DAQmxGetSampClkTimebaseTerm(self.taskHandle, data, bufferSize)

    def GetSampClkDigFltrEnable(self, data):
        return DAQmxGetSampClkDigFltrEnable(self.taskHandle, data)

    def SetSampClkDigFltrEnable(self, data):
        return DAQmxSetSampClkDigFltrEnable(self.taskHandle, data)

    def ResetSampClkDigFltrEnable(self):
        return DAQmxResetSampClkDigFltrEnable(self.taskHandle)

    def GetSampClkDigFltrMinPulseWidth(self, data):
        return DAQmxGetSampClkDigFltrMinPulseWidth(self.taskHandle, data)

    def SetSampClkDigFltrMinPulseWidth(self, data):
        return DAQmxSetSampClkDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetSampClkDigFltrMinPulseWidth(self):
        return DAQmxResetSampClkDigFltrMinPulseWidth(self.taskHandle)

    def GetSampClkDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetSampClkDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetSampClkDigFltrTimebaseSrc(self, data):
        return DAQmxSetSampClkDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetSampClkDigFltrTimebaseSrc(self):
        return DAQmxResetSampClkDigFltrTimebaseSrc(self.taskHandle)

    def GetSampClkDigFltrTimebaseRate(self, data):
        return DAQmxGetSampClkDigFltrTimebaseRate(self.taskHandle, data)

    def SetSampClkDigFltrTimebaseRate(self, data):
        return DAQmxSetSampClkDigFltrTimebaseRate(self.taskHandle, data)

    def ResetSampClkDigFltrTimebaseRate(self):
        return DAQmxResetSampClkDigFltrTimebaseRate(self.taskHandle)

    def GetSampClkDigSyncEnable(self, data):
        return DAQmxGetSampClkDigSyncEnable(self.taskHandle, data)

    def SetSampClkDigSyncEnable(self, data):
        return DAQmxSetSampClkDigSyncEnable(self.taskHandle, data)

    def ResetSampClkDigSyncEnable(self):
        return DAQmxResetSampClkDigSyncEnable(self.taskHandle)

    def GetHshkDelayAfterXfer(self, data):
        return DAQmxGetHshkDelayAfterXfer(self.taskHandle, data)

    def SetHshkDelayAfterXfer(self, data):
        return DAQmxSetHshkDelayAfterXfer(self.taskHandle, data)

    def ResetHshkDelayAfterXfer(self):
        return DAQmxResetHshkDelayAfterXfer(self.taskHandle)

    def GetHshkStartCond(self, data):
        return DAQmxGetHshkStartCond(self.taskHandle, data)

    def SetHshkStartCond(self, data):
        return DAQmxSetHshkStartCond(self.taskHandle, data)

    def ResetHshkStartCond(self):
        return DAQmxResetHshkStartCond(self.taskHandle)

    def GetHshkSampleInputDataWhen(self, data):
        return DAQmxGetHshkSampleInputDataWhen(self.taskHandle, data)

    def SetHshkSampleInputDataWhen(self, data):
        return DAQmxSetHshkSampleInputDataWhen(self.taskHandle, data)

    def ResetHshkSampleInputDataWhen(self):
        return DAQmxResetHshkSampleInputDataWhen(self.taskHandle)

    def GetChangeDetectDIRisingEdgePhysicalChans(self, data, bufferSize):
        return DAQmxGetChangeDetectDIRisingEdgePhysicalChans(self.taskHandle, data, bufferSize)

    def SetChangeDetectDIRisingEdgePhysicalChans(self, data):
        return DAQmxSetChangeDetectDIRisingEdgePhysicalChans(self.taskHandle, data)

    def ResetChangeDetectDIRisingEdgePhysicalChans(self):
        return DAQmxResetChangeDetectDIRisingEdgePhysicalChans(self.taskHandle)

    def GetChangeDetectDIFallingEdgePhysicalChans(self, data, bufferSize):
        return DAQmxGetChangeDetectDIFallingEdgePhysicalChans(self.taskHandle, data, bufferSize)

    def SetChangeDetectDIFallingEdgePhysicalChans(self, data):
        return DAQmxSetChangeDetectDIFallingEdgePhysicalChans(self.taskHandle, data)

    def ResetChangeDetectDIFallingEdgePhysicalChans(self):
        return DAQmxResetChangeDetectDIFallingEdgePhysicalChans(self.taskHandle)

    def GetChangeDetectDITristate(self, data):
        return DAQmxGetChangeDetectDITristate(self.taskHandle, data)

    def SetChangeDetectDITristate(self, data):
        return DAQmxSetChangeDetectDITristate(self.taskHandle, data)

    def ResetChangeDetectDITristate(self):
        return DAQmxResetChangeDetectDITristate(self.taskHandle)

    def GetOnDemandSimultaneousAOEnable(self, data):
        return DAQmxGetOnDemandSimultaneousAOEnable(self.taskHandle, data)

    def SetOnDemandSimultaneousAOEnable(self, data):
        return DAQmxSetOnDemandSimultaneousAOEnable(self.taskHandle, data)

    def ResetOnDemandSimultaneousAOEnable(self):
        return DAQmxResetOnDemandSimultaneousAOEnable(self.taskHandle)

    def GetImplicitUnderflowBehavior(self, data):
        return DAQmxGetImplicitUnderflowBehavior(self.taskHandle, data)

    def SetImplicitUnderflowBehavior(self, data):
        return DAQmxSetImplicitUnderflowBehavior(self.taskHandle, data)

    def ResetImplicitUnderflowBehavior(self):
        return DAQmxResetImplicitUnderflowBehavior(self.taskHandle)

    def GetAIConvRate(self, data):
        return DAQmxGetAIConvRate(self.taskHandle, data)

    def SetAIConvRate(self, data):
        return DAQmxSetAIConvRate(self.taskHandle, data)

    def ResetAIConvRate(self):
        return DAQmxResetAIConvRate(self.taskHandle)

    def GetAIConvRateEx(self, deviceNames, data):
        return DAQmxGetAIConvRateEx(self.taskHandle, deviceNames, data)

    def SetAIConvRateEx(self, deviceNames, data):
        return DAQmxSetAIConvRateEx(self.taskHandle, deviceNames, data)

    def ResetAIConvRateEx(self, deviceNames):
        return DAQmxResetAIConvRateEx(self.taskHandle, deviceNames)

    def GetAIConvMaxRate(self, data):
        return DAQmxGetAIConvMaxRate(self.taskHandle, data)

    def GetAIConvMaxRateEx(self, deviceNames, data):
        return DAQmxGetAIConvMaxRateEx(self.taskHandle, deviceNames, data)

    def GetAIConvSrc(self, data, bufferSize):
        return DAQmxGetAIConvSrc(self.taskHandle, data, bufferSize)

    def SetAIConvSrc(self, data):
        return DAQmxSetAIConvSrc(self.taskHandle, data)

    def ResetAIConvSrc(self):
        return DAQmxResetAIConvSrc(self.taskHandle)

    def GetAIConvSrcEx(self, deviceNames, data, bufferSize):
        return DAQmxGetAIConvSrcEx(self.taskHandle, deviceNames, data, bufferSize)

    def SetAIConvSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvSrcEx(self.taskHandle, deviceNames, data)

    def ResetAIConvSrcEx(self, deviceNames):
        return DAQmxResetAIConvSrcEx(self.taskHandle, deviceNames)

    def GetAIConvActiveEdge(self, data):
        return DAQmxGetAIConvActiveEdge(self.taskHandle, data)

    def SetAIConvActiveEdge(self, data):
        return DAQmxSetAIConvActiveEdge(self.taskHandle, data)

    def ResetAIConvActiveEdge(self):
        return DAQmxResetAIConvActiveEdge(self.taskHandle)

    def GetAIConvActiveEdgeEx(self, deviceNames, data):
        return DAQmxGetAIConvActiveEdgeEx(self.taskHandle, deviceNames, data)

    def SetAIConvActiveEdgeEx(self, deviceNames, data):
        return DAQmxSetAIConvActiveEdgeEx(self.taskHandle, deviceNames, data)

    def ResetAIConvActiveEdgeEx(self, deviceNames):
        return DAQmxResetAIConvActiveEdgeEx(self.taskHandle, deviceNames)

    def GetAIConvTimebaseDiv(self, data):
        return DAQmxGetAIConvTimebaseDiv(self.taskHandle, data)

    def SetAIConvTimebaseDiv(self, data):
        return DAQmxSetAIConvTimebaseDiv(self.taskHandle, data)

    def ResetAIConvTimebaseDiv(self):
        return DAQmxResetAIConvTimebaseDiv(self.taskHandle)

    def GetAIConvTimebaseDivEx(self, deviceNames, data):
        return DAQmxGetAIConvTimebaseDivEx(self.taskHandle, deviceNames, data)

    def SetAIConvTimebaseDivEx(self, deviceNames, data):
        return DAQmxSetAIConvTimebaseDivEx(self.taskHandle, deviceNames, data)

    def ResetAIConvTimebaseDivEx(self, deviceNames):
        return DAQmxResetAIConvTimebaseDivEx(self.taskHandle, deviceNames)

    def GetAIConvTimebaseSrc(self, data):
        return DAQmxGetAIConvTimebaseSrc(self.taskHandle, data)

    def SetAIConvTimebaseSrc(self, data):
        return DAQmxSetAIConvTimebaseSrc(self.taskHandle, data)

    def ResetAIConvTimebaseSrc(self):
        return DAQmxResetAIConvTimebaseSrc(self.taskHandle)

    def GetAIConvTimebaseSrcEx(self, deviceNames, data):
        return DAQmxGetAIConvTimebaseSrcEx(self.taskHandle, deviceNames, data)

    def SetAIConvTimebaseSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvTimebaseSrcEx(self.taskHandle, deviceNames, data)

    def ResetAIConvTimebaseSrcEx(self, deviceNames):
        return DAQmxResetAIConvTimebaseSrcEx(self.taskHandle, deviceNames)

    def GetDelayFromSampClkDelayUnits(self, data):
        return DAQmxGetDelayFromSampClkDelayUnits(self.taskHandle, data)

    def SetDelayFromSampClkDelayUnits(self, data):
        return DAQmxSetDelayFromSampClkDelayUnits(self.taskHandle, data)

    def ResetDelayFromSampClkDelayUnits(self):
        return DAQmxResetDelayFromSampClkDelayUnits(self.taskHandle)

    def GetDelayFromSampClkDelayUnitsEx(self, deviceNames, data):
        return DAQmxGetDelayFromSampClkDelayUnitsEx(self.taskHandle, deviceNames, data)

    def SetDelayFromSampClkDelayUnitsEx(self, deviceNames, data):
        return DAQmxSetDelayFromSampClkDelayUnitsEx(self.taskHandle, deviceNames, data)

    def ResetDelayFromSampClkDelayUnitsEx(self, deviceNames):
        return DAQmxResetDelayFromSampClkDelayUnitsEx(self.taskHandle, deviceNames)

    def GetDelayFromSampClkDelay(self, data):
        return DAQmxGetDelayFromSampClkDelay(self.taskHandle, data)

    def SetDelayFromSampClkDelay(self, data):
        return DAQmxSetDelayFromSampClkDelay(self.taskHandle, data)

    def ResetDelayFromSampClkDelay(self):
        return DAQmxResetDelayFromSampClkDelay(self.taskHandle)

    def GetDelayFromSampClkDelayEx(self, deviceNames, data):
        return DAQmxGetDelayFromSampClkDelayEx(self.taskHandle, deviceNames, data)

    def SetDelayFromSampClkDelayEx(self, deviceNames, data):
        return DAQmxSetDelayFromSampClkDelayEx(self.taskHandle, deviceNames, data)

    def ResetDelayFromSampClkDelayEx(self, deviceNames):
        return DAQmxResetDelayFromSampClkDelayEx(self.taskHandle, deviceNames)

    def GetAIConvDigFltrEnable(self, data):
        return DAQmxGetAIConvDigFltrEnable(self.taskHandle, data)

    def SetAIConvDigFltrEnable(self, data):
        return DAQmxSetAIConvDigFltrEnable(self.taskHandle, data)

    def ResetAIConvDigFltrEnable(self):
        return DAQmxResetAIConvDigFltrEnable(self.taskHandle)

    def GetAIConvDigFltrEnableEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrEnableEx(self.taskHandle, deviceNames, data)

    def SetAIConvDigFltrEnableEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrEnableEx(self.taskHandle, deviceNames, data)

    def ResetAIConvDigFltrEnableEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrEnableEx(self.taskHandle, deviceNames)

    def GetAIConvDigFltrMinPulseWidth(self, data):
        return DAQmxGetAIConvDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAIConvDigFltrMinPulseWidth(self, data):
        return DAQmxSetAIConvDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAIConvDigFltrMinPulseWidth(self):
        return DAQmxResetAIConvDigFltrMinPulseWidth(self.taskHandle)

    def GetAIConvDigFltrMinPulseWidthEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrMinPulseWidthEx(self.taskHandle, deviceNames, data)

    def SetAIConvDigFltrMinPulseWidthEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrMinPulseWidthEx(self.taskHandle, deviceNames, data)

    def ResetAIConvDigFltrMinPulseWidthEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrMinPulseWidthEx(self.taskHandle, deviceNames)

    def GetAIConvDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAIConvDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAIConvDigFltrTimebaseSrc(self, data):
        return DAQmxSetAIConvDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAIConvDigFltrTimebaseSrc(self):
        return DAQmxResetAIConvDigFltrTimebaseSrc(self.taskHandle)

    def GetAIConvDigFltrTimebaseSrcEx(self, deviceNames, data, bufferSize):
        return DAQmxGetAIConvDigFltrTimebaseSrcEx(self.taskHandle, deviceNames, data, bufferSize)

    def SetAIConvDigFltrTimebaseSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrTimebaseSrcEx(self.taskHandle, deviceNames, data)

    def ResetAIConvDigFltrTimebaseSrcEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrTimebaseSrcEx(self.taskHandle, deviceNames)

    def GetAIConvDigFltrTimebaseRate(self, data):
        return DAQmxGetAIConvDigFltrTimebaseRate(self.taskHandle, data)

    def SetAIConvDigFltrTimebaseRate(self, data):
        return DAQmxSetAIConvDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAIConvDigFltrTimebaseRate(self):
        return DAQmxResetAIConvDigFltrTimebaseRate(self.taskHandle)

    def GetAIConvDigFltrTimebaseRateEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrTimebaseRateEx(self.taskHandle, deviceNames, data)

    def SetAIConvDigFltrTimebaseRateEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrTimebaseRateEx(self.taskHandle, deviceNames, data)

    def ResetAIConvDigFltrTimebaseRateEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrTimebaseRateEx(self.taskHandle, deviceNames)

    def GetAIConvDigSyncEnable(self, data):
        return DAQmxGetAIConvDigSyncEnable(self.taskHandle, data)

    def SetAIConvDigSyncEnable(self, data):
        return DAQmxSetAIConvDigSyncEnable(self.taskHandle, data)

    def ResetAIConvDigSyncEnable(self):
        return DAQmxResetAIConvDigSyncEnable(self.taskHandle)

    def GetAIConvDigSyncEnableEx(self, deviceNames, data):
        return DAQmxGetAIConvDigSyncEnableEx(self.taskHandle, deviceNames, data)

    def SetAIConvDigSyncEnableEx(self, deviceNames, data):
        return DAQmxSetAIConvDigSyncEnableEx(self.taskHandle, deviceNames, data)

    def ResetAIConvDigSyncEnableEx(self, deviceNames):
        return DAQmxResetAIConvDigSyncEnableEx(self.taskHandle, deviceNames)

    def GetMasterTimebaseRate(self, data):
        return DAQmxGetMasterTimebaseRate(self.taskHandle, data)

    def SetMasterTimebaseRate(self, data):
        return DAQmxSetMasterTimebaseRate(self.taskHandle, data)

    def ResetMasterTimebaseRate(self):
        return DAQmxResetMasterTimebaseRate(self.taskHandle)

    def GetMasterTimebaseSrc(self, data, bufferSize):
        return DAQmxGetMasterTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetMasterTimebaseSrc(self, data):
        return DAQmxSetMasterTimebaseSrc(self.taskHandle, data)

    def ResetMasterTimebaseSrc(self):
        return DAQmxResetMasterTimebaseSrc(self.taskHandle)

    def GetRefClkRate(self, data):
        return DAQmxGetRefClkRate(self.taskHandle, data)

    def SetRefClkRate(self, data):
        return DAQmxSetRefClkRate(self.taskHandle, data)

    def ResetRefClkRate(self):
        return DAQmxResetRefClkRate(self.taskHandle)

    def GetRefClkSrc(self, data, bufferSize):
        return DAQmxGetRefClkSrc(self.taskHandle, data, bufferSize)

    def SetRefClkSrc(self, data):
        return DAQmxSetRefClkSrc(self.taskHandle, data)

    def ResetRefClkSrc(self):
        return DAQmxResetRefClkSrc(self.taskHandle)

    def GetSyncPulseSrc(self, data, bufferSize):
        return DAQmxGetSyncPulseSrc(self.taskHandle, data, bufferSize)

    def SetSyncPulseSrc(self, data):
        return DAQmxSetSyncPulseSrc(self.taskHandle, data)

    def ResetSyncPulseSrc(self):
        return DAQmxResetSyncPulseSrc(self.taskHandle)

    def GetSyncPulseSyncTime(self, data):
        return DAQmxGetSyncPulseSyncTime(self.taskHandle, data)

    def GetSyncPulseMinDelayToStart(self, data):
        return DAQmxGetSyncPulseMinDelayToStart(self.taskHandle, data)

    def SetSyncPulseMinDelayToStart(self, data):
        return DAQmxSetSyncPulseMinDelayToStart(self.taskHandle, data)

    def ResetSyncPulseMinDelayToStart(self):
        return DAQmxResetSyncPulseMinDelayToStart(self.taskHandle)

    def GetSyncPulseResetTime(self, data):
        return DAQmxGetSyncPulseResetTime(self.taskHandle, data)

    def GetSyncPulseResetDelay(self, data):
        return DAQmxGetSyncPulseResetDelay(self.taskHandle, data)

    def SetSyncPulseResetDelay(self, data):
        return DAQmxSetSyncPulseResetDelay(self.taskHandle, data)

    def ResetSyncPulseResetDelay(self):
        return DAQmxResetSyncPulseResetDelay(self.taskHandle)

    def GetSyncPulseTerm(self, data, bufferSize):
        return DAQmxGetSyncPulseTerm(self.taskHandle, data, bufferSize)

    def GetSyncClkInterval(self, data):
        return DAQmxGetSyncClkInterval(self.taskHandle, data)

    def SetSyncClkInterval(self, data):
        return DAQmxSetSyncClkInterval(self.taskHandle, data)

    def ResetSyncClkInterval(self):
        return DAQmxResetSyncClkInterval(self.taskHandle)

    def GetSampTimingEngine(self, data):
        return DAQmxGetSampTimingEngine(self.taskHandle, data)

    def SetSampTimingEngine(self, data):
        return DAQmxSetSampTimingEngine(self.taskHandle, data)

    def ResetSampTimingEngine(self):
        return DAQmxResetSampTimingEngine(self.taskHandle)

    def GetStartTrigType(self, data):
        return DAQmxGetStartTrigType(self.taskHandle, data)

    def SetStartTrigType(self, data):
        return DAQmxSetStartTrigType(self.taskHandle, data)

    def ResetStartTrigType(self):
        return DAQmxResetStartTrigType(self.taskHandle)

    def GetStartTrigTerm(self, data, bufferSize):
        return DAQmxGetStartTrigTerm(self.taskHandle, data, bufferSize)

    def GetDigEdgeStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeStartTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeStartTrigSrc(self, data):
        return DAQmxSetDigEdgeStartTrigSrc(self.taskHandle, data)

    def ResetDigEdgeStartTrigSrc(self):
        return DAQmxResetDigEdgeStartTrigSrc(self.taskHandle)

    def GetDigEdgeStartTrigEdge(self, data):
        return DAQmxGetDigEdgeStartTrigEdge(self.taskHandle, data)

    def SetDigEdgeStartTrigEdge(self, data):
        return DAQmxSetDigEdgeStartTrigEdge(self.taskHandle, data)

    def ResetDigEdgeStartTrigEdge(self):
        return DAQmxResetDigEdgeStartTrigEdge(self.taskHandle)

    def GetDigEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrEnable(self.taskHandle, data)

    def SetDigEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrEnable(self.taskHandle, data)

    def ResetDigEdgeStartTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeStartTrigDigFltrEnable(self.taskHandle)

    def GetDigEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetDigEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetDigEdgeStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetDigEdgeStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetDigEdgeStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetDigEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetDigEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetDigEdgeStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate(self.taskHandle)

    def GetDigEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeStartTrigDigSyncEnable(self.taskHandle, data)

    def SetDigEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeStartTrigDigSyncEnable(self.taskHandle, data)

    def ResetDigEdgeStartTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeStartTrigDigSyncEnable(self.taskHandle)

    def GetDigPatternStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternStartTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigPatternStartTrigSrc(self, data):
        return DAQmxSetDigPatternStartTrigSrc(self.taskHandle, data)

    def ResetDigPatternStartTrigSrc(self):
        return DAQmxResetDigPatternStartTrigSrc(self.taskHandle)

    def GetDigPatternStartTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternStartTrigPattern(self.taskHandle, data, bufferSize)

    def SetDigPatternStartTrigPattern(self, data):
        return DAQmxSetDigPatternStartTrigPattern(self.taskHandle, data)

    def ResetDigPatternStartTrigPattern(self):
        return DAQmxResetDigPatternStartTrigPattern(self.taskHandle)

    def GetDigPatternStartTrigWhen(self, data):
        return DAQmxGetDigPatternStartTrigWhen(self.taskHandle, data)

    def SetDigPatternStartTrigWhen(self, data):
        return DAQmxSetDigPatternStartTrigWhen(self.taskHandle, data)

    def ResetDigPatternStartTrigWhen(self):
        return DAQmxResetDigPatternStartTrigWhen(self.taskHandle)

    def GetAnlgEdgeStartTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeStartTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgEdgeStartTrigSrc(self, data):
        return DAQmxSetAnlgEdgeStartTrigSrc(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigSrc(self):
        return DAQmxResetAnlgEdgeStartTrigSrc(self.taskHandle)

    def GetAnlgEdgeStartTrigSlope(self, data):
        return DAQmxGetAnlgEdgeStartTrigSlope(self.taskHandle, data)

    def SetAnlgEdgeStartTrigSlope(self, data):
        return DAQmxSetAnlgEdgeStartTrigSlope(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigSlope(self):
        return DAQmxResetAnlgEdgeStartTrigSlope(self.taskHandle)

    def GetAnlgEdgeStartTrigLvl(self, data):
        return DAQmxGetAnlgEdgeStartTrigLvl(self.taskHandle, data)

    def SetAnlgEdgeStartTrigLvl(self, data):
        return DAQmxSetAnlgEdgeStartTrigLvl(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigLvl(self):
        return DAQmxResetAnlgEdgeStartTrigLvl(self.taskHandle)

    def GetAnlgEdgeStartTrigHyst(self, data):
        return DAQmxGetAnlgEdgeStartTrigHyst(self.taskHandle, data)

    def SetAnlgEdgeStartTrigHyst(self, data):
        return DAQmxSetAnlgEdgeStartTrigHyst(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigHyst(self):
        return DAQmxResetAnlgEdgeStartTrigHyst(self.taskHandle)

    def GetAnlgEdgeStartTrigCoupling(self, data):
        return DAQmxGetAnlgEdgeStartTrigCoupling(self.taskHandle, data)

    def SetAnlgEdgeStartTrigCoupling(self, data):
        return DAQmxSetAnlgEdgeStartTrigCoupling(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigCoupling(self):
        return DAQmxResetAnlgEdgeStartTrigCoupling(self.taskHandle)

    def GetAnlgEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigDigFltrEnable(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrEnable(self.taskHandle)

    def GetAnlgEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgEdgeStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgEdgeStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgEdgeStartTrigDigSyncEnable(self):
        return DAQmxResetAnlgEdgeStartTrigDigSyncEnable(self.taskHandle)

    def GetAnlgWinStartTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinStartTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinStartTrigSrc(self, data):
        return DAQmxSetAnlgWinStartTrigSrc(self.taskHandle, data)

    def ResetAnlgWinStartTrigSrc(self):
        return DAQmxResetAnlgWinStartTrigSrc(self.taskHandle)

    def GetAnlgWinStartTrigWhen(self, data):
        return DAQmxGetAnlgWinStartTrigWhen(self.taskHandle, data)

    def SetAnlgWinStartTrigWhen(self, data):
        return DAQmxSetAnlgWinStartTrigWhen(self.taskHandle, data)

    def ResetAnlgWinStartTrigWhen(self):
        return DAQmxResetAnlgWinStartTrigWhen(self.taskHandle)

    def GetAnlgWinStartTrigTop(self, data):
        return DAQmxGetAnlgWinStartTrigTop(self.taskHandle, data)

    def SetAnlgWinStartTrigTop(self, data):
        return DAQmxSetAnlgWinStartTrigTop(self.taskHandle, data)

    def ResetAnlgWinStartTrigTop(self):
        return DAQmxResetAnlgWinStartTrigTop(self.taskHandle)

    def GetAnlgWinStartTrigBtm(self, data):
        return DAQmxGetAnlgWinStartTrigBtm(self.taskHandle, data)

    def SetAnlgWinStartTrigBtm(self, data):
        return DAQmxSetAnlgWinStartTrigBtm(self.taskHandle, data)

    def ResetAnlgWinStartTrigBtm(self):
        return DAQmxResetAnlgWinStartTrigBtm(self.taskHandle)

    def GetAnlgWinStartTrigCoupling(self, data):
        return DAQmxGetAnlgWinStartTrigCoupling(self.taskHandle, data)

    def SetAnlgWinStartTrigCoupling(self, data):
        return DAQmxSetAnlgWinStartTrigCoupling(self.taskHandle, data)

    def ResetAnlgWinStartTrigCoupling(self):
        return DAQmxResetAnlgWinStartTrigCoupling(self.taskHandle)

    def GetAnlgWinStartTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgWinStartTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgWinStartTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinStartTrigDigFltrEnable(self.taskHandle)

    def GetAnlgWinStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgWinStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgWinStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgWinStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgWinStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgWinStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgWinStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgWinStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgWinStartTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinStartTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgWinStartTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinStartTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgWinStartTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinStartTrigDigSyncEnable(self.taskHandle)

    def GetStartTrigDelay(self, data):
        return DAQmxGetStartTrigDelay(self.taskHandle, data)

    def SetStartTrigDelay(self, data):
        return DAQmxSetStartTrigDelay(self.taskHandle, data)

    def ResetStartTrigDelay(self):
        return DAQmxResetStartTrigDelay(self.taskHandle)

    def GetStartTrigDelayUnits(self, data):
        return DAQmxGetStartTrigDelayUnits(self.taskHandle, data)

    def SetStartTrigDelayUnits(self, data):
        return DAQmxSetStartTrigDelayUnits(self.taskHandle, data)

    def ResetStartTrigDelayUnits(self):
        return DAQmxResetStartTrigDelayUnits(self.taskHandle)

    def GetStartTrigRetriggerable(self, data):
        return DAQmxGetStartTrigRetriggerable(self.taskHandle, data)

    def SetStartTrigRetriggerable(self, data):
        return DAQmxSetStartTrigRetriggerable(self.taskHandle, data)

    def ResetStartTrigRetriggerable(self):
        return DAQmxResetStartTrigRetriggerable(self.taskHandle)

    def GetRefTrigType(self, data):
        return DAQmxGetRefTrigType(self.taskHandle, data)

    def SetRefTrigType(self, data):
        return DAQmxSetRefTrigType(self.taskHandle, data)

    def ResetRefTrigType(self):
        return DAQmxResetRefTrigType(self.taskHandle)

    def GetRefTrigPretrigSamples(self, data):
        return DAQmxGetRefTrigPretrigSamples(self.taskHandle, data)

    def SetRefTrigPretrigSamples(self, data):
        return DAQmxSetRefTrigPretrigSamples(self.taskHandle, data)

    def ResetRefTrigPretrigSamples(self):
        return DAQmxResetRefTrigPretrigSamples(self.taskHandle)

    def GetRefTrigTerm(self, data, bufferSize):
        return DAQmxGetRefTrigTerm(self.taskHandle, data, bufferSize)

    def GetDigEdgeRefTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeRefTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeRefTrigSrc(self, data):
        return DAQmxSetDigEdgeRefTrigSrc(self.taskHandle, data)

    def ResetDigEdgeRefTrigSrc(self):
        return DAQmxResetDigEdgeRefTrigSrc(self.taskHandle)

    def GetDigEdgeRefTrigEdge(self, data):
        return DAQmxGetDigEdgeRefTrigEdge(self.taskHandle, data)

    def SetDigEdgeRefTrigEdge(self, data):
        return DAQmxSetDigEdgeRefTrigEdge(self.taskHandle, data)

    def ResetDigEdgeRefTrigEdge(self):
        return DAQmxResetDigEdgeRefTrigEdge(self.taskHandle)

    def GetDigEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrEnable(self.taskHandle, data)

    def SetDigEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrEnable(self.taskHandle, data)

    def ResetDigEdgeRefTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeRefTrigDigFltrEnable(self.taskHandle)

    def GetDigEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetDigEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetDigEdgeRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetDigEdgeRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetDigEdgeRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetDigEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetDigEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetDigEdgeRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate(self.taskHandle)

    def GetDigEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeRefTrigDigSyncEnable(self.taskHandle, data)

    def SetDigEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeRefTrigDigSyncEnable(self.taskHandle, data)

    def ResetDigEdgeRefTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeRefTrigDigSyncEnable(self.taskHandle)

    def GetDigPatternRefTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternRefTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigPatternRefTrigSrc(self, data):
        return DAQmxSetDigPatternRefTrigSrc(self.taskHandle, data)

    def ResetDigPatternRefTrigSrc(self):
        return DAQmxResetDigPatternRefTrigSrc(self.taskHandle)

    def GetDigPatternRefTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternRefTrigPattern(self.taskHandle, data, bufferSize)

    def SetDigPatternRefTrigPattern(self, data):
        return DAQmxSetDigPatternRefTrigPattern(self.taskHandle, data)

    def ResetDigPatternRefTrigPattern(self):
        return DAQmxResetDigPatternRefTrigPattern(self.taskHandle)

    def GetDigPatternRefTrigWhen(self, data):
        return DAQmxGetDigPatternRefTrigWhen(self.taskHandle, data)

    def SetDigPatternRefTrigWhen(self, data):
        return DAQmxSetDigPatternRefTrigWhen(self.taskHandle, data)

    def ResetDigPatternRefTrigWhen(self):
        return DAQmxResetDigPatternRefTrigWhen(self.taskHandle)

    def GetAnlgEdgeRefTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeRefTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgEdgeRefTrigSrc(self, data):
        return DAQmxSetAnlgEdgeRefTrigSrc(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigSrc(self):
        return DAQmxResetAnlgEdgeRefTrigSrc(self.taskHandle)

    def GetAnlgEdgeRefTrigSlope(self, data):
        return DAQmxGetAnlgEdgeRefTrigSlope(self.taskHandle, data)

    def SetAnlgEdgeRefTrigSlope(self, data):
        return DAQmxSetAnlgEdgeRefTrigSlope(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigSlope(self):
        return DAQmxResetAnlgEdgeRefTrigSlope(self.taskHandle)

    def GetAnlgEdgeRefTrigLvl(self, data):
        return DAQmxGetAnlgEdgeRefTrigLvl(self.taskHandle, data)

    def SetAnlgEdgeRefTrigLvl(self, data):
        return DAQmxSetAnlgEdgeRefTrigLvl(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigLvl(self):
        return DAQmxResetAnlgEdgeRefTrigLvl(self.taskHandle)

    def GetAnlgEdgeRefTrigHyst(self, data):
        return DAQmxGetAnlgEdgeRefTrigHyst(self.taskHandle, data)

    def SetAnlgEdgeRefTrigHyst(self, data):
        return DAQmxSetAnlgEdgeRefTrigHyst(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigHyst(self):
        return DAQmxResetAnlgEdgeRefTrigHyst(self.taskHandle)

    def GetAnlgEdgeRefTrigCoupling(self, data):
        return DAQmxGetAnlgEdgeRefTrigCoupling(self.taskHandle, data)

    def SetAnlgEdgeRefTrigCoupling(self, data):
        return DAQmxSetAnlgEdgeRefTrigCoupling(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigCoupling(self):
        return DAQmxResetAnlgEdgeRefTrigCoupling(self.taskHandle)

    def GetAnlgEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigDigFltrEnable(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrEnable(self.taskHandle)

    def GetAnlgEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgEdgeRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgEdgeRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgEdgeRefTrigDigSyncEnable(self):
        return DAQmxResetAnlgEdgeRefTrigDigSyncEnable(self.taskHandle)

    def GetAnlgWinRefTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinRefTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinRefTrigSrc(self, data):
        return DAQmxSetAnlgWinRefTrigSrc(self.taskHandle, data)

    def ResetAnlgWinRefTrigSrc(self):
        return DAQmxResetAnlgWinRefTrigSrc(self.taskHandle)

    def GetAnlgWinRefTrigWhen(self, data):
        return DAQmxGetAnlgWinRefTrigWhen(self.taskHandle, data)

    def SetAnlgWinRefTrigWhen(self, data):
        return DAQmxSetAnlgWinRefTrigWhen(self.taskHandle, data)

    def ResetAnlgWinRefTrigWhen(self):
        return DAQmxResetAnlgWinRefTrigWhen(self.taskHandle)

    def GetAnlgWinRefTrigTop(self, data):
        return DAQmxGetAnlgWinRefTrigTop(self.taskHandle, data)

    def SetAnlgWinRefTrigTop(self, data):
        return DAQmxSetAnlgWinRefTrigTop(self.taskHandle, data)

    def ResetAnlgWinRefTrigTop(self):
        return DAQmxResetAnlgWinRefTrigTop(self.taskHandle)

    def GetAnlgWinRefTrigBtm(self, data):
        return DAQmxGetAnlgWinRefTrigBtm(self.taskHandle, data)

    def SetAnlgWinRefTrigBtm(self, data):
        return DAQmxSetAnlgWinRefTrigBtm(self.taskHandle, data)

    def ResetAnlgWinRefTrigBtm(self):
        return DAQmxResetAnlgWinRefTrigBtm(self.taskHandle)

    def GetAnlgWinRefTrigCoupling(self, data):
        return DAQmxGetAnlgWinRefTrigCoupling(self.taskHandle, data)

    def SetAnlgWinRefTrigCoupling(self, data):
        return DAQmxSetAnlgWinRefTrigCoupling(self.taskHandle, data)

    def ResetAnlgWinRefTrigCoupling(self):
        return DAQmxResetAnlgWinRefTrigCoupling(self.taskHandle)

    def GetAnlgWinRefTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgWinRefTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgWinRefTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinRefTrigDigFltrEnable(self.taskHandle)

    def GetAnlgWinRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgWinRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgWinRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgWinRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgWinRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgWinRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgWinRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgWinRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgWinRefTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinRefTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgWinRefTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinRefTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgWinRefTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinRefTrigDigSyncEnable(self.taskHandle)

    def GetRefTrigAutoTrigEnable(self, data):
        return DAQmxGetRefTrigAutoTrigEnable(self.taskHandle, data)

    def SetRefTrigAutoTrigEnable(self, data):
        return DAQmxSetRefTrigAutoTrigEnable(self.taskHandle, data)

    def ResetRefTrigAutoTrigEnable(self):
        return DAQmxResetRefTrigAutoTrigEnable(self.taskHandle)

    def GetRefTrigAutoTriggered(self, data):
        return DAQmxGetRefTrigAutoTriggered(self.taskHandle, data)

    def GetRefTrigDelay(self, data):
        return DAQmxGetRefTrigDelay(self.taskHandle, data)

    def SetRefTrigDelay(self, data):
        return DAQmxSetRefTrigDelay(self.taskHandle, data)

    def ResetRefTrigDelay(self):
        return DAQmxResetRefTrigDelay(self.taskHandle)

    def GetAdvTrigType(self, data):
        return DAQmxGetAdvTrigType(self.taskHandle, data)

    def SetAdvTrigType(self, data):
        return DAQmxSetAdvTrigType(self.taskHandle, data)

    def ResetAdvTrigType(self):
        return DAQmxResetAdvTrigType(self.taskHandle)

    def GetDigEdgeAdvTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeAdvTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeAdvTrigSrc(self, data):
        return DAQmxSetDigEdgeAdvTrigSrc(self.taskHandle, data)

    def ResetDigEdgeAdvTrigSrc(self):
        return DAQmxResetDigEdgeAdvTrigSrc(self.taskHandle)

    def GetDigEdgeAdvTrigEdge(self, data):
        return DAQmxGetDigEdgeAdvTrigEdge(self.taskHandle, data)

    def SetDigEdgeAdvTrigEdge(self, data):
        return DAQmxSetDigEdgeAdvTrigEdge(self.taskHandle, data)

    def ResetDigEdgeAdvTrigEdge(self):
        return DAQmxResetDigEdgeAdvTrigEdge(self.taskHandle)

    def GetDigEdgeAdvTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeAdvTrigDigFltrEnable(self.taskHandle, data)

    def SetDigEdgeAdvTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeAdvTrigDigFltrEnable(self.taskHandle, data)

    def ResetDigEdgeAdvTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeAdvTrigDigFltrEnable(self.taskHandle)

    def GetHshkTrigType(self, data):
        return DAQmxGetHshkTrigType(self.taskHandle, data)

    def SetHshkTrigType(self, data):
        return DAQmxSetHshkTrigType(self.taskHandle, data)

    def ResetHshkTrigType(self):
        return DAQmxResetHshkTrigType(self.taskHandle)

    def GetInterlockedHshkTrigSrc(self, data, bufferSize):
        return DAQmxGetInterlockedHshkTrigSrc(self.taskHandle, data, bufferSize)

    def SetInterlockedHshkTrigSrc(self, data):
        return DAQmxSetInterlockedHshkTrigSrc(self.taskHandle, data)

    def ResetInterlockedHshkTrigSrc(self):
        return DAQmxResetInterlockedHshkTrigSrc(self.taskHandle)

    def GetInterlockedHshkTrigAssertedLvl(self, data):
        return DAQmxGetInterlockedHshkTrigAssertedLvl(self.taskHandle, data)

    def SetInterlockedHshkTrigAssertedLvl(self, data):
        return DAQmxSetInterlockedHshkTrigAssertedLvl(self.taskHandle, data)

    def ResetInterlockedHshkTrigAssertedLvl(self):
        return DAQmxResetInterlockedHshkTrigAssertedLvl(self.taskHandle)

    def GetPauseTrigType(self, data):
        return DAQmxGetPauseTrigType(self.taskHandle, data)

    def SetPauseTrigType(self, data):
        return DAQmxSetPauseTrigType(self.taskHandle, data)

    def ResetPauseTrigType(self):
        return DAQmxResetPauseTrigType(self.taskHandle)

    def GetPauseTrigTerm(self, data, bufferSize):
        return DAQmxGetPauseTrigTerm(self.taskHandle, data, bufferSize)

    def GetAnlgLvlPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgLvlPauseTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgLvlPauseTrigSrc(self, data):
        return DAQmxSetAnlgLvlPauseTrigSrc(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigSrc(self):
        return DAQmxResetAnlgLvlPauseTrigSrc(self.taskHandle)

    def GetAnlgLvlPauseTrigWhen(self, data):
        return DAQmxGetAnlgLvlPauseTrigWhen(self.taskHandle, data)

    def SetAnlgLvlPauseTrigWhen(self, data):
        return DAQmxSetAnlgLvlPauseTrigWhen(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigWhen(self):
        return DAQmxResetAnlgLvlPauseTrigWhen(self.taskHandle)

    def GetAnlgLvlPauseTrigLvl(self, data):
        return DAQmxGetAnlgLvlPauseTrigLvl(self.taskHandle, data)

    def SetAnlgLvlPauseTrigLvl(self, data):
        return DAQmxSetAnlgLvlPauseTrigLvl(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigLvl(self):
        return DAQmxResetAnlgLvlPauseTrigLvl(self.taskHandle)

    def GetAnlgLvlPauseTrigHyst(self, data):
        return DAQmxGetAnlgLvlPauseTrigHyst(self.taskHandle, data)

    def SetAnlgLvlPauseTrigHyst(self, data):
        return DAQmxSetAnlgLvlPauseTrigHyst(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigHyst(self):
        return DAQmxResetAnlgLvlPauseTrigHyst(self.taskHandle)

    def GetAnlgLvlPauseTrigCoupling(self, data):
        return DAQmxGetAnlgLvlPauseTrigCoupling(self.taskHandle, data)

    def SetAnlgLvlPauseTrigCoupling(self, data):
        return DAQmxSetAnlgLvlPauseTrigCoupling(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigCoupling(self):
        return DAQmxResetAnlgLvlPauseTrigCoupling(self.taskHandle)

    def GetAnlgLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigDigFltrEnable(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrEnable(self.taskHandle)

    def GetAnlgLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgLvlPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgLvlPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgLvlPauseTrigDigSyncEnable(self):
        return DAQmxResetAnlgLvlPauseTrigDigSyncEnable(self.taskHandle)

    def GetAnlgWinPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinPauseTrigSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinPauseTrigSrc(self, data):
        return DAQmxSetAnlgWinPauseTrigSrc(self.taskHandle, data)

    def ResetAnlgWinPauseTrigSrc(self):
        return DAQmxResetAnlgWinPauseTrigSrc(self.taskHandle)

    def GetAnlgWinPauseTrigWhen(self, data):
        return DAQmxGetAnlgWinPauseTrigWhen(self.taskHandle, data)

    def SetAnlgWinPauseTrigWhen(self, data):
        return DAQmxSetAnlgWinPauseTrigWhen(self.taskHandle, data)

    def ResetAnlgWinPauseTrigWhen(self):
        return DAQmxResetAnlgWinPauseTrigWhen(self.taskHandle)

    def GetAnlgWinPauseTrigTop(self, data):
        return DAQmxGetAnlgWinPauseTrigTop(self.taskHandle, data)

    def SetAnlgWinPauseTrigTop(self, data):
        return DAQmxSetAnlgWinPauseTrigTop(self.taskHandle, data)

    def ResetAnlgWinPauseTrigTop(self):
        return DAQmxResetAnlgWinPauseTrigTop(self.taskHandle)

    def GetAnlgWinPauseTrigBtm(self, data):
        return DAQmxGetAnlgWinPauseTrigBtm(self.taskHandle, data)

    def SetAnlgWinPauseTrigBtm(self, data):
        return DAQmxSetAnlgWinPauseTrigBtm(self.taskHandle, data)

    def ResetAnlgWinPauseTrigBtm(self):
        return DAQmxResetAnlgWinPauseTrigBtm(self.taskHandle)

    def GetAnlgWinPauseTrigCoupling(self, data):
        return DAQmxGetAnlgWinPauseTrigCoupling(self.taskHandle, data)

    def SetAnlgWinPauseTrigCoupling(self, data):
        return DAQmxSetAnlgWinPauseTrigCoupling(self.taskHandle, data)

    def ResetAnlgWinPauseTrigCoupling(self):
        return DAQmxResetAnlgWinPauseTrigCoupling(self.taskHandle)

    def GetAnlgWinPauseTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrEnable(self.taskHandle, data)

    def SetAnlgWinPauseTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrEnable(self.taskHandle, data)

    def ResetAnlgWinPauseTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrEnable(self.taskHandle)

    def GetAnlgWinPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetAnlgWinPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetAnlgWinPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetAnlgWinPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetAnlgWinPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetAnlgWinPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetAnlgWinPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetAnlgWinPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetAnlgWinPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate(self.taskHandle)

    def GetAnlgWinPauseTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinPauseTrigDigSyncEnable(self.taskHandle, data)

    def SetAnlgWinPauseTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinPauseTrigDigSyncEnable(self.taskHandle, data)

    def ResetAnlgWinPauseTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinPauseTrigDigSyncEnable(self.taskHandle)

    def GetDigLvlPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetDigLvlPauseTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigLvlPauseTrigSrc(self, data):
        return DAQmxSetDigLvlPauseTrigSrc(self.taskHandle, data)

    def ResetDigLvlPauseTrigSrc(self):
        return DAQmxResetDigLvlPauseTrigSrc(self.taskHandle)

    def GetDigLvlPauseTrigWhen(self, data):
        return DAQmxGetDigLvlPauseTrigWhen(self.taskHandle, data)

    def SetDigLvlPauseTrigWhen(self, data):
        return DAQmxSetDigLvlPauseTrigWhen(self.taskHandle, data)

    def ResetDigLvlPauseTrigWhen(self):
        return DAQmxResetDigLvlPauseTrigWhen(self.taskHandle)

    def GetDigLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrEnable(self.taskHandle, data)

    def SetDigLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrEnable(self.taskHandle, data)

    def ResetDigLvlPauseTrigDigFltrEnable(self):
        return DAQmxResetDigLvlPauseTrigDigFltrEnable(self.taskHandle)

    def GetDigLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetDigLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetDigLvlPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetDigLvlPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetDigLvlPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetDigLvlPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetDigLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetDigLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetDigLvlPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate(self.taskHandle)

    def GetDigLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxGetDigLvlPauseTrigDigSyncEnable(self.taskHandle, data)

    def SetDigLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxSetDigLvlPauseTrigDigSyncEnable(self.taskHandle, data)

    def ResetDigLvlPauseTrigDigSyncEnable(self):
        return DAQmxResetDigLvlPauseTrigDigSyncEnable(self.taskHandle)

    def GetDigPatternPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternPauseTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigPatternPauseTrigSrc(self, data):
        return DAQmxSetDigPatternPauseTrigSrc(self.taskHandle, data)

    def ResetDigPatternPauseTrigSrc(self):
        return DAQmxResetDigPatternPauseTrigSrc(self.taskHandle)

    def GetDigPatternPauseTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternPauseTrigPattern(self.taskHandle, data, bufferSize)

    def SetDigPatternPauseTrigPattern(self, data):
        return DAQmxSetDigPatternPauseTrigPattern(self.taskHandle, data)

    def ResetDigPatternPauseTrigPattern(self):
        return DAQmxResetDigPatternPauseTrigPattern(self.taskHandle)

    def GetDigPatternPauseTrigWhen(self, data):
        return DAQmxGetDigPatternPauseTrigWhen(self.taskHandle, data)

    def SetDigPatternPauseTrigWhen(self, data):
        return DAQmxSetDigPatternPauseTrigWhen(self.taskHandle, data)

    def ResetDigPatternPauseTrigWhen(self):
        return DAQmxResetDigPatternPauseTrigWhen(self.taskHandle)

    def GetArmStartTrigType(self, data):
        return DAQmxGetArmStartTrigType(self.taskHandle, data)

    def SetArmStartTrigType(self, data):
        return DAQmxSetArmStartTrigType(self.taskHandle, data)

    def ResetArmStartTrigType(self):
        return DAQmxResetArmStartTrigType(self.taskHandle)

    def GetArmStartTerm(self, data, bufferSize):
        return DAQmxGetArmStartTerm(self.taskHandle, data, bufferSize)

    def GetDigEdgeArmStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeArmStartTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeArmStartTrigSrc(self, data):
        return DAQmxSetDigEdgeArmStartTrigSrc(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigSrc(self):
        return DAQmxResetDigEdgeArmStartTrigSrc(self.taskHandle)

    def GetDigEdgeArmStartTrigEdge(self, data):
        return DAQmxGetDigEdgeArmStartTrigEdge(self.taskHandle, data)

    def SetDigEdgeArmStartTrigEdge(self, data):
        return DAQmxSetDigEdgeArmStartTrigEdge(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigEdge(self):
        return DAQmxResetDigEdgeArmStartTrigEdge(self.taskHandle)

    def GetDigEdgeArmStartTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrEnable(self.taskHandle, data)

    def SetDigEdgeArmStartTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrEnable(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrEnable(self.taskHandle)

    def GetDigEdgeArmStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def SetDigEdgeArmStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.taskHandle)

    def GetDigEdgeArmStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeArmStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.taskHandle)

    def GetDigEdgeArmStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def SetDigEdgeArmStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate(self.taskHandle)

    def GetDigEdgeArmStartTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigSyncEnable(self.taskHandle, data)

    def SetDigEdgeArmStartTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigSyncEnable(self.taskHandle, data)

    def ResetDigEdgeArmStartTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeArmStartTrigDigSyncEnable(self.taskHandle)

    def GetTriggerSyncType(self, data):
        return DAQmxGetTriggerSyncType(self.taskHandle, data)

    def SetTriggerSyncType(self, data):
        return DAQmxSetTriggerSyncType(self.taskHandle, data)

    def ResetTriggerSyncType(self):
        return DAQmxResetTriggerSyncType(self.taskHandle)

    def GetWatchdogTimeout(self, data):
        return DAQmxGetWatchdogTimeout(self.taskHandle, data)

    def SetWatchdogTimeout(self, data):
        return DAQmxSetWatchdogTimeout(self.taskHandle, data)

    def ResetWatchdogTimeout(self):
        return DAQmxResetWatchdogTimeout(self.taskHandle)

    def GetWatchdogExpirTrigType(self, data):
        return DAQmxGetWatchdogExpirTrigType(self.taskHandle, data)

    def SetWatchdogExpirTrigType(self, data):
        return DAQmxSetWatchdogExpirTrigType(self.taskHandle, data)

    def ResetWatchdogExpirTrigType(self):
        return DAQmxResetWatchdogExpirTrigType(self.taskHandle)

    def GetWatchdogExpirTrigTrigOnNetworkConnLoss(self, data):
        return DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss(self.taskHandle, data)

    def SetWatchdogExpirTrigTrigOnNetworkConnLoss(self, data):
        return DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss(self.taskHandle, data)

    def ResetWatchdogExpirTrigTrigOnNetworkConnLoss(self):
        return DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss(self.taskHandle)

    def GetDigEdgeWatchdogExpirTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeWatchdogExpirTrigSrc(self.taskHandle, data, bufferSize)

    def SetDigEdgeWatchdogExpirTrigSrc(self, data):
        return DAQmxSetDigEdgeWatchdogExpirTrigSrc(self.taskHandle, data)

    def ResetDigEdgeWatchdogExpirTrigSrc(self):
        return DAQmxResetDigEdgeWatchdogExpirTrigSrc(self.taskHandle)

    def GetDigEdgeWatchdogExpirTrigEdge(self, data):
        return DAQmxGetDigEdgeWatchdogExpirTrigEdge(self.taskHandle, data)

    def SetDigEdgeWatchdogExpirTrigEdge(self, data):
        return DAQmxSetDigEdgeWatchdogExpirTrigEdge(self.taskHandle, data)

    def ResetDigEdgeWatchdogExpirTrigEdge(self):
        return DAQmxResetDigEdgeWatchdogExpirTrigEdge(self.taskHandle)

    def GetWatchdogDOExpirState(self, lines, data):
        return DAQmxGetWatchdogDOExpirState(self.taskHandle, lines, data)

    def SetWatchdogDOExpirState(self, lines, data):
        return DAQmxSetWatchdogDOExpirState(self.taskHandle, lines, data)

    def ResetWatchdogDOExpirState(self, lines):
        return DAQmxResetWatchdogDOExpirState(self.taskHandle, lines)

    def GetWatchdogAOOutputType(self, lines, data):
        return DAQmxGetWatchdogAOOutputType(self.taskHandle, lines, data)

    def SetWatchdogAOOutputType(self, lines, data):
        return DAQmxSetWatchdogAOOutputType(self.taskHandle, lines, data)

    def ResetWatchdogAOOutputType(self, lines):
        return DAQmxResetWatchdogAOOutputType(self.taskHandle, lines)

    def GetWatchdogAOExpirState(self, lines, data):
        return DAQmxGetWatchdogAOExpirState(self.taskHandle, lines, data)

    def SetWatchdogAOExpirState(self, lines, data):
        return DAQmxSetWatchdogAOExpirState(self.taskHandle, lines, data)

    def ResetWatchdogAOExpirState(self, lines):
        return DAQmxResetWatchdogAOExpirState(self.taskHandle, lines)

    def GetWatchdogCOExpirState(self, lines, data):
        return DAQmxGetWatchdogCOExpirState(self.taskHandle, lines, data)

    def SetWatchdogCOExpirState(self, lines, data):
        return DAQmxSetWatchdogCOExpirState(self.taskHandle, lines, data)

    def ResetWatchdogCOExpirState(self, lines):
        return DAQmxResetWatchdogCOExpirState(self.taskHandle, lines)

    def GetWatchdogHasExpired(self, data):
        return DAQmxGetWatchdogHasExpired(self.taskHandle, data)

    def GetWriteRelativeTo(self, data):
        return DAQmxGetWriteRelativeTo(self.taskHandle, data)

    def SetWriteRelativeTo(self, data):
        return DAQmxSetWriteRelativeTo(self.taskHandle, data)

    def ResetWriteRelativeTo(self):
        return DAQmxResetWriteRelativeTo(self.taskHandle)

    def GetWriteOffset(self, data):
        return DAQmxGetWriteOffset(self.taskHandle, data)

    def SetWriteOffset(self, data):
        return DAQmxSetWriteOffset(self.taskHandle, data)

    def ResetWriteOffset(self):
        return DAQmxResetWriteOffset(self.taskHandle)

    def GetWriteRegenMode(self, data):
        return DAQmxGetWriteRegenMode(self.taskHandle, data)

    def SetWriteRegenMode(self, data):
        return DAQmxSetWriteRegenMode(self.taskHandle, data)

    def ResetWriteRegenMode(self):
        return DAQmxResetWriteRegenMode(self.taskHandle)

    def GetWriteCurrWritePos(self, data):
        return DAQmxGetWriteCurrWritePos(self.taskHandle, data)

    def GetWriteOvercurrentChansExist(self, data):
        return DAQmxGetWriteOvercurrentChansExist(self.taskHandle, data)

    def GetWriteOvercurrentChans(self, data, bufferSize):
        return DAQmxGetWriteOvercurrentChans(self.taskHandle, data, bufferSize)

    def GetWriteOvertemperatureChansExist(self, data):
        return DAQmxGetWriteOvertemperatureChansExist(self.taskHandle, data)

    def GetWriteOpenCurrentLoopChansExist(self, data):
        return DAQmxGetWriteOpenCurrentLoopChansExist(self.taskHandle, data)

    def GetWriteOpenCurrentLoopChans(self, data, bufferSize):
        return DAQmxGetWriteOpenCurrentLoopChans(self.taskHandle, data, bufferSize)

    def GetWritePowerSupplyFaultChansExist(self, data):
        return DAQmxGetWritePowerSupplyFaultChansExist(self.taskHandle, data)

    def GetWritePowerSupplyFaultChans(self, data, bufferSize):
        return DAQmxGetWritePowerSupplyFaultChans(self.taskHandle, data, bufferSize)

    def GetWriteSpaceAvail(self, data):
        return DAQmxGetWriteSpaceAvail(self.taskHandle, data)

    def GetWriteTotalSampPerChanGenerated(self, data):
        return DAQmxGetWriteTotalSampPerChanGenerated(self.taskHandle, data)

    def GetWriteRawDataWidth(self, data):
        return DAQmxGetWriteRawDataWidth(self.taskHandle, data)

    def GetWriteNumChans(self, data):
        return DAQmxGetWriteNumChans(self.taskHandle, data)

    def GetWriteWaitMode(self, data):
        return DAQmxGetWriteWaitMode(self.taskHandle, data)

    def SetWriteWaitMode(self, data):
        return DAQmxSetWriteWaitMode(self.taskHandle, data)

    def ResetWriteWaitMode(self):
        return DAQmxResetWriteWaitMode(self.taskHandle)

    def GetWriteSleepTime(self, data):
        return DAQmxGetWriteSleepTime(self.taskHandle, data)

    def SetWriteSleepTime(self, data):
        return DAQmxSetWriteSleepTime(self.taskHandle, data)

    def ResetWriteSleepTime(self):
        return DAQmxResetWriteSleepTime(self.taskHandle)

    def GetWriteNextWriteIsLast(self, data):
        return DAQmxGetWriteNextWriteIsLast(self.taskHandle, data)

    def SetWriteNextWriteIsLast(self, data):
        return DAQmxSetWriteNextWriteIsLast(self.taskHandle, data)

    def ResetWriteNextWriteIsLast(self):
        return DAQmxResetWriteNextWriteIsLast(self.taskHandle)

    def GetWriteDigitalLinesBytesPerChan(self, data):
        return DAQmxGetWriteDigitalLinesBytesPerChan(self.taskHandle, data)

    def GetWriteAccessoryInsertionOrRemovalDetected(self, data):
        return DAQmxGetWriteAccessoryInsertionOrRemovalDetected(self.taskHandle, data)

    def GetWriteDevsWithInsertedOrRemovedAccessories(self, data, bufferSize):
        return DAQmxGetWriteDevsWithInsertedOrRemovedAccessories(self.taskHandle, data, bufferSize)

    def GetSampClkTimingResponseMode(self, data):
        return DAQmxGetSampClkTimingResponseMode(self.taskHandle, data)

    def SetSampClkTimingResponseMode(self, data):
        return DAQmxSetSampClkTimingResponseMode(self.taskHandle, data)

    def ResetSampClkTimingResponseMode(self):
        return DAQmxResetSampClkTimingResponseMode(self.taskHandle)

    stop = StopTask
    start = StartTask

    def __del__(self):
        try:
            self.ClearTask()
        except DAQError:
            pass
